/* ----------------------------------------------------------------------------
 * This file was automatically generated by SWIG (https://www.swig.org).
 * Version 4.4.0
 *
 * Do not make changes to this file unless you know what you are doing - modify
 * the SWIG interface file instead.
 * ----------------------------------------------------------------------------- */

#define SWIG_PYTHON_STRICT_BYTE_CHAR
#if defined(_MSC_VER)
/* Prevent "non-constant aggregate initializer" errors from Python.h in
 * Python 3.9 */
# pragma warning(default : 4204)
#endif



#define SWIG_VERSION 0x040400
#define SWIGPYTHON
#define SWIG_PYTHON_DIRECTOR_NO_VTABLE

#define SWIG_name    "_ra"
/* -----------------------------------------------------------------------------
 *  This section contains generic SWIG labels for method/variable
 *  declarations/attributes, and other compiler dependent labels.
 * ----------------------------------------------------------------------------- */

/* template workaround for compilers that cannot correctly implement the C++ standard */
#ifndef SWIGTEMPLATEDISAMBIGUATOR
# if defined(__SUNPRO_CC) && (__SUNPRO_CC <= 0x560)
#  define SWIGTEMPLATEDISAMBIGUATOR template
# elif defined(__HP_aCC)
/* Needed even with `aCC -AA' when `aCC -V' reports HP ANSI C++ B3910B A.03.55 */
/* If we find a maximum version that requires this, the test would be __HP_aCC <= 35500 for A.03.55 */
#  define SWIGTEMPLATEDISAMBIGUATOR template
# else
#  define SWIGTEMPLATEDISAMBIGUATOR
# endif
#endif

/* inline attribute */
#ifndef SWIGINLINE
# if defined(__cplusplus) || (defined(__GNUC__) && !defined(__STRICT_ANSI__))
#   define SWIGINLINE inline
# else
#   define SWIGINLINE
# endif
#endif

/* attribute recognised by some compilers to avoid 'unused' warnings */
#ifndef SWIGUNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define SWIGUNUSED __attribute__ ((__unused__))
#   else
#     define SWIGUNUSED
#   endif
# elif defined(__ICC)
#   define SWIGUNUSED __attribute__ ((__unused__))
# else
#   define SWIGUNUSED
# endif
#endif

#ifndef SWIG_MSC_UNSUPPRESS_4505
# if defined(_MSC_VER)
#   pragma warning(disable : 4505) /* unreferenced local function has been removed */
# endif
#endif

#ifndef SWIGUNUSEDPARM
# ifdef __cplusplus
#   define SWIGUNUSEDPARM(p)
# else
#   define SWIGUNUSEDPARM(p) p SWIGUNUSED
# endif
#endif

/* internal SWIG method */
#ifndef SWIGINTERN
# define SWIGINTERN static SWIGUNUSED
#endif

/* internal inline SWIG method */
#ifndef SWIGINTERNINLINE
# define SWIGINTERNINLINE SWIGINTERN SWIGINLINE
#endif

/* exporting methods */
#if defined(__GNUC__)
#  if (__GNUC__ >= 4) || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4)
#    ifndef GCC_HASCLASSVISIBILITY
#      define GCC_HASCLASSVISIBILITY
#    endif
#  endif
#endif

#ifndef SWIGEXPORT
# if defined(_WIN32) || defined(__WIN32__) || defined(__CYGWIN__)
#   if defined(STATIC_LINKED)
#     define SWIGEXPORT
#   else
#     define SWIGEXPORT __declspec(dllexport)
#   endif
# else
#   if defined(__GNUC__) && defined(GCC_HASCLASSVISIBILITY)
#     define SWIGEXPORT __attribute__ ((visibility("default")))
#   else
#     define SWIGEXPORT
#   endif
# endif
#endif

/* calling conventions for Windows */
#ifndef SWIGSTDCALL
# if defined(_WIN32) || defined(__WIN32__) || defined(__CYGWIN__)
#   define SWIGSTDCALL __stdcall
# else
#   define SWIGSTDCALL
# endif
#endif

/* Deal with Microsoft's attempt at deprecating C standard runtime functions */
#if !defined(SWIG_NO_CRT_SECURE_NO_DEPRECATE) && defined(_MSC_VER) && !defined(_CRT_SECURE_NO_DEPRECATE)
# define _CRT_SECURE_NO_DEPRECATE
#endif

/* Deal with Microsoft's attempt at deprecating methods in the standard C++ library */
#if !defined(SWIG_NO_SCL_SECURE_NO_DEPRECATE) && defined(_MSC_VER) && !defined(_SCL_SECURE_NO_DEPRECATE)
# define _SCL_SECURE_NO_DEPRECATE
#endif

/* Deal with Apple's deprecated 'AssertMacros.h' from Carbon-framework */
#if defined(__APPLE__) && !defined(__ASSERT_MACROS_DEFINE_VERSIONS_WITHOUT_UNDERSCORES)
# define __ASSERT_MACROS_DEFINE_VERSIONS_WITHOUT_UNDERSCORES 0
#endif

/* Intel's compiler complains if a variable which was never initialised is
 * cast to void, which is a common idiom which we use to indicate that we
 * are aware a variable isn't used.  So we just silence that warning.
 * See: https://github.com/swig/swig/issues/192 for more discussion.
 */
#ifdef __INTEL_COMPILER
# pragma warning disable 592
#endif

#if defined(__cplusplus) && __cplusplus >=201103L
# define SWIG_NOEXCEPT noexcept
#else
# define SWIG_NOEXCEPT throw()
#endif 

/* -----------------------------------------------------------------------------
 * swigcompat.swg
 *
 * Macros to provide support compatibility with older C and C++ standards.
 *
 * Note that SWIG expects __cplusplus to be defined to the appropriate C++ standard.
 * MSVC users are urged to check and examine the /Zc:__cplusplus compiler option.
 * See https://learn.microsoft.com/en-us/cpp/build/reference/zc-cplusplus.
 * ----------------------------------------------------------------------------- */

/* C99 and C++11 should provide snprintf, but define SWIG_NO_SNPRINTF
 * if you're missing it.
 */
#if ((defined __STDC_VERSION__ && __STDC_VERSION__ >= 199901L) || \
     (defined __cplusplus && __cplusplus >= 201103L) || \
     defined SWIG_HAVE_SNPRINTF) && \
    !defined SWIG_NO_SNPRINTF
# define SWIG_snprintf(O,S,F,A) snprintf(O,S,F,A)
# define SWIG_snprintf2(O,S,F,A,B) snprintf(O,S,F,A,B)
#else
/* Fallback versions ignore the buffer size, but most of our uses either have a
 * fixed maximum possible size or dynamically allocate a buffer that's large
 * enough.
 */
# define SWIG_snprintf(O,S,F,A) sprintf(O,F,A)
# define SWIG_snprintf2(O,S,F,A,B) sprintf(O,F,A,B)
#endif


#if defined(__GNUC__) && defined(_WIN32) && !defined(SWIG_PYTHON_NO_HYPOT_WORKAROUND)
/* Workaround for '::hypot' has not been declared', see https://bugs.python.org/issue11566 */
# include <math.h>
#endif

#if !defined(PY_SSIZE_T_CLEAN) && !defined(SWIG_NO_PY_SSIZE_T_CLEAN)
#define PY_SSIZE_T_CLEAN
#endif

#if __GNUC__ >= 7
#pragma GCC diagnostic push
#if defined(__cplusplus) && __cplusplus >=201703L
#pragma GCC diagnostic ignored "-Wregister" /* For python-2.7 headers that use register */
#endif
#endif

#if defined(_DEBUG) && defined(SWIG_PYTHON_INTERPRETER_NO_DEBUG)
/* Use debug wrappers with the Python release dll */

#if defined(_MSC_VER) && _MSC_VER >= 1929
/* Workaround compilation errors when redefining _DEBUG in MSVC 2019 version 16.10 and later
 * See https://github.com/swig/swig/issues/2090 */
# include <corecrt.h>
#endif

# undef _DEBUG
# include <Python.h>
# define _DEBUG 1
#else
# include <Python.h>
#endif


#if __GNUC__ >= 7
#pragma GCC diagnostic pop
#endif

#include <stdio.h>
#include <stdlib.h>

/* -----------------------------------------------------------------------------
 * swigrun.swg
 *
 * This file contains generic C API SWIG runtime support for pointer
 * type checking.
 * ----------------------------------------------------------------------------- */

/* This should only be incremented when either the layout of swig_type_info changes,
   or for whatever reason, the runtime changes incompatibly */
#define SWIG_RUNTIME_VERSION "5"

/* define SWIG_TYPE_TABLE_NAME as "SWIG_TYPE_TABLE" */
#ifdef SWIG_TYPE_TABLE
# define SWIG_QUOTE_STRING(x) #x
# define SWIG_EXPAND_AND_QUOTE_STRING(x) SWIG_QUOTE_STRING(x)
# define SWIG_TYPE_TABLE_NAME SWIG_EXPAND_AND_QUOTE_STRING(SWIG_TYPE_TABLE)
#else
# define SWIG_TYPE_TABLE_NAME
#endif

/*
  You can use the SWIGRUNTIME and SWIGRUNTIMEINLINE macros for
  creating a static or dynamic library from the SWIG runtime code.
  In 99.9% of the cases, SWIG just needs to declare them as 'static'.

  But only do this if strictly necessary, ie, if you have problems
  with your compiler or suchlike.
*/

#ifndef SWIGRUNTIME
# define SWIGRUNTIME SWIGINTERN
#endif

#ifndef SWIGRUNTIMEINLINE
# define SWIGRUNTIMEINLINE SWIGRUNTIME SWIGINLINE
#endif

/*  Generic buffer size */
#ifndef SWIG_BUFFER_SIZE
# define SWIG_BUFFER_SIZE 1024
#endif

/* Flags for pointer conversions */
#define SWIG_POINTER_DISOWN        0x1
#define SWIG_CAST_NEW_MEMORY       0x2
#define SWIG_POINTER_NO_NULL       0x4
#define SWIG_POINTER_CLEAR         0x8
#define SWIG_POINTER_RELEASE       (SWIG_POINTER_CLEAR | SWIG_POINTER_DISOWN)

/* Flags for new pointer objects */
#define SWIG_POINTER_OWN           0x1


/*
   Flags/methods for returning states.

   The SWIG conversion methods, as ConvertPtr, return an integer
   that tells if the conversion was successful or not. And if not,
   an error code can be returned (see swigerrors.swg for the codes).

   Use the following macros/flags to set or process the returning
   states.

   In old versions of SWIG, code such as the following was usually written:

     if (SWIG_ConvertPtr(obj,vptr,ty.flags) != -1) {
       // success code
     } else {
       //fail code
     }

   Now you can be more explicit:

    int res = SWIG_ConvertPtr(obj,vptr,ty.flags);
    if (SWIG_IsOK(res)) {
      // success code
    } else {
      // fail code
    }

   which is the same really, but now you can also do

    Type *ptr;
    int res = SWIG_ConvertPtr(obj,(void **)(&ptr),ty.flags);
    if (SWIG_IsOK(res)) {
      // success code
      if (SWIG_IsNewObj(res) {
        ...
	delete *ptr;
      } else {
        ...
      }
    } else {
      // fail code
    }

   I.e., now SWIG_ConvertPtr can return new objects and you can
   identify the case and take care of the deallocation. Of course that
   also requires SWIG_ConvertPtr to return new result values, such as

      int SWIG_ConvertPtr(obj, ptr,...) {
        if (<obj is ok>) {
          if (<need new object>) {
            *ptr = <ptr to new allocated object>;
            return SWIG_NEWOBJ;
          } else {
            *ptr = <ptr to old object>;
            return SWIG_OLDOBJ;
          }
        } else {
          return SWIG_BADOBJ;
        }
      }

   Of course, returning the plain '0(success)/-1(fail)' still works, but you can be
   more explicit by returning SWIG_BADOBJ, SWIG_ERROR or any of the
   SWIG errors code.

   Finally, if the SWIG_CASTRANK_MODE is enabled, the result code
   allows returning the 'cast rank', for example, if you have this

       int food(double)
       int fooi(int);

   and you call

      food(1)   // cast rank '1'  (1 -> 1.0)
      fooi(1)   // cast rank '0'

   just use the SWIG_AddCast()/SWIG_CheckState()
*/

#define SWIG_OK                    (0)
/* Runtime errors are < 0 */
#define SWIG_ERROR                 (-1)
/* Errors in range -1 to -99 are in swigerrors.swg (errors for all languages including those not using the runtime) */
/* Errors in range -100 to -199 are language specific errors defined in *errors.swg */
/* Errors < -200 are generic runtime specific errors */
#define SWIG_ERROR_RELEASE_NOT_OWNED (-200)

#define SWIG_IsOK(r)               (r >= 0)
#define SWIG_ArgError(r)           ((r != SWIG_ERROR) ? r : SWIG_TypeError)

/* The CastRankLimit says how many bits are used for the cast rank */
#define SWIG_CASTRANKLIMIT         (1 << 8)
/* The NewMask denotes the object was created (using new/malloc) */
#define SWIG_NEWOBJMASK            (SWIG_CASTRANKLIMIT  << 1)
/* The TmpMask is for in/out typemaps that use temporary objects */
#define SWIG_TMPOBJMASK            (SWIG_NEWOBJMASK << 1)
/* Simple returning values */
#define SWIG_BADOBJ                (SWIG_ERROR)
#define SWIG_OLDOBJ                (SWIG_OK)
#define SWIG_NEWOBJ                (SWIG_OK | SWIG_NEWOBJMASK)
#define SWIG_TMPOBJ                (SWIG_OK | SWIG_TMPOBJMASK)
/* Check, add and del object mask methods */
#define SWIG_AddNewMask(r)         (SWIG_IsOK(r) ? (r | SWIG_NEWOBJMASK) : r)
#define SWIG_DelNewMask(r)         (SWIG_IsOK(r) ? (r & ~SWIG_NEWOBJMASK) : r)
#define SWIG_IsNewObj(r)           (SWIG_IsOK(r) && (r & SWIG_NEWOBJMASK))
#define SWIG_AddTmpMask(r)         (SWIG_IsOK(r) ? (r | SWIG_TMPOBJMASK) : r)
#define SWIG_DelTmpMask(r)         (SWIG_IsOK(r) ? (r & ~SWIG_TMPOBJMASK) : r)
#define SWIG_IsTmpObj(r)           (SWIG_IsOK(r) && (r & SWIG_TMPOBJMASK))

/* Cast-Rank Mode */
#if defined(SWIG_CASTRANK_MODE)
#  ifndef SWIG_TypeRank
#    define SWIG_TypeRank             unsigned long
#  endif
#  ifndef SWIG_MAXCASTRANK            /* Default cast allowed */
#    define SWIG_MAXCASTRANK          (2)
#  endif
#  define SWIG_CASTRANKMASK          ((SWIG_CASTRANKLIMIT) -1)
#  define SWIG_CastRank(r)           (r & SWIG_CASTRANKMASK)
SWIGINTERNINLINE int SWIG_AddCast(int r) {
  return SWIG_IsOK(r) ? ((SWIG_CastRank(r) < SWIG_MAXCASTRANK) ? (r + 1) : SWIG_ERROR) : r;
}
SWIGINTERNINLINE int SWIG_CheckState(int r) {
  return SWIG_IsOK(r) ? SWIG_CastRank(r) + 1 : 0;
}
#else /* no cast-rank mode */
#  define SWIG_AddCast(r) (r)
#  define SWIG_CheckState(r) (SWIG_IsOK(r) ? 1 : 0)
#endif


#include <string.h>

#ifdef __cplusplus
extern "C" {
#endif

typedef void *(*swig_converter_func)(void *, int *);
typedef struct swig_type_info *(*swig_dycast_func)(void **);

/* Structure to store information on one type */
typedef struct swig_type_info {
  const char             *name;			/* mangled name of this type */
  const char             *str;			/* human readable name of this type */
  swig_dycast_func        dcast;		/* dynamic cast function down a hierarchy */
  struct swig_cast_info  *cast;			/* linked list of types that can cast into this type */
  void                   *clientdata;		/* language specific type data */
  int                    owndata;		/* flag if the structure owns the clientdata */
} swig_type_info;

/* Structure to store a type and conversion function used for casting */
typedef struct swig_cast_info {
  swig_type_info         *type;			/* pointer to type that is equivalent to this type */
  swig_converter_func     converter;		/* function to cast the void pointers */
  struct swig_cast_info  *next;			/* pointer to next array of casts | pointer to cast hashed by value */
  unsigned int            value;		/* index of the last valid element in the array | typename hash value */
} swig_cast_info;

/* Structure used to store module information
 * Each module generates one structure like this, and the runtime collects
 * all of these structures and stores them in a circularly linked list.*/
typedef struct swig_module_info {
  swig_type_info         **types;		/* Array of pointers to swig_type_info structures that are in this module */
  size_t                 size;		        /* Number of types in this module */
  struct swig_module_info *next;		/* Pointer to next element in circularly linked list */
  swig_type_info         **type_initial;	/* Array of initially generated type structures */
  swig_cast_info         **cast_initial;	/* Array of initially generated casting structures */
  void                    *clientdata;		/* Language specific module data */
} swig_module_info;

/*
  Compare two type names skipping the space characters, therefore
  "char*" == "char *" and "Class<int>" == "Class<int >", etc.

  Return 0 when the two name types are equivalent, as in
  strncmp, but skipping ' '.
*/
SWIGRUNTIME int
SWIG_TypeNameComp(const char *f1, const char *l1,
		  const char *f2, const char *l2) {
  for (;(f1 != l1) && (f2 != l2); ++f1, ++f2) {
    while ((*f1 == ' ') && (f1 != l1)) ++f1;
    while ((*f2 == ' ') && (f2 != l2)) ++f2;
    if (*f1 != *f2) return (*f1 > *f2) ? 1 : -1;
  }
  return (int)((l1 - f1) - (l2 - f2));
}

/*
  Check type equivalence in a name list like <name1>|<name2>|...
  Return 0 if equal, -1 if nb < tb, 1 if nb > tb
*/
SWIGRUNTIME int
SWIG_TypeCmp(const char *nb, const char *tb) {
  int equiv = 1;
  const char* te = tb + strlen(tb);
  const char* ne = nb;
  while (equiv != 0 && *ne) {
    for (nb = ne; *ne; ++ne) {
      if (*ne == '|') break;
    }
    equiv = SWIG_TypeNameComp(nb, ne, tb, te);
    if (*ne) ++ne;
  }
  return equiv;
}

/*
  Check type equivalence in a name list like <name1>|<name2>|...
  Return 0 if not equal, 1 if equal
*/
SWIGRUNTIME int
SWIG_TypeEquiv(const char *nb, const char *tb) {
  return SWIG_TypeCmp(nb, tb) == 0 ? 1 : 0;
}

/*
 * Hash function for type name strings, based on maRushPrime1Hash (http://amsoftware.narod.ru/algo2.html)
 */
SWIGRUNTIME unsigned int SWIG_Hash(const char *str, unsigned int len) {
  const unsigned char *data = (const unsigned char *)str;
  unsigned int hash = len, i = 0, k;
  int rem = (int)len;

  while (rem >= (int)sizeof(unsigned int)) {
    k = *(unsigned int *)data;
    k += i++;
    hash ^= k;
    hash *= 171717;
    data += sizeof(unsigned int);
    rem -= sizeof(unsigned int);
  }

  switch (rem) {
    case 3: k = (unsigned int)(data[2]) << 16;
            k |= (unsigned int)(data[1]) << 8;
            k |= (unsigned int)(data[0]);
            k += i++;
            hash ^= k;
            hash *= 171717;
            break;
    case 2: k = (unsigned int)(data[1]) << 8;
            k |= (unsigned int)(data[0]);
            k += i++;
            hash ^= k;
            hash *= 171717;
            break;
    case 1: k = (unsigned int)(data[0]);
            k += i++;
            hash ^= k;
            hash *= 171717;
            break;
  }
  return hash;
}

/*
  Check the typename
*/
SWIGRUNTIME swig_cast_info *
SWIG_TypeCheck(const char *c, swig_type_info *ty) {
  static const unsigned int scan_threshold = 4;
  if (ty) {
    swig_cast_info *head = ty->cast;
    unsigned int hash_value = 0;
    int hashed = 0;

    while (head) {

      if (strcmp(head->type->name, c) == 0) {
        return head;
      }

      if (head->value) {
        swig_cast_info *iter;
        swig_cast_info *last = head + head->value;
        swig_cast_info *first = head + 1;
        int search = 1;

        if (!hashed) {
          if (head->value < scan_threshold) {
            for (iter = first; iter <= last; iter++) {
              if (strcmp(iter->type->name, c) == 0) {
                return iter;
              }
            }
            search = 0;
          } else {
            hashed = 1;
            hash_value = SWIG_Hash(c, (unsigned int)strlen(c));
          }
        }

        if (search) {
          /* Binary search over sorted <'next'|'value'> pairs */
          do {
            iter = first + ((last - first) >> 1);
            if (iter->value < hash_value) {
              first = iter + 1;
            } else if (iter->value == hash_value) {

              if (strcmp(iter->next->type->name, c) == 0) {
                return iter->next;
              }

              /* Hash collision check */
              for (last = iter + 1; last->next && last->value == hash_value; last++) {
                if (strcmp(last->next->type->name, c) == 0) {
                  return last->next;
                }
              }
              for (first = iter - 1; first != head && first->value == hash_value; first--) {
                if (strcmp(first->next->type->name, c) == 0) {
                  return first->next;
                }
              }
              break;
            } else
              last = iter - 1;
          } while (first <= last);
        }
      }
      head = head->next;
    }
  }
  return 0;
}

/*
  Check the type by type address
*/
SWIGRUNTIME swig_cast_info *
SWIG_TypeCheckStruct(const swig_type_info *from, swig_type_info *ty) {
  if (ty) {
    swig_cast_info *head = ty->cast;
    while (head) {
      if (head->type == from) {
        return head;
      }

      if (head->value) {
        swig_cast_info *iter;
        swig_cast_info *last = head + head->value;
        swig_cast_info *first = head + 1;

        /* Binary search over sorted array of casts */
        do {
          iter = first + ((last - first) >> 1);
          if (iter->type < from) {
            first = iter + 1;
          } else if (iter->type == from) {
            return iter;
          } else
            last = iter - 1;
        } while (first <= last);
      }
      head = head->next;
    }
  }
  return 0;
}

/*
  Cast a pointer up an inheritance hierarchy
*/
SWIGRUNTIMEINLINE void *
SWIG_TypeCast(swig_cast_info *ty, void *ptr, int *newmemory) {
  return ((!ty) || (!ty->converter)) ? ptr : (*ty->converter)(ptr, newmemory);
}

/*
   Dynamic pointer casting. Down an inheritance hierarchy
*/
SWIGRUNTIME swig_type_info *
SWIG_TypeDynamicCast(swig_type_info *ty, void **ptr) {
  swig_type_info *lastty = ty;
  if (!ty || !ty->dcast) return ty;
  while (ty && (ty->dcast)) {
    ty = (*ty->dcast)(ptr);
    if (ty) lastty = ty;
  }
  return lastty;
}

/*
  Return the name associated with this type
*/
SWIGRUNTIMEINLINE const char *
SWIG_TypeName(const swig_type_info *ty) {
  return ty->name;
}

/*
  Return the pretty name associated with this type,
  that is an unmangled type name in a form presentable to the user.
*/
SWIGRUNTIME const char *
SWIG_TypePrettyName(const swig_type_info *type) {
  /* The "str" field contains the equivalent pretty names of the
     type, separated by vertical-bar characters.  Choose the last
     name. It should be the most specific; a fully resolved name
     but not necessarily with default template parameters expanded. */
  if (!type) return NULL;
  if (type->str != NULL) {
    const char *last_name = type->str;
    const char *s;
    for (s = type->str; *s; s++)
      if (*s == '|') last_name = s+1;
    return last_name;
  }
  else
    return type->name;
}

/*
   Set the clientdata field for a type
*/
SWIGRUNTIME void
SWIG_TypeClientData(swig_type_info *ti, void *clientdata) {
  swig_cast_info *head = ti->cast;
  /* if (ti->clientdata == clientdata) return; */
  ti->clientdata = clientdata;

  while (head) {
    swig_cast_info *cast;
    for (cast = head; (cast - head) <= head->value; cast++) {
      if (!cast->converter) {
        swig_type_info *tc = cast->type;
        if (!tc->clientdata) {
          SWIG_TypeClientData(tc, clientdata);
        }
      }
    }
    head = head->next;
  }
}

SWIGRUNTIME void
SWIG_TypeNewClientData(swig_type_info *ti, void *clientdata) {
  SWIG_TypeClientData(ti, clientdata);
  ti->owndata = 1;
}

/*
  Search for a swig_type_info structure only by mangled name
  Search is a O(log #types)

  We start searching at module start, and finish searching when start == end.
  Note: if start == end at the beginning of the function, we go all the way around
  the circular list.
*/
SWIGRUNTIME swig_type_info *
SWIG_MangledTypeQueryModule(swig_module_info *start,
                            swig_module_info *end,
		            const char *name) {
  swig_module_info *iter = start;
  do {
    if (iter->size) {
      size_t l = 0;
      size_t r = iter->size - 1;
      do {
	/* since l+r >= 0, we can (>> 1) instead (/ 2) */
	size_t i = (l + r) >> 1;
	const char *iname = iter->types[i]->name;
	if (iname) {
	  int compare = strcmp(name, iname);
	  if (compare == 0) {
	    return iter->types[i];
	  } else if (compare < 0) {
	    if (i) {
	      r = i - 1;
	    } else {
	      break;
	    }
	  } else if (compare > 0) {
	    l = i + 1;
	  }
	} else {
	  break; /* should never happen */
	}
      } while (l <= r);
    }
    iter = iter->next;
  } while (iter != end);
  return 0;
}

/*
  Search for a swig_type_info structure for either a mangled name or a human readable name.
  It first searches the mangled names of the types, which is a O(log #types)
  If a type is not found it then searches the human readable names, which is O(#types).

  We start searching at module start, and finish searching when start == end.
  Note: if start == end at the beginning of the function, we go all the way around
  the circular list.
*/
SWIGRUNTIME swig_type_info *
SWIG_TypeQueryModule(swig_module_info *start,
                     swig_module_info *end,
		     const char *name) {
  /* STEP 1: Search the name field using binary search */
  swig_type_info *ret = SWIG_MangledTypeQueryModule(start, end, name);
  if (ret) {
    return ret;
  } else {
    /* STEP 2: If the type hasn't been found, do a complete search
       of the str field (the human readable name) */
    swig_module_info *iter = start;
    do {
      size_t i = 0;
      for (; i < iter->size; ++i) {
	if (iter->types[i]->str && (SWIG_TypeEquiv(iter->types[i]->str, name)))
	  return iter->types[i];
      }
      iter = iter->next;
    } while (iter != end);
  }

  /* neither found a match */
  return 0;
}

/*
   Pack binary data into a string
*/
SWIGRUNTIME char *
SWIG_PackData(char *c, void *ptr, size_t sz) {
  static const char hex[17] = "0123456789abcdef";
  const unsigned char *u = (unsigned char *) ptr;
  const unsigned char *eu =  u + sz;
  for (; u != eu; ++u) {
    unsigned char uu = *u;
    *(c++) = hex[(uu & 0xf0) >> 4];
    *(c++) = hex[uu & 0xf];
  }
  return c;
}

/*
   Unpack binary data from a string
*/
SWIGRUNTIME const char *
SWIG_UnpackData(const char *c, void *ptr, size_t sz) {
  unsigned char *u = (unsigned char *) ptr;
  const unsigned char *eu = u + sz;
  for (; u != eu; ++u) {
    char d = *(c++);
    unsigned char uu;
    if ((d >= '0') && (d <= '9'))
      uu = (unsigned char)((d - '0') << 4);
    else if ((d >= 'a') && (d <= 'f'))
      uu = (unsigned char)((d - ('a'-10)) << 4);
    else
      return (char *) 0;
    d = *(c++);
    if ((d >= '0') && (d <= '9'))
      uu |= (unsigned char)(d - '0');
    else if ((d >= 'a') && (d <= 'f'))
      uu |= (unsigned char)(d - ('a'-10));
    else
      return (char *) 0;
    *u = uu;
  }
  return c;
}

/*
   Pack 'void *' into a string buffer.
*/
SWIGRUNTIME char *
SWIG_PackVoidPtr(char *buff, void *ptr, const char *name, size_t bsz) {
  char *r = buff;
  if ((2*sizeof(void *) + 2) > bsz) return 0;
  *(r++) = '_';
  r = SWIG_PackData(r,&ptr,sizeof(void *));
  if (strlen(name) + 1 > (bsz - (r - buff))) return 0;
  strcpy(r,name);
  return buff;
}

SWIGRUNTIME const char *
SWIG_UnpackVoidPtr(const char *c, void **ptr, const char *name) {
  if (*c != '_') {
    if (strcmp(c,"NULL") == 0) {
      *ptr = (void *) 0;
      return name;
    } else {
      return 0;
    }
  }
  return SWIG_UnpackData(++c,ptr,sizeof(void *));
}

SWIGRUNTIME char *
SWIG_PackDataName(char *buff, void *ptr, size_t sz, const char *name, size_t bsz) {
  char *r = buff;
  size_t lname = (name ? strlen(name) : 0);
  if ((2*sz + 2 + lname) > bsz) return 0;
  *(r++) = '_';
  r = SWIG_PackData(r,ptr,sz);
  if (lname) {
    strncpy(r,name,lname+1);
  } else {
    *r = 0;
  }
  return buff;
}

SWIGRUNTIME const char *
SWIG_UnpackDataName(const char *c, void *ptr, size_t sz, const char *name) {
  if (*c != '_') {
    if (strcmp(c,"NULL") == 0) {
      memset(ptr,0,sz);
      return name;
    } else {
      return 0;
    }
  }
  return SWIG_UnpackData(++c,ptr,sz);
}

#ifdef __cplusplus
}
#endif

/* SWIG Errors applicable to all language modules, values are reserved from -1 to -99 */
#define  SWIG_UnknownError    	   -1
#define  SWIG_IOError        	   -2
#define  SWIG_RuntimeError   	   -3
#define  SWIG_IndexError     	   -4
#define  SWIG_TypeError      	   -5
#define  SWIG_DivisionByZero 	   -6
#define  SWIG_OverflowError  	   -7
#define  SWIG_SyntaxError    	   -8
#define  SWIG_ValueError     	   -9
#define  SWIG_SystemError    	   -10
#define  SWIG_AttributeError 	   -11
#define  SWIG_MemoryError    	   -12
#define  SWIG_NullReferenceError   -13


#if PY_VERSION_HEX >= 0x03030000 && !defined(SWIG_NO_HEAPTYPES)
#if !defined(SWIG_HEAPTYPES)
#define SWIG_HEAPTYPES
#endif
#endif

/* Compatibility macros for Python 3 */
#if PY_VERSION_HEX >= 0x03000000

#define PyClass_Check(obj) PyObject_IsInstance(obj, (PyObject *)&PyType_Type)
#define PyInt_Check(x) PyLong_Check(x)
#define PyInt_AsLong(x) PyLong_AsLong(x)
#define PyInt_FromLong(x) PyLong_FromLong(x)
#define PyInt_FromSize_t(x) PyLong_FromSize_t(x)
#define PyString_Check(name) PyBytes_Check(name)
#define PyString_FromString(x) PyUnicode_FromString(x)
#define PyString_Format(fmt, args)  PyUnicode_Format(fmt, args)
#define PyString_AsString(str) PyBytes_AsString(str)
#define PyString_Size(str) PyBytes_Size(str)	
#define PyString_InternFromString(key) PyUnicode_InternFromString(key)
#define Py_TPFLAGS_HAVE_CLASS Py_TPFLAGS_BASETYPE
#define _PyLong_FromSsize_t(x) PyLong_FromSsize_t(x)

#endif

/* SWIG APIs for compatibility of both Python 2 & 3 */

#if PY_VERSION_HEX >= 0x03000000
#  define SWIG_Python_str_FromFormat PyUnicode_FromFormat
#else
#  define SWIG_Python_str_FromFormat PyString_FromFormat
#endif

#if defined(SWIG_HEAPTYPES)
#if PY_VERSION_HEX < 0x030c0000
#include <structmember.h>
#define Py_READONLY READONLY
#define Py_T_PYSSIZET T_PYSSIZET
#endif
#endif

#include <stddef.h> /* For offsetof */


/* Wrapper around PyUnicode_AsUTF8AndSize - call Py_XDECREF on the returned pbytes when finished with the returned string */
SWIGINTERN const char *
SWIG_PyUnicode_AsUTF8AndSize(PyObject *str, Py_ssize_t *psize, PyObject **pbytes)
{
#if PY_VERSION_HEX >= 0x03030000
# if !defined(Py_LIMITED_API) || Py_LIMITED_API+0 >= 0x030A0000
  *pbytes = NULL;
  return PyUnicode_AsUTF8AndSize(str, psize);
# else
  const char *chars;
  *pbytes = PyUnicode_AsUTF8String(str);
  chars = *pbytes ? PyBytes_AsString(*pbytes) : NULL;
  if (chars && psize)
    *psize = PyBytes_Size(*pbytes);
  return chars;
# endif
#else
  char *chars = NULL;
  *pbytes = NULL;
  PyString_AsStringAndSize(str, &chars, psize);
  return chars;
#endif
}

SWIGINTERN PyObject *
SWIG_Python_str_FromChar(const char *c)
{
#if PY_VERSION_HEX >= 0x03000000
  return PyUnicode_FromString(c); 
#else
  return PyString_FromString(c);
#endif
}

#define SWIG_RUNTIME_MODULE "swig_runtime_data" SWIG_RUNTIME_VERSION

/* SWIGPY_USE_CAPSULE is no longer used within SWIG itself, but some user interface files check for it. */
# define SWIGPY_USE_CAPSULE
#ifdef SWIGPYTHON_BUILTIN
# define SWIGPY_CAPSULE_ATTR_NAME "type_pointer_capsule_builtin" SWIG_TYPE_TABLE_NAME
#else
# define SWIGPY_CAPSULE_ATTR_NAME "type_pointer_capsule" SWIG_TYPE_TABLE_NAME
#endif
#define SWIGPY_CAPSULE_NAME SWIG_RUNTIME_MODULE "." SWIGPY_CAPSULE_ATTR_NAME

#if PY_VERSION_HEX < 0x03020000
#define PyDescr_TYPE(x) (((PyDescrObject *)(x))->d_type)
#define PyDescr_NAME(x) (((PyDescrObject *)(x))->d_name)
#define Py_hash_t long
#endif

#if defined(Py_LIMITED_API)
# define PyTuple_GET_ITEM PyTuple_GetItem
/* Note that PyTuple_SetItem() has different semantics from PyTuple_SET_ITEM as it decref's the original tuple item, so in general they cannot be used
  interchangeably. However in SWIG-generated code PyTuple_SET_ITEM is only used with newly initialized tuples without any items and for them this does work. */
# define PyTuple_SET_ITEM PyTuple_SetItem
# define PyTuple_GET_SIZE PyTuple_Size
# define PyCFunction_GET_FLAGS PyCFunction_GetFlags
# define PyCFunction_GET_FUNCTION PyCFunction_GetFunction
# define PyCFunction_GET_SELF PyCFunction_GetSelf
# define PyList_GET_ITEM PyList_GetItem
# define PyList_SET_ITEM PyList_SetItem
# define PySliceObject PyObject
#endif

/* Increment and Decrement wrappers - for portability when using the stable abi and for performance otherwise */
#ifdef Py_LIMITED_API
# define SWIG_Py_INCREF Py_IncRef
# define SWIG_Py_XINCREF Py_IncRef
# define SWIG_Py_DECREF Py_DecRef
# define SWIG_Py_XDECREF Py_DecRef
#else
# define SWIG_Py_INCREF Py_INCREF
# define SWIG_Py_XINCREF Py_XINCREF
# define SWIG_Py_DECREF Py_DECREF
# define SWIG_Py_XDECREF Py_XDECREF
#endif

#if PY_VERSION_HEX >= 0x03000000
#if PY_VERSION_HEX < 0x030d00a6
SWIGINTERN PyObject *
SWIG_PyType_GetFullyQualifiedName(PyTypeObject *type) {
  PyObject *result = NULL;
  PyObject *qualname = PyObject_GetAttrString((PyObject *)type, "__qualname__");
  if (qualname) {
    PyObject *mod = PyObject_GetAttrString((PyObject *)type, "__module__");
    if (mod) {
      if (PyUnicode_Check(mod) && PyUnicode_CompareWithASCIIString(mod, "builtins") && PyUnicode_CompareWithASCIIString(mod, "__main__")) {
        result = PyUnicode_FromFormat("%U%c%U", mod, '.', qualname);
        SWIG_Py_DECREF(qualname);
      } else {
        result = qualname;
      }
      SWIG_Py_DECREF(mod);
    } else {
      result = qualname;
    }
  }

  return result;
}
#else
# define SWIG_PyType_GetFullyQualifiedName PyType_GetFullyQualifiedName
#endif
#endif

/* gh-114329 added PyList_GetItemRef() to Python 3.13.0a4 */
#if PY_VERSION_HEX < 0x030d00a4
SWIGINTERN PyObject *
SWIG_PyList_GetItemRef(PyObject *op, Py_ssize_t index) {
  PyObject *item = PyList_GetItem(op, index);
  Py_XINCREF(item);
  return item;
}
#else
# define SWIG_PyList_GetItemRef PyList_GetItemRef
#endif

/* gh-106004 added PyDict_GetItemRef() and PyDict_GetItemStringRef() to Python 3.13.0a1
   functions are renamed here for compatibility with abi3audit */
#if PY_VERSION_HEX < 0x030d00a1
SWIGINTERN int
SWIG_PyDict_GetItemRef(PyObject *mp, PyObject *key, PyObject **result) {
#if PY_VERSION_HEX >= 0x03000000
  PyObject *item = PyDict_GetItemWithError(mp, key);
#else
  PyObject *item = _PyDict_GetItemWithError(mp, key);
#endif
  if (item != NULL) {
    *result = (PyObject *)(item);
    SWIG_Py_INCREF(*result);
    return 1;
  }
  if (!PyErr_Occurred()) {
    *result = NULL;
    return 0;
  }
  *result = NULL;
  return -1;
}

SWIGINTERN int
SWIG_PyDict_GetItemStringRef(PyObject *mp, const char *key, PyObject **result) {
  int res;
#if PY_VERSION_HEX >= 0x03000000
  PyObject *key_obj = PyUnicode_FromString(key);
#else
  PyObject *key_obj = PyString_FromString(key);
#endif
  if (key_obj == NULL) {
    *result = NULL;
    return -1;
  }
  res = SWIG_PyDict_GetItemRef(mp, key_obj, result);
  Py_DECREF(key_obj);
  return res;
}
#else
# define SWIG_PyDict_GetItemRef PyDict_GetItemRef
# define SWIG_PyDict_GetItemStringRef PyDict_GetItemStringRef
#endif

/* -----------------------------------------------------------------------------
 * error manipulation
 * ----------------------------------------------------------------------------- */

SWIGRUNTIME PyObject*
SWIG_Python_ErrorType(int code) {
  PyObject* type = 0;
  switch(code) {
  case SWIG_MemoryError:
    type = PyExc_MemoryError;
    break;
  case SWIG_IOError:
    type = PyExc_IOError;
    break;
  case SWIG_RuntimeError:
    type = PyExc_RuntimeError;
    break;
  case SWIG_IndexError:
    type = PyExc_IndexError;
    break;
  case SWIG_TypeError:
    type = PyExc_TypeError;
    break;
  case SWIG_DivisionByZero:
    type = PyExc_ZeroDivisionError;
    break;
  case SWIG_OverflowError:
    type = PyExc_OverflowError;
    break;
  case SWIG_SyntaxError:
    type = PyExc_SyntaxError;
    break;
  case SWIG_ValueError:
    type = PyExc_ValueError;
    break;
  case SWIG_SystemError:
    type = PyExc_SystemError;
    break;
  case SWIG_AttributeError:
    type = PyExc_AttributeError;
    break;
  case SWIG_NullReferenceError:
    type = PyExc_TypeError;
    break;
  default:
    type = PyExc_RuntimeError;
  }
  return type;
}


SWIGRUNTIME void
SWIG_Python_AddErrorMsg(const char* mesg)
{
  PyObject *type = 0;
  PyObject *value = 0;
  PyObject *traceback = 0;

  if (PyErr_Occurred())
    PyErr_Fetch(&type, &value, &traceback);
  if (value) {
    PyObject *old_str = PyObject_Str(value);
    PyObject *bytes = NULL;
    const char *tmp = SWIG_PyUnicode_AsUTF8AndSize(old_str, NULL, &bytes);
    PyErr_Clear();
    SWIG_Py_XINCREF(type);
    if (tmp)
      PyErr_Format(type, "%s %s", tmp, mesg);
    else
      PyErr_Format(type, "%s", mesg);
    SWIG_Py_XDECREF(bytes);
    SWIG_Py_DECREF(old_str);
    SWIG_Py_DECREF(value);
  } else {
    PyErr_SetString(PyExc_RuntimeError, mesg);
  }
}

SWIGRUNTIME int
SWIG_Python_TypeErrorOccurred(PyObject *obj)
{
  PyObject *error;
  if (obj)
    return 0;
  error = PyErr_Occurred();
  return error && PyErr_GivenExceptionMatches(error, PyExc_TypeError);
}

SWIGRUNTIME void
SWIG_Python_RaiseOrModifyTypeError(const char *message)
{
  if (SWIG_Python_TypeErrorOccurred(NULL)) {
    /* Use existing TypeError to preserve stacktrace and enhance with given message */
    PyObject *newvalue;
    PyObject *type = NULL, *value = NULL, *traceback = NULL;
    PyErr_Fetch(&type, &value, &traceback);
#if PY_VERSION_HEX >= 0x03000000
    newvalue = PyUnicode_FromFormat("%S\nAdditional information:\n%s", value, message);
#else
    newvalue = PyString_FromFormat("%s\nAdditional information:\n%s", PyString_AsString(value), message);
#endif
    if (newvalue) {
      SWIG_Py_XDECREF(value);
      PyErr_Restore(type, newvalue, traceback);
    } else {
      PyErr_Restore(type, value, traceback);
    }
  } else {
    /* Raise TypeError using given message */
    PyErr_SetString(PyExc_TypeError, message);
  }
}

#if defined(SWIG_PYTHON_NO_THREADS)
#  if defined(SWIG_PYTHON_THREADS)
#    undef SWIG_PYTHON_THREADS
#  endif
#endif
#if defined(SWIG_PYTHON_THREADS) /* Threading support is enabled */
#  if !defined(SWIG_PYTHON_USE_GIL) && !defined(SWIG_PYTHON_NO_USE_GIL)
#    define SWIG_PYTHON_USE_GIL
#  endif
#  if defined(SWIG_PYTHON_USE_GIL) /* Use PyGILState threads calls */
#    if !defined(SWIG_PYTHON_INITIALIZE_THREADS)
#      if PY_VERSION_HEX < 0x03070000
#        define SWIG_PYTHON_INITIALIZE_THREADS PyEval_InitThreads()
#      else
#        define SWIG_PYTHON_INITIALIZE_THREADS
#      endif
#    endif
#    ifdef __cplusplus /* C++ code */
       class SWIG_Python_Thread_Block {
         bool status;
         PyGILState_STATE state;
       public:
         void end() { if (status) { PyGILState_Release(state); status = false;} }
         SWIG_Python_Thread_Block() : status(true), state(PyGILState_Ensure()) {}
         ~SWIG_Python_Thread_Block() { end(); }
       };
       class SWIG_Python_Thread_Allow {
         bool status;
         PyThreadState *save;
       public:
         void end() { if (status) { status = false; PyEval_RestoreThread(save); }}
         SWIG_Python_Thread_Allow() : status(true), save(PyEval_SaveThread()) {}
         ~SWIG_Python_Thread_Allow() { end(); }
       };
#      define SWIG_PYTHON_THREAD_BEGIN_BLOCK   SWIG_Python_Thread_Block _swig_thread_block
#      define SWIG_PYTHON_THREAD_END_BLOCK     _swig_thread_block.end()
#      define SWIG_PYTHON_THREAD_BEGIN_ALLOW   SWIG_Python_Thread_Allow _swig_thread_allow
#      define SWIG_PYTHON_THREAD_END_ALLOW     _swig_thread_allow.end()
#    else /* C code */
#      define SWIG_PYTHON_THREAD_BEGIN_BLOCK   PyGILState_STATE _swig_thread_block = PyGILState_Ensure()
#      define SWIG_PYTHON_THREAD_END_BLOCK     PyGILState_Release(_swig_thread_block)
#      define SWIG_PYTHON_THREAD_BEGIN_ALLOW   PyThreadState *_swig_thread_allow = PyEval_SaveThread()
#      define SWIG_PYTHON_THREAD_END_ALLOW     PyEval_RestoreThread(_swig_thread_allow)
#    endif
#  else /* Old thread way, not implemented, user must provide it */
#    if !defined(SWIG_PYTHON_INITIALIZE_THREADS)
#      define SWIG_PYTHON_INITIALIZE_THREADS
#    endif
#    if !defined(SWIG_PYTHON_THREAD_BEGIN_BLOCK)
#      define SWIG_PYTHON_THREAD_BEGIN_BLOCK
#    endif
#    if !defined(SWIG_PYTHON_THREAD_END_BLOCK)
#      define SWIG_PYTHON_THREAD_END_BLOCK
#    endif
#    if !defined(SWIG_PYTHON_THREAD_BEGIN_ALLOW)
#      define SWIG_PYTHON_THREAD_BEGIN_ALLOW
#    endif
#    if !defined(SWIG_PYTHON_THREAD_END_ALLOW)
#      define SWIG_PYTHON_THREAD_END_ALLOW
#    endif
#  endif
#else /* No thread support */
#  define SWIG_PYTHON_INITIALIZE_THREADS
#  define SWIG_PYTHON_THREAD_BEGIN_BLOCK
#  define SWIG_PYTHON_THREAD_END_BLOCK
#  define SWIG_PYTHON_THREAD_BEGIN_ALLOW
#  define SWIG_PYTHON_THREAD_END_ALLOW
#endif

/* -----------------------------------------------------------------------------
 * Python API portion that goes into the runtime
 * ----------------------------------------------------------------------------- */

#ifdef __cplusplus
extern "C" {
#endif

/* -----------------------------------------------------------------------------
 * Constant declarations
 * ----------------------------------------------------------------------------- */

/* Constant Types */
#define SWIG_PY_POINTER 4
#define SWIG_PY_BINARY  5

/* Constant information structure */
typedef struct swig_const_info {
  int type;
  const char *name;
  long lvalue;
  double dvalue;
  void   *pvalue;
  swig_type_info **ptype;
} swig_const_info;

#ifdef __cplusplus
}
#endif


/* -----------------------------------------------------------------------------
 * pyrun.swg
 *
 * This file contains the runtime support for Python modules
 * and includes code for managing global variables and pointer
 * type checking.
 *
 * ----------------------------------------------------------------------------- */

#if PY_VERSION_HEX < 0x02070000 /* 2.7.0 */
# error "This version of SWIG only supports Python >= 2.7"
#endif

#if PY_VERSION_HEX >= 0x03000000 && PY_VERSION_HEX < 0x03050000
# error "This version of SWIG only supports Python 3 >= 3.5"
#endif

/* Common SWIG API */

/* for raw pointers */
#define SWIG_Python_ConvertPtr(obj, pptr, type, flags)  SWIG_Python_ConvertPtrAndOwn(obj, pptr, type, flags, 0)
#define SWIG_ConvertPtr(obj, pptr, type, flags)         SWIG_Python_ConvertPtr(obj, pptr, type, flags)
#define SWIG_ConvertPtrAndOwn(obj,pptr,type,flags,own)  SWIG_Python_ConvertPtrAndOwn(obj, pptr, type, flags, own)

#ifdef SWIGPYTHON_BUILTIN
#define SWIG_NewPointerObj(ptr, type, flags)            SWIG_Python_NewPointerObj(self, ptr, type, flags)
#else
#define SWIG_NewPointerObj(ptr, type, flags)            SWIG_Python_NewPointerObj(NULL, ptr, type, flags)
#endif

#define SWIG_InternalNewPointerObj(ptr, type, flags)	SWIG_Python_NewPointerObj(NULL, ptr, type, flags)

#define SWIG_CheckImplicit(ty)                          SWIG_Python_CheckImplicit(ty) 
#define SWIG_AcquirePtr(ptr, src)                       SWIG_Python_AcquirePtr(ptr, src)
#define swig_owntype                                    int

/* for raw packed data */
#define SWIG_ConvertPacked(obj, ptr, sz, ty)            SWIG_Python_ConvertPacked(obj, ptr, sz, ty)
#define SWIG_NewPackedObj(ptr, sz, type)                SWIG_Python_NewPackedObj(ptr, sz, type)

/* for class or struct pointers */
#define SWIG_ConvertInstance(obj, pptr, type, flags)    SWIG_ConvertPtr(obj, pptr, type, flags)
#define SWIG_NewInstanceObj(ptr, type, flags)           SWIG_NewPointerObj(ptr, type, flags)

/* for C or C++ function pointers */
#define SWIG_ConvertFunctionPtr(obj, pptr, type)        SWIG_Python_ConvertFunctionPtr(obj, pptr, type)
#define SWIG_NewFunctionPtrObj(ptr, type)               SWIG_Python_NewPointerObj(NULL, ptr, type, 0)

/* for C++ member pointers, ie, member methods */
#define SWIG_ConvertMember(obj, ptr, sz, ty)            SWIG_Python_ConvertPacked(obj, ptr, sz, ty)
#define SWIG_NewMemberObj(ptr, sz, type)                SWIG_Python_NewPackedObj(ptr, sz, type)


/* Runtime API */

#define SWIG_GetModule(clientdata)                      SWIG_Python_GetModule(clientdata)
#define SWIG_SetModule(clientdata, pointer)             SWIG_Python_SetModule(pointer)
#define SWIG_NewClientData(obj)                         SwigPyClientData_New(obj)

#define SWIG_SetErrorObj                                SWIG_Python_SetErrorObj                            
#define SWIG_SetErrorMsg                        	SWIG_Python_SetErrorMsg				   
#define SWIG_ErrorType(code)                    	SWIG_Python_ErrorType(code)                        
#define SWIG_Error(code, msg)            		SWIG_Python_SetErrorMsg(SWIG_ErrorType(code), msg) 
#define SWIG_fail                        		goto fail					   

/* Runtime API implementation */

/* Error manipulation */

SWIGINTERN void 
SWIG_Python_SetErrorObj(PyObject *errtype, PyObject *obj) {
  SWIG_PYTHON_THREAD_BEGIN_BLOCK; 
  PyErr_SetObject(errtype, obj);
  SWIG_Py_DECREF(obj);
  SWIG_PYTHON_THREAD_END_BLOCK;
}

SWIGINTERN void 
SWIG_Python_SetErrorMsg(PyObject *errtype, const char *msg) {
  SWIG_PYTHON_THREAD_BEGIN_BLOCK;
  PyErr_SetString(errtype, msg);
  SWIG_PYTHON_THREAD_END_BLOCK;
}

#define SWIG_Python_Raise(obj, type, desc)  SWIG_Python_SetErrorObj(SWIG_Python_ExceptionType(desc), obj)

/* Set a constant value */

#if defined(SWIGPYTHON_BUILTIN)

SWIGINTERN void
SwigPyBuiltin_AddPublicSymbol(PyObject *seq, const char *key) {
  PyObject *s = PyString_InternFromString(key);
  PyList_Append(seq, s);
  SWIG_Py_DECREF(s);
}

SWIGINTERN void
SWIG_Python_SetConstant(PyObject *d, PyObject *public_interface, const char *name, PyObject *obj) {   
  PyDict_SetItemString(d, name, obj);
  SWIG_Py_DECREF(obj);
  if (public_interface)
    SwigPyBuiltin_AddPublicSymbol(public_interface, name);
}

#else

SWIGINTERN void
SWIG_Python_SetConstant(PyObject *d, const char *name, PyObject *obj) {   
  PyDict_SetItemString(d, name, obj);
  SWIG_Py_DECREF(obj);
}

#endif

/* SWIG runtime data Python module */
static PyObject *Swig_runtime_data_module_global = NULL;

/* Create/obtain the single swig_runtime_data module which is used across different SWIG generated modules */
SWIGINTERN PyObject *
SWIG_runtime_data_module(void) {
  if (!Swig_runtime_data_module_global) {
#if PY_VERSION_HEX >= 0x030d0000
    /* free-threading note: the GIL is always enabled when this function is first called
       by SWIG_init, so there's no risk of race conditions */
    Swig_runtime_data_module_global = PyImport_AddModuleRef(SWIG_RUNTIME_MODULE);
#elif PY_VERSION_HEX >= 0x03000000
    Swig_runtime_data_module_global = PyImport_AddModule(SWIG_RUNTIME_MODULE);
    SWIG_Py_XINCREF(Swig_runtime_data_module_global);
#else
    static PyMethodDef swig_empty_runtime_method_table[] = { {NULL, NULL, 0, NULL} }; /* Sentinel */
    Swig_runtime_data_module_global = Py_InitModule(SWIG_RUNTIME_MODULE, swig_empty_runtime_method_table);
    SWIG_Py_XINCREF(Swig_runtime_data_module_global);
#endif
  }
  assert(Swig_runtime_data_module_global);
  return Swig_runtime_data_module_global;
}

/* Append a value to the result obj */
SWIGINTERN PyObject*
SWIG_Python_AppendOutput(PyObject* result, PyObject* obj, int is_void) {
  if (!result) {
    result = obj;
  } else if (result == Py_None && is_void) {
    SWIG_Py_DECREF(result);
    result = obj;
  } else {
    if (!PyList_Check(result)) {
      PyObject *o2 = result;
      result = PyList_New(1);
      if (result) {
        PyList_SET_ITEM(result, 0, o2);
      } else {
        SWIG_Py_DECREF(obj);
        return o2;
      }
    }
    PyList_Append(result,obj);
    SWIG_Py_DECREF(obj);
  }
  return result;
}

/* Unpack the argument tuple */

SWIGINTERN Py_ssize_t
SWIG_Python_UnpackTuple(PyObject *args, const char *name, Py_ssize_t min, Py_ssize_t max, PyObject **objs)
{
  if (!args) {
    if (!min && !max) {
      return 1;
    } else {
      PyErr_Format(PyExc_TypeError, "%s expected %s%d arguments, got none", 
		   name, (min == max ? "" : "at least "), (int)min);
      return 0;
    }
  }  
  if (!PyTuple_Check(args)) {
    if (min <= 1 && max >= 1) {
      Py_ssize_t i;
      objs[0] = args;
      for (i = 1; i < max; ++i) {
	objs[i] = 0;
      }
      return 2;
    }
    PyErr_SetString(PyExc_SystemError, "UnpackTuple() argument list is not a tuple");
    return 0;
  } else {
    Py_ssize_t l = PyTuple_GET_SIZE(args);
    if (l < min) {
      PyErr_Format(PyExc_TypeError, "%s expected %s%d arguments, got %d", 
		   name, (min == max ? "" : "at least "), (int)min, (int)l);
      return 0;
    } else if (l > max) {
      PyErr_Format(PyExc_TypeError, "%s expected %s%d arguments, got %d", 
		   name, (min == max ? "" : "at most "), (int)max, (int)l);
      return 0;
    } else {
      Py_ssize_t i;
      for (i = 0; i < l; ++i) {
	objs[i] = PyTuple_GET_ITEM(args, i);
      }
      for (; l < max; ++l) {
	objs[l] = 0;
      }
      return i + 1;
    }    
  }
}

SWIGINTERN int
SWIG_Python_CheckNoKeywords(PyObject *kwargs, const char *name) {
  int no_kwargs = 1;
  if (kwargs) {
    assert(PyDict_Check(kwargs));
    if (PyDict_Size(kwargs) > 0) {
      PyErr_Format(PyExc_TypeError, "%s() does not take keyword arguments", name);
      no_kwargs = 0;
    }
  }
  return no_kwargs;
}

/* A functor is a function object with one single object argument */
#define SWIG_Python_CallFunctor(functor, obj)	        PyObject_CallFunctionObjArgs(functor, obj, NULL);

/*
  Helper for static pointer initialization for both C and C++ code, for example
  static PyObject *SWIG_STATIC_POINTER(MyVar) = NewSomething(...);
*/
#ifdef __cplusplus
#define SWIG_STATIC_POINTER(var)  var
#else
#define SWIG_STATIC_POINTER(var)  var = 0; if (!var) var
#endif

#ifdef __cplusplus
extern "C" {
#endif

/* Python-specific SWIG API */
#define SWIG_newvarlink()                             SWIG_Python_newvarlink()
#define SWIG_addvarlink(p, name, get_attr, set_attr)  SWIG_Python_addvarlink(p, name, get_attr, set_attr)
#define SWIG_InstallConstants(d, constants)           SWIG_Python_InstallConstants(d, constants)
 
/* -----------------------------------------------------------------------------
 * global variable support code.
 * ----------------------------------------------------------------------------- */
 
typedef struct swig_globalvar {   
  char       *name;                  /* Name of global variable */
  PyObject *(*get_attr)(void);       /* Return the current value */
  int       (*set_attr)(PyObject *); /* Set the value */
  struct swig_globalvar *next;
} swig_globalvar;

typedef struct swig_varlinkobject {
  PyObject_HEAD
  swig_globalvar *vars;
} swig_varlinkobject;

SWIGINTERN PyObject *
SwigVarLink_repr(PyObject *SWIGUNUSEDPARM(v)) {
#if PY_VERSION_HEX >= 0x03000000
  return PyUnicode_InternFromString("<Swig global variables>");
#else
  return PyString_FromString("<Swig global variables>");
#endif
}

SWIGINTERN PyObject *
SwigVarLink_str(PyObject *o) {
  swig_varlinkobject *v = (swig_varlinkobject *) o;
#if PY_VERSION_HEX >= 0x03000000
  PyObject *str = PyUnicode_InternFromString("(");
  PyObject *tail;
  PyObject *joined;
  swig_globalvar *var;
  for (var = v->vars; var; var=var->next) {
    tail = PyUnicode_FromString(var->name);
    joined = PyUnicode_Concat(str, tail);
    SWIG_Py_DECREF(str);
    SWIG_Py_DECREF(tail);
    str = joined;
    if (var->next) {
        tail = PyUnicode_InternFromString(", ");
        joined = PyUnicode_Concat(str, tail);
        SWIG_Py_DECREF(str);
        SWIG_Py_DECREF(tail);
        str = joined;
    }
  }
  tail = PyUnicode_InternFromString(")");
  joined = PyUnicode_Concat(str, tail);
  SWIG_Py_DECREF(str);
  SWIG_Py_DECREF(tail);
  str = joined;
#else
  PyObject *str = PyString_FromString("(");
  swig_globalvar *var;
  for (var = v->vars; var; var=var->next) {
    PyString_ConcatAndDel(&str,PyString_FromString(var->name));
    if (var->next) PyString_ConcatAndDel(&str,PyString_FromString(", "));
  }
  PyString_ConcatAndDel(&str,PyString_FromString(")"));
#endif
  return str;
}

SWIGINTERN void
SwigVarLink_dealloc(PyObject *o) {
  swig_varlinkobject *v = (swig_varlinkobject *) o;
  swig_globalvar *var = v->vars;
  while (var) {
    swig_globalvar *n = var->next;
    free(var->name);
    free(var);
    var = n;
  }
}

SWIGINTERN PyObject *
SwigVarLink_getattr(PyObject *o, char *n) {
  swig_varlinkobject *v = (swig_varlinkobject *) o;
  PyObject *res = NULL;
  swig_globalvar *var = v->vars;
  while (var) {
    if (strcmp(var->name,n) == 0) {
      res = (*var->get_attr)();
      break;
    }
    var = var->next;
  }
  if (res == NULL && !PyErr_Occurred()) {
    PyErr_Format(PyExc_AttributeError, "Unknown C global variable '%s'", n);
  }
  return res;
}

SWIGINTERN int
SwigVarLink_setattr(PyObject *o, char *n, PyObject *p) {
  swig_varlinkobject *v = (swig_varlinkobject *) o;
  int res = 1;
  swig_globalvar *var = v->vars;
  while (var) {
    if (strcmp(var->name,n) == 0) {
      res = (*var->set_attr)(p);
      break;
    }
    var = var->next;
  }
  if (res == 1 && !PyErr_Occurred()) {
    PyErr_Format(PyExc_AttributeError, "Unknown C global variable '%s'", n);
  }
  return res;
}

SWIGINTERN PyTypeObject*
SwigVarLink_TypeOnce(void) {
  static char SwigVarLink_doc[] = "Swig variable link object";
#ifndef SWIG_HEAPTYPES
  static PyTypeObject varlink_type;
  static int type_init = 0;
  if (!type_init) {
    const PyTypeObject tmp = {
#if PY_VERSION_HEX >= 0x03000000
      PyVarObject_HEAD_INIT(NULL, 0)
#else
      PyObject_HEAD_INIT(NULL)
      0,                                  /* ob_size */
#endif
      SWIG_RUNTIME_MODULE ".SwigVarLink", /* tp_name */
      sizeof(swig_varlinkobject),         /* tp_basicsize */
      0,                                  /* tp_itemsize */
      (destructor) SwigVarLink_dealloc,   /* tp_dealloc */
#if PY_VERSION_HEX < 0x030800b4
      (printfunc)0,                       /* tp_print */
#else
      (Py_ssize_t)0,                      /* tp_vectorcall_offset */
#endif
      (getattrfunc) SwigVarLink_getattr,  /* tp_getattr */
      (setattrfunc) SwigVarLink_setattr,  /* tp_setattr */
      0,                                  /* tp_compare */
      (reprfunc) SwigVarLink_repr,        /* tp_repr */
      0,                                  /* tp_as_number */
      0,                                  /* tp_as_sequence */
      0,                                  /* tp_as_mapping */
      0,                                  /* tp_hash */
      0,                                  /* tp_call */
      (reprfunc) SwigVarLink_str,         /* tp_str */
      0,                                  /* tp_getattro */
      0,                                  /* tp_setattro */
      0,                                  /* tp_as_buffer */
      0,                                  /* tp_flags */
      SwigVarLink_doc,                    /* tp_doc */
      0,                                  /* tp_traverse */
      0,                                  /* tp_clear */
      0,                                  /* tp_richcompare */
      0,                                  /* tp_weaklistoffset */
      0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, /* tp_iter -> tp_weaklist */
      0,                                  /* tp_del */
      0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
      0,                                  /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
      0,                                  /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
      0,                                  /* tp_print */
#endif
#if PY_VERSION_HEX >= 0x030c0000
      0,                                  /* tp_watched */
#endif
#if PY_VERSION_HEX >= 0x030d00a4
      0,                                  /* tp_versions_used */
#endif
#ifdef COUNT_ALLOCS
      0,                                  /* tp_allocs */
      0,                                  /* tp_frees */
      0,                                  /* tp_maxalloc */
      0,                                  /* tp_prev */
      0                                   /* tp_next */
#endif
    };
    PyObject *runtime_data_module = SWIG_runtime_data_module();
    varlink_type = tmp;
    type_init = 1;
    if (PyType_Ready(&varlink_type) < 0)
      return NULL;
    if (PyModule_AddObject(runtime_data_module, "SwigVarLink", (PyObject *)&varlink_type) == 0)
      SWIG_Py_INCREF((PyObject *)&varlink_type);
  }
  return &varlink_type;
#else
  PyType_Slot slots[] = {
    { Py_tp_dealloc, (void *)SwigVarLink_dealloc },
    { Py_tp_repr, (void *)SwigVarLink_repr },
    { Py_tp_getattr, (void *)SwigVarLink_getattr },
    { Py_tp_setattr, (void *)SwigVarLink_setattr },
    { Py_tp_str, (void *)SwigVarLink_str },
    { Py_tp_doc, (void *)SwigVarLink_doc },
    { 0, NULL }
  };
  PyType_Spec spec = {
    SWIG_RUNTIME_MODULE ".SwigVarLink",
    sizeof(swig_varlinkobject),
    0,
    Py_TPFLAGS_DEFAULT,
    slots
  };
  PyObject *pytype = PyType_FromSpec(&spec);
  PyObject *runtime_data_module = SWIG_runtime_data_module();
  if (pytype && PyModule_AddObject(runtime_data_module, "SwigVarLink", pytype) == 0)
    SWIG_Py_INCREF(pytype);
  return (PyTypeObject *)pytype;
#endif
}

SWIGRUNTIME PyTypeObject*
SwigVarLink_Type(void) {
  static PyTypeObject *SWIG_STATIC_POINTER(type) = SwigVarLink_TypeOnce();
  return type;
}

/* Create a variable linking object for use later */
SWIGINTERN PyObject *
SWIG_Python_newvarlink(void) {
  swig_varlinkobject *result = PyObject_New(swig_varlinkobject, SwigVarLink_Type());
  if (result) {
    result->vars = 0;
  }
  return ((PyObject*) result);
}

SWIGINTERN void 
SWIG_Python_addvarlink(PyObject *p, const char *name, PyObject *(*get_attr)(void), int (*set_attr)(PyObject *p)) {
  swig_varlinkobject *v = (swig_varlinkobject *) p;
  swig_globalvar *gv = (swig_globalvar *) malloc(sizeof(swig_globalvar));
  if (gv) {
    size_t size = strlen(name)+1;
    gv->name = (char *)malloc(size);
    if (gv->name) {
      memcpy(gv->name, name, size);
      gv->get_attr = get_attr;
      gv->set_attr = set_attr;
      gv->next = v->vars;
    }
  }
  v->vars = gv;
}


static PyObject *Swig_Globals_global = NULL;
  
SWIGINTERN PyObject *
SWIG_globals(void) {
  if (Swig_Globals_global == NULL) {
    Swig_Globals_global = SWIG_newvarlink();
  }
  return Swig_Globals_global;
}

#ifdef __cplusplus
}
#endif

/* -----------------------------------------------------------------------------
 * Pointer declarations
 * ----------------------------------------------------------------------------- */

/* Flags for new pointer objects */
#define SWIG_POINTER_NOSHADOW       (SWIG_POINTER_OWN      << 1)
#define SWIG_POINTER_NEW            (SWIG_POINTER_NOSHADOW | SWIG_POINTER_OWN)

#define SWIG_POINTER_IMPLICIT_CONV  (SWIG_POINTER_DISOWN   << 1)

#define SWIG_BUILTIN_TP_INIT	    (SWIG_POINTER_OWN << 2)
#define SWIG_BUILTIN_INIT	    (SWIG_BUILTIN_TP_INIT | SWIG_POINTER_OWN)

#ifdef __cplusplus
extern "C" {
#endif

/* The python void return value */

SWIGRUNTIMEINLINE PyObject * 
SWIG_Py_Void(void)
{
  PyObject *none = Py_None;
  SWIG_Py_INCREF(none);
  return none;
}

/* SwigPyClientData */

typedef struct {
  PyObject *klass;
  PyObject *newraw;
  PyObject *newargs;
  PyObject *destroy;
  int delargs;
  int implicitconv;
  PyTypeObject *pytype;
} SwigPyClientData;

SWIGRUNTIMEINLINE int 
SWIG_Python_CheckImplicit(swig_type_info *ty)
{
  SwigPyClientData *data = (SwigPyClientData *)ty->clientdata;
  int fail = data ? data->implicitconv : 0;
  if (fail)
    PyErr_SetString(PyExc_TypeError, "Implicit conversion is prohibited for explicit constructors.");
  return fail;
}

SWIGRUNTIMEINLINE PyObject *
SWIG_Python_ExceptionType(swig_type_info *desc) {
  SwigPyClientData *data = desc ? (SwigPyClientData *) desc->clientdata : 0;
  PyObject *klass = data ? data->klass : 0;
  return (klass ? klass : PyExc_RuntimeError);
}


SWIGRUNTIME SwigPyClientData * 
SwigPyClientData_New(PyObject* obj)
{
  if (!obj) {
    return 0;
  } else {
    SwigPyClientData *data = (SwigPyClientData *)malloc(sizeof(SwigPyClientData));
    /* the klass element */
    data->klass = obj;
    SWIG_Py_INCREF(data->klass);
    /* the newraw method and newargs arguments used to create a new raw instance */
    if (PyClass_Check(obj)) {
      data->newraw = 0;
      SWIG_Py_INCREF(obj);
      data->newargs = obj;
    } else {
      data->newraw = PyObject_GetAttrString(data->klass, "__new__");
      if (data->newraw) {
        data->newargs = PyTuple_New(1);
        if (data->newargs) {
          SWIG_Py_INCREF(obj);
          PyTuple_SET_ITEM(data->newargs, 0, obj);
        } else {
          SWIG_Py_DECREF(data->newraw);
          SWIG_Py_DECREF(data->klass);
          free(data);
          return 0;
        }
      } else {
        SWIG_Py_INCREF(obj);
        data->newargs = obj;
      }
    }
    /* the destroy method, aka as the C++ delete method */
    data->destroy = PyObject_GetAttrString(data->klass, "__swig_destroy__");
    if (PyErr_Occurred()) {
      PyErr_Clear();
      data->destroy = 0;
    }
    if (data->destroy) {
      data->delargs = !(PyCFunction_GET_FLAGS(data->destroy) & METH_O);
    } else {
      data->delargs = 0;
    }
    data->implicitconv = 0;
    data->pytype = 0;
    return data;
  }
}

SWIGRUNTIME void 
SwigPyClientData_Del(SwigPyClientData *data)
{
  SWIG_Py_XDECREF(data->klass);
  SWIG_Py_XDECREF(data->newraw);
  SWIG_Py_XDECREF(data->newargs);
  SWIG_Py_XDECREF(data->destroy);
  free(data);
}

/* =============== SwigPyObject =====================*/

typedef struct {
  PyObject_HEAD
  void *ptr;
  swig_type_info *ty;
  int own;
  PyObject *next;
  PyObject *swigdict;
  PyObject *weakreflist;
} SwigPyObject;


#ifdef SWIGPYTHON_BUILTIN

SWIGRUNTIME PyObject *
SwigPyObject_get___dict__(PyObject *v, PyObject *SWIGUNUSEDPARM(args))
{
  SwigPyObject *sobj = (SwigPyObject *)v;

  if (!sobj->swigdict)
    sobj->swigdict = PyDict_New();

  SWIG_Py_XINCREF(sobj->swigdict);
  return sobj->swigdict;
}

#endif

SWIGRUNTIME PyObject *
SwigPyObject_long(SwigPyObject *v)
{
  return PyLong_FromVoidPtr(v->ptr);
}

SWIGRUNTIME PyObject *
SwigPyObject_format(const char* fmt, SwigPyObject *v)
{
  PyObject *res = NULL;
  PyObject *args = PyTuple_New(1);
  if (args) {
    PyObject *val = SwigPyObject_long(v);
    if (val) {
      PyObject *ofmt;
      PyTuple_SET_ITEM(args, 0, val);
      ofmt = SWIG_Python_str_FromChar(fmt);
      if (ofmt) {
#if PY_VERSION_HEX >= 0x03000000
        res = PyUnicode_Format(ofmt,args);
#else
        res = PyString_Format(ofmt,args);
#endif
        SWIG_Py_DECREF(ofmt);
      }
    }
    SWIG_Py_DECREF(args);
  }
  return res;
}

SWIGRUNTIME PyObject *
SwigPyObject_oct(SwigPyObject *v)
{
  return SwigPyObject_format("%o",v);
}

SWIGRUNTIME PyObject *
SwigPyObject_hex(SwigPyObject *v)
{
  return SwigPyObject_format("%x",v);
}

SWIGRUNTIME PyObject *
SwigPyObject_repr(SwigPyObject *v)
{
  const char *name = SWIG_TypePrettyName(v->ty);
  PyObject *repr = SWIG_Python_str_FromFormat("<Swig Object of type '%s' at %p>", (name ? name : "unknown"), (void *)v);
  if (repr && v->next) {
    PyObject *nrep = SwigPyObject_repr((SwigPyObject *)v->next);
    if (nrep) {
# if PY_VERSION_HEX >= 0x03000000
      PyObject *joined = PyUnicode_Concat(repr, nrep);
      SWIG_Py_DECREF(repr);
      SWIG_Py_DECREF(nrep);
      repr = joined;
# else
      PyString_ConcatAndDel(&repr,nrep);
# endif
    } else {
      SWIG_Py_DECREF(repr);
      repr = NULL;
    }
  }
  return repr;
}

/* We need a version taking two PyObject* parameters so it's a valid
 * PyCFunction to use in SwigPyObject_methods[]. */
SWIGRUNTIME PyObject *
SwigPyObject_repr2(PyObject *v, PyObject *SWIGUNUSEDPARM(args))
{
  return SwigPyObject_repr((SwigPyObject*)v);
}

SWIGRUNTIME int
SwigPyObject_compare(PyObject *v, PyObject *w)
{
  /* tp_compare is only called when both objects have the same type, so
   * the casts are guaranteed to be ok. */
  void *i = ((SwigPyObject *)v)->ptr;
  void *j = ((SwigPyObject *)w)->ptr;
  return (i < j) ? -1 : ((i > j) ? 1 : 0);
}

SWIGRUNTIMEINLINE int SwigPyObject_Check(PyObject *);

/* Added for Python 3.x, would it also be useful for Python 2.x? */
SWIGRUNTIME PyObject*
SwigPyObject_richcompare(PyObject *v, PyObject *w, int op)
{
  PyObject* res = NULL;
  if (!PyErr_Occurred()) {
    /* Per https://docs.python.org/3/c-api/typeobj.html#c.PyTypeObject.tp_richcompare
     * the first argument is guaranteed to be an instance of SwigPyObject, but the
     * second is not, so we typecheck that one. */
    if ((op != Py_EQ && op != Py_NE) || !SwigPyObject_Check(w)) {
      SWIG_Py_INCREF(Py_NotImplemented);
      return Py_NotImplemented;
    }
    res = PyBool_FromLong( (SwigPyObject_compare(v, w)==0) == (op == Py_EQ) ? 1 : 0);
  }
  return res;
}


SWIGRUNTIME PyTypeObject* SwigPyObject_TypeOnce(void);

#ifdef SWIGPYTHON_BUILTIN
static swig_type_info *SwigPyObject_stype = 0;
SWIGRUNTIME PyTypeObject*
SwigPyObject_Type(void) {
    SwigPyClientData *cd;
    assert(SwigPyObject_stype);
    cd = (SwigPyClientData*) SwigPyObject_stype->clientdata;
    assert(cd);
    assert(cd->pytype);
    return cd->pytype;
}
#else
SWIGRUNTIME PyTypeObject*
SwigPyObject_Type(void) {
  static PyTypeObject *SWIG_STATIC_POINTER(type) = SwigPyObject_TypeOnce();
  return type;
}
#endif

SWIGRUNTIMEINLINE int
SwigPyObject_Check(PyObject *op) {
  PyTypeObject *target_tp = SwigPyObject_Type();
  PyTypeObject *op_type = Py_TYPE(op);
#ifdef SWIGPYTHON_BUILTIN
  /* Only builtin types have SwigPyObject as a base type */
  return PyType_IsSubtype(op_type, target_tp);
#else
  /* Check for an exact match to SwigPyObject */
  if (op_type == target_tp) {
    return 1;
  } else {
    /* Fallback for multiple modules */
#if PY_VERSION_HEX >= 0x03000000
    int cmp;
    PyObject *tpname = SWIG_PyType_GetFullyQualifiedName(op_type);
    if (!tpname)
      return 0;
    cmp = PyUnicode_CompareWithASCIIString(tpname, SWIG_RUNTIME_MODULE ".SwigPyObject");
    SWIG_Py_DECREF(tpname);
    return cmp == 0;
#else
    return strcmp(op_type->tp_name, SWIG_RUNTIME_MODULE ".SwigPyObject") == 0;
#endif
  }
#endif
}

SWIGRUNTIME PyObject *
SwigPyObject_New(void *ptr, swig_type_info *ty, int own);

static PyObject* Swig_Capsule_global = NULL;

SWIGRUNTIME void
SwigPyObject_dealloc(PyObject *v)
{
  SwigPyObject *sobj = (SwigPyObject *) v;
  PyObject *next = sobj->next;
  if (sobj->own == SWIG_POINTER_OWN) {
    swig_type_info *ty = sobj->ty;
    SwigPyClientData *data = ty ? (SwigPyClientData *) ty->clientdata : 0;
    PyObject *destroy = data ? data->destroy : 0;
    if (destroy) {
      /* destroy is always a VARARGS method */
      PyObject *res;

      /* PyObject_CallFunction() has the potential to silently drop
         the active exception.  In cases of unnamed temporary
         variable or where we just finished iterating over a generator
         StopIteration will be active right now, and this needs to
         remain true upon return from SwigPyObject_dealloc.  So save
         and restore. */
      
      PyObject *type = NULL, *value = NULL, *traceback = NULL;
      PyErr_Fetch(&type, &value, &traceback);

      if (data->delargs) {
        /* we need to create a temporary object to carry the destroy operation */
        PyObject *tmp = SwigPyObject_New(sobj->ptr, ty, 0);
        if (tmp) {
          res = SWIG_Python_CallFunctor(destroy, tmp);
        } else {
          res = 0;
        }
        SWIG_Py_XDECREF(tmp);
      } else {
        PyCFunction meth = PyCFunction_GET_FUNCTION(destroy);
        PyObject *mself = PyCFunction_GET_SELF(destroy);
        res = ((*meth)(mself, v));
      }
      if (!res)
        PyErr_WriteUnraisable(destroy);

      PyErr_Restore(type, value, traceback);

      SWIG_Py_XDECREF(res);
    } 
#if !defined(SWIG_PYTHON_SILENT_MEMLEAK)
    else {
      const char *name = SWIG_TypePrettyName(ty);
      printf("swig/python detected a memory leak of type '%s', no destructor found.\n", (name ? name : "unknown"));
    }
#endif
    SWIG_Py_XDECREF(Swig_Capsule_global);
  }
  SWIG_Py_XDECREF(next);
  SWIG_Py_XDECREF(sobj->swigdict);
  PyObject_Free(v);
}

SWIGRUNTIME PyObject* 
SwigPyObject_append(PyObject* v, PyObject* next)
{
  SwigPyObject *sobj = (SwigPyObject *) v;
  if (!SwigPyObject_Check(next)) {
    PyErr_SetString(PyExc_TypeError, "Attempt to append a non SwigPyObject");
    return NULL;
  }
  ((SwigPyObject *)next)->next = sobj->next;
  sobj->next = next;
  SWIG_Py_INCREF(next);
  return SWIG_Py_Void();
}

SWIGRUNTIME PyObject* 
SwigPyObject_next(PyObject* v, PyObject *SWIGUNUSEDPARM(args))
{
  SwigPyObject *sobj = (SwigPyObject *) v;
  if (sobj->next) {    
    SWIG_Py_INCREF(sobj->next);
    return sobj->next;
  } else {
    return SWIG_Py_Void();
  }
}

SWIGINTERN PyObject*
SwigPyObject_disown(PyObject* v, PyObject *SWIGUNUSEDPARM(args))
{
  SwigPyObject *sobj = (SwigPyObject *)v;
  sobj->own = 0;
  return SWIG_Py_Void();
}

SWIGINTERN PyObject*
SwigPyObject_acquire(PyObject* v, PyObject *SWIGUNUSEDPARM(args))
{
  SwigPyObject *sobj = (SwigPyObject *)v;
  sobj->own = SWIG_POINTER_OWN;
  return SWIG_Py_Void();
}

SWIGINTERN PyObject*
SwigPyObject_own(PyObject *v, PyObject *args)
{
  PyObject *val = 0;
  if (!PyArg_UnpackTuple(args, "own", 0, 1, &val)) {
    return NULL;
  } else {
    SwigPyObject *sobj = (SwigPyObject *)v;
    PyObject *obj = PyBool_FromLong(sobj->own);
    if (val) {
      if (PyObject_IsTrue(val)) {
        SWIG_Py_DECREF(SwigPyObject_acquire(v,args));
      } else {
        SWIG_Py_DECREF(SwigPyObject_disown(v,args));
      }
    } 
    return obj;
  }
}

static PyMethodDef
SwigPyObject_methods[] = {
  {"disown",  SwigPyObject_disown,  METH_NOARGS,  "releases ownership of the pointer"},
  {"acquire", SwigPyObject_acquire, METH_NOARGS,  "acquires ownership of the pointer"},
  {"own",     SwigPyObject_own,     METH_VARARGS, "returns/sets ownership of the pointer"},
  {"append",  SwigPyObject_append,  METH_O,       "appends another 'this' object"},
  {"next",    SwigPyObject_next,    METH_NOARGS,  "returns the next 'this' object"},
  {"__repr__",SwigPyObject_repr2,   METH_NOARGS,  "returns object representation"},
  {0, 0, 0, 0}  
};

SWIGRUNTIME PyTypeObject*
SwigPyObject_TypeOnce(void) {
  static char SwigPyObject_doc[] = "Swig object holding a C/C++ pointer";
#ifndef SWIG_HEAPTYPES
  static PyNumberMethods SwigPyObject_as_number = {
    (binaryfunc)0, /*nb_add*/
    (binaryfunc)0, /*nb_subtract*/
    (binaryfunc)0, /*nb_multiply*/
    /* nb_divide removed in Python 3 */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc)0, /*nb_divide*/
#endif
    (binaryfunc)0, /*nb_remainder*/
    (binaryfunc)0, /*nb_divmod*/
    (ternaryfunc)0,/*nb_power*/
    (unaryfunc)0,  /*nb_negative*/
    (unaryfunc)0,  /*nb_positive*/
    (unaryfunc)0,  /*nb_absolute*/
    (inquiry)0,    /*nb_nonzero*/
    0,		   /*nb_invert*/
    0,		   /*nb_lshift*/
    0,		   /*nb_rshift*/
    0,		   /*nb_and*/
    0,		   /*nb_xor*/
    0,		   /*nb_or*/
#if PY_VERSION_HEX < 0x03000000
    0,   /*nb_coerce*/
#endif
    (unaryfunc)SwigPyObject_long, /*nb_int*/
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc)SwigPyObject_long, /*nb_long*/
#else
    0, /*nb_reserved*/
#endif
    (unaryfunc)0,                 /*nb_float*/
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc)SwigPyObject_oct,  /*nb_oct*/
    (unaryfunc)SwigPyObject_hex,  /*nb_hex*/
#endif
#if PY_VERSION_HEX >= 0x03050000 /* 3.5 */
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 /* nb_inplace_add -> nb_inplace_matrix_multiply */
#elif PY_VERSION_HEX >= 0x03000000 /* 3.0 */
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 /* nb_inplace_add -> nb_index, nb_inplace_divide removed */
#else
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 /* nb_inplace_add -> nb_index */
#endif
  };

  static PyTypeObject swigpyobject_type;
  static int type_init = 0;
  if (!type_init) {
    const PyTypeObject tmp = {
#if PY_VERSION_HEX >= 0x03000000
      PyVarObject_HEAD_INIT(NULL, 0)
#else
      PyObject_HEAD_INIT(NULL)
      0,                                    /* ob_size */
#endif
      SWIG_RUNTIME_MODULE ".SwigPyObject",  /* tp_name */
      sizeof(SwigPyObject),                 /* tp_basicsize */
      0,                                    /* tp_itemsize */
      (destructor)SwigPyObject_dealloc,     /* tp_dealloc */
#if PY_VERSION_HEX < 0x030800b4
      (printfunc)0,                         /* tp_print */
#else
      (Py_ssize_t)0,                        /* tp_vectorcall_offset */
#endif
      (getattrfunc)0,                       /* tp_getattr */
      (setattrfunc)0,                       /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
      0, /* tp_reserved in 3.0.1, tp_compare in 3.0.0 but not used */
#else
      (cmpfunc)SwigPyObject_compare,        /* tp_compare */
#endif
      (reprfunc)SwigPyObject_repr,          /* tp_repr */
      &SwigPyObject_as_number,              /* tp_as_number */
      0,                                    /* tp_as_sequence */
      0,                                    /* tp_as_mapping */
      (hashfunc)0,                          /* tp_hash */
      (ternaryfunc)0,                       /* tp_call */
      0,                                    /* tp_str */
      PyObject_GenericGetAttr,              /* tp_getattro */
      0,                                    /* tp_setattro */
      0,                                    /* tp_as_buffer */
      Py_TPFLAGS_DEFAULT,                   /* tp_flags */
      SwigPyObject_doc,                     /* tp_doc */
      0,                                    /* tp_traverse */
      0,                                    /* tp_clear */
      (richcmpfunc)SwigPyObject_richcompare,/* tp_richcompare */
      offsetof(SwigPyObject, weakreflist),  /* tp_weaklistoffset */
      0,                                    /* tp_iter */
      0,                                    /* tp_iternext */
      SwigPyObject_methods,                 /* tp_methods */
      0,                                    /* tp_members */
      0,                                    /* tp_getset */
      0,                                    /* tp_base */
      0,                                    /* tp_dict */
      0,                                    /* tp_descr_get */
      0,                                    /* tp_descr_set */
      0,                                    /* tp_dictoffset */
      0,                                    /* tp_init */
      0,                                    /* tp_alloc */
      0,                                    /* tp_new */
      0,                                    /* tp_free */
      0,                                    /* tp_is_gc */
      0,                                    /* tp_bases */
      0,                                    /* tp_mro */
      0,                                    /* tp_cache */
      0,                                    /* tp_subclasses */
      0,                                    /* tp_weaklist */
      0,                                    /* tp_del */
      0,                                    /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
      0,                                    /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
      0,                                    /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
      0,                                    /* tp_print */
#endif
#if PY_VERSION_HEX >= 0x030c0000
      0,                                    /* tp_watched */
#endif
#if PY_VERSION_HEX >= 0x030d00a4
      0,                                    /* tp_versions_used */
#endif
#ifdef COUNT_ALLOCS
      0,                                    /* tp_allocs */
      0,                                    /* tp_frees */
      0,                                    /* tp_maxalloc */
      0,                                    /* tp_prev */
      0                                     /* tp_next */
#endif
    };
    PyObject *runtime_data_module = SWIG_runtime_data_module();
    swigpyobject_type = tmp;
    type_init = 1;
    if (PyType_Ready(&swigpyobject_type) != 0)
      return NULL;
    if (PyModule_AddObject(runtime_data_module, "SwigPyObject", (PyObject *)&swigpyobject_type) == 0)
      SWIG_Py_INCREF((PyObject *)&swigpyobject_type);
  }
  return &swigpyobject_type;
#else
  static PyMemberDef SwigPyObject_members[] = {
      { (char *)"__dictoffset__", Py_T_PYSSIZET, offsetof(SwigPyObject, swigdict), Py_READONLY, NULL },
      { (char *)"__weaklistoffset__", Py_T_PYSSIZET, offsetof(SwigPyObject, weakreflist), Py_READONLY, NULL },
      { NULL, 0, 0, 0, NULL }
  };
  PyType_Slot slots[] = {
    { Py_tp_dealloc, (void *)SwigPyObject_dealloc },
    { Py_tp_repr, (void *)SwigPyObject_repr },
    { Py_tp_getattro, (void *)PyObject_GenericGetAttr },
    { Py_tp_doc, (void *)SwigPyObject_doc },
    { Py_tp_richcompare, (void *)SwigPyObject_richcompare },
    { Py_tp_methods, (void *)SwigPyObject_methods },
    { Py_nb_int, (void *)SwigPyObject_long },
    { Py_tp_members, (void *)SwigPyObject_members },
    { 0, NULL }
  };
  PyType_Spec spec = {
    SWIG_RUNTIME_MODULE ".SwigPyObject",
    sizeof(SwigPyObject),
    0,
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,
    slots
  };
  PyObject *pytype = PyType_FromSpec(&spec);
  PyObject *runtime_data_module = SWIG_runtime_data_module();
#if !defined(Py_LIMITED_API)
/* While this __dictoffset__ is only used with the builtin wrappers, SwigPyObject ought to be
   identical when created for use by proxy class wrappers in case it is shared across multiple modules. */
#if PY_VERSION_HEX < 0x03090000
    /* Workaround as __dictoffset__ and __weaklistoffset__ above are only supported from python-3.9 */
  if (pytype) {
    ((PyTypeObject *)pytype)->tp_dictoffset = offsetof(SwigPyObject, swigdict);
    ((PyTypeObject *)pytype)->tp_weaklistoffset = offsetof(SwigPyObject, weakreflist);
  }
#endif
#endif
  if (pytype && PyModule_AddObject(runtime_data_module, "SwigPyObject", pytype) == 0)
    SWIG_Py_INCREF(pytype);
  return (PyTypeObject *)pytype;
#endif
}

SWIGRUNTIME PyObject *
SwigPyObject_New(void *ptr, swig_type_info *ty, int own)
{
  SwigPyObject *sobj = PyObject_New(SwigPyObject, SwigPyObject_Type());
  if (sobj) {
    sobj->ptr  = ptr;
    sobj->ty   = ty;
    sobj->own  = own;
    sobj->next = 0;
    sobj->swigdict = 0;
    sobj->weakreflist = 0;
    if (own == SWIG_POINTER_OWN) {
      /* Obtain a reference to the Python capsule wrapping the module information, so that the
       * module information is correctly destroyed after all SWIG python objects have been freed
       * by the GC (and corresponding destructors invoked) */
      SWIG_Py_XINCREF(Swig_Capsule_global);
    }
  }
  return (PyObject *)sobj;
}

/* -----------------------------------------------------------------------------
 * Implements a simple Swig Packed type, and use it instead of string
 * ----------------------------------------------------------------------------- */

typedef struct {
  PyObject_HEAD
  void *pack;
  swig_type_info *ty;
  size_t size;
} SwigPyPacked;

SWIGRUNTIME PyObject *
SwigPyPacked_repr(SwigPyPacked *v)
{
  char result[SWIG_BUFFER_SIZE];
  if (SWIG_PackDataName(result, v->pack, v->size, 0, sizeof(result))) {
    return SWIG_Python_str_FromFormat("<Swig Packed at %s%s>", result, v->ty->name);
  } else {
    return SWIG_Python_str_FromFormat("<Swig Packed %s>", v->ty->name);
  }  
}

SWIGRUNTIME PyObject *
SwigPyPacked_str(SwigPyPacked *v)
{
  char result[SWIG_BUFFER_SIZE];
  if (SWIG_PackDataName(result, v->pack, v->size, 0, sizeof(result))){
    return SWIG_Python_str_FromFormat("%s%s", result, v->ty->name);
  } else {
    return SWIG_Python_str_FromChar(v->ty->name);
  }  
}

SWIGRUNTIME int
SwigPyPacked_compare(SwigPyPacked *v, SwigPyPacked *w)
{
  size_t i = v->size;
  size_t j = w->size;
  int s = (i < j) ? -1 : ((i > j) ? 1 : 0);
  return s ? s : strncmp((const char *)v->pack, (const char *)w->pack, 2*v->size);
}

SWIGRUNTIME PyTypeObject* SwigPyPacked_TypeOnce(void);

SWIGRUNTIME PyTypeObject*
SwigPyPacked_Type(void) {
  static PyTypeObject *SWIG_STATIC_POINTER(type) = SwigPyPacked_TypeOnce();
  return type;
}

SWIGRUNTIMEINLINE int
SwigPyPacked_Check(PyObject *op) {
  PyTypeObject *target_tp = SwigPyPacked_Type();
  PyTypeObject *op_type = Py_TYPE(op);
  /* Check for an exact match to SwigPyPacked */
  if (op_type == target_tp) {
    return 1;
  } else {
    /* Fallback for multiple modules */
#if PY_VERSION_HEX >= 0x03000000
    int cmp;
    PyObject *tpname = SWIG_PyType_GetFullyQualifiedName(op_type);
    if (!tpname)
      return 0;
    cmp = PyUnicode_CompareWithASCIIString(tpname, SWIG_RUNTIME_MODULE ".SwigPyPacked");
    SWIG_Py_DECREF(tpname);
    return cmp == 0;
#else
    return strcmp(op_type->tp_name, SWIG_RUNTIME_MODULE ".SwigPyPacked") == 0;
#endif
  }
}

SWIGRUNTIME void
SwigPyPacked_dealloc(PyObject *v)
{
  if (SwigPyPacked_Check(v)) {
    SwigPyPacked *sobj = (SwigPyPacked *) v;
    free(sobj->pack);
  }
  PyObject_Free(v);
}

SWIGRUNTIME PyTypeObject*
SwigPyPacked_TypeOnce(void) {
  static char SwigPyPacked_doc[] = "Swig object holding a C/C++ function pointer";
#ifndef SWIG_HEAPTYPES
  static PyTypeObject swigpypacked_type;
  static int type_init = 0;
  if (!type_init) {
    const PyTypeObject tmp = {
#if PY_VERSION_HEX >= 0x03000000
      PyVarObject_HEAD_INIT(NULL, 0)
#else
      PyObject_HEAD_INIT(NULL)
      0,                                    /* ob_size */
#endif
      SWIG_RUNTIME_MODULE ".SwigPyPacked",  /* tp_name */
      sizeof(SwigPyPacked),                 /* tp_basicsize */
      0,                                    /* tp_itemsize */
      (destructor)SwigPyPacked_dealloc,     /* tp_dealloc */
#if PY_VERSION_HEX < 0x030800b4
      (printfunc)0,                         /* tp_print */
#else
      (Py_ssize_t)0,                        /* tp_vectorcall_offset */
#endif
      (getattrfunc)0,                       /* tp_getattr */
      (setattrfunc)0,                       /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
      0, /* tp_reserved in 3.0.1 */
#else
      (cmpfunc)SwigPyPacked_compare,        /* tp_compare */
#endif
      (reprfunc)SwigPyPacked_repr,          /* tp_repr */
      0,                                    /* tp_as_number */
      0,                                    /* tp_as_sequence */
      0,                                    /* tp_as_mapping */
      (hashfunc)0,                          /* tp_hash */
      (ternaryfunc)0,                       /* tp_call */
      (reprfunc)SwigPyPacked_str,           /* tp_str */
      PyObject_GenericGetAttr,              /* tp_getattro */
      0,                                    /* tp_setattro */
      0,                                    /* tp_as_buffer */
      Py_TPFLAGS_DEFAULT,                   /* tp_flags */
      SwigPyPacked_doc,                     /* tp_doc */
      0,                                    /* tp_traverse */
      0,                                    /* tp_clear */
      0,                                    /* tp_richcompare */
      0,                                    /* tp_weaklistoffset */
      0,                                    /* tp_iter */
      0,                                    /* tp_iternext */
      0,                                    /* tp_methods */
      0,                                    /* tp_members */
      0,                                    /* tp_getset */
      0,                                    /* tp_base */
      0,                                    /* tp_dict */
      0,                                    /* tp_descr_get */
      0,                                    /* tp_descr_set */
      0,                                    /* tp_dictoffset */
      0,                                    /* tp_init */
      0,                                    /* tp_alloc */
      0,                                    /* tp_new */
      0,                                    /* tp_free */
      0,                                    /* tp_is_gc */
      0,                                    /* tp_bases */
      0,                                    /* tp_mro */
      0,                                    /* tp_cache */
      0,                                    /* tp_subclasses */
      0,                                    /* tp_weaklist */
      0,                                    /* tp_del */
      0,                                    /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
      0,                                    /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
      0,                                    /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
      0,                                    /* tp_print */
#endif
#if PY_VERSION_HEX >= 0x030c0000
      0,                                    /* tp_watched */
#endif
#if PY_VERSION_HEX >= 0x030d00a4
      0,                                    /* tp_versions_used */
#endif
#ifdef COUNT_ALLOCS
      0,                                    /* tp_allocs */
      0,                                    /* tp_frees */
      0,                                    /* tp_maxalloc */
      0,                                    /* tp_prev */
      0                                     /* tp_next */
#endif
    };
    PyObject *runtime_data_module = SWIG_runtime_data_module();
    swigpypacked_type = tmp;
    type_init = 1;
    if (PyType_Ready(&swigpypacked_type) != 0)
      return NULL;
    if (PyModule_AddObject(runtime_data_module, "SwigPyPacked", (PyObject *)&swigpypacked_type) == 0)
      SWIG_Py_INCREF((PyObject *)&swigpypacked_type);
  }
  return &swigpypacked_type;
#else
  PyType_Slot slots[] = {
    { Py_tp_dealloc, (void *)SwigPyPacked_dealloc },
    { Py_tp_repr, (void *)SwigPyPacked_repr },
    { Py_tp_str, (void *)SwigPyPacked_str },
    { Py_tp_getattro, (void *)PyObject_GenericGetAttr },
    { Py_tp_doc, (void *)SwigPyPacked_doc },
    { 0, NULL }
  };
  PyType_Spec spec = {
    SWIG_RUNTIME_MODULE ".SwigPyPacked",
    sizeof(SwigPyPacked),
    0,
    Py_TPFLAGS_DEFAULT,
    slots
  };
  PyObject *pytype = PyType_FromSpec(&spec);
  PyObject *runtime_data_module = SWIG_runtime_data_module();
  if (pytype && PyModule_AddObject(runtime_data_module, "SwigPyPacked", pytype) == 0)
    SWIG_Py_INCREF(pytype);
  return (PyTypeObject *)pytype;
#endif
}

SWIGRUNTIME PyObject *
SwigPyPacked_New(void *ptr, size_t size, swig_type_info *ty)
{
  SwigPyPacked *sobj = PyObject_New(SwigPyPacked, SwigPyPacked_Type());
  if (sobj) {
    void *pack = malloc(size);
    if (pack) {
      memcpy(pack, ptr, size);
      sobj->pack = pack;
      sobj->ty   = ty;
      sobj->size = size;
    } else {
      PyObject_Free((PyObject *)sobj);
      sobj = 0;
    }
  }
  return (PyObject *) sobj;
}

SWIGRUNTIME swig_type_info *
SwigPyPacked_UnpackData(PyObject *obj, void *ptr, size_t size)
{
  if (SwigPyPacked_Check(obj)) {
    SwigPyPacked *sobj = (SwigPyPacked *)obj;
    if (sobj->size != size) return 0;
    memcpy(ptr, sobj->pack, size);
    return sobj->ty;
  } else {
    return 0;
  }
}

/* -----------------------------------------------------------------------------
 * pointers/data manipulation
 * ----------------------------------------------------------------------------- */

static PyObject *Swig_This_global = NULL;

SWIGRUNTIME PyObject *
SWIG_This(void)
{
  if (Swig_This_global == NULL)
    Swig_This_global = SWIG_Python_str_FromChar("this");
  return Swig_This_global;
}

/* #define SWIG_PYTHON_SLOW_GETSET_THIS */

/* TODO: I don't know how to implement the fast getset in Python 3 right now */
#if PY_VERSION_HEX >= 0x03000000
#define SWIG_PYTHON_SLOW_GETSET_THIS 
#endif

/* Returns a borrowed reference to the 'this' object */
SWIGRUNTIME SwigPyObject *
SWIG_Python_GetSwigThis(PyObject *pyobj) 
{
  PyObject *obj;

  if (SwigPyObject_Check(pyobj))
    return (SwigPyObject *) pyobj;

#ifdef SWIGPYTHON_BUILTIN
  (void)obj;
  if (PyWeakref_CheckProxy(pyobj)) {
#if PY_VERSION_HEX >= 0x030d0000
    if (PyWeakref_GetRef(pyobj, &pyobj) > 0)
      Py_DECREF(pyobj);
    else
      pyobj = NULL;
#else
    pyobj = PyWeakref_GetObject(pyobj);
#endif
    if (pyobj && SwigPyObject_Check(pyobj))
      return (SwigPyObject*) pyobj;
  }
  return NULL;
#else

  obj = 0;

#if !defined(SWIG_PYTHON_SLOW_GETSET_THIS)
  if (PyInstance_Check(pyobj)) {
    obj = _PyInstance_Lookup(pyobj, SWIG_This());      
  } else {
    PyObject **dictptr = _PyObject_GetDictPtr(pyobj);
    if (dictptr != NULL) {
      PyObject *dict = *dictptr;
      obj = dict ? PyDict_GetItem(dict, SWIG_This()) : 0;
    } else {
      if (PyWeakref_CheckProxy(pyobj)) {
        PyObject *wobj = PyWeakref_GET_OBJECT(pyobj);
	return wobj ? SWIG_Python_GetSwigThis(wobj) : 0;
      }
      obj = PyObject_GetAttr(pyobj, SWIG_This());
      if (obj) {
	SWIG_Py_DECREF(obj);
      } else {
	if (PyErr_Occurred()) PyErr_Clear();
	return 0;
      }
    }
  }
#else
  obj = PyObject_GetAttr(pyobj, SWIG_This());
  if (obj) {
    SWIG_Py_DECREF(obj);
  } else {
    if (PyErr_Occurred()) PyErr_Clear();
    return 0;
  }
#endif
  if (obj && !SwigPyObject_Check(obj)) {
    /* a PyObject is called 'this', try to get the 'real this'
       SwigPyObject from it */ 
    return SWIG_Python_GetSwigThis(obj);
  }
  return (SwigPyObject *)obj;
#endif
}

/* Acquire a pointer value */

SWIGRUNTIME int
SWIG_Python_AcquirePtr(PyObject *obj, int own) {
  if (own == SWIG_POINTER_OWN) {
    SwigPyObject *sobj = SWIG_Python_GetSwigThis(obj);
    if (sobj) {
      int oldown = sobj->own;
      sobj->own = own;
      return oldown;
    }
  }
  return 0;
}

/* Convert a pointer value */

SWIGRUNTIME int
SWIG_Python_ConvertPtrAndOwn(PyObject *obj, void **ptr, swig_type_info *ty, int flags, int *own) {
  int res;
  SwigPyObject *sobj;
  int implicit_conv = (flags & SWIG_POINTER_IMPLICIT_CONV) != 0;

  if (!obj)
    return SWIG_ERROR;
  if (obj == Py_None && !implicit_conv) {
    if (ptr)
      *ptr = 0;
    return (flags & SWIG_POINTER_NO_NULL) ? SWIG_NullReferenceError : SWIG_OK;
  }

  res = SWIG_ERROR;

  sobj = SWIG_Python_GetSwigThis(obj);
  if (own)
    *own = 0;
  while (sobj) {
    void *vptr = sobj->ptr;
    if (ty) {
      swig_type_info *to = sobj->ty;
      if (to == ty) {
        /* no type cast needed */
        if (ptr) *ptr = vptr;
        break;
      } else {
        swig_cast_info *tc = SWIG_TypeCheck(to->name,ty);
        if (!tc) {
          sobj = (SwigPyObject *)sobj->next;
        } else {
          if (ptr) {
            int newmemory = 0;
            *ptr = SWIG_TypeCast(tc,vptr,&newmemory);
            if (newmemory == SWIG_CAST_NEW_MEMORY) {
              assert(own); /* badly formed typemap which will lead to a memory leak - it must set and use own to delete *ptr */
              if (own)
                *own = *own | SWIG_CAST_NEW_MEMORY;
            }
          }
          break;
        }
      }
    } else {
      if (ptr) *ptr = vptr;
      break;
    }
  }
  if (sobj) {
    if (((flags & SWIG_POINTER_RELEASE) == SWIG_POINTER_RELEASE) && !sobj->own) {
      res = SWIG_ERROR_RELEASE_NOT_OWNED;
    } else {
      if (own)
        *own = *own | sobj->own;
      if (flags & SWIG_POINTER_DISOWN) {
        sobj->own = 0;
      }
      if (flags & SWIG_POINTER_CLEAR) {
        sobj->ptr = 0;
      }
      res = SWIG_OK;
    }
  } else {
    if (implicit_conv) {
      SwigPyClientData *data = ty ? (SwigPyClientData *) ty->clientdata : 0;
      if (data && !data->implicitconv) {
        PyObject *klass = data->klass;
        if (klass) {
          PyObject *impconv;
          data->implicitconv = 1; /* avoid recursion and call 'explicit' constructors*/
          impconv = SWIG_Python_CallFunctor(klass, obj);
          data->implicitconv = 0;
          if (PyErr_Occurred()) {
            PyErr_Clear();
            impconv = 0;
          }
          if (impconv) {
            SwigPyObject *iobj = SWIG_Python_GetSwigThis(impconv);
            if (iobj) {
              void *vptr;
              res = SWIG_Python_ConvertPtrAndOwn((PyObject*)iobj, &vptr, ty, 0, 0);
              if (SWIG_IsOK(res)) {
                if (ptr) {
                  *ptr = vptr;
                  /* transfer the ownership to 'ptr' */
                  iobj->own = 0;
                  res = SWIG_AddCast(res);
                  res = SWIG_AddNewMask(res);
                } else {
                  res = SWIG_AddCast(res);		    
                }
              }
            }
            SWIG_Py_DECREF(impconv);
          }
        }
      }
      if (!SWIG_IsOK(res) && obj == Py_None) {
        if (ptr)
          *ptr = 0;
        if (PyErr_Occurred())
          PyErr_Clear();
        res = SWIG_OK;
      }
    }
  }
  return res;
}

/* Convert a function ptr value */

SWIGRUNTIME int
SWIG_Python_ConvertFunctionPtr(PyObject *obj, void **ptr, swig_type_info *ty) {
  if (!PyCFunction_Check(obj)) {
    return SWIG_ConvertPtr(obj, ptr, ty, 0);
  } else {
    void *vptr = 0;
    swig_cast_info *tc;

    /* here we get the method pointer for callbacks */
#ifndef Py_LIMITED_API
    const char *doc = (((PyCFunctionObject *)obj) -> m_ml -> ml_doc);
#else
    PyObject* pystr_doc = PyObject_GetAttrString(obj, "__doc__");
    PyObject *bytes = NULL;
    const char *doc = pystr_doc ? SWIG_PyUnicode_AsUTF8AndSize(pystr_doc, NULL, &bytes) : 0;
#endif
    const char *desc = doc ? strstr(doc, "swig_ptr: ") : 0;
    if (desc)
      desc = ty ? SWIG_UnpackVoidPtr(desc + 10, &vptr, ty->name) : 0;
#ifdef Py_LIMITED_API
    SWIG_Py_XDECREF(bytes);
    SWIG_Py_XDECREF(pystr_doc);
#endif
    if (!desc)
      return SWIG_ERROR;
    tc = SWIG_TypeCheck(desc,ty);
    if (tc) {
      int newmemory = 0;
      *ptr = SWIG_TypeCast(tc,vptr,&newmemory);
      assert(!newmemory); /* newmemory handling not yet implemented */
    } else {
      return SWIG_ERROR;
    }
    return SWIG_OK;
  }
}

/* Convert a packed pointer value */

SWIGRUNTIME int
SWIG_Python_ConvertPacked(PyObject *obj, void *ptr, size_t sz, swig_type_info *ty) {
  swig_type_info *to = SwigPyPacked_UnpackData(obj, ptr, sz);
  if (!to) return SWIG_ERROR;
  if (ty) {
    if (to != ty) {
      /* check type cast? */
      swig_cast_info *tc = SWIG_TypeCheck(to->name,ty);
      if (!tc) return SWIG_ERROR;
    }
  }
  return SWIG_OK;
}  

/* -----------------------------------------------------------------------------
 * Create a new pointer object
 * ----------------------------------------------------------------------------- */

/*
  Create a new instance object, without calling __init__, and set the
  'this' attribute.
*/

SWIGRUNTIME PyObject* 
SWIG_Python_NewShadowInstance(SwigPyClientData *data, PyObject *swig_this)
{
  PyObject *inst = 0;
  PyObject *newraw = data->newraw;
  if (newraw) {
    inst = PyObject_Call(newraw, data->newargs, NULL);
    if (inst) {
#if !defined(SWIG_PYTHON_SLOW_GETSET_THIS)
      PyObject **dictptr = _PyObject_GetDictPtr(inst);
      if (dictptr != NULL) {
        PyObject *dict = *dictptr;
        if (dict == NULL) {
          dict = PyDict_New();
          *dictptr = dict;
        }
        if (dict) {
          PyDict_SetItem(dict, SWIG_This(), swig_this);
        } else{
          SWIG_Py_DECREF(inst);
          inst = 0;
        }
      }
#else
      if (PyObject_SetAttr(inst, SWIG_This(), swig_this) == -1) {
        SWIG_Py_DECREF(inst);
        inst = 0;
      }
#endif
    }
  } else {
#if PY_VERSION_HEX >= 0x03000000
    PyObject *empty_args = PyTuple_New(0);
    if (empty_args) {
      PyObject *empty_kwargs = PyDict_New();
      if (empty_kwargs) {
#ifndef Py_LIMITED_API
        newfunc newfn = ((PyTypeObject *)data->newargs)->tp_new;
#else
        newfunc newfn = (newfunc)PyType_GetSlot((PyTypeObject *)data->newargs, Py_tp_new);
#endif
        inst = newfn((PyTypeObject *)data->newargs, empty_args, empty_kwargs);
        SWIG_Py_DECREF(empty_kwargs);
        if (inst) {
          if (PyObject_SetAttr(inst, SWIG_This(), swig_this) == -1) {
            SWIG_Py_DECREF(inst);
            inst = 0;
          } else {
            PyType_Modified(Py_TYPE(inst));
          }
        }
      }
      SWIG_Py_DECREF(empty_args);
    }
#else
    PyObject *dict = PyDict_New();
    if (dict) {
      PyDict_SetItem(dict, SWIG_This(), swig_this);
      inst = PyInstance_NewRaw(data->newargs, dict);
      SWIG_Py_DECREF(dict);
    }
#endif
  }
  return inst;
}

SWIGRUNTIME int
SWIG_Python_SetSwigThis(PyObject *inst, PyObject *swig_this)
{
#if !defined(SWIG_PYTHON_SLOW_GETSET_THIS)
  PyObject **dictptr = _PyObject_GetDictPtr(inst);
  if (dictptr != NULL) {
    PyObject *dict = *dictptr;
    if (dict == NULL) {
      dict = PyDict_New();
      *dictptr = dict;
    }
    if (dict) {
      return PyDict_SetItem(dict, SWIG_This(), swig_this);
    } else{
      return -1;
    }
  }
#endif
  return PyObject_SetAttr(inst, SWIG_This(), swig_this);
} 


SWIGINTERN PyObject *
SWIG_Python_InitShadowInstance(PyObject *args) {
  PyObject *obj[2];
  if (!SWIG_Python_UnpackTuple(args, "swiginit", 2, 2, obj)) {
    return NULL;
  } else {
    SwigPyObject *sthis = SWIG_Python_GetSwigThis(obj[0]);
    if (sthis) {
      SWIG_Py_DECREF(SwigPyObject_append((PyObject*) sthis, obj[1]));
    } else {
      if (SWIG_Python_SetSwigThis(obj[0], obj[1]) != 0)
        return NULL;
    }
    return SWIG_Py_Void();
  }
}

/* Create a new pointer object */

SWIGRUNTIME PyObject *
SWIG_Python_NewPointerObj(PyObject *self, void *ptr, swig_type_info *type, int flags) {
  SwigPyClientData *clientdata;
  PyObject * robj;
  int own;

  if (!ptr)
    return SWIG_Py_Void();

  clientdata = type ? (SwigPyClientData *)(type->clientdata) : 0;
  own = (flags & SWIG_POINTER_OWN) ? SWIG_POINTER_OWN : 0;
  if (clientdata && clientdata->pytype) {
    SwigPyObject *newobj;
    if (flags & SWIG_BUILTIN_TP_INIT) {
      newobj = (SwigPyObject*) self;
      if (newobj->ptr) {
#ifndef Py_LIMITED_API
        allocfunc alloc = clientdata->pytype->tp_alloc;
#else
        allocfunc alloc = (allocfunc)PyType_GetSlot(clientdata->pytype, Py_tp_alloc);
#endif
        PyObject *next_self = alloc(clientdata->pytype, 0);
        while (newobj->next)
	  newobj = (SwigPyObject *) newobj->next;
        newobj->next = next_self;
        newobj = (SwigPyObject *)next_self;
        newobj->swigdict = 0;
        newobj->weakreflist = 0;
      }
    } else {
      newobj = PyObject_New(SwigPyObject, clientdata->pytype);
      if (newobj) {
        newobj->swigdict = 0;
        newobj->weakreflist = 0;
      }
    }
    if (newobj) {
      newobj->ptr = ptr;
      newobj->ty = type;
      newobj->own = own;
      newobj->next = 0;
      return (PyObject*) newobj;
    }
    return SWIG_Py_Void();
  }

  assert(!(flags & SWIG_BUILTIN_TP_INIT));

  robj = SwigPyObject_New(ptr, type, own);
  if (robj && clientdata && !(flags & SWIG_POINTER_NOSHADOW)) {
    PyObject *inst = SWIG_Python_NewShadowInstance(clientdata, robj);
    SWIG_Py_DECREF(robj);
    robj = inst;
  }
  return robj;
}

/* Create a new packed object */

SWIGRUNTIMEINLINE PyObject *
SWIG_Python_NewPackedObj(void *ptr, size_t sz, swig_type_info *type) {
  return ptr ? SwigPyPacked_New((void *) ptr, sz, type) : SWIG_Py_Void();
}

/* -----------------------------------------------------------------------------*
 *  Get type list 
 * -----------------------------------------------------------------------------*/

#ifdef SWIG_LINK_RUNTIME
void *SWIG_ReturnGlobalTypeList(void *);
#endif

static PyObject *Swig_TypeCache_global = NULL;

/* The python cached type query */
SWIGRUNTIME PyObject *
SWIG_Python_TypeCache(void) {
  if (Swig_TypeCache_global == NULL) {
    Swig_TypeCache_global = PyDict_New();
  }
  return Swig_TypeCache_global;
}

SWIGRUNTIME swig_module_info *
SWIG_Python_GetModule(void *SWIGUNUSEDPARM(clientdata)) {
#ifdef SWIG_LINK_RUNTIME
  static void *type_pointer = (void *)0;
  /* first check if module already created */
  if (!type_pointer) {
    type_pointer = SWIG_ReturnGlobalTypeList((void *)0);
  }
#else
  void *type_pointer = PyCapsule_Import(SWIGPY_CAPSULE_NAME, 0);
  if (PyErr_Occurred()) {
    PyErr_Clear();
    type_pointer = (void *)0;
  }
#endif
  return (swig_module_info *) type_pointer;
}


#if defined(SWIG_REFCNT_DEBUG)
#define SWIG_PYOBJ_REFCNT(OBJ) fprintf(stdout, "" #OBJ " count %ld\n", (OBJ ? Py_REFCNT(OBJ) : 0))
#else
#define SWIG_PYOBJ_REFCNT(OBJ)
#endif

static int interpreter_counter = 0; /* how many (sub-)interpreters are using swig_module's types */

SWIGRUNTIME void
SWIG_Python_DestroyModule(PyObject *obj)
{
  swig_module_info *swig_module = (swig_module_info *) PyCapsule_GetPointer(obj, SWIGPY_CAPSULE_NAME);
  swig_type_info **types = swig_module->types;
  size_t i;
  if (--interpreter_counter != 0) /* another sub-interpreter may still be using the swig_module's types */
    return;
  for (i = 0; i < swig_module->size; ++i) {
    swig_type_info *ty = types[i];
    if (ty->owndata) {
      SwigPyClientData *data = (SwigPyClientData *) ty->clientdata;
      ty->clientdata = 0;
      if (data) SwigPyClientData_Del(data);
    }
  }
  SWIG_PYOBJ_REFCNT(Swig_This_global);
  SWIG_Py_XDECREF(Swig_This_global);
  Swig_This_global = NULL;

  SWIG_PYOBJ_REFCNT(Swig_Globals_global);
  SWIG_Py_XDECREF(Swig_Globals_global);
  Swig_Globals_global = NULL;

  SWIG_PYOBJ_REFCNT(Swig_TypeCache_global);
  SWIG_Py_XDECREF(Swig_TypeCache_global);
  Swig_TypeCache_global = NULL;

  SWIG_PYOBJ_REFCNT(Swig_Capsule_global);
  Swig_Capsule_global = NULL;

  SWIG_PYOBJ_REFCNT(Swig_runtime_data_module_global);
  SWIG_Py_XDECREF(Swig_runtime_data_module_global);
  Swig_runtime_data_module_global = NULL;
}

SWIGRUNTIME void
SWIG_Python_SetModule(swig_module_info *swig_module) {
  PyObject *runtime_data_module = SWIG_runtime_data_module();
  PyObject *pointer = PyCapsule_New((void *) swig_module, SWIGPY_CAPSULE_NAME, SWIG_Python_DestroyModule);
  if (pointer && runtime_data_module) {
    if (PyModule_AddObject(runtime_data_module, SWIGPY_CAPSULE_ATTR_NAME, pointer) == 0) {
      ++interpreter_counter;
      Swig_Capsule_global = pointer;
    } else {
      SWIG_Py_DECREF(pointer);
    }
  } else {
    SWIG_Py_XDECREF(pointer);
  }
}

SWIGRUNTIME swig_type_info *
SWIG_Python_TypeQuery(const char *type)
{
  swig_type_info *descriptor;
  PyObject *cache = SWIG_Python_TypeCache();
  PyObject *obj;
  SWIG_PyDict_GetItemStringRef(cache, type, &obj);
  if (obj) {
    descriptor = (swig_type_info *) PyCapsule_GetPointer(obj, NULL);
  } else {
    swig_module_info *swig_module = SWIG_GetModule(0);
    descriptor = SWIG_TypeQueryModule(swig_module, swig_module, type);
    if (descriptor) {
      obj = PyCapsule_New((void*) descriptor, NULL, NULL);
      if (obj) PyDict_SetItemString(cache, type, obj);
    }
  }
  SWIG_Py_XDECREF(obj);
  return descriptor;
}

/* 
   For backward compatibility only
*/
#define SWIG_POINTER_EXCEPTION  0
#define SWIG_arg_fail(arg)      SWIG_Python_ArgFail(arg)
#define SWIG_MustGetPtr(p, type, argnum, flags)  SWIG_Python_MustGetPtr(p, type, argnum, flags)

SWIGRUNTIME int
SWIG_Python_AddErrMesg(const char* mesg, int infront)
{  
  if (PyErr_Occurred()) {
    PyObject *type = 0;
    PyObject *value = 0;
    PyObject *traceback = 0;
    PyErr_Fetch(&type, &value, &traceback);
    if (value) {
      PyObject *old_str = PyObject_Str(value);
      PyObject *bytes = NULL;
      const char *tmp = SWIG_PyUnicode_AsUTF8AndSize(old_str, NULL, &bytes);
      const char *errmesg = tmp ? tmp : "Invalid error message";
      SWIG_Py_XINCREF(type);
      PyErr_Clear();
      if (infront) {
	PyErr_Format(type, "%s %s", mesg, errmesg);
      } else {
	PyErr_Format(type, "%s %s", errmesg, mesg);
      }
      SWIG_Py_XDECREF(bytes);
      SWIG_Py_DECREF(old_str);
    }
    return 1;
  } else {
    return 0;
  }
}
  
SWIGRUNTIME int
SWIG_Python_ArgFail(int argnum)
{
  if (PyErr_Occurred()) {
    /* add information about failing argument */
    char mesg[256];
    PyOS_snprintf(mesg, sizeof(mesg), "argument number %d:", argnum);
    return SWIG_Python_AddErrMesg(mesg, 1);
  } else {
    return 0;
  }
}

SWIGRUNTIMEINLINE const char *
SwigPyObject_GetDesc(PyObject *self)
{
  SwigPyObject *v = (SwigPyObject *)self;
  swig_type_info *ty = v ? v->ty : 0;
  return ty ? ty->str : "";
}

/* Convert a pointer value, signal an exception on a type mismatch */
SWIGRUNTIME void *
SWIG_Python_MustGetPtr(PyObject *obj, swig_type_info *ty, int SWIGUNUSEDPARM(argnum), int flags) {
  void *result;
  if (SWIG_Python_ConvertPtr(obj, &result, ty, flags) == -1) {
    PyErr_Clear();
  }
  return result;
}

#ifdef SWIGPYTHON_BUILTIN
SWIGRUNTIME int
SWIG_Python_NonDynamicSetAttr(PyObject *obj, PyObject *name, PyObject *value) {
  PyTypeObject *tp = Py_TYPE(obj);
  PyObject *descr;
  PyObject *encoded_name;
  descrsetfunc f;
  int res = -1;

# ifdef Py_USING_UNICODE
  if (PyString_Check(name)) {
    name = PyUnicode_Decode(PyString_AsString(name), PyString_Size(name), NULL, NULL);
    if (!name)
      return -1;
  } else if (!PyUnicode_Check(name))
# else
  if (!PyString_Check(name))
# endif
  {
#if PY_VERSION_HEX >= 0x03000000
    PyObject *tpname = SWIG_PyType_GetFullyQualifiedName(Py_TYPE(name));
    PyErr_Format(PyExc_TypeError, "attribute name must be string, not '%S'", tpname);
    SWIG_Py_DECREF(tpname);
#else
    PyErr_Format(PyExc_TypeError, "attribute name must be string, not '%s'", Py_TYPE(name)->tp_name);
#endif
    return -1;
  } else {
    SWIG_Py_INCREF(name);
  }

  if (!tp->tp_dict) {
    if (PyType_Ready(tp) != 0)
      goto done;
  }

  descr = _PyType_Lookup(tp, name);
  f = NULL;
  if (descr != NULL)
    f = Py_TYPE(descr)->tp_descr_set;
  if (!f) {
    if (PyString_Check(name)) {
      encoded_name = name;
      SWIG_Py_INCREF(name);
    } else {
      encoded_name = PyUnicode_AsUTF8String(name);
      if (!encoded_name)
        goto done;
    }
#if PY_VERSION_HEX >= 0x03000000
    {
      PyObject *tpname = SWIG_PyType_GetFullyQualifiedName(tp);
      PyErr_Format(PyExc_AttributeError, "'%S' object has no attribute '%s'", tpname, PyString_AsString(encoded_name));
      SWIG_Py_DECREF(tpname);
    }
#else
    PyErr_Format(PyExc_AttributeError, "'%s' object has no attribute '%s'", tp->tp_name, PyString_AsString(encoded_name));
#endif
    SWIG_Py_DECREF(encoded_name);
  } else {
    res = f(descr, obj, value);
  }
  
  done:
  SWIG_Py_DECREF(name);
  return res;
}
#endif


#ifdef __cplusplus
}
#endif



#define SWIG_exception_fail(code, msg) do { SWIG_Error(code, msg); SWIG_fail; } while(0) 

#define SWIG_contract_assert(expr, msg) do { if (!(expr)) { SWIG_Error(SWIG_RuntimeError, msg); SWIG_fail; } } while (0) 



  #define SWIG_exception(code, msg) do { SWIG_Error(code, msg); SWIG_fail;; } while(0) 


/* -------- TYPES TABLE (BEGIN) -------- */

#define SWIGTYPE_p_apr_array_header_t swig_types[0]
#define SWIGTYPE_p_apr_file_t swig_types[1]
#define SWIGTYPE_p_apr_hash_t swig_types[2]
#define SWIGTYPE_p_apr_int32_t swig_types[3]
#define SWIGTYPE_p_apr_int64_t swig_types[4]
#define SWIGTYPE_p_apr_off_t swig_types[5]
#define SWIGTYPE_p_apr_pool_t swig_types[6]
#define SWIGTYPE_p_char swig_types[7]
#define SWIGTYPE_p_f_apr_off_t_apr_off_t_p_void_p_apr_pool_t__void swig_types[8]
#define SWIGTYPE_p_f_int_p_apr_pool_t_p_apr_hash_t__p_svn_error_t swig_types[9]
#define SWIGTYPE_p_f_p_apr_getopt_t_p_void_p_apr_pool_t__p_svn_error_t swig_types[10]
#define SWIGTYPE_p_f_p_p_apr_file_t_p_void_p_apr_pool_t__p_svn_error_t swig_types[11]
#define SWIGTYPE_p_f_p_p_svn_stream_t_p_p_svn_stream_t_p_svn_ra_close_tunnel_func_t_p_p_void_p_void_p_q_const__char_p_q_const__char_p_q_const__char_int_svn_cancel_func_t_p_void_p_apr_pool_t__p_svn_error_t swig_types[12]
#define SWIGTYPE_p_f_p_p_void_p_q_const__char_p_q_const__svn_ra_callbacks_t_p_void_p_apr_hash_t_p_apr_pool_t__p_svn_error_t swig_types[13]
#define SWIGTYPE_p_f_p_q_const__char_p_svn_dirent_t_p_void_p_apr_pool_t__p_svn_error_t swig_types[14]
#define SWIGTYPE_p_f_p_q_const__svn_commit_info_t_p_void_p_apr_pool_t__p_svn_error_t swig_types[15]
#define SWIGTYPE_p_f_p_svn_location_segment_t_p_void_p_apr_pool_t__p_svn_error_t swig_types[16]
#define SWIGTYPE_p_f_p_void__p_svn_error_t swig_types[17]
#define SWIGTYPE_p_f_p_void_p_apr_hash_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t swig_types[18]
#define SWIGTYPE_p_f_p_void_p_apr_pool_t__p_svn_error_t swig_types[19]
#define SWIGTYPE_p_f_p_void_p_p_apr_hash_t_p_q_const__char_svn_revnum_t_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t swig_types[20]
#define SWIGTYPE_p_f_p_void_p_p_q_const__char_p_apr_pool_t__p_svn_error_t swig_types[21]
#define SWIGTYPE_p_f_p_void_p_p_q_const__svn_delta_editor_t_p_p_void_p_q_const__char_svn_commit_callback_t_p_void_p_apr_pool_t__p_svn_error_t swig_types[22]
#define SWIGTYPE_p_f_p_void_p_p_q_const__svn_ra_reporter_t_p_p_void_p_q_const__char_svn_revnum_t_svn_boolean_t_p_q_const__svn_delta_editor_t_p_void_p_apr_pool_t__p_svn_error_t swig_types[23]
#define SWIGTYPE_p_f_p_void_p_p_q_const__svn_ra_reporter_t_p_p_void_svn_revnum_t_p_q_const__char_svn_boolean_t_p_q_const__char_p_q_const__svn_delta_editor_t_p_void_p_apr_pool_t__p_svn_error_t swig_types[24]
#define SWIGTYPE_p_f_p_void_p_p_q_const__svn_ra_reporter_t_p_p_void_svn_revnum_t_p_q_const__char_svn_boolean_t_p_q_const__svn_delta_editor_t_p_void_p_apr_pool_t__p_svn_error_t swig_types[25]
#define SWIGTYPE_p_f_p_void_p_p_q_const__svn_ra_reporter_t_p_p_void_svn_revnum_t_p_q_const__char_svn_boolean_t_svn_boolean_t_p_q_const__char_p_q_const__svn_delta_editor_t_p_void_p_apr_pool_t__p_svn_error_t swig_types[26]
#define SWIGTYPE_p_f_p_void_p_p_svn_stream_t_p_q_const__svn_checksum_t_p_apr_pool_t__p_svn_error_t swig_types[27]
#define SWIGTYPE_p_f_p_void_p_q_const__apr_array_header_t_svn_revnum_t_svn_revnum_t_svn_boolean_t_svn_boolean_t_svn_log_message_receiver_t_p_void_p_apr_pool_t__p_svn_error_t swig_types[28]
#define SWIGTYPE_p_f_p_void_p_q_const__char__int swig_types[29]
#define SWIGTYPE_p_f_p_void_p_q_const__char_p_apr_pool_t__p_svn_error_t swig_types[30]
#define SWIGTYPE_p_f_p_void_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t swig_types[31]
#define SWIGTYPE_p_f_p_void_p_q_const__char_p_q_const__char_p_p_q_const__svn_string_t_p_apr_pool_t__p_svn_error_t swig_types[32]
#define SWIGTYPE_p_f_p_void_p_q_const__char_p_q_const__char_p_q_const__svn_string_t_p_apr_pool_t__p_svn_error_t swig_types[33]
#define SWIGTYPE_p_f_p_void_p_q_const__char_p_q_const__char_svn_revnum_t_svn_boolean_t_p_apr_pool_t__p_svn_error_t swig_types[34]
#define SWIGTYPE_p_f_p_void_p_q_const__char_p_q_const__char_svn_revnum_t_svn_boolean_t_p_q_const__char_p_apr_pool_t__p_svn_error_t swig_types[35]
#define SWIGTYPE_p_f_p_void_p_q_const__char_p_q_const__char_svn_revnum_t_svn_depth_t_svn_boolean_t_p_q_const__char_p_apr_pool_t__p_svn_error_t swig_types[36]
#define SWIGTYPE_p_f_p_void_p_q_const__char_svn_boolean_t_p_q_const__svn_lock_t_p_svn_error_t_p_apr_pool_t__p_svn_error_t swig_types[37]
#define SWIGTYPE_p_f_p_void_p_q_const__char_svn_revnum_t_p_apr_hash_t_p_svn_txdelta_window_handler_t_p_p_void_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t swig_types[38]
#define SWIGTYPE_p_f_p_void_p_q_const__char_svn_revnum_t_p_apr_hash_t_svn_boolean_t_p_svn_txdelta_window_handler_t_p_p_void_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t swig_types[39]
#define SWIGTYPE_p_f_p_void_p_q_const__char_svn_revnum_t_p_p_apr_hash_t_p_svn_revnum_t_p_p_apr_hash_t_p_apr_pool_t__p_svn_error_t swig_types[40]
#define SWIGTYPE_p_f_p_void_p_q_const__char_svn_revnum_t_p_svn_node_kind_t_p_apr_pool_t__p_svn_error_t swig_types[41]
#define SWIGTYPE_p_f_p_void_p_q_const__char_svn_revnum_t_p_svn_stream_t_p_svn_revnum_t_p_p_apr_hash_t_p_apr_pool_t__p_svn_error_t swig_types[42]
#define SWIGTYPE_p_f_p_void_p_q_const__char_svn_revnum_t_svn_boolean_t_p_apr_pool_t__p_svn_error_t swig_types[43]
#define SWIGTYPE_p_f_p_void_p_q_const__char_svn_revnum_t_svn_boolean_t_p_q_const__char_p_apr_pool_t__p_svn_error_t swig_types[44]
#define SWIGTYPE_p_f_p_void_p_q_const__char_svn_revnum_t_svn_depth_t_svn_boolean_t_p_q_const__char_p_apr_pool_t__p_svn_error_t swig_types[45]
#define SWIGTYPE_p_f_p_void_p_q_const__char_svn_revnum_t_svn_revnum_t_svn_ra_file_rev_handler_t_p_void_p_apr_pool_t__p_svn_error_t swig_types[46]
#define SWIGTYPE_p_f_p_void_p_svn_log_entry_t_p_apr_pool_t__p_svn_error_t swig_types[47]
#define SWIGTYPE_p_f_p_void_p_svn_revnum_t__p_svn_error_t swig_types[48]
#define SWIGTYPE_p_f_p_void_p_svn_revnum_t_apr_time_t_p_apr_pool_t__p_svn_error_t swig_types[49]
#define SWIGTYPE_p_f_p_void_p_svn_revnum_t_p_apr_pool_t__p_svn_error_t swig_types[50]
#define SWIGTYPE_p_f_p_void_p_void__void swig_types[51]
#define SWIGTYPE_p_f_p_void_svn_revnum_t_p_p_apr_hash_t_p_apr_pool_t__p_svn_error_t swig_types[52]
#define SWIGTYPE_p_f_p_void_svn_revnum_t_p_q_const__char_p_p_svn_string_t_p_apr_pool_t__p_svn_error_t swig_types[53]
#define SWIGTYPE_p_f_p_void_svn_revnum_t_p_q_const__char_p_q_const__svn_string_t_p_apr_pool_t__p_svn_error_t swig_types[54]
#define SWIGTYPE_p_f_svn_revnum_t_p_q_const__char_p_q_const__char_p_void__p_svn_error_t swig_types[55]
#define SWIGTYPE_p_f_svn_revnum_t_p_void_p_p_q_const__svn_delta_editor_t_p_p_void_p_apr_hash_t_p_apr_pool_t__p_svn_error_t swig_types[56]
#define SWIGTYPE_p_f_svn_revnum_t_p_void_p_q_const__svn_delta_editor_t_p_void_p_apr_hash_t_p_apr_pool_t__p_svn_error_t swig_types[57]
#define SWIGTYPE_p_f_void__p_svn_version_t swig_types[58]
#define SWIGTYPE_p_int swig_types[59]
#define SWIGTYPE_p_long swig_types[60]
#define SWIGTYPE_p_p_apr_array_header_t swig_types[61]
#define SWIGTYPE_p_p_apr_file_t swig_types[62]
#define SWIGTYPE_p_p_apr_hash_t swig_types[63]
#define SWIGTYPE_p_p_char swig_types[64]
#define SWIGTYPE_p_p_f_apr_off_t_apr_off_t_p_void_p_apr_pool_t__void swig_types[65]
#define SWIGTYPE_p_p_f_int_p_apr_pool_t_p_apr_hash_t__p_svn_error_t swig_types[66]
#define SWIGTYPE_p_p_f_p_p_svn_stream_t_p_p_svn_stream_t_p_svn_ra_close_tunnel_func_t_p_p_void_p_void_p_q_const__char_p_q_const__char_p_q_const__char_int_svn_cancel_func_t_p_void_p_apr_pool_t__p_svn_error_t swig_types[67]
#define SWIGTYPE_p_p_f_p_svn_txdelta_window_t_p_void__p_svn_error_t swig_types[68]
#define SWIGTYPE_p_p_f_p_void__p_svn_error_t swig_types[69]
#define SWIGTYPE_p_p_f_p_void_p_p_q_const__char_p_apr_pool_t__p_svn_error_t swig_types[70]
#define SWIGTYPE_p_p_f_p_void_p_p_svn_stream_t_p_q_const__svn_checksum_t_p_apr_pool_t__p_svn_error_t swig_types[71]
#define SWIGTYPE_p_p_f_p_void_p_q_const__char__int swig_types[72]
#define SWIGTYPE_p_p_f_p_void_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t swig_types[73]
#define SWIGTYPE_p_p_f_p_void_p_q_const__char_p_q_const__char_p_p_q_const__svn_string_t_p_apr_pool_t__p_svn_error_t swig_types[74]
#define SWIGTYPE_p_p_f_p_void_p_q_const__char_p_q_const__char_p_q_const__svn_string_t_p_apr_pool_t__p_svn_error_t swig_types[75]
#define SWIGTYPE_p_p_f_p_void_p_q_const__char_svn_boolean_t_p_q_const__svn_lock_t_p_svn_error_t_p_apr_pool_t__p_svn_error_t swig_types[76]
#define SWIGTYPE_p_p_f_p_void_p_q_const__char_svn_revnum_t_p_apr_hash_t_p_svn_txdelta_window_handler_t_p_p_void_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t swig_types[77]
#define SWIGTYPE_p_p_f_p_void_p_q_const__char_svn_revnum_t_p_apr_hash_t_svn_boolean_t_p_svn_txdelta_window_handler_t_p_p_void_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t swig_types[78]
#define SWIGTYPE_p_p_f_p_void_p_svn_revnum_t__p_svn_error_t swig_types[79]
#define SWIGTYPE_p_p_f_p_void_p_void__void swig_types[80]
#define SWIGTYPE_p_p_f_svn_revnum_t_p_void_p_p_q_const__svn_delta_editor_t_p_p_void_p_apr_hash_t_p_apr_pool_t__p_svn_error_t swig_types[81]
#define SWIGTYPE_p_p_f_svn_revnum_t_p_void_p_q_const__svn_delta_editor_t_p_void_p_apr_hash_t_p_apr_pool_t__p_svn_error_t swig_types[82]
#define SWIGTYPE_p_p_svn_delta_editor_t swig_types[83]
#define SWIGTYPE_p_p_svn_dirent_t swig_types[84]
#define SWIGTYPE_p_p_svn_lock_t swig_types[85]
#define SWIGTYPE_p_p_svn_ra_callbacks2_t swig_types[86]
#define SWIGTYPE_p_p_svn_ra_plugin_t swig_types[87]
#define SWIGTYPE_p_p_svn_ra_reporter2_t swig_types[88]
#define SWIGTYPE_p_p_svn_ra_reporter3_t swig_types[89]
#define SWIGTYPE_p_p_svn_ra_reporter_t swig_types[90]
#define SWIGTYPE_p_p_svn_ra_session_t swig_types[91]
#define SWIGTYPE_p_p_svn_stream_t swig_types[92]
#define SWIGTYPE_p_p_svn_string_t swig_types[93]
#define SWIGTYPE_p_p_svn_stringbuf_t swig_types[94]
#define SWIGTYPE_p_p_void swig_types[95]
#define SWIGTYPE_p_svn_auth_baton_t swig_types[96]
#define SWIGTYPE_p_svn_auth_cred_simple_t swig_types[97]
#define SWIGTYPE_p_svn_auth_cred_ssl_client_cert_pw_t swig_types[98]
#define SWIGTYPE_p_svn_auth_cred_ssl_client_cert_t swig_types[99]
#define SWIGTYPE_p_svn_auth_cred_ssl_server_trust_t swig_types[100]
#define SWIGTYPE_p_svn_auth_cred_username_t swig_types[101]
#define SWIGTYPE_p_svn_auth_iterstate_t swig_types[102]
#define SWIGTYPE_p_svn_auth_provider_object_t swig_types[103]
#define SWIGTYPE_p_svn_auth_provider_t swig_types[104]
#define SWIGTYPE_p_svn_auth_ssl_server_cert_info_t swig_types[105]
#define SWIGTYPE_p_svn_cache_config_t swig_types[106]
#define SWIGTYPE_p_svn_checksum_ctx_t swig_types[107]
#define SWIGTYPE_p_svn_checksum_kind_t swig_types[108]
#define SWIGTYPE_p_svn_checksum_t swig_types[109]
#define SWIGTYPE_p_svn_client__shelf_version_t swig_types[110]
#define SWIGTYPE_p_svn_client_conflict_option_t swig_types[111]
#define SWIGTYPE_p_svn_commit_info_t swig_types[112]
#define SWIGTYPE_p_svn_config_t swig_types[113]
#define SWIGTYPE_p_svn_delta_editor_t swig_types[114]
#define SWIGTYPE_p_svn_delta_path_driver_state_t swig_types[115]
#define SWIGTYPE_p_svn_depth_t swig_types[116]
#define SWIGTYPE_p_svn_diff_hunk_t swig_types[117]
#define SWIGTYPE_p_svn_dirent_t swig_types[118]
#define SWIGTYPE_p_svn_errno_t swig_types[119]
#define SWIGTYPE_p_svn_error_t swig_types[120]
#define SWIGTYPE_p_svn_fs_lock_target_t swig_types[121]
#define SWIGTYPE_p_svn_io_dirent2_t swig_types[122]
#define SWIGTYPE_p_svn_io_dirent_t swig_types[123]
#define SWIGTYPE_p_svn_io_file_del_t swig_types[124]
#define SWIGTYPE_p_svn_location_segment_t swig_types[125]
#define SWIGTYPE_p_svn_lock_t swig_types[126]
#define SWIGTYPE_p_svn_log_changed_path2_t swig_types[127]
#define SWIGTYPE_p_svn_log_changed_path_t swig_types[128]
#define SWIGTYPE_p_svn_log_entry_t swig_types[129]
#define SWIGTYPE_p_svn_merge_range_t swig_types[130]
#define SWIGTYPE_p_svn_mergeinfo_inheritance_t swig_types[131]
#define SWIGTYPE_p_svn_node_kind_t swig_types[132]
#define SWIGTYPE_p_svn_opt_revision_range_t swig_types[133]
#define SWIGTYPE_p_svn_opt_revision_t swig_types[134]
#define SWIGTYPE_p_svn_opt_revision_value_t swig_types[135]
#define SWIGTYPE_p_svn_opt_subcommand_desc2_t swig_types[136]
#define SWIGTYPE_p_svn_opt_subcommand_desc3_t swig_types[137]
#define SWIGTYPE_p_svn_opt_subcommand_desc_t swig_types[138]
#define SWIGTYPE_p_svn_prop_inherited_item_t swig_types[139]
#define SWIGTYPE_p_svn_prop_kind swig_types[140]
#define SWIGTYPE_p_svn_ra_callbacks2_t swig_types[141]
#define SWIGTYPE_p_svn_ra_callbacks_t swig_types[142]
#define SWIGTYPE_p_svn_ra_plugin_t swig_types[143]
#define SWIGTYPE_p_svn_ra_reporter2_t swig_types[144]
#define SWIGTYPE_p_svn_ra_reporter3_t swig_types[145]
#define SWIGTYPE_p_svn_ra_reporter_t swig_types[146]
#define SWIGTYPE_p_svn_ra_session_t swig_types[147]
#define SWIGTYPE_p_svn_stream_mark_t swig_types[148]
#define SWIGTYPE_p_svn_stream_t swig_types[149]
#define SWIGTYPE_p_svn_string_t swig_types[150]
#define SWIGTYPE_p_svn_stringbuf_t swig_types[151]
#define SWIGTYPE_p_svn_tristate_t swig_types[152]
#define SWIGTYPE_p_svn_txdelta_op_t swig_types[153]
#define SWIGTYPE_p_svn_txdelta_stream_t swig_types[154]
#define SWIGTYPE_p_svn_txdelta_window_t swig_types[155]
#define SWIGTYPE_p_svn_version_checklist_t swig_types[156]
#define SWIGTYPE_p_svn_version_ext_linked_lib_t swig_types[157]
#define SWIGTYPE_p_svn_version_ext_loaded_lib_t swig_types[158]
#define SWIGTYPE_p_svn_version_extended_t swig_types[159]
#define SWIGTYPE_p_svn_version_t swig_types[160]
#define SWIGTYPE_p_svn_wc_external_item2_t swig_types[161]
#define SWIGTYPE_p_unsigned_long swig_types[162]
#define SWIGTYPE_p_void swig_types[163]
static swig_type_info *swig_types[165];
static swig_module_info swig_module = {swig_types, 164, 0, 0, 0, 0};
#define SWIG_TypeQuery(name) SWIG_TypeQueryModule(&swig_module, &swig_module, name)
#define SWIG_MangledTypeQuery(name) SWIG_MangledTypeQueryModule(&swig_module, &swig_module, name)

/* -------- TYPES TABLE (END) -------- */

#ifdef SWIG_TypeQuery
# undef SWIG_TypeQuery
#endif
#define SWIG_TypeQuery SWIG_Python_TypeQuery

/*-----------------------------------------------
              @(target):= _ra.so
  ------------------------------------------------*/
#if PY_VERSION_HEX >= 0x03000000
#  define SWIG_init    PyInit__ra

#else
#  define SWIG_init    init_ra

#endif

#define SWIG_as_voidptr(a) (void *)((const void *)(a)) 
#define SWIG_as_voidptrptr(a) ((void)SWIG_as_voidptr(*a),(void**)(a)) 


#include "svn_time.h"
#include "svn_pools.h"


#include "swigutil_py.h"
#include "swigutil_py3c.h"


static PyObject * _global_py_pool = NULL;


#define svn_argnum_obj0 1
#define svn_argnum_obj1 2
#define svn_argnum_obj2 3
#define svn_argnum_obj3 4
#define svn_argnum_obj4 5
#define svn_argnum_obj5 6
#define svn_argnum_obj6 7
#define svn_argnum_obj7 8
#define svn_argnum_obj8 9
#define svn_argnum_obj9 10
#define svn_argnum_obj10 11
#define svn_argnum_obj11 12
#define svn_argnum_obj12 13
#define svn_argnum_obj13 14
#define svn_argnum_obj14 15
#define svn_argnum_obj15 16
#define svn_argnum_obj16 17
#define svn_argnum_obj17 18
#define svn_argnum_obj18 19
#define svn_argnum_obj19 20
#define svn_argnum_obj20 21
#define svn_argnum_obj21 22
#define svn_argnum_obj22 23
#define svn_argnum_obj23 24
#define svn_argnum_obj24 25
#define svn_argnum_obj25 26
#define svn_argnum_obj26 27
#define svn_argnum_obj27 28
#define svn_argnum_obj28 29
#define svn_argnum_obj29 30
#define svn_argnum_obj30 31
#define svn_argnum_obj31 32
#define svn_argnum_obj32 33
#define svn_argnum_obj33 34
#define svn_argnum_obj34 35
#define svn_argnum_obj35 36
#define svn_argnum_obj36 37
#define svn_argnum_obj37 38
#define svn_argnum_obj38 39
#define svn_argnum_obj39 40


#include "svn_ra.h"


SWIGINTERN swig_type_info*
SWIG_pchar_descriptor(void)
{
  static swig_type_info* info = 0;
  if (!info) {
    info = SWIG_TypeQuery("_p_char");
  }
  return info;
}


/* Return string from Python obj. NOTE: obj must remain in scope in order
   to use the returned cptr (but only when alloc is set to SWIG_OLDOBJ) */
SWIGINTERN int
SWIG_AsCharPtrAndSize(PyObject *obj, char **cptr, size_t *psize, int *alloc)
{
#if PY_VERSION_HEX >= 0x03000000
#if defined(SWIG_PYTHON_STRICT_BYTE_CHAR)
  if (PyBytes_Check(obj))
#else
  if (PyUnicode_Check(obj))
#endif
#else  
  if (PyString_Check(obj))
#endif
  {
    char *cstr; Py_ssize_t len;
    PyObject *bytes = NULL;
    int ret = SWIG_OK;
    if (alloc)
      *alloc = SWIG_OLDOBJ;
#if PY_VERSION_HEX >= 0x03000000 && defined(SWIG_PYTHON_STRICT_BYTE_CHAR)
    if (PyBytes_AsStringAndSize(obj, &cstr, &len) == -1)
      return SWIG_TypeError;
#else
    cstr = (char *)SWIG_PyUnicode_AsUTF8AndSize(obj, &len, &bytes);
    if (!cstr)
      return SWIG_TypeError;
    /* The returned string is only duplicated if the char * returned is not owned and memory managed by obj */
    if (bytes && cptr) {
      if (alloc) {
        cstr = (char *)memcpy(malloc((len + 1)*sizeof(char)), cstr, sizeof(char)*(len + 1));
        *alloc = SWIG_NEWOBJ;
      } else {
        /* alloc must be set in order to clean up allocated memory */
        return SWIG_RuntimeError;
      }
    }
#endif
    if (cptr) *cptr = cstr;
    if (psize) *psize = len + 1;
    SWIG_Py_XDECREF(bytes);
    return ret;
  } else {
#if defined(SWIG_PYTHON_2_UNICODE)
#if defined(SWIG_PYTHON_STRICT_BYTE_CHAR)
#error "Cannot use both SWIG_PYTHON_2_UNICODE and SWIG_PYTHON_STRICT_BYTE_CHAR at once"
#endif
#if PY_VERSION_HEX < 0x03000000
    if (PyUnicode_Check(obj)) {
      char *cstr; Py_ssize_t len;
      if (!alloc && cptr) {
        return SWIG_RuntimeError;
      }
      obj = PyUnicode_AsUTF8String(obj);
      if (!obj)
        return SWIG_TypeError;
      if (PyString_AsStringAndSize(obj, &cstr, &len) != -1) {
        if (cptr) {
          if (alloc) *alloc = SWIG_NEWOBJ;
          *cptr = (char *)memcpy(malloc((len + 1)*sizeof(char)), cstr, sizeof(char)*(len + 1));
        }
        if (psize) *psize = len + 1;

        SWIG_Py_XDECREF(obj);
        return SWIG_OK;
      } else {
        SWIG_Py_XDECREF(obj);
      }
    }
#endif
#endif

    swig_type_info* pchar_descriptor = SWIG_pchar_descriptor();
    if (pchar_descriptor) {
      void* vptr = 0;
      if (SWIG_ConvertPtr(obj, &vptr, pchar_descriptor, 0) == SWIG_OK) {
	if (cptr) *cptr = (char *) vptr;
	if (psize) *psize = vptr ? (strlen((char *)vptr) + 1) : 0;
	if (alloc) *alloc = SWIG_OLDOBJ;
	return SWIG_OK;
      }
    }
  }
  return SWIG_TypeError;
}





  #define SWIG_From_long   PyInt_FromLong 


SWIGINTERN int
SWIG_AsVal_double (PyObject *obj, double *val)
{
  int res = SWIG_TypeError;
  if (PyFloat_Check(obj)) {
    if (val) *val = PyFloat_AsDouble(obj);
    return SWIG_OK;
#if PY_VERSION_HEX < 0x03000000
  } else if (PyInt_Check(obj)) {
    if (val) *val = (double) PyInt_AsLong(obj);
    return SWIG_OK;
#endif
  } else if (PyLong_Check(obj)) {
    double v = PyLong_AsDouble(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = v;
      return SWIG_OK;
    } else {
      PyErr_Clear();
    }
  }
#ifdef SWIG_PYTHON_CAST_MODE
  {
    int dispatch = 0;
    double d = PyFloat_AsDouble(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = d;
      return SWIG_AddCast(SWIG_OK);
    } else {
      PyErr_Clear();
    }
    if (!dispatch) {
      long v = PyLong_AsLong(obj);
      if (!PyErr_Occurred()) {
	if (val) *val = v;
	return SWIG_AddCast(SWIG_AddCast(SWIG_OK));
      } else {
	PyErr_Clear();
      }
    }
  }
#endif
  return res;
}


#include <errno.h>


#include <float.h>


#include <math.h>


SWIGINTERNINLINE int
SWIG_CanCastAsInteger(double *d, double min, double max) {
  double x = *d;
  if ((min <= x && x <= max)) {
   double fx, cx, rd;
   errno = 0;
   fx = floor(x);
   cx = ceil(x);
   rd =  ((x - fx) < 0.5) ? fx : cx; /* simple rint */
   if ((errno == EDOM) || (errno == ERANGE)) {
     errno = 0;
   } else {
     double summ, reps, diff;
     if (rd < x) {
       diff = x - rd;
     } else if (rd > x) {
       diff = rd - x;
     } else {
       return 1;
     }
     summ = rd + x;
     reps = diff/summ;
     if (reps < 8*DBL_EPSILON) {
       *d = rd;
       return 1;
     }
   }
  }
  return 0;
}


SWIGINTERN int
SWIG_AsVal_long (PyObject *obj, long* val)
{
#if PY_VERSION_HEX < 0x03000000
  if (PyInt_Check(obj)) {
    if (val) *val = PyInt_AsLong(obj);
    return SWIG_OK;
  } else
#endif
  if (PyLong_Check(obj)) {
    long v = PyLong_AsLong(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = v;
      return SWIG_OK;
    } else {
      PyErr_Clear();
      return SWIG_OverflowError;
    }
  }
#ifdef SWIG_PYTHON_CAST_MODE
  {
    int dispatch = 0;
    long v = PyInt_AsLong(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = v;
      return SWIG_AddCast(SWIG_OK);
    } else {
      PyErr_Clear();
    }
    if (!dispatch) {
      double d;
      int res = SWIG_AddCast(SWIG_AsVal_double (obj,&d));
      // Largest double not larger than LONG_MAX (not portably calculated easily)
      // Note that double(LONG_MAX) is stored in a double rounded up by one (for 64-bit long)
      // 0x7ffffffffffffc00LL == (int64_t)std::nextafter(double(__uint128_t(LONG_MAX)+1), double(0))
      const double long_max = sizeof(long) == 8 ? 0x7ffffffffffffc00LL : LONG_MAX;
      // No equivalent needed for 64-bit double(LONG_MIN) is exactly LONG_MIN
      if (SWIG_IsOK(res) && SWIG_CanCastAsInteger(&d, LONG_MIN, long_max)) {
	if (val) *val = (long)(d);
	return res;
      }
    }
  }
#endif
  return SWIG_TypeError;
}


SWIGINTERNINLINE long
SWIG_As_long (PyObject* obj)
{
  long v;
  int res = SWIG_AsVal_long (obj, &v);
  if (!SWIG_IsOK(res)) {
    /*
      this is needed to make valgrind/purify happier. 
     */
    memset((void*)&v, 0, sizeof(long));
    SWIG_Error(res, "");
  }
  return v;
}


SWIGINTERN int
SWIG_AsVal_unsigned_SS_long (PyObject *obj, unsigned long *val) 
{
#if PY_VERSION_HEX < 0x03000000
  if (PyInt_Check(obj)) {
    long v = PyInt_AsLong(obj);
    if (v >= 0) {
      if (val) *val = v;
      return SWIG_OK;
    } else {
      return SWIG_OverflowError;
    }
  } else
#endif
  if (PyLong_Check(obj)) {
    unsigned long v = PyLong_AsUnsignedLong(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = v;
      return SWIG_OK;
    } else {
      PyErr_Clear();
      return SWIG_OverflowError;
    }
  }
#ifdef SWIG_PYTHON_CAST_MODE
  {
    int dispatch = 0;
    unsigned long v = PyLong_AsUnsignedLong(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = v;
      return SWIG_AddCast(SWIG_OK);
    } else {
      PyErr_Clear();
    }
    if (!dispatch) {
      double d;
      int res = SWIG_AddCast(SWIG_AsVal_double (obj,&d));
      // Largest double not larger than ULONG_MAX (not portably calculated easily)
      // Note that double(ULONG_MAX) is stored in a double rounded up by one (for 64-bit unsigned long)
      // 0xfffffffffffff800ULL == (uint64_t)std::nextafter(double(__uint128_t(ULONG_MAX)+1), double(0))
      const double ulong_max = sizeof(unsigned long) == 8 ? 0xfffffffffffff800ULL : ULONG_MAX;
      if (SWIG_IsOK(res) && SWIG_CanCastAsInteger(&d, 0, ulong_max)) {
	if (val) *val = (unsigned long)(d);
	return res;
      }
    }
  }
#endif
  return SWIG_TypeError;
}


SWIGINTERNINLINE unsigned long
SWIG_As_unsigned_SS_long (PyObject* obj)
{
  unsigned long v;
  int res = SWIG_AsVal_unsigned_SS_long (obj, &v);
  if (!SWIG_IsOK(res)) {
    /*
      this is needed to make valgrind/purify happier. 
     */
    memset((void*)&v, 0, sizeof(unsigned long));
    SWIG_Error(res, "");
  }
  return v;
}


SWIGINTERNINLINE PyObject*
  SWIG_From_int  (int value)
{
  return PyInt_FromLong((long) value);
}


SWIGINTERNINLINE PyObject *
SWIG_FromCharPtrAndSize(const char* carray, size_t size)
{
  if (carray) {
    if (size > (size_t)PY_SSIZE_T_MAX) {
      swig_type_info* pchar_descriptor = SWIG_pchar_descriptor();
      return pchar_descriptor ? 
	SWIG_InternalNewPointerObj((char *)(carray), pchar_descriptor, 0) : SWIG_Py_Void();
    } else {
#if PY_VERSION_HEX >= 0x03000000
#if defined(SWIG_PYTHON_STRICT_BYTE_CHAR)
      return PyBytes_FromStringAndSize(carray, (Py_ssize_t)(size));
#else
      return PyUnicode_DecodeUTF8(carray, (Py_ssize_t)(size), "surrogateescape");
#endif
#else
      return PyString_FromStringAndSize(carray, (Py_ssize_t)(size));
#endif
    }
  } else {
    return SWIG_Py_Void();
  }
}


SWIGINTERNINLINE PyObject * 
SWIG_FromCharPtr(const char *cptr)
{ 
  return SWIG_FromCharPtrAndSize(cptr, (cptr ? strlen(cptr) : 0));
}


static svn_error_t * svn_ra_reporter3_invoke_set_path(
  svn_ra_reporter3_t * _obj, void *report_baton, const char *path, svn_revnum_t revision, svn_depth_t depth, svn_boolean_t start_empty, const char *lock_token, apr_pool_t *pool) {
  return (_obj->set_path)(report_baton, path, revision, depth, start_empty, lock_token, pool);
}

static svn_error_t * svn_ra_reporter3_invoke_delete_path(
  svn_ra_reporter3_t * _obj, void *report_baton, const char *path, apr_pool_t *pool) {
  return (_obj->delete_path)(report_baton, path, pool);
}

static svn_error_t * svn_ra_reporter3_invoke_link_path(
  svn_ra_reporter3_t * _obj, void *report_baton, const char *path, const char *url, svn_revnum_t revision, svn_depth_t depth, svn_boolean_t start_empty, const char *lock_token, apr_pool_t *pool) {
  return (_obj->link_path)(report_baton, path, url, revision, depth, start_empty, lock_token, pool);
}

static svn_error_t * svn_ra_reporter3_invoke_finish_report(
  svn_ra_reporter3_t * _obj, void *report_baton, apr_pool_t *pool) {
  return (_obj->finish_report)(report_baton, pool);
}

static svn_error_t * svn_ra_reporter3_invoke_abort_report(
  svn_ra_reporter3_t * _obj, void *report_baton, apr_pool_t *pool) {
  return (_obj->abort_report)(report_baton, pool);
}

static svn_error_t * svn_ra_reporter2_invoke_set_path(
  svn_ra_reporter2_t * _obj, void *report_baton, const char *path, svn_revnum_t revision, svn_boolean_t start_empty, const char *lock_token, apr_pool_t *pool) {
  return (_obj->set_path)(report_baton, path, revision, start_empty, lock_token, pool);
}

static svn_error_t * svn_ra_reporter2_invoke_delete_path(
  svn_ra_reporter2_t * _obj, void *report_baton, const char *path, apr_pool_t *pool) {
  return (_obj->delete_path)(report_baton, path, pool);
}

static svn_error_t * svn_ra_reporter2_invoke_link_path(
  svn_ra_reporter2_t * _obj, void *report_baton, const char *path, const char *url, svn_revnum_t revision, svn_boolean_t start_empty, const char *lock_token, apr_pool_t *pool) {
  return (_obj->link_path)(report_baton, path, url, revision, start_empty, lock_token, pool);
}

static svn_error_t * svn_ra_reporter2_invoke_finish_report(
  svn_ra_reporter2_t * _obj, void *report_baton, apr_pool_t *pool) {
  return (_obj->finish_report)(report_baton, pool);
}

static svn_error_t * svn_ra_reporter2_invoke_abort_report(
  svn_ra_reporter2_t * _obj, void *report_baton, apr_pool_t *pool) {
  return (_obj->abort_report)(report_baton, pool);
}

static svn_error_t * svn_ra_reporter_invoke_set_path(
  svn_ra_reporter_t * _obj, void *report_baton, const char *path, svn_revnum_t revision, svn_boolean_t start_empty, apr_pool_t *pool) {
  return (_obj->set_path)(report_baton, path, revision, start_empty, pool);
}

static svn_error_t * svn_ra_reporter_invoke_delete_path(
  svn_ra_reporter_t * _obj, void *report_baton, const char *path, apr_pool_t *pool) {
  return (_obj->delete_path)(report_baton, path, pool);
}

static svn_error_t * svn_ra_reporter_invoke_link_path(
  svn_ra_reporter_t * _obj, void *report_baton, const char *path, const char *url, svn_revnum_t revision, svn_boolean_t start_empty, apr_pool_t *pool) {
  return (_obj->link_path)(report_baton, path, url, revision, start_empty, pool);
}

static svn_error_t * svn_ra_reporter_invoke_finish_report(
  svn_ra_reporter_t * _obj, void *report_baton, apr_pool_t *pool) {
  return (_obj->finish_report)(report_baton, pool);
}

static svn_error_t * svn_ra_reporter_invoke_abort_report(
  svn_ra_reporter_t * _obj, void *report_baton, apr_pool_t *pool) {
  return (_obj->abort_report)(report_baton, pool);
}

static svn_error_t * svn_ra_callbacks2_invoke_open_tmp_file(
  svn_ra_callbacks2_t * _obj, apr_file_t **fp, void *callback_baton, apr_pool_t *pool) {
  return (_obj->open_tmp_file)(fp, callback_baton, pool);
}

static svn_error_t * svn_ra_callbacks_invoke_open_tmp_file(
  svn_ra_callbacks_t * _obj, apr_file_t **fp, void *callback_baton, apr_pool_t *pool) {
  return (_obj->open_tmp_file)(fp, callback_baton, pool);
}

static svn_error_t * svn_ra_plugin_invoke_open(
  svn_ra_plugin_t * _obj, void **session_baton, const char *repos_URL, const svn_ra_callbacks_t *callbacks, void *callback_baton, apr_hash_t *config, apr_pool_t *pool) {
  return (_obj->open)(session_baton, repos_URL, callbacks, callback_baton, config, pool);
}

static svn_error_t * svn_ra_plugin_invoke_get_latest_revnum(
  svn_ra_plugin_t * _obj, void *session_baton, svn_revnum_t *latest_revnum, apr_pool_t *pool) {
  return (_obj->get_latest_revnum)(session_baton, latest_revnum, pool);
}

static svn_error_t * svn_ra_plugin_invoke_get_dated_revision(
  svn_ra_plugin_t * _obj, void *session_baton, svn_revnum_t *revision, apr_time_t tm, apr_pool_t *pool) {
  return (_obj->get_dated_revision)(session_baton, revision, tm, pool);
}

static svn_error_t * svn_ra_plugin_invoke_change_rev_prop(
  svn_ra_plugin_t * _obj, void *session_baton, svn_revnum_t rev, const char *name, const svn_string_t *value, apr_pool_t *pool) {
  return (_obj->change_rev_prop)(session_baton, rev, name, value, pool);
}

static svn_error_t * svn_ra_plugin_invoke_rev_proplist(
  svn_ra_plugin_t * _obj, void *session_baton, svn_revnum_t rev, apr_hash_t **props, apr_pool_t *pool) {
  return (_obj->rev_proplist)(session_baton, rev, props, pool);
}

static svn_error_t * svn_ra_plugin_invoke_rev_prop(
  svn_ra_plugin_t * _obj, void *session_baton, svn_revnum_t rev, const char *name, svn_string_t **value, apr_pool_t *pool) {
  return (_obj->rev_prop)(session_baton, rev, name, value, pool);
}

static svn_error_t * svn_ra_plugin_invoke_get_commit_editor(
  svn_ra_plugin_t * _obj, void *session_baton, const svn_delta_editor_t **editor, void **edit_baton, const char *log_msg, svn_commit_callback_t callback, void *callback_baton, apr_pool_t *pool) {
  return (_obj->get_commit_editor)(session_baton, editor, edit_baton, log_msg, callback, callback_baton, pool);
}

static svn_error_t * svn_ra_plugin_invoke_get_file(
  svn_ra_plugin_t * _obj, void *session_baton, const char *path, svn_revnum_t revision, svn_stream_t *stream, svn_revnum_t *fetched_rev, apr_hash_t **props, apr_pool_t *pool) {
  return (_obj->get_file)(session_baton, path, revision, stream, fetched_rev, props, pool);
}

static svn_error_t * svn_ra_plugin_invoke_get_dir(
  svn_ra_plugin_t * _obj, void *session_baton, const char *path, svn_revnum_t revision, apr_hash_t **dirents, svn_revnum_t *fetched_rev, apr_hash_t **props, apr_pool_t *pool) {
  return (_obj->get_dir)(session_baton, path, revision, dirents, fetched_rev, props, pool);
}

static svn_error_t * svn_ra_plugin_invoke_do_update(
  svn_ra_plugin_t * _obj, void *session_baton, const svn_ra_reporter_t **reporter, void **report_baton, svn_revnum_t revision_to_update_to, const char *update_target, svn_boolean_t recurse, const svn_delta_editor_t *update_editor, void *update_baton, apr_pool_t *pool) {
  return (_obj->do_update)(session_baton, reporter, report_baton, revision_to_update_to, update_target, recurse, update_editor, update_baton, pool);
}

static svn_error_t * svn_ra_plugin_invoke_do_switch(
  svn_ra_plugin_t * _obj, void *session_baton, const svn_ra_reporter_t **reporter, void **report_baton, svn_revnum_t revision_to_switch_to, const char *switch_target, svn_boolean_t recurse, const char *switch_url, const svn_delta_editor_t *switch_editor, void *switch_baton, apr_pool_t *pool) {
  return (_obj->do_switch)(session_baton, reporter, report_baton, revision_to_switch_to, switch_target, recurse, switch_url, switch_editor, switch_baton, pool);
}

static svn_error_t * svn_ra_plugin_invoke_do_status(
  svn_ra_plugin_t * _obj, void *session_baton, const svn_ra_reporter_t **reporter, void **report_baton, const char *status_target, svn_revnum_t revision, svn_boolean_t recurse, const svn_delta_editor_t *status_editor, void *status_baton, apr_pool_t *pool) {
  return (_obj->do_status)(session_baton, reporter, report_baton, status_target, revision, recurse, status_editor, status_baton, pool);
}

static svn_error_t * svn_ra_plugin_invoke_do_diff(
  svn_ra_plugin_t * _obj, void *session_baton, const svn_ra_reporter_t **reporter, void **report_baton, svn_revnum_t revision, const char *diff_target, svn_boolean_t recurse, svn_boolean_t ignore_ancestry, const char *versus_url, const svn_delta_editor_t *diff_editor, void *diff_baton, apr_pool_t *pool) {
  return (_obj->do_diff)(session_baton, reporter, report_baton, revision, diff_target, recurse, ignore_ancestry, versus_url, diff_editor, diff_baton, pool);
}

static svn_error_t * svn_ra_plugin_invoke_get_log(
  svn_ra_plugin_t * _obj, void *session_baton, const apr_array_header_t *paths, svn_revnum_t start, svn_revnum_t end, svn_boolean_t discover_changed_paths, svn_boolean_t strict_node_history, svn_log_message_receiver_t receiver, void *receiver_baton, apr_pool_t *pool) {
  return (_obj->get_log)(session_baton, paths, start, end, discover_changed_paths, strict_node_history, receiver, receiver_baton, pool);
}

static svn_error_t * svn_ra_plugin_invoke_check_path(
  svn_ra_plugin_t * _obj, void *session_baton, const char *path, svn_revnum_t revision, svn_node_kind_t *kind, apr_pool_t *pool) {
  return (_obj->check_path)(session_baton, path, revision, kind, pool);
}

static svn_error_t * svn_ra_plugin_invoke_get_uuid(
  svn_ra_plugin_t * _obj, void *session_baton, const char **uuid, apr_pool_t *pool) {
  return (_obj->get_uuid)(session_baton, uuid, pool);
}

static svn_error_t * svn_ra_plugin_invoke_get_repos_root(
  svn_ra_plugin_t * _obj, void *session_baton, const char **url, apr_pool_t *pool) {
  return (_obj->get_repos_root)(session_baton, url, pool);
}

static svn_error_t * svn_ra_plugin_invoke_get_locations(
  svn_ra_plugin_t * _obj, void *session_baton, apr_hash_t **locations, const char *path, svn_revnum_t peg_revision, apr_array_header_t *location_revisions, apr_pool_t *pool) {
  return (_obj->get_locations)(session_baton, locations, path, peg_revision, location_revisions, pool);
}

static svn_error_t * svn_ra_plugin_invoke_get_file_revs(
  svn_ra_plugin_t * _obj, void *session_baton, const char *path, svn_revnum_t start, svn_revnum_t end, svn_ra_file_rev_handler_t handler, void *handler_baton, apr_pool_t *pool) {
  return (_obj->get_file_revs)(session_baton, path, start, end, handler, handler_baton, pool);
}

static const svn_version_t * svn_ra_plugin_invoke_get_version(
  svn_ra_plugin_t * _obj) {
  return (_obj->get_version)();
}

static svn_error_t * svn_ra_invoke_get_wc_prop_func(
  svn_ra_get_wc_prop_func_t _obj, void *baton, const char *path, const char *name, const svn_string_t **value, apr_pool_t *pool) {
  return _obj(baton, path, name, value, pool);
}

static svn_error_t * svn_ra_invoke_set_wc_prop_func(
  svn_ra_set_wc_prop_func_t _obj, void *baton, const char *path, const char *name, const svn_string_t *value, apr_pool_t *pool) {
  return _obj(baton, path, name, value, pool);
}

static svn_error_t * svn_ra_invoke_push_wc_prop_func(
  svn_ra_push_wc_prop_func_t _obj, void *baton, const char *path, const char *name, const svn_string_t *value, apr_pool_t *pool) {
  return _obj(baton, path, name, value, pool);
}

static svn_error_t * svn_ra_invoke_invalidate_wc_props_func(
  svn_ra_invalidate_wc_props_func_t _obj, void *baton, const char *path, const char *name, apr_pool_t *pool) {
  return _obj(baton, path, name, pool);
}

static svn_error_t * svn_ra_invoke_get_wc_contents_func(
  svn_ra_get_wc_contents_func_t _obj, void *baton, svn_stream_t **contents, const svn_checksum_t *checksum, apr_pool_t *pool) {
  return _obj(baton, contents, checksum, pool);
}

static svn_error_t * svn_ra_invoke_get_latest_revnum_func(
  svn_ra_get_latest_revnum_func_t _obj, void *session_baton, svn_revnum_t *latest_revnum) {
  return _obj(session_baton, latest_revnum);
}

static svn_error_t * svn_ra_invoke_get_client_string_func(
  svn_ra_get_client_string_func_t _obj, void *baton, const char **name, apr_pool_t *pool) {
  return _obj(baton, name, pool);
}

static svn_error_t * svn_ra_invoke_file_rev_handler(
  svn_ra_file_rev_handler_t _obj, void *baton, const char *path, svn_revnum_t rev, apr_hash_t *rev_props, svn_txdelta_window_handler_t *delta_handler, void **delta_baton, apr_array_header_t *prop_diffs, apr_pool_t *pool) {
  return _obj(baton, path, rev, rev_props, delta_handler, delta_baton, prop_diffs, pool);
}

static svn_error_t * svn_ra_invoke_lock_callback(
  svn_ra_lock_callback_t _obj, void *baton, const char *path, svn_boolean_t do_lock, const svn_lock_t *lock, svn_error_t *ra_err, apr_pool_t *pool) {
  return _obj(baton, path, do_lock, lock, ra_err, pool);
}

static void svn_ra_invoke_progress_notify_func(
  svn_ra_progress_notify_func_t _obj, apr_off_t progress, apr_off_t total, void *baton, apr_pool_t *pool) {
  _obj(progress, total, baton, pool);
}

static svn_error_t * svn_ra_invoke_replay_revstart_callback(
  svn_ra_replay_revstart_callback_t _obj, svn_revnum_t revision, void *replay_baton, const svn_delta_editor_t **editor, void **edit_baton, apr_hash_t *rev_props, apr_pool_t *pool) {
  return _obj(revision, replay_baton, editor, edit_baton, rev_props, pool);
}

static svn_error_t * svn_ra_invoke_replay_revfinish_callback(
  svn_ra_replay_revfinish_callback_t _obj, svn_revnum_t revision, void *replay_baton, const svn_delta_editor_t *editor, void *edit_baton, apr_hash_t *rev_props, apr_pool_t *pool) {
  return _obj(revision, replay_baton, editor, edit_baton, rev_props, pool);
}

static svn_boolean_t svn_ra_invoke_check_tunnel_func(
  svn_ra_check_tunnel_func_t _obj, void *tunnel_baton, const char *tunnel_name) {
  return _obj(tunnel_baton, tunnel_name);
}

static void svn_ra_invoke_close_tunnel_func(
  svn_ra_close_tunnel_func_t _obj, void *close_baton, void *tunnel_baton) {
  _obj(close_baton, tunnel_baton);
}

static svn_error_t * svn_ra_invoke_open_tunnel_func(
  svn_ra_open_tunnel_func_t _obj, svn_stream_t **request, svn_stream_t **response, svn_ra_close_tunnel_func_t *close_func, void **close_baton, void *tunnel_baton, const char *tunnel_name, const char *user, const char *hostname, int port, svn_cancel_func_t cancel_func, void *cancel_baton, apr_pool_t *pool) {
  return _obj(request, response, close_func, close_baton, tunnel_baton, tunnel_name, user, hostname, port, cancel_func, cancel_baton, pool);
}

static svn_error_t * svn_ra_invoke_init_func(
  svn_ra_init_func_t _obj, int abi_version, apr_pool_t *pool, apr_hash_t *hash) {
  return _obj(abi_version, pool, hash);
}


#ifdef __cplusplus
extern "C" {
#endif
SWIGINTERN PyObject *_wrap_svn_ra_version(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_version_t *result = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_version", 0, 0)) SWIG_fail;
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_version_t *)svn_ra_version();
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_version_t,
    _global_py_pool, args);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_reporter3_t_set_path_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_reporter3_t *arg1 = 0 ;
  svn_error_t *(*arg2)(void *,char const *,svn_revnum_t,svn_depth_t,svn_boolean_t,char const *,apr_pool_t *) = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_reporter3_t_set_path_set", 2, 2, &obj0, &obj1)) SWIG_fail;
  {
    arg1 = (struct svn_ra_reporter3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_reporter3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_void_p_q_const__char_svn_revnum_t_svn_depth_t_svn_boolean_t_p_q_const__char_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_ra_reporter3_t_set_path_set" "', argument " "2"" of type '" "svn_error_t *(*)(void *,char const *,svn_revnum_t,svn_depth_t,svn_boolean_t,char const *,apr_pool_t *)""'"); 
    }
  }
  if (arg1) (arg1)->set_path = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_reporter3_t_set_path_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_reporter3_t *arg1 = 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(void *,char const *,svn_revnum_t,svn_depth_t,svn_boolean_t,char const *,apr_pool_t *) = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_reporter3_t_set_path_get", 1, 1, &obj0)) SWIG_fail;
  {
    arg1 = (struct svn_ra_reporter3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_reporter3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(void *,char const *,svn_revnum_t,svn_depth_t,svn_boolean_t,char const *,apr_pool_t *)) ((arg1)->set_path);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_void_p_q_const__char_svn_revnum_t_svn_depth_t_svn_boolean_t_p_q_const__char_p_apr_pool_t__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_reporter3_t_delete_path_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_reporter3_t *arg1 = 0 ;
  svn_error_t *(*arg2)(void *,char const *,apr_pool_t *) = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_reporter3_t_delete_path_set", 2, 2, &obj0, &obj1)) SWIG_fail;
  {
    arg1 = (struct svn_ra_reporter3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_reporter3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_void_p_q_const__char_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_ra_reporter3_t_delete_path_set" "', argument " "2"" of type '" "svn_error_t *(*)(void *,char const *,apr_pool_t *)""'"); 
    }
  }
  if (arg1) (arg1)->delete_path = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_reporter3_t_delete_path_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_reporter3_t *arg1 = 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(void *,char const *,apr_pool_t *) = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_reporter3_t_delete_path_get", 1, 1, &obj0)) SWIG_fail;
  {
    arg1 = (struct svn_ra_reporter3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_reporter3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(void *,char const *,apr_pool_t *)) ((arg1)->delete_path);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_void_p_q_const__char_p_apr_pool_t__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_reporter3_t_link_path_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_reporter3_t *arg1 = 0 ;
  svn_error_t *(*arg2)(void *,char const *,char const *,svn_revnum_t,svn_depth_t,svn_boolean_t,char const *,apr_pool_t *) = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_reporter3_t_link_path_set", 2, 2, &obj0, &obj1)) SWIG_fail;
  {
    arg1 = (struct svn_ra_reporter3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_reporter3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_void_p_q_const__char_p_q_const__char_svn_revnum_t_svn_depth_t_svn_boolean_t_p_q_const__char_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_ra_reporter3_t_link_path_set" "', argument " "2"" of type '" "svn_error_t *(*)(void *,char const *,char const *,svn_revnum_t,svn_depth_t,svn_boolean_t,char const *,apr_pool_t *)""'"); 
    }
  }
  if (arg1) (arg1)->link_path = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_reporter3_t_link_path_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_reporter3_t *arg1 = 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(void *,char const *,char const *,svn_revnum_t,svn_depth_t,svn_boolean_t,char const *,apr_pool_t *) = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_reporter3_t_link_path_get", 1, 1, &obj0)) SWIG_fail;
  {
    arg1 = (struct svn_ra_reporter3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_reporter3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(void *,char const *,char const *,svn_revnum_t,svn_depth_t,svn_boolean_t,char const *,apr_pool_t *)) ((arg1)->link_path);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_void_p_q_const__char_p_q_const__char_svn_revnum_t_svn_depth_t_svn_boolean_t_p_q_const__char_p_apr_pool_t__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_reporter3_t_finish_report_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_reporter3_t *arg1 = 0 ;
  svn_error_t *(*arg2)(void *,apr_pool_t *) = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_reporter3_t_finish_report_set", 2, 2, &obj0, &obj1)) SWIG_fail;
  {
    arg1 = (struct svn_ra_reporter3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_reporter3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_void_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_ra_reporter3_t_finish_report_set" "', argument " "2"" of type '" "svn_error_t *(*)(void *,apr_pool_t *)""'"); 
    }
  }
  if (arg1) (arg1)->finish_report = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_reporter3_t_finish_report_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_reporter3_t *arg1 = 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(void *,apr_pool_t *) = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_reporter3_t_finish_report_get", 1, 1, &obj0)) SWIG_fail;
  {
    arg1 = (struct svn_ra_reporter3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_reporter3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(void *,apr_pool_t *)) ((arg1)->finish_report);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_void_p_apr_pool_t__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_reporter3_t_abort_report_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_reporter3_t *arg1 = 0 ;
  svn_error_t *(*arg2)(void *,apr_pool_t *) = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_reporter3_t_abort_report_set", 2, 2, &obj0, &obj1)) SWIG_fail;
  {
    arg1 = (struct svn_ra_reporter3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_reporter3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_void_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_ra_reporter3_t_abort_report_set" "', argument " "2"" of type '" "svn_error_t *(*)(void *,apr_pool_t *)""'"); 
    }
  }
  if (arg1) (arg1)->abort_report = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_reporter3_t_abort_report_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_reporter3_t *arg1 = 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(void *,apr_pool_t *) = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_reporter3_t_abort_report_get", 1, 1, &obj0)) SWIG_fail;
  {
    arg1 = (struct svn_ra_reporter3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_reporter3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(void *,apr_pool_t *)) ((arg1)->abort_report);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_void_p_apr_pool_t__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_svn_ra_reporter3_t(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_reporter3_t *result = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "new_svn_ra_reporter3_t", 0, 0)) SWIG_fail;
  {
    svn_swig_py_release_py_lock();
    
    result = (struct svn_ra_reporter3_t *)calloc(1, sizeof(struct svn_ra_reporter3_t));
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_ra_reporter3_t,
    _global_py_pool, args);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_svn_ra_reporter3_t(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_reporter3_t *arg1 = 0 ;
  PyObject * obj0 = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "delete_svn_ra_reporter3_t", 1, 1, &obj0)) SWIG_fail;
  {
    arg1 = (struct svn_ra_reporter3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_reporter3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    free((char *) arg1);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *svn_ra_reporter3_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_svn_ra_reporter3_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *svn_ra_reporter3_t_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_svn_ra_reporter2_t_set_path_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_reporter2_t *arg1 = 0 ;
  svn_error_t *(*arg2)(void *,char const *,svn_revnum_t,svn_boolean_t,char const *,apr_pool_t *) = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_reporter2_t_set_path_set", 2, 2, &obj0, &obj1)) SWIG_fail;
  {
    arg1 = (struct svn_ra_reporter2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_reporter2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_void_p_q_const__char_svn_revnum_t_svn_boolean_t_p_q_const__char_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_ra_reporter2_t_set_path_set" "', argument " "2"" of type '" "svn_error_t *(*)(void *,char const *,svn_revnum_t,svn_boolean_t,char const *,apr_pool_t *)""'"); 
    }
  }
  if (arg1) (arg1)->set_path = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_reporter2_t_set_path_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_reporter2_t *arg1 = 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(void *,char const *,svn_revnum_t,svn_boolean_t,char const *,apr_pool_t *) = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_reporter2_t_set_path_get", 1, 1, &obj0)) SWIG_fail;
  {
    arg1 = (struct svn_ra_reporter2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_reporter2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(void *,char const *,svn_revnum_t,svn_boolean_t,char const *,apr_pool_t *)) ((arg1)->set_path);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_void_p_q_const__char_svn_revnum_t_svn_boolean_t_p_q_const__char_p_apr_pool_t__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_reporter2_t_delete_path_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_reporter2_t *arg1 = 0 ;
  svn_error_t *(*arg2)(void *,char const *,apr_pool_t *) = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_reporter2_t_delete_path_set", 2, 2, &obj0, &obj1)) SWIG_fail;
  {
    arg1 = (struct svn_ra_reporter2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_reporter2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_void_p_q_const__char_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_ra_reporter2_t_delete_path_set" "', argument " "2"" of type '" "svn_error_t *(*)(void *,char const *,apr_pool_t *)""'"); 
    }
  }
  if (arg1) (arg1)->delete_path = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_reporter2_t_delete_path_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_reporter2_t *arg1 = 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(void *,char const *,apr_pool_t *) = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_reporter2_t_delete_path_get", 1, 1, &obj0)) SWIG_fail;
  {
    arg1 = (struct svn_ra_reporter2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_reporter2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(void *,char const *,apr_pool_t *)) ((arg1)->delete_path);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_void_p_q_const__char_p_apr_pool_t__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_reporter2_t_link_path_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_reporter2_t *arg1 = 0 ;
  svn_error_t *(*arg2)(void *,char const *,char const *,svn_revnum_t,svn_boolean_t,char const *,apr_pool_t *) = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_reporter2_t_link_path_set", 2, 2, &obj0, &obj1)) SWIG_fail;
  {
    arg1 = (struct svn_ra_reporter2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_reporter2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_void_p_q_const__char_p_q_const__char_svn_revnum_t_svn_boolean_t_p_q_const__char_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_ra_reporter2_t_link_path_set" "', argument " "2"" of type '" "svn_error_t *(*)(void *,char const *,char const *,svn_revnum_t,svn_boolean_t,char const *,apr_pool_t *)""'"); 
    }
  }
  if (arg1) (arg1)->link_path = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_reporter2_t_link_path_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_reporter2_t *arg1 = 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(void *,char const *,char const *,svn_revnum_t,svn_boolean_t,char const *,apr_pool_t *) = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_reporter2_t_link_path_get", 1, 1, &obj0)) SWIG_fail;
  {
    arg1 = (struct svn_ra_reporter2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_reporter2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(void *,char const *,char const *,svn_revnum_t,svn_boolean_t,char const *,apr_pool_t *)) ((arg1)->link_path);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_void_p_q_const__char_p_q_const__char_svn_revnum_t_svn_boolean_t_p_q_const__char_p_apr_pool_t__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_reporter2_t_finish_report_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_reporter2_t *arg1 = 0 ;
  svn_error_t *(*arg2)(void *,apr_pool_t *) = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_reporter2_t_finish_report_set", 2, 2, &obj0, &obj1)) SWIG_fail;
  {
    arg1 = (struct svn_ra_reporter2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_reporter2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_void_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_ra_reporter2_t_finish_report_set" "', argument " "2"" of type '" "svn_error_t *(*)(void *,apr_pool_t *)""'"); 
    }
  }
  if (arg1) (arg1)->finish_report = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_reporter2_t_finish_report_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_reporter2_t *arg1 = 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(void *,apr_pool_t *) = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_reporter2_t_finish_report_get", 1, 1, &obj0)) SWIG_fail;
  {
    arg1 = (struct svn_ra_reporter2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_reporter2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(void *,apr_pool_t *)) ((arg1)->finish_report);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_void_p_apr_pool_t__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_reporter2_t_abort_report_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_reporter2_t *arg1 = 0 ;
  svn_error_t *(*arg2)(void *,apr_pool_t *) = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_reporter2_t_abort_report_set", 2, 2, &obj0, &obj1)) SWIG_fail;
  {
    arg1 = (struct svn_ra_reporter2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_reporter2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_void_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_ra_reporter2_t_abort_report_set" "', argument " "2"" of type '" "svn_error_t *(*)(void *,apr_pool_t *)""'"); 
    }
  }
  if (arg1) (arg1)->abort_report = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_reporter2_t_abort_report_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_reporter2_t *arg1 = 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(void *,apr_pool_t *) = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_reporter2_t_abort_report_get", 1, 1, &obj0)) SWIG_fail;
  {
    arg1 = (struct svn_ra_reporter2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_reporter2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(void *,apr_pool_t *)) ((arg1)->abort_report);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_void_p_apr_pool_t__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_svn_ra_reporter2_t(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_reporter2_t *result = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "new_svn_ra_reporter2_t", 0, 0)) SWIG_fail;
  {
    svn_swig_py_release_py_lock();
    
    result = (struct svn_ra_reporter2_t *)calloc(1, sizeof(struct svn_ra_reporter2_t));
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_ra_reporter2_t,
    _global_py_pool, args);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_svn_ra_reporter2_t(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_reporter2_t *arg1 = 0 ;
  PyObject * obj0 = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "delete_svn_ra_reporter2_t", 1, 1, &obj0)) SWIG_fail;
  {
    arg1 = (struct svn_ra_reporter2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_reporter2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    free((char *) arg1);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *svn_ra_reporter2_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_svn_ra_reporter2_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *svn_ra_reporter2_t_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_svn_ra_reporter_t_set_path_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_reporter_t *arg1 = 0 ;
  svn_error_t *(*arg2)(void *,char const *,svn_revnum_t,svn_boolean_t,apr_pool_t *) = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_reporter_t_set_path_set", 2, 2, &obj0, &obj1)) SWIG_fail;
  {
    arg1 = (struct svn_ra_reporter_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_reporter_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_void_p_q_const__char_svn_revnum_t_svn_boolean_t_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_ra_reporter_t_set_path_set" "', argument " "2"" of type '" "svn_error_t *(*)(void *,char const *,svn_revnum_t,svn_boolean_t,apr_pool_t *)""'"); 
    }
  }
  if (arg1) (arg1)->set_path = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_reporter_t_set_path_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_reporter_t *arg1 = 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(void *,char const *,svn_revnum_t,svn_boolean_t,apr_pool_t *) = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_reporter_t_set_path_get", 1, 1, &obj0)) SWIG_fail;
  {
    arg1 = (struct svn_ra_reporter_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_reporter_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(void *,char const *,svn_revnum_t,svn_boolean_t,apr_pool_t *)) ((arg1)->set_path);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_void_p_q_const__char_svn_revnum_t_svn_boolean_t_p_apr_pool_t__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_reporter_t_delete_path_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_reporter_t *arg1 = 0 ;
  svn_error_t *(*arg2)(void *,char const *,apr_pool_t *) = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_reporter_t_delete_path_set", 2, 2, &obj0, &obj1)) SWIG_fail;
  {
    arg1 = (struct svn_ra_reporter_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_reporter_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_void_p_q_const__char_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_ra_reporter_t_delete_path_set" "', argument " "2"" of type '" "svn_error_t *(*)(void *,char const *,apr_pool_t *)""'"); 
    }
  }
  if (arg1) (arg1)->delete_path = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_reporter_t_delete_path_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_reporter_t *arg1 = 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(void *,char const *,apr_pool_t *) = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_reporter_t_delete_path_get", 1, 1, &obj0)) SWIG_fail;
  {
    arg1 = (struct svn_ra_reporter_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_reporter_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(void *,char const *,apr_pool_t *)) ((arg1)->delete_path);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_void_p_q_const__char_p_apr_pool_t__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_reporter_t_link_path_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_reporter_t *arg1 = 0 ;
  svn_error_t *(*arg2)(void *,char const *,char const *,svn_revnum_t,svn_boolean_t,apr_pool_t *) = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_reporter_t_link_path_set", 2, 2, &obj0, &obj1)) SWIG_fail;
  {
    arg1 = (struct svn_ra_reporter_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_reporter_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_void_p_q_const__char_p_q_const__char_svn_revnum_t_svn_boolean_t_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_ra_reporter_t_link_path_set" "', argument " "2"" of type '" "svn_error_t *(*)(void *,char const *,char const *,svn_revnum_t,svn_boolean_t,apr_pool_t *)""'"); 
    }
  }
  if (arg1) (arg1)->link_path = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_reporter_t_link_path_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_reporter_t *arg1 = 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(void *,char const *,char const *,svn_revnum_t,svn_boolean_t,apr_pool_t *) = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_reporter_t_link_path_get", 1, 1, &obj0)) SWIG_fail;
  {
    arg1 = (struct svn_ra_reporter_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_reporter_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(void *,char const *,char const *,svn_revnum_t,svn_boolean_t,apr_pool_t *)) ((arg1)->link_path);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_void_p_q_const__char_p_q_const__char_svn_revnum_t_svn_boolean_t_p_apr_pool_t__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_reporter_t_finish_report_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_reporter_t *arg1 = 0 ;
  svn_error_t *(*arg2)(void *,apr_pool_t *) = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_reporter_t_finish_report_set", 2, 2, &obj0, &obj1)) SWIG_fail;
  {
    arg1 = (struct svn_ra_reporter_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_reporter_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_void_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_ra_reporter_t_finish_report_set" "', argument " "2"" of type '" "svn_error_t *(*)(void *,apr_pool_t *)""'"); 
    }
  }
  if (arg1) (arg1)->finish_report = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_reporter_t_finish_report_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_reporter_t *arg1 = 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(void *,apr_pool_t *) = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_reporter_t_finish_report_get", 1, 1, &obj0)) SWIG_fail;
  {
    arg1 = (struct svn_ra_reporter_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_reporter_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(void *,apr_pool_t *)) ((arg1)->finish_report);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_void_p_apr_pool_t__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_reporter_t_abort_report_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_reporter_t *arg1 = 0 ;
  svn_error_t *(*arg2)(void *,apr_pool_t *) = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_reporter_t_abort_report_set", 2, 2, &obj0, &obj1)) SWIG_fail;
  {
    arg1 = (struct svn_ra_reporter_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_reporter_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_void_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_ra_reporter_t_abort_report_set" "', argument " "2"" of type '" "svn_error_t *(*)(void *,apr_pool_t *)""'"); 
    }
  }
  if (arg1) (arg1)->abort_report = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_reporter_t_abort_report_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_reporter_t *arg1 = 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(void *,apr_pool_t *) = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_reporter_t_abort_report_get", 1, 1, &obj0)) SWIG_fail;
  {
    arg1 = (struct svn_ra_reporter_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_reporter_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(void *,apr_pool_t *)) ((arg1)->abort_report);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_void_p_apr_pool_t__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_svn_ra_reporter_t(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_reporter_t *result = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "new_svn_ra_reporter_t", 0, 0)) SWIG_fail;
  {
    svn_swig_py_release_py_lock();
    
    result = (struct svn_ra_reporter_t *)calloc(1, sizeof(struct svn_ra_reporter_t));
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_ra_reporter_t,
    _global_py_pool, args);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_svn_ra_reporter_t(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_reporter_t *arg1 = 0 ;
  PyObject * obj0 = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "delete_svn_ra_reporter_t", 1, 1, &obj0)) SWIG_fail;
  {
    arg1 = (struct svn_ra_reporter_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_reporter_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    free((char *) arg1);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *svn_ra_reporter_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_svn_ra_reporter_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *svn_ra_reporter_t_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_svn_ra_callbacks2_t_open_tmp_file_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_callbacks2_t *arg1 = 0 ;
  svn_error_t *(*arg2)(apr_file_t **,void *,apr_pool_t *) = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_callbacks2_t_open_tmp_file_set", 2, 2, &obj0, &obj1)) SWIG_fail;
  {
    arg1 = (struct svn_ra_callbacks2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_callbacks2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_p_apr_file_t_p_void_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_ra_callbacks2_t_open_tmp_file_set" "', argument " "2"" of type '" "svn_error_t *(*)(apr_file_t **,void *,apr_pool_t *)""'"); 
    }
  }
  if (arg1) (arg1)->open_tmp_file = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_callbacks2_t_open_tmp_file_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_callbacks2_t *arg1 = 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(apr_file_t **,void *,apr_pool_t *) = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_callbacks2_t_open_tmp_file_get", 1, 1, &obj0)) SWIG_fail;
  {
    arg1 = (struct svn_ra_callbacks2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_callbacks2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(apr_file_t **,void *,apr_pool_t *)) ((arg1)->open_tmp_file);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_p_apr_file_t_p_void_p_apr_pool_t__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_callbacks2_t_auth_baton_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_callbacks2_t *arg1 = 0 ;
  svn_auth_baton_t *arg2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_callbacks2_t_auth_baton_set", 2, 2, &obj0, &obj1)) SWIG_fail;
  {
    arg1 = (struct svn_ra_callbacks2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_callbacks2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_auth_baton_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_auth_baton_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->auth_baton = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_callbacks2_t_auth_baton_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_callbacks2_t *arg1 = 0 ;
  PyObject * obj0 = 0 ;
  svn_auth_baton_t *result = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_callbacks2_t_auth_baton_get", 1, 1, &obj0)) SWIG_fail;
  {
    arg1 = (struct svn_ra_callbacks2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_callbacks2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_auth_baton_t *) ((arg1)->auth_baton);
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_auth_baton_t,
    _global_py_pool, args);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_callbacks2_t_get_wc_prop_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_callbacks2_t *arg1 = 0 ;
  svn_ra_get_wc_prop_func_t arg2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_callbacks2_t_get_wc_prop_set", 2, 2, &obj0, &obj1)) SWIG_fail;
  {
    arg1 = (struct svn_ra_callbacks2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_callbacks2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    svn_ra_get_wc_prop_func_t * tmp =
    svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_p_f_p_void_p_q_const__char_p_q_const__char_p_p_q_const__svn_string_t_p_apr_pool_t__p_svn_error_t, svn_argnum_obj1);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg2 = *tmp;
  }
  if (arg1) (arg1)->get_wc_prop = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_callbacks2_t_get_wc_prop_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_callbacks2_t *arg1 = 0 ;
  PyObject * obj0 = 0 ;
  svn_ra_get_wc_prop_func_t result;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_callbacks2_t_get_wc_prop_get", 1, 1, &obj0)) SWIG_fail;
  {
    arg1 = (struct svn_ra_callbacks2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_callbacks2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_ra_get_wc_prop_func_t) ((arg1)->get_wc_prop);
  {
    PyObject *py_pool = NULL;
    apr_pool_t *pool = NULL;
    
    if (svn_swig_py_get_parent_pool(args, SWIGTYPE_p_apr_pool_t,
        &py_pool, &pool))
    SWIG_fail;
    if (result == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    } else {
      svn_ra_get_wc_prop_func_t * tmp = apr_palloc(pool, sizeof(svn_ra_get_wc_prop_func_t));
      if (tmp == NULL) {
        SWIG_fail;
      }
      *tmp = (svn_ra_get_wc_prop_func_t) result;
      resultobj = svn_swig_py_new_pointer_obj(tmp, SWIGTYPE_p_p_f_p_void_p_q_const__char_p_q_const__char_p_p_q_const__svn_string_t_p_apr_pool_t__p_svn_error_t,
        py_pool, args);
    }
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_callbacks2_t_set_wc_prop_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_callbacks2_t *arg1 = 0 ;
  svn_ra_set_wc_prop_func_t arg2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_callbacks2_t_set_wc_prop_set", 2, 2, &obj0, &obj1)) SWIG_fail;
  {
    arg1 = (struct svn_ra_callbacks2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_callbacks2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    svn_ra_set_wc_prop_func_t * tmp =
    svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_p_f_p_void_p_q_const__char_p_q_const__char_p_q_const__svn_string_t_p_apr_pool_t__p_svn_error_t, svn_argnum_obj1);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg2 = *tmp;
  }
  if (arg1) (arg1)->set_wc_prop = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_callbacks2_t_set_wc_prop_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_callbacks2_t *arg1 = 0 ;
  PyObject * obj0 = 0 ;
  svn_ra_set_wc_prop_func_t result;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_callbacks2_t_set_wc_prop_get", 1, 1, &obj0)) SWIG_fail;
  {
    arg1 = (struct svn_ra_callbacks2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_callbacks2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_ra_set_wc_prop_func_t) ((arg1)->set_wc_prop);
  {
    PyObject *py_pool = NULL;
    apr_pool_t *pool = NULL;
    
    if (svn_swig_py_get_parent_pool(args, SWIGTYPE_p_apr_pool_t,
        &py_pool, &pool))
    SWIG_fail;
    if (result == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    } else {
      svn_ra_set_wc_prop_func_t * tmp = apr_palloc(pool, sizeof(svn_ra_set_wc_prop_func_t));
      if (tmp == NULL) {
        SWIG_fail;
      }
      *tmp = (svn_ra_set_wc_prop_func_t) result;
      resultobj = svn_swig_py_new_pointer_obj(tmp, SWIGTYPE_p_p_f_p_void_p_q_const__char_p_q_const__char_p_q_const__svn_string_t_p_apr_pool_t__p_svn_error_t,
        py_pool, args);
    }
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_callbacks2_t_push_wc_prop_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_callbacks2_t *arg1 = 0 ;
  svn_ra_push_wc_prop_func_t arg2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_callbacks2_t_push_wc_prop_set", 2, 2, &obj0, &obj1)) SWIG_fail;
  {
    arg1 = (struct svn_ra_callbacks2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_callbacks2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    svn_ra_push_wc_prop_func_t * tmp =
    svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_p_f_p_void_p_q_const__char_p_q_const__char_p_q_const__svn_string_t_p_apr_pool_t__p_svn_error_t, svn_argnum_obj1);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg2 = *tmp;
  }
  if (arg1) (arg1)->push_wc_prop = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_callbacks2_t_push_wc_prop_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_callbacks2_t *arg1 = 0 ;
  PyObject * obj0 = 0 ;
  svn_ra_push_wc_prop_func_t result;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_callbacks2_t_push_wc_prop_get", 1, 1, &obj0)) SWIG_fail;
  {
    arg1 = (struct svn_ra_callbacks2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_callbacks2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_ra_push_wc_prop_func_t) ((arg1)->push_wc_prop);
  {
    PyObject *py_pool = NULL;
    apr_pool_t *pool = NULL;
    
    if (svn_swig_py_get_parent_pool(args, SWIGTYPE_p_apr_pool_t,
        &py_pool, &pool))
    SWIG_fail;
    if (result == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    } else {
      svn_ra_push_wc_prop_func_t * tmp = apr_palloc(pool, sizeof(svn_ra_push_wc_prop_func_t));
      if (tmp == NULL) {
        SWIG_fail;
      }
      *tmp = (svn_ra_push_wc_prop_func_t) result;
      resultobj = svn_swig_py_new_pointer_obj(tmp, SWIGTYPE_p_p_f_p_void_p_q_const__char_p_q_const__char_p_q_const__svn_string_t_p_apr_pool_t__p_svn_error_t,
        py_pool, args);
    }
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_callbacks2_t_invalidate_wc_props_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_callbacks2_t *arg1 = 0 ;
  svn_ra_invalidate_wc_props_func_t arg2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_callbacks2_t_invalidate_wc_props_set", 2, 2, &obj0, &obj1)) SWIG_fail;
  {
    arg1 = (struct svn_ra_callbacks2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_callbacks2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    svn_ra_invalidate_wc_props_func_t * tmp =
    svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_p_f_p_void_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t, svn_argnum_obj1);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg2 = *tmp;
  }
  if (arg1) (arg1)->invalidate_wc_props = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_callbacks2_t_invalidate_wc_props_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_callbacks2_t *arg1 = 0 ;
  PyObject * obj0 = 0 ;
  svn_ra_invalidate_wc_props_func_t result;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_callbacks2_t_invalidate_wc_props_get", 1, 1, &obj0)) SWIG_fail;
  {
    arg1 = (struct svn_ra_callbacks2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_callbacks2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_ra_invalidate_wc_props_func_t) ((arg1)->invalidate_wc_props);
  {
    PyObject *py_pool = NULL;
    apr_pool_t *pool = NULL;
    
    if (svn_swig_py_get_parent_pool(args, SWIGTYPE_p_apr_pool_t,
        &py_pool, &pool))
    SWIG_fail;
    if (result == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    } else {
      svn_ra_invalidate_wc_props_func_t * tmp = apr_palloc(pool, sizeof(svn_ra_invalidate_wc_props_func_t));
      if (tmp == NULL) {
        SWIG_fail;
      }
      *tmp = (svn_ra_invalidate_wc_props_func_t) result;
      resultobj = svn_swig_py_new_pointer_obj(tmp, SWIGTYPE_p_p_f_p_void_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t,
        py_pool, args);
    }
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_callbacks2_t_progress_func_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_callbacks2_t *arg1 = 0 ;
  svn_ra_progress_notify_func_t arg2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_callbacks2_t_progress_func_set", 2, 2, &obj0, &obj1)) SWIG_fail;
  {
    arg1 = (struct svn_ra_callbacks2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_callbacks2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    svn_ra_progress_notify_func_t * tmp =
    svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_p_f_apr_off_t_apr_off_t_p_void_p_apr_pool_t__void, svn_argnum_obj1);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg2 = *tmp;
  }
  if (arg1) (arg1)->progress_func = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_callbacks2_t_progress_func_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_callbacks2_t *arg1 = 0 ;
  PyObject * obj0 = 0 ;
  svn_ra_progress_notify_func_t result;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_callbacks2_t_progress_func_get", 1, 1, &obj0)) SWIG_fail;
  {
    arg1 = (struct svn_ra_callbacks2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_callbacks2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_ra_progress_notify_func_t) ((arg1)->progress_func);
  {
    PyObject *py_pool = NULL;
    apr_pool_t *pool = NULL;
    
    if (svn_swig_py_get_parent_pool(args, SWIGTYPE_p_apr_pool_t,
        &py_pool, &pool))
    SWIG_fail;
    if (result == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    } else {
      svn_ra_progress_notify_func_t * tmp = apr_palloc(pool, sizeof(svn_ra_progress_notify_func_t));
      if (tmp == NULL) {
        SWIG_fail;
      }
      *tmp = (svn_ra_progress_notify_func_t) result;
      resultobj = svn_swig_py_new_pointer_obj(tmp, SWIGTYPE_p_p_f_apr_off_t_apr_off_t_p_void_p_apr_pool_t__void,
        py_pool, args);
    }
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_callbacks2_t_progress_baton_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_callbacks2_t *arg1 = 0 ;
  void *arg2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_callbacks2_t_progress_baton_set", 2, 2, &obj0, &obj1)) SWIG_fail;
  {
    arg1 = (struct svn_ra_callbacks2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_callbacks2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  if (arg1) (arg1)->progress_baton = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_callbacks2_t_progress_baton_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_callbacks2_t *arg1 = 0 ;
  PyObject * obj0 = 0 ;
  void *result = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_callbacks2_t_progress_baton_get", 1, 1, &obj0)) SWIG_fail;
  {
    arg1 = (struct svn_ra_callbacks2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_callbacks2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (void *) ((arg1)->progress_baton);
  {
    PyObject *ownerObj = obj0;
    PyObject *members = PyObject_GetAttrString(ownerObj, "_members");
    
    resultobj = NULL;
    if (members != NULL) {
      resultobj = PyDict_GetItemString(members, "progress_baton");
      Py_XINCREF(resultobj);
      Py_DECREF(members);
    }
    
    if (resultobj == NULL) {
      if (result == NULL) {
        resultobj = Py_None;
        Py_INCREF(resultobj);
      } else {
        /* We don't know the type of this reference, so we'll have to
               * treat it as an opaque void pointer.
               */
        resultobj = svn_swig_py_new_pointer_obj(result, SWIGTYPE_p_void,
          _global_py_pool, args);
      }
    }
    
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_callbacks2_t_cancel_func_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_callbacks2_t *arg1 = 0 ;
  svn_cancel_func_t arg2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_callbacks2_t_cancel_func_set", 2, 2, &obj0, &obj1)) SWIG_fail;
  {
    arg1 = (struct svn_ra_callbacks2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_callbacks2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    svn_cancel_func_t * tmp =
    svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_p_f_p_void__p_svn_error_t, svn_argnum_obj1);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg2 = *tmp;
  }
  if (arg1) (arg1)->cancel_func = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_callbacks2_t_cancel_func_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_callbacks2_t *arg1 = 0 ;
  PyObject * obj0 = 0 ;
  svn_cancel_func_t result;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_callbacks2_t_cancel_func_get", 1, 1, &obj0)) SWIG_fail;
  {
    arg1 = (struct svn_ra_callbacks2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_callbacks2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_cancel_func_t) ((arg1)->cancel_func);
  {
    PyObject *py_pool = NULL;
    apr_pool_t *pool = NULL;
    
    if (svn_swig_py_get_parent_pool(args, SWIGTYPE_p_apr_pool_t,
        &py_pool, &pool))
    SWIG_fail;
    if (result == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    } else {
      svn_cancel_func_t * tmp = apr_palloc(pool, sizeof(svn_cancel_func_t));
      if (tmp == NULL) {
        SWIG_fail;
      }
      *tmp = (svn_cancel_func_t) result;
      resultobj = svn_swig_py_new_pointer_obj(tmp, SWIGTYPE_p_p_f_p_void__p_svn_error_t,
        py_pool, args);
    }
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_callbacks2_t_get_client_string_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_callbacks2_t *arg1 = 0 ;
  svn_ra_get_client_string_func_t arg2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_callbacks2_t_get_client_string_set", 2, 2, &obj0, &obj1)) SWIG_fail;
  {
    arg1 = (struct svn_ra_callbacks2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_callbacks2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    svn_ra_get_client_string_func_t * tmp =
    svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_p_f_p_void_p_p_q_const__char_p_apr_pool_t__p_svn_error_t, svn_argnum_obj1);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg2 = *tmp;
  }
  if (arg1) (arg1)->get_client_string = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_callbacks2_t_get_client_string_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_callbacks2_t *arg1 = 0 ;
  PyObject * obj0 = 0 ;
  svn_ra_get_client_string_func_t result;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_callbacks2_t_get_client_string_get", 1, 1, &obj0)) SWIG_fail;
  {
    arg1 = (struct svn_ra_callbacks2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_callbacks2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_ra_get_client_string_func_t) ((arg1)->get_client_string);
  {
    PyObject *py_pool = NULL;
    apr_pool_t *pool = NULL;
    
    if (svn_swig_py_get_parent_pool(args, SWIGTYPE_p_apr_pool_t,
        &py_pool, &pool))
    SWIG_fail;
    if (result == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    } else {
      svn_ra_get_client_string_func_t * tmp = apr_palloc(pool, sizeof(svn_ra_get_client_string_func_t));
      if (tmp == NULL) {
        SWIG_fail;
      }
      *tmp = (svn_ra_get_client_string_func_t) result;
      resultobj = svn_swig_py_new_pointer_obj(tmp, SWIGTYPE_p_p_f_p_void_p_p_q_const__char_p_apr_pool_t__p_svn_error_t,
        py_pool, args);
    }
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_callbacks2_t_get_wc_contents_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_callbacks2_t *arg1 = 0 ;
  svn_ra_get_wc_contents_func_t arg2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_callbacks2_t_get_wc_contents_set", 2, 2, &obj0, &obj1)) SWIG_fail;
  {
    arg1 = (struct svn_ra_callbacks2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_callbacks2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    svn_ra_get_wc_contents_func_t * tmp =
    svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_p_f_p_void_p_p_svn_stream_t_p_q_const__svn_checksum_t_p_apr_pool_t__p_svn_error_t, svn_argnum_obj1);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg2 = *tmp;
  }
  if (arg1) (arg1)->get_wc_contents = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_callbacks2_t_get_wc_contents_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_callbacks2_t *arg1 = 0 ;
  PyObject * obj0 = 0 ;
  svn_ra_get_wc_contents_func_t result;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_callbacks2_t_get_wc_contents_get", 1, 1, &obj0)) SWIG_fail;
  {
    arg1 = (struct svn_ra_callbacks2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_callbacks2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_ra_get_wc_contents_func_t) ((arg1)->get_wc_contents);
  {
    PyObject *py_pool = NULL;
    apr_pool_t *pool = NULL;
    
    if (svn_swig_py_get_parent_pool(args, SWIGTYPE_p_apr_pool_t,
        &py_pool, &pool))
    SWIG_fail;
    if (result == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    } else {
      svn_ra_get_wc_contents_func_t * tmp = apr_palloc(pool, sizeof(svn_ra_get_wc_contents_func_t));
      if (tmp == NULL) {
        SWIG_fail;
      }
      *tmp = (svn_ra_get_wc_contents_func_t) result;
      resultobj = svn_swig_py_new_pointer_obj(tmp, SWIGTYPE_p_p_f_p_void_p_p_svn_stream_t_p_q_const__svn_checksum_t_p_apr_pool_t__p_svn_error_t,
        py_pool, args);
    }
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_callbacks2_t_check_tunnel_func_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_callbacks2_t *arg1 = 0 ;
  svn_ra_check_tunnel_func_t arg2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_callbacks2_t_check_tunnel_func_set", 2, 2, &obj0, &obj1)) SWIG_fail;
  {
    arg1 = (struct svn_ra_callbacks2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_callbacks2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    svn_ra_check_tunnel_func_t * tmp =
    svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_p_f_p_void_p_q_const__char__int, svn_argnum_obj1);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg2 = *tmp;
  }
  if (arg1) (arg1)->check_tunnel_func = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_callbacks2_t_check_tunnel_func_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_callbacks2_t *arg1 = 0 ;
  PyObject * obj0 = 0 ;
  svn_ra_check_tunnel_func_t result;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_callbacks2_t_check_tunnel_func_get", 1, 1, &obj0)) SWIG_fail;
  {
    arg1 = (struct svn_ra_callbacks2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_callbacks2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_ra_check_tunnel_func_t) ((arg1)->check_tunnel_func);
  {
    PyObject *py_pool = NULL;
    apr_pool_t *pool = NULL;
    
    if (svn_swig_py_get_parent_pool(args, SWIGTYPE_p_apr_pool_t,
        &py_pool, &pool))
    SWIG_fail;
    if (result == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    } else {
      svn_ra_check_tunnel_func_t * tmp = apr_palloc(pool, sizeof(svn_ra_check_tunnel_func_t));
      if (tmp == NULL) {
        SWIG_fail;
      }
      *tmp = (svn_ra_check_tunnel_func_t) result;
      resultobj = svn_swig_py_new_pointer_obj(tmp, SWIGTYPE_p_p_f_p_void_p_q_const__char__int,
        py_pool, args);
    }
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_callbacks2_t_open_tunnel_func_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_callbacks2_t *arg1 = 0 ;
  svn_ra_open_tunnel_func_t arg2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_callbacks2_t_open_tunnel_func_set", 2, 2, &obj0, &obj1)) SWIG_fail;
  {
    arg1 = (struct svn_ra_callbacks2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_callbacks2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    svn_ra_open_tunnel_func_t * tmp =
    svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_p_f_p_p_svn_stream_t_p_p_svn_stream_t_p_svn_ra_close_tunnel_func_t_p_p_void_p_void_p_q_const__char_p_q_const__char_p_q_const__char_int_svn_cancel_func_t_p_void_p_apr_pool_t__p_svn_error_t, svn_argnum_obj1);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg2 = *tmp;
  }
  if (arg1) (arg1)->open_tunnel_func = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_callbacks2_t_open_tunnel_func_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_callbacks2_t *arg1 = 0 ;
  PyObject * obj0 = 0 ;
  svn_ra_open_tunnel_func_t result;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_callbacks2_t_open_tunnel_func_get", 1, 1, &obj0)) SWIG_fail;
  {
    arg1 = (struct svn_ra_callbacks2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_callbacks2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_ra_open_tunnel_func_t) ((arg1)->open_tunnel_func);
  {
    PyObject *py_pool = NULL;
    apr_pool_t *pool = NULL;
    
    if (svn_swig_py_get_parent_pool(args, SWIGTYPE_p_apr_pool_t,
        &py_pool, &pool))
    SWIG_fail;
    if (result == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    } else {
      svn_ra_open_tunnel_func_t * tmp = apr_palloc(pool, sizeof(svn_ra_open_tunnel_func_t));
      if (tmp == NULL) {
        SWIG_fail;
      }
      *tmp = (svn_ra_open_tunnel_func_t) result;
      resultobj = svn_swig_py_new_pointer_obj(tmp, SWIGTYPE_p_p_f_p_p_svn_stream_t_p_p_svn_stream_t_p_svn_ra_close_tunnel_func_t_p_p_void_p_void_p_q_const__char_p_q_const__char_p_q_const__char_int_svn_cancel_func_t_p_void_p_apr_pool_t__p_svn_error_t,
        py_pool, args);
    }
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_callbacks2_t_tunnel_baton_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_callbacks2_t *arg1 = 0 ;
  void *arg2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_callbacks2_t_tunnel_baton_set", 2, 2, &obj0, &obj1)) SWIG_fail;
  {
    arg1 = (struct svn_ra_callbacks2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_callbacks2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  if (arg1) (arg1)->tunnel_baton = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_callbacks2_t_tunnel_baton_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_callbacks2_t *arg1 = 0 ;
  PyObject * obj0 = 0 ;
  void *result = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_callbacks2_t_tunnel_baton_get", 1, 1, &obj0)) SWIG_fail;
  {
    arg1 = (struct svn_ra_callbacks2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_callbacks2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (void *) ((arg1)->tunnel_baton);
  {
    PyObject *ownerObj = obj0;
    PyObject *members = PyObject_GetAttrString(ownerObj, "_members");
    
    resultobj = NULL;
    if (members != NULL) {
      resultobj = PyDict_GetItemString(members, "tunnel_baton");
      Py_XINCREF(resultobj);
      Py_DECREF(members);
    }
    
    if (resultobj == NULL) {
      if (result == NULL) {
        resultobj = Py_None;
        Py_INCREF(resultobj);
      } else {
        /* We don't know the type of this reference, so we'll have to
               * treat it as an opaque void pointer.
               */
        resultobj = svn_swig_py_new_pointer_obj(result, SWIGTYPE_p_void,
          _global_py_pool, args);
      }
    }
    
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_svn_ra_callbacks2_t(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_callbacks2_t *result = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "new_svn_ra_callbacks2_t", 0, 0)) SWIG_fail;
  {
    svn_swig_py_release_py_lock();
    
    result = (struct svn_ra_callbacks2_t *)calloc(1, sizeof(struct svn_ra_callbacks2_t));
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_ra_callbacks2_t,
    _global_py_pool, args);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_svn_ra_callbacks2_t(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_callbacks2_t *arg1 = 0 ;
  PyObject * obj0 = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "delete_svn_ra_callbacks2_t", 1, 1, &obj0)) SWIG_fail;
  {
    arg1 = (struct svn_ra_callbacks2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_callbacks2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    free((char *) arg1);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *svn_ra_callbacks2_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_svn_ra_callbacks2_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *svn_ra_callbacks2_t_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_svn_ra_callbacks_t_open_tmp_file_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_callbacks_t *arg1 = 0 ;
  svn_error_t *(*arg2)(apr_file_t **,void *,apr_pool_t *) = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_callbacks_t_open_tmp_file_set", 2, 2, &obj0, &obj1)) SWIG_fail;
  {
    arg1 = (struct svn_ra_callbacks_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_callbacks_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_p_apr_file_t_p_void_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_ra_callbacks_t_open_tmp_file_set" "', argument " "2"" of type '" "svn_error_t *(*)(apr_file_t **,void *,apr_pool_t *)""'"); 
    }
  }
  if (arg1) (arg1)->open_tmp_file = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_callbacks_t_open_tmp_file_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_callbacks_t *arg1 = 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(apr_file_t **,void *,apr_pool_t *) = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_callbacks_t_open_tmp_file_get", 1, 1, &obj0)) SWIG_fail;
  {
    arg1 = (struct svn_ra_callbacks_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_callbacks_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(apr_file_t **,void *,apr_pool_t *)) ((arg1)->open_tmp_file);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_p_apr_file_t_p_void_p_apr_pool_t__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_callbacks_t_auth_baton_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_callbacks_t *arg1 = 0 ;
  svn_auth_baton_t *arg2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_callbacks_t_auth_baton_set", 2, 2, &obj0, &obj1)) SWIG_fail;
  {
    arg1 = (struct svn_ra_callbacks_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_callbacks_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_auth_baton_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_auth_baton_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (arg1) (arg1)->auth_baton = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_callbacks_t_auth_baton_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_callbacks_t *arg1 = 0 ;
  PyObject * obj0 = 0 ;
  svn_auth_baton_t *result = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_callbacks_t_auth_baton_get", 1, 1, &obj0)) SWIG_fail;
  {
    arg1 = (struct svn_ra_callbacks_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_callbacks_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_auth_baton_t *) ((arg1)->auth_baton);
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_auth_baton_t,
    _global_py_pool, args);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_callbacks_t_get_wc_prop_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_callbacks_t *arg1 = 0 ;
  svn_ra_get_wc_prop_func_t arg2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_callbacks_t_get_wc_prop_set", 2, 2, &obj0, &obj1)) SWIG_fail;
  {
    arg1 = (struct svn_ra_callbacks_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_callbacks_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    svn_ra_get_wc_prop_func_t * tmp =
    svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_p_f_p_void_p_q_const__char_p_q_const__char_p_p_q_const__svn_string_t_p_apr_pool_t__p_svn_error_t, svn_argnum_obj1);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg2 = *tmp;
  }
  if (arg1) (arg1)->get_wc_prop = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_callbacks_t_get_wc_prop_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_callbacks_t *arg1 = 0 ;
  PyObject * obj0 = 0 ;
  svn_ra_get_wc_prop_func_t result;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_callbacks_t_get_wc_prop_get", 1, 1, &obj0)) SWIG_fail;
  {
    arg1 = (struct svn_ra_callbacks_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_callbacks_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_ra_get_wc_prop_func_t) ((arg1)->get_wc_prop);
  {
    PyObject *py_pool = NULL;
    apr_pool_t *pool = NULL;
    
    if (svn_swig_py_get_parent_pool(args, SWIGTYPE_p_apr_pool_t,
        &py_pool, &pool))
    SWIG_fail;
    if (result == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    } else {
      svn_ra_get_wc_prop_func_t * tmp = apr_palloc(pool, sizeof(svn_ra_get_wc_prop_func_t));
      if (tmp == NULL) {
        SWIG_fail;
      }
      *tmp = (svn_ra_get_wc_prop_func_t) result;
      resultobj = svn_swig_py_new_pointer_obj(tmp, SWIGTYPE_p_p_f_p_void_p_q_const__char_p_q_const__char_p_p_q_const__svn_string_t_p_apr_pool_t__p_svn_error_t,
        py_pool, args);
    }
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_callbacks_t_set_wc_prop_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_callbacks_t *arg1 = 0 ;
  svn_ra_set_wc_prop_func_t arg2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_callbacks_t_set_wc_prop_set", 2, 2, &obj0, &obj1)) SWIG_fail;
  {
    arg1 = (struct svn_ra_callbacks_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_callbacks_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    svn_ra_set_wc_prop_func_t * tmp =
    svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_p_f_p_void_p_q_const__char_p_q_const__char_p_q_const__svn_string_t_p_apr_pool_t__p_svn_error_t, svn_argnum_obj1);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg2 = *tmp;
  }
  if (arg1) (arg1)->set_wc_prop = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_callbacks_t_set_wc_prop_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_callbacks_t *arg1 = 0 ;
  PyObject * obj0 = 0 ;
  svn_ra_set_wc_prop_func_t result;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_callbacks_t_set_wc_prop_get", 1, 1, &obj0)) SWIG_fail;
  {
    arg1 = (struct svn_ra_callbacks_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_callbacks_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_ra_set_wc_prop_func_t) ((arg1)->set_wc_prop);
  {
    PyObject *py_pool = NULL;
    apr_pool_t *pool = NULL;
    
    if (svn_swig_py_get_parent_pool(args, SWIGTYPE_p_apr_pool_t,
        &py_pool, &pool))
    SWIG_fail;
    if (result == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    } else {
      svn_ra_set_wc_prop_func_t * tmp = apr_palloc(pool, sizeof(svn_ra_set_wc_prop_func_t));
      if (tmp == NULL) {
        SWIG_fail;
      }
      *tmp = (svn_ra_set_wc_prop_func_t) result;
      resultobj = svn_swig_py_new_pointer_obj(tmp, SWIGTYPE_p_p_f_p_void_p_q_const__char_p_q_const__char_p_q_const__svn_string_t_p_apr_pool_t__p_svn_error_t,
        py_pool, args);
    }
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_callbacks_t_push_wc_prop_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_callbacks_t *arg1 = 0 ;
  svn_ra_push_wc_prop_func_t arg2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_callbacks_t_push_wc_prop_set", 2, 2, &obj0, &obj1)) SWIG_fail;
  {
    arg1 = (struct svn_ra_callbacks_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_callbacks_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    svn_ra_push_wc_prop_func_t * tmp =
    svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_p_f_p_void_p_q_const__char_p_q_const__char_p_q_const__svn_string_t_p_apr_pool_t__p_svn_error_t, svn_argnum_obj1);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg2 = *tmp;
  }
  if (arg1) (arg1)->push_wc_prop = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_callbacks_t_push_wc_prop_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_callbacks_t *arg1 = 0 ;
  PyObject * obj0 = 0 ;
  svn_ra_push_wc_prop_func_t result;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_callbacks_t_push_wc_prop_get", 1, 1, &obj0)) SWIG_fail;
  {
    arg1 = (struct svn_ra_callbacks_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_callbacks_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_ra_push_wc_prop_func_t) ((arg1)->push_wc_prop);
  {
    PyObject *py_pool = NULL;
    apr_pool_t *pool = NULL;
    
    if (svn_swig_py_get_parent_pool(args, SWIGTYPE_p_apr_pool_t,
        &py_pool, &pool))
    SWIG_fail;
    if (result == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    } else {
      svn_ra_push_wc_prop_func_t * tmp = apr_palloc(pool, sizeof(svn_ra_push_wc_prop_func_t));
      if (tmp == NULL) {
        SWIG_fail;
      }
      *tmp = (svn_ra_push_wc_prop_func_t) result;
      resultobj = svn_swig_py_new_pointer_obj(tmp, SWIGTYPE_p_p_f_p_void_p_q_const__char_p_q_const__char_p_q_const__svn_string_t_p_apr_pool_t__p_svn_error_t,
        py_pool, args);
    }
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_callbacks_t_invalidate_wc_props_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_callbacks_t *arg1 = 0 ;
  svn_ra_invalidate_wc_props_func_t arg2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_callbacks_t_invalidate_wc_props_set", 2, 2, &obj0, &obj1)) SWIG_fail;
  {
    arg1 = (struct svn_ra_callbacks_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_callbacks_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    svn_ra_invalidate_wc_props_func_t * tmp =
    svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_p_f_p_void_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t, svn_argnum_obj1);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg2 = *tmp;
  }
  if (arg1) (arg1)->invalidate_wc_props = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_callbacks_t_invalidate_wc_props_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_callbacks_t *arg1 = 0 ;
  PyObject * obj0 = 0 ;
  svn_ra_invalidate_wc_props_func_t result;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_callbacks_t_invalidate_wc_props_get", 1, 1, &obj0)) SWIG_fail;
  {
    arg1 = (struct svn_ra_callbacks_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_callbacks_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_ra_invalidate_wc_props_func_t) ((arg1)->invalidate_wc_props);
  {
    PyObject *py_pool = NULL;
    apr_pool_t *pool = NULL;
    
    if (svn_swig_py_get_parent_pool(args, SWIGTYPE_p_apr_pool_t,
        &py_pool, &pool))
    SWIG_fail;
    if (result == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    } else {
      svn_ra_invalidate_wc_props_func_t * tmp = apr_palloc(pool, sizeof(svn_ra_invalidate_wc_props_func_t));
      if (tmp == NULL) {
        SWIG_fail;
      }
      *tmp = (svn_ra_invalidate_wc_props_func_t) result;
      resultobj = svn_swig_py_new_pointer_obj(tmp, SWIGTYPE_p_p_f_p_void_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t,
        py_pool, args);
    }
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_svn_ra_callbacks_t(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_callbacks_t *result = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "new_svn_ra_callbacks_t", 0, 0)) SWIG_fail;
  {
    svn_swig_py_release_py_lock();
    
    result = (struct svn_ra_callbacks_t *)calloc(1, sizeof(struct svn_ra_callbacks_t));
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_ra_callbacks_t,
    _global_py_pool, args);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_svn_ra_callbacks_t(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_callbacks_t *arg1 = 0 ;
  PyObject * obj0 = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "delete_svn_ra_callbacks_t", 1, 1, &obj0)) SWIG_fail;
  {
    arg1 = (struct svn_ra_callbacks_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_callbacks_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    free((char *) arg1);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *svn_ra_callbacks_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_svn_ra_callbacks_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *svn_ra_callbacks_t_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_svn_ra_initialize(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  apr_pool_t *arg1 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg1 = _global_pool;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_initialize", 0, 1, &obj0)) SWIG_fail;
  if (obj0) {
    /* Verify that the user supplied a valid pool */
    if (obj0 != Py_None && obj0 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj0);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_initialize(arg1);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_create_callbacks(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_callbacks2_t **arg1 = 0 ;
  apr_pool_t *arg2 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_ra_callbacks2_t *temp1 ;
  PyObject * obj0 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg2 = _global_pool;
  arg1 = &temp1;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_create_callbacks", 0, 1, &obj0)) SWIG_fail;
  if (obj0) {
    /* Verify that the user supplied a valid pool */
    if (obj0 != Py_None && obj0 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj0);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_create_callbacks(arg1,arg2);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg1, SWIGTYPE_p_svn_ra_callbacks2_t,
        _global_py_pool, args), 0)
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_open5(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_session_t **arg1 = 0 ;
  char **arg2 = 0 ;
  char **arg3 = 0 ;
  char *arg4 = 0 ;
  char *arg5 = 0 ;
  svn_ra_callbacks2_t *arg6 = 0 ;
  void *arg7 = 0 ;
  apr_hash_t *arg8 = 0 ;
  apr_pool_t *arg9 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_ra_session_t *temp1 ;
  char *temp2 ;
  char *temp3 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg9 = _global_pool;
  arg1 = &temp1;
  arg2 = &temp2;
  arg3 = &temp3;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_open5", 4, 5, &obj0, &obj1, &obj2, &obj3, &obj4)) SWIG_fail;
  {
    arg4 = svn_swig_py_string_to_cstring(obj0, FALSE, "svn_ra_open5", "repos_URL");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    arg5 = svn_swig_py_string_to_cstring(obj1, FALSE, "svn_ra_open5", "uuid");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    svn_swig_py_setup_ra_callbacks(&arg6, &arg7, obj2, _global_pool);
  }
  {
    if (_global_pool == NULL)
    {
      if (svn_swig_py_get_parent_pool(args, SWIGTYPE_p_apr_pool_t,
          &_global_py_pool, &_global_pool))
      SWIG_fail;
    }
    
    arg8 = svn_swig_py_struct_ptr_hash_from_dict(obj3,
      SWIGTYPE_p_svn_config_t, _global_pool);
    if (PyErr_Occurred())
    SWIG_fail;
  }
  if (obj4) {
    /* Verify that the user supplied a valid pool */
    if (obj4 != Py_None && obj4 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj4);
      SWIG_fail;
    }
  }
  {
    if (!arg6) {
      SWIG_exception(SWIG_NullReferenceError,"Received a NULL pointer.");
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_open5(arg1,(char const **)arg2,(char const **)arg3,(char const *)arg4,(char const *)arg5,(struct svn_ra_callbacks2_t const *)arg6,arg7,arg8,arg9);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg1, SWIGTYPE_p_svn_ra_session_t,
        _global_py_pool, args), 0)
    
    ;
  }
  {
    PyObject *s;
    if (*arg2 == NULL) {
      Py_INCREF(Py_None);
      s = Py_None;
    }
    else {
      s = PyBytes_FromString(*arg2);
      if (s == NULL)
      SWIG_fail;
    }
    resultobj = SWIG_Python_AppendOutput(resultobj, s, 0);
  }
  {
    PyObject *s;
    if (*arg3 == NULL) {
      Py_INCREF(Py_None);
      s = Py_None;
    }
    else {
      s = PyBytes_FromString(*arg3);
      if (s == NULL)
      SWIG_fail;
    }
    resultobj = SWIG_Python_AppendOutput(resultobj, s, 0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_open4(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_session_t **arg1 = 0 ;
  char **arg2 = 0 ;
  char *arg3 = 0 ;
  char *arg4 = 0 ;
  svn_ra_callbacks2_t *arg5 = 0 ;
  void *arg6 = 0 ;
  apr_hash_t *arg7 = 0 ;
  apr_pool_t *arg8 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_ra_session_t *temp1 ;
  char *temp2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg8 = _global_pool;
  arg1 = &temp1;
  arg2 = &temp2;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_open4", 4, 5, &obj0, &obj1, &obj2, &obj3, &obj4)) SWIG_fail;
  {
    arg3 = svn_swig_py_string_to_cstring(obj0, FALSE, "svn_ra_open4", "repos_URL");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    arg4 = svn_swig_py_string_to_cstring(obj1, FALSE, "svn_ra_open4", "uuid");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    svn_swig_py_setup_ra_callbacks(&arg5, &arg6, obj2, _global_pool);
  }
  {
    if (_global_pool == NULL)
    {
      if (svn_swig_py_get_parent_pool(args, SWIGTYPE_p_apr_pool_t,
          &_global_py_pool, &_global_pool))
      SWIG_fail;
    }
    
    arg7 = svn_swig_py_struct_ptr_hash_from_dict(obj3,
      SWIGTYPE_p_svn_config_t, _global_pool);
    if (PyErr_Occurred())
    SWIG_fail;
  }
  if (obj4) {
    /* Verify that the user supplied a valid pool */
    if (obj4 != Py_None && obj4 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj4);
      SWIG_fail;
    }
  }
  {
    if (!arg5) {
      SWIG_exception(SWIG_NullReferenceError,"Received a NULL pointer.");
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_open4(arg1,(char const **)arg2,(char const *)arg3,(char const *)arg4,(struct svn_ra_callbacks2_t const *)arg5,arg6,arg7,arg8);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg1, SWIGTYPE_p_svn_ra_session_t,
        _global_py_pool, args), 0)
    
    ;
  }
  {
    PyObject *s;
    if (*arg2 == NULL) {
      Py_INCREF(Py_None);
      s = Py_None;
    }
    else {
      s = PyBytes_FromString(*arg2);
      if (s == NULL)
      SWIG_fail;
    }
    resultobj = SWIG_Python_AppendOutput(resultobj, s, 0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_open3(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_session_t **arg1 = 0 ;
  char *arg2 = 0 ;
  char *arg3 = 0 ;
  svn_ra_callbacks2_t *arg4 = 0 ;
  void *arg5 = 0 ;
  apr_hash_t *arg6 = 0 ;
  apr_pool_t *arg7 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_ra_session_t *temp1 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg7 = _global_pool;
  arg1 = &temp1;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_open3", 4, 5, &obj0, &obj1, &obj2, &obj3, &obj4)) SWIG_fail;
  {
    arg2 = svn_swig_py_string_to_cstring(obj0, FALSE, "svn_ra_open3", "repos_URL");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    arg3 = svn_swig_py_string_to_cstring(obj1, FALSE, "svn_ra_open3", "uuid");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    svn_swig_py_setup_ra_callbacks(&arg4, &arg5, obj2, _global_pool);
  }
  {
    if (_global_pool == NULL)
    {
      if (svn_swig_py_get_parent_pool(args, SWIGTYPE_p_apr_pool_t,
          &_global_py_pool, &_global_pool))
      SWIG_fail;
    }
    
    arg6 = svn_swig_py_struct_ptr_hash_from_dict(obj3,
      SWIGTYPE_p_svn_config_t, _global_pool);
    if (PyErr_Occurred())
    SWIG_fail;
  }
  if (obj4) {
    /* Verify that the user supplied a valid pool */
    if (obj4 != Py_None && obj4 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj4);
      SWIG_fail;
    }
  }
  {
    if (!arg4) {
      SWIG_exception(SWIG_NullReferenceError,"Received a NULL pointer.");
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_open3(arg1,(char const *)arg2,(char const *)arg3,(struct svn_ra_callbacks2_t const *)arg4,arg5,arg6,arg7);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg1, SWIGTYPE_p_svn_ra_session_t,
        _global_py_pool, args), 0)
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_open2(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_session_t **arg1 = 0 ;
  char *arg2 = 0 ;
  svn_ra_callbacks2_t *arg3 = 0 ;
  void *arg4 = 0 ;
  apr_hash_t *arg5 = 0 ;
  apr_pool_t *arg6 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_ra_session_t *temp1 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg6 = _global_pool;
  arg1 = &temp1;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_open2", 3, 4, &obj0, &obj1, &obj2, &obj3)) SWIG_fail;
  {
    arg2 = svn_swig_py_string_to_cstring(obj0, FALSE, "svn_ra_open2", "repos_URL");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    svn_swig_py_setup_ra_callbacks(&arg3, &arg4, obj1, _global_pool);
  }
  {
    if (_global_pool == NULL)
    {
      if (svn_swig_py_get_parent_pool(args, SWIGTYPE_p_apr_pool_t,
          &_global_py_pool, &_global_pool))
      SWIG_fail;
    }
    
    arg5 = svn_swig_py_struct_ptr_hash_from_dict(obj2,
      SWIGTYPE_p_svn_config_t, _global_pool);
    if (PyErr_Occurred())
    SWIG_fail;
  }
  if (obj3) {
    /* Verify that the user supplied a valid pool */
    if (obj3 != Py_None && obj3 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj3);
      SWIG_fail;
    }
  }
  {
    if (!arg3) {
      SWIG_exception(SWIG_NullReferenceError,"Received a NULL pointer.");
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_open2(arg1,(char const *)arg2,(struct svn_ra_callbacks2_t const *)arg3,arg4,arg5,arg6);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg1, SWIGTYPE_p_svn_ra_session_t,
        _global_py_pool, args), 0)
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_open(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_session_t **arg1 = 0 ;
  char *arg2 = 0 ;
  svn_ra_callbacks_t *arg3 = 0 ;
  void *arg4 = 0 ;
  apr_hash_t *arg5 = 0 ;
  apr_pool_t *arg6 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_ra_session_t *temp1 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg6 = _global_pool;
  arg1 = &temp1;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_open", 4, 5, &obj0, &obj1, &obj2, &obj3, &obj4)) SWIG_fail;
  {
    arg2 = svn_swig_py_string_to_cstring(obj0, FALSE, "svn_ra_open", "repos_URL");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    arg3 = (svn_ra_callbacks_t *)svn_swig_py_must_get_ptr(obj1, SWIGTYPE_p_svn_ra_callbacks_t, svn_argnum_obj1);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj2 == Py_None) {
      arg4 = NULL;
    } else if (SWIG_ConvertPtr(obj2, (void **) &arg4, 0, 0) == -1) {
      arg4 = (void *) obj2;
      PyErr_Clear();
    }
  }
  {
    if (_global_pool == NULL)
    {
      if (svn_swig_py_get_parent_pool(args, SWIGTYPE_p_apr_pool_t,
          &_global_py_pool, &_global_pool))
      SWIG_fail;
    }
    
    arg5 = svn_swig_py_struct_ptr_hash_from_dict(obj3,
      SWIGTYPE_p_svn_config_t, _global_pool);
    if (PyErr_Occurred())
    SWIG_fail;
  }
  if (obj4) {
    /* Verify that the user supplied a valid pool */
    if (obj4 != Py_None && obj4 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj4);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_open(arg1,(char const *)arg2,(struct svn_ra_callbacks_t const *)arg3,arg4,arg5,arg6);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg1, SWIGTYPE_p_svn_ra_session_t,
        _global_py_pool, args), 0)
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_reparent(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_session_t *arg1 = 0 ;
  char *arg2 = 0 ;
  apr_pool_t *arg3 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg3 = _global_pool;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_reparent", 2, 3, &obj0, &obj1, &obj2)) SWIG_fail;
  {
    arg1 = (svn_ra_session_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_session_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = svn_swig_py_string_to_cstring(obj1, FALSE, "svn_ra_reparent", "url");
    if (PyErr_Occurred()) SWIG_fail;
  }
  if (obj2) {
    /* Verify that the user supplied a valid pool */
    if (obj2 != Py_None && obj2 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj2);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_reparent(arg1,(char const *)arg2,arg3);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_get_session_url(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_session_t *arg1 = 0 ;
  char **arg2 = 0 ;
  apr_pool_t *arg3 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  char *temp2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg3 = _global_pool;
  arg2 = &temp2;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_get_session_url", 1, 2, &obj0, &obj1)) SWIG_fail;
  {
    arg1 = (svn_ra_session_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_session_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj1) {
    /* Verify that the user supplied a valid pool */
    if (obj1 != Py_None && obj1 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj1);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_get_session_url(arg1,(char const **)arg2,arg3);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    PyObject *s;
    if (*arg2 == NULL) {
      Py_INCREF(Py_None);
      s = Py_None;
    }
    else {
      s = PyBytes_FromString(*arg2);
      if (s == NULL)
      SWIG_fail;
    }
    resultobj = SWIG_Python_AppendOutput(resultobj, s, 0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_get_path_relative_to_session(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_session_t *arg1 = 0 ;
  char **arg2 = 0 ;
  char *arg3 = 0 ;
  apr_pool_t *arg4 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  char *temp2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg4 = _global_pool;
  arg2 = &temp2;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_get_path_relative_to_session", 2, 3, &obj0, &obj1, &obj2)) SWIG_fail;
  {
    arg1 = (svn_ra_session_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_session_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = svn_swig_py_string_to_cstring(obj1, FALSE, "svn_ra_get_path_relative_to_session", "url");
    if (PyErr_Occurred()) SWIG_fail;
  }
  if (obj2) {
    /* Verify that the user supplied a valid pool */
    if (obj2 != Py_None && obj2 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj2);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_get_path_relative_to_session(arg1,(char const **)arg2,(char const *)arg3,arg4);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    PyObject *s;
    if (*arg2 == NULL) {
      Py_INCREF(Py_None);
      s = Py_None;
    }
    else {
      s = PyBytes_FromString(*arg2);
      if (s == NULL)
      SWIG_fail;
    }
    resultobj = SWIG_Python_AppendOutput(resultobj, s, 0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_get_path_relative_to_root(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_session_t *arg1 = 0 ;
  char **arg2 = 0 ;
  char *arg3 = 0 ;
  apr_pool_t *arg4 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  char *temp2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg4 = _global_pool;
  arg2 = &temp2;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_get_path_relative_to_root", 2, 3, &obj0, &obj1, &obj2)) SWIG_fail;
  {
    arg1 = (svn_ra_session_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_session_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = svn_swig_py_string_to_cstring(obj1, FALSE, "svn_ra_get_path_relative_to_root", "url");
    if (PyErr_Occurred()) SWIG_fail;
  }
  if (obj2) {
    /* Verify that the user supplied a valid pool */
    if (obj2 != Py_None && obj2 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj2);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_get_path_relative_to_root(arg1,(char const **)arg2,(char const *)arg3,arg4);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    PyObject *s;
    if (*arg2 == NULL) {
      Py_INCREF(Py_None);
      s = Py_None;
    }
    else {
      s = PyBytes_FromString(*arg2);
      if (s == NULL)
      SWIG_fail;
    }
    resultobj = SWIG_Python_AppendOutput(resultobj, s, 0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_get_latest_revnum(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_session_t *arg1 = 0 ;
  svn_revnum_t *arg2 = 0 ;
  apr_pool_t *arg3 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_revnum_t temp2 ;
  int res2 = SWIG_TMPOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg3 = _global_pool;
  arg2 = &temp2;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_get_latest_revnum", 1, 2, &obj0, &obj1)) SWIG_fail;
  {
    arg1 = (svn_ra_session_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_session_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj1) {
    /* Verify that the user supplied a valid pool */
    if (obj1 != Py_None && obj1 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj1);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_get_latest_revnum(arg1,arg2,arg3);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  if (SWIG_IsTmpObj(res2)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_long((*arg2)), 0);
  } else {
    int new_flags = SWIG_IsNewObj(res2) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg2), SWIGTYPE_p_long, new_flags), 0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_get_dated_revision(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_session_t *arg1 = 0 ;
  svn_revnum_t *arg2 = 0 ;
  apr_time_t arg3 ;
  apr_pool_t *arg4 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_revnum_t temp2 ;
  int res2 = SWIG_TMPOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg4 = _global_pool;
  arg2 = &temp2;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_get_dated_revision", 2, 3, &obj0, &obj1, &obj2)) SWIG_fail;
  {
    arg1 = (svn_ra_session_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_session_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  arg3 = (apr_time_t) PyLong_AsLongLong(obj1); 
  if (obj2) {
    /* Verify that the user supplied a valid pool */
    if (obj2 != Py_None && obj2 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj2);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_get_dated_revision(arg1,arg2,arg3,arg4);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  if (SWIG_IsTmpObj(res2)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_long((*arg2)), 0);
  } else {
    int new_flags = SWIG_IsNewObj(res2) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg2), SWIGTYPE_p_long, new_flags), 0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_change_rev_prop2(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_session_t *arg1 = 0 ;
  svn_revnum_t arg2 ;
  char *arg3 = 0 ;
  svn_string_t **arg4 = 0 ;
  svn_string_t *arg5 = 0 ;
  apr_pool_t *arg6 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_string_t value5 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg6 = _global_pool;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_change_rev_prop2", 5, 6, &obj0, &obj1, &obj2, &obj3, &obj4, &obj5)) SWIG_fail;
  {
    arg1 = (svn_ra_session_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_session_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_revnum_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  {
    arg3 = svn_swig_py_string_to_cstring(obj2, FALSE, "svn_ra_change_rev_prop2", "name");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    SWIG_exception(SWIG_ValueError, "svn_ra_change_rev_prop2 is not implemented yet");
    
  }
  {
    if (obj4 == Py_None)
    arg5 = NULL;
    else {
      Py_ssize_t pyStrLen;
      if (!PyBytes_Check(obj4)) {
        PyErr_SetString(PyExc_TypeError, "not a bytes object");
        SWIG_fail;
      }
      if (PyBytes_AsStringAndSize(obj4, (char **)&(value5.data),
          &pyStrLen) == -1) {
        SWIG_fail;
      }
      value5.len = pyStrLen;
      arg5 = &value5;
    }
  }
  if (obj5) {
    /* Verify that the user supplied a valid pool */
    if (obj5 != Py_None && obj5 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj5);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_change_rev_prop2(arg1,arg2,(char const *)arg3,(struct svn_string_t const *const *)arg4,(struct svn_string_t const *)arg5,arg6);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    PyObject *s;
    if (*arg4 == NULL) {
      Py_INCREF(Py_None);
      s = Py_None;
    } else {
      s = PyBytes_FromStringAndSize((*arg4)->data, (*arg4)->len);
      if (s == NULL)
      SWIG_fail;
    }
    resultobj = SWIG_Python_AppendOutput(resultobj, s, 0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_change_rev_prop(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_session_t *arg1 = 0 ;
  svn_revnum_t arg2 ;
  char *arg3 = 0 ;
  svn_string_t *arg4 = 0 ;
  apr_pool_t *arg5 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_string_t value4 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg5 = _global_pool;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_change_rev_prop", 4, 5, &obj0, &obj1, &obj2, &obj3, &obj4)) SWIG_fail;
  {
    arg1 = (svn_ra_session_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_session_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_revnum_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  {
    arg3 = svn_swig_py_string_to_cstring(obj2, FALSE, "svn_ra_change_rev_prop", "name");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    if (obj3 == Py_None)
    arg4 = NULL;
    else {
      Py_ssize_t pyStrLen;
      if (!PyBytes_Check(obj3)) {
        PyErr_SetString(PyExc_TypeError, "not a bytes object");
        SWIG_fail;
      }
      if (PyBytes_AsStringAndSize(obj3, (char **)&(value4.data),
          &pyStrLen) == -1) {
        SWIG_fail;
      }
      value4.len = pyStrLen;
      arg4 = &value4;
    }
  }
  if (obj4) {
    /* Verify that the user supplied a valid pool */
    if (obj4 != Py_None && obj4 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj4);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_change_rev_prop(arg1,arg2,(char const *)arg3,(struct svn_string_t const *)arg4,arg5);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_rev_proplist(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_session_t *arg1 = 0 ;
  svn_revnum_t arg2 ;
  apr_hash_t **arg3 = 0 ;
  apr_pool_t *arg4 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  apr_hash_t *temp3 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg4 = _global_pool;
  arg3 = &temp3;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_rev_proplist", 2, 3, &obj0, &obj1, &obj2)) SWIG_fail;
  {
    arg1 = (svn_ra_session_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_session_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_revnum_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (obj2) {
    /* Verify that the user supplied a valid pool */
    if (obj2 != Py_None && obj2 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj2);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_rev_proplist(arg1,arg2,arg3,arg4);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_prophash_to_dict(*arg3), 0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_rev_prop(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_session_t *arg1 = 0 ;
  svn_revnum_t arg2 ;
  char *arg3 = 0 ;
  svn_string_t **arg4 = 0 ;
  apr_pool_t *arg5 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_string_t *temp4 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg5 = _global_pool;
  arg4 = &temp4;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_rev_prop", 3, 4, &obj0, &obj1, &obj2, &obj3)) SWIG_fail;
  {
    arg1 = (svn_ra_session_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_session_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_revnum_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  {
    arg3 = svn_swig_py_string_to_cstring(obj2, FALSE, "svn_ra_rev_prop", "name");
    if (PyErr_Occurred()) SWIG_fail;
  }
  if (obj3) {
    /* Verify that the user supplied a valid pool */
    if (obj3 != Py_None && obj3 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj3);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_rev_prop(arg1,arg2,(char const *)arg3,arg4,arg5);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    PyObject *s;
    if (*arg4 == NULL) {
      Py_INCREF(Py_None);
      s = Py_None;
    } else {
      s = PyBytes_FromStringAndSize((*arg4)->data, (*arg4)->len);
      if (s == NULL)
      SWIG_fail;
    }
    resultobj = SWIG_Python_AppendOutput(resultobj, s, 0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_get_commit_editor3(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_session_t *arg1 = 0 ;
  svn_delta_editor_t **arg2 = 0 ;
  void **arg3 = 0 ;
  apr_hash_t *arg4 = 0 ;
  svn_commit_callback2_t arg5 = 0 ;
  void *arg6 = 0 ;
  apr_hash_t *arg7 = 0 ;
  svn_boolean_t arg8 ;
  apr_pool_t *arg9 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_delta_editor_t *temp2 ;
  void *temp3 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg9 = _global_pool;
  arg2 = &temp2;
  arg3 = &temp3;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_get_commit_editor3", 5, 6, &obj0, &obj1, &obj2, &obj3, &obj4, &obj5)) SWIG_fail;
  {
    arg1 = (svn_ra_session_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_session_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (_global_pool == NULL)
    {
      if (svn_swig_py_get_parent_pool(args, SWIGTYPE_p_apr_pool_t,
          &_global_py_pool, &_global_pool))
      SWIG_fail;
    }
    
    arg4 = svn_swig_py_prophash_from_dict(obj1, _global_pool);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg5 = svn_swig_py_commit_callback2;
    arg6 = (void *)obj2;
  }
  {
    /* PYTHON-FIXME: Handle None -> NULL. */
    arg7 = svn_swig_py_stringhash_from_dict(obj3, _global_pool);
  }
  {
    arg8 = (svn_boolean_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  if (obj5) {
    /* Verify that the user supplied a valid pool */
    if (obj5 != Py_None && obj5 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj5);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_get_commit_editor3(arg1,(struct svn_delta_editor_t const **)arg2,arg3,arg4,arg5,arg6,arg7,arg8,arg9);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg2, SWIGTYPE_p_svn_delta_editor_t,
        _global_py_pool, args), 0)
    
    ;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg3, SWIGTYPE_p_void,
        _global_py_pool, args), 0)
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_get_commit_editor2(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_session_t *arg1 = 0 ;
  svn_delta_editor_t **arg2 = 0 ;
  void **arg3 = 0 ;
  char *arg4 = 0 ;
  svn_commit_callback2_t arg5 = 0 ;
  void *arg6 = 0 ;
  apr_hash_t *arg7 = 0 ;
  svn_boolean_t arg8 ;
  apr_pool_t *arg9 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_delta_editor_t *temp2 ;
  void *temp3 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg9 = _global_pool;
  arg2 = &temp2;
  arg3 = &temp3;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_get_commit_editor2", 5, 6, &obj0, &obj1, &obj2, &obj3, &obj4, &obj5)) SWIG_fail;
  {
    arg1 = (svn_ra_session_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_session_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg4 = svn_swig_py_string_to_cstring(obj1, FALSE, "svn_ra_get_commit_editor2", "log_msg");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    arg5 = svn_swig_py_commit_callback2;
    arg6 = (void *)obj2;
  }
  {
    /* PYTHON-FIXME: Handle None -> NULL. */
    arg7 = svn_swig_py_stringhash_from_dict(obj3, _global_pool);
  }
  {
    arg8 = (svn_boolean_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  if (obj5) {
    /* Verify that the user supplied a valid pool */
    if (obj5 != Py_None && obj5 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj5);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_get_commit_editor2(arg1,(struct svn_delta_editor_t const **)arg2,arg3,(char const *)arg4,arg5,arg6,arg7,arg8,arg9);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg2, SWIGTYPE_p_svn_delta_editor_t,
        _global_py_pool, args), 0)
    
    ;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg3, SWIGTYPE_p_void,
        _global_py_pool, args), 0)
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_get_commit_editor(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_session_t *arg1 = 0 ;
  svn_delta_editor_t **arg2 = 0 ;
  void **arg3 = 0 ;
  char *arg4 = 0 ;
  svn_commit_callback_t arg5 = 0 ;
  void *arg6 = 0 ;
  apr_hash_t *arg7 = 0 ;
  svn_boolean_t arg8 ;
  apr_pool_t *arg9 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_delta_editor_t *temp2 ;
  void *temp3 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg9 = _global_pool;
  arg2 = &temp2;
  arg3 = &temp3;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_get_commit_editor", 5, 6, &obj0, &obj1, &obj2, &obj3, &obj4, &obj5)) SWIG_fail;
  {
    arg1 = (svn_ra_session_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_session_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg4 = svn_swig_py_string_to_cstring(obj1, FALSE, "svn_ra_get_commit_editor", "log_msg");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    arg5 = svn_swig_py_commit_callback;
    arg6 = (void *)obj2;
  }
  {
    /* PYTHON-FIXME: Handle None -> NULL. */
    arg7 = svn_swig_py_stringhash_from_dict(obj3, _global_pool);
  }
  {
    arg8 = (svn_boolean_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  if (obj5) {
    /* Verify that the user supplied a valid pool */
    if (obj5 != Py_None && obj5 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj5);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_get_commit_editor(arg1,(struct svn_delta_editor_t const **)arg2,arg3,(char const *)arg4,arg5,arg6,arg7,arg8,arg9);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg2, SWIGTYPE_p_svn_delta_editor_t,
        _global_py_pool, args), 0)
    
    ;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg3, SWIGTYPE_p_void,
        _global_py_pool, args), 0)
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_get_file(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_session_t *arg1 = 0 ;
  char *arg2 = 0 ;
  svn_revnum_t arg3 ;
  svn_stream_t *arg4 = 0 ;
  svn_revnum_t *arg5 = 0 ;
  apr_hash_t **arg6 = 0 ;
  apr_pool_t *arg7 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_revnum_t temp5 ;
  int res5 = SWIG_TMPOBJ ;
  apr_hash_t *temp6 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg7 = _global_pool;
  arg5 = &temp5;
  arg6 = &temp6;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_get_file", 4, 5, &obj0, &obj1, &obj2, &obj3, &obj4)) SWIG_fail;
  {
    arg1 = (svn_ra_session_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_session_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = svn_swig_py_string_to_cstring(obj1, FALSE, "svn_ra_get_file", "path");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    arg3 = (svn_revnum_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  {
    if (obj3 == Py_None) {
      arg4 = NULL;
    }
    else {
      arg4 = svn_swig_py_make_stream (obj3, _global_pool);
      if (arg4 == NULL) {
        SWIG_fail;
      }
    }
  }
  if (obj4) {
    /* Verify that the user supplied a valid pool */
    if (obj4 != Py_None && obj4 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj4);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_get_file(arg1,(char const *)arg2,arg3,arg4,arg5,arg6,arg7);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  if (SWIG_IsTmpObj(res5)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_long((*arg5)), 0);
  } else {
    int new_flags = SWIG_IsNewObj(res5) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg5), SWIGTYPE_p_long, new_flags), 0);
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_prophash_to_dict(*arg6), 0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_get_dir2(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_session_t *arg1 = 0 ;
  apr_hash_t **arg2 = 0 ;
  svn_revnum_t *arg3 = 0 ;
  apr_hash_t **arg4 = 0 ;
  char *arg5 = 0 ;
  svn_revnum_t arg6 ;
  apr_uint32_t arg7 ;
  apr_pool_t *arg8 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  apr_hash_t *temp2 ;
  svn_revnum_t temp3 ;
  int res3 = SWIG_TMPOBJ ;
  apr_hash_t *temp4 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg8 = _global_pool;
  arg2 = &temp2;
  arg3 = &temp3;
  arg4 = &temp4;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_get_dir2", 4, 5, &obj0, &obj1, &obj2, &obj3, &obj4)) SWIG_fail;
  {
    arg1 = (svn_ra_session_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_session_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg5 = svn_swig_py_string_to_cstring(obj1, FALSE, "svn_ra_get_dir2", "path");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    arg6 = (svn_revnum_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  {
    arg7 = (apr_uint32_t)SWIG_As_unsigned_SS_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  if (obj4) {
    /* Verify that the user supplied a valid pool */
    if (obj4 != Py_None && obj4 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj4);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_get_dir2(arg1,arg2,arg3,arg4,(char const *)arg5,arg6,arg7,arg8);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_convert_hash(*arg2,
        SWIGTYPE_p_svn_dirent_t,
        _global_py_pool), 0)
    
    
    
    ;
  }
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_long((*arg3)), 0);
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_long, new_flags), 0);
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_prophash_to_dict(*arg4), 0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_get_dir(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_session_t *arg1 = 0 ;
  char *arg2 = 0 ;
  svn_revnum_t arg3 ;
  apr_hash_t **arg4 = 0 ;
  svn_revnum_t *arg5 = 0 ;
  apr_hash_t **arg6 = 0 ;
  apr_pool_t *arg7 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  apr_hash_t *temp4 ;
  svn_revnum_t temp5 ;
  int res5 = SWIG_TMPOBJ ;
  apr_hash_t *temp6 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg7 = _global_pool;
  arg4 = &temp4;
  arg5 = &temp5;
  arg6 = &temp6;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_get_dir", 3, 4, &obj0, &obj1, &obj2, &obj3)) SWIG_fail;
  {
    arg1 = (svn_ra_session_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_session_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = svn_swig_py_string_to_cstring(obj1, FALSE, "svn_ra_get_dir", "path");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    arg3 = (svn_revnum_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  if (obj3) {
    /* Verify that the user supplied a valid pool */
    if (obj3 != Py_None && obj3 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj3);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_get_dir(arg1,(char const *)arg2,arg3,arg4,arg5,arg6,arg7);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_convert_hash(*arg4,
        SWIGTYPE_p_svn_dirent_t,
        _global_py_pool), 0)
    
    
    
    ;
  }
  if (SWIG_IsTmpObj(res5)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_long((*arg5)), 0);
  } else {
    int new_flags = SWIG_IsNewObj(res5) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg5), SWIGTYPE_p_long, new_flags), 0);
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_prophash_to_dict(*arg6), 0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_list(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_session_t *arg1 = 0 ;
  char *arg2 = 0 ;
  svn_revnum_t arg3 ;
  apr_array_header_t *arg4 = 0 ;
  svn_depth_t arg5 ;
  apr_uint32_t arg6 ;
  svn_ra_dirent_receiver_t arg7 = 0 ;
  void *arg8 = 0 ;
  apr_pool_t *arg9 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg9 = _global_pool;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_list", 8, 9, &obj0, &obj1, &obj2, &obj3, &obj4, &obj5, &obj6, &obj7, &obj8)) SWIG_fail;
  {
    arg1 = (svn_ra_session_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_session_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = svn_swig_py_string_to_cstring(obj1, FALSE, "svn_ra_list", "path");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    arg3 = (svn_revnum_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  {
    arg4 = (apr_array_header_t *)svn_swig_py_must_get_ptr(obj3, SWIGTYPE_p_apr_array_header_t, svn_argnum_obj3);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg5 = (svn_depth_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  {
    arg6 = (apr_uint32_t)SWIG_As_unsigned_SS_long (obj5);
    if (SWIG_arg_fail(svn_argnum_obj5)) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj6, (void**)(&arg7), SWIGTYPE_p_f_p_q_const__char_p_svn_dirent_t_p_void_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_ra_list" "', argument " "7"" of type '" "svn_ra_dirent_receiver_t""'"); 
    }
  }
  {
    if (obj7 == Py_None) {
      arg8 = NULL;
    } else if (SWIG_ConvertPtr(obj7, (void **) &arg8, 0, 0) == -1) {
      arg8 = (void *) obj7;
      PyErr_Clear();
    }
  }
  if (obj8) {
    /* Verify that the user supplied a valid pool */
    if (obj8 != Py_None && obj8 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj8);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_list(arg1,(char const *)arg2,arg3,(apr_array_header_t const *)arg4,arg5,arg6,arg7,arg8,arg9);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_get_mergeinfo(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_session_t *arg1 = 0 ;
  svn_mergeinfo_catalog_t *arg2 = 0 ;
  apr_array_header_t *arg3 = 0 ;
  svn_revnum_t arg4 ;
  svn_mergeinfo_inheritance_t arg5 ;
  svn_boolean_t arg6 ;
  apr_pool_t *arg7 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_mergeinfo_catalog_t temp2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg7 = _global_pool;
  arg2 = &temp2;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_get_mergeinfo", 5, 6, &obj0, &obj1, &obj2, &obj3, &obj4, &obj5)) SWIG_fail;
  {
    arg1 = (svn_ra_session_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_session_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = (apr_array_header_t *) svn_swig_py_seq_to_array(obj1,
      sizeof(const char *),
      svn_swig_py_unwrap_string,
      NULL,
      _global_pool);
    if (PyErr_Occurred())
    SWIG_fail;
  }
  {
    arg4 = (svn_revnum_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  {
    arg5 = (svn_mergeinfo_inheritance_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg6 = (svn_boolean_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  if (obj5) {
    /* Verify that the user supplied a valid pool */
    if (obj5 != Py_None && obj5 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj5);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_get_mergeinfo(arg1,arg2,(apr_array_header_t const *)arg3,arg4,arg5,arg6,arg7);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_mergeinfo_catalog_to_dict(*arg2,
        SWIGTYPE_p_svn_merge_range_t,
        _global_py_pool), 0)
    
    
    
    ;
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_do_update3(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_session_t *arg1 = 0 ;
  svn_ra_reporter3_t **arg2 = 0 ;
  void **arg3 = 0 ;
  svn_revnum_t arg4 ;
  char *arg5 = 0 ;
  svn_depth_t arg6 ;
  svn_boolean_t arg7 ;
  svn_boolean_t arg8 ;
  svn_delta_editor_t *arg9 = 0 ;
  void *arg10 = 0 ;
  apr_pool_t *arg11 = 0 ;
  apr_pool_t *arg12 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_ra_reporter3_t *temp2 ;
  void *temp3 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg11 = _global_pool;
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg12 = _global_pool;
  arg2 = &temp2;
  arg3 = &temp3;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_do_update3", 8, 10, &obj0, &obj1, &obj2, &obj3, &obj4, &obj5, &obj6, &obj7, &obj8, &obj9)) SWIG_fail;
  {
    arg1 = (svn_ra_session_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_session_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_revnum_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  {
    arg5 = svn_swig_py_string_to_cstring(obj2, FALSE, "svn_ra_do_update3", "update_target");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    arg6 = (svn_depth_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg7 = (svn_boolean_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  {
    arg8 = (svn_boolean_t)SWIG_As_long (obj5);
    if (SWIG_arg_fail(svn_argnum_obj5)) {
      SWIG_fail;
    }
  }
  {
    arg9 = (svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj6, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj6);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj7 == Py_None) {
      arg10 = NULL;
    } else if (SWIG_ConvertPtr(obj7, (void **) &arg10, 0, 0) == -1) {
      arg10 = (void *) obj7;
      PyErr_Clear();
    }
  }
  if (obj8) {
    /* Verify that the user supplied a valid pool */
    if (obj8 != Py_None && obj8 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj8);
      SWIG_fail;
    }
  }
  if (obj9) {
    /* Verify that the user supplied a valid pool */
    if (obj9 != Py_None && obj9 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj9);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_do_update3(arg1,(struct svn_ra_reporter3_t const **)arg2,arg3,arg4,(char const *)arg5,arg6,arg7,arg8,(struct svn_delta_editor_t const *)arg9,arg10,arg11,arg12);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg2, SWIGTYPE_p_svn_ra_reporter3_t,
        _global_py_pool, args), 0)
    
    ;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg3, SWIGTYPE_p_void,
        _global_py_pool, args), 0)
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_do_update2(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_session_t *arg1 = 0 ;
  svn_ra_reporter3_t **arg2 = 0 ;
  void **arg3 = 0 ;
  svn_revnum_t arg4 ;
  char *arg5 = 0 ;
  svn_depth_t arg6 ;
  svn_boolean_t arg7 ;
  svn_delta_editor_t *arg8 = 0 ;
  void *arg9 = 0 ;
  apr_pool_t *arg10 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_ra_reporter3_t *temp2 ;
  void *temp3 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg10 = _global_pool;
  arg2 = &temp2;
  arg3 = &temp3;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_do_update2", 7, 8, &obj0, &obj1, &obj2, &obj3, &obj4, &obj5, &obj6, &obj7)) SWIG_fail;
  {
    arg1 = (svn_ra_session_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_session_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_revnum_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  {
    arg5 = svn_swig_py_string_to_cstring(obj2, FALSE, "svn_ra_do_update2", "update_target");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    arg6 = (svn_depth_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg7 = (svn_boolean_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  {
    arg8 = (svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj5, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj5);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj6 == Py_None) {
      arg9 = NULL;
    } else if (SWIG_ConvertPtr(obj6, (void **) &arg9, 0, 0) == -1) {
      arg9 = (void *) obj6;
      PyErr_Clear();
    }
  }
  if (obj7) {
    /* Verify that the user supplied a valid pool */
    if (obj7 != Py_None && obj7 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj7);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_do_update2(arg1,(struct svn_ra_reporter3_t const **)arg2,arg3,arg4,(char const *)arg5,arg6,arg7,(struct svn_delta_editor_t const *)arg8,arg9,arg10);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg2, SWIGTYPE_p_svn_ra_reporter3_t,
        _global_py_pool, args), 0)
    
    ;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg3, SWIGTYPE_p_void,
        _global_py_pool, args), 0)
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_do_update(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_session_t *arg1 = 0 ;
  svn_ra_reporter2_t **arg2 = 0 ;
  void **arg3 = 0 ;
  svn_revnum_t arg4 ;
  char *arg5 = 0 ;
  svn_boolean_t arg6 ;
  svn_delta_editor_t *arg7 = 0 ;
  void *arg8 = 0 ;
  apr_pool_t *arg9 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_ra_reporter2_t *temp2 ;
  void *temp3 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg9 = _global_pool;
  arg2 = &temp2;
  arg3 = &temp3;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_do_update", 6, 7, &obj0, &obj1, &obj2, &obj3, &obj4, &obj5, &obj6)) SWIG_fail;
  {
    arg1 = (svn_ra_session_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_session_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_revnum_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  {
    arg5 = svn_swig_py_string_to_cstring(obj2, FALSE, "svn_ra_do_update", "update_target");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    arg6 = (svn_boolean_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg7 = (svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj4, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj4);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj5 == Py_None) {
      arg8 = NULL;
    } else if (SWIG_ConvertPtr(obj5, (void **) &arg8, 0, 0) == -1) {
      arg8 = (void *) obj5;
      PyErr_Clear();
    }
  }
  if (obj6) {
    /* Verify that the user supplied a valid pool */
    if (obj6 != Py_None && obj6 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj6);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_do_update(arg1,(struct svn_ra_reporter2_t const **)arg2,arg3,arg4,(char const *)arg5,arg6,(struct svn_delta_editor_t const *)arg7,arg8,arg9);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg2, SWIGTYPE_p_svn_ra_reporter2_t,
        _global_py_pool, args), 0)
    
    ;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg3, SWIGTYPE_p_void,
        _global_py_pool, args), 0)
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_do_switch3(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_session_t *arg1 = 0 ;
  svn_ra_reporter3_t **arg2 = 0 ;
  void **arg3 = 0 ;
  svn_revnum_t arg4 ;
  char *arg5 = 0 ;
  svn_depth_t arg6 ;
  char *arg7 = 0 ;
  svn_boolean_t arg8 ;
  svn_boolean_t arg9 ;
  svn_delta_editor_t *arg10 = 0 ;
  void *arg11 = 0 ;
  apr_pool_t *arg12 = 0 ;
  apr_pool_t *arg13 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_ra_reporter3_t *temp2 ;
  void *temp3 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  PyObject * obj10 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg12 = _global_pool;
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg13 = _global_pool;
  arg2 = &temp2;
  arg3 = &temp3;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_do_switch3", 9, 11, &obj0, &obj1, &obj2, &obj3, &obj4, &obj5, &obj6, &obj7, &obj8, &obj9, &obj10)) SWIG_fail;
  {
    arg1 = (svn_ra_session_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_session_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_revnum_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  {
    arg5 = svn_swig_py_string_to_cstring(obj2, FALSE, "svn_ra_do_switch3", "switch_target");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    arg6 = (svn_depth_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg7 = svn_swig_py_string_to_cstring(obj4, FALSE, "svn_ra_do_switch3", "switch_url");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    arg8 = (svn_boolean_t)SWIG_As_long (obj5);
    if (SWIG_arg_fail(svn_argnum_obj5)) {
      SWIG_fail;
    }
  }
  {
    arg9 = (svn_boolean_t)SWIG_As_long (obj6);
    if (SWIG_arg_fail(svn_argnum_obj6)) {
      SWIG_fail;
    }
  }
  {
    arg10 = (svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj7, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj7);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj8 == Py_None) {
      arg11 = NULL;
    } else if (SWIG_ConvertPtr(obj8, (void **) &arg11, 0, 0) == -1) {
      arg11 = (void *) obj8;
      PyErr_Clear();
    }
  }
  if (obj9) {
    /* Verify that the user supplied a valid pool */
    if (obj9 != Py_None && obj9 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj9);
      SWIG_fail;
    }
  }
  if (obj10) {
    /* Verify that the user supplied a valid pool */
    if (obj10 != Py_None && obj10 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj10);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_do_switch3(arg1,(struct svn_ra_reporter3_t const **)arg2,arg3,arg4,(char const *)arg5,arg6,(char const *)arg7,arg8,arg9,(struct svn_delta_editor_t const *)arg10,arg11,arg12,arg13);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg2, SWIGTYPE_p_svn_ra_reporter3_t,
        _global_py_pool, args), 0)
    
    ;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg3, SWIGTYPE_p_void,
        _global_py_pool, args), 0)
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_do_switch2(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_session_t *arg1 = 0 ;
  svn_ra_reporter3_t **arg2 = 0 ;
  void **arg3 = 0 ;
  svn_revnum_t arg4 ;
  char *arg5 = 0 ;
  svn_depth_t arg6 ;
  char *arg7 = 0 ;
  svn_delta_editor_t *arg8 = 0 ;
  void *arg9 = 0 ;
  apr_pool_t *arg10 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_ra_reporter3_t *temp2 ;
  void *temp3 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg10 = _global_pool;
  arg2 = &temp2;
  arg3 = &temp3;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_do_switch2", 7, 8, &obj0, &obj1, &obj2, &obj3, &obj4, &obj5, &obj6, &obj7)) SWIG_fail;
  {
    arg1 = (svn_ra_session_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_session_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_revnum_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  {
    arg5 = svn_swig_py_string_to_cstring(obj2, FALSE, "svn_ra_do_switch2", "switch_target");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    arg6 = (svn_depth_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg7 = svn_swig_py_string_to_cstring(obj4, FALSE, "svn_ra_do_switch2", "switch_url");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    arg8 = (svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj5, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj5);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj6 == Py_None) {
      arg9 = NULL;
    } else if (SWIG_ConvertPtr(obj6, (void **) &arg9, 0, 0) == -1) {
      arg9 = (void *) obj6;
      PyErr_Clear();
    }
  }
  if (obj7) {
    /* Verify that the user supplied a valid pool */
    if (obj7 != Py_None && obj7 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj7);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_do_switch2(arg1,(struct svn_ra_reporter3_t const **)arg2,arg3,arg4,(char const *)arg5,arg6,(char const *)arg7,(struct svn_delta_editor_t const *)arg8,arg9,arg10);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg2, SWIGTYPE_p_svn_ra_reporter3_t,
        _global_py_pool, args), 0)
    
    ;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg3, SWIGTYPE_p_void,
        _global_py_pool, args), 0)
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_do_switch(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_session_t *arg1 = 0 ;
  svn_ra_reporter2_t **arg2 = 0 ;
  void **arg3 = 0 ;
  svn_revnum_t arg4 ;
  char *arg5 = 0 ;
  svn_boolean_t arg6 ;
  char *arg7 = 0 ;
  svn_delta_editor_t *arg8 = 0 ;
  void *arg9 = 0 ;
  apr_pool_t *arg10 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_ra_reporter2_t *temp2 ;
  void *temp3 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg10 = _global_pool;
  arg2 = &temp2;
  arg3 = &temp3;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_do_switch", 7, 8, &obj0, &obj1, &obj2, &obj3, &obj4, &obj5, &obj6, &obj7)) SWIG_fail;
  {
    arg1 = (svn_ra_session_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_session_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_revnum_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  {
    arg5 = svn_swig_py_string_to_cstring(obj2, FALSE, "svn_ra_do_switch", "switch_target");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    arg6 = (svn_boolean_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg7 = svn_swig_py_string_to_cstring(obj4, FALSE, "svn_ra_do_switch", "switch_url");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    arg8 = (svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj5, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj5);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj6 == Py_None) {
      arg9 = NULL;
    } else if (SWIG_ConvertPtr(obj6, (void **) &arg9, 0, 0) == -1) {
      arg9 = (void *) obj6;
      PyErr_Clear();
    }
  }
  if (obj7) {
    /* Verify that the user supplied a valid pool */
    if (obj7 != Py_None && obj7 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj7);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_do_switch(arg1,(struct svn_ra_reporter2_t const **)arg2,arg3,arg4,(char const *)arg5,arg6,(char const *)arg7,(struct svn_delta_editor_t const *)arg8,arg9,arg10);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg2, SWIGTYPE_p_svn_ra_reporter2_t,
        _global_py_pool, args), 0)
    
    ;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg3, SWIGTYPE_p_void,
        _global_py_pool, args), 0)
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_do_status2(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_session_t *arg1 = 0 ;
  svn_ra_reporter3_t **arg2 = 0 ;
  void **arg3 = 0 ;
  char *arg4 = 0 ;
  svn_revnum_t arg5 ;
  svn_depth_t arg6 ;
  svn_delta_editor_t *arg7 = 0 ;
  void *arg8 = 0 ;
  apr_pool_t *arg9 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_ra_reporter3_t *temp2 ;
  void *temp3 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg9 = _global_pool;
  arg2 = &temp2;
  arg3 = &temp3;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_do_status2", 6, 7, &obj0, &obj1, &obj2, &obj3, &obj4, &obj5, &obj6)) SWIG_fail;
  {
    arg1 = (svn_ra_session_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_session_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg4 = svn_swig_py_string_to_cstring(obj1, FALSE, "svn_ra_do_status2", "status_target");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    arg5 = (svn_revnum_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  {
    arg6 = (svn_depth_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg7 = (svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj4, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj4);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj5 == Py_None) {
      arg8 = NULL;
    } else if (SWIG_ConvertPtr(obj5, (void **) &arg8, 0, 0) == -1) {
      arg8 = (void *) obj5;
      PyErr_Clear();
    }
  }
  if (obj6) {
    /* Verify that the user supplied a valid pool */
    if (obj6 != Py_None && obj6 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj6);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_do_status2(arg1,(struct svn_ra_reporter3_t const **)arg2,arg3,(char const *)arg4,arg5,arg6,(struct svn_delta_editor_t const *)arg7,arg8,arg9);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg2, SWIGTYPE_p_svn_ra_reporter3_t,
        _global_py_pool, args), 0)
    
    ;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg3, SWIGTYPE_p_void,
        _global_py_pool, args), 0)
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_do_status(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_session_t *arg1 = 0 ;
  svn_ra_reporter2_t **arg2 = 0 ;
  void **arg3 = 0 ;
  char *arg4 = 0 ;
  svn_revnum_t arg5 ;
  svn_boolean_t arg6 ;
  svn_delta_editor_t *arg7 = 0 ;
  void *arg8 = 0 ;
  apr_pool_t *arg9 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_ra_reporter2_t *temp2 ;
  void *temp3 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg9 = _global_pool;
  arg2 = &temp2;
  arg3 = &temp3;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_do_status", 6, 7, &obj0, &obj1, &obj2, &obj3, &obj4, &obj5, &obj6)) SWIG_fail;
  {
    arg1 = (svn_ra_session_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_session_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg4 = svn_swig_py_string_to_cstring(obj1, FALSE, "svn_ra_do_status", "status_target");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    arg5 = (svn_revnum_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  {
    arg6 = (svn_boolean_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg7 = (svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj4, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj4);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj5 == Py_None) {
      arg8 = NULL;
    } else if (SWIG_ConvertPtr(obj5, (void **) &arg8, 0, 0) == -1) {
      arg8 = (void *) obj5;
      PyErr_Clear();
    }
  }
  if (obj6) {
    /* Verify that the user supplied a valid pool */
    if (obj6 != Py_None && obj6 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj6);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_do_status(arg1,(struct svn_ra_reporter2_t const **)arg2,arg3,(char const *)arg4,arg5,arg6,(struct svn_delta_editor_t const *)arg7,arg8,arg9);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg2, SWIGTYPE_p_svn_ra_reporter2_t,
        _global_py_pool, args), 0)
    
    ;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg3, SWIGTYPE_p_void,
        _global_py_pool, args), 0)
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_do_diff3(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_session_t *arg1 = 0 ;
  svn_ra_reporter3_t **arg2 = 0 ;
  void **arg3 = 0 ;
  svn_revnum_t arg4 ;
  char *arg5 = 0 ;
  svn_depth_t arg6 ;
  svn_boolean_t arg7 ;
  svn_boolean_t arg8 ;
  char *arg9 = 0 ;
  svn_delta_editor_t *arg10 = 0 ;
  void *arg11 = 0 ;
  apr_pool_t *arg12 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_ra_reporter3_t *temp2 ;
  void *temp3 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg12 = _global_pool;
  arg2 = &temp2;
  arg3 = &temp3;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_do_diff3", 9, 10, &obj0, &obj1, &obj2, &obj3, &obj4, &obj5, &obj6, &obj7, &obj8, &obj9)) SWIG_fail;
  {
    arg1 = (svn_ra_session_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_session_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_revnum_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  {
    arg5 = svn_swig_py_string_to_cstring(obj2, FALSE, "svn_ra_do_diff3", "diff_target");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    arg6 = (svn_depth_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg7 = (svn_boolean_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  {
    arg8 = (svn_boolean_t)SWIG_As_long (obj5);
    if (SWIG_arg_fail(svn_argnum_obj5)) {
      SWIG_fail;
    }
  }
  {
    arg9 = svn_swig_py_string_to_cstring(obj6, FALSE, "svn_ra_do_diff3", "versus_url");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    arg10 = (svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj7, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj7);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj8 == Py_None) {
      arg11 = NULL;
    } else if (SWIG_ConvertPtr(obj8, (void **) &arg11, 0, 0) == -1) {
      arg11 = (void *) obj8;
      PyErr_Clear();
    }
  }
  if (obj9) {
    /* Verify that the user supplied a valid pool */
    if (obj9 != Py_None && obj9 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj9);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_do_diff3(arg1,(struct svn_ra_reporter3_t const **)arg2,arg3,arg4,(char const *)arg5,arg6,arg7,arg8,(char const *)arg9,(struct svn_delta_editor_t const *)arg10,arg11,arg12);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg2, SWIGTYPE_p_svn_ra_reporter3_t,
        _global_py_pool, args), 0)
    
    ;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg3, SWIGTYPE_p_void,
        _global_py_pool, args), 0)
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_do_diff2(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_session_t *arg1 = 0 ;
  svn_ra_reporter2_t **arg2 = 0 ;
  void **arg3 = 0 ;
  svn_revnum_t arg4 ;
  char *arg5 = 0 ;
  svn_boolean_t arg6 ;
  svn_boolean_t arg7 ;
  svn_boolean_t arg8 ;
  char *arg9 = 0 ;
  svn_delta_editor_t *arg10 = 0 ;
  void *arg11 = 0 ;
  apr_pool_t *arg12 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_ra_reporter2_t *temp2 ;
  void *temp3 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg12 = _global_pool;
  arg2 = &temp2;
  arg3 = &temp3;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_do_diff2", 9, 10, &obj0, &obj1, &obj2, &obj3, &obj4, &obj5, &obj6, &obj7, &obj8, &obj9)) SWIG_fail;
  {
    arg1 = (svn_ra_session_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_session_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_revnum_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  {
    arg5 = svn_swig_py_string_to_cstring(obj2, FALSE, "svn_ra_do_diff2", "diff_target");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    arg6 = (svn_boolean_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg7 = (svn_boolean_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  {
    arg8 = (svn_boolean_t)SWIG_As_long (obj5);
    if (SWIG_arg_fail(svn_argnum_obj5)) {
      SWIG_fail;
    }
  }
  {
    arg9 = svn_swig_py_string_to_cstring(obj6, FALSE, "svn_ra_do_diff2", "versus_url");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    arg10 = (svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj7, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj7);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj8 == Py_None) {
      arg11 = NULL;
    } else if (SWIG_ConvertPtr(obj8, (void **) &arg11, 0, 0) == -1) {
      arg11 = (void *) obj8;
      PyErr_Clear();
    }
  }
  if (obj9) {
    /* Verify that the user supplied a valid pool */
    if (obj9 != Py_None && obj9 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj9);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_do_diff2(arg1,(struct svn_ra_reporter2_t const **)arg2,arg3,arg4,(char const *)arg5,arg6,arg7,arg8,(char const *)arg9,(struct svn_delta_editor_t const *)arg10,arg11,arg12);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg2, SWIGTYPE_p_svn_ra_reporter2_t,
        _global_py_pool, args), 0)
    
    ;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg3, SWIGTYPE_p_void,
        _global_py_pool, args), 0)
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_do_diff(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_session_t *arg1 = 0 ;
  svn_ra_reporter2_t **arg2 = 0 ;
  void **arg3 = 0 ;
  svn_revnum_t arg4 ;
  char *arg5 = 0 ;
  svn_boolean_t arg6 ;
  svn_boolean_t arg7 ;
  char *arg8 = 0 ;
  svn_delta_editor_t *arg9 = 0 ;
  void *arg10 = 0 ;
  apr_pool_t *arg11 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_ra_reporter2_t *temp2 ;
  void *temp3 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg11 = _global_pool;
  arg2 = &temp2;
  arg3 = &temp3;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_do_diff", 8, 9, &obj0, &obj1, &obj2, &obj3, &obj4, &obj5, &obj6, &obj7, &obj8)) SWIG_fail;
  {
    arg1 = (svn_ra_session_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_session_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_revnum_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  {
    arg5 = svn_swig_py_string_to_cstring(obj2, FALSE, "svn_ra_do_diff", "diff_target");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    arg6 = (svn_boolean_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg7 = (svn_boolean_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  {
    arg8 = svn_swig_py_string_to_cstring(obj5, FALSE, "svn_ra_do_diff", "versus_url");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    arg9 = (svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj6, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj6);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj7 == Py_None) {
      arg10 = NULL;
    } else if (SWIG_ConvertPtr(obj7, (void **) &arg10, 0, 0) == -1) {
      arg10 = (void *) obj7;
      PyErr_Clear();
    }
  }
  if (obj8) {
    /* Verify that the user supplied a valid pool */
    if (obj8 != Py_None && obj8 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj8);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_do_diff(arg1,(struct svn_ra_reporter2_t const **)arg2,arg3,arg4,(char const *)arg5,arg6,arg7,(char const *)arg8,(struct svn_delta_editor_t const *)arg9,arg10,arg11);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg2, SWIGTYPE_p_svn_ra_reporter2_t,
        _global_py_pool, args), 0)
    
    ;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg3, SWIGTYPE_p_void,
        _global_py_pool, args), 0)
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_get_log2(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_session_t *arg1 = 0 ;
  apr_array_header_t *arg2 = 0 ;
  svn_revnum_t arg3 ;
  svn_revnum_t arg4 ;
  int arg5 ;
  svn_boolean_t arg6 ;
  svn_boolean_t arg7 ;
  svn_boolean_t arg8 ;
  apr_array_header_t *arg9 = 0 ;
  svn_log_entry_receiver_t arg10 = 0 ;
  void *arg11 = 0 ;
  apr_pool_t *arg12 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  PyObject * obj10 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg12 = _global_pool;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_get_log2", 10, 11, &obj0, &obj1, &obj2, &obj3, &obj4, &obj5, &obj6, &obj7, &obj8, &obj9, &obj10)) SWIG_fail;
  {
    arg1 = (svn_ra_session_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_session_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (apr_array_header_t *) svn_swig_py_seq_to_array(obj1,
      sizeof(const char *),
      svn_swig_py_unwrap_string,
      NULL,
      _global_pool);
    if (PyErr_Occurred())
    SWIG_fail;
  }
  {
    arg3 = (svn_revnum_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_revnum_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg5 = (int)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  {
    arg6 = (svn_boolean_t)SWIG_As_long (obj5);
    if (SWIG_arg_fail(svn_argnum_obj5)) {
      SWIG_fail;
    }
  }
  {
    arg7 = (svn_boolean_t)SWIG_As_long (obj6);
    if (SWIG_arg_fail(svn_argnum_obj6)) {
      SWIG_fail;
    }
  }
  {
    arg8 = (svn_boolean_t)SWIG_As_long (obj7);
    if (SWIG_arg_fail(svn_argnum_obj7)) {
      SWIG_fail;
    }
  }
  {
    arg9 = (apr_array_header_t *) svn_swig_py_seq_to_array(obj8,
      sizeof(const char *),
      svn_swig_py_unwrap_string,
      NULL,
      _global_pool);
    if (PyErr_Occurred())
    SWIG_fail;
  }
  {
    arg10 = (svn_log_entry_receiver_t) svn_swig_py_log_entry_receiver;
    arg11 = obj9;
  }
  if (obj10) {
    /* Verify that the user supplied a valid pool */
    if (obj10 != Py_None && obj10 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj10);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_get_log2(arg1,(apr_array_header_t const *)arg2,arg3,arg4,arg5,arg6,arg7,arg8,(apr_array_header_t const *)arg9,arg10,arg11,arg12);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_get_log(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_session_t *arg1 = 0 ;
  apr_array_header_t *arg2 = 0 ;
  svn_revnum_t arg3 ;
  svn_revnum_t arg4 ;
  int arg5 ;
  svn_boolean_t arg6 ;
  svn_boolean_t arg7 ;
  svn_log_message_receiver_t arg8 = 0 ;
  void *arg9 = 0 ;
  apr_pool_t *arg10 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg10 = _global_pool;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_get_log", 8, 9, &obj0, &obj1, &obj2, &obj3, &obj4, &obj5, &obj6, &obj7, &obj8)) SWIG_fail;
  {
    arg1 = (svn_ra_session_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_session_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (apr_array_header_t *) svn_swig_py_seq_to_array(obj1,
      sizeof(const char *),
      svn_swig_py_unwrap_string,
      NULL,
      _global_pool);
    if (PyErr_Occurred())
    SWIG_fail;
  }
  {
    arg3 = (svn_revnum_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_revnum_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg5 = (int)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  {
    arg6 = (svn_boolean_t)SWIG_As_long (obj5);
    if (SWIG_arg_fail(svn_argnum_obj5)) {
      SWIG_fail;
    }
  }
  {
    arg7 = (svn_boolean_t)SWIG_As_long (obj6);
    if (SWIG_arg_fail(svn_argnum_obj6)) {
      SWIG_fail;
    }
  }
  {
    arg8 = (svn_log_message_receiver_t) svn_swig_py_log_receiver;
    arg9 = obj7;
  }
  if (obj8) {
    /* Verify that the user supplied a valid pool */
    if (obj8 != Py_None && obj8 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj8);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_get_log(arg1,(apr_array_header_t const *)arg2,arg3,arg4,arg5,arg6,arg7,arg8,arg9,arg10);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_check_path(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_session_t *arg1 = 0 ;
  char *arg2 = 0 ;
  svn_revnum_t arg3 ;
  svn_node_kind_t *arg4 = 0 ;
  apr_pool_t *arg5 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_node_kind_t temp4 ;
  int res4 = SWIG_TMPOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg5 = _global_pool;
  arg4 = &temp4;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_check_path", 3, 4, &obj0, &obj1, &obj2, &obj3)) SWIG_fail;
  {
    arg1 = (svn_ra_session_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_session_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = svn_swig_py_string_to_cstring(obj1, FALSE, "svn_ra_check_path", "path");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    arg3 = (svn_revnum_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  if (obj3) {
    /* Verify that the user supplied a valid pool */
    if (obj3 != Py_None && obj3 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj3);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_check_path(arg1,(char const *)arg2,arg3,arg4,arg5);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  if (SWIG_IsTmpObj(res4)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_long((*arg4)), 0);
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_svn_node_kind_t, new_flags), 0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_stat(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_session_t *arg1 = 0 ;
  char *arg2 = 0 ;
  svn_revnum_t arg3 ;
  svn_dirent_t **arg4 = 0 ;
  apr_pool_t *arg5 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_dirent_t *temp4 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg5 = _global_pool;
  arg4 = &temp4;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_stat", 3, 4, &obj0, &obj1, &obj2, &obj3)) SWIG_fail;
  {
    arg1 = (svn_ra_session_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_session_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = svn_swig_py_string_to_cstring(obj1, FALSE, "svn_ra_stat", "path");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    arg3 = (svn_revnum_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  if (obj3) {
    /* Verify that the user supplied a valid pool */
    if (obj3 != Py_None && obj3 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj3);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_stat(arg1,(char const *)arg2,arg3,arg4,arg5);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg4, SWIGTYPE_p_svn_dirent_t,
        _global_py_pool, args), 0)
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_get_uuid2(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_session_t *arg1 = 0 ;
  char **arg2 = 0 ;
  apr_pool_t *arg3 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  char *temp2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg3 = _global_pool;
  arg2 = &temp2;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_get_uuid2", 1, 2, &obj0, &obj1)) SWIG_fail;
  {
    arg1 = (svn_ra_session_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_session_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj1) {
    /* Verify that the user supplied a valid pool */
    if (obj1 != Py_None && obj1 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj1);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_get_uuid2(arg1,(char const **)arg2,arg3);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    PyObject *s;
    if (*arg2 == NULL) {
      Py_INCREF(Py_None);
      s = Py_None;
    }
    else {
      s = PyBytes_FromString(*arg2);
      if (s == NULL)
      SWIG_fail;
    }
    resultobj = SWIG_Python_AppendOutput(resultobj, s, 0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_get_uuid(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_session_t *arg1 = 0 ;
  char **arg2 = 0 ;
  apr_pool_t *arg3 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  char *temp2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg3 = _global_pool;
  arg2 = &temp2;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_get_uuid", 1, 2, &obj0, &obj1)) SWIG_fail;
  {
    arg1 = (svn_ra_session_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_session_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj1) {
    /* Verify that the user supplied a valid pool */
    if (obj1 != Py_None && obj1 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj1);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_get_uuid(arg1,(char const **)arg2,arg3);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    PyObject *s;
    if (*arg2 == NULL) {
      Py_INCREF(Py_None);
      s = Py_None;
    }
    else {
      s = PyBytes_FromString(*arg2);
      if (s == NULL)
      SWIG_fail;
    }
    resultobj = SWIG_Python_AppendOutput(resultobj, s, 0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_get_repos_root2(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_session_t *arg1 = 0 ;
  char **arg2 = 0 ;
  apr_pool_t *arg3 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  char *temp2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg3 = _global_pool;
  arg2 = &temp2;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_get_repos_root2", 1, 2, &obj0, &obj1)) SWIG_fail;
  {
    arg1 = (svn_ra_session_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_session_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj1) {
    /* Verify that the user supplied a valid pool */
    if (obj1 != Py_None && obj1 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj1);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_get_repos_root2(arg1,(char const **)arg2,arg3);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    PyObject *s;
    if (*arg2 == NULL) {
      Py_INCREF(Py_None);
      s = Py_None;
    }
    else {
      s = PyBytes_FromString(*arg2);
      if (s == NULL)
      SWIG_fail;
    }
    resultobj = SWIG_Python_AppendOutput(resultobj, s, 0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_get_repos_root(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_session_t *arg1 = 0 ;
  char **arg2 = 0 ;
  apr_pool_t *arg3 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  char *temp2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg3 = _global_pool;
  arg2 = &temp2;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_get_repos_root", 1, 2, &obj0, &obj1)) SWIG_fail;
  {
    arg1 = (svn_ra_session_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_session_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj1) {
    /* Verify that the user supplied a valid pool */
    if (obj1 != Py_None && obj1 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj1);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_get_repos_root(arg1,(char const **)arg2,arg3);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    PyObject *s;
    if (*arg2 == NULL) {
      Py_INCREF(Py_None);
      s = Py_None;
    }
    else {
      s = PyBytes_FromString(*arg2);
      if (s == NULL)
      SWIG_fail;
    }
    resultobj = SWIG_Python_AppendOutput(resultobj, s, 0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_get_locations(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_session_t *arg1 = 0 ;
  apr_hash_t **arg2 = 0 ;
  char *arg3 = 0 ;
  svn_revnum_t arg4 ;
  apr_array_header_t *arg5 = 0 ;
  apr_pool_t *arg6 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  apr_hash_t *temp2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg6 = _global_pool;
  arg2 = &temp2;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_get_locations", 4, 5, &obj0, &obj1, &obj2, &obj3, &obj4)) SWIG_fail;
  {
    arg1 = (svn_ra_session_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_session_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = svn_swig_py_string_to_cstring(obj1, FALSE, "svn_ra_get_locations", "path");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    arg4 = (svn_revnum_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  {
    arg5 = (apr_array_header_t *) svn_swig_py_seq_to_array(obj3,
      sizeof(svn_revnum_t),
      svn_swig_py_unwrap_revnum,
      NULL,
      _global_pool);
    if (PyErr_Occurred())
    SWIG_fail;
  }
  if (obj4) {
    /* Verify that the user supplied a valid pool */
    if (obj4 != Py_None && obj4 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj4);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_get_locations(arg1,arg2,(char const *)arg3,arg4,(apr_array_header_t const *)arg5,arg6);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_locationhash_to_dict(*arg2), 0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_get_location_segments(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_session_t *arg1 = 0 ;
  char *arg2 = 0 ;
  svn_revnum_t arg3 ;
  svn_revnum_t arg4 ;
  svn_revnum_t arg5 ;
  svn_location_segment_receiver_t arg6 = 0 ;
  void *arg7 = 0 ;
  apr_pool_t *arg8 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg8 = _global_pool;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_get_location_segments", 6, 7, &obj0, &obj1, &obj2, &obj3, &obj4, &obj5, &obj6)) SWIG_fail;
  {
    arg1 = (svn_ra_session_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_session_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = svn_swig_py_string_to_cstring(obj1, FALSE, "svn_ra_get_location_segments", "path");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    arg3 = (svn_revnum_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_revnum_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg5 = (svn_revnum_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  {
    arg6 = (svn_location_segment_receiver_t) svn_swig_py_location_segment_receiver_func;
    arg7 = obj5;
  }
  if (obj6) {
    /* Verify that the user supplied a valid pool */
    if (obj6 != Py_None && obj6 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj6);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_get_location_segments(arg1,(char const *)arg2,arg3,arg4,arg5,arg6,arg7,arg8);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_get_file_revs2(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_session_t *arg1 = 0 ;
  char *arg2 = 0 ;
  svn_revnum_t arg3 ;
  svn_revnum_t arg4 ;
  svn_boolean_t arg5 ;
  svn_file_rev_handler_t arg6 = 0 ;
  void *arg7 = 0 ;
  apr_pool_t *arg8 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg8 = _global_pool;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_get_file_revs2", 7, 8, &obj0, &obj1, &obj2, &obj3, &obj4, &obj5, &obj6, &obj7)) SWIG_fail;
  {
    arg1 = (svn_ra_session_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_session_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = svn_swig_py_string_to_cstring(obj1, FALSE, "svn_ra_get_file_revs2", "path");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    arg3 = (svn_revnum_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_revnum_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg5 = (svn_boolean_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  {
    svn_file_rev_handler_t * tmp =
    svn_swig_py_must_get_ptr(obj5, SWIGTYPE_p_p_f_p_void_p_q_const__char_svn_revnum_t_p_apr_hash_t_svn_boolean_t_p_svn_txdelta_window_handler_t_p_p_void_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t, svn_argnum_obj5);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg6 = *tmp;
  }
  {
    if (obj6 == Py_None) {
      arg7 = NULL;
    } else if (SWIG_ConvertPtr(obj6, (void **) &arg7, 0, 0) == -1) {
      arg7 = (void *) obj6;
      PyErr_Clear();
    }
  }
  if (obj7) {
    /* Verify that the user supplied a valid pool */
    if (obj7 != Py_None && obj7 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj7);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_get_file_revs2(arg1,(char const *)arg2,arg3,arg4,arg5,arg6,arg7,arg8);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_get_file_revs(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_session_t *arg1 = 0 ;
  char *arg2 = 0 ;
  svn_revnum_t arg3 ;
  svn_revnum_t arg4 ;
  svn_ra_file_rev_handler_t arg5 = 0 ;
  void *arg6 = 0 ;
  apr_pool_t *arg7 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg7 = _global_pool;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_get_file_revs", 5, 6, &obj0, &obj1, &obj2, &obj3, &obj4, &obj5)) SWIG_fail;
  {
    arg1 = (svn_ra_session_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_session_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = svn_swig_py_string_to_cstring(obj1, FALSE, "svn_ra_get_file_revs", "path");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    arg3 = (svn_revnum_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_revnum_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg5 = (svn_ra_file_rev_handler_t) svn_swig_py_ra_file_rev_handler_func;
    arg6 = obj4;
  }
  if (obj5) {
    /* Verify that the user supplied a valid pool */
    if (obj5 != Py_None && obj5 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj5);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_get_file_revs(arg1,(char const *)arg2,arg3,arg4,arg5,arg6,arg7);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_lock(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_session_t *arg1 = 0 ;
  apr_hash_t *arg2 = 0 ;
  char *arg3 = 0 ;
  svn_boolean_t arg4 ;
  svn_ra_lock_callback_t arg5 = 0 ;
  void *arg6 = 0 ;
  apr_pool_t *arg7 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg7 = _global_pool;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_lock", 5, 6, &obj0, &obj1, &obj2, &obj3, &obj4, &obj5)) SWIG_fail;
  {
    arg1 = (svn_ra_session_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_session_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = svn_swig_py_path_revs_hash_from_dict(obj1, _global_pool);
  }
  {
    arg3 = svn_swig_py_string_to_cstring(obj2, TRUE, "svn_ra_lock", "comment");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    arg4 = (svn_boolean_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg5 = (svn_ra_lock_callback_t) svn_swig_py_ra_lock_callback;
    arg6 = obj4;
  }
  if (obj5) {
    /* Verify that the user supplied a valid pool */
    if (obj5 != Py_None && obj5 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj5);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_lock(arg1,arg2,(char const *)arg3,arg4,arg5,arg6,arg7);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_unlock(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_session_t *arg1 = 0 ;
  apr_hash_t *arg2 = 0 ;
  svn_boolean_t arg3 ;
  svn_ra_lock_callback_t arg4 = 0 ;
  void *arg5 = 0 ;
  apr_pool_t *arg6 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg6 = _global_pool;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_unlock", 4, 5, &obj0, &obj1, &obj2, &obj3, &obj4)) SWIG_fail;
  {
    arg1 = (svn_ra_session_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_session_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = svn_swig_py_stringhash_from_dict(obj1, _global_pool);
  }
  {
    arg3 = (svn_boolean_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_ra_lock_callback_t) svn_swig_py_ra_lock_callback;
    arg5 = obj3;
  }
  if (obj4) {
    /* Verify that the user supplied a valid pool */
    if (obj4 != Py_None && obj4 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj4);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_unlock(arg1,arg2,arg3,arg4,arg5,arg6);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_get_lock(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_session_t *arg1 = 0 ;
  svn_lock_t **arg2 = 0 ;
  char *arg3 = 0 ;
  apr_pool_t *arg4 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_lock_t *temp2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg4 = _global_pool;
  arg2 = &temp2;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_get_lock", 2, 3, &obj0, &obj1, &obj2)) SWIG_fail;
  {
    arg1 = (svn_ra_session_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_session_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = svn_swig_py_string_to_cstring(obj1, FALSE, "svn_ra_get_lock", "path");
    if (PyErr_Occurred()) SWIG_fail;
  }
  if (obj2) {
    /* Verify that the user supplied a valid pool */
    if (obj2 != Py_None && obj2 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj2);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_get_lock(arg1,arg2,(char const *)arg3,arg4);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg2, SWIGTYPE_p_svn_lock_t,
        _global_py_pool, args), 0)
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_get_locks2(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_session_t *arg1 = 0 ;
  apr_hash_t **arg2 = 0 ;
  char *arg3 = 0 ;
  svn_depth_t arg4 ;
  apr_pool_t *arg5 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  apr_hash_t *temp2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg5 = _global_pool;
  arg2 = &temp2;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_get_locks2", 3, 4, &obj0, &obj1, &obj2, &obj3)) SWIG_fail;
  {
    arg1 = (svn_ra_session_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_session_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = svn_swig_py_string_to_cstring(obj1, FALSE, "svn_ra_get_locks2", "path");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    arg4 = (svn_depth_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  if (obj3) {
    /* Verify that the user supplied a valid pool */
    if (obj3 != Py_None && obj3 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj3);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_get_locks2(arg1,arg2,(char const *)arg3,arg4,arg5);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_convert_hash(*arg2,
        SWIGTYPE_p_svn_lock_t,
        _global_py_pool), 0)
    
    
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_get_locks(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_session_t *arg1 = 0 ;
  apr_hash_t **arg2 = 0 ;
  char *arg3 = 0 ;
  apr_pool_t *arg4 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  apr_hash_t *temp2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg4 = _global_pool;
  arg2 = &temp2;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_get_locks", 2, 3, &obj0, &obj1, &obj2)) SWIG_fail;
  {
    arg1 = (svn_ra_session_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_session_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = svn_swig_py_string_to_cstring(obj1, FALSE, "svn_ra_get_locks", "path");
    if (PyErr_Occurred()) SWIG_fail;
  }
  if (obj2) {
    /* Verify that the user supplied a valid pool */
    if (obj2 != Py_None && obj2 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj2);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_get_locks(arg1,arg2,(char const *)arg3,arg4);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_convert_hash(*arg2,
        SWIGTYPE_p_svn_lock_t,
        _global_py_pool), 0)
    
    
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_replay_range(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_session_t *arg1 = 0 ;
  svn_revnum_t arg2 ;
  svn_revnum_t arg3 ;
  svn_revnum_t arg4 ;
  svn_boolean_t arg5 ;
  svn_ra_replay_revstart_callback_t arg6 = 0 ;
  svn_ra_replay_revfinish_callback_t arg7 = 0 ;
  void *arg8 = 0 ;
  apr_pool_t *arg9 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg9 = _global_pool;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_replay_range", 8, 9, &obj0, &obj1, &obj2, &obj3, &obj4, &obj5, &obj6, &obj7, &obj8)) SWIG_fail;
  {
    arg1 = (svn_ra_session_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_session_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_revnum_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_revnum_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_revnum_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg5 = (svn_boolean_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  {
    svn_ra_replay_revstart_callback_t * tmp =
    svn_swig_py_must_get_ptr(obj5, SWIGTYPE_p_p_f_svn_revnum_t_p_void_p_p_q_const__svn_delta_editor_t_p_p_void_p_apr_hash_t_p_apr_pool_t__p_svn_error_t, svn_argnum_obj5);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg6 = *tmp;
  }
  {
    svn_ra_replay_revfinish_callback_t * tmp =
    svn_swig_py_must_get_ptr(obj6, SWIGTYPE_p_p_f_svn_revnum_t_p_void_p_q_const__svn_delta_editor_t_p_void_p_apr_hash_t_p_apr_pool_t__p_svn_error_t, svn_argnum_obj6);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg7 = *tmp;
  }
  {
    if (obj7 == Py_None) {
      arg8 = NULL;
    } else if (SWIG_ConvertPtr(obj7, (void **) &arg8, 0, 0) == -1) {
      arg8 = (void *) obj7;
      PyErr_Clear();
    }
  }
  if (obj8) {
    /* Verify that the user supplied a valid pool */
    if (obj8 != Py_None && obj8 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj8);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_replay_range(arg1,arg2,arg3,arg4,arg5,arg6,arg7,arg8,arg9);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_replay(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_session_t *arg1 = 0 ;
  svn_revnum_t arg2 ;
  svn_revnum_t arg3 ;
  svn_boolean_t arg4 ;
  svn_delta_editor_t *arg5 = 0 ;
  void *arg6 = 0 ;
  apr_pool_t *arg7 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg7 = _global_pool;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_replay", 6, 7, &obj0, &obj1, &obj2, &obj3, &obj4, &obj5, &obj6)) SWIG_fail;
  {
    arg1 = (svn_ra_session_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_session_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = (svn_revnum_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  {
    arg3 = (svn_revnum_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_boolean_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg5 = (svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj4, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj4);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj5 == Py_None) {
      arg6 = NULL;
    } else if (SWIG_ConvertPtr(obj5, (void **) &arg6, 0, 0) == -1) {
      arg6 = (void *) obj5;
      PyErr_Clear();
    }
  }
  if (obj6) {
    /* Verify that the user supplied a valid pool */
    if (obj6 != Py_None && obj6 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj6);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_replay(arg1,arg2,arg3,arg4,(struct svn_delta_editor_t const *)arg5,arg6,arg7);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_get_deleted_rev(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_session_t *arg1 = 0 ;
  char *arg2 = 0 ;
  svn_revnum_t arg3 ;
  svn_revnum_t arg4 ;
  svn_revnum_t *arg5 = 0 ;
  apr_pool_t *arg6 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_revnum_t temp5 ;
  int res5 = SWIG_TMPOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg6 = _global_pool;
  arg5 = &temp5;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_get_deleted_rev", 4, 5, &obj0, &obj1, &obj2, &obj3, &obj4)) SWIG_fail;
  {
    arg1 = (svn_ra_session_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_session_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = svn_swig_py_string_to_cstring(obj1, FALSE, "svn_ra_get_deleted_rev", "path");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    arg3 = (svn_revnum_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  {
    arg4 = (svn_revnum_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  if (obj4) {
    /* Verify that the user supplied a valid pool */
    if (obj4 != Py_None && obj4 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj4);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_get_deleted_rev(arg1,(char const *)arg2,arg3,arg4,arg5,arg6);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  if (SWIG_IsTmpObj(res5)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_long((*arg5)), 0);
  } else {
    int new_flags = SWIG_IsNewObj(res5) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg5), SWIGTYPE_p_long, new_flags), 0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_get_inherited_props(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_session_t *arg1 = 0 ;
  apr_array_header_t **arg2 = 0 ;
  char *arg3 = 0 ;
  svn_revnum_t arg4 ;
  apr_pool_t *arg5 = 0 ;
  apr_pool_t *arg6 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  apr_array_header_t *temp2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg5 = _global_pool;
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg6 = _global_pool;
  arg2 = &temp2;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_get_inherited_props", 3, 5, &obj0, &obj1, &obj2, &obj3, &obj4)) SWIG_fail;
  {
    arg1 = (svn_ra_session_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_session_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = svn_swig_py_string_to_cstring(obj1, FALSE, "svn_ra_get_inherited_props", "path");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    arg4 = (svn_revnum_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  if (obj3) {
    /* Verify that the user supplied a valid pool */
    if (obj3 != Py_None && obj3 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj3);
      SWIG_fail;
    }
  }
  if (obj4) {
    /* Verify that the user supplied a valid pool */
    if (obj4 != Py_None && obj4 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj4);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_get_inherited_props(arg1,arg2,(char const *)arg3,arg4,arg5,arg6);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_propinheriteditemarray_to_dict(*arg2), 0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_has_capability(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_session_t *arg1 = 0 ;
  svn_boolean_t *arg2 = 0 ;
  char *arg3 = 0 ;
  apr_pool_t *arg4 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_boolean_t temp2 ;
  int res2 = SWIG_TMPOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg4 = _global_pool;
  arg2 = &temp2;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_has_capability", 2, 3, &obj0, &obj1, &obj2)) SWIG_fail;
  {
    arg1 = (svn_ra_session_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_session_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = svn_swig_py_string_to_cstring(obj1, FALSE, "svn_ra_has_capability", "capability");
    if (PyErr_Occurred()) SWIG_fail;
  }
  if (obj2) {
    /* Verify that the user supplied a valid pool */
    if (obj2 != Py_None && obj2 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj2);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_has_capability(arg1,arg2,(char const *)arg3,arg4);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  if (SWIG_IsTmpObj(res2)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_int((*arg2)), 0);
  } else {
    int new_flags = SWIG_IsNewObj(res2) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg2), SWIGTYPE_p_int, new_flags), 0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_print_modules(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_stringbuf_t *arg1 = 0 ;
  apr_pool_t *arg2 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg2 = _global_pool;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_print_modules", 1, 2, &obj0, &obj1)) SWIG_fail;
  {
    if (!PyBytes_Check(obj0)) {
      PyErr_SetString(PyExc_TypeError, "not a bytes object");
      SWIG_fail;
    }
    {
      Py_ssize_t strBufLen;
      char *strBufChar;
      if (-1 == PyBytes_AsStringAndSize(obj0, &strBufChar, &strBufLen)) {
        SWIG_fail;
      }
      arg1 = svn_stringbuf_ncreate(strBufChar, strBufLen,
        /* ### gah... what pool to use? */
        _global_pool);
    }
  }
  if (obj1) {
    /* Verify that the user supplied a valid pool */
    if (obj1 != Py_None && obj1 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj1);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_print_modules(arg1,arg2);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_print_ra_libraries(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_stringbuf_t **arg1 = 0 ;
  void *arg2 = 0 ;
  apr_pool_t *arg3 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_stringbuf_t *temp1 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg3 = _global_pool;
  arg1 = &temp1;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_print_ra_libraries", 1, 2, &obj0, &obj1)) SWIG_fail;
  {
    if (obj0 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj0, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj0;
      PyErr_Clear();
    }
  }
  if (obj1) {
    /* Verify that the user supplied a valid pool */
    if (obj1 != Py_None && obj1 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj1);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_print_ra_libraries(arg1,arg2,arg3);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    PyObject *s;
    if (*arg1 == NULL) {
      Py_INCREF(Py_None);
      s = Py_None;
    } else {
      s = PyBytes_FromStringAndSize((*arg1)->data, (*arg1)->len);
      if (s == NULL)
      SWIG_fail;
    }
    resultobj = SWIG_Python_AppendOutput(resultobj, s, 0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_plugin_t_name_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_plugin_t *arg1 = 0 ;
  char *arg2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_plugin_t_name_set", 2, 2, &obj0, &obj1)) SWIG_fail;
  {
    arg1 = (struct svn_ra_plugin_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_plugin_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = svn_swig_py_string_to_cstring(obj1, FALSE, "svn_ra_plugin_t_name_set", "name");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->name) free((char *)arg1->name);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->name = copied;
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_plugin_t_name_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_plugin_t *arg1 = 0 ;
  PyObject * obj0 = 0 ;
  char *result = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_plugin_t_name_get", 1, 1, &obj0)) SWIG_fail;
  {
    arg1 = (struct svn_ra_plugin_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_plugin_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (char *) ((arg1)->name);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_plugin_t_description_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_plugin_t *arg1 = 0 ;
  char *arg2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_plugin_t_description_set", 2, 2, &obj0, &obj1)) SWIG_fail;
  {
    arg1 = (struct svn_ra_plugin_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_plugin_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg2 = svn_swig_py_string_to_cstring(obj1, FALSE, "svn_ra_plugin_t_description_set", "description");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->description) free((char *)arg1->description);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->description = copied;
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_plugin_t_description_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_plugin_t *arg1 = 0 ;
  PyObject * obj0 = 0 ;
  char *result = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_plugin_t_description_get", 1, 1, &obj0)) SWIG_fail;
  {
    arg1 = (struct svn_ra_plugin_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_plugin_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (char *) ((arg1)->description);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_plugin_t_open_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_plugin_t *arg1 = 0 ;
  svn_error_t *(*arg2)(void **,char const *,svn_ra_callbacks_t const *,void *,apr_hash_t *,apr_pool_t *) = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_plugin_t_open_set", 2, 2, &obj0, &obj1)) SWIG_fail;
  {
    arg1 = (struct svn_ra_plugin_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_plugin_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_p_void_p_q_const__char_p_q_const__svn_ra_callbacks_t_p_void_p_apr_hash_t_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_ra_plugin_t_open_set" "', argument " "2"" of type '" "svn_error_t *(*)(void **,char const *,svn_ra_callbacks_t const *,void *,apr_hash_t *,apr_pool_t *)""'"); 
    }
  }
  if (arg1) (arg1)->open = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_plugin_t_open_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_plugin_t *arg1 = 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(void **,char const *,svn_ra_callbacks_t const *,void *,apr_hash_t *,apr_pool_t *) = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_plugin_t_open_get", 1, 1, &obj0)) SWIG_fail;
  {
    arg1 = (struct svn_ra_plugin_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_plugin_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(void **,char const *,svn_ra_callbacks_t const *,void *,apr_hash_t *,apr_pool_t *)) ((arg1)->open);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_p_void_p_q_const__char_p_q_const__svn_ra_callbacks_t_p_void_p_apr_hash_t_p_apr_pool_t__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_plugin_t_get_latest_revnum_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_plugin_t *arg1 = 0 ;
  svn_error_t *(*arg2)(void *,svn_revnum_t *,apr_pool_t *) = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_plugin_t_get_latest_revnum_set", 2, 2, &obj0, &obj1)) SWIG_fail;
  {
    arg1 = (struct svn_ra_plugin_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_plugin_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_void_p_svn_revnum_t_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_ra_plugin_t_get_latest_revnum_set" "', argument " "2"" of type '" "svn_error_t *(*)(void *,svn_revnum_t *,apr_pool_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_latest_revnum = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_plugin_t_get_latest_revnum_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_plugin_t *arg1 = 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(void *,svn_revnum_t *,apr_pool_t *) = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_plugin_t_get_latest_revnum_get", 1, 1, &obj0)) SWIG_fail;
  {
    arg1 = (struct svn_ra_plugin_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_plugin_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(void *,svn_revnum_t *,apr_pool_t *)) ((arg1)->get_latest_revnum);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_void_p_svn_revnum_t_p_apr_pool_t__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_plugin_t_get_dated_revision_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_plugin_t *arg1 = 0 ;
  svn_error_t *(*arg2)(void *,svn_revnum_t *,apr_time_t,apr_pool_t *) = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_plugin_t_get_dated_revision_set", 2, 2, &obj0, &obj1)) SWIG_fail;
  {
    arg1 = (struct svn_ra_plugin_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_plugin_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_void_p_svn_revnum_t_apr_time_t_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_ra_plugin_t_get_dated_revision_set" "', argument " "2"" of type '" "svn_error_t *(*)(void *,svn_revnum_t *,apr_time_t,apr_pool_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_dated_revision = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_plugin_t_get_dated_revision_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_plugin_t *arg1 = 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(void *,svn_revnum_t *,apr_time_t,apr_pool_t *) = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_plugin_t_get_dated_revision_get", 1, 1, &obj0)) SWIG_fail;
  {
    arg1 = (struct svn_ra_plugin_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_plugin_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(void *,svn_revnum_t *,apr_time_t,apr_pool_t *)) ((arg1)->get_dated_revision);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_void_p_svn_revnum_t_apr_time_t_p_apr_pool_t__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_plugin_t_change_rev_prop_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_plugin_t *arg1 = 0 ;
  svn_error_t *(*arg2)(void *,svn_revnum_t,char const *,svn_string_t const *,apr_pool_t *) = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_plugin_t_change_rev_prop_set", 2, 2, &obj0, &obj1)) SWIG_fail;
  {
    arg1 = (struct svn_ra_plugin_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_plugin_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_void_svn_revnum_t_p_q_const__char_p_q_const__svn_string_t_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_ra_plugin_t_change_rev_prop_set" "', argument " "2"" of type '" "svn_error_t *(*)(void *,svn_revnum_t,char const *,svn_string_t const *,apr_pool_t *)""'"); 
    }
  }
  if (arg1) (arg1)->change_rev_prop = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_plugin_t_change_rev_prop_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_plugin_t *arg1 = 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(void *,svn_revnum_t,char const *,svn_string_t const *,apr_pool_t *) = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_plugin_t_change_rev_prop_get", 1, 1, &obj0)) SWIG_fail;
  {
    arg1 = (struct svn_ra_plugin_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_plugin_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(void *,svn_revnum_t,char const *,svn_string_t const *,apr_pool_t *)) ((arg1)->change_rev_prop);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_void_svn_revnum_t_p_q_const__char_p_q_const__svn_string_t_p_apr_pool_t__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_plugin_t_rev_proplist_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_plugin_t *arg1 = 0 ;
  svn_error_t *(*arg2)(void *,svn_revnum_t,apr_hash_t **,apr_pool_t *) = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_plugin_t_rev_proplist_set", 2, 2, &obj0, &obj1)) SWIG_fail;
  {
    arg1 = (struct svn_ra_plugin_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_plugin_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_void_svn_revnum_t_p_p_apr_hash_t_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_ra_plugin_t_rev_proplist_set" "', argument " "2"" of type '" "svn_error_t *(*)(void *,svn_revnum_t,apr_hash_t **,apr_pool_t *)""'"); 
    }
  }
  if (arg1) (arg1)->rev_proplist = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_plugin_t_rev_proplist_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_plugin_t *arg1 = 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(void *,svn_revnum_t,apr_hash_t **,apr_pool_t *) = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_plugin_t_rev_proplist_get", 1, 1, &obj0)) SWIG_fail;
  {
    arg1 = (struct svn_ra_plugin_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_plugin_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(void *,svn_revnum_t,apr_hash_t **,apr_pool_t *)) ((arg1)->rev_proplist);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_void_svn_revnum_t_p_p_apr_hash_t_p_apr_pool_t__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_plugin_t_rev_prop_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_plugin_t *arg1 = 0 ;
  svn_error_t *(*arg2)(void *,svn_revnum_t,char const *,svn_string_t **,apr_pool_t *) = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_plugin_t_rev_prop_set", 2, 2, &obj0, &obj1)) SWIG_fail;
  {
    arg1 = (struct svn_ra_plugin_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_plugin_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_void_svn_revnum_t_p_q_const__char_p_p_svn_string_t_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_ra_plugin_t_rev_prop_set" "', argument " "2"" of type '" "svn_error_t *(*)(void *,svn_revnum_t,char const *,svn_string_t **,apr_pool_t *)""'"); 
    }
  }
  if (arg1) (arg1)->rev_prop = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_plugin_t_rev_prop_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_plugin_t *arg1 = 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(void *,svn_revnum_t,char const *,svn_string_t **,apr_pool_t *) = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_plugin_t_rev_prop_get", 1, 1, &obj0)) SWIG_fail;
  {
    arg1 = (struct svn_ra_plugin_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_plugin_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(void *,svn_revnum_t,char const *,svn_string_t **,apr_pool_t *)) ((arg1)->rev_prop);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_void_svn_revnum_t_p_q_const__char_p_p_svn_string_t_p_apr_pool_t__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_plugin_t_get_commit_editor_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_plugin_t *arg1 = 0 ;
  svn_error_t *(*arg2)(void *,svn_delta_editor_t const **,void **,char const *,svn_commit_callback_t,void *,apr_pool_t *) = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_plugin_t_get_commit_editor_set", 2, 2, &obj0, &obj1)) SWIG_fail;
  {
    arg1 = (struct svn_ra_plugin_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_plugin_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_void_p_p_q_const__svn_delta_editor_t_p_p_void_p_q_const__char_svn_commit_callback_t_p_void_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_ra_plugin_t_get_commit_editor_set" "', argument " "2"" of type '" "svn_error_t *(*)(void *,svn_delta_editor_t const **,void **,char const *,svn_commit_callback_t,void *,apr_pool_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_commit_editor = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_plugin_t_get_commit_editor_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_plugin_t *arg1 = 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(void *,svn_delta_editor_t const **,void **,char const *,svn_commit_callback_t,void *,apr_pool_t *) = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_plugin_t_get_commit_editor_get", 1, 1, &obj0)) SWIG_fail;
  {
    arg1 = (struct svn_ra_plugin_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_plugin_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(void *,svn_delta_editor_t const **,void **,char const *,svn_commit_callback_t,void *,apr_pool_t *)) ((arg1)->get_commit_editor);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_void_p_p_q_const__svn_delta_editor_t_p_p_void_p_q_const__char_svn_commit_callback_t_p_void_p_apr_pool_t__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_plugin_t_get_file_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_plugin_t *arg1 = 0 ;
  svn_error_t *(*arg2)(void *,char const *,svn_revnum_t,svn_stream_t *,svn_revnum_t *,apr_hash_t **,apr_pool_t *) = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_plugin_t_get_file_set", 2, 2, &obj0, &obj1)) SWIG_fail;
  {
    arg1 = (struct svn_ra_plugin_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_plugin_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_void_p_q_const__char_svn_revnum_t_p_svn_stream_t_p_svn_revnum_t_p_p_apr_hash_t_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_ra_plugin_t_get_file_set" "', argument " "2"" of type '" "svn_error_t *(*)(void *,char const *,svn_revnum_t,svn_stream_t *,svn_revnum_t *,apr_hash_t **,apr_pool_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_file = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_plugin_t_get_file_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_plugin_t *arg1 = 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(void *,char const *,svn_revnum_t,svn_stream_t *,svn_revnum_t *,apr_hash_t **,apr_pool_t *) = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_plugin_t_get_file_get", 1, 1, &obj0)) SWIG_fail;
  {
    arg1 = (struct svn_ra_plugin_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_plugin_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(void *,char const *,svn_revnum_t,svn_stream_t *,svn_revnum_t *,apr_hash_t **,apr_pool_t *)) ((arg1)->get_file);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_void_p_q_const__char_svn_revnum_t_p_svn_stream_t_p_svn_revnum_t_p_p_apr_hash_t_p_apr_pool_t__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_plugin_t_get_dir_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_plugin_t *arg1 = 0 ;
  svn_error_t *(*arg2)(void *,char const *,svn_revnum_t,apr_hash_t **,svn_revnum_t *,apr_hash_t **,apr_pool_t *) = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_plugin_t_get_dir_set", 2, 2, &obj0, &obj1)) SWIG_fail;
  {
    arg1 = (struct svn_ra_plugin_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_plugin_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_void_p_q_const__char_svn_revnum_t_p_p_apr_hash_t_p_svn_revnum_t_p_p_apr_hash_t_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_ra_plugin_t_get_dir_set" "', argument " "2"" of type '" "svn_error_t *(*)(void *,char const *,svn_revnum_t,apr_hash_t **,svn_revnum_t *,apr_hash_t **,apr_pool_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_dir = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_plugin_t_get_dir_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_plugin_t *arg1 = 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(void *,char const *,svn_revnum_t,apr_hash_t **,svn_revnum_t *,apr_hash_t **,apr_pool_t *) = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_plugin_t_get_dir_get", 1, 1, &obj0)) SWIG_fail;
  {
    arg1 = (struct svn_ra_plugin_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_plugin_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(void *,char const *,svn_revnum_t,apr_hash_t **,svn_revnum_t *,apr_hash_t **,apr_pool_t *)) ((arg1)->get_dir);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_void_p_q_const__char_svn_revnum_t_p_p_apr_hash_t_p_svn_revnum_t_p_p_apr_hash_t_p_apr_pool_t__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_plugin_t_do_update_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_plugin_t *arg1 = 0 ;
  svn_error_t *(*arg2)(void *,svn_ra_reporter_t const **,void **,svn_revnum_t,char const *,svn_boolean_t,svn_delta_editor_t const *,void *,apr_pool_t *) = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_plugin_t_do_update_set", 2, 2, &obj0, &obj1)) SWIG_fail;
  {
    arg1 = (struct svn_ra_plugin_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_plugin_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_void_p_p_q_const__svn_ra_reporter_t_p_p_void_svn_revnum_t_p_q_const__char_svn_boolean_t_p_q_const__svn_delta_editor_t_p_void_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_ra_plugin_t_do_update_set" "', argument " "2"" of type '" "svn_error_t *(*)(void *,svn_ra_reporter_t const **,void **,svn_revnum_t,char const *,svn_boolean_t,svn_delta_editor_t const *,void *,apr_pool_t *)""'"); 
    }
  }
  if (arg1) (arg1)->do_update = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_plugin_t_do_update_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_plugin_t *arg1 = 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(void *,svn_ra_reporter_t const **,void **,svn_revnum_t,char const *,svn_boolean_t,svn_delta_editor_t const *,void *,apr_pool_t *) = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_plugin_t_do_update_get", 1, 1, &obj0)) SWIG_fail;
  {
    arg1 = (struct svn_ra_plugin_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_plugin_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(void *,svn_ra_reporter_t const **,void **,svn_revnum_t,char const *,svn_boolean_t,svn_delta_editor_t const *,void *,apr_pool_t *)) ((arg1)->do_update);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_void_p_p_q_const__svn_ra_reporter_t_p_p_void_svn_revnum_t_p_q_const__char_svn_boolean_t_p_q_const__svn_delta_editor_t_p_void_p_apr_pool_t__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_plugin_t_do_switch_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_plugin_t *arg1 = 0 ;
  svn_error_t *(*arg2)(void *,svn_ra_reporter_t const **,void **,svn_revnum_t,char const *,svn_boolean_t,char const *,svn_delta_editor_t const *,void *,apr_pool_t *) = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_plugin_t_do_switch_set", 2, 2, &obj0, &obj1)) SWIG_fail;
  {
    arg1 = (struct svn_ra_plugin_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_plugin_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_void_p_p_q_const__svn_ra_reporter_t_p_p_void_svn_revnum_t_p_q_const__char_svn_boolean_t_p_q_const__char_p_q_const__svn_delta_editor_t_p_void_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_ra_plugin_t_do_switch_set" "', argument " "2"" of type '" "svn_error_t *(*)(void *,svn_ra_reporter_t const **,void **,svn_revnum_t,char const *,svn_boolean_t,char const *,svn_delta_editor_t const *,void *,apr_pool_t *)""'"); 
    }
  }
  if (arg1) (arg1)->do_switch = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_plugin_t_do_switch_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_plugin_t *arg1 = 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(void *,svn_ra_reporter_t const **,void **,svn_revnum_t,char const *,svn_boolean_t,char const *,svn_delta_editor_t const *,void *,apr_pool_t *) = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_plugin_t_do_switch_get", 1, 1, &obj0)) SWIG_fail;
  {
    arg1 = (struct svn_ra_plugin_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_plugin_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(void *,svn_ra_reporter_t const **,void **,svn_revnum_t,char const *,svn_boolean_t,char const *,svn_delta_editor_t const *,void *,apr_pool_t *)) ((arg1)->do_switch);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_void_p_p_q_const__svn_ra_reporter_t_p_p_void_svn_revnum_t_p_q_const__char_svn_boolean_t_p_q_const__char_p_q_const__svn_delta_editor_t_p_void_p_apr_pool_t__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_plugin_t_do_status_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_plugin_t *arg1 = 0 ;
  svn_error_t *(*arg2)(void *,svn_ra_reporter_t const **,void **,char const *,svn_revnum_t,svn_boolean_t,svn_delta_editor_t const *,void *,apr_pool_t *) = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_plugin_t_do_status_set", 2, 2, &obj0, &obj1)) SWIG_fail;
  {
    arg1 = (struct svn_ra_plugin_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_plugin_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_void_p_p_q_const__svn_ra_reporter_t_p_p_void_p_q_const__char_svn_revnum_t_svn_boolean_t_p_q_const__svn_delta_editor_t_p_void_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_ra_plugin_t_do_status_set" "', argument " "2"" of type '" "svn_error_t *(*)(void *,svn_ra_reporter_t const **,void **,char const *,svn_revnum_t,svn_boolean_t,svn_delta_editor_t const *,void *,apr_pool_t *)""'"); 
    }
  }
  if (arg1) (arg1)->do_status = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_plugin_t_do_status_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_plugin_t *arg1 = 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(void *,svn_ra_reporter_t const **,void **,char const *,svn_revnum_t,svn_boolean_t,svn_delta_editor_t const *,void *,apr_pool_t *) = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_plugin_t_do_status_get", 1, 1, &obj0)) SWIG_fail;
  {
    arg1 = (struct svn_ra_plugin_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_plugin_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(void *,svn_ra_reporter_t const **,void **,char const *,svn_revnum_t,svn_boolean_t,svn_delta_editor_t const *,void *,apr_pool_t *)) ((arg1)->do_status);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_void_p_p_q_const__svn_ra_reporter_t_p_p_void_p_q_const__char_svn_revnum_t_svn_boolean_t_p_q_const__svn_delta_editor_t_p_void_p_apr_pool_t__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_plugin_t_do_diff_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_plugin_t *arg1 = 0 ;
  svn_error_t *(*arg2)(void *,svn_ra_reporter_t const **,void **,svn_revnum_t,char const *,svn_boolean_t,svn_boolean_t,char const *,svn_delta_editor_t const *,void *,apr_pool_t *) = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_plugin_t_do_diff_set", 2, 2, &obj0, &obj1)) SWIG_fail;
  {
    arg1 = (struct svn_ra_plugin_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_plugin_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_void_p_p_q_const__svn_ra_reporter_t_p_p_void_svn_revnum_t_p_q_const__char_svn_boolean_t_svn_boolean_t_p_q_const__char_p_q_const__svn_delta_editor_t_p_void_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_ra_plugin_t_do_diff_set" "', argument " "2"" of type '" "svn_error_t *(*)(void *,svn_ra_reporter_t const **,void **,svn_revnum_t,char const *,svn_boolean_t,svn_boolean_t,char const *,svn_delta_editor_t const *,void *,apr_pool_t *)""'"); 
    }
  }
  if (arg1) (arg1)->do_diff = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_plugin_t_do_diff_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_plugin_t *arg1 = 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(void *,svn_ra_reporter_t const **,void **,svn_revnum_t,char const *,svn_boolean_t,svn_boolean_t,char const *,svn_delta_editor_t const *,void *,apr_pool_t *) = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_plugin_t_do_diff_get", 1, 1, &obj0)) SWIG_fail;
  {
    arg1 = (struct svn_ra_plugin_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_plugin_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(void *,svn_ra_reporter_t const **,void **,svn_revnum_t,char const *,svn_boolean_t,svn_boolean_t,char const *,svn_delta_editor_t const *,void *,apr_pool_t *)) ((arg1)->do_diff);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_void_p_p_q_const__svn_ra_reporter_t_p_p_void_svn_revnum_t_p_q_const__char_svn_boolean_t_svn_boolean_t_p_q_const__char_p_q_const__svn_delta_editor_t_p_void_p_apr_pool_t__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_plugin_t_get_log_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_plugin_t *arg1 = 0 ;
  svn_error_t *(*arg2)(void *,apr_array_header_t const *,svn_revnum_t,svn_revnum_t,svn_boolean_t,svn_boolean_t,svn_log_message_receiver_t,void *,apr_pool_t *) = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_plugin_t_get_log_set", 2, 2, &obj0, &obj1)) SWIG_fail;
  {
    arg1 = (struct svn_ra_plugin_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_plugin_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_void_p_q_const__apr_array_header_t_svn_revnum_t_svn_revnum_t_svn_boolean_t_svn_boolean_t_svn_log_message_receiver_t_p_void_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_ra_plugin_t_get_log_set" "', argument " "2"" of type '" "svn_error_t *(*)(void *,apr_array_header_t const *,svn_revnum_t,svn_revnum_t,svn_boolean_t,svn_boolean_t,svn_log_message_receiver_t,void *,apr_pool_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_log = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_plugin_t_get_log_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_plugin_t *arg1 = 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(void *,apr_array_header_t const *,svn_revnum_t,svn_revnum_t,svn_boolean_t,svn_boolean_t,svn_log_message_receiver_t,void *,apr_pool_t *) = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_plugin_t_get_log_get", 1, 1, &obj0)) SWIG_fail;
  {
    arg1 = (struct svn_ra_plugin_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_plugin_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(void *,apr_array_header_t const *,svn_revnum_t,svn_revnum_t,svn_boolean_t,svn_boolean_t,svn_log_message_receiver_t,void *,apr_pool_t *)) ((arg1)->get_log);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_void_p_q_const__apr_array_header_t_svn_revnum_t_svn_revnum_t_svn_boolean_t_svn_boolean_t_svn_log_message_receiver_t_p_void_p_apr_pool_t__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_plugin_t_check_path_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_plugin_t *arg1 = 0 ;
  svn_error_t *(*arg2)(void *,char const *,svn_revnum_t,svn_node_kind_t *,apr_pool_t *) = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_plugin_t_check_path_set", 2, 2, &obj0, &obj1)) SWIG_fail;
  {
    arg1 = (struct svn_ra_plugin_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_plugin_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_void_p_q_const__char_svn_revnum_t_p_svn_node_kind_t_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_ra_plugin_t_check_path_set" "', argument " "2"" of type '" "svn_error_t *(*)(void *,char const *,svn_revnum_t,svn_node_kind_t *,apr_pool_t *)""'"); 
    }
  }
  if (arg1) (arg1)->check_path = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_plugin_t_check_path_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_plugin_t *arg1 = 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(void *,char const *,svn_revnum_t,svn_node_kind_t *,apr_pool_t *) = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_plugin_t_check_path_get", 1, 1, &obj0)) SWIG_fail;
  {
    arg1 = (struct svn_ra_plugin_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_plugin_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(void *,char const *,svn_revnum_t,svn_node_kind_t *,apr_pool_t *)) ((arg1)->check_path);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_void_p_q_const__char_svn_revnum_t_p_svn_node_kind_t_p_apr_pool_t__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_plugin_t_get_uuid_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_plugin_t *arg1 = 0 ;
  svn_error_t *(*arg2)(void *,char const **,apr_pool_t *) = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_plugin_t_get_uuid_set", 2, 2, &obj0, &obj1)) SWIG_fail;
  {
    arg1 = (struct svn_ra_plugin_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_plugin_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_void_p_p_q_const__char_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_ra_plugin_t_get_uuid_set" "', argument " "2"" of type '" "svn_error_t *(*)(void *,char const **,apr_pool_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_uuid = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_plugin_t_get_uuid_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_plugin_t *arg1 = 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(void *,char const **,apr_pool_t *) = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_plugin_t_get_uuid_get", 1, 1, &obj0)) SWIG_fail;
  {
    arg1 = (struct svn_ra_plugin_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_plugin_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(void *,char const **,apr_pool_t *)) ((arg1)->get_uuid);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_void_p_p_q_const__char_p_apr_pool_t__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_plugin_t_get_repos_root_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_plugin_t *arg1 = 0 ;
  svn_error_t *(*arg2)(void *,char const **,apr_pool_t *) = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_plugin_t_get_repos_root_set", 2, 2, &obj0, &obj1)) SWIG_fail;
  {
    arg1 = (struct svn_ra_plugin_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_plugin_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_void_p_p_q_const__char_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_ra_plugin_t_get_repos_root_set" "', argument " "2"" of type '" "svn_error_t *(*)(void *,char const **,apr_pool_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_repos_root = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_plugin_t_get_repos_root_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_plugin_t *arg1 = 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(void *,char const **,apr_pool_t *) = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_plugin_t_get_repos_root_get", 1, 1, &obj0)) SWIG_fail;
  {
    arg1 = (struct svn_ra_plugin_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_plugin_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(void *,char const **,apr_pool_t *)) ((arg1)->get_repos_root);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_void_p_p_q_const__char_p_apr_pool_t__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_plugin_t_get_locations_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_plugin_t *arg1 = 0 ;
  svn_error_t *(*arg2)(void *,apr_hash_t **,char const *,svn_revnum_t,apr_array_header_t *,apr_pool_t *) = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_plugin_t_get_locations_set", 2, 2, &obj0, &obj1)) SWIG_fail;
  {
    arg1 = (struct svn_ra_plugin_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_plugin_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_void_p_p_apr_hash_t_p_q_const__char_svn_revnum_t_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_ra_plugin_t_get_locations_set" "', argument " "2"" of type '" "svn_error_t *(*)(void *,apr_hash_t **,char const *,svn_revnum_t,apr_array_header_t *,apr_pool_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_locations = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_plugin_t_get_locations_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_plugin_t *arg1 = 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(void *,apr_hash_t **,char const *,svn_revnum_t,apr_array_header_t *,apr_pool_t *) = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_plugin_t_get_locations_get", 1, 1, &obj0)) SWIG_fail;
  {
    arg1 = (struct svn_ra_plugin_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_plugin_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(void *,apr_hash_t **,char const *,svn_revnum_t,apr_array_header_t *,apr_pool_t *)) ((arg1)->get_locations);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_void_p_p_apr_hash_t_p_q_const__char_svn_revnum_t_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_plugin_t_get_file_revs_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_plugin_t *arg1 = 0 ;
  svn_error_t *(*arg2)(void *,char const *,svn_revnum_t,svn_revnum_t,svn_ra_file_rev_handler_t,void *,apr_pool_t *) = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_plugin_t_get_file_revs_set", 2, 2, &obj0, &obj1)) SWIG_fail;
  {
    arg1 = (struct svn_ra_plugin_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_plugin_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_p_void_p_q_const__char_svn_revnum_t_svn_revnum_t_svn_ra_file_rev_handler_t_p_void_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_ra_plugin_t_get_file_revs_set" "', argument " "2"" of type '" "svn_error_t *(*)(void *,char const *,svn_revnum_t,svn_revnum_t,svn_ra_file_rev_handler_t,void *,apr_pool_t *)""'"); 
    }
  }
  if (arg1) (arg1)->get_file_revs = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_plugin_t_get_file_revs_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_plugin_t *arg1 = 0 ;
  PyObject * obj0 = 0 ;
  svn_error_t *(*result)(void *,char const *,svn_revnum_t,svn_revnum_t,svn_ra_file_rev_handler_t,void *,apr_pool_t *) = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_plugin_t_get_file_revs_get", 1, 1, &obj0)) SWIG_fail;
  {
    arg1 = (struct svn_ra_plugin_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_plugin_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_error_t *(*)(void *,char const *,svn_revnum_t,svn_revnum_t,svn_ra_file_rev_handler_t,void *,apr_pool_t *)) ((arg1)->get_file_revs);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_void_p_q_const__char_svn_revnum_t_svn_revnum_t_svn_ra_file_rev_handler_t_p_void_p_apr_pool_t__p_svn_error_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_plugin_t_get_version_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_plugin_t *arg1 = 0 ;
  svn_version_t *(*arg2)(void) = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_plugin_t_get_version_set", 2, 2, &obj0, &obj1)) SWIG_fail;
  {
    arg1 = (struct svn_ra_plugin_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_plugin_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    int res = SWIG_ConvertFunctionPtr(obj1, (void**)(&arg2), SWIGTYPE_p_f_void__p_svn_version_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), "in method '" "svn_ra_plugin_t_get_version_set" "', argument " "2"" of type '" "svn_version_t const *(*)(void)""'"); 
    }
  }
  if (arg1) (arg1)->get_version = (svn_version_t const *(*)(void))arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_plugin_t_get_version_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_plugin_t *arg1 = 0 ;
  PyObject * obj0 = 0 ;
  svn_version_t *(*result)(void) = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_plugin_t_get_version_get", 1, 1, &obj0)) SWIG_fail;
  {
    arg1 = (struct svn_ra_plugin_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_plugin_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  result = (svn_version_t *(*)(void)) ((arg1)->get_version);
  resultobj = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_void__p_svn_version_t);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_svn_ra_plugin_t(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_plugin_t *result = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "new_svn_ra_plugin_t", 0, 0)) SWIG_fail;
  {
    svn_swig_py_release_py_lock();
    
    result = (struct svn_ra_plugin_t *)calloc(1, sizeof(struct svn_ra_plugin_t));
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_ra_plugin_t,
    _global_py_pool, args);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_svn_ra_plugin_t(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  struct svn_ra_plugin_t *arg1 = 0 ;
  PyObject * obj0 = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "delete_svn_ra_plugin_t", 1, 1, &obj0)) SWIG_fail;
  {
    arg1 = (struct svn_ra_plugin_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_plugin_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    free((char *) arg1);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *svn_ra_plugin_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_svn_ra_plugin_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *svn_ra_plugin_t_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_svn_ra_init_ra_libs(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  void **arg1 = 0 ;
  apr_pool_t *arg2 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  void *temp1 ;
  PyObject * obj0 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg2 = _global_pool;
  arg1 = &temp1;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_init_ra_libs", 0, 1, &obj0)) SWIG_fail;
  if (obj0) {
    /* Verify that the user supplied a valid pool */
    if (obj0 != Py_None && obj0 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj0);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_init_ra_libs(arg1,arg2);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg1, SWIGTYPE_p_void,
        _global_py_pool, args), 0)
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_get_ra_library(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_plugin_t **arg1 = 0 ;
  void *arg2 = 0 ;
  char *arg3 = 0 ;
  apr_pool_t *arg4 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_ra_plugin_t *temp1 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg4 = _global_pool;
  arg1 = &temp1;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_get_ra_library", 2, 3, &obj0, &obj1, &obj2)) SWIG_fail;
  {
    if (obj0 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj0, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj0;
      PyErr_Clear();
    }
  }
  {
    arg3 = svn_swig_py_string_to_cstring(obj1, FALSE, "svn_ra_get_ra_library", "url");
    if (PyErr_Occurred()) SWIG_fail;
  }
  if (obj2) {
    /* Verify that the user supplied a valid pool */
    if (obj2 != Py_None && obj2 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj2);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_get_ra_library(arg1,arg2,(char const *)arg3,arg4);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg1, SWIGTYPE_p_svn_ra_plugin_t,
        _global_py_pool, args), 0)
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *svn_ra_session_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_svn_ra_session_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_svn_ra_reporter3_invoke_set_path(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_reporter3_t *arg1 = 0 ;
  void *arg2 = 0 ;
  char *arg3 = 0 ;
  svn_revnum_t arg4 ;
  svn_depth_t arg5 ;
  svn_boolean_t arg6 ;
  char *arg7 = 0 ;
  apr_pool_t *arg8 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg8 = _global_pool;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_reporter3_invoke_set_path", 7, 8, &obj0, &obj1, &obj2, &obj3, &obj4, &obj5, &obj6, &obj7)) SWIG_fail;
  {
    arg1 = (svn_ra_reporter3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_reporter3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  {
    arg3 = svn_swig_py_string_to_cstring(obj2, FALSE, "svn_ra_reporter3_invoke_set_path", "path");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    arg4 = (svn_revnum_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg5 = (svn_depth_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  {
    arg6 = (svn_boolean_t)SWIG_As_long (obj5);
    if (SWIG_arg_fail(svn_argnum_obj5)) {
      SWIG_fail;
    }
  }
  {
    arg7 = svn_swig_py_string_to_cstring(obj6, TRUE, "svn_ra_reporter3_invoke_set_path", "lock_token");
    if (PyErr_Occurred()) SWIG_fail;
  }
  if (obj7) {
    /* Verify that the user supplied a valid pool */
    if (obj7 != Py_None && obj7 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj7);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_reporter3_invoke_set_path(arg1,arg2,(char const *)arg3,arg4,arg5,arg6,(char const *)arg7,arg8);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_reporter3_invoke_delete_path(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_reporter3_t *arg1 = 0 ;
  void *arg2 = 0 ;
  char *arg3 = 0 ;
  apr_pool_t *arg4 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg4 = _global_pool;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_reporter3_invoke_delete_path", 3, 4, &obj0, &obj1, &obj2, &obj3)) SWIG_fail;
  {
    arg1 = (svn_ra_reporter3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_reporter3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  {
    arg3 = svn_swig_py_string_to_cstring(obj2, FALSE, "svn_ra_reporter3_invoke_delete_path", "path");
    if (PyErr_Occurred()) SWIG_fail;
  }
  if (obj3) {
    /* Verify that the user supplied a valid pool */
    if (obj3 != Py_None && obj3 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj3);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_reporter3_invoke_delete_path(arg1,arg2,(char const *)arg3,arg4);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_reporter3_invoke_link_path(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_reporter3_t *arg1 = 0 ;
  void *arg2 = 0 ;
  char *arg3 = 0 ;
  char *arg4 = 0 ;
  svn_revnum_t arg5 ;
  svn_depth_t arg6 ;
  svn_boolean_t arg7 ;
  char *arg8 = 0 ;
  apr_pool_t *arg9 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg9 = _global_pool;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_reporter3_invoke_link_path", 8, 9, &obj0, &obj1, &obj2, &obj3, &obj4, &obj5, &obj6, &obj7, &obj8)) SWIG_fail;
  {
    arg1 = (svn_ra_reporter3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_reporter3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  {
    arg3 = svn_swig_py_string_to_cstring(obj2, FALSE, "svn_ra_reporter3_invoke_link_path", "path");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    arg4 = svn_swig_py_string_to_cstring(obj3, FALSE, "svn_ra_reporter3_invoke_link_path", "url");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    arg5 = (svn_revnum_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  {
    arg6 = (svn_depth_t)SWIG_As_long (obj5);
    if (SWIG_arg_fail(svn_argnum_obj5)) {
      SWIG_fail;
    }
  }
  {
    arg7 = (svn_boolean_t)SWIG_As_long (obj6);
    if (SWIG_arg_fail(svn_argnum_obj6)) {
      SWIG_fail;
    }
  }
  {
    arg8 = svn_swig_py_string_to_cstring(obj7, TRUE, "svn_ra_reporter3_invoke_link_path", "lock_token");
    if (PyErr_Occurred()) SWIG_fail;
  }
  if (obj8) {
    /* Verify that the user supplied a valid pool */
    if (obj8 != Py_None && obj8 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj8);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_reporter3_invoke_link_path(arg1,arg2,(char const *)arg3,(char const *)arg4,arg5,arg6,arg7,(char const *)arg8,arg9);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_reporter3_invoke_finish_report(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_reporter3_t *arg1 = 0 ;
  void *arg2 = 0 ;
  apr_pool_t *arg3 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg3 = _global_pool;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_reporter3_invoke_finish_report", 2, 3, &obj0, &obj1, &obj2)) SWIG_fail;
  {
    arg1 = (svn_ra_reporter3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_reporter3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  if (obj2) {
    /* Verify that the user supplied a valid pool */
    if (obj2 != Py_None && obj2 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj2);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_reporter3_invoke_finish_report(arg1,arg2,arg3);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_reporter3_invoke_abort_report(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_reporter3_t *arg1 = 0 ;
  void *arg2 = 0 ;
  apr_pool_t *arg3 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg3 = _global_pool;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_reporter3_invoke_abort_report", 2, 3, &obj0, &obj1, &obj2)) SWIG_fail;
  {
    arg1 = (svn_ra_reporter3_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_reporter3_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  if (obj2) {
    /* Verify that the user supplied a valid pool */
    if (obj2 != Py_None && obj2 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj2);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_reporter3_invoke_abort_report(arg1,arg2,arg3);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_reporter2_invoke_set_path(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_reporter2_t *arg1 = 0 ;
  void *arg2 = 0 ;
  char *arg3 = 0 ;
  svn_revnum_t arg4 ;
  svn_boolean_t arg5 ;
  char *arg6 = 0 ;
  apr_pool_t *arg7 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg7 = _global_pool;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_reporter2_invoke_set_path", 6, 7, &obj0, &obj1, &obj2, &obj3, &obj4, &obj5, &obj6)) SWIG_fail;
  {
    arg1 = (svn_ra_reporter2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_reporter2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  {
    arg3 = svn_swig_py_string_to_cstring(obj2, FALSE, "svn_ra_reporter2_invoke_set_path", "path");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    arg4 = (svn_revnum_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg5 = (svn_boolean_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  {
    arg6 = svn_swig_py_string_to_cstring(obj5, TRUE, "svn_ra_reporter2_invoke_set_path", "lock_token");
    if (PyErr_Occurred()) SWIG_fail;
  }
  if (obj6) {
    /* Verify that the user supplied a valid pool */
    if (obj6 != Py_None && obj6 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj6);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_reporter2_invoke_set_path(arg1,arg2,(char const *)arg3,arg4,arg5,(char const *)arg6,arg7);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_reporter2_invoke_delete_path(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_reporter2_t *arg1 = 0 ;
  void *arg2 = 0 ;
  char *arg3 = 0 ;
  apr_pool_t *arg4 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg4 = _global_pool;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_reporter2_invoke_delete_path", 3, 4, &obj0, &obj1, &obj2, &obj3)) SWIG_fail;
  {
    arg1 = (svn_ra_reporter2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_reporter2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  {
    arg3 = svn_swig_py_string_to_cstring(obj2, FALSE, "svn_ra_reporter2_invoke_delete_path", "path");
    if (PyErr_Occurred()) SWIG_fail;
  }
  if (obj3) {
    /* Verify that the user supplied a valid pool */
    if (obj3 != Py_None && obj3 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj3);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_reporter2_invoke_delete_path(arg1,arg2,(char const *)arg3,arg4);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_reporter2_invoke_link_path(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_reporter2_t *arg1 = 0 ;
  void *arg2 = 0 ;
  char *arg3 = 0 ;
  char *arg4 = 0 ;
  svn_revnum_t arg5 ;
  svn_boolean_t arg6 ;
  char *arg7 = 0 ;
  apr_pool_t *arg8 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg8 = _global_pool;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_reporter2_invoke_link_path", 7, 8, &obj0, &obj1, &obj2, &obj3, &obj4, &obj5, &obj6, &obj7)) SWIG_fail;
  {
    arg1 = (svn_ra_reporter2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_reporter2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  {
    arg3 = svn_swig_py_string_to_cstring(obj2, FALSE, "svn_ra_reporter2_invoke_link_path", "path");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    arg4 = svn_swig_py_string_to_cstring(obj3, FALSE, "svn_ra_reporter2_invoke_link_path", "url");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    arg5 = (svn_revnum_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  {
    arg6 = (svn_boolean_t)SWIG_As_long (obj5);
    if (SWIG_arg_fail(svn_argnum_obj5)) {
      SWIG_fail;
    }
  }
  {
    arg7 = svn_swig_py_string_to_cstring(obj6, TRUE, "svn_ra_reporter2_invoke_link_path", "lock_token");
    if (PyErr_Occurred()) SWIG_fail;
  }
  if (obj7) {
    /* Verify that the user supplied a valid pool */
    if (obj7 != Py_None && obj7 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj7);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_reporter2_invoke_link_path(arg1,arg2,(char const *)arg3,(char const *)arg4,arg5,arg6,(char const *)arg7,arg8);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_reporter2_invoke_finish_report(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_reporter2_t *arg1 = 0 ;
  void *arg2 = 0 ;
  apr_pool_t *arg3 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg3 = _global_pool;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_reporter2_invoke_finish_report", 2, 3, &obj0, &obj1, &obj2)) SWIG_fail;
  {
    arg1 = (svn_ra_reporter2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_reporter2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  if (obj2) {
    /* Verify that the user supplied a valid pool */
    if (obj2 != Py_None && obj2 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj2);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_reporter2_invoke_finish_report(arg1,arg2,arg3);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_reporter2_invoke_abort_report(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_reporter2_t *arg1 = 0 ;
  void *arg2 = 0 ;
  apr_pool_t *arg3 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg3 = _global_pool;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_reporter2_invoke_abort_report", 2, 3, &obj0, &obj1, &obj2)) SWIG_fail;
  {
    arg1 = (svn_ra_reporter2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_reporter2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  if (obj2) {
    /* Verify that the user supplied a valid pool */
    if (obj2 != Py_None && obj2 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj2);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_reporter2_invoke_abort_report(arg1,arg2,arg3);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_reporter_invoke_set_path(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_reporter_t *arg1 = 0 ;
  void *arg2 = 0 ;
  char *arg3 = 0 ;
  svn_revnum_t arg4 ;
  svn_boolean_t arg5 ;
  apr_pool_t *arg6 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg6 = _global_pool;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_reporter_invoke_set_path", 5, 6, &obj0, &obj1, &obj2, &obj3, &obj4, &obj5)) SWIG_fail;
  {
    arg1 = (svn_ra_reporter_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_reporter_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  {
    arg3 = svn_swig_py_string_to_cstring(obj2, FALSE, "svn_ra_reporter_invoke_set_path", "path");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    arg4 = (svn_revnum_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg5 = (svn_boolean_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  if (obj5) {
    /* Verify that the user supplied a valid pool */
    if (obj5 != Py_None && obj5 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj5);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_reporter_invoke_set_path(arg1,arg2,(char const *)arg3,arg4,arg5,arg6);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_reporter_invoke_delete_path(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_reporter_t *arg1 = 0 ;
  void *arg2 = 0 ;
  char *arg3 = 0 ;
  apr_pool_t *arg4 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg4 = _global_pool;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_reporter_invoke_delete_path", 3, 4, &obj0, &obj1, &obj2, &obj3)) SWIG_fail;
  {
    arg1 = (svn_ra_reporter_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_reporter_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  {
    arg3 = svn_swig_py_string_to_cstring(obj2, FALSE, "svn_ra_reporter_invoke_delete_path", "path");
    if (PyErr_Occurred()) SWIG_fail;
  }
  if (obj3) {
    /* Verify that the user supplied a valid pool */
    if (obj3 != Py_None && obj3 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj3);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_reporter_invoke_delete_path(arg1,arg2,(char const *)arg3,arg4);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_reporter_invoke_link_path(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_reporter_t *arg1 = 0 ;
  void *arg2 = 0 ;
  char *arg3 = 0 ;
  char *arg4 = 0 ;
  svn_revnum_t arg5 ;
  svn_boolean_t arg6 ;
  apr_pool_t *arg7 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg7 = _global_pool;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_reporter_invoke_link_path", 6, 7, &obj0, &obj1, &obj2, &obj3, &obj4, &obj5, &obj6)) SWIG_fail;
  {
    arg1 = (svn_ra_reporter_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_reporter_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  {
    arg3 = svn_swig_py_string_to_cstring(obj2, FALSE, "svn_ra_reporter_invoke_link_path", "path");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    arg4 = svn_swig_py_string_to_cstring(obj3, FALSE, "svn_ra_reporter_invoke_link_path", "url");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    arg5 = (svn_revnum_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  {
    arg6 = (svn_boolean_t)SWIG_As_long (obj5);
    if (SWIG_arg_fail(svn_argnum_obj5)) {
      SWIG_fail;
    }
  }
  if (obj6) {
    /* Verify that the user supplied a valid pool */
    if (obj6 != Py_None && obj6 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj6);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_reporter_invoke_link_path(arg1,arg2,(char const *)arg3,(char const *)arg4,arg5,arg6,arg7);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_reporter_invoke_finish_report(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_reporter_t *arg1 = 0 ;
  void *arg2 = 0 ;
  apr_pool_t *arg3 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg3 = _global_pool;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_reporter_invoke_finish_report", 2, 3, &obj0, &obj1, &obj2)) SWIG_fail;
  {
    arg1 = (svn_ra_reporter_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_reporter_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  if (obj2) {
    /* Verify that the user supplied a valid pool */
    if (obj2 != Py_None && obj2 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj2);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_reporter_invoke_finish_report(arg1,arg2,arg3);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_reporter_invoke_abort_report(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_reporter_t *arg1 = 0 ;
  void *arg2 = 0 ;
  apr_pool_t *arg3 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg3 = _global_pool;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_reporter_invoke_abort_report", 2, 3, &obj0, &obj1, &obj2)) SWIG_fail;
  {
    arg1 = (svn_ra_reporter_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_reporter_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  if (obj2) {
    /* Verify that the user supplied a valid pool */
    if (obj2 != Py_None && obj2 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj2);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_reporter_invoke_abort_report(arg1,arg2,arg3);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_callbacks2_invoke_open_tmp_file(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_callbacks2_t *arg1 = 0 ;
  apr_file_t **arg2 = 0 ;
  void *arg3 = 0 ;
  apr_pool_t *arg4 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  apr_file_t *temp2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg4 = _global_pool;
  arg2 = &temp2;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_callbacks2_invoke_open_tmp_file", 2, 3, &obj0, &obj1, &obj2)) SWIG_fail;
  {
    arg1 = (svn_ra_callbacks2_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_callbacks2_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj1 == Py_None) {
      arg3 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg3, 0, 0) == -1) {
      arg3 = (void *) obj1;
      PyErr_Clear();
    }
  }
  if (obj2) {
    /* Verify that the user supplied a valid pool */
    if (obj2 != Py_None && obj2 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj2);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_callbacks2_invoke_open_tmp_file(arg1,arg2,arg3,arg4);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg2, SWIGTYPE_p_apr_file_t,
        _global_py_pool, args), 0)
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_callbacks_invoke_open_tmp_file(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_callbacks_t *arg1 = 0 ;
  apr_file_t **arg2 = 0 ;
  void *arg3 = 0 ;
  apr_pool_t *arg4 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  apr_file_t *temp2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg4 = _global_pool;
  arg2 = &temp2;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_callbacks_invoke_open_tmp_file", 2, 3, &obj0, &obj1, &obj2)) SWIG_fail;
  {
    arg1 = (svn_ra_callbacks_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_callbacks_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj1 == Py_None) {
      arg3 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg3, 0, 0) == -1) {
      arg3 = (void *) obj1;
      PyErr_Clear();
    }
  }
  if (obj2) {
    /* Verify that the user supplied a valid pool */
    if (obj2 != Py_None && obj2 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj2);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_callbacks_invoke_open_tmp_file(arg1,arg2,arg3,arg4);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg2, SWIGTYPE_p_apr_file_t,
        _global_py_pool, args), 0)
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_plugin_invoke_open(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_plugin_t *arg1 = 0 ;
  void **arg2 = 0 ;
  char *arg3 = 0 ;
  svn_ra_callbacks_t *arg4 = 0 ;
  void *arg5 = 0 ;
  apr_hash_t *arg6 = 0 ;
  apr_pool_t *arg7 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  void *temp2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg7 = _global_pool;
  arg2 = &temp2;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_plugin_invoke_open", 5, 6, &obj0, &obj1, &obj2, &obj3, &obj4, &obj5)) SWIG_fail;
  {
    arg1 = (svn_ra_plugin_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_plugin_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg3 = svn_swig_py_string_to_cstring(obj1, FALSE, "svn_ra_plugin_invoke_open", "repos_URL");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    arg4 = (svn_ra_callbacks_t *)svn_swig_py_must_get_ptr(obj2, SWIGTYPE_p_svn_ra_callbacks_t, svn_argnum_obj2);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj3 == Py_None) {
      arg5 = NULL;
    } else if (SWIG_ConvertPtr(obj3, (void **) &arg5, 0, 0) == -1) {
      arg5 = (void *) obj3;
      PyErr_Clear();
    }
  }
  {
    if (_global_pool == NULL)
    {
      if (svn_swig_py_get_parent_pool(args, SWIGTYPE_p_apr_pool_t,
          &_global_py_pool, &_global_pool))
      SWIG_fail;
    }
    
    arg6 = svn_swig_py_struct_ptr_hash_from_dict(obj4,
      SWIGTYPE_p_svn_config_t, _global_pool);
    if (PyErr_Occurred())
    SWIG_fail;
  }
  if (obj5) {
    /* Verify that the user supplied a valid pool */
    if (obj5 != Py_None && obj5 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj5);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_plugin_invoke_open(arg1,arg2,(char const *)arg3,(struct svn_ra_callbacks_t const *)arg4,arg5,arg6,arg7);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg2, SWIGTYPE_p_void,
        _global_py_pool, args), 0)
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_plugin_invoke_get_latest_revnum(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_plugin_t *arg1 = 0 ;
  void *arg2 = 0 ;
  svn_revnum_t *arg3 = 0 ;
  apr_pool_t *arg4 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_revnum_t temp3 ;
  int res3 = SWIG_TMPOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg4 = _global_pool;
  arg3 = &temp3;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_plugin_invoke_get_latest_revnum", 2, 3, &obj0, &obj1, &obj2)) SWIG_fail;
  {
    arg1 = (svn_ra_plugin_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_plugin_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  if (obj2) {
    /* Verify that the user supplied a valid pool */
    if (obj2 != Py_None && obj2 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj2);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_plugin_invoke_get_latest_revnum(arg1,arg2,arg3,arg4);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_long((*arg3)), 0);
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_long, new_flags), 0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_plugin_invoke_get_dated_revision(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_plugin_t *arg1 = 0 ;
  void *arg2 = 0 ;
  svn_revnum_t *arg3 = 0 ;
  apr_time_t arg4 ;
  apr_pool_t *arg5 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_revnum_t temp3 ;
  int res3 = SWIG_TMPOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg5 = _global_pool;
  arg3 = &temp3;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_plugin_invoke_get_dated_revision", 3, 4, &obj0, &obj1, &obj2, &obj3)) SWIG_fail;
  {
    arg1 = (svn_ra_plugin_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_plugin_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  arg4 = (apr_time_t) PyLong_AsLongLong(obj2); 
  if (obj3) {
    /* Verify that the user supplied a valid pool */
    if (obj3 != Py_None && obj3 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj3);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_plugin_invoke_get_dated_revision(arg1,arg2,arg3,arg4,arg5);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_long((*arg3)), 0);
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_long, new_flags), 0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_plugin_invoke_change_rev_prop(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_plugin_t *arg1 = 0 ;
  void *arg2 = 0 ;
  svn_revnum_t arg3 ;
  char *arg4 = 0 ;
  svn_string_t *arg5 = 0 ;
  apr_pool_t *arg6 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_string_t value5 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg6 = _global_pool;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_plugin_invoke_change_rev_prop", 5, 6, &obj0, &obj1, &obj2, &obj3, &obj4, &obj5)) SWIG_fail;
  {
    arg1 = (svn_ra_plugin_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_plugin_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  {
    arg3 = (svn_revnum_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  {
    arg4 = svn_swig_py_string_to_cstring(obj3, FALSE, "svn_ra_plugin_invoke_change_rev_prop", "name");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    if (obj4 == Py_None)
    arg5 = NULL;
    else {
      Py_ssize_t pyStrLen;
      if (!PyBytes_Check(obj4)) {
        PyErr_SetString(PyExc_TypeError, "not a bytes object");
        SWIG_fail;
      }
      if (PyBytes_AsStringAndSize(obj4, (char **)&(value5.data),
          &pyStrLen) == -1) {
        SWIG_fail;
      }
      value5.len = pyStrLen;
      arg5 = &value5;
    }
  }
  if (obj5) {
    /* Verify that the user supplied a valid pool */
    if (obj5 != Py_None && obj5 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj5);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_plugin_invoke_change_rev_prop(arg1,arg2,arg3,(char const *)arg4,(struct svn_string_t const *)arg5,arg6);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_plugin_invoke_rev_proplist(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_plugin_t *arg1 = 0 ;
  void *arg2 = 0 ;
  svn_revnum_t arg3 ;
  apr_hash_t **arg4 = 0 ;
  apr_pool_t *arg5 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  apr_hash_t *temp4 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg5 = _global_pool;
  arg4 = &temp4;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_plugin_invoke_rev_proplist", 3, 4, &obj0, &obj1, &obj2, &obj3)) SWIG_fail;
  {
    arg1 = (svn_ra_plugin_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_plugin_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  {
    arg3 = (svn_revnum_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  if (obj3) {
    /* Verify that the user supplied a valid pool */
    if (obj3 != Py_None && obj3 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj3);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_plugin_invoke_rev_proplist(arg1,arg2,arg3,arg4,arg5);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_prophash_to_dict(*arg4), 0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_plugin_invoke_rev_prop(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_plugin_t *arg1 = 0 ;
  void *arg2 = 0 ;
  svn_revnum_t arg3 ;
  char *arg4 = 0 ;
  svn_string_t **arg5 = 0 ;
  apr_pool_t *arg6 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_string_t *temp5 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg6 = _global_pool;
  arg5 = &temp5;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_plugin_invoke_rev_prop", 4, 5, &obj0, &obj1, &obj2, &obj3, &obj4)) SWIG_fail;
  {
    arg1 = (svn_ra_plugin_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_plugin_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  {
    arg3 = (svn_revnum_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  {
    arg4 = svn_swig_py_string_to_cstring(obj3, FALSE, "svn_ra_plugin_invoke_rev_prop", "name");
    if (PyErr_Occurred()) SWIG_fail;
  }
  if (obj4) {
    /* Verify that the user supplied a valid pool */
    if (obj4 != Py_None && obj4 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj4);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_plugin_invoke_rev_prop(arg1,arg2,arg3,(char const *)arg4,arg5,arg6);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    PyObject *s;
    if (*arg5 == NULL) {
      Py_INCREF(Py_None);
      s = Py_None;
    } else {
      s = PyBytes_FromStringAndSize((*arg5)->data, (*arg5)->len);
      if (s == NULL)
      SWIG_fail;
    }
    resultobj = SWIG_Python_AppendOutput(resultobj, s, 0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_plugin_invoke_get_commit_editor(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_plugin_t *arg1 = 0 ;
  void *arg2 = 0 ;
  svn_delta_editor_t **arg3 = 0 ;
  void **arg4 = 0 ;
  char *arg5 = 0 ;
  svn_commit_callback_t arg6 = 0 ;
  void *arg7 = 0 ;
  apr_pool_t *arg8 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_delta_editor_t *temp3 ;
  void *temp4 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg8 = _global_pool;
  arg3 = &temp3;
  arg4 = &temp4;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_plugin_invoke_get_commit_editor", 4, 5, &obj0, &obj1, &obj2, &obj3, &obj4)) SWIG_fail;
  {
    arg1 = (svn_ra_plugin_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_plugin_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  {
    arg5 = svn_swig_py_string_to_cstring(obj2, FALSE, "svn_ra_plugin_invoke_get_commit_editor", "log_msg");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    arg6 = svn_swig_py_commit_callback;
    arg7 = (void *)obj3;
  }
  if (obj4) {
    /* Verify that the user supplied a valid pool */
    if (obj4 != Py_None && obj4 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj4);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_plugin_invoke_get_commit_editor(arg1,arg2,(struct svn_delta_editor_t const **)arg3,arg4,(char const *)arg5,arg6,arg7,arg8);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg3, SWIGTYPE_p_svn_delta_editor_t,
        _global_py_pool, args), 0)
    
    ;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg4, SWIGTYPE_p_void,
        _global_py_pool, args), 0)
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_plugin_invoke_get_file(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_plugin_t *arg1 = 0 ;
  void *arg2 = 0 ;
  char *arg3 = 0 ;
  svn_revnum_t arg4 ;
  svn_stream_t *arg5 = 0 ;
  svn_revnum_t *arg6 = 0 ;
  apr_hash_t **arg7 = 0 ;
  apr_pool_t *arg8 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_revnum_t temp6 ;
  int res6 = SWIG_TMPOBJ ;
  apr_hash_t *temp7 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg8 = _global_pool;
  arg6 = &temp6;
  arg7 = &temp7;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_plugin_invoke_get_file", 5, 6, &obj0, &obj1, &obj2, &obj3, &obj4, &obj5)) SWIG_fail;
  {
    arg1 = (svn_ra_plugin_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_plugin_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  {
    arg3 = svn_swig_py_string_to_cstring(obj2, FALSE, "svn_ra_plugin_invoke_get_file", "path");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    arg4 = (svn_revnum_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    if (obj4 == Py_None) {
      arg5 = NULL;
    }
    else {
      arg5 = svn_swig_py_make_stream (obj4, _global_pool);
      if (arg5 == NULL) {
        SWIG_fail;
      }
    }
  }
  if (obj5) {
    /* Verify that the user supplied a valid pool */
    if (obj5 != Py_None && obj5 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj5);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_plugin_invoke_get_file(arg1,arg2,(char const *)arg3,arg4,arg5,arg6,arg7,arg8);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  if (SWIG_IsTmpObj(res6)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_long((*arg6)), 0);
  } else {
    int new_flags = SWIG_IsNewObj(res6) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg6), SWIGTYPE_p_long, new_flags), 0);
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_prophash_to_dict(*arg7), 0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_plugin_invoke_get_dir(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_plugin_t *arg1 = 0 ;
  void *arg2 = 0 ;
  char *arg3 = 0 ;
  svn_revnum_t arg4 ;
  apr_hash_t **arg5 = 0 ;
  svn_revnum_t *arg6 = 0 ;
  apr_hash_t **arg7 = 0 ;
  apr_pool_t *arg8 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  apr_hash_t *temp5 ;
  svn_revnum_t temp6 ;
  int res6 = SWIG_TMPOBJ ;
  apr_hash_t *temp7 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg8 = _global_pool;
  arg5 = &temp5;
  arg6 = &temp6;
  arg7 = &temp7;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_plugin_invoke_get_dir", 4, 5, &obj0, &obj1, &obj2, &obj3, &obj4)) SWIG_fail;
  {
    arg1 = (svn_ra_plugin_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_plugin_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  {
    arg3 = svn_swig_py_string_to_cstring(obj2, FALSE, "svn_ra_plugin_invoke_get_dir", "path");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    arg4 = (svn_revnum_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  if (obj4) {
    /* Verify that the user supplied a valid pool */
    if (obj4 != Py_None && obj4 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj4);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_plugin_invoke_get_dir(arg1,arg2,(char const *)arg3,arg4,arg5,arg6,arg7,arg8);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_convert_hash(*arg5,
        SWIGTYPE_p_svn_dirent_t,
        _global_py_pool), 0)
    
    
    
    ;
  }
  if (SWIG_IsTmpObj(res6)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_long((*arg6)), 0);
  } else {
    int new_flags = SWIG_IsNewObj(res6) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg6), SWIGTYPE_p_long, new_flags), 0);
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_prophash_to_dict(*arg7), 0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_plugin_invoke_do_update(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_plugin_t *arg1 = 0 ;
  void *arg2 = 0 ;
  svn_ra_reporter_t **arg3 = 0 ;
  void **arg4 = 0 ;
  svn_revnum_t arg5 ;
  char *arg6 = 0 ;
  svn_boolean_t arg7 ;
  svn_delta_editor_t *arg8 = 0 ;
  void *arg9 = 0 ;
  apr_pool_t *arg10 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_ra_reporter_t *temp3 ;
  void *temp4 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg10 = _global_pool;
  arg3 = &temp3;
  arg4 = &temp4;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_plugin_invoke_do_update", 7, 8, &obj0, &obj1, &obj2, &obj3, &obj4, &obj5, &obj6, &obj7)) SWIG_fail;
  {
    arg1 = (svn_ra_plugin_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_plugin_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  {
    arg5 = (svn_revnum_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  {
    arg6 = svn_swig_py_string_to_cstring(obj3, FALSE, "svn_ra_plugin_invoke_do_update", "update_target");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    arg7 = (svn_boolean_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  {
    arg8 = (svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj5, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj5);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj6 == Py_None) {
      arg9 = NULL;
    } else if (SWIG_ConvertPtr(obj6, (void **) &arg9, 0, 0) == -1) {
      arg9 = (void *) obj6;
      PyErr_Clear();
    }
  }
  if (obj7) {
    /* Verify that the user supplied a valid pool */
    if (obj7 != Py_None && obj7 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj7);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_plugin_invoke_do_update(arg1,arg2,(struct svn_ra_reporter_t const **)arg3,arg4,arg5,(char const *)arg6,arg7,(struct svn_delta_editor_t const *)arg8,arg9,arg10);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg3, SWIGTYPE_p_svn_ra_reporter_t,
        _global_py_pool, args), 0)
    
    ;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg4, SWIGTYPE_p_void,
        _global_py_pool, args), 0)
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_plugin_invoke_do_switch(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_plugin_t *arg1 = 0 ;
  void *arg2 = 0 ;
  svn_ra_reporter_t **arg3 = 0 ;
  void **arg4 = 0 ;
  svn_revnum_t arg5 ;
  char *arg6 = 0 ;
  svn_boolean_t arg7 ;
  char *arg8 = 0 ;
  svn_delta_editor_t *arg9 = 0 ;
  void *arg10 = 0 ;
  apr_pool_t *arg11 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_ra_reporter_t *temp3 ;
  void *temp4 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg11 = _global_pool;
  arg3 = &temp3;
  arg4 = &temp4;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_plugin_invoke_do_switch", 8, 9, &obj0, &obj1, &obj2, &obj3, &obj4, &obj5, &obj6, &obj7, &obj8)) SWIG_fail;
  {
    arg1 = (svn_ra_plugin_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_plugin_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  {
    arg5 = (svn_revnum_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  {
    arg6 = svn_swig_py_string_to_cstring(obj3, FALSE, "svn_ra_plugin_invoke_do_switch", "switch_target");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    arg7 = (svn_boolean_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  {
    arg8 = svn_swig_py_string_to_cstring(obj5, FALSE, "svn_ra_plugin_invoke_do_switch", "switch_url");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    arg9 = (svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj6, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj6);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj7 == Py_None) {
      arg10 = NULL;
    } else if (SWIG_ConvertPtr(obj7, (void **) &arg10, 0, 0) == -1) {
      arg10 = (void *) obj7;
      PyErr_Clear();
    }
  }
  if (obj8) {
    /* Verify that the user supplied a valid pool */
    if (obj8 != Py_None && obj8 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj8);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_plugin_invoke_do_switch(arg1,arg2,(struct svn_ra_reporter_t const **)arg3,arg4,arg5,(char const *)arg6,arg7,(char const *)arg8,(struct svn_delta_editor_t const *)arg9,arg10,arg11);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg3, SWIGTYPE_p_svn_ra_reporter_t,
        _global_py_pool, args), 0)
    
    ;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg4, SWIGTYPE_p_void,
        _global_py_pool, args), 0)
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_plugin_invoke_do_status(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_plugin_t *arg1 = 0 ;
  void *arg2 = 0 ;
  svn_ra_reporter_t **arg3 = 0 ;
  void **arg4 = 0 ;
  char *arg5 = 0 ;
  svn_revnum_t arg6 ;
  svn_boolean_t arg7 ;
  svn_delta_editor_t *arg8 = 0 ;
  void *arg9 = 0 ;
  apr_pool_t *arg10 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_ra_reporter_t *temp3 ;
  void *temp4 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg10 = _global_pool;
  arg3 = &temp3;
  arg4 = &temp4;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_plugin_invoke_do_status", 7, 8, &obj0, &obj1, &obj2, &obj3, &obj4, &obj5, &obj6, &obj7)) SWIG_fail;
  {
    arg1 = (svn_ra_plugin_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_plugin_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  {
    arg5 = svn_swig_py_string_to_cstring(obj2, FALSE, "svn_ra_plugin_invoke_do_status", "status_target");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    arg6 = (svn_revnum_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg7 = (svn_boolean_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  {
    arg8 = (svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj5, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj5);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj6 == Py_None) {
      arg9 = NULL;
    } else if (SWIG_ConvertPtr(obj6, (void **) &arg9, 0, 0) == -1) {
      arg9 = (void *) obj6;
      PyErr_Clear();
    }
  }
  if (obj7) {
    /* Verify that the user supplied a valid pool */
    if (obj7 != Py_None && obj7 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj7);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_plugin_invoke_do_status(arg1,arg2,(struct svn_ra_reporter_t const **)arg3,arg4,(char const *)arg5,arg6,arg7,(struct svn_delta_editor_t const *)arg8,arg9,arg10);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg3, SWIGTYPE_p_svn_ra_reporter_t,
        _global_py_pool, args), 0)
    
    ;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg4, SWIGTYPE_p_void,
        _global_py_pool, args), 0)
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_plugin_invoke_do_diff(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_plugin_t *arg1 = 0 ;
  void *arg2 = 0 ;
  svn_ra_reporter_t **arg3 = 0 ;
  void **arg4 = 0 ;
  svn_revnum_t arg5 ;
  char *arg6 = 0 ;
  svn_boolean_t arg7 ;
  svn_boolean_t arg8 ;
  char *arg9 = 0 ;
  svn_delta_editor_t *arg10 = 0 ;
  void *arg11 = 0 ;
  apr_pool_t *arg12 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_ra_reporter_t *temp3 ;
  void *temp4 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg12 = _global_pool;
  arg3 = &temp3;
  arg4 = &temp4;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_plugin_invoke_do_diff", 9, 10, &obj0, &obj1, &obj2, &obj3, &obj4, &obj5, &obj6, &obj7, &obj8, &obj9)) SWIG_fail;
  {
    arg1 = (svn_ra_plugin_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_plugin_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  {
    arg5 = (svn_revnum_t)SWIG_As_long (obj2);
    if (SWIG_arg_fail(svn_argnum_obj2)) {
      SWIG_fail;
    }
  }
  {
    arg6 = svn_swig_py_string_to_cstring(obj3, FALSE, "svn_ra_plugin_invoke_do_diff", "diff_target");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    arg7 = (svn_boolean_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  {
    arg8 = (svn_boolean_t)SWIG_As_long (obj5);
    if (SWIG_arg_fail(svn_argnum_obj5)) {
      SWIG_fail;
    }
  }
  {
    arg9 = svn_swig_py_string_to_cstring(obj6, FALSE, "svn_ra_plugin_invoke_do_diff", "versus_url");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    arg10 = (svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj7, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj7);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj8 == Py_None) {
      arg11 = NULL;
    } else if (SWIG_ConvertPtr(obj8, (void **) &arg11, 0, 0) == -1) {
      arg11 = (void *) obj8;
      PyErr_Clear();
    }
  }
  if (obj9) {
    /* Verify that the user supplied a valid pool */
    if (obj9 != Py_None && obj9 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj9);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_plugin_invoke_do_diff(arg1,arg2,(struct svn_ra_reporter_t const **)arg3,arg4,arg5,(char const *)arg6,arg7,arg8,(char const *)arg9,(struct svn_delta_editor_t const *)arg10,arg11,arg12);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg3, SWIGTYPE_p_svn_ra_reporter_t,
        _global_py_pool, args), 0)
    
    ;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg4, SWIGTYPE_p_void,
        _global_py_pool, args), 0)
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_plugin_invoke_get_log(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_plugin_t *arg1 = 0 ;
  void *arg2 = 0 ;
  apr_array_header_t *arg3 = 0 ;
  svn_revnum_t arg4 ;
  svn_revnum_t arg5 ;
  svn_boolean_t arg6 ;
  svn_boolean_t arg7 ;
  svn_log_message_receiver_t arg8 = 0 ;
  void *arg9 = 0 ;
  apr_pool_t *arg10 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg10 = _global_pool;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_plugin_invoke_get_log", 8, 9, &obj0, &obj1, &obj2, &obj3, &obj4, &obj5, &obj6, &obj7, &obj8)) SWIG_fail;
  {
    arg1 = (svn_ra_plugin_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_plugin_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  {
    arg3 = (apr_array_header_t *) svn_swig_py_seq_to_array(obj2,
      sizeof(const char *),
      svn_swig_py_unwrap_string,
      NULL,
      _global_pool);
    if (PyErr_Occurred())
    SWIG_fail;
  }
  {
    arg4 = (svn_revnum_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg5 = (svn_revnum_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  {
    arg6 = (svn_boolean_t)SWIG_As_long (obj5);
    if (SWIG_arg_fail(svn_argnum_obj5)) {
      SWIG_fail;
    }
  }
  {
    arg7 = (svn_boolean_t)SWIG_As_long (obj6);
    if (SWIG_arg_fail(svn_argnum_obj6)) {
      SWIG_fail;
    }
  }
  {
    arg8 = (svn_log_message_receiver_t) svn_swig_py_log_receiver;
    arg9 = obj7;
  }
  if (obj8) {
    /* Verify that the user supplied a valid pool */
    if (obj8 != Py_None && obj8 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj8);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_plugin_invoke_get_log(arg1,arg2,(apr_array_header_t const *)arg3,arg4,arg5,arg6,arg7,arg8,arg9,arg10);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_plugin_invoke_check_path(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_plugin_t *arg1 = 0 ;
  void *arg2 = 0 ;
  char *arg3 = 0 ;
  svn_revnum_t arg4 ;
  svn_node_kind_t *arg5 = 0 ;
  apr_pool_t *arg6 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_node_kind_t temp5 ;
  int res5 = SWIG_TMPOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg6 = _global_pool;
  arg5 = &temp5;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_plugin_invoke_check_path", 4, 5, &obj0, &obj1, &obj2, &obj3, &obj4)) SWIG_fail;
  {
    arg1 = (svn_ra_plugin_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_plugin_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  {
    arg3 = svn_swig_py_string_to_cstring(obj2, FALSE, "svn_ra_plugin_invoke_check_path", "path");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    arg4 = (svn_revnum_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  if (obj4) {
    /* Verify that the user supplied a valid pool */
    if (obj4 != Py_None && obj4 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj4);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_plugin_invoke_check_path(arg1,arg2,(char const *)arg3,arg4,arg5,arg6);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  if (SWIG_IsTmpObj(res5)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_long((*arg5)), 0);
  } else {
    int new_flags = SWIG_IsNewObj(res5) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg5), SWIGTYPE_p_svn_node_kind_t, new_flags), 0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_plugin_invoke_get_uuid(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_plugin_t *arg1 = 0 ;
  void *arg2 = 0 ;
  char **arg3 = 0 ;
  apr_pool_t *arg4 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  char *temp3 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg4 = _global_pool;
  arg3 = &temp3;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_plugin_invoke_get_uuid", 2, 3, &obj0, &obj1, &obj2)) SWIG_fail;
  {
    arg1 = (svn_ra_plugin_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_plugin_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  if (obj2) {
    /* Verify that the user supplied a valid pool */
    if (obj2 != Py_None && obj2 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj2);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_plugin_invoke_get_uuid(arg1,arg2,(char const **)arg3,arg4);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    PyObject *s;
    if (*arg3 == NULL) {
      Py_INCREF(Py_None);
      s = Py_None;
    }
    else {
      s = PyBytes_FromString(*arg3);
      if (s == NULL)
      SWIG_fail;
    }
    resultobj = SWIG_Python_AppendOutput(resultobj, s, 0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_plugin_invoke_get_repos_root(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_plugin_t *arg1 = 0 ;
  void *arg2 = 0 ;
  char **arg3 = 0 ;
  apr_pool_t *arg4 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  char *temp3 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg4 = _global_pool;
  arg3 = &temp3;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_plugin_invoke_get_repos_root", 2, 3, &obj0, &obj1, &obj2)) SWIG_fail;
  {
    arg1 = (svn_ra_plugin_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_plugin_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  if (obj2) {
    /* Verify that the user supplied a valid pool */
    if (obj2 != Py_None && obj2 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj2);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_plugin_invoke_get_repos_root(arg1,arg2,(char const **)arg3,arg4);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    PyObject *s;
    if (*arg3 == NULL) {
      Py_INCREF(Py_None);
      s = Py_None;
    }
    else {
      s = PyBytes_FromString(*arg3);
      if (s == NULL)
      SWIG_fail;
    }
    resultobj = SWIG_Python_AppendOutput(resultobj, s, 0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_plugin_invoke_get_locations(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_plugin_t *arg1 = 0 ;
  void *arg2 = 0 ;
  apr_hash_t **arg3 = 0 ;
  char *arg4 = 0 ;
  svn_revnum_t arg5 ;
  apr_array_header_t *arg6 = 0 ;
  apr_pool_t *arg7 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  apr_hash_t *temp3 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg7 = _global_pool;
  arg3 = &temp3;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_plugin_invoke_get_locations", 5, 6, &obj0, &obj1, &obj2, &obj3, &obj4, &obj5)) SWIG_fail;
  {
    arg1 = (svn_ra_plugin_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_plugin_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  {
    arg4 = svn_swig_py_string_to_cstring(obj2, FALSE, "svn_ra_plugin_invoke_get_locations", "path");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    arg5 = (svn_revnum_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg6 = (apr_array_header_t *) svn_swig_py_seq_to_array(obj4,
      sizeof(svn_revnum_t),
      svn_swig_py_unwrap_revnum,
      NULL,
      _global_pool);
    if (PyErr_Occurred())
    SWIG_fail;
  }
  if (obj5) {
    /* Verify that the user supplied a valid pool */
    if (obj5 != Py_None && obj5 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj5);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_plugin_invoke_get_locations(arg1,arg2,arg3,(char const *)arg4,arg5,arg6,arg7);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_locationhash_to_dict(*arg3), 0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_plugin_invoke_get_file_revs(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_plugin_t *arg1 = 0 ;
  void *arg2 = 0 ;
  char *arg3 = 0 ;
  svn_revnum_t arg4 ;
  svn_revnum_t arg5 ;
  svn_ra_file_rev_handler_t arg6 = 0 ;
  void *arg7 = 0 ;
  apr_pool_t *arg8 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg8 = _global_pool;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_plugin_invoke_get_file_revs", 6, 7, &obj0, &obj1, &obj2, &obj3, &obj4, &obj5, &obj6)) SWIG_fail;
  {
    arg1 = (svn_ra_plugin_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_plugin_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  {
    arg3 = svn_swig_py_string_to_cstring(obj2, FALSE, "svn_ra_plugin_invoke_get_file_revs", "path");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    arg4 = (svn_revnum_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg5 = (svn_revnum_t)SWIG_As_long (obj4);
    if (SWIG_arg_fail(svn_argnum_obj4)) {
      SWIG_fail;
    }
  }
  {
    arg6 = (svn_ra_file_rev_handler_t) svn_swig_py_ra_file_rev_handler_func;
    arg7 = obj5;
  }
  if (obj6) {
    /* Verify that the user supplied a valid pool */
    if (obj6 != Py_None && obj6 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj6);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_plugin_invoke_get_file_revs(arg1,arg2,(char const *)arg3,arg4,arg5,arg6,arg7,arg8);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_plugin_invoke_get_version(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_plugin_t *arg1 = 0 ;
  PyObject * obj0 = 0 ;
  svn_version_t *result = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_plugin_invoke_get_version", 1, 1, &obj0)) SWIG_fail;
  {
    arg1 = (svn_ra_plugin_t *)svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_svn_ra_plugin_t, svn_argnum_obj0);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_version_t *)svn_ra_plugin_invoke_get_version(arg1);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = svn_swig_py_new_pointer_obj((void*)(result), SWIGTYPE_p_svn_version_t,
    _global_py_pool, args);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_invoke_get_wc_prop_func(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_get_wc_prop_func_t arg1 = 0 ;
  void *arg2 = 0 ;
  char *arg3 = 0 ;
  char *arg4 = 0 ;
  svn_string_t **arg5 = 0 ;
  apr_pool_t *arg6 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_string_t *temp5 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg6 = _global_pool;
  arg5 = &temp5;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_invoke_get_wc_prop_func", 4, 5, &obj0, &obj1, &obj2, &obj3, &obj4)) SWIG_fail;
  {
    svn_ra_get_wc_prop_func_t * tmp =
    svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_p_f_p_void_p_q_const__char_p_q_const__char_p_p_q_const__svn_string_t_p_apr_pool_t__p_svn_error_t, svn_argnum_obj0);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg1 = *tmp;
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  {
    arg3 = svn_swig_py_string_to_cstring(obj2, FALSE, "svn_ra_invoke_get_wc_prop_func", "path");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    arg4 = svn_swig_py_string_to_cstring(obj3, FALSE, "svn_ra_invoke_get_wc_prop_func", "name");
    if (PyErr_Occurred()) SWIG_fail;
  }
  if (obj4) {
    /* Verify that the user supplied a valid pool */
    if (obj4 != Py_None && obj4 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj4);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_invoke_get_wc_prop_func(arg1,arg2,(char const *)arg3,(char const *)arg4,(struct svn_string_t const **)arg5,arg6);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    PyObject *s;
    if (*arg5 == NULL) {
      Py_INCREF(Py_None);
      s = Py_None;
    } else {
      s = PyBytes_FromStringAndSize((*arg5)->data, (*arg5)->len);
      if (s == NULL)
      SWIG_fail;
    }
    resultobj = SWIG_Python_AppendOutput(resultobj, s, 0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_invoke_set_wc_prop_func(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_set_wc_prop_func_t arg1 = 0 ;
  void *arg2 = 0 ;
  char *arg3 = 0 ;
  char *arg4 = 0 ;
  svn_string_t *arg5 = 0 ;
  apr_pool_t *arg6 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_string_t value5 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg6 = _global_pool;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_invoke_set_wc_prop_func", 5, 6, &obj0, &obj1, &obj2, &obj3, &obj4, &obj5)) SWIG_fail;
  {
    svn_ra_set_wc_prop_func_t * tmp =
    svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_p_f_p_void_p_q_const__char_p_q_const__char_p_q_const__svn_string_t_p_apr_pool_t__p_svn_error_t, svn_argnum_obj0);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg1 = *tmp;
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  {
    arg3 = svn_swig_py_string_to_cstring(obj2, FALSE, "svn_ra_invoke_set_wc_prop_func", "path");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    arg4 = svn_swig_py_string_to_cstring(obj3, FALSE, "svn_ra_invoke_set_wc_prop_func", "name");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    if (obj4 == Py_None)
    arg5 = NULL;
    else {
      Py_ssize_t pyStrLen;
      if (!PyBytes_Check(obj4)) {
        PyErr_SetString(PyExc_TypeError, "not a bytes object");
        SWIG_fail;
      }
      if (PyBytes_AsStringAndSize(obj4, (char **)&(value5.data),
          &pyStrLen) == -1) {
        SWIG_fail;
      }
      value5.len = pyStrLen;
      arg5 = &value5;
    }
  }
  if (obj5) {
    /* Verify that the user supplied a valid pool */
    if (obj5 != Py_None && obj5 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj5);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_invoke_set_wc_prop_func(arg1,arg2,(char const *)arg3,(char const *)arg4,(struct svn_string_t const *)arg5,arg6);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_invoke_push_wc_prop_func(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_push_wc_prop_func_t arg1 = 0 ;
  void *arg2 = 0 ;
  char *arg3 = 0 ;
  char *arg4 = 0 ;
  svn_string_t *arg5 = 0 ;
  apr_pool_t *arg6 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_string_t value5 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg6 = _global_pool;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_invoke_push_wc_prop_func", 5, 6, &obj0, &obj1, &obj2, &obj3, &obj4, &obj5)) SWIG_fail;
  {
    svn_ra_push_wc_prop_func_t * tmp =
    svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_p_f_p_void_p_q_const__char_p_q_const__char_p_q_const__svn_string_t_p_apr_pool_t__p_svn_error_t, svn_argnum_obj0);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg1 = *tmp;
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  {
    arg3 = svn_swig_py_string_to_cstring(obj2, FALSE, "svn_ra_invoke_push_wc_prop_func", "path");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    arg4 = svn_swig_py_string_to_cstring(obj3, FALSE, "svn_ra_invoke_push_wc_prop_func", "name");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    if (obj4 == Py_None)
    arg5 = NULL;
    else {
      Py_ssize_t pyStrLen;
      if (!PyBytes_Check(obj4)) {
        PyErr_SetString(PyExc_TypeError, "not a bytes object");
        SWIG_fail;
      }
      if (PyBytes_AsStringAndSize(obj4, (char **)&(value5.data),
          &pyStrLen) == -1) {
        SWIG_fail;
      }
      value5.len = pyStrLen;
      arg5 = &value5;
    }
  }
  if (obj5) {
    /* Verify that the user supplied a valid pool */
    if (obj5 != Py_None && obj5 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj5);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_invoke_push_wc_prop_func(arg1,arg2,(char const *)arg3,(char const *)arg4,(struct svn_string_t const *)arg5,arg6);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_invoke_invalidate_wc_props_func(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_invalidate_wc_props_func_t arg1 = 0 ;
  void *arg2 = 0 ;
  char *arg3 = 0 ;
  char *arg4 = 0 ;
  apr_pool_t *arg5 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg5 = _global_pool;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_invoke_invalidate_wc_props_func", 4, 5, &obj0, &obj1, &obj2, &obj3, &obj4)) SWIG_fail;
  {
    svn_ra_invalidate_wc_props_func_t * tmp =
    svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_p_f_p_void_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t, svn_argnum_obj0);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg1 = *tmp;
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  {
    arg3 = svn_swig_py_string_to_cstring(obj2, FALSE, "svn_ra_invoke_invalidate_wc_props_func", "path");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    arg4 = svn_swig_py_string_to_cstring(obj3, FALSE, "svn_ra_invoke_invalidate_wc_props_func", "name");
    if (PyErr_Occurred()) SWIG_fail;
  }
  if (obj4) {
    /* Verify that the user supplied a valid pool */
    if (obj4 != Py_None && obj4 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj4);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_invoke_invalidate_wc_props_func(arg1,arg2,(char const *)arg3,(char const *)arg4,arg5);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_invoke_get_wc_contents_func(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_get_wc_contents_func_t arg1 = 0 ;
  void *arg2 = 0 ;
  svn_stream_t **arg3 = 0 ;
  svn_checksum_t *arg4 = 0 ;
  apr_pool_t *arg5 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_stream_t *temp3 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg5 = _global_pool;
  arg3 = &temp3;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_invoke_get_wc_contents_func", 3, 4, &obj0, &obj1, &obj2, &obj3)) SWIG_fail;
  {
    svn_ra_get_wc_contents_func_t * tmp =
    svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_p_f_p_void_p_p_svn_stream_t_p_q_const__svn_checksum_t_p_apr_pool_t__p_svn_error_t, svn_argnum_obj0);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg1 = *tmp;
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  {
    arg4 = (svn_checksum_t *)svn_swig_py_must_get_ptr(obj2, SWIGTYPE_p_svn_checksum_t, svn_argnum_obj2);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj3) {
    /* Verify that the user supplied a valid pool */
    if (obj3 != Py_None && obj3 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj3);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_invoke_get_wc_contents_func(arg1,arg2,arg3,(struct svn_checksum_t const *)arg4,arg5);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg3, SWIGTYPE_p_svn_stream_t,
        _global_py_pool, args), 0)
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_invoke_get_latest_revnum_func(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_get_latest_revnum_func_t arg1 = 0 ;
  void *arg2 = 0 ;
  svn_revnum_t *arg3 = 0 ;
  svn_revnum_t temp3 ;
  int res3 = SWIG_TMPOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  svn_error_t *result = 0 ;
  
  arg3 = &temp3;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_invoke_get_latest_revnum_func", 2, 2, &obj0, &obj1)) SWIG_fail;
  {
    svn_ra_get_latest_revnum_func_t * tmp =
    svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_p_f_p_void_p_svn_revnum_t__p_svn_error_t, svn_argnum_obj0);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg1 = *tmp;
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_invoke_get_latest_revnum_func(arg1,arg2,arg3);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  if (SWIG_IsTmpObj(res3)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_long((*arg3)), 0);
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_long, new_flags), 0);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_invoke_get_client_string_func(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_get_client_string_func_t arg1 = 0 ;
  void *arg2 = 0 ;
  char **arg3 = 0 ;
  apr_pool_t *arg4 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  char *temp3 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg4 = _global_pool;
  arg3 = &temp3;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_invoke_get_client_string_func", 2, 3, &obj0, &obj1, &obj2)) SWIG_fail;
  {
    svn_ra_get_client_string_func_t * tmp =
    svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_p_f_p_void_p_p_q_const__char_p_apr_pool_t__p_svn_error_t, svn_argnum_obj0);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg1 = *tmp;
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  if (obj2) {
    /* Verify that the user supplied a valid pool */
    if (obj2 != Py_None && obj2 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj2);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_invoke_get_client_string_func(arg1,arg2,(char const **)arg3,arg4);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    PyObject *s;
    if (*arg3 == NULL) {
      Py_INCREF(Py_None);
      s = Py_None;
    }
    else {
      s = PyBytes_FromString(*arg3);
      if (s == NULL)
      SWIG_fail;
    }
    resultobj = SWIG_Python_AppendOutput(resultobj, s, 0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_invoke_file_rev_handler(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_file_rev_handler_t arg1 = 0 ;
  void *arg2 = 0 ;
  char *arg3 = 0 ;
  svn_revnum_t arg4 ;
  apr_hash_t *arg5 = 0 ;
  svn_txdelta_window_handler_t *arg6 = 0 ;
  void **arg7 = 0 ;
  apr_array_header_t *arg8 = 0 ;
  apr_pool_t *arg9 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  void *temp7 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg9 = _global_pool;
  {
    if (_global_pool == NULL)
    {
      if (svn_swig_py_get_parent_pool(args, SWIGTYPE_p_apr_pool_t,
          &_global_py_pool, &_global_pool))
      SWIG_fail;
    }
    arg6 = (svn_txdelta_window_handler_t *) apr_pcalloc(_global_pool, sizeof(svn_txdelta_window_handler_t));
    if (arg6 == NULL) SWIG_fail;
  }
  arg7 = &temp7;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_invoke_file_rev_handler", 6, 7, &obj0, &obj1, &obj2, &obj3, &obj4, &obj5, &obj6)) SWIG_fail;
  {
    svn_ra_file_rev_handler_t * tmp =
    svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_p_f_p_void_p_q_const__char_svn_revnum_t_p_apr_hash_t_p_svn_txdelta_window_handler_t_p_p_void_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t, svn_argnum_obj0);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg1 = *tmp;
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  {
    arg3 = svn_swig_py_string_to_cstring(obj2, FALSE, "svn_ra_invoke_file_rev_handler", "path");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    arg4 = (svn_revnum_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg5 = (apr_hash_t *)svn_swig_py_must_get_ptr(obj4, SWIGTYPE_p_apr_hash_t, svn_argnum_obj4);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg8 = (apr_array_header_t *)svn_swig_py_must_get_ptr(obj5, SWIGTYPE_p_apr_array_header_t, svn_argnum_obj5);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj6) {
    /* Verify that the user supplied a valid pool */
    if (obj6 != Py_None && obj6 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj6);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_invoke_file_rev_handler(arg1,arg2,(char const *)arg3,arg4,arg5,arg6,arg7,arg8,arg9);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(arg6, SWIGTYPE_p_p_f_p_svn_txdelta_window_t_p_void__p_svn_error_t,
        _global_py_pool, args), 0)
    
    ;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg7, SWIGTYPE_p_void,
        _global_py_pool, args), 0)
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_invoke_lock_callback(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_lock_callback_t arg1 = 0 ;
  void *arg2 = 0 ;
  char *arg3 = 0 ;
  svn_boolean_t arg4 ;
  svn_lock_t *arg5 = 0 ;
  svn_error_t *arg6 = 0 ;
  apr_pool_t *arg7 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg7 = _global_pool;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_invoke_lock_callback", 6, 7, &obj0, &obj1, &obj2, &obj3, &obj4, &obj5, &obj6)) SWIG_fail;
  {
    svn_ra_lock_callback_t * tmp =
    svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_p_f_p_void_p_q_const__char_svn_boolean_t_p_q_const__svn_lock_t_p_svn_error_t_p_apr_pool_t__p_svn_error_t, svn_argnum_obj0);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg1 = *tmp;
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  {
    arg3 = svn_swig_py_string_to_cstring(obj2, FALSE, "svn_ra_invoke_lock_callback", "path");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    arg4 = (svn_boolean_t)SWIG_As_long (obj3);
    if (SWIG_arg_fail(svn_argnum_obj3)) {
      SWIG_fail;
    }
  }
  {
    arg5 = (svn_lock_t *)svn_swig_py_must_get_ptr(obj4, SWIGTYPE_p_svn_lock_t, svn_argnum_obj4);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    arg6 = (svn_error_t *)svn_swig_py_must_get_ptr(obj5, SWIGTYPE_p_svn_error_t, svn_argnum_obj5);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj6) {
    /* Verify that the user supplied a valid pool */
    if (obj6 != Py_None && obj6 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj6);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_invoke_lock_callback(arg1,arg2,(char const *)arg3,arg4,(struct svn_lock_t const *)arg5,arg6,arg7);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_invoke_progress_notify_func(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_progress_notify_func_t arg1 = 0 ;
  apr_off_t arg2 ;
  apr_off_t arg3 ;
  void *arg4 = 0 ;
  apr_pool_t *arg5 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg5 = _global_pool;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_invoke_progress_notify_func", 4, 5, &obj0, &obj1, &obj2, &obj3, &obj4)) SWIG_fail;
  {
    svn_ra_progress_notify_func_t * tmp =
    svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_p_f_apr_off_t_apr_off_t_p_void_p_apr_pool_t__void, svn_argnum_obj0);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg1 = *tmp;
  }
  arg2 = (apr_off_t) PyLong_AsLongLong(obj1); 
  arg3 = (apr_off_t) PyLong_AsLongLong(obj2); 
  {
    if (obj3 == Py_None) {
      arg4 = NULL;
    } else if (SWIG_ConvertPtr(obj3, (void **) &arg4, 0, 0) == -1) {
      arg4 = (void *) obj3;
      PyErr_Clear();
    }
  }
  if (obj4) {
    /* Verify that the user supplied a valid pool */
    if (obj4 != Py_None && obj4 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj4);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    svn_ra_invoke_progress_notify_func(arg1,arg2,arg3,arg4,arg5);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = SWIG_Py_Void();
  {
    Py_XDECREF(_global_py_pool);
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_invoke_replay_revstart_callback(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_replay_revstart_callback_t arg1 = 0 ;
  svn_revnum_t arg2 ;
  void *arg3 = 0 ;
  svn_delta_editor_t **arg4 = 0 ;
  void **arg5 = 0 ;
  apr_hash_t *arg6 = 0 ;
  apr_pool_t *arg7 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_delta_editor_t *temp4 ;
  void *temp5 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg7 = _global_pool;
  arg4 = &temp4;
  arg5 = &temp5;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_invoke_replay_revstart_callback", 4, 5, &obj0, &obj1, &obj2, &obj3, &obj4)) SWIG_fail;
  {
    svn_ra_replay_revstart_callback_t * tmp =
    svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_p_f_svn_revnum_t_p_void_p_p_q_const__svn_delta_editor_t_p_p_void_p_apr_hash_t_p_apr_pool_t__p_svn_error_t, svn_argnum_obj0);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg1 = *tmp;
  }
  {
    arg2 = (svn_revnum_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  {
    if (obj2 == Py_None) {
      arg3 = NULL;
    } else if (SWIG_ConvertPtr(obj2, (void **) &arg3, 0, 0) == -1) {
      arg3 = (void *) obj2;
      PyErr_Clear();
    }
  }
  {
    arg6 = (apr_hash_t *)svn_swig_py_must_get_ptr(obj3, SWIGTYPE_p_apr_hash_t, svn_argnum_obj3);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj4) {
    /* Verify that the user supplied a valid pool */
    if (obj4 != Py_None && obj4 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj4);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_invoke_replay_revstart_callback(arg1,arg2,arg3,(struct svn_delta_editor_t const **)arg4,arg5,arg6,arg7);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg4, SWIGTYPE_p_svn_delta_editor_t,
        _global_py_pool, args), 0)
    
    ;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg5, SWIGTYPE_p_void,
        _global_py_pool, args), 0)
    
    ;
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_invoke_replay_revfinish_callback(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_replay_revfinish_callback_t arg1 = 0 ;
  svn_revnum_t arg2 ;
  void *arg3 = 0 ;
  svn_delta_editor_t *arg4 = 0 ;
  void *arg5 = 0 ;
  apr_hash_t *arg6 = 0 ;
  apr_pool_t *arg7 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg7 = _global_pool;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_invoke_replay_revfinish_callback", 6, 7, &obj0, &obj1, &obj2, &obj3, &obj4, &obj5, &obj6)) SWIG_fail;
  {
    svn_ra_replay_revfinish_callback_t * tmp =
    svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_p_f_svn_revnum_t_p_void_p_q_const__svn_delta_editor_t_p_void_p_apr_hash_t_p_apr_pool_t__p_svn_error_t, svn_argnum_obj0);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg1 = *tmp;
  }
  {
    arg2 = (svn_revnum_t)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  {
    if (obj2 == Py_None) {
      arg3 = NULL;
    } else if (SWIG_ConvertPtr(obj2, (void **) &arg3, 0, 0) == -1) {
      arg3 = (void *) obj2;
      PyErr_Clear();
    }
  }
  {
    arg4 = (svn_delta_editor_t *)svn_swig_py_must_get_ptr(obj3, SWIGTYPE_p_svn_delta_editor_t, svn_argnum_obj3);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  {
    if (obj4 == Py_None) {
      arg5 = NULL;
    } else if (SWIG_ConvertPtr(obj4, (void **) &arg5, 0, 0) == -1) {
      arg5 = (void *) obj4;
      PyErr_Clear();
    }
  }
  {
    arg6 = (apr_hash_t *)svn_swig_py_must_get_ptr(obj5, SWIGTYPE_p_apr_hash_t, svn_argnum_obj5);
    if (PyErr_Occurred()) {
      SWIG_fail;
    }
  }
  if (obj6) {
    /* Verify that the user supplied a valid pool */
    if (obj6 != Py_None && obj6 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj6);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_invoke_replay_revfinish_callback(arg1,arg2,arg3,(struct svn_delta_editor_t const *)arg4,arg5,arg6,arg7);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_invoke_check_tunnel_func(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_check_tunnel_func_t arg1 = 0 ;
  void *arg2 = 0 ;
  char *arg3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  svn_boolean_t result;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_invoke_check_tunnel_func", 3, 3, &obj0, &obj1, &obj2)) SWIG_fail;
  {
    svn_ra_check_tunnel_func_t * tmp =
    svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_p_f_p_void_p_q_const__char__int, svn_argnum_obj0);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg1 = *tmp;
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  {
    arg3 = svn_swig_py_string_to_cstring(obj2, FALSE, "svn_ra_invoke_check_tunnel_func", "tunnel_name");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_boolean_t)svn_ra_invoke_check_tunnel_func(arg1,arg2,(char const *)arg3);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = SWIG_From_long((long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_invoke_close_tunnel_func(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_close_tunnel_func_t arg1 = 0 ;
  void *arg2 = 0 ;
  void *arg3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_invoke_close_tunnel_func", 3, 3, &obj0, &obj1, &obj2)) SWIG_fail;
  {
    svn_ra_close_tunnel_func_t * tmp =
    svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_p_f_p_void_p_void__void, svn_argnum_obj0);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg1 = *tmp;
  }
  {
    if (obj1 == Py_None) {
      arg2 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg2, 0, 0) == -1) {
      arg2 = (void *) obj1;
      PyErr_Clear();
    }
  }
  {
    if (obj2 == Py_None) {
      arg3 = NULL;
    } else if (SWIG_ConvertPtr(obj2, (void **) &arg3, 0, 0) == -1) {
      arg3 = (void *) obj2;
      PyErr_Clear();
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    svn_ra_invoke_close_tunnel_func(arg1,arg2,arg3);
    
    svn_swig_py_acquire_py_lock();
    
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_invoke_open_tunnel_func(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_open_tunnel_func_t arg1 = 0 ;
  svn_stream_t **arg2 = 0 ;
  svn_stream_t **arg3 = 0 ;
  svn_ra_close_tunnel_func_t *arg4 = 0 ;
  void **arg5 = 0 ;
  void *arg6 = 0 ;
  char *arg7 = 0 ;
  char *arg8 = 0 ;
  char *arg9 = 0 ;
  int arg10 ;
  svn_cancel_func_t arg11 = 0 ;
  void *arg12 = 0 ;
  apr_pool_t *arg13 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  svn_stream_t *temp2 ;
  svn_stream_t *temp3 ;
  void *temp5 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg13 = _global_pool;
  arg2 = &temp2;
  arg3 = &temp3;
  {
    if (_global_pool == NULL)
    {
      if (svn_swig_py_get_parent_pool(args, SWIGTYPE_p_apr_pool_t,
          &_global_py_pool, &_global_pool))
      SWIG_fail;
    }
    arg4 = (svn_ra_close_tunnel_func_t *) apr_pcalloc(_global_pool, sizeof(svn_ra_close_tunnel_func_t));
    if (arg4 == NULL) SWIG_fail;
  }
  arg5 = &temp5;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_invoke_open_tunnel_func", 7, 8, &obj0, &obj1, &obj2, &obj3, &obj4, &obj5, &obj6, &obj7)) SWIG_fail;
  {
    svn_ra_open_tunnel_func_t * tmp =
    svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_p_f_p_p_svn_stream_t_p_p_svn_stream_t_p_svn_ra_close_tunnel_func_t_p_p_void_p_void_p_q_const__char_p_q_const__char_p_q_const__char_int_svn_cancel_func_t_p_void_p_apr_pool_t__p_svn_error_t, svn_argnum_obj0);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg1 = *tmp;
  }
  {
    if (obj1 == Py_None) {
      arg6 = NULL;
    } else if (SWIG_ConvertPtr(obj1, (void **) &arg6, 0, 0) == -1) {
      arg6 = (void *) obj1;
      PyErr_Clear();
    }
  }
  {
    arg7 = svn_swig_py_string_to_cstring(obj2, FALSE, "svn_ra_invoke_open_tunnel_func", "tunnel_name");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    arg8 = svn_swig_py_string_to_cstring(obj3, FALSE, "svn_ra_invoke_open_tunnel_func", "user");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    arg9 = svn_swig_py_string_to_cstring(obj4, FALSE, "svn_ra_invoke_open_tunnel_func", "hostname");
    if (PyErr_Occurred()) SWIG_fail;
  }
  {
    arg10 = (int)SWIG_As_long (obj5);
    if (SWIG_arg_fail(svn_argnum_obj5)) {
      SWIG_fail;
    }
  }
  {
    arg11 = (svn_cancel_func_t) svn_swig_py_cancel_func;
    arg12 = obj6;
  }
  if (obj7) {
    /* Verify that the user supplied a valid pool */
    if (obj7 != Py_None && obj7 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj7);
      SWIG_fail;
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_invoke_open_tunnel_func(arg1,arg2,arg3,arg4,arg5,arg6,(char const *)arg7,(char const *)arg8,(char const *)arg9,arg10,arg11,arg12,arg13);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg2, SWIGTYPE_p_svn_stream_t,
        _global_py_pool, args), 0)
    
    ;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(*arg3, SWIGTYPE_p_svn_stream_t,
        _global_py_pool, args), 0)
    
    ;
  }
  {
    resultobj = SWIG_Python_AppendOutput(resultobj, svn_swig_py_new_pointer_obj(arg4, SWIGTYPE_p_p_f_p_void_p_void__void,
        _global_py_pool, args), 0)
    
    ;
  }
  {
    /* FIXME: Missing argout typemap: svn_ra_invoke_open_tunnel_func arg 5 (void **) */
    
    
    
    
    SWIG_exception(SWIG_ValueError, "svn_ra_invoke_open_tunnel_func is not implemented yet");
    
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_svn_ra_invoke_init_func(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  svn_ra_init_func_t arg1 = 0 ;
  int arg2 ;
  apr_pool_t *arg3 = 0 ;
  apr_hash_t *arg4 = 0 ;
  apr_pool_t *_global_pool = NULL ;
  PyObject *_global_py_pool = NULL ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  svn_error_t *result = 0 ;
  
  if (svn_swig_py_get_pool_arg(args, SWIGTYPE_p_apr_pool_t,
      &_global_py_pool, &_global_pool))
  SWIG_fail;
  arg3 = _global_pool;
  (void)self;
  if (!PyArg_UnpackTuple(args, "svn_ra_invoke_init_func", 2, 4, &obj0, &obj1, &obj2, &obj3)) SWIG_fail;
  {
    svn_ra_init_func_t * tmp =
    svn_swig_py_must_get_ptr(obj0, SWIGTYPE_p_p_f_int_p_apr_pool_t_p_apr_hash_t__p_svn_error_t, svn_argnum_obj0);
    if (tmp == NULL || PyErr_Occurred()) {
      SWIG_fail;
    }
    arg1 = *tmp;
  }
  {
    arg2 = (int)SWIG_As_long (obj1);
    if (SWIG_arg_fail(svn_argnum_obj1)) {
      SWIG_fail;
    }
  }
  if (obj2) {
    /* Verify that the user supplied a valid pool */
    if (obj2 != Py_None && obj2 != _global_py_pool) {
      SWIG_Error(SWIG_TypeError,  "a 'apr_pool_t *' is expected");
      SWIG_arg_fail(svn_argnum_obj2);
      SWIG_fail;
    }
  }
  if (obj3) {
    {
      if (_global_pool == NULL)
      {
        if (svn_swig_py_get_parent_pool(args, SWIGTYPE_p_apr_pool_t,
            &_global_py_pool, &_global_pool))
        SWIG_fail;
      }
      
      arg4 = svn_swig_py_prophash_from_dict(obj3, _global_pool);
      if (PyErr_Occurred()) {
        SWIG_fail;
      }
    }
  }
  {
    svn_swig_py_release_py_lock();
    
    result = (svn_error_t *)svn_ra_invoke_init_func(arg1,arg2,arg3,arg4);
    
    svn_swig_py_acquire_py_lock();
    
  }
  {
    if (result != NULL) {
      if (result->apr_err != SVN_ERR_SWIG_PY_EXCEPTION_SET)
      svn_swig_py_svn_exception(result);
      else
      svn_error_clear(result);
      SWIG_fail;
    }
    Py_XDECREF(resultobj);
    resultobj = PyList_New(0);
  }
  {
    Py_XDECREF(_global_py_pool);
  }
  {
    if (resultobj == NULL) {
      resultobj = Py_None;
      Py_INCREF(resultobj);
    }
    else {
      switch (PyList_Size(resultobj)) {
      case 0:
        resultobj = Py_None;
        Py_INCREF(resultobj);
        break;
      case 1:
        {
          PyObject *tmp = resultobj;
          resultobj = PyList_GetItem(tmp, 0);
          Py_INCREF(resultobj);
          Py_DECREF(tmp);
        }
        break;
      }
    }
  }
  return resultobj;
fail:
  {
    Py_XDECREF(_global_py_pool);
  }
  return NULL;
}


SWIGINTERN PyObject *svn_ra_get_wc_prop_func_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_p_f_p_void_p_q_const__char_p_q_const__char_p_p_q_const__svn_string_t_p_apr_pool_t__p_svn_error_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *svn_ra_set_wc_prop_func_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_p_f_p_void_p_q_const__char_p_q_const__char_p_q_const__svn_string_t_p_apr_pool_t__p_svn_error_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *svn_ra_push_wc_prop_func_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_p_f_p_void_p_q_const__char_p_q_const__char_p_q_const__svn_string_t_p_apr_pool_t__p_svn_error_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *svn_ra_invalidate_wc_props_func_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_p_f_p_void_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *svn_ra_get_wc_contents_func_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_p_f_p_void_p_p_svn_stream_t_p_q_const__svn_checksum_t_p_apr_pool_t__p_svn_error_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *svn_ra_get_latest_revnum_func_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_p_f_p_void_p_svn_revnum_t__p_svn_error_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *svn_ra_get_client_string_func_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_p_f_p_void_p_p_q_const__char_p_apr_pool_t__p_svn_error_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *svn_ra_file_rev_handler_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_p_f_p_void_p_q_const__char_svn_revnum_t_p_apr_hash_t_p_svn_txdelta_window_handler_t_p_p_void_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *svn_ra_lock_callback_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_p_f_p_void_p_q_const__char_svn_boolean_t_p_q_const__svn_lock_t_p_svn_error_t_p_apr_pool_t__p_svn_error_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *svn_ra_progress_notify_func_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_p_f_apr_off_t_apr_off_t_p_void_p_apr_pool_t__void, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *svn_ra_replay_revstart_callback_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_p_f_svn_revnum_t_p_void_p_p_q_const__svn_delta_editor_t_p_p_void_p_apr_hash_t_p_apr_pool_t__p_svn_error_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *svn_ra_replay_revfinish_callback_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_p_f_svn_revnum_t_p_void_p_q_const__svn_delta_editor_t_p_void_p_apr_hash_t_p_apr_pool_t__p_svn_error_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *svn_ra_check_tunnel_func_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_p_f_p_void_p_q_const__char__int, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *svn_ra_close_tunnel_func_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_p_f_p_void_p_void__void, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *svn_ra_open_tunnel_func_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_p_f_p_p_svn_stream_t_p_p_svn_stream_t_p_svn_ra_close_tunnel_func_t_p_p_void_p_void_p_q_const__char_p_q_const__char_p_q_const__char_int_svn_cancel_func_t_p_void_p_apr_pool_t__p_svn_error_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *svn_ra_init_func_t_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_p_f_int_p_apr_pool_t_p_apr_hash_t__p_svn_error_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

static PyMethodDef SwigMethods[] = {
	 { "svn_ra_version", _wrap_svn_ra_version, METH_VARARGS, "svn_ra_version() -> svn_version_t"},
	 { "svn_ra_reporter3_t_set_path_set", _wrap_svn_ra_reporter3_t_set_path_set, METH_VARARGS, "svn_ra_reporter3_t_set_path_set(svn_ra_reporter3_t self, svn_error_t *(*)(void *,char const *,svn_revnum_t,svn_depth_t,svn_boolean_t,char const *,apr_pool_t *) set_path)"},
	 { "svn_ra_reporter3_t_set_path_get", _wrap_svn_ra_reporter3_t_set_path_get, METH_VARARGS, "svn_ra_reporter3_t_set_path_get(svn_ra_reporter3_t self) -> svn_error_t *(*)(void *,char const *,svn_revnum_t,svn_depth_t,svn_boolean_t,char const *,apr_pool_t *)"},
	 { "svn_ra_reporter3_t_delete_path_set", _wrap_svn_ra_reporter3_t_delete_path_set, METH_VARARGS, "svn_ra_reporter3_t_delete_path_set(svn_ra_reporter3_t self, svn_error_t *(*)(void *,char const *,apr_pool_t *) delete_path)"},
	 { "svn_ra_reporter3_t_delete_path_get", _wrap_svn_ra_reporter3_t_delete_path_get, METH_VARARGS, "svn_ra_reporter3_t_delete_path_get(svn_ra_reporter3_t self) -> svn_error_t *(*)(void *,char const *,apr_pool_t *)"},
	 { "svn_ra_reporter3_t_link_path_set", _wrap_svn_ra_reporter3_t_link_path_set, METH_VARARGS, "svn_ra_reporter3_t_link_path_set(svn_ra_reporter3_t self, svn_error_t *(*)(void *,char const *,char const *,svn_revnum_t,svn_depth_t,svn_boolean_t,char const *,apr_pool_t *) link_path)"},
	 { "svn_ra_reporter3_t_link_path_get", _wrap_svn_ra_reporter3_t_link_path_get, METH_VARARGS, "svn_ra_reporter3_t_link_path_get(svn_ra_reporter3_t self) -> svn_error_t *(*)(void *,char const *,char const *,svn_revnum_t,svn_depth_t,svn_boolean_t,char const *,apr_pool_t *)"},
	 { "svn_ra_reporter3_t_finish_report_set", _wrap_svn_ra_reporter3_t_finish_report_set, METH_VARARGS, "svn_ra_reporter3_t_finish_report_set(svn_ra_reporter3_t self, svn_error_t *(*)(void *,apr_pool_t *) finish_report)"},
	 { "svn_ra_reporter3_t_finish_report_get", _wrap_svn_ra_reporter3_t_finish_report_get, METH_VARARGS, "svn_ra_reporter3_t_finish_report_get(svn_ra_reporter3_t self) -> svn_error_t *(*)(void *,apr_pool_t *)"},
	 { "svn_ra_reporter3_t_abort_report_set", _wrap_svn_ra_reporter3_t_abort_report_set, METH_VARARGS, "svn_ra_reporter3_t_abort_report_set(svn_ra_reporter3_t self, svn_error_t *(*)(void *,apr_pool_t *) abort_report)"},
	 { "svn_ra_reporter3_t_abort_report_get", _wrap_svn_ra_reporter3_t_abort_report_get, METH_VARARGS, "svn_ra_reporter3_t_abort_report_get(svn_ra_reporter3_t self) -> svn_error_t *(*)(void *,apr_pool_t *)"},
	 { "new_svn_ra_reporter3_t", _wrap_new_svn_ra_reporter3_t, METH_VARARGS, "new_svn_ra_reporter3_t() -> svn_ra_reporter3_t"},
	 { "delete_svn_ra_reporter3_t", _wrap_delete_svn_ra_reporter3_t, METH_VARARGS, "delete_svn_ra_reporter3_t(svn_ra_reporter3_t self)"},
	 { "svn_ra_reporter3_t_swigregister", svn_ra_reporter3_t_swigregister, METH_O, NULL},
	 { "svn_ra_reporter3_t_swiginit", svn_ra_reporter3_t_swiginit, METH_VARARGS, NULL},
	 { "svn_ra_reporter2_t_set_path_set", _wrap_svn_ra_reporter2_t_set_path_set, METH_VARARGS, "svn_ra_reporter2_t_set_path_set(svn_ra_reporter2_t self, svn_error_t *(*)(void *,char const *,svn_revnum_t,svn_boolean_t,char const *,apr_pool_t *) set_path)"},
	 { "svn_ra_reporter2_t_set_path_get", _wrap_svn_ra_reporter2_t_set_path_get, METH_VARARGS, "svn_ra_reporter2_t_set_path_get(svn_ra_reporter2_t self) -> svn_error_t *(*)(void *,char const *,svn_revnum_t,svn_boolean_t,char const *,apr_pool_t *)"},
	 { "svn_ra_reporter2_t_delete_path_set", _wrap_svn_ra_reporter2_t_delete_path_set, METH_VARARGS, "svn_ra_reporter2_t_delete_path_set(svn_ra_reporter2_t self, svn_error_t *(*)(void *,char const *,apr_pool_t *) delete_path)"},
	 { "svn_ra_reporter2_t_delete_path_get", _wrap_svn_ra_reporter2_t_delete_path_get, METH_VARARGS, "svn_ra_reporter2_t_delete_path_get(svn_ra_reporter2_t self) -> svn_error_t *(*)(void *,char const *,apr_pool_t *)"},
	 { "svn_ra_reporter2_t_link_path_set", _wrap_svn_ra_reporter2_t_link_path_set, METH_VARARGS, "svn_ra_reporter2_t_link_path_set(svn_ra_reporter2_t self, svn_error_t *(*)(void *,char const *,char const *,svn_revnum_t,svn_boolean_t,char const *,apr_pool_t *) link_path)"},
	 { "svn_ra_reporter2_t_link_path_get", _wrap_svn_ra_reporter2_t_link_path_get, METH_VARARGS, "svn_ra_reporter2_t_link_path_get(svn_ra_reporter2_t self) -> svn_error_t *(*)(void *,char const *,char const *,svn_revnum_t,svn_boolean_t,char const *,apr_pool_t *)"},
	 { "svn_ra_reporter2_t_finish_report_set", _wrap_svn_ra_reporter2_t_finish_report_set, METH_VARARGS, "svn_ra_reporter2_t_finish_report_set(svn_ra_reporter2_t self, svn_error_t *(*)(void *,apr_pool_t *) finish_report)"},
	 { "svn_ra_reporter2_t_finish_report_get", _wrap_svn_ra_reporter2_t_finish_report_get, METH_VARARGS, "svn_ra_reporter2_t_finish_report_get(svn_ra_reporter2_t self) -> svn_error_t *(*)(void *,apr_pool_t *)"},
	 { "svn_ra_reporter2_t_abort_report_set", _wrap_svn_ra_reporter2_t_abort_report_set, METH_VARARGS, "svn_ra_reporter2_t_abort_report_set(svn_ra_reporter2_t self, svn_error_t *(*)(void *,apr_pool_t *) abort_report)"},
	 { "svn_ra_reporter2_t_abort_report_get", _wrap_svn_ra_reporter2_t_abort_report_get, METH_VARARGS, "svn_ra_reporter2_t_abort_report_get(svn_ra_reporter2_t self) -> svn_error_t *(*)(void *,apr_pool_t *)"},
	 { "new_svn_ra_reporter2_t", _wrap_new_svn_ra_reporter2_t, METH_VARARGS, "new_svn_ra_reporter2_t() -> svn_ra_reporter2_t"},
	 { "delete_svn_ra_reporter2_t", _wrap_delete_svn_ra_reporter2_t, METH_VARARGS, "delete_svn_ra_reporter2_t(svn_ra_reporter2_t self)"},
	 { "svn_ra_reporter2_t_swigregister", svn_ra_reporter2_t_swigregister, METH_O, NULL},
	 { "svn_ra_reporter2_t_swiginit", svn_ra_reporter2_t_swiginit, METH_VARARGS, NULL},
	 { "svn_ra_reporter_t_set_path_set", _wrap_svn_ra_reporter_t_set_path_set, METH_VARARGS, "svn_ra_reporter_t_set_path_set(svn_ra_reporter_t self, svn_error_t *(*)(void *,char const *,svn_revnum_t,svn_boolean_t,apr_pool_t *) set_path)"},
	 { "svn_ra_reporter_t_set_path_get", _wrap_svn_ra_reporter_t_set_path_get, METH_VARARGS, "svn_ra_reporter_t_set_path_get(svn_ra_reporter_t self) -> svn_error_t *(*)(void *,char const *,svn_revnum_t,svn_boolean_t,apr_pool_t *)"},
	 { "svn_ra_reporter_t_delete_path_set", _wrap_svn_ra_reporter_t_delete_path_set, METH_VARARGS, "svn_ra_reporter_t_delete_path_set(svn_ra_reporter_t self, svn_error_t *(*)(void *,char const *,apr_pool_t *) delete_path)"},
	 { "svn_ra_reporter_t_delete_path_get", _wrap_svn_ra_reporter_t_delete_path_get, METH_VARARGS, "svn_ra_reporter_t_delete_path_get(svn_ra_reporter_t self) -> svn_error_t *(*)(void *,char const *,apr_pool_t *)"},
	 { "svn_ra_reporter_t_link_path_set", _wrap_svn_ra_reporter_t_link_path_set, METH_VARARGS, "svn_ra_reporter_t_link_path_set(svn_ra_reporter_t self, svn_error_t *(*)(void *,char const *,char const *,svn_revnum_t,svn_boolean_t,apr_pool_t *) link_path)"},
	 { "svn_ra_reporter_t_link_path_get", _wrap_svn_ra_reporter_t_link_path_get, METH_VARARGS, "svn_ra_reporter_t_link_path_get(svn_ra_reporter_t self) -> svn_error_t *(*)(void *,char const *,char const *,svn_revnum_t,svn_boolean_t,apr_pool_t *)"},
	 { "svn_ra_reporter_t_finish_report_set", _wrap_svn_ra_reporter_t_finish_report_set, METH_VARARGS, "svn_ra_reporter_t_finish_report_set(svn_ra_reporter_t self, svn_error_t *(*)(void *,apr_pool_t *) finish_report)"},
	 { "svn_ra_reporter_t_finish_report_get", _wrap_svn_ra_reporter_t_finish_report_get, METH_VARARGS, "svn_ra_reporter_t_finish_report_get(svn_ra_reporter_t self) -> svn_error_t *(*)(void *,apr_pool_t *)"},
	 { "svn_ra_reporter_t_abort_report_set", _wrap_svn_ra_reporter_t_abort_report_set, METH_VARARGS, "svn_ra_reporter_t_abort_report_set(svn_ra_reporter_t self, svn_error_t *(*)(void *,apr_pool_t *) abort_report)"},
	 { "svn_ra_reporter_t_abort_report_get", _wrap_svn_ra_reporter_t_abort_report_get, METH_VARARGS, "svn_ra_reporter_t_abort_report_get(svn_ra_reporter_t self) -> svn_error_t *(*)(void *,apr_pool_t *)"},
	 { "new_svn_ra_reporter_t", _wrap_new_svn_ra_reporter_t, METH_VARARGS, "new_svn_ra_reporter_t() -> svn_ra_reporter_t"},
	 { "delete_svn_ra_reporter_t", _wrap_delete_svn_ra_reporter_t, METH_VARARGS, "delete_svn_ra_reporter_t(svn_ra_reporter_t self)"},
	 { "svn_ra_reporter_t_swigregister", svn_ra_reporter_t_swigregister, METH_O, NULL},
	 { "svn_ra_reporter_t_swiginit", svn_ra_reporter_t_swiginit, METH_VARARGS, NULL},
	 { "svn_ra_callbacks2_t_open_tmp_file_set", _wrap_svn_ra_callbacks2_t_open_tmp_file_set, METH_VARARGS, "svn_ra_callbacks2_t_open_tmp_file_set(svn_ra_callbacks2_t self, svn_error_t *(*)(apr_file_t **,void *,apr_pool_t *) open_tmp_file)"},
	 { "svn_ra_callbacks2_t_open_tmp_file_get", _wrap_svn_ra_callbacks2_t_open_tmp_file_get, METH_VARARGS, "svn_ra_callbacks2_t_open_tmp_file_get(svn_ra_callbacks2_t self) -> svn_error_t *(*)(apr_file_t **,void *,apr_pool_t *)"},
	 { "svn_ra_callbacks2_t_auth_baton_set", _wrap_svn_ra_callbacks2_t_auth_baton_set, METH_VARARGS, "svn_ra_callbacks2_t_auth_baton_set(svn_ra_callbacks2_t self, svn_auth_baton_t auth_baton)"},
	 { "svn_ra_callbacks2_t_auth_baton_get", _wrap_svn_ra_callbacks2_t_auth_baton_get, METH_VARARGS, "svn_ra_callbacks2_t_auth_baton_get(svn_ra_callbacks2_t self) -> svn_auth_baton_t"},
	 { "svn_ra_callbacks2_t_get_wc_prop_set", _wrap_svn_ra_callbacks2_t_get_wc_prop_set, METH_VARARGS, "svn_ra_callbacks2_t_get_wc_prop_set(svn_ra_callbacks2_t self, svn_ra_get_wc_prop_func_t get_wc_prop)"},
	 { "svn_ra_callbacks2_t_get_wc_prop_get", _wrap_svn_ra_callbacks2_t_get_wc_prop_get, METH_VARARGS, "svn_ra_callbacks2_t_get_wc_prop_get(svn_ra_callbacks2_t self) -> svn_ra_get_wc_prop_func_t"},
	 { "svn_ra_callbacks2_t_set_wc_prop_set", _wrap_svn_ra_callbacks2_t_set_wc_prop_set, METH_VARARGS, "svn_ra_callbacks2_t_set_wc_prop_set(svn_ra_callbacks2_t self, svn_ra_set_wc_prop_func_t set_wc_prop)"},
	 { "svn_ra_callbacks2_t_set_wc_prop_get", _wrap_svn_ra_callbacks2_t_set_wc_prop_get, METH_VARARGS, "svn_ra_callbacks2_t_set_wc_prop_get(svn_ra_callbacks2_t self) -> svn_ra_set_wc_prop_func_t"},
	 { "svn_ra_callbacks2_t_push_wc_prop_set", _wrap_svn_ra_callbacks2_t_push_wc_prop_set, METH_VARARGS, "svn_ra_callbacks2_t_push_wc_prop_set(svn_ra_callbacks2_t self, svn_ra_push_wc_prop_func_t push_wc_prop)"},
	 { "svn_ra_callbacks2_t_push_wc_prop_get", _wrap_svn_ra_callbacks2_t_push_wc_prop_get, METH_VARARGS, "svn_ra_callbacks2_t_push_wc_prop_get(svn_ra_callbacks2_t self) -> svn_ra_push_wc_prop_func_t"},
	 { "svn_ra_callbacks2_t_invalidate_wc_props_set", _wrap_svn_ra_callbacks2_t_invalidate_wc_props_set, METH_VARARGS, "svn_ra_callbacks2_t_invalidate_wc_props_set(svn_ra_callbacks2_t self, svn_ra_invalidate_wc_props_func_t invalidate_wc_props)"},
	 { "svn_ra_callbacks2_t_invalidate_wc_props_get", _wrap_svn_ra_callbacks2_t_invalidate_wc_props_get, METH_VARARGS, "svn_ra_callbacks2_t_invalidate_wc_props_get(svn_ra_callbacks2_t self) -> svn_ra_invalidate_wc_props_func_t"},
	 { "svn_ra_callbacks2_t_progress_func_set", _wrap_svn_ra_callbacks2_t_progress_func_set, METH_VARARGS, "svn_ra_callbacks2_t_progress_func_set(svn_ra_callbacks2_t self, svn_ra_progress_notify_func_t progress_func)"},
	 { "svn_ra_callbacks2_t_progress_func_get", _wrap_svn_ra_callbacks2_t_progress_func_get, METH_VARARGS, "svn_ra_callbacks2_t_progress_func_get(svn_ra_callbacks2_t self) -> svn_ra_progress_notify_func_t"},
	 { "svn_ra_callbacks2_t_progress_baton_set", _wrap_svn_ra_callbacks2_t_progress_baton_set, METH_VARARGS, "svn_ra_callbacks2_t_progress_baton_set(svn_ra_callbacks2_t self, void * progress_baton)"},
	 { "svn_ra_callbacks2_t_progress_baton_get", _wrap_svn_ra_callbacks2_t_progress_baton_get, METH_VARARGS, "svn_ra_callbacks2_t_progress_baton_get(svn_ra_callbacks2_t self) -> void *"},
	 { "svn_ra_callbacks2_t_cancel_func_set", _wrap_svn_ra_callbacks2_t_cancel_func_set, METH_VARARGS, "svn_ra_callbacks2_t_cancel_func_set(svn_ra_callbacks2_t self, svn_cancel_func_t cancel_func)"},
	 { "svn_ra_callbacks2_t_cancel_func_get", _wrap_svn_ra_callbacks2_t_cancel_func_get, METH_VARARGS, "svn_ra_callbacks2_t_cancel_func_get(svn_ra_callbacks2_t self) -> svn_cancel_func_t"},
	 { "svn_ra_callbacks2_t_get_client_string_set", _wrap_svn_ra_callbacks2_t_get_client_string_set, METH_VARARGS, "svn_ra_callbacks2_t_get_client_string_set(svn_ra_callbacks2_t self, svn_ra_get_client_string_func_t get_client_string)"},
	 { "svn_ra_callbacks2_t_get_client_string_get", _wrap_svn_ra_callbacks2_t_get_client_string_get, METH_VARARGS, "svn_ra_callbacks2_t_get_client_string_get(svn_ra_callbacks2_t self) -> svn_ra_get_client_string_func_t"},
	 { "svn_ra_callbacks2_t_get_wc_contents_set", _wrap_svn_ra_callbacks2_t_get_wc_contents_set, METH_VARARGS, "svn_ra_callbacks2_t_get_wc_contents_set(svn_ra_callbacks2_t self, svn_ra_get_wc_contents_func_t get_wc_contents)"},
	 { "svn_ra_callbacks2_t_get_wc_contents_get", _wrap_svn_ra_callbacks2_t_get_wc_contents_get, METH_VARARGS, "svn_ra_callbacks2_t_get_wc_contents_get(svn_ra_callbacks2_t self) -> svn_ra_get_wc_contents_func_t"},
	 { "svn_ra_callbacks2_t_check_tunnel_func_set", _wrap_svn_ra_callbacks2_t_check_tunnel_func_set, METH_VARARGS, "svn_ra_callbacks2_t_check_tunnel_func_set(svn_ra_callbacks2_t self, svn_ra_check_tunnel_func_t check_tunnel_func)"},
	 { "svn_ra_callbacks2_t_check_tunnel_func_get", _wrap_svn_ra_callbacks2_t_check_tunnel_func_get, METH_VARARGS, "svn_ra_callbacks2_t_check_tunnel_func_get(svn_ra_callbacks2_t self) -> svn_ra_check_tunnel_func_t"},
	 { "svn_ra_callbacks2_t_open_tunnel_func_set", _wrap_svn_ra_callbacks2_t_open_tunnel_func_set, METH_VARARGS, "svn_ra_callbacks2_t_open_tunnel_func_set(svn_ra_callbacks2_t self, svn_ra_open_tunnel_func_t open_tunnel_func)"},
	 { "svn_ra_callbacks2_t_open_tunnel_func_get", _wrap_svn_ra_callbacks2_t_open_tunnel_func_get, METH_VARARGS, "svn_ra_callbacks2_t_open_tunnel_func_get(svn_ra_callbacks2_t self) -> svn_ra_open_tunnel_func_t"},
	 { "svn_ra_callbacks2_t_tunnel_baton_set", _wrap_svn_ra_callbacks2_t_tunnel_baton_set, METH_VARARGS, "svn_ra_callbacks2_t_tunnel_baton_set(svn_ra_callbacks2_t self, void * tunnel_baton)"},
	 { "svn_ra_callbacks2_t_tunnel_baton_get", _wrap_svn_ra_callbacks2_t_tunnel_baton_get, METH_VARARGS, "svn_ra_callbacks2_t_tunnel_baton_get(svn_ra_callbacks2_t self) -> void *"},
	 { "new_svn_ra_callbacks2_t", _wrap_new_svn_ra_callbacks2_t, METH_VARARGS, "new_svn_ra_callbacks2_t() -> svn_ra_callbacks2_t"},
	 { "delete_svn_ra_callbacks2_t", _wrap_delete_svn_ra_callbacks2_t, METH_VARARGS, "delete_svn_ra_callbacks2_t(svn_ra_callbacks2_t self)"},
	 { "svn_ra_callbacks2_t_swigregister", svn_ra_callbacks2_t_swigregister, METH_O, NULL},
	 { "svn_ra_callbacks2_t_swiginit", svn_ra_callbacks2_t_swiginit, METH_VARARGS, NULL},
	 { "svn_ra_callbacks_t_open_tmp_file_set", _wrap_svn_ra_callbacks_t_open_tmp_file_set, METH_VARARGS, "svn_ra_callbacks_t_open_tmp_file_set(svn_ra_callbacks_t self, svn_error_t *(*)(apr_file_t **,void *,apr_pool_t *) open_tmp_file)"},
	 { "svn_ra_callbacks_t_open_tmp_file_get", _wrap_svn_ra_callbacks_t_open_tmp_file_get, METH_VARARGS, "svn_ra_callbacks_t_open_tmp_file_get(svn_ra_callbacks_t self) -> svn_error_t *(*)(apr_file_t **,void *,apr_pool_t *)"},
	 { "svn_ra_callbacks_t_auth_baton_set", _wrap_svn_ra_callbacks_t_auth_baton_set, METH_VARARGS, "svn_ra_callbacks_t_auth_baton_set(svn_ra_callbacks_t self, svn_auth_baton_t auth_baton)"},
	 { "svn_ra_callbacks_t_auth_baton_get", _wrap_svn_ra_callbacks_t_auth_baton_get, METH_VARARGS, "svn_ra_callbacks_t_auth_baton_get(svn_ra_callbacks_t self) -> svn_auth_baton_t"},
	 { "svn_ra_callbacks_t_get_wc_prop_set", _wrap_svn_ra_callbacks_t_get_wc_prop_set, METH_VARARGS, "svn_ra_callbacks_t_get_wc_prop_set(svn_ra_callbacks_t self, svn_ra_get_wc_prop_func_t get_wc_prop)"},
	 { "svn_ra_callbacks_t_get_wc_prop_get", _wrap_svn_ra_callbacks_t_get_wc_prop_get, METH_VARARGS, "svn_ra_callbacks_t_get_wc_prop_get(svn_ra_callbacks_t self) -> svn_ra_get_wc_prop_func_t"},
	 { "svn_ra_callbacks_t_set_wc_prop_set", _wrap_svn_ra_callbacks_t_set_wc_prop_set, METH_VARARGS, "svn_ra_callbacks_t_set_wc_prop_set(svn_ra_callbacks_t self, svn_ra_set_wc_prop_func_t set_wc_prop)"},
	 { "svn_ra_callbacks_t_set_wc_prop_get", _wrap_svn_ra_callbacks_t_set_wc_prop_get, METH_VARARGS, "svn_ra_callbacks_t_set_wc_prop_get(svn_ra_callbacks_t self) -> svn_ra_set_wc_prop_func_t"},
	 { "svn_ra_callbacks_t_push_wc_prop_set", _wrap_svn_ra_callbacks_t_push_wc_prop_set, METH_VARARGS, "svn_ra_callbacks_t_push_wc_prop_set(svn_ra_callbacks_t self, svn_ra_push_wc_prop_func_t push_wc_prop)"},
	 { "svn_ra_callbacks_t_push_wc_prop_get", _wrap_svn_ra_callbacks_t_push_wc_prop_get, METH_VARARGS, "svn_ra_callbacks_t_push_wc_prop_get(svn_ra_callbacks_t self) -> svn_ra_push_wc_prop_func_t"},
	 { "svn_ra_callbacks_t_invalidate_wc_props_set", _wrap_svn_ra_callbacks_t_invalidate_wc_props_set, METH_VARARGS, "svn_ra_callbacks_t_invalidate_wc_props_set(svn_ra_callbacks_t self, svn_ra_invalidate_wc_props_func_t invalidate_wc_props)"},
	 { "svn_ra_callbacks_t_invalidate_wc_props_get", _wrap_svn_ra_callbacks_t_invalidate_wc_props_get, METH_VARARGS, "svn_ra_callbacks_t_invalidate_wc_props_get(svn_ra_callbacks_t self) -> svn_ra_invalidate_wc_props_func_t"},
	 { "new_svn_ra_callbacks_t", _wrap_new_svn_ra_callbacks_t, METH_VARARGS, "new_svn_ra_callbacks_t() -> svn_ra_callbacks_t"},
	 { "delete_svn_ra_callbacks_t", _wrap_delete_svn_ra_callbacks_t, METH_VARARGS, "delete_svn_ra_callbacks_t(svn_ra_callbacks_t self)"},
	 { "svn_ra_callbacks_t_swigregister", svn_ra_callbacks_t_swigregister, METH_O, NULL},
	 { "svn_ra_callbacks_t_swiginit", svn_ra_callbacks_t_swiginit, METH_VARARGS, NULL},
	 { "svn_ra_initialize", _wrap_svn_ra_initialize, METH_VARARGS, "svn_ra_initialize(apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_create_callbacks", _wrap_svn_ra_create_callbacks, METH_VARARGS, "svn_ra_create_callbacks(apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_open5", _wrap_svn_ra_open5, METH_VARARGS, "svn_ra_open5(char const * repos_URL, char const * uuid, svn_ra_callbacks2_t callbacks, apr_hash_t config, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_open4", _wrap_svn_ra_open4, METH_VARARGS, "svn_ra_open4(char const * repos_URL, char const * uuid, svn_ra_callbacks2_t callbacks, apr_hash_t config, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_open3", _wrap_svn_ra_open3, METH_VARARGS, "svn_ra_open3(char const * repos_URL, char const * uuid, svn_ra_callbacks2_t callbacks, apr_hash_t config, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_open2", _wrap_svn_ra_open2, METH_VARARGS, "svn_ra_open2(char const * repos_URL, svn_ra_callbacks2_t callbacks, apr_hash_t config, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_open", _wrap_svn_ra_open, METH_VARARGS, "svn_ra_open(char const * repos_URL, svn_ra_callbacks_t callbacks, void * callback_baton, apr_hash_t config, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_reparent", _wrap_svn_ra_reparent, METH_VARARGS, "svn_ra_reparent(svn_ra_session_t ra_session, char const * url, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_get_session_url", _wrap_svn_ra_get_session_url, METH_VARARGS, "svn_ra_get_session_url(svn_ra_session_t ra_session, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_get_path_relative_to_session", _wrap_svn_ra_get_path_relative_to_session, METH_VARARGS, "svn_ra_get_path_relative_to_session(svn_ra_session_t ra_session, char const * url, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_get_path_relative_to_root", _wrap_svn_ra_get_path_relative_to_root, METH_VARARGS, "svn_ra_get_path_relative_to_root(svn_ra_session_t ra_session, char const * url, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_get_latest_revnum", _wrap_svn_ra_get_latest_revnum, METH_VARARGS, "svn_ra_get_latest_revnum(svn_ra_session_t session, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_get_dated_revision", _wrap_svn_ra_get_dated_revision, METH_VARARGS, "svn_ra_get_dated_revision(svn_ra_session_t session, apr_time_t tm, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_change_rev_prop2", _wrap_svn_ra_change_rev_prop2, METH_VARARGS, "svn_ra_change_rev_prop2(svn_ra_session_t session, svn_revnum_t rev, char const * name, svn_string_t const *const * old_value_p, svn_string_t const * value, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_change_rev_prop", _wrap_svn_ra_change_rev_prop, METH_VARARGS, "svn_ra_change_rev_prop(svn_ra_session_t session, svn_revnum_t rev, char const * name, svn_string_t const * value, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_rev_proplist", _wrap_svn_ra_rev_proplist, METH_VARARGS, "svn_ra_rev_proplist(svn_ra_session_t session, svn_revnum_t rev, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_rev_prop", _wrap_svn_ra_rev_prop, METH_VARARGS, "svn_ra_rev_prop(svn_ra_session_t session, svn_revnum_t rev, char const * name, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_get_commit_editor3", _wrap_svn_ra_get_commit_editor3, METH_VARARGS, "svn_ra_get_commit_editor3(svn_ra_session_t session, apr_hash_t revprop_table, svn_commit_callback2_t commit_callback, apr_hash_t lock_tokens, svn_boolean_t keep_locks, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_get_commit_editor2", _wrap_svn_ra_get_commit_editor2, METH_VARARGS, "svn_ra_get_commit_editor2(svn_ra_session_t session, char const * log_msg, svn_commit_callback2_t commit_callback, apr_hash_t lock_tokens, svn_boolean_t keep_locks, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_get_commit_editor", _wrap_svn_ra_get_commit_editor, METH_VARARGS, "svn_ra_get_commit_editor(svn_ra_session_t session, char const * log_msg, svn_commit_callback_t callback, apr_hash_t lock_tokens, svn_boolean_t keep_locks, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_get_file", _wrap_svn_ra_get_file, METH_VARARGS, "svn_ra_get_file(svn_ra_session_t session, char const * path, svn_revnum_t revision, svn_stream_t stream, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_get_dir2", _wrap_svn_ra_get_dir2, METH_VARARGS, "svn_ra_get_dir2(svn_ra_session_t session, char const * path, svn_revnum_t revision, apr_uint32_t dirent_fields, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_get_dir", _wrap_svn_ra_get_dir, METH_VARARGS, "svn_ra_get_dir(svn_ra_session_t session, char const * path, svn_revnum_t revision, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_list", _wrap_svn_ra_list, METH_VARARGS, "svn_ra_list(svn_ra_session_t session, char const * path, svn_revnum_t revision, apr_array_header_t patterns, svn_depth_t depth, apr_uint32_t dirent_fields, svn_ra_dirent_receiver_t receiver, void * receiver_baton, apr_pool_t scratch_pool) -> svn_error_t"},
	 { "svn_ra_get_mergeinfo", _wrap_svn_ra_get_mergeinfo, METH_VARARGS, "svn_ra_get_mergeinfo(svn_ra_session_t session, apr_array_header_t paths, svn_revnum_t revision, svn_mergeinfo_inheritance_t inherit, svn_boolean_t include_descendants, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_do_update3", _wrap_svn_ra_do_update3, METH_VARARGS, "svn_ra_do_update3(svn_ra_session_t session, svn_revnum_t revision_to_update_to, char const * update_target, svn_depth_t depth, svn_boolean_t send_copyfrom_args, svn_boolean_t ignore_ancestry, svn_delta_editor_t update_editor, void * update_baton, apr_pool_t result_pool, apr_pool_t scratch_pool) -> svn_error_t"},
	 { "svn_ra_do_update2", _wrap_svn_ra_do_update2, METH_VARARGS, "svn_ra_do_update2(svn_ra_session_t session, svn_revnum_t revision_to_update_to, char const * update_target, svn_depth_t depth, svn_boolean_t send_copyfrom_args, svn_delta_editor_t update_editor, void * update_baton, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_do_update", _wrap_svn_ra_do_update, METH_VARARGS, "svn_ra_do_update(svn_ra_session_t session, svn_revnum_t revision_to_update_to, char const * update_target, svn_boolean_t recurse, svn_delta_editor_t update_editor, void * update_baton, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_do_switch3", _wrap_svn_ra_do_switch3, METH_VARARGS, "svn_ra_do_switch3(svn_ra_session_t session, svn_revnum_t revision_to_switch_to, char const * switch_target, svn_depth_t depth, char const * switch_url, svn_boolean_t send_copyfrom_args, svn_boolean_t ignore_ancestry, svn_delta_editor_t switch_editor, void * switch_baton, apr_pool_t result_pool, apr_pool_t scratch_pool) -> svn_error_t"},
	 { "svn_ra_do_switch2", _wrap_svn_ra_do_switch2, METH_VARARGS, "svn_ra_do_switch2(svn_ra_session_t session, svn_revnum_t revision_to_switch_to, char const * switch_target, svn_depth_t depth, char const * switch_url, svn_delta_editor_t switch_editor, void * switch_baton, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_do_switch", _wrap_svn_ra_do_switch, METH_VARARGS, "svn_ra_do_switch(svn_ra_session_t session, svn_revnum_t revision_to_switch_to, char const * switch_target, svn_boolean_t recurse, char const * switch_url, svn_delta_editor_t switch_editor, void * switch_baton, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_do_status2", _wrap_svn_ra_do_status2, METH_VARARGS, "svn_ra_do_status2(svn_ra_session_t session, char const * status_target, svn_revnum_t revision, svn_depth_t depth, svn_delta_editor_t status_editor, void * status_baton, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_do_status", _wrap_svn_ra_do_status, METH_VARARGS, "svn_ra_do_status(svn_ra_session_t session, char const * status_target, svn_revnum_t revision, svn_boolean_t recurse, svn_delta_editor_t status_editor, void * status_baton, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_do_diff3", _wrap_svn_ra_do_diff3, METH_VARARGS, "svn_ra_do_diff3(svn_ra_session_t session, svn_revnum_t revision, char const * diff_target, svn_depth_t depth, svn_boolean_t ignore_ancestry, svn_boolean_t text_deltas, char const * versus_url, svn_delta_editor_t diff_editor, void * diff_baton, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_do_diff2", _wrap_svn_ra_do_diff2, METH_VARARGS, "svn_ra_do_diff2(svn_ra_session_t session, svn_revnum_t revision, char const * diff_target, svn_boolean_t recurse, svn_boolean_t ignore_ancestry, svn_boolean_t text_deltas, char const * versus_url, svn_delta_editor_t diff_editor, void * diff_baton, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_do_diff", _wrap_svn_ra_do_diff, METH_VARARGS, "svn_ra_do_diff(svn_ra_session_t session, svn_revnum_t revision, char const * diff_target, svn_boolean_t recurse, svn_boolean_t ignore_ancestry, char const * versus_url, svn_delta_editor_t diff_editor, void * diff_baton, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_get_log2", _wrap_svn_ra_get_log2, METH_VARARGS, "svn_ra_get_log2(svn_ra_session_t session, apr_array_header_t paths, svn_revnum_t start, svn_revnum_t end, int limit, svn_boolean_t discover_changed_paths, svn_boolean_t strict_node_history, svn_boolean_t include_merged_revisions, apr_array_header_t revprops, svn_log_entry_receiver_t receiver, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_get_log", _wrap_svn_ra_get_log, METH_VARARGS, "svn_ra_get_log(svn_ra_session_t session, apr_array_header_t paths, svn_revnum_t start, svn_revnum_t end, int limit, svn_boolean_t discover_changed_paths, svn_boolean_t strict_node_history, svn_log_message_receiver_t receiver, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_check_path", _wrap_svn_ra_check_path, METH_VARARGS, "svn_ra_check_path(svn_ra_session_t session, char const * path, svn_revnum_t revision, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_stat", _wrap_svn_ra_stat, METH_VARARGS, "svn_ra_stat(svn_ra_session_t session, char const * path, svn_revnum_t revision, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_get_uuid2", _wrap_svn_ra_get_uuid2, METH_VARARGS, "svn_ra_get_uuid2(svn_ra_session_t session, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_get_uuid", _wrap_svn_ra_get_uuid, METH_VARARGS, "svn_ra_get_uuid(svn_ra_session_t session, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_get_repos_root2", _wrap_svn_ra_get_repos_root2, METH_VARARGS, "svn_ra_get_repos_root2(svn_ra_session_t session, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_get_repos_root", _wrap_svn_ra_get_repos_root, METH_VARARGS, "svn_ra_get_repos_root(svn_ra_session_t session, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_get_locations", _wrap_svn_ra_get_locations, METH_VARARGS, "svn_ra_get_locations(svn_ra_session_t session, char const * path, svn_revnum_t peg_revision, apr_array_header_t location_revisions, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_get_location_segments", _wrap_svn_ra_get_location_segments, METH_VARARGS, "svn_ra_get_location_segments(svn_ra_session_t session, char const * path, svn_revnum_t peg_revision, svn_revnum_t start_rev, svn_revnum_t end_rev, svn_location_segment_receiver_t receiver, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_get_file_revs2", _wrap_svn_ra_get_file_revs2, METH_VARARGS, "svn_ra_get_file_revs2(svn_ra_session_t session, char const * path, svn_revnum_t start, svn_revnum_t end, svn_boolean_t include_merged_revisions, svn_file_rev_handler_t handler, void * handler_baton, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_get_file_revs", _wrap_svn_ra_get_file_revs, METH_VARARGS, "svn_ra_get_file_revs(svn_ra_session_t session, char const * path, svn_revnum_t start, svn_revnum_t end, svn_ra_file_rev_handler_t handler, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_lock", _wrap_svn_ra_lock, METH_VARARGS, "svn_ra_lock(svn_ra_session_t session, apr_hash_t path_revs, char const * comment, svn_boolean_t steal_lock, svn_ra_lock_callback_t lock_func, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_unlock", _wrap_svn_ra_unlock, METH_VARARGS, "svn_ra_unlock(svn_ra_session_t session, apr_hash_t path_tokens, svn_boolean_t break_lock, svn_ra_lock_callback_t lock_func, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_get_lock", _wrap_svn_ra_get_lock, METH_VARARGS, "svn_ra_get_lock(svn_ra_session_t session, char const * path, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_get_locks2", _wrap_svn_ra_get_locks2, METH_VARARGS, "svn_ra_get_locks2(svn_ra_session_t session, char const * path, svn_depth_t depth, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_get_locks", _wrap_svn_ra_get_locks, METH_VARARGS, "svn_ra_get_locks(svn_ra_session_t session, char const * path, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_replay_range", _wrap_svn_ra_replay_range, METH_VARARGS, "svn_ra_replay_range(svn_ra_session_t session, svn_revnum_t start_revision, svn_revnum_t end_revision, svn_revnum_t low_water_mark, svn_boolean_t send_deltas, svn_ra_replay_revstart_callback_t revstart_func, svn_ra_replay_revfinish_callback_t revfinish_func, void * replay_baton, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_replay", _wrap_svn_ra_replay, METH_VARARGS, "svn_ra_replay(svn_ra_session_t session, svn_revnum_t revision, svn_revnum_t low_water_mark, svn_boolean_t send_deltas, svn_delta_editor_t editor, void * edit_baton, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_get_deleted_rev", _wrap_svn_ra_get_deleted_rev, METH_VARARGS, "svn_ra_get_deleted_rev(svn_ra_session_t session, char const * path, svn_revnum_t peg_revision, svn_revnum_t end_revision, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_get_inherited_props", _wrap_svn_ra_get_inherited_props, METH_VARARGS, "svn_ra_get_inherited_props(svn_ra_session_t session, char const * path, svn_revnum_t revision, apr_pool_t result_pool, apr_pool_t scratch_pool) -> svn_error_t"},
	 { "svn_ra_has_capability", _wrap_svn_ra_has_capability, METH_VARARGS, "svn_ra_has_capability(svn_ra_session_t session, char const * capability, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_print_modules", _wrap_svn_ra_print_modules, METH_VARARGS, "svn_ra_print_modules(svn_stringbuf_t * output, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_print_ra_libraries", _wrap_svn_ra_print_ra_libraries, METH_VARARGS, "svn_ra_print_ra_libraries(void * ra_baton, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_plugin_t_name_set", _wrap_svn_ra_plugin_t_name_set, METH_VARARGS, "svn_ra_plugin_t_name_set(svn_ra_plugin_t self, char const * name)"},
	 { "svn_ra_plugin_t_name_get", _wrap_svn_ra_plugin_t_name_get, METH_VARARGS, "svn_ra_plugin_t_name_get(svn_ra_plugin_t self) -> char const *"},
	 { "svn_ra_plugin_t_description_set", _wrap_svn_ra_plugin_t_description_set, METH_VARARGS, "svn_ra_plugin_t_description_set(svn_ra_plugin_t self, char const * description)"},
	 { "svn_ra_plugin_t_description_get", _wrap_svn_ra_plugin_t_description_get, METH_VARARGS, "svn_ra_plugin_t_description_get(svn_ra_plugin_t self) -> char const *"},
	 { "svn_ra_plugin_t_open_set", _wrap_svn_ra_plugin_t_open_set, METH_VARARGS, "svn_ra_plugin_t_open_set(svn_ra_plugin_t self, svn_error_t *(*)(void **,char const *,svn_ra_callbacks_t const *,void *,apr_hash_t *,apr_pool_t *) open)"},
	 { "svn_ra_plugin_t_open_get", _wrap_svn_ra_plugin_t_open_get, METH_VARARGS, "svn_ra_plugin_t_open_get(svn_ra_plugin_t self) -> svn_error_t *(*)(void **,char const *,svn_ra_callbacks_t const *,void *,apr_hash_t *,apr_pool_t *)"},
	 { "svn_ra_plugin_t_get_latest_revnum_set", _wrap_svn_ra_plugin_t_get_latest_revnum_set, METH_VARARGS, "svn_ra_plugin_t_get_latest_revnum_set(svn_ra_plugin_t self, svn_error_t *(*)(void *,svn_revnum_t *,apr_pool_t *) get_latest_revnum)"},
	 { "svn_ra_plugin_t_get_latest_revnum_get", _wrap_svn_ra_plugin_t_get_latest_revnum_get, METH_VARARGS, "svn_ra_plugin_t_get_latest_revnum_get(svn_ra_plugin_t self) -> svn_error_t *(*)(void *,svn_revnum_t *,apr_pool_t *)"},
	 { "svn_ra_plugin_t_get_dated_revision_set", _wrap_svn_ra_plugin_t_get_dated_revision_set, METH_VARARGS, "svn_ra_plugin_t_get_dated_revision_set(svn_ra_plugin_t self, svn_error_t *(*)(void *,svn_revnum_t *,apr_time_t,apr_pool_t *) get_dated_revision)"},
	 { "svn_ra_plugin_t_get_dated_revision_get", _wrap_svn_ra_plugin_t_get_dated_revision_get, METH_VARARGS, "svn_ra_plugin_t_get_dated_revision_get(svn_ra_plugin_t self) -> svn_error_t *(*)(void *,svn_revnum_t *,apr_time_t,apr_pool_t *)"},
	 { "svn_ra_plugin_t_change_rev_prop_set", _wrap_svn_ra_plugin_t_change_rev_prop_set, METH_VARARGS, "svn_ra_plugin_t_change_rev_prop_set(svn_ra_plugin_t self, svn_error_t *(*)(void *,svn_revnum_t,char const *,svn_string_t const *,apr_pool_t *) change_rev_prop)"},
	 { "svn_ra_plugin_t_change_rev_prop_get", _wrap_svn_ra_plugin_t_change_rev_prop_get, METH_VARARGS, "svn_ra_plugin_t_change_rev_prop_get(svn_ra_plugin_t self) -> svn_error_t *(*)(void *,svn_revnum_t,char const *,svn_string_t const *,apr_pool_t *)"},
	 { "svn_ra_plugin_t_rev_proplist_set", _wrap_svn_ra_plugin_t_rev_proplist_set, METH_VARARGS, "svn_ra_plugin_t_rev_proplist_set(svn_ra_plugin_t self, svn_error_t *(*)(void *,svn_revnum_t,apr_hash_t **,apr_pool_t *) rev_proplist)"},
	 { "svn_ra_plugin_t_rev_proplist_get", _wrap_svn_ra_plugin_t_rev_proplist_get, METH_VARARGS, "svn_ra_plugin_t_rev_proplist_get(svn_ra_plugin_t self) -> svn_error_t *(*)(void *,svn_revnum_t,apr_hash_t **,apr_pool_t *)"},
	 { "svn_ra_plugin_t_rev_prop_set", _wrap_svn_ra_plugin_t_rev_prop_set, METH_VARARGS, "svn_ra_plugin_t_rev_prop_set(svn_ra_plugin_t self, svn_error_t *(*)(void *,svn_revnum_t,char const *,svn_string_t **,apr_pool_t *) rev_prop)"},
	 { "svn_ra_plugin_t_rev_prop_get", _wrap_svn_ra_plugin_t_rev_prop_get, METH_VARARGS, "svn_ra_plugin_t_rev_prop_get(svn_ra_plugin_t self) -> svn_error_t *(*)(void *,svn_revnum_t,char const *,svn_string_t **,apr_pool_t *)"},
	 { "svn_ra_plugin_t_get_commit_editor_set", _wrap_svn_ra_plugin_t_get_commit_editor_set, METH_VARARGS, "svn_ra_plugin_t_get_commit_editor_set(svn_ra_plugin_t self, svn_error_t *(*)(void *,svn_delta_editor_t const **,void **,char const *,svn_commit_callback_t,void *,apr_pool_t *) get_commit_editor)"},
	 { "svn_ra_plugin_t_get_commit_editor_get", _wrap_svn_ra_plugin_t_get_commit_editor_get, METH_VARARGS, "svn_ra_plugin_t_get_commit_editor_get(svn_ra_plugin_t self) -> svn_error_t *(*)(void *,svn_delta_editor_t const **,void **,char const *,svn_commit_callback_t,void *,apr_pool_t *)"},
	 { "svn_ra_plugin_t_get_file_set", _wrap_svn_ra_plugin_t_get_file_set, METH_VARARGS, "svn_ra_plugin_t_get_file_set(svn_ra_plugin_t self, svn_error_t *(*)(void *,char const *,svn_revnum_t,svn_stream_t *,svn_revnum_t *,apr_hash_t **,apr_pool_t *) get_file)"},
	 { "svn_ra_plugin_t_get_file_get", _wrap_svn_ra_plugin_t_get_file_get, METH_VARARGS, "svn_ra_plugin_t_get_file_get(svn_ra_plugin_t self) -> svn_error_t *(*)(void *,char const *,svn_revnum_t,svn_stream_t *,svn_revnum_t *,apr_hash_t **,apr_pool_t *)"},
	 { "svn_ra_plugin_t_get_dir_set", _wrap_svn_ra_plugin_t_get_dir_set, METH_VARARGS, "svn_ra_plugin_t_get_dir_set(svn_ra_plugin_t self, svn_error_t *(*)(void *,char const *,svn_revnum_t,apr_hash_t **,svn_revnum_t *,apr_hash_t **,apr_pool_t *) get_dir)"},
	 { "svn_ra_plugin_t_get_dir_get", _wrap_svn_ra_plugin_t_get_dir_get, METH_VARARGS, "svn_ra_plugin_t_get_dir_get(svn_ra_plugin_t self) -> svn_error_t *(*)(void *,char const *,svn_revnum_t,apr_hash_t **,svn_revnum_t *,apr_hash_t **,apr_pool_t *)"},
	 { "svn_ra_plugin_t_do_update_set", _wrap_svn_ra_plugin_t_do_update_set, METH_VARARGS, "svn_ra_plugin_t_do_update_set(svn_ra_plugin_t self, svn_error_t *(*)(void *,svn_ra_reporter_t const **,void **,svn_revnum_t,char const *,svn_boolean_t,svn_delta_editor_t const *,void *,apr_pool_t *) do_update)"},
	 { "svn_ra_plugin_t_do_update_get", _wrap_svn_ra_plugin_t_do_update_get, METH_VARARGS, "svn_ra_plugin_t_do_update_get(svn_ra_plugin_t self) -> svn_error_t *(*)(void *,svn_ra_reporter_t const **,void **,svn_revnum_t,char const *,svn_boolean_t,svn_delta_editor_t const *,void *,apr_pool_t *)"},
	 { "svn_ra_plugin_t_do_switch_set", _wrap_svn_ra_plugin_t_do_switch_set, METH_VARARGS, "svn_ra_plugin_t_do_switch_set(svn_ra_plugin_t self, svn_error_t *(*)(void *,svn_ra_reporter_t const **,void **,svn_revnum_t,char const *,svn_boolean_t,char const *,svn_delta_editor_t const *,void *,apr_pool_t *) do_switch)"},
	 { "svn_ra_plugin_t_do_switch_get", _wrap_svn_ra_plugin_t_do_switch_get, METH_VARARGS, "svn_ra_plugin_t_do_switch_get(svn_ra_plugin_t self) -> svn_error_t *(*)(void *,svn_ra_reporter_t const **,void **,svn_revnum_t,char const *,svn_boolean_t,char const *,svn_delta_editor_t const *,void *,apr_pool_t *)"},
	 { "svn_ra_plugin_t_do_status_set", _wrap_svn_ra_plugin_t_do_status_set, METH_VARARGS, "svn_ra_plugin_t_do_status_set(svn_ra_plugin_t self, svn_error_t *(*)(void *,svn_ra_reporter_t const **,void **,char const *,svn_revnum_t,svn_boolean_t,svn_delta_editor_t const *,void *,apr_pool_t *) do_status)"},
	 { "svn_ra_plugin_t_do_status_get", _wrap_svn_ra_plugin_t_do_status_get, METH_VARARGS, "svn_ra_plugin_t_do_status_get(svn_ra_plugin_t self) -> svn_error_t *(*)(void *,svn_ra_reporter_t const **,void **,char const *,svn_revnum_t,svn_boolean_t,svn_delta_editor_t const *,void *,apr_pool_t *)"},
	 { "svn_ra_plugin_t_do_diff_set", _wrap_svn_ra_plugin_t_do_diff_set, METH_VARARGS, "svn_ra_plugin_t_do_diff_set(svn_ra_plugin_t self, svn_error_t *(*)(void *,svn_ra_reporter_t const **,void **,svn_revnum_t,char const *,svn_boolean_t,svn_boolean_t,char const *,svn_delta_editor_t const *,void *,apr_pool_t *) do_diff)"},
	 { "svn_ra_plugin_t_do_diff_get", _wrap_svn_ra_plugin_t_do_diff_get, METH_VARARGS, "svn_ra_plugin_t_do_diff_get(svn_ra_plugin_t self) -> svn_error_t *(*)(void *,svn_ra_reporter_t const **,void **,svn_revnum_t,char const *,svn_boolean_t,svn_boolean_t,char const *,svn_delta_editor_t const *,void *,apr_pool_t *)"},
	 { "svn_ra_plugin_t_get_log_set", _wrap_svn_ra_plugin_t_get_log_set, METH_VARARGS, "svn_ra_plugin_t_get_log_set(svn_ra_plugin_t self, svn_error_t *(*)(void *,apr_array_header_t const *,svn_revnum_t,svn_revnum_t,svn_boolean_t,svn_boolean_t,svn_log_message_receiver_t,void *,apr_pool_t *) get_log)"},
	 { "svn_ra_plugin_t_get_log_get", _wrap_svn_ra_plugin_t_get_log_get, METH_VARARGS, "svn_ra_plugin_t_get_log_get(svn_ra_plugin_t self) -> svn_error_t *(*)(void *,apr_array_header_t const *,svn_revnum_t,svn_revnum_t,svn_boolean_t,svn_boolean_t,svn_log_message_receiver_t,void *,apr_pool_t *)"},
	 { "svn_ra_plugin_t_check_path_set", _wrap_svn_ra_plugin_t_check_path_set, METH_VARARGS, "svn_ra_plugin_t_check_path_set(svn_ra_plugin_t self, svn_error_t *(*)(void *,char const *,svn_revnum_t,svn_node_kind_t *,apr_pool_t *) check_path)"},
	 { "svn_ra_plugin_t_check_path_get", _wrap_svn_ra_plugin_t_check_path_get, METH_VARARGS, "svn_ra_plugin_t_check_path_get(svn_ra_plugin_t self) -> svn_error_t *(*)(void *,char const *,svn_revnum_t,svn_node_kind_t *,apr_pool_t *)"},
	 { "svn_ra_plugin_t_get_uuid_set", _wrap_svn_ra_plugin_t_get_uuid_set, METH_VARARGS, "svn_ra_plugin_t_get_uuid_set(svn_ra_plugin_t self, svn_error_t *(*)(void *,char const **,apr_pool_t *) get_uuid)"},
	 { "svn_ra_plugin_t_get_uuid_get", _wrap_svn_ra_plugin_t_get_uuid_get, METH_VARARGS, "svn_ra_plugin_t_get_uuid_get(svn_ra_plugin_t self) -> svn_error_t *(*)(void *,char const **,apr_pool_t *)"},
	 { "svn_ra_plugin_t_get_repos_root_set", _wrap_svn_ra_plugin_t_get_repos_root_set, METH_VARARGS, "svn_ra_plugin_t_get_repos_root_set(svn_ra_plugin_t self, svn_error_t *(*)(void *,char const **,apr_pool_t *) get_repos_root)"},
	 { "svn_ra_plugin_t_get_repos_root_get", _wrap_svn_ra_plugin_t_get_repos_root_get, METH_VARARGS, "svn_ra_plugin_t_get_repos_root_get(svn_ra_plugin_t self) -> svn_error_t *(*)(void *,char const **,apr_pool_t *)"},
	 { "svn_ra_plugin_t_get_locations_set", _wrap_svn_ra_plugin_t_get_locations_set, METH_VARARGS, "svn_ra_plugin_t_get_locations_set(svn_ra_plugin_t self, svn_error_t *(*)(void *,apr_hash_t **,char const *,svn_revnum_t,apr_array_header_t *,apr_pool_t *) get_locations)"},
	 { "svn_ra_plugin_t_get_locations_get", _wrap_svn_ra_plugin_t_get_locations_get, METH_VARARGS, "svn_ra_plugin_t_get_locations_get(svn_ra_plugin_t self) -> svn_error_t *(*)(void *,apr_hash_t **,char const *,svn_revnum_t,apr_array_header_t *,apr_pool_t *)"},
	 { "svn_ra_plugin_t_get_file_revs_set", _wrap_svn_ra_plugin_t_get_file_revs_set, METH_VARARGS, "svn_ra_plugin_t_get_file_revs_set(svn_ra_plugin_t self, svn_error_t *(*)(void *,char const *,svn_revnum_t,svn_revnum_t,svn_ra_file_rev_handler_t,void *,apr_pool_t *) get_file_revs)"},
	 { "svn_ra_plugin_t_get_file_revs_get", _wrap_svn_ra_plugin_t_get_file_revs_get, METH_VARARGS, "svn_ra_plugin_t_get_file_revs_get(svn_ra_plugin_t self) -> svn_error_t *(*)(void *,char const *,svn_revnum_t,svn_revnum_t,svn_ra_file_rev_handler_t,void *,apr_pool_t *)"},
	 { "svn_ra_plugin_t_get_version_set", _wrap_svn_ra_plugin_t_get_version_set, METH_VARARGS, "svn_ra_plugin_t_get_version_set(svn_ra_plugin_t self, svn_version_t const *(*)(void) get_version)"},
	 { "svn_ra_plugin_t_get_version_get", _wrap_svn_ra_plugin_t_get_version_get, METH_VARARGS, "svn_ra_plugin_t_get_version_get(svn_ra_plugin_t self) -> svn_version_t const *(*)(void)"},
	 { "new_svn_ra_plugin_t", _wrap_new_svn_ra_plugin_t, METH_VARARGS, "new_svn_ra_plugin_t() -> svn_ra_plugin_t"},
	 { "delete_svn_ra_plugin_t", _wrap_delete_svn_ra_plugin_t, METH_VARARGS, "delete_svn_ra_plugin_t(svn_ra_plugin_t self)"},
	 { "svn_ra_plugin_t_swigregister", svn_ra_plugin_t_swigregister, METH_O, NULL},
	 { "svn_ra_plugin_t_swiginit", svn_ra_plugin_t_swiginit, METH_VARARGS, NULL},
	 { "svn_ra_init_ra_libs", _wrap_svn_ra_init_ra_libs, METH_VARARGS, "svn_ra_init_ra_libs(apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_get_ra_library", _wrap_svn_ra_get_ra_library, METH_VARARGS, "svn_ra_get_ra_library(void * ra_baton, char const * url, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_session_t_swigregister", svn_ra_session_t_swigregister, METH_O, NULL},
	 { "svn_ra_reporter3_invoke_set_path", _wrap_svn_ra_reporter3_invoke_set_path, METH_VARARGS, "svn_ra_reporter3_invoke_set_path(svn_ra_reporter3_t _obj, void * report_baton, char const * path, svn_revnum_t revision, svn_depth_t depth, svn_boolean_t start_empty, char const * lock_token, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_reporter3_invoke_delete_path", _wrap_svn_ra_reporter3_invoke_delete_path, METH_VARARGS, "svn_ra_reporter3_invoke_delete_path(svn_ra_reporter3_t _obj, void * report_baton, char const * path, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_reporter3_invoke_link_path", _wrap_svn_ra_reporter3_invoke_link_path, METH_VARARGS, "svn_ra_reporter3_invoke_link_path(svn_ra_reporter3_t _obj, void * report_baton, char const * path, char const * url, svn_revnum_t revision, svn_depth_t depth, svn_boolean_t start_empty, char const * lock_token, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_reporter3_invoke_finish_report", _wrap_svn_ra_reporter3_invoke_finish_report, METH_VARARGS, "svn_ra_reporter3_invoke_finish_report(svn_ra_reporter3_t _obj, void * report_baton, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_reporter3_invoke_abort_report", _wrap_svn_ra_reporter3_invoke_abort_report, METH_VARARGS, "svn_ra_reporter3_invoke_abort_report(svn_ra_reporter3_t _obj, void * report_baton, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_reporter2_invoke_set_path", _wrap_svn_ra_reporter2_invoke_set_path, METH_VARARGS, "svn_ra_reporter2_invoke_set_path(svn_ra_reporter2_t _obj, void * report_baton, char const * path, svn_revnum_t revision, svn_boolean_t start_empty, char const * lock_token, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_reporter2_invoke_delete_path", _wrap_svn_ra_reporter2_invoke_delete_path, METH_VARARGS, "svn_ra_reporter2_invoke_delete_path(svn_ra_reporter2_t _obj, void * report_baton, char const * path, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_reporter2_invoke_link_path", _wrap_svn_ra_reporter2_invoke_link_path, METH_VARARGS, "svn_ra_reporter2_invoke_link_path(svn_ra_reporter2_t _obj, void * report_baton, char const * path, char const * url, svn_revnum_t revision, svn_boolean_t start_empty, char const * lock_token, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_reporter2_invoke_finish_report", _wrap_svn_ra_reporter2_invoke_finish_report, METH_VARARGS, "svn_ra_reporter2_invoke_finish_report(svn_ra_reporter2_t _obj, void * report_baton, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_reporter2_invoke_abort_report", _wrap_svn_ra_reporter2_invoke_abort_report, METH_VARARGS, "svn_ra_reporter2_invoke_abort_report(svn_ra_reporter2_t _obj, void * report_baton, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_reporter_invoke_set_path", _wrap_svn_ra_reporter_invoke_set_path, METH_VARARGS, "svn_ra_reporter_invoke_set_path(svn_ra_reporter_t _obj, void * report_baton, char const * path, svn_revnum_t revision, svn_boolean_t start_empty, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_reporter_invoke_delete_path", _wrap_svn_ra_reporter_invoke_delete_path, METH_VARARGS, "svn_ra_reporter_invoke_delete_path(svn_ra_reporter_t _obj, void * report_baton, char const * path, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_reporter_invoke_link_path", _wrap_svn_ra_reporter_invoke_link_path, METH_VARARGS, "svn_ra_reporter_invoke_link_path(svn_ra_reporter_t _obj, void * report_baton, char const * path, char const * url, svn_revnum_t revision, svn_boolean_t start_empty, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_reporter_invoke_finish_report", _wrap_svn_ra_reporter_invoke_finish_report, METH_VARARGS, "svn_ra_reporter_invoke_finish_report(svn_ra_reporter_t _obj, void * report_baton, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_reporter_invoke_abort_report", _wrap_svn_ra_reporter_invoke_abort_report, METH_VARARGS, "svn_ra_reporter_invoke_abort_report(svn_ra_reporter_t _obj, void * report_baton, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_callbacks2_invoke_open_tmp_file", _wrap_svn_ra_callbacks2_invoke_open_tmp_file, METH_VARARGS, "svn_ra_callbacks2_invoke_open_tmp_file(svn_ra_callbacks2_t _obj, void * callback_baton, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_callbacks_invoke_open_tmp_file", _wrap_svn_ra_callbacks_invoke_open_tmp_file, METH_VARARGS, "svn_ra_callbacks_invoke_open_tmp_file(svn_ra_callbacks_t _obj, void * callback_baton, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_plugin_invoke_open", _wrap_svn_ra_plugin_invoke_open, METH_VARARGS, "svn_ra_plugin_invoke_open(svn_ra_plugin_t _obj, char const * repos_URL, svn_ra_callbacks_t callbacks, void * callback_baton, apr_hash_t config, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_plugin_invoke_get_latest_revnum", _wrap_svn_ra_plugin_invoke_get_latest_revnum, METH_VARARGS, "svn_ra_plugin_invoke_get_latest_revnum(svn_ra_plugin_t _obj, void * session_baton, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_plugin_invoke_get_dated_revision", _wrap_svn_ra_plugin_invoke_get_dated_revision, METH_VARARGS, "svn_ra_plugin_invoke_get_dated_revision(svn_ra_plugin_t _obj, void * session_baton, apr_time_t tm, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_plugin_invoke_change_rev_prop", _wrap_svn_ra_plugin_invoke_change_rev_prop, METH_VARARGS, "svn_ra_plugin_invoke_change_rev_prop(svn_ra_plugin_t _obj, void * session_baton, svn_revnum_t rev, char const * name, svn_string_t const * value, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_plugin_invoke_rev_proplist", _wrap_svn_ra_plugin_invoke_rev_proplist, METH_VARARGS, "svn_ra_plugin_invoke_rev_proplist(svn_ra_plugin_t _obj, void * session_baton, svn_revnum_t rev, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_plugin_invoke_rev_prop", _wrap_svn_ra_plugin_invoke_rev_prop, METH_VARARGS, "svn_ra_plugin_invoke_rev_prop(svn_ra_plugin_t _obj, void * session_baton, svn_revnum_t rev, char const * name, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_plugin_invoke_get_commit_editor", _wrap_svn_ra_plugin_invoke_get_commit_editor, METH_VARARGS, "svn_ra_plugin_invoke_get_commit_editor(svn_ra_plugin_t _obj, void * session_baton, char const * log_msg, svn_commit_callback_t callback, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_plugin_invoke_get_file", _wrap_svn_ra_plugin_invoke_get_file, METH_VARARGS, "svn_ra_plugin_invoke_get_file(svn_ra_plugin_t _obj, void * session_baton, char const * path, svn_revnum_t revision, svn_stream_t stream, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_plugin_invoke_get_dir", _wrap_svn_ra_plugin_invoke_get_dir, METH_VARARGS, "svn_ra_plugin_invoke_get_dir(svn_ra_plugin_t _obj, void * session_baton, char const * path, svn_revnum_t revision, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_plugin_invoke_do_update", _wrap_svn_ra_plugin_invoke_do_update, METH_VARARGS, "svn_ra_plugin_invoke_do_update(svn_ra_plugin_t _obj, void * session_baton, svn_revnum_t revision_to_update_to, char const * update_target, svn_boolean_t recurse, svn_delta_editor_t update_editor, void * update_baton, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_plugin_invoke_do_switch", _wrap_svn_ra_plugin_invoke_do_switch, METH_VARARGS, "svn_ra_plugin_invoke_do_switch(svn_ra_plugin_t _obj, void * session_baton, svn_revnum_t revision_to_switch_to, char const * switch_target, svn_boolean_t recurse, char const * switch_url, svn_delta_editor_t switch_editor, void * switch_baton, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_plugin_invoke_do_status", _wrap_svn_ra_plugin_invoke_do_status, METH_VARARGS, "svn_ra_plugin_invoke_do_status(svn_ra_plugin_t _obj, void * session_baton, char const * status_target, svn_revnum_t revision, svn_boolean_t recurse, svn_delta_editor_t status_editor, void * status_baton, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_plugin_invoke_do_diff", _wrap_svn_ra_plugin_invoke_do_diff, METH_VARARGS, "svn_ra_plugin_invoke_do_diff(svn_ra_plugin_t _obj, void * session_baton, svn_revnum_t revision, char const * diff_target, svn_boolean_t recurse, svn_boolean_t ignore_ancestry, char const * versus_url, svn_delta_editor_t diff_editor, void * diff_baton, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_plugin_invoke_get_log", _wrap_svn_ra_plugin_invoke_get_log, METH_VARARGS, "svn_ra_plugin_invoke_get_log(svn_ra_plugin_t _obj, void * session_baton, apr_array_header_t paths, svn_revnum_t start, svn_revnum_t end, svn_boolean_t discover_changed_paths, svn_boolean_t strict_node_history, svn_log_message_receiver_t receiver, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_plugin_invoke_check_path", _wrap_svn_ra_plugin_invoke_check_path, METH_VARARGS, "svn_ra_plugin_invoke_check_path(svn_ra_plugin_t _obj, void * session_baton, char const * path, svn_revnum_t revision, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_plugin_invoke_get_uuid", _wrap_svn_ra_plugin_invoke_get_uuid, METH_VARARGS, "svn_ra_plugin_invoke_get_uuid(svn_ra_plugin_t _obj, void * session_baton, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_plugin_invoke_get_repos_root", _wrap_svn_ra_plugin_invoke_get_repos_root, METH_VARARGS, "svn_ra_plugin_invoke_get_repos_root(svn_ra_plugin_t _obj, void * session_baton, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_plugin_invoke_get_locations", _wrap_svn_ra_plugin_invoke_get_locations, METH_VARARGS, "svn_ra_plugin_invoke_get_locations(svn_ra_plugin_t _obj, void * session_baton, char const * path, svn_revnum_t peg_revision, apr_array_header_t location_revisions, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_plugin_invoke_get_file_revs", _wrap_svn_ra_plugin_invoke_get_file_revs, METH_VARARGS, "svn_ra_plugin_invoke_get_file_revs(svn_ra_plugin_t _obj, void * session_baton, char const * path, svn_revnum_t start, svn_revnum_t end, svn_ra_file_rev_handler_t handler, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_plugin_invoke_get_version", _wrap_svn_ra_plugin_invoke_get_version, METH_VARARGS, "svn_ra_plugin_invoke_get_version(svn_ra_plugin_t _obj) -> svn_version_t"},
	 { "svn_ra_invoke_get_wc_prop_func", _wrap_svn_ra_invoke_get_wc_prop_func, METH_VARARGS, "svn_ra_invoke_get_wc_prop_func(svn_ra_get_wc_prop_func_t _obj, void * baton, char const * path, char const * name, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_invoke_set_wc_prop_func", _wrap_svn_ra_invoke_set_wc_prop_func, METH_VARARGS, "svn_ra_invoke_set_wc_prop_func(svn_ra_set_wc_prop_func_t _obj, void * baton, char const * path, char const * name, svn_string_t const * value, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_invoke_push_wc_prop_func", _wrap_svn_ra_invoke_push_wc_prop_func, METH_VARARGS, "svn_ra_invoke_push_wc_prop_func(svn_ra_push_wc_prop_func_t _obj, void * baton, char const * path, char const * name, svn_string_t const * value, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_invoke_invalidate_wc_props_func", _wrap_svn_ra_invoke_invalidate_wc_props_func, METH_VARARGS, "svn_ra_invoke_invalidate_wc_props_func(svn_ra_invalidate_wc_props_func_t _obj, void * baton, char const * path, char const * name, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_invoke_get_wc_contents_func", _wrap_svn_ra_invoke_get_wc_contents_func, METH_VARARGS, "svn_ra_invoke_get_wc_contents_func(svn_ra_get_wc_contents_func_t _obj, void * baton, svn_checksum_t checksum, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_invoke_get_latest_revnum_func", _wrap_svn_ra_invoke_get_latest_revnum_func, METH_VARARGS, "svn_ra_invoke_get_latest_revnum_func(svn_ra_get_latest_revnum_func_t _obj, void * session_baton) -> svn_error_t"},
	 { "svn_ra_invoke_get_client_string_func", _wrap_svn_ra_invoke_get_client_string_func, METH_VARARGS, "svn_ra_invoke_get_client_string_func(svn_ra_get_client_string_func_t _obj, void * baton, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_invoke_file_rev_handler", _wrap_svn_ra_invoke_file_rev_handler, METH_VARARGS, "svn_ra_invoke_file_rev_handler(svn_ra_file_rev_handler_t _obj, void * baton, char const * path, svn_revnum_t rev, apr_hash_t rev_props, apr_array_header_t prop_diffs, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_invoke_lock_callback", _wrap_svn_ra_invoke_lock_callback, METH_VARARGS, "svn_ra_invoke_lock_callback(svn_ra_lock_callback_t _obj, void * baton, char const * path, svn_boolean_t do_lock, svn_lock_t lock, svn_error_t ra_err, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_invoke_progress_notify_func", _wrap_svn_ra_invoke_progress_notify_func, METH_VARARGS, "svn_ra_invoke_progress_notify_func(svn_ra_progress_notify_func_t _obj, apr_off_t progress, apr_off_t total, void * baton, apr_pool_t pool)"},
	 { "svn_ra_invoke_replay_revstart_callback", _wrap_svn_ra_invoke_replay_revstart_callback, METH_VARARGS, "svn_ra_invoke_replay_revstart_callback(svn_ra_replay_revstart_callback_t _obj, svn_revnum_t revision, void * replay_baton, apr_hash_t rev_props, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_invoke_replay_revfinish_callback", _wrap_svn_ra_invoke_replay_revfinish_callback, METH_VARARGS, "svn_ra_invoke_replay_revfinish_callback(svn_ra_replay_revfinish_callback_t _obj, svn_revnum_t revision, void * replay_baton, svn_delta_editor_t editor, void * edit_baton, apr_hash_t rev_props, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_invoke_check_tunnel_func", _wrap_svn_ra_invoke_check_tunnel_func, METH_VARARGS, "svn_ra_invoke_check_tunnel_func(svn_ra_check_tunnel_func_t _obj, void * tunnel_baton, char const * tunnel_name) -> svn_boolean_t"},
	 { "svn_ra_invoke_close_tunnel_func", _wrap_svn_ra_invoke_close_tunnel_func, METH_VARARGS, "svn_ra_invoke_close_tunnel_func(svn_ra_close_tunnel_func_t _obj, void * close_baton, void * tunnel_baton)"},
	 { "svn_ra_invoke_open_tunnel_func", _wrap_svn_ra_invoke_open_tunnel_func, METH_VARARGS, "svn_ra_invoke_open_tunnel_func(svn_ra_open_tunnel_func_t _obj, void * tunnel_baton, char const * tunnel_name, char const * user, char const * hostname, int port, svn_cancel_func_t cancel_func, apr_pool_t pool) -> svn_error_t"},
	 { "svn_ra_invoke_init_func", _wrap_svn_ra_invoke_init_func, METH_VARARGS, "svn_ra_invoke_init_func(svn_ra_init_func_t _obj, int abi_version, apr_pool_t pool, apr_hash_t hash) -> svn_error_t"},
	 { "svn_ra_get_wc_prop_func_t_swigregister", svn_ra_get_wc_prop_func_t_swigregister, METH_O, NULL},
	 { "svn_ra_set_wc_prop_func_t_swigregister", svn_ra_set_wc_prop_func_t_swigregister, METH_O, NULL},
	 { "svn_ra_push_wc_prop_func_t_swigregister", svn_ra_push_wc_prop_func_t_swigregister, METH_O, NULL},
	 { "svn_ra_invalidate_wc_props_func_t_swigregister", svn_ra_invalidate_wc_props_func_t_swigregister, METH_O, NULL},
	 { "svn_ra_get_wc_contents_func_t_swigregister", svn_ra_get_wc_contents_func_t_swigregister, METH_O, NULL},
	 { "svn_ra_get_latest_revnum_func_t_swigregister", svn_ra_get_latest_revnum_func_t_swigregister, METH_O, NULL},
	 { "svn_ra_get_client_string_func_t_swigregister", svn_ra_get_client_string_func_t_swigregister, METH_O, NULL},
	 { "svn_ra_file_rev_handler_t_swigregister", svn_ra_file_rev_handler_t_swigregister, METH_O, NULL},
	 { "svn_ra_lock_callback_t_swigregister", svn_ra_lock_callback_t_swigregister, METH_O, NULL},
	 { "svn_ra_progress_notify_func_t_swigregister", svn_ra_progress_notify_func_t_swigregister, METH_O, NULL},
	 { "svn_ra_replay_revstart_callback_t_swigregister", svn_ra_replay_revstart_callback_t_swigregister, METH_O, NULL},
	 { "svn_ra_replay_revfinish_callback_t_swigregister", svn_ra_replay_revfinish_callback_t_swigregister, METH_O, NULL},
	 { "svn_ra_check_tunnel_func_t_swigregister", svn_ra_check_tunnel_func_t_swigregister, METH_O, NULL},
	 { "svn_ra_close_tunnel_func_t_swigregister", svn_ra_close_tunnel_func_t_swigregister, METH_O, NULL},
	 { "svn_ra_open_tunnel_func_t_swigregister", svn_ra_open_tunnel_func_t_swigregister, METH_O, NULL},
	 { "svn_ra_init_func_t_swigregister", svn_ra_init_func_t_swigregister, METH_O, NULL},
	 { NULL, NULL, 0, NULL }
};


/* -------- TYPE CONVERSION AND EQUIVALENCE RULES (BEGIN) -------- */

static swig_type_info _swigt__p_apr_array_header_t = {"_p_apr_array_header_t", "svn_rangelist_t *|apr_array_header_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_apr_file_t = {"_p_apr_file_t", "apr_file_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_apr_hash_t = {"_p_apr_hash_t", "apr_hash_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_apr_int32_t = {"_p_apr_int32_t", "time_t *|apr_int32_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_apr_int64_t = {"_p_apr_int64_t", "apr_time_t *|svn_filesize_t *|apr_int64_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_apr_off_t = {"_p_apr_off_t", "apr_off_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_apr_pool_t = {"_p_apr_pool_t", "apr_pool_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_char = {"_p_char", "char *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_apr_off_t_apr_off_t_p_void_p_apr_pool_t__void = {"_p_f_apr_off_t_apr_off_t_p_void_p_apr_pool_t__void", "svn_ra_progress_notify_func_t|void (*)(apr_off_t,apr_off_t,void *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_int_p_apr_pool_t_p_apr_hash_t__p_svn_error_t = {"_p_f_int_p_apr_pool_t_p_apr_hash_t__p_svn_error_t", "svn_ra_init_func_t|svn_error_t *(*)(int,apr_pool_t *,apr_hash_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_apr_getopt_t_p_void_p_apr_pool_t__p_svn_error_t = {"_p_f_p_apr_getopt_t_p_void_p_apr_pool_t__p_svn_error_t", "svn_opt_subcommand_t *|svn_error_t *(*)(apr_getopt_t *,void *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_p_apr_file_t_p_void_p_apr_pool_t__p_svn_error_t = {"_p_f_p_p_apr_file_t_p_void_p_apr_pool_t__p_svn_error_t", "svn_error_t *(*)(apr_file_t **,void *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_p_svn_stream_t_p_p_svn_stream_t_p_svn_ra_close_tunnel_func_t_p_p_void_p_void_p_q_const__char_p_q_const__char_p_q_const__char_int_svn_cancel_func_t_p_void_p_apr_pool_t__p_svn_error_t = {"_p_f_p_p_svn_stream_t_p_p_svn_stream_t_p_svn_ra_close_tunnel_func_t_p_p_void_p_void_p_q_const__char_p_q_const__char_p_q_const__char_int_svn_cancel_func_t_p_void_p_apr_pool_t__p_svn_error_t", "svn_ra_open_tunnel_func_t|svn_error_t *(*)(svn_stream_t **,svn_stream_t **,svn_ra_close_tunnel_func_t *,void **,void *,char const *,char const *,char const *,int,svn_cancel_func_t,void *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_p_void_p_q_const__char_p_q_const__svn_ra_callbacks_t_p_void_p_apr_hash_t_p_apr_pool_t__p_svn_error_t = {"_p_f_p_p_void_p_q_const__char_p_q_const__svn_ra_callbacks_t_p_void_p_apr_hash_t_p_apr_pool_t__p_svn_error_t", "svn_error_t *(*)(void **,char const *,svn_ra_callbacks_t const *,void *,apr_hash_t *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_q_const__char_p_svn_dirent_t_p_void_p_apr_pool_t__p_svn_error_t = {"_p_f_p_q_const__char_p_svn_dirent_t_p_void_p_apr_pool_t__p_svn_error_t", "svn_ra_dirent_receiver_t|svn_error_t *(*)(char const *,svn_dirent_t *,void *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_q_const__svn_commit_info_t_p_void_p_apr_pool_t__p_svn_error_t = {"_p_f_p_q_const__svn_commit_info_t_p_void_p_apr_pool_t__p_svn_error_t", "svn_commit_callback2_t|svn_error_t *(*)(svn_commit_info_t const *,void *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_svn_location_segment_t_p_void_p_apr_pool_t__p_svn_error_t = {"_p_f_p_svn_location_segment_t_p_void_p_apr_pool_t__p_svn_error_t", "svn_location_segment_receiver_t|svn_error_t *(*)(svn_location_segment_t *,void *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void__p_svn_error_t = {"_p_f_p_void__p_svn_error_t", "svn_cancel_func_t|svn_error_t *(*)(void *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_apr_hash_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_p_apr_hash_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t", "svn_log_message_receiver_t|svn_error_t *(*)(void *,apr_hash_t *,svn_revnum_t,char const *,char const *,char const *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_p_apr_pool_t__p_svn_error_t", "svn_error_t *(*)(void *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_p_apr_hash_t_p_q_const__char_svn_revnum_t_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_p_p_apr_hash_t_p_q_const__char_svn_revnum_t_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t", "svn_error_t *(*)(void *,apr_hash_t **,char const *,svn_revnum_t,apr_array_header_t *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_p_q_const__char_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_p_p_q_const__char_p_apr_pool_t__p_svn_error_t", "svn_ra_get_client_string_func_t|svn_error_t *(*)(void *,char const **,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_p_q_const__svn_delta_editor_t_p_p_void_p_q_const__char_svn_commit_callback_t_p_void_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_p_p_q_const__svn_delta_editor_t_p_p_void_p_q_const__char_svn_commit_callback_t_p_void_p_apr_pool_t__p_svn_error_t", "svn_error_t *(*)(void *,svn_delta_editor_t const **,void **,char const *,svn_commit_callback_t,void *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_p_q_const__svn_ra_reporter_t_p_p_void_p_q_const__char_svn_revnum_t_svn_boolean_t_p_q_const__svn_delta_editor_t_p_void_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_p_p_q_const__svn_ra_reporter_t_p_p_void_p_q_const__char_svn_revnum_t_svn_boolean_t_p_q_const__svn_delta_editor_t_p_void_p_apr_pool_t__p_svn_error_t", "svn_error_t *(*)(void *,svn_ra_reporter_t const **,void **,char const *,svn_revnum_t,svn_boolean_t,svn_delta_editor_t const *,void *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_p_q_const__svn_ra_reporter_t_p_p_void_svn_revnum_t_p_q_const__char_svn_boolean_t_p_q_const__char_p_q_const__svn_delta_editor_t_p_void_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_p_p_q_const__svn_ra_reporter_t_p_p_void_svn_revnum_t_p_q_const__char_svn_boolean_t_p_q_const__char_p_q_const__svn_delta_editor_t_p_void_p_apr_pool_t__p_svn_error_t", "svn_error_t *(*)(void *,svn_ra_reporter_t const **,void **,svn_revnum_t,char const *,svn_boolean_t,char const *,svn_delta_editor_t const *,void *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_p_q_const__svn_ra_reporter_t_p_p_void_svn_revnum_t_p_q_const__char_svn_boolean_t_p_q_const__svn_delta_editor_t_p_void_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_p_p_q_const__svn_ra_reporter_t_p_p_void_svn_revnum_t_p_q_const__char_svn_boolean_t_p_q_const__svn_delta_editor_t_p_void_p_apr_pool_t__p_svn_error_t", "svn_error_t *(*)(void *,svn_ra_reporter_t const **,void **,svn_revnum_t,char const *,svn_boolean_t,svn_delta_editor_t const *,void *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_p_q_const__svn_ra_reporter_t_p_p_void_svn_revnum_t_p_q_const__char_svn_boolean_t_svn_boolean_t_p_q_const__char_p_q_const__svn_delta_editor_t_p_void_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_p_p_q_const__svn_ra_reporter_t_p_p_void_svn_revnum_t_p_q_const__char_svn_boolean_t_svn_boolean_t_p_q_const__char_p_q_const__svn_delta_editor_t_p_void_p_apr_pool_t__p_svn_error_t", "svn_error_t *(*)(void *,svn_ra_reporter_t const **,void **,svn_revnum_t,char const *,svn_boolean_t,svn_boolean_t,char const *,svn_delta_editor_t const *,void *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_p_svn_stream_t_p_q_const__svn_checksum_t_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_p_p_svn_stream_t_p_q_const__svn_checksum_t_p_apr_pool_t__p_svn_error_t", "svn_ra_get_wc_contents_func_t|svn_error_t *(*)(void *,svn_stream_t **,svn_checksum_t const *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_q_const__apr_array_header_t_svn_revnum_t_svn_revnum_t_svn_boolean_t_svn_boolean_t_svn_log_message_receiver_t_p_void_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_p_q_const__apr_array_header_t_svn_revnum_t_svn_revnum_t_svn_boolean_t_svn_boolean_t_svn_log_message_receiver_t_p_void_p_apr_pool_t__p_svn_error_t", "svn_error_t *(*)(void *,apr_array_header_t const *,svn_revnum_t,svn_revnum_t,svn_boolean_t,svn_boolean_t,svn_log_message_receiver_t,void *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_q_const__char__int = {"_p_f_p_void_p_q_const__char__int", "svn_ra_check_tunnel_func_t|int (*)(void *,char const *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_q_const__char_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_p_q_const__char_p_apr_pool_t__p_svn_error_t", "svn_error_t *(*)(void *,char const *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t", "svn_ra_invalidate_wc_props_func_t|svn_error_t *(*)(void *,char const *,char const *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_q_const__char_p_q_const__char_p_p_q_const__svn_string_t_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_p_q_const__char_p_q_const__char_p_p_q_const__svn_string_t_p_apr_pool_t__p_svn_error_t", "svn_ra_get_wc_prop_func_t|svn_error_t *(*)(void *,char const *,char const *,svn_string_t const **,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_q_const__char_p_q_const__char_p_q_const__svn_string_t_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_p_q_const__char_p_q_const__char_p_q_const__svn_string_t_p_apr_pool_t__p_svn_error_t", "svn_ra_push_wc_prop_func_t|svn_ra_set_wc_prop_func_t|svn_error_t *(*)(void *,char const *,char const *,svn_string_t const *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_q_const__char_p_q_const__char_svn_revnum_t_svn_boolean_t_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_p_q_const__char_p_q_const__char_svn_revnum_t_svn_boolean_t_p_apr_pool_t__p_svn_error_t", "svn_error_t *(*)(void *,char const *,char const *,svn_revnum_t,svn_boolean_t,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_q_const__char_p_q_const__char_svn_revnum_t_svn_boolean_t_p_q_const__char_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_p_q_const__char_p_q_const__char_svn_revnum_t_svn_boolean_t_p_q_const__char_p_apr_pool_t__p_svn_error_t", "svn_error_t *(*)(void *,char const *,char const *,svn_revnum_t,svn_boolean_t,char const *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_q_const__char_p_q_const__char_svn_revnum_t_svn_depth_t_svn_boolean_t_p_q_const__char_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_p_q_const__char_p_q_const__char_svn_revnum_t_svn_depth_t_svn_boolean_t_p_q_const__char_p_apr_pool_t__p_svn_error_t", "svn_error_t *(*)(void *,char const *,char const *,svn_revnum_t,svn_depth_t,svn_boolean_t,char const *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_q_const__char_svn_boolean_t_p_q_const__svn_lock_t_p_svn_error_t_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_p_q_const__char_svn_boolean_t_p_q_const__svn_lock_t_p_svn_error_t_p_apr_pool_t__p_svn_error_t", "svn_ra_lock_callback_t|svn_error_t *(*)(void *,char const *,svn_boolean_t,svn_lock_t const *,svn_error_t *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_q_const__char_svn_revnum_t_p_apr_hash_t_p_svn_txdelta_window_handler_t_p_p_void_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_p_q_const__char_svn_revnum_t_p_apr_hash_t_p_svn_txdelta_window_handler_t_p_p_void_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t", "svn_ra_file_rev_handler_t|svn_error_t *(*)(void *,char const *,svn_revnum_t,apr_hash_t *,svn_txdelta_window_handler_t *,void **,apr_array_header_t *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_q_const__char_svn_revnum_t_p_apr_hash_t_svn_boolean_t_p_svn_txdelta_window_handler_t_p_p_void_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_p_q_const__char_svn_revnum_t_p_apr_hash_t_svn_boolean_t_p_svn_txdelta_window_handler_t_p_p_void_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t", "svn_file_rev_handler_t|svn_error_t *(*)(void *,char const *,svn_revnum_t,apr_hash_t *,svn_boolean_t,svn_txdelta_window_handler_t *,void **,apr_array_header_t *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_q_const__char_svn_revnum_t_p_p_apr_hash_t_p_svn_revnum_t_p_p_apr_hash_t_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_p_q_const__char_svn_revnum_t_p_p_apr_hash_t_p_svn_revnum_t_p_p_apr_hash_t_p_apr_pool_t__p_svn_error_t", "svn_error_t *(*)(void *,char const *,svn_revnum_t,apr_hash_t **,svn_revnum_t *,apr_hash_t **,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_q_const__char_svn_revnum_t_p_svn_node_kind_t_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_p_q_const__char_svn_revnum_t_p_svn_node_kind_t_p_apr_pool_t__p_svn_error_t", "svn_error_t *(*)(void *,char const *,svn_revnum_t,svn_node_kind_t *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_q_const__char_svn_revnum_t_p_svn_stream_t_p_svn_revnum_t_p_p_apr_hash_t_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_p_q_const__char_svn_revnum_t_p_svn_stream_t_p_svn_revnum_t_p_p_apr_hash_t_p_apr_pool_t__p_svn_error_t", "svn_error_t *(*)(void *,char const *,svn_revnum_t,svn_stream_t *,svn_revnum_t *,apr_hash_t **,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_q_const__char_svn_revnum_t_svn_boolean_t_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_p_q_const__char_svn_revnum_t_svn_boolean_t_p_apr_pool_t__p_svn_error_t", "svn_error_t *(*)(void *,char const *,svn_revnum_t,svn_boolean_t,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_q_const__char_svn_revnum_t_svn_boolean_t_p_q_const__char_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_p_q_const__char_svn_revnum_t_svn_boolean_t_p_q_const__char_p_apr_pool_t__p_svn_error_t", "svn_error_t *(*)(void *,char const *,svn_revnum_t,svn_boolean_t,char const *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_q_const__char_svn_revnum_t_svn_depth_t_svn_boolean_t_p_q_const__char_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_p_q_const__char_svn_revnum_t_svn_depth_t_svn_boolean_t_p_q_const__char_p_apr_pool_t__p_svn_error_t", "svn_error_t *(*)(void *,char const *,svn_revnum_t,svn_depth_t,svn_boolean_t,char const *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_q_const__char_svn_revnum_t_svn_revnum_t_svn_ra_file_rev_handler_t_p_void_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_p_q_const__char_svn_revnum_t_svn_revnum_t_svn_ra_file_rev_handler_t_p_void_p_apr_pool_t__p_svn_error_t", "svn_error_t *(*)(void *,char const *,svn_revnum_t,svn_revnum_t,svn_ra_file_rev_handler_t,void *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_svn_log_entry_t_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_p_svn_log_entry_t_p_apr_pool_t__p_svn_error_t", "svn_log_entry_receiver_t|svn_error_t *(*)(void *,svn_log_entry_t *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_svn_revnum_t__p_svn_error_t = {"_p_f_p_void_p_svn_revnum_t__p_svn_error_t", "svn_ra_get_latest_revnum_func_t|svn_error_t *(*)(void *,svn_revnum_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_svn_revnum_t_apr_time_t_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_p_svn_revnum_t_apr_time_t_p_apr_pool_t__p_svn_error_t", "svn_error_t *(*)(void *,svn_revnum_t *,apr_time_t,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_svn_revnum_t_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_p_svn_revnum_t_p_apr_pool_t__p_svn_error_t", "svn_error_t *(*)(void *,svn_revnum_t *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_void__void = {"_p_f_p_void_p_void__void", "svn_ra_close_tunnel_func_t|void (*)(void *,void *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_svn_revnum_t_p_p_apr_hash_t_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_svn_revnum_t_p_p_apr_hash_t_p_apr_pool_t__p_svn_error_t", "svn_error_t *(*)(void *,svn_revnum_t,apr_hash_t **,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_svn_revnum_t_p_q_const__char_p_p_svn_string_t_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_svn_revnum_t_p_q_const__char_p_p_svn_string_t_p_apr_pool_t__p_svn_error_t", "svn_error_t *(*)(void *,svn_revnum_t,char const *,svn_string_t **,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_svn_revnum_t_p_q_const__char_p_q_const__svn_string_t_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_svn_revnum_t_p_q_const__char_p_q_const__svn_string_t_p_apr_pool_t__p_svn_error_t", "svn_error_t *(*)(void *,svn_revnum_t,char const *,svn_string_t const *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_svn_revnum_t_p_q_const__char_p_q_const__char_p_void__p_svn_error_t = {"_p_f_svn_revnum_t_p_q_const__char_p_q_const__char_p_void__p_svn_error_t", "svn_commit_callback_t|svn_error_t *(*)(svn_revnum_t,char const *,char const *,void *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_svn_revnum_t_p_void_p_p_q_const__svn_delta_editor_t_p_p_void_p_apr_hash_t_p_apr_pool_t__p_svn_error_t = {"_p_f_svn_revnum_t_p_void_p_p_q_const__svn_delta_editor_t_p_p_void_p_apr_hash_t_p_apr_pool_t__p_svn_error_t", "svn_ra_replay_revstart_callback_t|svn_error_t *(*)(svn_revnum_t,void *,svn_delta_editor_t const **,void **,apr_hash_t *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_svn_revnum_t_p_void_p_q_const__svn_delta_editor_t_p_void_p_apr_hash_t_p_apr_pool_t__p_svn_error_t = {"_p_f_svn_revnum_t_p_void_p_q_const__svn_delta_editor_t_p_void_p_apr_hash_t_p_apr_pool_t__p_svn_error_t", "svn_ra_replay_revfinish_callback_t|svn_error_t *(*)(svn_revnum_t,void *,svn_delta_editor_t const *,void *,apr_hash_t *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_void__p_svn_version_t = {"_p_f_void__p_svn_version_t", "svn_version_t *(*)(void)|struct svn_version_t *(*)(void)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_int = {"_p_int", "apr_status_t *|svn_boolean_t *|int *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_long = {"_p_long", "svn_revnum_t *|long *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_apr_array_header_t = {"_p_p_apr_array_header_t", "apr_array_header_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_apr_file_t = {"_p_p_apr_file_t", "apr_file_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_apr_hash_t = {"_p_p_apr_hash_t", "svn_mergeinfo_catalog_t *|apr_hash_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_char = {"_p_p_char", "char **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_f_apr_off_t_apr_off_t_p_void_p_apr_pool_t__void = {"_p_p_f_apr_off_t_apr_off_t_p_void_p_apr_pool_t__void", "svn_ra_progress_notify_func_t *|void (**)(apr_off_t,apr_off_t,void *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_f_int_p_apr_pool_t_p_apr_hash_t__p_svn_error_t = {"_p_p_f_int_p_apr_pool_t_p_apr_hash_t__p_svn_error_t", "svn_ra_init_func_t *|svn_error_t *(**)(int,apr_pool_t *,apr_hash_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_f_p_p_svn_stream_t_p_p_svn_stream_t_p_svn_ra_close_tunnel_func_t_p_p_void_p_void_p_q_const__char_p_q_const__char_p_q_const__char_int_svn_cancel_func_t_p_void_p_apr_pool_t__p_svn_error_t = {"_p_p_f_p_p_svn_stream_t_p_p_svn_stream_t_p_svn_ra_close_tunnel_func_t_p_p_void_p_void_p_q_const__char_p_q_const__char_p_q_const__char_int_svn_cancel_func_t_p_void_p_apr_pool_t__p_svn_error_t", "svn_ra_open_tunnel_func_t *|svn_error_t *(**)(svn_stream_t **,svn_stream_t **,svn_ra_close_tunnel_func_t *,void **,void *,char const *,char const *,char const *,int,svn_cancel_func_t,void *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_f_p_svn_txdelta_window_t_p_void__p_svn_error_t = {"_p_p_f_p_svn_txdelta_window_t_p_void__p_svn_error_t", "svn_txdelta_window_handler_t *|svn_error_t *(**)(svn_txdelta_window_t *,void *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_f_p_void__p_svn_error_t = {"_p_p_f_p_void__p_svn_error_t", "svn_cancel_func_t *|svn_error_t *(**)(void *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_f_p_void_p_p_q_const__char_p_apr_pool_t__p_svn_error_t = {"_p_p_f_p_void_p_p_q_const__char_p_apr_pool_t__p_svn_error_t", "svn_ra_get_client_string_func_t *|svn_error_t *(**)(void *,char const **,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_f_p_void_p_p_svn_stream_t_p_q_const__svn_checksum_t_p_apr_pool_t__p_svn_error_t = {"_p_p_f_p_void_p_p_svn_stream_t_p_q_const__svn_checksum_t_p_apr_pool_t__p_svn_error_t", "svn_ra_get_wc_contents_func_t *|svn_error_t *(**)(void *,svn_stream_t **,svn_checksum_t const *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_f_p_void_p_q_const__char__int = {"_p_p_f_p_void_p_q_const__char__int", "svn_ra_check_tunnel_func_t *|int (**)(void *,char const *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_f_p_void_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t = {"_p_p_f_p_void_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t", "svn_ra_invalidate_wc_props_func_t *|svn_error_t *(**)(void *,char const *,char const *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_f_p_void_p_q_const__char_p_q_const__char_p_p_q_const__svn_string_t_p_apr_pool_t__p_svn_error_t = {"_p_p_f_p_void_p_q_const__char_p_q_const__char_p_p_q_const__svn_string_t_p_apr_pool_t__p_svn_error_t", "svn_ra_get_wc_prop_func_t *|svn_error_t *(**)(void *,char const *,char const *,svn_string_t const **,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_f_p_void_p_q_const__char_p_q_const__char_p_q_const__svn_string_t_p_apr_pool_t__p_svn_error_t = {"_p_p_f_p_void_p_q_const__char_p_q_const__char_p_q_const__svn_string_t_p_apr_pool_t__p_svn_error_t", "svn_ra_push_wc_prop_func_t *|svn_ra_set_wc_prop_func_t *|svn_error_t *(**)(void *,char const *,char const *,svn_string_t const *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_f_p_void_p_q_const__char_svn_boolean_t_p_q_const__svn_lock_t_p_svn_error_t_p_apr_pool_t__p_svn_error_t = {"_p_p_f_p_void_p_q_const__char_svn_boolean_t_p_q_const__svn_lock_t_p_svn_error_t_p_apr_pool_t__p_svn_error_t", "svn_ra_lock_callback_t *|svn_error_t *(**)(void *,char const *,svn_boolean_t,svn_lock_t const *,svn_error_t *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_f_p_void_p_q_const__char_svn_revnum_t_p_apr_hash_t_p_svn_txdelta_window_handler_t_p_p_void_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t = {"_p_p_f_p_void_p_q_const__char_svn_revnum_t_p_apr_hash_t_p_svn_txdelta_window_handler_t_p_p_void_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t", "svn_ra_file_rev_handler_t *|svn_error_t *(**)(void *,char const *,svn_revnum_t,apr_hash_t *,svn_txdelta_window_handler_t *,void **,apr_array_header_t *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_f_p_void_p_q_const__char_svn_revnum_t_p_apr_hash_t_svn_boolean_t_p_svn_txdelta_window_handler_t_p_p_void_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t = {"_p_p_f_p_void_p_q_const__char_svn_revnum_t_p_apr_hash_t_svn_boolean_t_p_svn_txdelta_window_handler_t_p_p_void_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t", "svn_file_rev_handler_t *|svn_error_t *(**)(void *,char const *,svn_revnum_t,apr_hash_t *,svn_boolean_t,svn_txdelta_window_handler_t *,void **,apr_array_header_t *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_f_p_void_p_svn_revnum_t__p_svn_error_t = {"_p_p_f_p_void_p_svn_revnum_t__p_svn_error_t", "svn_ra_get_latest_revnum_func_t *|svn_error_t *(**)(void *,svn_revnum_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_f_p_void_p_void__void = {"_p_p_f_p_void_p_void__void", "svn_ra_close_tunnel_func_t *|void (**)(void *,void *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_f_svn_revnum_t_p_void_p_p_q_const__svn_delta_editor_t_p_p_void_p_apr_hash_t_p_apr_pool_t__p_svn_error_t = {"_p_p_f_svn_revnum_t_p_void_p_p_q_const__svn_delta_editor_t_p_p_void_p_apr_hash_t_p_apr_pool_t__p_svn_error_t", "svn_ra_replay_revstart_callback_t *|svn_error_t *(**)(svn_revnum_t,void *,svn_delta_editor_t const **,void **,apr_hash_t *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_f_svn_revnum_t_p_void_p_q_const__svn_delta_editor_t_p_void_p_apr_hash_t_p_apr_pool_t__p_svn_error_t = {"_p_p_f_svn_revnum_t_p_void_p_q_const__svn_delta_editor_t_p_void_p_apr_hash_t_p_apr_pool_t__p_svn_error_t", "svn_ra_replay_revfinish_callback_t *|svn_error_t *(**)(svn_revnum_t,void *,svn_delta_editor_t const *,void *,apr_hash_t *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_svn_delta_editor_t = {"_p_p_svn_delta_editor_t", "svn_delta_editor_t **|struct svn_delta_editor_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_svn_dirent_t = {"_p_p_svn_dirent_t", "svn_dirent_t **|struct svn_dirent_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_svn_lock_t = {"_p_p_svn_lock_t", "svn_lock_t **|struct svn_lock_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_svn_ra_callbacks2_t = {"_p_p_svn_ra_callbacks2_t", "svn_ra_callbacks2_t **|struct svn_ra_callbacks2_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_svn_ra_plugin_t = {"_p_p_svn_ra_plugin_t", "svn_ra_plugin_t **|struct svn_ra_plugin_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_svn_ra_reporter2_t = {"_p_p_svn_ra_reporter2_t", "svn_ra_reporter2_t **|struct svn_ra_reporter2_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_svn_ra_reporter3_t = {"_p_p_svn_ra_reporter3_t", "svn_ra_reporter3_t **|struct svn_ra_reporter3_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_svn_ra_reporter_t = {"_p_p_svn_ra_reporter_t", "svn_ra_reporter_t **|struct svn_ra_reporter_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_svn_ra_session_t = {"_p_p_svn_ra_session_t", "svn_ra_session_t **|struct svn_ra_session_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_svn_stream_t = {"_p_p_svn_stream_t", "svn_stream_t **|struct svn_stream_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_svn_string_t = {"_p_p_svn_string_t", "svn_string_t **|struct svn_string_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_svn_stringbuf_t = {"_p_p_svn_stringbuf_t", "svn_stringbuf_t **|struct svn_stringbuf_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_void = {"_p_p_void", "void **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_auth_baton_t = {"_p_svn_auth_baton_t", "svn_auth_baton_t *|struct svn_auth_baton_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_auth_cred_simple_t = {"_p_svn_auth_cred_simple_t", "svn_auth_cred_simple_t *|struct svn_auth_cred_simple_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_auth_cred_ssl_client_cert_pw_t = {"_p_svn_auth_cred_ssl_client_cert_pw_t", "svn_auth_cred_ssl_client_cert_pw_t *|struct svn_auth_cred_ssl_client_cert_pw_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_auth_cred_ssl_client_cert_t = {"_p_svn_auth_cred_ssl_client_cert_t", "svn_auth_cred_ssl_client_cert_t *|struct svn_auth_cred_ssl_client_cert_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_auth_cred_ssl_server_trust_t = {"_p_svn_auth_cred_ssl_server_trust_t", "svn_auth_cred_ssl_server_trust_t *|struct svn_auth_cred_ssl_server_trust_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_auth_cred_username_t = {"_p_svn_auth_cred_username_t", "svn_auth_cred_username_t *|struct svn_auth_cred_username_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_auth_iterstate_t = {"_p_svn_auth_iterstate_t", "svn_auth_iterstate_t *|struct svn_auth_iterstate_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_auth_provider_object_t = {"_p_svn_auth_provider_object_t", "svn_auth_provider_object_t *|struct svn_auth_provider_object_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_auth_provider_t = {"_p_svn_auth_provider_t", "svn_auth_provider_t *|struct svn_auth_provider_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_auth_ssl_server_cert_info_t = {"_p_svn_auth_ssl_server_cert_info_t", "svn_auth_ssl_server_cert_info_t *|struct svn_auth_ssl_server_cert_info_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_cache_config_t = {"_p_svn_cache_config_t", "svn_cache_config_t *|struct svn_cache_config_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_checksum_ctx_t = {"_p_svn_checksum_ctx_t", "svn_checksum_ctx_t *|struct svn_checksum_ctx_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_checksum_kind_t = {"_p_svn_checksum_kind_t", "svn_checksum_kind_t *|enum svn_checksum_kind_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_checksum_t = {"_p_svn_checksum_t", "svn_checksum_t *|struct svn_checksum_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_client__shelf_version_t = {"_p_svn_client__shelf_version_t", "svn_client__shelf_version_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_client_conflict_option_t = {"_p_svn_client_conflict_option_t", "svn_client_conflict_option_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_commit_info_t = {"_p_svn_commit_info_t", "svn_commit_info_t *|struct svn_commit_info_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_config_t = {"_p_svn_config_t", "svn_config_t *|struct svn_config_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_delta_editor_t = {"_p_svn_delta_editor_t", "svn_delta_editor_t *|struct svn_delta_editor_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_delta_path_driver_state_t = {"_p_svn_delta_path_driver_state_t", "svn_delta_path_driver_state_t *|struct svn_delta_path_driver_state_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_depth_t = {"_p_svn_depth_t", "svn_depth_t *|enum svn_depth_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_diff_hunk_t = {"_p_svn_diff_hunk_t", "svn_diff_hunk_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_dirent_t = {"_p_svn_dirent_t", "svn_dirent_t *|struct svn_dirent_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_errno_t = {"_p_svn_errno_t", "svn_errno_t *|enum svn_errno_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_error_t = {"_p_svn_error_t", "svn_error_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_fs_lock_target_t = {"_p_svn_fs_lock_target_t", "svn_fs_lock_target_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_io_dirent2_t = {"_p_svn_io_dirent2_t", "svn_io_dirent2_t *|struct svn_io_dirent2_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_io_dirent_t = {"_p_svn_io_dirent_t", "svn_io_dirent_t *|struct svn_io_dirent_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_io_file_del_t = {"_p_svn_io_file_del_t", "svn_io_file_del_t *|enum svn_io_file_del_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_location_segment_t = {"_p_svn_location_segment_t", "svn_location_segment_t *|struct svn_location_segment_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_lock_t = {"_p_svn_lock_t", "svn_lock_t *|struct svn_lock_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_log_changed_path2_t = {"_p_svn_log_changed_path2_t", "svn_log_changed_path2_t *|struct svn_log_changed_path2_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_log_changed_path_t = {"_p_svn_log_changed_path_t", "svn_log_changed_path_t *|struct svn_log_changed_path_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_log_entry_t = {"_p_svn_log_entry_t", "svn_log_entry_t *|struct svn_log_entry_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_merge_range_t = {"_p_svn_merge_range_t", "svn_merge_range_t *|struct svn_merge_range_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_mergeinfo_inheritance_t = {"_p_svn_mergeinfo_inheritance_t", "svn_mergeinfo_inheritance_t *|enum svn_mergeinfo_inheritance_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_node_kind_t = {"_p_svn_node_kind_t", "svn_node_kind_t *|enum svn_node_kind_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_opt_revision_range_t = {"_p_svn_opt_revision_range_t", "svn_opt_revision_range_t *|struct svn_opt_revision_range_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_opt_revision_t = {"_p_svn_opt_revision_t", "svn_opt_revision_t *|struct svn_opt_revision_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_opt_revision_value_t = {"_p_svn_opt_revision_value_t", "svn_opt_revision_value_t *|union svn_opt_revision_value_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_opt_subcommand_desc2_t = {"_p_svn_opt_subcommand_desc2_t", "svn_opt_subcommand_desc2_t *|struct svn_opt_subcommand_desc2_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_opt_subcommand_desc3_t = {"_p_svn_opt_subcommand_desc3_t", "svn_opt_subcommand_desc3_t *|struct svn_opt_subcommand_desc3_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_opt_subcommand_desc_t = {"_p_svn_opt_subcommand_desc_t", "svn_opt_subcommand_desc_t *|struct svn_opt_subcommand_desc_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_prop_inherited_item_t = {"_p_svn_prop_inherited_item_t", "svn_prop_inherited_item_t *|struct svn_prop_inherited_item_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_prop_kind = {"_p_svn_prop_kind", "svn_prop_kind_t *|enum svn_prop_kind *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_ra_callbacks2_t = {"_p_svn_ra_callbacks2_t", "svn_ra_callbacks2_t *|struct svn_ra_callbacks2_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_ra_callbacks_t = {"_p_svn_ra_callbacks_t", "svn_ra_callbacks_t *|struct svn_ra_callbacks_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_ra_plugin_t = {"_p_svn_ra_plugin_t", "svn_ra_plugin_t *|struct svn_ra_plugin_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_ra_reporter2_t = {"_p_svn_ra_reporter2_t", "svn_ra_reporter2_t *|struct svn_ra_reporter2_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_ra_reporter3_t = {"_p_svn_ra_reporter3_t", "svn_ra_reporter3_t *|struct svn_ra_reporter3_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_ra_reporter_t = {"_p_svn_ra_reporter_t", "svn_ra_reporter_t *|struct svn_ra_reporter_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_ra_session_t = {"_p_svn_ra_session_t", "svn_ra_session_t *|struct svn_ra_session_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_stream_mark_t = {"_p_svn_stream_mark_t", "svn_stream_mark_t *|struct svn_stream_mark_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_stream_t = {"_p_svn_stream_t", "svn_stream_t *|struct svn_stream_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_string_t = {"_p_svn_string_t", "svn_string_t *|struct svn_string_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_stringbuf_t = {"_p_svn_stringbuf_t", "svn_stringbuf_t *|struct svn_stringbuf_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_tristate_t = {"_p_svn_tristate_t", "svn_tristate_t *|enum svn_tristate_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_txdelta_op_t = {"_p_svn_txdelta_op_t", "svn_txdelta_op_t *|struct svn_txdelta_op_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_txdelta_stream_t = {"_p_svn_txdelta_stream_t", "svn_txdelta_stream_t *|struct svn_txdelta_stream_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_txdelta_window_t = {"_p_svn_txdelta_window_t", "svn_txdelta_window_t *|struct svn_txdelta_window_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_version_checklist_t = {"_p_svn_version_checklist_t", "svn_version_checklist_t *|struct svn_version_checklist_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_version_ext_linked_lib_t = {"_p_svn_version_ext_linked_lib_t", "svn_version_ext_linked_lib_t *|struct svn_version_ext_linked_lib_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_version_ext_loaded_lib_t = {"_p_svn_version_ext_loaded_lib_t", "svn_version_ext_loaded_lib_t *|struct svn_version_ext_loaded_lib_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_version_extended_t = {"_p_svn_version_extended_t", "svn_version_extended_t *|struct svn_version_extended_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_version_t = {"_p_svn_version_t", "svn_version_t *|struct svn_version_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_external_item2_t = {"_p_svn_wc_external_item2_t", "svn_wc_external_item2_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_unsigned_long = {"_p_unsigned_long", "svn_linenum_t *|unsigned long *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_void = {"_p_void", "void *", 0, 0, (void*)0, 0};

static swig_type_info *swig_type_initial[] = {
  &_swigt__p_apr_array_header_t,
  &_swigt__p_apr_file_t,
  &_swigt__p_apr_hash_t,
  &_swigt__p_apr_int32_t,
  &_swigt__p_apr_int64_t,
  &_swigt__p_apr_off_t,
  &_swigt__p_apr_pool_t,
  &_swigt__p_char,
  &_swigt__p_f_apr_off_t_apr_off_t_p_void_p_apr_pool_t__void,
  &_swigt__p_f_int_p_apr_pool_t_p_apr_hash_t__p_svn_error_t,
  &_swigt__p_f_p_apr_getopt_t_p_void_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_p_apr_file_t_p_void_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_p_svn_stream_t_p_p_svn_stream_t_p_svn_ra_close_tunnel_func_t_p_p_void_p_void_p_q_const__char_p_q_const__char_p_q_const__char_int_svn_cancel_func_t_p_void_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_p_void_p_q_const__char_p_q_const__svn_ra_callbacks_t_p_void_p_apr_hash_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_q_const__char_p_svn_dirent_t_p_void_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_q_const__svn_commit_info_t_p_void_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_svn_location_segment_t_p_void_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void__p_svn_error_t,
  &_swigt__p_f_p_void_p_apr_hash_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_p_apr_hash_t_p_q_const__char_svn_revnum_t_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_p_q_const__char_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_p_q_const__svn_delta_editor_t_p_p_void_p_q_const__char_svn_commit_callback_t_p_void_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_p_q_const__svn_ra_reporter_t_p_p_void_p_q_const__char_svn_revnum_t_svn_boolean_t_p_q_const__svn_delta_editor_t_p_void_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_p_q_const__svn_ra_reporter_t_p_p_void_svn_revnum_t_p_q_const__char_svn_boolean_t_p_q_const__char_p_q_const__svn_delta_editor_t_p_void_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_p_q_const__svn_ra_reporter_t_p_p_void_svn_revnum_t_p_q_const__char_svn_boolean_t_p_q_const__svn_delta_editor_t_p_void_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_p_q_const__svn_ra_reporter_t_p_p_void_svn_revnum_t_p_q_const__char_svn_boolean_t_svn_boolean_t_p_q_const__char_p_q_const__svn_delta_editor_t_p_void_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_p_svn_stream_t_p_q_const__svn_checksum_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_q_const__apr_array_header_t_svn_revnum_t_svn_revnum_t_svn_boolean_t_svn_boolean_t_svn_log_message_receiver_t_p_void_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_q_const__char__int,
  &_swigt__p_f_p_void_p_q_const__char_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_q_const__char_p_q_const__char_p_p_q_const__svn_string_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_q_const__char_p_q_const__char_p_q_const__svn_string_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_q_const__char_p_q_const__char_svn_revnum_t_svn_boolean_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_q_const__char_p_q_const__char_svn_revnum_t_svn_boolean_t_p_q_const__char_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_q_const__char_p_q_const__char_svn_revnum_t_svn_depth_t_svn_boolean_t_p_q_const__char_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_q_const__char_svn_boolean_t_p_q_const__svn_lock_t_p_svn_error_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_q_const__char_svn_revnum_t_p_apr_hash_t_p_svn_txdelta_window_handler_t_p_p_void_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_q_const__char_svn_revnum_t_p_apr_hash_t_svn_boolean_t_p_svn_txdelta_window_handler_t_p_p_void_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_q_const__char_svn_revnum_t_p_p_apr_hash_t_p_svn_revnum_t_p_p_apr_hash_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_q_const__char_svn_revnum_t_p_svn_node_kind_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_q_const__char_svn_revnum_t_p_svn_stream_t_p_svn_revnum_t_p_p_apr_hash_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_q_const__char_svn_revnum_t_svn_boolean_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_q_const__char_svn_revnum_t_svn_boolean_t_p_q_const__char_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_q_const__char_svn_revnum_t_svn_depth_t_svn_boolean_t_p_q_const__char_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_q_const__char_svn_revnum_t_svn_revnum_t_svn_ra_file_rev_handler_t_p_void_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_svn_log_entry_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_svn_revnum_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_svn_revnum_t_apr_time_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_svn_revnum_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_void__void,
  &_swigt__p_f_p_void_svn_revnum_t_p_p_apr_hash_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void_svn_revnum_t_p_q_const__char_p_p_svn_string_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void_svn_revnum_t_p_q_const__char_p_q_const__svn_string_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_svn_revnum_t_p_q_const__char_p_q_const__char_p_void__p_svn_error_t,
  &_swigt__p_f_svn_revnum_t_p_void_p_p_q_const__svn_delta_editor_t_p_p_void_p_apr_hash_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_svn_revnum_t_p_void_p_q_const__svn_delta_editor_t_p_void_p_apr_hash_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_void__p_svn_version_t,
  &_swigt__p_int,
  &_swigt__p_long,
  &_swigt__p_p_apr_array_header_t,
  &_swigt__p_p_apr_file_t,
  &_swigt__p_p_apr_hash_t,
  &_swigt__p_p_char,
  &_swigt__p_p_f_apr_off_t_apr_off_t_p_void_p_apr_pool_t__void,
  &_swigt__p_p_f_int_p_apr_pool_t_p_apr_hash_t__p_svn_error_t,
  &_swigt__p_p_f_p_p_svn_stream_t_p_p_svn_stream_t_p_svn_ra_close_tunnel_func_t_p_p_void_p_void_p_q_const__char_p_q_const__char_p_q_const__char_int_svn_cancel_func_t_p_void_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_p_f_p_svn_txdelta_window_t_p_void__p_svn_error_t,
  &_swigt__p_p_f_p_void__p_svn_error_t,
  &_swigt__p_p_f_p_void_p_p_q_const__char_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_p_f_p_void_p_p_svn_stream_t_p_q_const__svn_checksum_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_p_f_p_void_p_q_const__char__int,
  &_swigt__p_p_f_p_void_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_p_f_p_void_p_q_const__char_p_q_const__char_p_p_q_const__svn_string_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_p_f_p_void_p_q_const__char_p_q_const__char_p_q_const__svn_string_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_p_f_p_void_p_q_const__char_svn_boolean_t_p_q_const__svn_lock_t_p_svn_error_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_p_f_p_void_p_q_const__char_svn_revnum_t_p_apr_hash_t_p_svn_txdelta_window_handler_t_p_p_void_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_p_f_p_void_p_q_const__char_svn_revnum_t_p_apr_hash_t_svn_boolean_t_p_svn_txdelta_window_handler_t_p_p_void_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_p_f_p_void_p_svn_revnum_t__p_svn_error_t,
  &_swigt__p_p_f_p_void_p_void__void,
  &_swigt__p_p_f_svn_revnum_t_p_void_p_p_q_const__svn_delta_editor_t_p_p_void_p_apr_hash_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_p_f_svn_revnum_t_p_void_p_q_const__svn_delta_editor_t_p_void_p_apr_hash_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_p_svn_delta_editor_t,
  &_swigt__p_p_svn_dirent_t,
  &_swigt__p_p_svn_lock_t,
  &_swigt__p_p_svn_ra_callbacks2_t,
  &_swigt__p_p_svn_ra_plugin_t,
  &_swigt__p_p_svn_ra_reporter2_t,
  &_swigt__p_p_svn_ra_reporter3_t,
  &_swigt__p_p_svn_ra_reporter_t,
  &_swigt__p_p_svn_ra_session_t,
  &_swigt__p_p_svn_stream_t,
  &_swigt__p_p_svn_string_t,
  &_swigt__p_p_svn_stringbuf_t,
  &_swigt__p_p_void,
  &_swigt__p_svn_auth_baton_t,
  &_swigt__p_svn_auth_cred_simple_t,
  &_swigt__p_svn_auth_cred_ssl_client_cert_pw_t,
  &_swigt__p_svn_auth_cred_ssl_client_cert_t,
  &_swigt__p_svn_auth_cred_ssl_server_trust_t,
  &_swigt__p_svn_auth_cred_username_t,
  &_swigt__p_svn_auth_iterstate_t,
  &_swigt__p_svn_auth_provider_object_t,
  &_swigt__p_svn_auth_provider_t,
  &_swigt__p_svn_auth_ssl_server_cert_info_t,
  &_swigt__p_svn_cache_config_t,
  &_swigt__p_svn_checksum_ctx_t,
  &_swigt__p_svn_checksum_kind_t,
  &_swigt__p_svn_checksum_t,
  &_swigt__p_svn_client__shelf_version_t,
  &_swigt__p_svn_client_conflict_option_t,
  &_swigt__p_svn_commit_info_t,
  &_swigt__p_svn_config_t,
  &_swigt__p_svn_delta_editor_t,
  &_swigt__p_svn_delta_path_driver_state_t,
  &_swigt__p_svn_depth_t,
  &_swigt__p_svn_diff_hunk_t,
  &_swigt__p_svn_dirent_t,
  &_swigt__p_svn_errno_t,
  &_swigt__p_svn_error_t,
  &_swigt__p_svn_fs_lock_target_t,
  &_swigt__p_svn_io_dirent2_t,
  &_swigt__p_svn_io_dirent_t,
  &_swigt__p_svn_io_file_del_t,
  &_swigt__p_svn_location_segment_t,
  &_swigt__p_svn_lock_t,
  &_swigt__p_svn_log_changed_path2_t,
  &_swigt__p_svn_log_changed_path_t,
  &_swigt__p_svn_log_entry_t,
  &_swigt__p_svn_merge_range_t,
  &_swigt__p_svn_mergeinfo_inheritance_t,
  &_swigt__p_svn_node_kind_t,
  &_swigt__p_svn_opt_revision_range_t,
  &_swigt__p_svn_opt_revision_t,
  &_swigt__p_svn_opt_revision_value_t,
  &_swigt__p_svn_opt_subcommand_desc2_t,
  &_swigt__p_svn_opt_subcommand_desc3_t,
  &_swigt__p_svn_opt_subcommand_desc_t,
  &_swigt__p_svn_prop_inherited_item_t,
  &_swigt__p_svn_prop_kind,
  &_swigt__p_svn_ra_callbacks2_t,
  &_swigt__p_svn_ra_callbacks_t,
  &_swigt__p_svn_ra_plugin_t,
  &_swigt__p_svn_ra_reporter2_t,
  &_swigt__p_svn_ra_reporter3_t,
  &_swigt__p_svn_ra_reporter_t,
  &_swigt__p_svn_ra_session_t,
  &_swigt__p_svn_stream_mark_t,
  &_swigt__p_svn_stream_t,
  &_swigt__p_svn_string_t,
  &_swigt__p_svn_stringbuf_t,
  &_swigt__p_svn_tristate_t,
  &_swigt__p_svn_txdelta_op_t,
  &_swigt__p_svn_txdelta_stream_t,
  &_swigt__p_svn_txdelta_window_t,
  &_swigt__p_svn_version_checklist_t,
  &_swigt__p_svn_version_ext_linked_lib_t,
  &_swigt__p_svn_version_ext_loaded_lib_t,
  &_swigt__p_svn_version_extended_t,
  &_swigt__p_svn_version_t,
  &_swigt__p_svn_wc_external_item2_t,
  &_swigt__p_unsigned_long,
  &_swigt__p_void,
};

static swig_cast_info _swigc__p_apr_array_header_t[] = {  {&_swigt__p_apr_array_header_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_apr_file_t[] = {  {&_swigt__p_apr_file_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_apr_hash_t[] = {  {&_swigt__p_apr_hash_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_apr_int32_t[] = {  {&_swigt__p_apr_int32_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_apr_int64_t[] = {  {&_swigt__p_apr_int64_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_apr_off_t[] = {  {&_swigt__p_apr_off_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_apr_pool_t[] = {  {&_swigt__p_apr_pool_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_char[] = {  {&_swigt__p_char, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_apr_off_t_apr_off_t_p_void_p_apr_pool_t__void[] = {  {&_swigt__p_f_apr_off_t_apr_off_t_p_void_p_apr_pool_t__void, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_int_p_apr_pool_t_p_apr_hash_t__p_svn_error_t[] = {  {&_swigt__p_f_int_p_apr_pool_t_p_apr_hash_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_apr_getopt_t_p_void_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_apr_getopt_t_p_void_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_p_apr_file_t_p_void_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_p_apr_file_t_p_void_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_p_svn_stream_t_p_p_svn_stream_t_p_svn_ra_close_tunnel_func_t_p_p_void_p_void_p_q_const__char_p_q_const__char_p_q_const__char_int_svn_cancel_func_t_p_void_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_p_svn_stream_t_p_p_svn_stream_t_p_svn_ra_close_tunnel_func_t_p_p_void_p_void_p_q_const__char_p_q_const__char_p_q_const__char_int_svn_cancel_func_t_p_void_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_p_void_p_q_const__char_p_q_const__svn_ra_callbacks_t_p_void_p_apr_hash_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_p_void_p_q_const__char_p_q_const__svn_ra_callbacks_t_p_void_p_apr_hash_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_q_const__char_p_svn_dirent_t_p_void_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_q_const__char_p_svn_dirent_t_p_void_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_q_const__svn_commit_info_t_p_void_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_q_const__svn_commit_info_t_p_void_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_svn_location_segment_t_p_void_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_svn_location_segment_t_p_void_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void__p_svn_error_t[] = {  {&_swigt__p_f_p_void__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_apr_hash_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_apr_hash_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_p_apr_hash_t_p_q_const__char_svn_revnum_t_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_p_apr_hash_t_p_q_const__char_svn_revnum_t_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_p_q_const__char_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_p_q_const__char_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_p_q_const__svn_delta_editor_t_p_p_void_p_q_const__char_svn_commit_callback_t_p_void_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_p_q_const__svn_delta_editor_t_p_p_void_p_q_const__char_svn_commit_callback_t_p_void_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_p_q_const__svn_ra_reporter_t_p_p_void_p_q_const__char_svn_revnum_t_svn_boolean_t_p_q_const__svn_delta_editor_t_p_void_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_p_q_const__svn_ra_reporter_t_p_p_void_p_q_const__char_svn_revnum_t_svn_boolean_t_p_q_const__svn_delta_editor_t_p_void_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_p_q_const__svn_ra_reporter_t_p_p_void_svn_revnum_t_p_q_const__char_svn_boolean_t_p_q_const__char_p_q_const__svn_delta_editor_t_p_void_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_p_q_const__svn_ra_reporter_t_p_p_void_svn_revnum_t_p_q_const__char_svn_boolean_t_p_q_const__char_p_q_const__svn_delta_editor_t_p_void_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_p_q_const__svn_ra_reporter_t_p_p_void_svn_revnum_t_p_q_const__char_svn_boolean_t_p_q_const__svn_delta_editor_t_p_void_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_p_q_const__svn_ra_reporter_t_p_p_void_svn_revnum_t_p_q_const__char_svn_boolean_t_p_q_const__svn_delta_editor_t_p_void_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_p_q_const__svn_ra_reporter_t_p_p_void_svn_revnum_t_p_q_const__char_svn_boolean_t_svn_boolean_t_p_q_const__char_p_q_const__svn_delta_editor_t_p_void_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_p_q_const__svn_ra_reporter_t_p_p_void_svn_revnum_t_p_q_const__char_svn_boolean_t_svn_boolean_t_p_q_const__char_p_q_const__svn_delta_editor_t_p_void_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_p_svn_stream_t_p_q_const__svn_checksum_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_p_svn_stream_t_p_q_const__svn_checksum_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_q_const__apr_array_header_t_svn_revnum_t_svn_revnum_t_svn_boolean_t_svn_boolean_t_svn_log_message_receiver_t_p_void_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_q_const__apr_array_header_t_svn_revnum_t_svn_revnum_t_svn_boolean_t_svn_boolean_t_svn_log_message_receiver_t_p_void_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_q_const__char__int[] = {  {&_swigt__p_f_p_void_p_q_const__char__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_q_const__char_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_q_const__char_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_q_const__char_p_q_const__char_p_p_q_const__svn_string_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_q_const__char_p_q_const__char_p_p_q_const__svn_string_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_q_const__char_p_q_const__char_p_q_const__svn_string_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_q_const__char_p_q_const__char_p_q_const__svn_string_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_q_const__char_p_q_const__char_svn_revnum_t_svn_boolean_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_q_const__char_p_q_const__char_svn_revnum_t_svn_boolean_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_q_const__char_p_q_const__char_svn_revnum_t_svn_boolean_t_p_q_const__char_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_q_const__char_p_q_const__char_svn_revnum_t_svn_boolean_t_p_q_const__char_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_q_const__char_p_q_const__char_svn_revnum_t_svn_depth_t_svn_boolean_t_p_q_const__char_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_q_const__char_p_q_const__char_svn_revnum_t_svn_depth_t_svn_boolean_t_p_q_const__char_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_q_const__char_svn_boolean_t_p_q_const__svn_lock_t_p_svn_error_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_q_const__char_svn_boolean_t_p_q_const__svn_lock_t_p_svn_error_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_q_const__char_svn_revnum_t_p_apr_hash_t_p_svn_txdelta_window_handler_t_p_p_void_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_q_const__char_svn_revnum_t_p_apr_hash_t_p_svn_txdelta_window_handler_t_p_p_void_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_q_const__char_svn_revnum_t_p_apr_hash_t_svn_boolean_t_p_svn_txdelta_window_handler_t_p_p_void_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_q_const__char_svn_revnum_t_p_apr_hash_t_svn_boolean_t_p_svn_txdelta_window_handler_t_p_p_void_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_q_const__char_svn_revnum_t_p_p_apr_hash_t_p_svn_revnum_t_p_p_apr_hash_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_q_const__char_svn_revnum_t_p_p_apr_hash_t_p_svn_revnum_t_p_p_apr_hash_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_q_const__char_svn_revnum_t_p_svn_node_kind_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_q_const__char_svn_revnum_t_p_svn_node_kind_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_q_const__char_svn_revnum_t_p_svn_stream_t_p_svn_revnum_t_p_p_apr_hash_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_q_const__char_svn_revnum_t_p_svn_stream_t_p_svn_revnum_t_p_p_apr_hash_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_q_const__char_svn_revnum_t_svn_boolean_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_q_const__char_svn_revnum_t_svn_boolean_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_q_const__char_svn_revnum_t_svn_boolean_t_p_q_const__char_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_q_const__char_svn_revnum_t_svn_boolean_t_p_q_const__char_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_q_const__char_svn_revnum_t_svn_depth_t_svn_boolean_t_p_q_const__char_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_q_const__char_svn_revnum_t_svn_depth_t_svn_boolean_t_p_q_const__char_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_q_const__char_svn_revnum_t_svn_revnum_t_svn_ra_file_rev_handler_t_p_void_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_q_const__char_svn_revnum_t_svn_revnum_t_svn_ra_file_rev_handler_t_p_void_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_svn_log_entry_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_svn_log_entry_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_svn_revnum_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_svn_revnum_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_svn_revnum_t_apr_time_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_svn_revnum_t_apr_time_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_svn_revnum_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_svn_revnum_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_void__void[] = {  {&_swigt__p_f_p_void_p_void__void, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_svn_revnum_t_p_p_apr_hash_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_svn_revnum_t_p_p_apr_hash_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_svn_revnum_t_p_q_const__char_p_p_svn_string_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_svn_revnum_t_p_q_const__char_p_p_svn_string_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_svn_revnum_t_p_q_const__char_p_q_const__svn_string_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_svn_revnum_t_p_q_const__char_p_q_const__svn_string_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_svn_revnum_t_p_q_const__char_p_q_const__char_p_void__p_svn_error_t[] = {  {&_swigt__p_f_svn_revnum_t_p_q_const__char_p_q_const__char_p_void__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_svn_revnum_t_p_void_p_p_q_const__svn_delta_editor_t_p_p_void_p_apr_hash_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_svn_revnum_t_p_void_p_p_q_const__svn_delta_editor_t_p_p_void_p_apr_hash_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_svn_revnum_t_p_void_p_q_const__svn_delta_editor_t_p_void_p_apr_hash_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_svn_revnum_t_p_void_p_q_const__svn_delta_editor_t_p_void_p_apr_hash_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_void__p_svn_version_t[] = {  {&_swigt__p_f_void__p_svn_version_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_int[] = {  {&_swigt__p_int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_long[] = {  {&_swigt__p_long, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_apr_array_header_t[] = {  {&_swigt__p_p_apr_array_header_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_apr_file_t[] = {  {&_swigt__p_p_apr_file_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_apr_hash_t[] = {  {&_swigt__p_p_apr_hash_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_char[] = {  {&_swigt__p_p_char, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_f_apr_off_t_apr_off_t_p_void_p_apr_pool_t__void[] = {  {&_swigt__p_p_f_apr_off_t_apr_off_t_p_void_p_apr_pool_t__void, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_f_int_p_apr_pool_t_p_apr_hash_t__p_svn_error_t[] = {  {&_swigt__p_p_f_int_p_apr_pool_t_p_apr_hash_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_f_p_p_svn_stream_t_p_p_svn_stream_t_p_svn_ra_close_tunnel_func_t_p_p_void_p_void_p_q_const__char_p_q_const__char_p_q_const__char_int_svn_cancel_func_t_p_void_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_p_f_p_p_svn_stream_t_p_p_svn_stream_t_p_svn_ra_close_tunnel_func_t_p_p_void_p_void_p_q_const__char_p_q_const__char_p_q_const__char_int_svn_cancel_func_t_p_void_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_f_p_svn_txdelta_window_t_p_void__p_svn_error_t[] = {  {&_swigt__p_p_f_p_svn_txdelta_window_t_p_void__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_f_p_void__p_svn_error_t[] = {  {&_swigt__p_p_f_p_void__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_f_p_void_p_p_q_const__char_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_p_f_p_void_p_p_q_const__char_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_f_p_void_p_p_svn_stream_t_p_q_const__svn_checksum_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_p_f_p_void_p_p_svn_stream_t_p_q_const__svn_checksum_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_f_p_void_p_q_const__char__int[] = {  {&_swigt__p_p_f_p_void_p_q_const__char__int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_f_p_void_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_p_f_p_void_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_f_p_void_p_q_const__char_p_q_const__char_p_p_q_const__svn_string_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_p_f_p_void_p_q_const__char_p_q_const__char_p_p_q_const__svn_string_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_f_p_void_p_q_const__char_p_q_const__char_p_q_const__svn_string_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_p_f_p_void_p_q_const__char_p_q_const__char_p_q_const__svn_string_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_f_p_void_p_q_const__char_svn_boolean_t_p_q_const__svn_lock_t_p_svn_error_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_p_f_p_void_p_q_const__char_svn_boolean_t_p_q_const__svn_lock_t_p_svn_error_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_f_p_void_p_q_const__char_svn_revnum_t_p_apr_hash_t_p_svn_txdelta_window_handler_t_p_p_void_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_p_f_p_void_p_q_const__char_svn_revnum_t_p_apr_hash_t_p_svn_txdelta_window_handler_t_p_p_void_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_f_p_void_p_q_const__char_svn_revnum_t_p_apr_hash_t_svn_boolean_t_p_svn_txdelta_window_handler_t_p_p_void_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_p_f_p_void_p_q_const__char_svn_revnum_t_p_apr_hash_t_svn_boolean_t_p_svn_txdelta_window_handler_t_p_p_void_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_f_p_void_p_svn_revnum_t__p_svn_error_t[] = {  {&_swigt__p_p_f_p_void_p_svn_revnum_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_f_p_void_p_void__void[] = {  {&_swigt__p_p_f_p_void_p_void__void, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_f_svn_revnum_t_p_void_p_p_q_const__svn_delta_editor_t_p_p_void_p_apr_hash_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_p_f_svn_revnum_t_p_void_p_p_q_const__svn_delta_editor_t_p_p_void_p_apr_hash_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_f_svn_revnum_t_p_void_p_q_const__svn_delta_editor_t_p_void_p_apr_hash_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_p_f_svn_revnum_t_p_void_p_q_const__svn_delta_editor_t_p_void_p_apr_hash_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_svn_delta_editor_t[] = {  {&_swigt__p_p_svn_delta_editor_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_svn_dirent_t[] = {  {&_swigt__p_p_svn_dirent_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_svn_lock_t[] = {  {&_swigt__p_p_svn_lock_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_svn_ra_callbacks2_t[] = {  {&_swigt__p_p_svn_ra_callbacks2_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_svn_ra_plugin_t[] = {  {&_swigt__p_p_svn_ra_plugin_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_svn_ra_reporter2_t[] = {  {&_swigt__p_p_svn_ra_reporter2_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_svn_ra_reporter3_t[] = {  {&_swigt__p_p_svn_ra_reporter3_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_svn_ra_reporter_t[] = {  {&_swigt__p_p_svn_ra_reporter_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_svn_ra_session_t[] = {  {&_swigt__p_p_svn_ra_session_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_svn_stream_t[] = {  {&_swigt__p_p_svn_stream_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_svn_string_t[] = {  {&_swigt__p_p_svn_string_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_svn_stringbuf_t[] = {  {&_swigt__p_p_svn_stringbuf_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_void[] = {  {&_swigt__p_p_void, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_auth_baton_t[] = {  {&_swigt__p_svn_auth_baton_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_auth_cred_simple_t[] = {  {&_swigt__p_svn_auth_cred_simple_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_auth_cred_ssl_client_cert_pw_t[] = {  {&_swigt__p_svn_auth_cred_ssl_client_cert_pw_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_auth_cred_ssl_client_cert_t[] = {  {&_swigt__p_svn_auth_cred_ssl_client_cert_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_auth_cred_ssl_server_trust_t[] = {  {&_swigt__p_svn_auth_cred_ssl_server_trust_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_auth_cred_username_t[] = {  {&_swigt__p_svn_auth_cred_username_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_auth_iterstate_t[] = {  {&_swigt__p_svn_auth_iterstate_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_auth_provider_object_t[] = {  {&_swigt__p_svn_auth_provider_object_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_auth_provider_t[] = {  {&_swigt__p_svn_auth_provider_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_auth_ssl_server_cert_info_t[] = {  {&_swigt__p_svn_auth_ssl_server_cert_info_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_cache_config_t[] = {  {&_swigt__p_svn_cache_config_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_checksum_ctx_t[] = {  {&_swigt__p_svn_checksum_ctx_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_checksum_kind_t[] = {  {&_swigt__p_svn_checksum_kind_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_checksum_t[] = {  {&_swigt__p_svn_checksum_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_client__shelf_version_t[] = {  {&_swigt__p_svn_client__shelf_version_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_client_conflict_option_t[] = {  {&_swigt__p_svn_client_conflict_option_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_commit_info_t[] = {  {&_swigt__p_svn_commit_info_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_config_t[] = {  {&_swigt__p_svn_config_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_delta_editor_t[] = {  {&_swigt__p_svn_delta_editor_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_delta_path_driver_state_t[] = {  {&_swigt__p_svn_delta_path_driver_state_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_depth_t[] = {  {&_swigt__p_svn_depth_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_diff_hunk_t[] = {  {&_swigt__p_svn_diff_hunk_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_dirent_t[] = {  {&_swigt__p_svn_dirent_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_errno_t[] = {  {&_swigt__p_svn_errno_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_error_t[] = {  {&_swigt__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_fs_lock_target_t[] = {  {&_swigt__p_svn_fs_lock_target_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_io_dirent2_t[] = {  {&_swigt__p_svn_io_dirent2_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_io_dirent_t[] = {  {&_swigt__p_svn_io_dirent_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_io_file_del_t[] = {  {&_swigt__p_svn_io_file_del_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_location_segment_t[] = {  {&_swigt__p_svn_location_segment_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_lock_t[] = {  {&_swigt__p_svn_lock_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_log_changed_path2_t[] = {  {&_swigt__p_svn_log_changed_path2_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_log_changed_path_t[] = {  {&_swigt__p_svn_log_changed_path_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_log_entry_t[] = {  {&_swigt__p_svn_log_entry_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_merge_range_t[] = {  {&_swigt__p_svn_merge_range_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_mergeinfo_inheritance_t[] = {  {&_swigt__p_svn_mergeinfo_inheritance_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_node_kind_t[] = {  {&_swigt__p_svn_node_kind_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_opt_revision_range_t[] = {  {&_swigt__p_svn_opt_revision_range_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_opt_revision_t[] = {  {&_swigt__p_svn_opt_revision_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_opt_revision_value_t[] = {  {&_swigt__p_svn_opt_revision_value_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_opt_subcommand_desc2_t[] = {  {&_swigt__p_svn_opt_subcommand_desc2_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_opt_subcommand_desc3_t[] = {  {&_swigt__p_svn_opt_subcommand_desc3_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_opt_subcommand_desc_t[] = {  {&_swigt__p_svn_opt_subcommand_desc_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_prop_inherited_item_t[] = {  {&_swigt__p_svn_prop_inherited_item_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_prop_kind[] = {  {&_swigt__p_svn_prop_kind, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_ra_callbacks2_t[] = {  {&_swigt__p_svn_ra_callbacks2_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_ra_callbacks_t[] = {  {&_swigt__p_svn_ra_callbacks_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_ra_plugin_t[] = {  {&_swigt__p_svn_ra_plugin_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_ra_reporter2_t[] = {  {&_swigt__p_svn_ra_reporter2_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_ra_reporter3_t[] = {  {&_swigt__p_svn_ra_reporter3_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_ra_reporter_t[] = {  {&_swigt__p_svn_ra_reporter_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_ra_session_t[] = {  {&_swigt__p_svn_ra_session_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_stream_mark_t[] = {  {&_swigt__p_svn_stream_mark_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_stream_t[] = {  {&_swigt__p_svn_stream_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_string_t[] = {  {&_swigt__p_svn_string_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_stringbuf_t[] = {  {&_swigt__p_svn_stringbuf_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_tristate_t[] = {  {&_swigt__p_svn_tristate_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_txdelta_op_t[] = {  {&_swigt__p_svn_txdelta_op_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_txdelta_stream_t[] = {  {&_swigt__p_svn_txdelta_stream_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_txdelta_window_t[] = {  {&_swigt__p_svn_txdelta_window_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_version_checklist_t[] = {  {&_swigt__p_svn_version_checklist_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_version_ext_linked_lib_t[] = {  {&_swigt__p_svn_version_ext_linked_lib_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_version_ext_loaded_lib_t[] = {  {&_swigt__p_svn_version_ext_loaded_lib_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_version_extended_t[] = {  {&_swigt__p_svn_version_extended_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_version_t[] = {  {&_swigt__p_svn_version_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_external_item2_t[] = {  {&_swigt__p_svn_wc_external_item2_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_unsigned_long[] = {  {&_swigt__p_unsigned_long, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_void[] = {  {&_swigt__p_void, 0, 0, 0},{0, 0, 0, 0}};

static swig_cast_info *swig_cast_initial[] = {
  _swigc__p_apr_array_header_t,
  _swigc__p_apr_file_t,
  _swigc__p_apr_hash_t,
  _swigc__p_apr_int32_t,
  _swigc__p_apr_int64_t,
  _swigc__p_apr_off_t,
  _swigc__p_apr_pool_t,
  _swigc__p_char,
  _swigc__p_f_apr_off_t_apr_off_t_p_void_p_apr_pool_t__void,
  _swigc__p_f_int_p_apr_pool_t_p_apr_hash_t__p_svn_error_t,
  _swigc__p_f_p_apr_getopt_t_p_void_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_p_apr_file_t_p_void_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_p_svn_stream_t_p_p_svn_stream_t_p_svn_ra_close_tunnel_func_t_p_p_void_p_void_p_q_const__char_p_q_const__char_p_q_const__char_int_svn_cancel_func_t_p_void_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_p_void_p_q_const__char_p_q_const__svn_ra_callbacks_t_p_void_p_apr_hash_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_q_const__char_p_svn_dirent_t_p_void_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_q_const__svn_commit_info_t_p_void_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_svn_location_segment_t_p_void_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void__p_svn_error_t,
  _swigc__p_f_p_void_p_apr_hash_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void_p_p_apr_hash_t_p_q_const__char_svn_revnum_t_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void_p_p_q_const__char_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void_p_p_q_const__svn_delta_editor_t_p_p_void_p_q_const__char_svn_commit_callback_t_p_void_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void_p_p_q_const__svn_ra_reporter_t_p_p_void_p_q_const__char_svn_revnum_t_svn_boolean_t_p_q_const__svn_delta_editor_t_p_void_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void_p_p_q_const__svn_ra_reporter_t_p_p_void_svn_revnum_t_p_q_const__char_svn_boolean_t_p_q_const__char_p_q_const__svn_delta_editor_t_p_void_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void_p_p_q_const__svn_ra_reporter_t_p_p_void_svn_revnum_t_p_q_const__char_svn_boolean_t_p_q_const__svn_delta_editor_t_p_void_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void_p_p_q_const__svn_ra_reporter_t_p_p_void_svn_revnum_t_p_q_const__char_svn_boolean_t_svn_boolean_t_p_q_const__char_p_q_const__svn_delta_editor_t_p_void_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void_p_p_svn_stream_t_p_q_const__svn_checksum_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void_p_q_const__apr_array_header_t_svn_revnum_t_svn_revnum_t_svn_boolean_t_svn_boolean_t_svn_log_message_receiver_t_p_void_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void_p_q_const__char__int,
  _swigc__p_f_p_void_p_q_const__char_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void_p_q_const__char_p_q_const__char_p_p_q_const__svn_string_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void_p_q_const__char_p_q_const__char_p_q_const__svn_string_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void_p_q_const__char_p_q_const__char_svn_revnum_t_svn_boolean_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void_p_q_const__char_p_q_const__char_svn_revnum_t_svn_boolean_t_p_q_const__char_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void_p_q_const__char_p_q_const__char_svn_revnum_t_svn_depth_t_svn_boolean_t_p_q_const__char_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void_p_q_const__char_svn_boolean_t_p_q_const__svn_lock_t_p_svn_error_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void_p_q_const__char_svn_revnum_t_p_apr_hash_t_p_svn_txdelta_window_handler_t_p_p_void_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void_p_q_const__char_svn_revnum_t_p_apr_hash_t_svn_boolean_t_p_svn_txdelta_window_handler_t_p_p_void_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void_p_q_const__char_svn_revnum_t_p_p_apr_hash_t_p_svn_revnum_t_p_p_apr_hash_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void_p_q_const__char_svn_revnum_t_p_svn_node_kind_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void_p_q_const__char_svn_revnum_t_p_svn_stream_t_p_svn_revnum_t_p_p_apr_hash_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void_p_q_const__char_svn_revnum_t_svn_boolean_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void_p_q_const__char_svn_revnum_t_svn_boolean_t_p_q_const__char_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void_p_q_const__char_svn_revnum_t_svn_depth_t_svn_boolean_t_p_q_const__char_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void_p_q_const__char_svn_revnum_t_svn_revnum_t_svn_ra_file_rev_handler_t_p_void_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void_p_svn_log_entry_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void_p_svn_revnum_t__p_svn_error_t,
  _swigc__p_f_p_void_p_svn_revnum_t_apr_time_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void_p_svn_revnum_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void_p_void__void,
  _swigc__p_f_p_void_svn_revnum_t_p_p_apr_hash_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void_svn_revnum_t_p_q_const__char_p_p_svn_string_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void_svn_revnum_t_p_q_const__char_p_q_const__svn_string_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_svn_revnum_t_p_q_const__char_p_q_const__char_p_void__p_svn_error_t,
  _swigc__p_f_svn_revnum_t_p_void_p_p_q_const__svn_delta_editor_t_p_p_void_p_apr_hash_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_svn_revnum_t_p_void_p_q_const__svn_delta_editor_t_p_void_p_apr_hash_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_void__p_svn_version_t,
  _swigc__p_int,
  _swigc__p_long,
  _swigc__p_p_apr_array_header_t,
  _swigc__p_p_apr_file_t,
  _swigc__p_p_apr_hash_t,
  _swigc__p_p_char,
  _swigc__p_p_f_apr_off_t_apr_off_t_p_void_p_apr_pool_t__void,
  _swigc__p_p_f_int_p_apr_pool_t_p_apr_hash_t__p_svn_error_t,
  _swigc__p_p_f_p_p_svn_stream_t_p_p_svn_stream_t_p_svn_ra_close_tunnel_func_t_p_p_void_p_void_p_q_const__char_p_q_const__char_p_q_const__char_int_svn_cancel_func_t_p_void_p_apr_pool_t__p_svn_error_t,
  _swigc__p_p_f_p_svn_txdelta_window_t_p_void__p_svn_error_t,
  _swigc__p_p_f_p_void__p_svn_error_t,
  _swigc__p_p_f_p_void_p_p_q_const__char_p_apr_pool_t__p_svn_error_t,
  _swigc__p_p_f_p_void_p_p_svn_stream_t_p_q_const__svn_checksum_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_p_f_p_void_p_q_const__char__int,
  _swigc__p_p_f_p_void_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t,
  _swigc__p_p_f_p_void_p_q_const__char_p_q_const__char_p_p_q_const__svn_string_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_p_f_p_void_p_q_const__char_p_q_const__char_p_q_const__svn_string_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_p_f_p_void_p_q_const__char_svn_boolean_t_p_q_const__svn_lock_t_p_svn_error_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_p_f_p_void_p_q_const__char_svn_revnum_t_p_apr_hash_t_p_svn_txdelta_window_handler_t_p_p_void_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_p_f_p_void_p_q_const__char_svn_revnum_t_p_apr_hash_t_svn_boolean_t_p_svn_txdelta_window_handler_t_p_p_void_p_apr_array_header_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_p_f_p_void_p_svn_revnum_t__p_svn_error_t,
  _swigc__p_p_f_p_void_p_void__void,
  _swigc__p_p_f_svn_revnum_t_p_void_p_p_q_const__svn_delta_editor_t_p_p_void_p_apr_hash_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_p_f_svn_revnum_t_p_void_p_q_const__svn_delta_editor_t_p_void_p_apr_hash_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_p_svn_delta_editor_t,
  _swigc__p_p_svn_dirent_t,
  _swigc__p_p_svn_lock_t,
  _swigc__p_p_svn_ra_callbacks2_t,
  _swigc__p_p_svn_ra_plugin_t,
  _swigc__p_p_svn_ra_reporter2_t,
  _swigc__p_p_svn_ra_reporter3_t,
  _swigc__p_p_svn_ra_reporter_t,
  _swigc__p_p_svn_ra_session_t,
  _swigc__p_p_svn_stream_t,
  _swigc__p_p_svn_string_t,
  _swigc__p_p_svn_stringbuf_t,
  _swigc__p_p_void,
  _swigc__p_svn_auth_baton_t,
  _swigc__p_svn_auth_cred_simple_t,
  _swigc__p_svn_auth_cred_ssl_client_cert_pw_t,
  _swigc__p_svn_auth_cred_ssl_client_cert_t,
  _swigc__p_svn_auth_cred_ssl_server_trust_t,
  _swigc__p_svn_auth_cred_username_t,
  _swigc__p_svn_auth_iterstate_t,
  _swigc__p_svn_auth_provider_object_t,
  _swigc__p_svn_auth_provider_t,
  _swigc__p_svn_auth_ssl_server_cert_info_t,
  _swigc__p_svn_cache_config_t,
  _swigc__p_svn_checksum_ctx_t,
  _swigc__p_svn_checksum_kind_t,
  _swigc__p_svn_checksum_t,
  _swigc__p_svn_client__shelf_version_t,
  _swigc__p_svn_client_conflict_option_t,
  _swigc__p_svn_commit_info_t,
  _swigc__p_svn_config_t,
  _swigc__p_svn_delta_editor_t,
  _swigc__p_svn_delta_path_driver_state_t,
  _swigc__p_svn_depth_t,
  _swigc__p_svn_diff_hunk_t,
  _swigc__p_svn_dirent_t,
  _swigc__p_svn_errno_t,
  _swigc__p_svn_error_t,
  _swigc__p_svn_fs_lock_target_t,
  _swigc__p_svn_io_dirent2_t,
  _swigc__p_svn_io_dirent_t,
  _swigc__p_svn_io_file_del_t,
  _swigc__p_svn_location_segment_t,
  _swigc__p_svn_lock_t,
  _swigc__p_svn_log_changed_path2_t,
  _swigc__p_svn_log_changed_path_t,
  _swigc__p_svn_log_entry_t,
  _swigc__p_svn_merge_range_t,
  _swigc__p_svn_mergeinfo_inheritance_t,
  _swigc__p_svn_node_kind_t,
  _swigc__p_svn_opt_revision_range_t,
  _swigc__p_svn_opt_revision_t,
  _swigc__p_svn_opt_revision_value_t,
  _swigc__p_svn_opt_subcommand_desc2_t,
  _swigc__p_svn_opt_subcommand_desc3_t,
  _swigc__p_svn_opt_subcommand_desc_t,
  _swigc__p_svn_prop_inherited_item_t,
  _swigc__p_svn_prop_kind,
  _swigc__p_svn_ra_callbacks2_t,
  _swigc__p_svn_ra_callbacks_t,
  _swigc__p_svn_ra_plugin_t,
  _swigc__p_svn_ra_reporter2_t,
  _swigc__p_svn_ra_reporter3_t,
  _swigc__p_svn_ra_reporter_t,
  _swigc__p_svn_ra_session_t,
  _swigc__p_svn_stream_mark_t,
  _swigc__p_svn_stream_t,
  _swigc__p_svn_string_t,
  _swigc__p_svn_stringbuf_t,
  _swigc__p_svn_tristate_t,
  _swigc__p_svn_txdelta_op_t,
  _swigc__p_svn_txdelta_stream_t,
  _swigc__p_svn_txdelta_window_t,
  _swigc__p_svn_version_checklist_t,
  _swigc__p_svn_version_ext_linked_lib_t,
  _swigc__p_svn_version_ext_loaded_lib_t,
  _swigc__p_svn_version_extended_t,
  _swigc__p_svn_version_t,
  _swigc__p_svn_wc_external_item2_t,
  _swigc__p_unsigned_long,
  _swigc__p_void,
};


/* -------- TYPE CONVERSION AND EQUIVALENCE RULES (END) -------- */

static swig_const_info swig_const_table[] = {
{0, 0, 0, 0.0, 0, 0}};

#ifdef __cplusplus
}
#endif
/* -----------------------------------------------------------------------------
 * Type initialization:
 * This problem is made tough by the requirement that no dynamic
 * memory is used. Also, since swig_type_info structures store pointers to
 * swig_cast_info structures and swig_cast_info structures store pointers back
 * to swig_type_info structures, we need some lookup code at initialization.
 * The idea is that swig generates all the structures that are needed.
 * The runtime then collects these partially filled structures.
 * The SWIG_InitializeModule function takes these initial arrays out of
 * swig_module, and does all the lookup, filling in the swig_module.types
 * array with the correct data and linking the correct swig_cast_info
 * structures together.
 *
 * The generated swig_type_info structures are assigned statically to an initial
 * array. We just loop through that array, and handle each type individually.
 * First we lookup if this type has been already loaded, and if so, use the
 * loaded structure instead of the generated one. Then we have to fill in the
 * cast dependencies. The cast data is initially stored in something like a
 * two-dimensional array. Each row corresponds to a type (there are the same
 * number of rows as there are in the swig_type_initial array). Each entry in
 * a column is one of the swig_cast_info structures for that type.
 * The cast_initial array is actually an array of arrays, because each row has
 * a variable number of columns.
 *
 * We loop through the cast[] array associated with the type and mark casts
 * which have not been defined in previously loaded modules by assigning
 * cast pointer value to cast->next. We also hash cast->type->name string
 * and store the value in the cast->value field. If we encounter swig_cast_info
 * structure that represents a cast to self we move it to the beginning
 * of the cast array. One trick we need to do is making sure the type pointer
 * in the swig_cast_info struct is correct.
 *
 * First off, we lookup the cast->type name to see if it is already loaded.
 * There are three cases to handle:
 *  1) If the cast->type has already been loaded AND the type we are adding
 *     casting info to has not been loaded (it is in this module), THEN we
 *     replace the cast->type pointer with the type pointer that has already
 *     been loaded.
 *  2) If BOTH types (the one we are adding casting info to, and the
 *     cast->type) are loaded, THEN the cast info has already been loaded by
 *     the previous module so we just ignore it.
 *  3) Finally, if cast->type has not already been loaded, then we add that
 *     swig_cast_info to the list (because the cast->type) pointer will
 *     be correct.
 *
 * Once the cast array has been set up AND it does have new casts that need
 * to be added we sort non-self cast entries to move filtered out entries
 * to the end of the array and to arrange the rest in the increasing order
 * of their type pointer values. We store the index of the last added entry
 * in the cast->value field of the entry[0] (overwriting the name hash). Then
 * we sort <next|value> fields of the remaining entries to arrange hash values
 * in the increasing order. This way cast->next->type->name field matches
 * the cast->value hash.
 *
 * Example:
 * Array of casts for type stored at 0x5000, cast to type stored at 0x3000
 * has already been loaded
 *
 *      After sweep-and-hash:   After sort-by-type:    After sort-by-hash:
 *        ________________       ________________       ________________
 *       |                |     |                |     |                |
 * Entry |  type = 0x5000 |     |  type = 0x5000 |     |  type = 0x5000 |
 *   0   |                |     |                |     |                |
 *       |  next = Entry0 |     |  next = Entry0 |     |  next = Entry0 |
 *       | value = 1212   |     | value = 3      |     | value = 3      |
 *       |                |     |                |     |                |
 *       |================|     |================|     |================|
 *       |                |     |                |     |                |
 * Entry |  type = 0x2000 |     |  type = 0x1000 |     |  type = 0x1000 |
 *   1   |                |     |                |     |                |
 *       |  next = Entry1 |     |  next = Entry1 |     |  next = Entry3 |
 *       | value = 3434   |     | value = 4545   |     | value = 2323   |
 *       |________________|     |________________|     |________________|
 *       |                |     |                |     |                |
 * Entry |  type = 0x3000 |     |  type = 0x2000 |     |  type = 0x2000 |
 *   2   |                |     |                |     |                |
 *       |  next = 0      |     |  next = Entry2 |     |  next = Entry2 |
 *       | value = 0      |     | value = 3434   |     | value = 3434   |
 *       |________________|     |________________|     |________________|
 *       |                |     |                |     |                |
 * Entry |  type = 0x1000 |     |  type = 0x4000 |     |  type = 0x4000 |
 *   3   |                |     |                |     |                |
 *       |  next = Entry3 |     |  next = Entry3 |     |  next = Entry1 |
 *       | value = 4545   |     | value = 2323   |     | value = 4545   |
 *       |________________|     |________________|     |________________|
 *       |                |     |                |     |                |
 * Entry |  type = 0x4000 |     |  type = 0x3000 |     |  type = 0x3000 |
 *   4   |                |     |                |     |                |
 *       |  next = Entry4 |     |  next = 0      |     |  next = 0      |
 *       | value = 2323   |     | value = 0      |     | value = 0      |
 *       |________________|     |________________|     |________________|
 *
 * Once the cast array has been initialized, we use cast[0]->next field to link
 * it into the list of cast arrays for the type.
 *              ____    ____    ____
 * type->cast->|next|->|next|->|next|->0
 *             |----|  |----|  |----|
 *             |----|  |----|  |----|
 *             |----|  |----|  |----|
 *
 * Subsequent cast resolution works as follows:
 *
 * 1. Check whether the type matches the first entry in the current cast array.
 * 2. If not, then do a binary search over the (0:cast->value] interval using
 *    either type address or the hash value of the type name.
 * 3. If not found, then move over to the next cast array (cast[0]->next).
 *
 * ----------------------------------------------------------------------------- */

#ifdef __cplusplus
extern "C" {
#if 0
} /* c-mode */
#endif
#endif

#if 0
#define SWIGRUNTIME_DEBUG
#endif

#ifndef SWIG_INIT_CLIENT_DATA_TYPE
#define SWIG_INIT_CLIENT_DATA_TYPE void *
#endif

/*
 * Sort function that puts cast entries with nonzero 'next' at the front
 * of the array while ordering them by addresses of their 'type' structs.
 */
SWIGINTERN int SWIG_CastCmpStruct(const void *pa, const void *pb) {
  swig_cast_info *pca = (swig_cast_info *)pa;
  swig_cast_info *pcb = (swig_cast_info *)pb;
  if (pca->type < pcb->type)
  return (pca->next || pcb->next == 0) ? -1 : 1;
  if (pca->type > pcb->type)
  return (pcb->next || pca->next == 0) ? 1 : -1;
  return 0;
}

/*
 * Shell-sort 'next' and 'value' field pairs to order them by 'value'.
 */
SWIGINTERN void SWIG_CastHashSort(swig_cast_info *cast, int size) {
  const int hmax = size/9;
  int h, i;
  for(h = 1; h <= hmax; h = 3*h+1);
  for(; h > 0; h /= 3)
  {
    for(i = h; i < size; ++i)
    {
      swig_cast_info *p = cast[i].next;
      unsigned int hash = cast[i].value;
      int j = i;
      while(j >= h && hash < cast[j-h].value)
      {
        cast[j].next = cast[j-h].next;
        cast[j].value = cast[j-h].value;
        j -= h;
      }
      cast[j].next = p;
      cast[j].value = hash;
    }
  }
}

SWIGRUNTIME void
SWIG_InitializeModule(SWIG_INIT_CLIENT_DATA_TYPE clientdata) {
  size_t i;
  swig_module_info *module_head, *iter;
  int init;
  
  /* check to see if the circular list has been setup, if not, set it up */
  if (swig_module.next==0) {
    /* Initialize the swig_module */
    swig_module.type_initial = swig_type_initial;
    swig_module.cast_initial = swig_cast_initial;
    swig_module.next = &swig_module;
    init = 1;
  } else {
    init = 0;
  }
  
  /* Try and load any already created modules */
  module_head = SWIG_GetModule(clientdata);
  if (!module_head) {
    /* This is the first module loaded for this interpreter */
    /* so set the swig module into the interpreter */
    SWIG_SetModule(clientdata, &swig_module);
  } else {
    /* the interpreter has loaded a SWIG module, but has it loaded this one? */
    iter=module_head;
    do {
      if (iter==&swig_module) {
        /* Our module is already in the list, so there's nothing more to do. */
        return;
      }
      iter=iter->next;
    } while (iter!= module_head);
    
    /* otherwise we must add our module into the list */
    swig_module.next = module_head->next;
    module_head->next = &swig_module;
  }
  
  /* When multiple interpreters are used, a module could have already been initialized in
       a different interpreter, but not yet have a pointer in this interpreter.
       In this case, we do not want to continue adding types... everything should be
       set up already */
  if (init == 0) return;
  
  /* Now work on filling in swig_module.types */
#ifdef SWIGRUNTIME_DEBUG
  printf("SWIG_InitializeModule: size %lu\n", (unsigned long)swig_module.size);
#endif
  for (i = 0; i < swig_module.size; ++i) {
    swig_type_info *type = 0;
    swig_type_info *target_type;
    swig_cast_info *cast, *first;
    int num_mapped = 0;
    
#ifdef SWIGRUNTIME_DEBUG
    printf("SWIG_InitializeModule: type %lu %s\n", (unsigned long)i, swig_module.type_initial[i]->name);
#endif
    
    /* if there is another module already loaded */
    if (swig_module.next != &swig_module) {
      type = SWIG_MangledTypeQueryModule(swig_module.next, &swig_module, swig_module.type_initial[i]->name);
    }
    if (type) {
      /* Overwrite clientdata field */
#ifdef SWIGRUNTIME_DEBUG
      printf("SWIG_InitializeModule: found type %s\n", type->name);
#endif
      if (swig_module.type_initial[i]->clientdata) {
        type->clientdata = swig_module.type_initial[i]->clientdata;
#ifdef SWIGRUNTIME_DEBUG
        printf("SWIG_InitializeModule: found and overwrite type %s \n", type->name);
#endif
      }
    } else {
      type = swig_module.type_initial[i];
    }
    
    /* Insert casting types */
    cast = first = swig_module.cast_initial[i];
    while (cast->type) {
      /* Don't need to add information already in the list */
      target_type = 0;
#ifdef SWIGRUNTIME_DEBUG
      printf("SWIG_InitializeModule: look cast %s\n", cast->type->name);
#endif
      if (swig_module.next != &swig_module) {
        target_type = SWIG_MangledTypeQueryModule(swig_module.next, &swig_module, cast->type->name);
        if (target_type) {
          /* Target type already defined in another module */
#ifdef SWIGRUNTIME_DEBUG
          printf("SWIG_InitializeModule: found cast %s\n", target_type->name);
#endif
          if (type == swig_module.type_initial[i]) {
#ifdef SWIGRUNTIME_DEBUG
            printf("SWIG_InitializeModule: skip old type %s\n", target_type->name);
#endif
            cast->type = target_type;
            target_type = 0;
          } else {
            /* Check if this cast is already in the list */
            swig_cast_info *ocast = SWIG_TypeCheck(target_type->name, type);
#ifdef SWIGRUNTIME_DEBUG
            if (ocast) printf("SWIG_InitializeModule: skip old cast %s\n", target_type->name);
#endif
            if (!ocast) target_type = 0;
          }
        }
      }
      
      if (!target_type) {
#ifdef SWIGRUNTIME_DEBUG
        printf("SWIG_InitializeModule: adding cast %s\n", cast->type->name);
#endif
        /* Set inclusion mark for sorting */
        cast->next = cast;
        num_mapped++;
        
        if (type == cast->type) {
#ifdef SWIGRUNTIME_DEBUG
          printf("%s : self cast at pos [%li]\n", type->name, cast - first);
#endif
          if (cast - first) {
            /* Move cast to itself to the first entry in the array */
            
            swig_cast_info tmp = *cast;
            *cast = *first;
            *first = tmp;
          }
          first++;
          
        } else {
          cast->value = SWIG_Hash(cast->type->name, (unsigned int)strlen(cast->type->name));
        }
      }
      cast++;
    }
    
    if (num_mapped) {
      if (cast - first) {
        swig_cast_info *tmp;
        
        /* Sort casts by type address for binary search in SWIG_TypeCheckStruct */
        qsort(first, cast - first, sizeof(swig_cast_info), SWIG_CastCmpStruct);
        
        /* Remap back links for added entries */
        cast = swig_module.cast_initial[i] + num_mapped;
        for (tmp = first; tmp < cast; tmp++) {
          tmp->next = tmp;
        }
      }
      
      /* Set the value field of the first entry to the index of the last added entry */
      cast = swig_module.cast_initial[i];
      cast->value = num_mapped - 1;
      
      num_mapped -= (int)(first - cast);
      if (num_mapped > 1) {
        /* Sort <'next','value'> pairs by 'value' for binary search in SWIG_TypeCheck */
        
        SWIG_CastHashSort(first, num_mapped);
      }
      
      first = type->cast;
      if (first) {
        /* Link the current set into the list of cast arrays */
        cast->next = first->next;
        first->next = cast;
      } else {
        cast->next = 0;
        type->cast = cast;
      }
    }
    
    /* Set entry in modules->types array equal to the type */
    swig_module.types[i] = type;
  }
  swig_module.types[i] = 0;
  
#ifdef SWIGRUNTIME_DEBUG
  printf("**** SWIG_InitializeModule: Cast List ******\n");
  for (i = 0; i < swig_module.size; ++i) {
    int j = 0;
    swig_cast_info *cast = swig_module.cast_initial[i];
    printf("SWIG_InitializeModule: type %lu %s\n", (unsigned long)i, swig_module.type_initial[i]->name);
    while (cast->type) {
      printf("SWIG_InitializeModule: cast type %s\n", cast->type->name);
      cast++;
      ++j;
    }
    printf("---- Total casts: %d\n",j);
  }
  printf("**** SWIG_InitializeModule: Cast List ******\n");
#endif
}

/* This function will propagate the clientdata field of type to
* any new swig_type_info structures that have been added into the list
* of equivalent types.  It is like calling
* SWIG_TypeClientData(type, clientdata) a second time.
*/
SWIGRUNTIME void
SWIG_PropagateClientData(void) {
  size_t i;
  static int init_run = 0;
  
  if (init_run) return;
  init_run = 1;
  
  for (i = 0; i < swig_module.size; i++) {
    if (swig_module.types[i]->clientdata) {
      swig_cast_info *head, *cast;
      head = swig_module.types[i]->cast;
      while (head) {
        for (cast = head; (cast - head) <= head->value; cast++) {
          if (!cast->converter) {
            if (cast->type && !cast->type->clientdata)
            SWIG_TypeClientData(cast->type, swig_module.types[i]->clientdata);
          }
        }
        head = head->next;
      }
    }
  }
}

#ifdef __cplusplus
#if 0
{
  /* c-mode */
#endif
}
#endif



#ifdef __cplusplus
extern "C" {
#endif
  
  /* -----------------------------------------------------------------------------
   * constants/methods manipulation
   * ----------------------------------------------------------------------------- */
  
  /* Install Constants */
  SWIGINTERN void
  SWIG_Python_InstallConstants(PyObject *d, swig_const_info constants[]) {
    PyObject *obj = 0;
    size_t i;
    for (i = 0; constants[i].type; ++i) {
      switch(constants[i].type) {
      case SWIG_PY_POINTER:
        obj = SWIG_InternalNewPointerObj(constants[i].pvalue, *(constants[i]).ptype,0);
        break;
      case SWIG_PY_BINARY:
        obj = SWIG_NewPackedObj(constants[i].pvalue, constants[i].lvalue, *(constants[i].ptype));
        break;
      default:
        obj = 0;
        break;
      }
      if (obj) {
        PyDict_SetItemString(d, constants[i].name, obj);
        SWIG_Py_DECREF(obj);
      }
    }
  }
  
  /* -----------------------------------------------------------------------------
   * Patch %callback methods' docstrings to hold the callback ptrs
   * -----------------------------------------------------------------------------*/
  
  SWIGINTERN void
  SWIG_Python_FixMethods(PyMethodDef *methods, const swig_const_info *const_table, swig_type_info **types, swig_type_info **types_initial) {
    size_t i;
    for (i = 0; methods[i].ml_name; ++i) {
      const char *c = methods[i].ml_doc;
      if (!c) continue;
      c = strstr(c, "swig_ptr: ");
      if (c) {
        int j;
        const swig_const_info *ci = 0;
        const char *name = c + 10;
        for (j = 0; const_table[j].type; ++j) {
          if (strncmp(const_table[j].name, name, 
              strlen(const_table[j].name)) == 0) {
            ci = &(const_table[j]);
            break;
          }
        }
        if (ci) {
          void *ptr = (ci->type == SWIG_PY_POINTER) ? ci->pvalue : 0;
          if (ptr) {
            size_t shift = (ci->ptype) - types;
            swig_type_info *ty = types_initial[shift];
            size_t ldoc = (c - methods[i].ml_doc);
            size_t lptr = strlen(ty->name)+2*sizeof(void*)+2;
            char *ndoc = (char*)malloc(ldoc + lptr + 10);
            if (ndoc) {
              char *buff = ndoc;
              memcpy(buff, methods[i].ml_doc, ldoc);
              buff += ldoc;
              memcpy(buff, "swig_ptr: ", 10);
              buff += 10;
              SWIG_PackVoidPtr(buff, ptr, ty->name, lptr);
              methods[i].ml_doc = ndoc;
            }
          }
        }
      }
    }
  } 
  
#ifdef __cplusplus
}
#endif




/* -----------------------------------------------------------------------------*
 *  Partial Init method
 * -----------------------------------------------------------------------------*/

SWIGINTERN int SWIG_mod_exec(PyObject *module);

#ifdef __cplusplus
extern "C"
#endif

SWIGEXPORT 
#if PY_VERSION_HEX >= 0x03000000
PyObject*
#else
void
#endif
SWIG_init(void) {
#if PY_VERSION_HEX >= 0x03000000
  static PyModuleDef_Slot SwigSlots[] = {
    {
      Py_mod_exec, (void *)SWIG_mod_exec 
    },
#ifdef SWIGPYTHON_NOGIL
#ifdef Py_GIL_DISABLED
    {
      Py_mod_gil, Py_MOD_GIL_NOT_USED 
    },
#endif
#endif
    {
      0, NULL 
    }
  };
  
  static struct PyModuleDef SWIG_module = {
    PyModuleDef_HEAD_INIT,
    SWIG_name,
    NULL,
    0,
    SwigMethods,
    SwigSlots,
    NULL,
    NULL,
    NULL
  };
  
  return PyModuleDef_Init(&SWIG_module);
#else
  PyObject *m = Py_InitModule(SWIG_name, SwigMethods);
  if (m && SWIG_mod_exec(m) != 0) {
    Py_DECREF(m);
  }
#endif
}

SWIGINTERN int SWIG_mod_exec(PyObject *m) {
  PyObject *d, *md, *globals;
  
#if defined(SWIGPYTHON_BUILTIN)
  static SwigPyClientData SwigPyObject_clientdata = {
    0, 0, 0, 0, 0, 0, 0
  };
  static PyGetSetDef this_getset_def = {
    (char *)"this", &SwigPyBuiltin_ThisClosure, NULL, NULL, NULL
  };
  static SwigPyGetSet thisown_getset_closure = {
    SwigPyObject_own,
    SwigPyObject_own
  };
  static PyGetSetDef thisown_getset_def = {
    (char *)"thisown", SwigPyBuiltin_GetterClosure, SwigPyBuiltin_SetterClosure, NULL, &thisown_getset_closure
  };
  PyTypeObject *builtin_pytype;
  int builtin_base_count;
  swig_type_info *builtin_basetype;
  PyObject *tuple;
  PyGetSetDescrObject *static_getset;
  PyTypeObject *metatype;
  PyTypeObject *swigpyobject;
  SwigPyClientData *cd;
  PyObject *public_interface, *public_symbol;
  PyObject *this_descr;
  PyObject *thisown_descr;
  PyObject *self = 0;
  int i;
  
  (void)builtin_pytype;
  (void)builtin_base_count;
  (void)builtin_basetype;
  (void)tuple;
  (void)static_getset;
  (void)self;
  
  /* Metaclass is used to implement static member variables */
  metatype = SwigPyObjectType_Type();
  assert(metatype);
  
  SwigPyStaticVar_Type();
#endif
  
  (void)globals;
  
  /* Create singletons now to avoid potential deadlocks with multi-threaded usage after module initialization */
  SWIG_runtime_data_module();
  SWIG_This();
  SWIG_Python_TypeCache();
  SwigPyPacked_Type();
  SwigVarLink_Type();
#ifndef SWIGPYTHON_BUILTIN
  SwigPyObject_Type();
#endif
  
  /* Fix SwigMethods to carry the callback ptrs when needed */
  SWIG_Python_FixMethods(SwigMethods, swig_const_table, swig_types, swig_type_initial);
  
#ifdef SWIGPYTHON_NOGIL
#ifdef Py_GIL_DISABLED
  PyUnstable_Module_SetGIL(m, Py_MOD_GIL_NOT_USED);
#endif
#endif
  
  md = d = PyModule_GetDict(m);
  (void)md;
  
  SWIG_InitializeModule(0);
  
#ifdef SWIGPYTHON_BUILTIN
  swigpyobject = SwigPyObject_TypeOnce();
  
  SwigPyObject_stype = SWIG_MangledTypeQuery("_p_SwigPyObject");
  assert(SwigPyObject_stype);
  cd = (SwigPyClientData*) SwigPyObject_stype->clientdata;
  if (!cd) {
    SwigPyObject_stype->clientdata = &SwigPyObject_clientdata;
    SwigPyObject_clientdata.pytype = swigpyobject;
  } else if (swigpyobject->tp_basicsize != cd->pytype->tp_basicsize) {
    PyErr_SetString(PyExc_RuntimeError, "Import error: attempted to load two incompatible swig-generated modules.");
    return -1;
  }
  
  /* All objects have a 'this' attribute */
  this_descr = PyDescr_NewGetSet(SwigPyObject_Type(), &this_getset_def);
  (void)this_descr;
  
  /* All objects have a 'thisown' attribute */
  thisown_descr = PyDescr_NewGetSet(SwigPyObject_Type(), &thisown_getset_def);
  (void)thisown_descr;
  
  public_interface = PyList_New(0);
  public_symbol = 0;
  (void)public_symbol;
  
  PyDict_SetItemString(md, "__all__", public_interface);
  SWIG_Py_DECREF(public_interface);
  for (i = 0; SwigMethods[i].ml_name != NULL; ++i)
  SwigPyBuiltin_AddPublicSymbol(public_interface, SwigMethods[i].ml_name);
  for (i = 0; swig_const_table[i].name != 0; ++i)
  SwigPyBuiltin_AddPublicSymbol(public_interface, swig_const_table[i].name);
#endif
  
  SWIG_InstallConstants(d,swig_const_table);
  
  SWIG_Python_SetConstant(d, "SVN_RA_CAPABILITY_DEPTH",SWIG_FromCharPtr("depth"));
  SWIG_Python_SetConstant(d, "SVN_RA_CAPABILITY_MERGEINFO",SWIG_FromCharPtr("mergeinfo"));
  SWIG_Python_SetConstant(d, "SVN_RA_CAPABILITY_LOG_REVPROPS",SWIG_FromCharPtr("log-revprops"));
  SWIG_Python_SetConstant(d, "SVN_RA_CAPABILITY_PARTIAL_REPLAY",SWIG_FromCharPtr("partial-replay"));
  SWIG_Python_SetConstant(d, "SVN_RA_CAPABILITY_COMMIT_REVPROPS",SWIG_FromCharPtr("commit-revprops"));
  SWIG_Python_SetConstant(d, "SVN_RA_CAPABILITY_ATOMIC_REVPROPS",SWIG_FromCharPtr("atomic-revprops"));
  SWIG_Python_SetConstant(d, "SVN_RA_CAPABILITY_INHERITED_PROPS",SWIG_FromCharPtr("inherited-props"));
  SWIG_Python_SetConstant(d, "SVN_RA_CAPABILITY_EPHEMERAL_TXNPROPS",SWIG_FromCharPtr("ephemeral-txnprops"));
  SWIG_Python_SetConstant(d, "SVN_RA_CAPABILITY_GET_FILE_REVS_REVERSE",SWIG_FromCharPtr("get-file-revs-reversed"));
  SWIG_Python_SetConstant(d, "SVN_RA_CAPABILITY_LIST",SWIG_FromCharPtr("list"));
  SWIG_Python_SetConstant(d, "SVN_RA_ABI_VERSION",SWIG_From_long((long)(2)));
  return 0;
}

