/* ----------------------------------------------------------------------------
 * This file was automatically generated by SWIG (https://www.swig.org).
 * Version 4.4.0
 *
 * Do not make changes to this file unless you know what you are doing - modify
 * the SWIG interface file instead.
 * ----------------------------------------------------------------------------- */


#define SWIG_VERSION 0x040400
#define SWIGRUBY

/* -----------------------------------------------------------------------------
 *  This section contains generic SWIG labels for method/variable
 *  declarations/attributes, and other compiler dependent labels.
 * ----------------------------------------------------------------------------- */

/* template workaround for compilers that cannot correctly implement the C++ standard */
#ifndef SWIGTEMPLATEDISAMBIGUATOR
# if defined(__SUNPRO_CC) && (__SUNPRO_CC <= 0x560)
#  define SWIGTEMPLATEDISAMBIGUATOR template
# elif defined(__HP_aCC)
/* Needed even with `aCC -AA' when `aCC -V' reports HP ANSI C++ B3910B A.03.55 */
/* If we find a maximum version that requires this, the test would be __HP_aCC <= 35500 for A.03.55 */
#  define SWIGTEMPLATEDISAMBIGUATOR template
# else
#  define SWIGTEMPLATEDISAMBIGUATOR
# endif
#endif

/* inline attribute */
#ifndef SWIGINLINE
# if defined(__cplusplus) || (defined(__GNUC__) && !defined(__STRICT_ANSI__))
#   define SWIGINLINE inline
# else
#   define SWIGINLINE
# endif
#endif

/* attribute recognised by some compilers to avoid 'unused' warnings */
#ifndef SWIGUNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define SWIGUNUSED __attribute__ ((__unused__))
#   else
#     define SWIGUNUSED
#   endif
# elif defined(__ICC)
#   define SWIGUNUSED __attribute__ ((__unused__))
# else
#   define SWIGUNUSED
# endif
#endif

#ifndef SWIG_MSC_UNSUPPRESS_4505
# if defined(_MSC_VER)
#   pragma warning(disable : 4505) /* unreferenced local function has been removed */
# endif
#endif

#ifndef SWIGUNUSEDPARM
# ifdef __cplusplus
#   define SWIGUNUSEDPARM(p)
# else
#   define SWIGUNUSEDPARM(p) p SWIGUNUSED
# endif
#endif

/* internal SWIG method */
#ifndef SWIGINTERN
# define SWIGINTERN static SWIGUNUSED
#endif

/* internal inline SWIG method */
#ifndef SWIGINTERNINLINE
# define SWIGINTERNINLINE SWIGINTERN SWIGINLINE
#endif

/* exporting methods */
#if defined(__GNUC__)
#  if (__GNUC__ >= 4) || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4)
#    ifndef GCC_HASCLASSVISIBILITY
#      define GCC_HASCLASSVISIBILITY
#    endif
#  endif
#endif

#ifndef SWIGEXPORT
# if defined(_WIN32) || defined(__WIN32__) || defined(__CYGWIN__)
#   if defined(STATIC_LINKED)
#     define SWIGEXPORT
#   else
#     define SWIGEXPORT __declspec(dllexport)
#   endif
# else
#   if defined(__GNUC__) && defined(GCC_HASCLASSVISIBILITY)
#     define SWIGEXPORT __attribute__ ((visibility("default")))
#   else
#     define SWIGEXPORT
#   endif
# endif
#endif

/* calling conventions for Windows */
#ifndef SWIGSTDCALL
# if defined(_WIN32) || defined(__WIN32__) || defined(__CYGWIN__)
#   define SWIGSTDCALL __stdcall
# else
#   define SWIGSTDCALL
# endif
#endif

/* Deal with Microsoft's attempt at deprecating C standard runtime functions */
#if !defined(SWIG_NO_CRT_SECURE_NO_DEPRECATE) && defined(_MSC_VER) && !defined(_CRT_SECURE_NO_DEPRECATE)
# define _CRT_SECURE_NO_DEPRECATE
#endif

/* Deal with Microsoft's attempt at deprecating methods in the standard C++ library */
#if !defined(SWIG_NO_SCL_SECURE_NO_DEPRECATE) && defined(_MSC_VER) && !defined(_SCL_SECURE_NO_DEPRECATE)
# define _SCL_SECURE_NO_DEPRECATE
#endif

/* Deal with Apple's deprecated 'AssertMacros.h' from Carbon-framework */
#if defined(__APPLE__) && !defined(__ASSERT_MACROS_DEFINE_VERSIONS_WITHOUT_UNDERSCORES)
# define __ASSERT_MACROS_DEFINE_VERSIONS_WITHOUT_UNDERSCORES 0
#endif

/* Intel's compiler complains if a variable which was never initialised is
 * cast to void, which is a common idiom which we use to indicate that we
 * are aware a variable isn't used.  So we just silence that warning.
 * See: https://github.com/swig/swig/issues/192 for more discussion.
 */
#ifdef __INTEL_COMPILER
# pragma warning disable 592
#endif

#if defined(__cplusplus) && __cplusplus >=201103L
# define SWIG_NOEXCEPT noexcept
#else
# define SWIG_NOEXCEPT throw()
#endif 

/* -----------------------------------------------------------------------------
 * swigcompat.swg
 *
 * Macros to provide support compatibility with older C and C++ standards.
 *
 * Note that SWIG expects __cplusplus to be defined to the appropriate C++ standard.
 * MSVC users are urged to check and examine the /Zc:__cplusplus compiler option.
 * See https://learn.microsoft.com/en-us/cpp/build/reference/zc-cplusplus.
 * ----------------------------------------------------------------------------- */

/* C99 and C++11 should provide snprintf, but define SWIG_NO_SNPRINTF
 * if you're missing it.
 */
#if ((defined __STDC_VERSION__ && __STDC_VERSION__ >= 199901L) || \
     (defined __cplusplus && __cplusplus >= 201103L) || \
     defined SWIG_HAVE_SNPRINTF) && \
    !defined SWIG_NO_SNPRINTF
# define SWIG_snprintf(O,S,F,A) snprintf(O,S,F,A)
# define SWIG_snprintf2(O,S,F,A,B) snprintf(O,S,F,A,B)
#else
/* Fallback versions ignore the buffer size, but most of our uses either have a
 * fixed maximum possible size or dynamically allocate a buffer that's large
 * enough.
 */
# define SWIG_snprintf(O,S,F,A) sprintf(O,F,A)
# define SWIG_snprintf2(O,S,F,A,B) sprintf(O,F,A,B)
#endif

/* -----------------------------------------------------------------------------
 *  This section contains generic SWIG labels for method/variable
 *  declarations/attributes, and other compiler dependent labels.
 * ----------------------------------------------------------------------------- */

/* template workaround for compilers that cannot correctly implement the C++ standard */
#ifndef SWIGTEMPLATEDISAMBIGUATOR
# if defined(__SUNPRO_CC) && (__SUNPRO_CC <= 0x560)
#  define SWIGTEMPLATEDISAMBIGUATOR template
# elif defined(__HP_aCC)
/* Needed even with `aCC -AA' when `aCC -V' reports HP ANSI C++ B3910B A.03.55 */
/* If we find a maximum version that requires this, the test would be __HP_aCC <= 35500 for A.03.55 */
#  define SWIGTEMPLATEDISAMBIGUATOR template
# else
#  define SWIGTEMPLATEDISAMBIGUATOR
# endif
#endif

/* inline attribute */
#ifndef SWIGINLINE
# if defined(__cplusplus) || (defined(__GNUC__) && !defined(__STRICT_ANSI__))
#   define SWIGINLINE inline
# else
#   define SWIGINLINE
# endif
#endif

/* attribute recognised by some compilers to avoid 'unused' warnings */
#ifndef SWIGUNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define SWIGUNUSED __attribute__ ((__unused__))
#   else
#     define SWIGUNUSED
#   endif
# elif defined(__ICC)
#   define SWIGUNUSED __attribute__ ((__unused__))
# else
#   define SWIGUNUSED
# endif
#endif

#ifndef SWIG_MSC_UNSUPPRESS_4505
# if defined(_MSC_VER)
#   pragma warning(disable : 4505) /* unreferenced local function has been removed */
# endif
#endif

#ifndef SWIGUNUSEDPARM
# ifdef __cplusplus
#   define SWIGUNUSEDPARM(p)
# else
#   define SWIGUNUSEDPARM(p) p SWIGUNUSED
# endif
#endif

/* internal SWIG method */
#ifndef SWIGINTERN
# define SWIGINTERN static SWIGUNUSED
#endif

/* internal inline SWIG method */
#ifndef SWIGINTERNINLINE
# define SWIGINTERNINLINE SWIGINTERN SWIGINLINE
#endif

/* exporting methods */
#if defined(__GNUC__)
#  if (__GNUC__ >= 4) || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4)
#    ifndef GCC_HASCLASSVISIBILITY
#      define GCC_HASCLASSVISIBILITY
#    endif
#  endif
#endif

#ifndef SWIGEXPORT
# if defined(_WIN32) || defined(__WIN32__) || defined(__CYGWIN__)
#   if defined(STATIC_LINKED)
#     define SWIGEXPORT
#   else
#     define SWIGEXPORT __declspec(dllexport)
#   endif
# else
#   if defined(__GNUC__) && defined(GCC_HASCLASSVISIBILITY)
#     define SWIGEXPORT __attribute__ ((visibility("default")))
#   else
#     define SWIGEXPORT
#   endif
# endif
#endif

/* calling conventions for Windows */
#ifndef SWIGSTDCALL
# if defined(_WIN32) || defined(__WIN32__) || defined(__CYGWIN__)
#   define SWIGSTDCALL __stdcall
# else
#   define SWIGSTDCALL
# endif
#endif

/* Deal with Microsoft's attempt at deprecating C standard runtime functions */
#if !defined(SWIG_NO_CRT_SECURE_NO_DEPRECATE) && defined(_MSC_VER) && !defined(_CRT_SECURE_NO_DEPRECATE)
# define _CRT_SECURE_NO_DEPRECATE
#endif

/* Deal with Microsoft's attempt at deprecating methods in the standard C++ library */
#if !defined(SWIG_NO_SCL_SECURE_NO_DEPRECATE) && defined(_MSC_VER) && !defined(_SCL_SECURE_NO_DEPRECATE)
# define _SCL_SECURE_NO_DEPRECATE
#endif

/* Deal with Apple's deprecated 'AssertMacros.h' from Carbon-framework */
#if defined(__APPLE__) && !defined(__ASSERT_MACROS_DEFINE_VERSIONS_WITHOUT_UNDERSCORES)
# define __ASSERT_MACROS_DEFINE_VERSIONS_WITHOUT_UNDERSCORES 0
#endif

/* Intel's compiler complains if a variable which was never initialised is
 * cast to void, which is a common idiom which we use to indicate that we
 * are aware a variable isn't used.  So we just silence that warning.
 * See: https://github.com/swig/swig/issues/192 for more discussion.
 */
#ifdef __INTEL_COMPILER
# pragma warning disable 592
#endif

#if defined(__cplusplus) && __cplusplus >=201103L
# define SWIG_NOEXCEPT noexcept
#else
# define SWIG_NOEXCEPT throw()
#endif 

/* -----------------------------------------------------------------------------
 * swigrun.swg
 *
 * This file contains generic C API SWIG runtime support for pointer
 * type checking.
 * ----------------------------------------------------------------------------- */

/* This should only be incremented when either the layout of swig_type_info changes,
   or for whatever reason, the runtime changes incompatibly */
#define SWIG_RUNTIME_VERSION "5"

/* define SWIG_TYPE_TABLE_NAME as "SWIG_TYPE_TABLE" */
#ifdef SWIG_TYPE_TABLE
# define SWIG_QUOTE_STRING(x) #x
# define SWIG_EXPAND_AND_QUOTE_STRING(x) SWIG_QUOTE_STRING(x)
# define SWIG_TYPE_TABLE_NAME SWIG_EXPAND_AND_QUOTE_STRING(SWIG_TYPE_TABLE)
#else
# define SWIG_TYPE_TABLE_NAME
#endif

/*
  You can use the SWIGRUNTIME and SWIGRUNTIMEINLINE macros for
  creating a static or dynamic library from the SWIG runtime code.
  In 99.9% of the cases, SWIG just needs to declare them as 'static'.

  But only do this if strictly necessary, ie, if you have problems
  with your compiler or suchlike.
*/

#ifndef SWIGRUNTIME
# define SWIGRUNTIME SWIGINTERN
#endif

#ifndef SWIGRUNTIMEINLINE
# define SWIGRUNTIMEINLINE SWIGRUNTIME SWIGINLINE
#endif

/*  Generic buffer size */
#ifndef SWIG_BUFFER_SIZE
# define SWIG_BUFFER_SIZE 1024
#endif

/* Flags for pointer conversions */
#define SWIG_POINTER_DISOWN        0x1
#define SWIG_CAST_NEW_MEMORY       0x2
#define SWIG_POINTER_NO_NULL       0x4
#define SWIG_POINTER_CLEAR         0x8
#define SWIG_POINTER_RELEASE       (SWIG_POINTER_CLEAR | SWIG_POINTER_DISOWN)

/* Flags for new pointer objects */
#define SWIG_POINTER_OWN           0x1


/*
   Flags/methods for returning states.

   The SWIG conversion methods, as ConvertPtr, return an integer
   that tells if the conversion was successful or not. And if not,
   an error code can be returned (see swigerrors.swg for the codes).

   Use the following macros/flags to set or process the returning
   states.

   In old versions of SWIG, code such as the following was usually written:

     if (SWIG_ConvertPtr(obj,vptr,ty.flags) != -1) {
       // success code
     } else {
       //fail code
     }

   Now you can be more explicit:

    int res = SWIG_ConvertPtr(obj,vptr,ty.flags);
    if (SWIG_IsOK(res)) {
      // success code
    } else {
      // fail code
    }

   which is the same really, but now you can also do

    Type *ptr;
    int res = SWIG_ConvertPtr(obj,(void **)(&ptr),ty.flags);
    if (SWIG_IsOK(res)) {
      // success code
      if (SWIG_IsNewObj(res) {
        ...
	delete *ptr;
      } else {
        ...
      }
    } else {
      // fail code
    }

   I.e., now SWIG_ConvertPtr can return new objects and you can
   identify the case and take care of the deallocation. Of course that
   also requires SWIG_ConvertPtr to return new result values, such as

      int SWIG_ConvertPtr(obj, ptr,...) {
        if (<obj is ok>) {
          if (<need new object>) {
            *ptr = <ptr to new allocated object>;
            return SWIG_NEWOBJ;
          } else {
            *ptr = <ptr to old object>;
            return SWIG_OLDOBJ;
          }
        } else {
          return SWIG_BADOBJ;
        }
      }

   Of course, returning the plain '0(success)/-1(fail)' still works, but you can be
   more explicit by returning SWIG_BADOBJ, SWIG_ERROR or any of the
   SWIG errors code.

   Finally, if the SWIG_CASTRANK_MODE is enabled, the result code
   allows returning the 'cast rank', for example, if you have this

       int food(double)
       int fooi(int);

   and you call

      food(1)   // cast rank '1'  (1 -> 1.0)
      fooi(1)   // cast rank '0'

   just use the SWIG_AddCast()/SWIG_CheckState()
*/

#define SWIG_OK                    (0)
/* Runtime errors are < 0 */
#define SWIG_ERROR                 (-1)
/* Errors in range -1 to -99 are in swigerrors.swg (errors for all languages including those not using the runtime) */
/* Errors in range -100 to -199 are language specific errors defined in *errors.swg */
/* Errors < -200 are generic runtime specific errors */
#define SWIG_ERROR_RELEASE_NOT_OWNED (-200)

#define SWIG_IsOK(r)               (r >= 0)
#define SWIG_ArgError(r)           ((r != SWIG_ERROR) ? r : SWIG_TypeError)

/* The CastRankLimit says how many bits are used for the cast rank */
#define SWIG_CASTRANKLIMIT         (1 << 8)
/* The NewMask denotes the object was created (using new/malloc) */
#define SWIG_NEWOBJMASK            (SWIG_CASTRANKLIMIT  << 1)
/* The TmpMask is for in/out typemaps that use temporary objects */
#define SWIG_TMPOBJMASK            (SWIG_NEWOBJMASK << 1)
/* Simple returning values */
#define SWIG_BADOBJ                (SWIG_ERROR)
#define SWIG_OLDOBJ                (SWIG_OK)
#define SWIG_NEWOBJ                (SWIG_OK | SWIG_NEWOBJMASK)
#define SWIG_TMPOBJ                (SWIG_OK | SWIG_TMPOBJMASK)
/* Check, add and del object mask methods */
#define SWIG_AddNewMask(r)         (SWIG_IsOK(r) ? (r | SWIG_NEWOBJMASK) : r)
#define SWIG_DelNewMask(r)         (SWIG_IsOK(r) ? (r & ~SWIG_NEWOBJMASK) : r)
#define SWIG_IsNewObj(r)           (SWIG_IsOK(r) && (r & SWIG_NEWOBJMASK))
#define SWIG_AddTmpMask(r)         (SWIG_IsOK(r) ? (r | SWIG_TMPOBJMASK) : r)
#define SWIG_DelTmpMask(r)         (SWIG_IsOK(r) ? (r & ~SWIG_TMPOBJMASK) : r)
#define SWIG_IsTmpObj(r)           (SWIG_IsOK(r) && (r & SWIG_TMPOBJMASK))

/* Cast-Rank Mode */
#if defined(SWIG_CASTRANK_MODE)
#  ifndef SWIG_TypeRank
#    define SWIG_TypeRank             unsigned long
#  endif
#  ifndef SWIG_MAXCASTRANK            /* Default cast allowed */
#    define SWIG_MAXCASTRANK          (2)
#  endif
#  define SWIG_CASTRANKMASK          ((SWIG_CASTRANKLIMIT) -1)
#  define SWIG_CastRank(r)           (r & SWIG_CASTRANKMASK)
SWIGINTERNINLINE int SWIG_AddCast(int r) {
  return SWIG_IsOK(r) ? ((SWIG_CastRank(r) < SWIG_MAXCASTRANK) ? (r + 1) : SWIG_ERROR) : r;
}
SWIGINTERNINLINE int SWIG_CheckState(int r) {
  return SWIG_IsOK(r) ? SWIG_CastRank(r) + 1 : 0;
}
#else /* no cast-rank mode */
#  define SWIG_AddCast(r) (r)
#  define SWIG_CheckState(r) (SWIG_IsOK(r) ? 1 : 0)
#endif


#include <string.h>

#ifdef __cplusplus
extern "C" {
#endif

typedef void *(*swig_converter_func)(void *, int *);
typedef struct swig_type_info *(*swig_dycast_func)(void **);

/* Structure to store information on one type */
typedef struct swig_type_info {
  const char             *name;			/* mangled name of this type */
  const char             *str;			/* human readable name of this type */
  swig_dycast_func        dcast;		/* dynamic cast function down a hierarchy */
  struct swig_cast_info  *cast;			/* linked list of types that can cast into this type */
  void                   *clientdata;		/* language specific type data */
  int                    owndata;		/* flag if the structure owns the clientdata */
} swig_type_info;

/* Structure to store a type and conversion function used for casting */
typedef struct swig_cast_info {
  swig_type_info         *type;			/* pointer to type that is equivalent to this type */
  swig_converter_func     converter;		/* function to cast the void pointers */
  struct swig_cast_info  *next;			/* pointer to next array of casts | pointer to cast hashed by value */
  unsigned int            value;		/* index of the last valid element in the array | typename hash value */
} swig_cast_info;

/* Structure used to store module information
 * Each module generates one structure like this, and the runtime collects
 * all of these structures and stores them in a circularly linked list.*/
typedef struct swig_module_info {
  swig_type_info         **types;		/* Array of pointers to swig_type_info structures that are in this module */
  size_t                 size;		        /* Number of types in this module */
  struct swig_module_info *next;		/* Pointer to next element in circularly linked list */
  swig_type_info         **type_initial;	/* Array of initially generated type structures */
  swig_cast_info         **cast_initial;	/* Array of initially generated casting structures */
  void                    *clientdata;		/* Language specific module data */
} swig_module_info;

/*
  Compare two type names skipping the space characters, therefore
  "char*" == "char *" and "Class<int>" == "Class<int >", etc.

  Return 0 when the two name types are equivalent, as in
  strncmp, but skipping ' '.
*/
SWIGRUNTIME int
SWIG_TypeNameComp(const char *f1, const char *l1,
		  const char *f2, const char *l2) {
  for (;(f1 != l1) && (f2 != l2); ++f1, ++f2) {
    while ((*f1 == ' ') && (f1 != l1)) ++f1;
    while ((*f2 == ' ') && (f2 != l2)) ++f2;
    if (*f1 != *f2) return (*f1 > *f2) ? 1 : -1;
  }
  return (int)((l1 - f1) - (l2 - f2));
}

/*
  Check type equivalence in a name list like <name1>|<name2>|...
  Return 0 if equal, -1 if nb < tb, 1 if nb > tb
*/
SWIGRUNTIME int
SWIG_TypeCmp(const char *nb, const char *tb) {
  int equiv = 1;
  const char* te = tb + strlen(tb);
  const char* ne = nb;
  while (equiv != 0 && *ne) {
    for (nb = ne; *ne; ++ne) {
      if (*ne == '|') break;
    }
    equiv = SWIG_TypeNameComp(nb, ne, tb, te);
    if (*ne) ++ne;
  }
  return equiv;
}

/*
  Check type equivalence in a name list like <name1>|<name2>|...
  Return 0 if not equal, 1 if equal
*/
SWIGRUNTIME int
SWIG_TypeEquiv(const char *nb, const char *tb) {
  return SWIG_TypeCmp(nb, tb) == 0 ? 1 : 0;
}

/*
 * Hash function for type name strings, based on maRushPrime1Hash (http://amsoftware.narod.ru/algo2.html)
 */
SWIGRUNTIME unsigned int SWIG_Hash(const char *str, unsigned int len) {
  const unsigned char *data = (const unsigned char *)str;
  unsigned int hash = len, i = 0, k;
  int rem = (int)len;

  while (rem >= (int)sizeof(unsigned int)) {
    k = *(unsigned int *)data;
    k += i++;
    hash ^= k;
    hash *= 171717;
    data += sizeof(unsigned int);
    rem -= sizeof(unsigned int);
  }

  switch (rem) {
    case 3: k = (unsigned int)(data[2]) << 16;
            k |= (unsigned int)(data[1]) << 8;
            k |= (unsigned int)(data[0]);
            k += i++;
            hash ^= k;
            hash *= 171717;
            break;
    case 2: k = (unsigned int)(data[1]) << 8;
            k |= (unsigned int)(data[0]);
            k += i++;
            hash ^= k;
            hash *= 171717;
            break;
    case 1: k = (unsigned int)(data[0]);
            k += i++;
            hash ^= k;
            hash *= 171717;
            break;
  }
  return hash;
}

/*
  Check the typename
*/
SWIGRUNTIME swig_cast_info *
SWIG_TypeCheck(const char *c, swig_type_info *ty) {
  static const unsigned int scan_threshold = 4;
  if (ty) {
    swig_cast_info *head = ty->cast;
    unsigned int hash_value = 0;
    int hashed = 0;

    while (head) {

      if (strcmp(head->type->name, c) == 0) {
        return head;
      }

      if (head->value) {
        swig_cast_info *iter;
        swig_cast_info *last = head + head->value;
        swig_cast_info *first = head + 1;
        int search = 1;

        if (!hashed) {
          if (head->value < scan_threshold) {
            for (iter = first; iter <= last; iter++) {
              if (strcmp(iter->type->name, c) == 0) {
                return iter;
              }
            }
            search = 0;
          } else {
            hashed = 1;
            hash_value = SWIG_Hash(c, (unsigned int)strlen(c));
          }
        }

        if (search) {
          /* Binary search over sorted <'next'|'value'> pairs */
          do {
            iter = first + ((last - first) >> 1);
            if (iter->value < hash_value) {
              first = iter + 1;
            } else if (iter->value == hash_value) {

              if (strcmp(iter->next->type->name, c) == 0) {
                return iter->next;
              }

              /* Hash collision check */
              for (last = iter + 1; last->next && last->value == hash_value; last++) {
                if (strcmp(last->next->type->name, c) == 0) {
                  return last->next;
                }
              }
              for (first = iter - 1; first != head && first->value == hash_value; first--) {
                if (strcmp(first->next->type->name, c) == 0) {
                  return first->next;
                }
              }
              break;
            } else
              last = iter - 1;
          } while (first <= last);
        }
      }
      head = head->next;
    }
  }
  return 0;
}

/*
  Check the type by type address
*/
SWIGRUNTIME swig_cast_info *
SWIG_TypeCheckStruct(const swig_type_info *from, swig_type_info *ty) {
  if (ty) {
    swig_cast_info *head = ty->cast;
    while (head) {
      if (head->type == from) {
        return head;
      }

      if (head->value) {
        swig_cast_info *iter;
        swig_cast_info *last = head + head->value;
        swig_cast_info *first = head + 1;

        /* Binary search over sorted array of casts */
        do {
          iter = first + ((last - first) >> 1);
          if (iter->type < from) {
            first = iter + 1;
          } else if (iter->type == from) {
            return iter;
          } else
            last = iter - 1;
        } while (first <= last);
      }
      head = head->next;
    }
  }
  return 0;
}

/*
  Cast a pointer up an inheritance hierarchy
*/
SWIGRUNTIMEINLINE void *
SWIG_TypeCast(swig_cast_info *ty, void *ptr, int *newmemory) {
  return ((!ty) || (!ty->converter)) ? ptr : (*ty->converter)(ptr, newmemory);
}

/*
   Dynamic pointer casting. Down an inheritance hierarchy
*/
SWIGRUNTIME swig_type_info *
SWIG_TypeDynamicCast(swig_type_info *ty, void **ptr) {
  swig_type_info *lastty = ty;
  if (!ty || !ty->dcast) return ty;
  while (ty && (ty->dcast)) {
    ty = (*ty->dcast)(ptr);
    if (ty) lastty = ty;
  }
  return lastty;
}

/*
  Return the name associated with this type
*/
SWIGRUNTIMEINLINE const char *
SWIG_TypeName(const swig_type_info *ty) {
  return ty->name;
}

/*
  Return the pretty name associated with this type,
  that is an unmangled type name in a form presentable to the user.
*/
SWIGRUNTIME const char *
SWIG_TypePrettyName(const swig_type_info *type) {
  /* The "str" field contains the equivalent pretty names of the
     type, separated by vertical-bar characters.  Choose the last
     name. It should be the most specific; a fully resolved name
     but not necessarily with default template parameters expanded. */
  if (!type) return NULL;
  if (type->str != NULL) {
    const char *last_name = type->str;
    const char *s;
    for (s = type->str; *s; s++)
      if (*s == '|') last_name = s+1;
    return last_name;
  }
  else
    return type->name;
}

/*
   Set the clientdata field for a type
*/
SWIGRUNTIME void
SWIG_TypeClientData(swig_type_info *ti, void *clientdata) {
  swig_cast_info *head = ti->cast;
  /* if (ti->clientdata == clientdata) return; */
  ti->clientdata = clientdata;

  while (head) {
    swig_cast_info *cast;
    for (cast = head; (cast - head) <= head->value; cast++) {
      if (!cast->converter) {
        swig_type_info *tc = cast->type;
        if (!tc->clientdata) {
          SWIG_TypeClientData(tc, clientdata);
        }
      }
    }
    head = head->next;
  }
}

SWIGRUNTIME void
SWIG_TypeNewClientData(swig_type_info *ti, void *clientdata) {
  SWIG_TypeClientData(ti, clientdata);
  ti->owndata = 1;
}

/*
  Search for a swig_type_info structure only by mangled name
  Search is a O(log #types)

  We start searching at module start, and finish searching when start == end.
  Note: if start == end at the beginning of the function, we go all the way around
  the circular list.
*/
SWIGRUNTIME swig_type_info *
SWIG_MangledTypeQueryModule(swig_module_info *start,
                            swig_module_info *end,
		            const char *name) {
  swig_module_info *iter = start;
  do {
    if (iter->size) {
      size_t l = 0;
      size_t r = iter->size - 1;
      do {
	/* since l+r >= 0, we can (>> 1) instead (/ 2) */
	size_t i = (l + r) >> 1;
	const char *iname = iter->types[i]->name;
	if (iname) {
	  int compare = strcmp(name, iname);
	  if (compare == 0) {
	    return iter->types[i];
	  } else if (compare < 0) {
	    if (i) {
	      r = i - 1;
	    } else {
	      break;
	    }
	  } else if (compare > 0) {
	    l = i + 1;
	  }
	} else {
	  break; /* should never happen */
	}
      } while (l <= r);
    }
    iter = iter->next;
  } while (iter != end);
  return 0;
}

/*
  Search for a swig_type_info structure for either a mangled name or a human readable name.
  It first searches the mangled names of the types, which is a O(log #types)
  If a type is not found it then searches the human readable names, which is O(#types).

  We start searching at module start, and finish searching when start == end.
  Note: if start == end at the beginning of the function, we go all the way around
  the circular list.
*/
SWIGRUNTIME swig_type_info *
SWIG_TypeQueryModule(swig_module_info *start,
                     swig_module_info *end,
		     const char *name) {
  /* STEP 1: Search the name field using binary search */
  swig_type_info *ret = SWIG_MangledTypeQueryModule(start, end, name);
  if (ret) {
    return ret;
  } else {
    /* STEP 2: If the type hasn't been found, do a complete search
       of the str field (the human readable name) */
    swig_module_info *iter = start;
    do {
      size_t i = 0;
      for (; i < iter->size; ++i) {
	if (iter->types[i]->str && (SWIG_TypeEquiv(iter->types[i]->str, name)))
	  return iter->types[i];
      }
      iter = iter->next;
    } while (iter != end);
  }

  /* neither found a match */
  return 0;
}

/*
   Pack binary data into a string
*/
SWIGRUNTIME char *
SWIG_PackData(char *c, void *ptr, size_t sz) {
  static const char hex[17] = "0123456789abcdef";
  const unsigned char *u = (unsigned char *) ptr;
  const unsigned char *eu =  u + sz;
  for (; u != eu; ++u) {
    unsigned char uu = *u;
    *(c++) = hex[(uu & 0xf0) >> 4];
    *(c++) = hex[uu & 0xf];
  }
  return c;
}

/*
   Unpack binary data from a string
*/
SWIGRUNTIME const char *
SWIG_UnpackData(const char *c, void *ptr, size_t sz) {
  unsigned char *u = (unsigned char *) ptr;
  const unsigned char *eu = u + sz;
  for (; u != eu; ++u) {
    char d = *(c++);
    unsigned char uu;
    if ((d >= '0') && (d <= '9'))
      uu = (unsigned char)((d - '0') << 4);
    else if ((d >= 'a') && (d <= 'f'))
      uu = (unsigned char)((d - ('a'-10)) << 4);
    else
      return (char *) 0;
    d = *(c++);
    if ((d >= '0') && (d <= '9'))
      uu |= (unsigned char)(d - '0');
    else if ((d >= 'a') && (d <= 'f'))
      uu |= (unsigned char)(d - ('a'-10));
    else
      return (char *) 0;
    *u = uu;
  }
  return c;
}

/*
   Pack 'void *' into a string buffer.
*/
SWIGRUNTIME char *
SWIG_PackVoidPtr(char *buff, void *ptr, const char *name, size_t bsz) {
  char *r = buff;
  if ((2*sizeof(void *) + 2) > bsz) return 0;
  *(r++) = '_';
  r = SWIG_PackData(r,&ptr,sizeof(void *));
  if (strlen(name) + 1 > (bsz - (r - buff))) return 0;
  strcpy(r,name);
  return buff;
}

SWIGRUNTIME const char *
SWIG_UnpackVoidPtr(const char *c, void **ptr, const char *name) {
  if (*c != '_') {
    if (strcmp(c,"NULL") == 0) {
      *ptr = (void *) 0;
      return name;
    } else {
      return 0;
    }
  }
  return SWIG_UnpackData(++c,ptr,sizeof(void *));
}

SWIGRUNTIME char *
SWIG_PackDataName(char *buff, void *ptr, size_t sz, const char *name, size_t bsz) {
  char *r = buff;
  size_t lname = (name ? strlen(name) : 0);
  if ((2*sz + 2 + lname) > bsz) return 0;
  *(r++) = '_';
  r = SWIG_PackData(r,ptr,sz);
  if (lname) {
    strncpy(r,name,lname+1);
  } else {
    *r = 0;
  }
  return buff;
}

SWIGRUNTIME const char *
SWIG_UnpackDataName(const char *c, void *ptr, size_t sz, const char *name) {
  if (*c != '_') {
    if (strcmp(c,"NULL") == 0) {
      memset(ptr,0,sz);
      return name;
    } else {
      return 0;
    }
  }
  return SWIG_UnpackData(++c,ptr,sz);
}

#ifdef __cplusplus
}
#endif

/* SWIG Errors applicable to all language modules, values are reserved from -1 to -99 */
#define  SWIG_UnknownError    	   -1
#define  SWIG_IOError        	   -2
#define  SWIG_RuntimeError   	   -3
#define  SWIG_IndexError     	   -4
#define  SWIG_TypeError      	   -5
#define  SWIG_DivisionByZero 	   -6
#define  SWIG_OverflowError  	   -7
#define  SWIG_SyntaxError    	   -8
#define  SWIG_ValueError     	   -9
#define  SWIG_SystemError    	   -10
#define  SWIG_AttributeError 	   -11
#define  SWIG_MemoryError    	   -12
#define  SWIG_NullReferenceError   -13


#if __GNUC__ >= 7
#pragma GCC diagnostic push
#if defined(__cplusplus)
#pragma GCC diagnostic ignored "-Wregister"
#if __GNUC__ >= 10
#pragma GCC diagnostic ignored "-Wvolatile"
#if __GNUC__ >= 11
#pragma GCC diagnostic ignored "-Wdeprecated-enum-enum-conversion"
#endif
#endif
#endif
#endif

#include <ruby.h>
#include <ruby/version.h> /* For RUBY_API_VERSION_CODE */

#if __GNUC__ >= 7
#pragma GCC diagnostic pop
#endif

/* Ruby 1.9.1 has a "memoisation optimisation" when compiling with GCC which
 * breaks using rb_intern as an lvalue, as SWIG does.  We work around this
 * issue for now by disabling this.
 * https://sourceforge.net/tracker/?func=detail&aid=2859614&group_id=1645&atid=101645
 */
#ifdef rb_intern
# undef rb_intern
#endif

/* Remove global macros defined in Ruby's win32.h */
#ifdef write
# undef write
#endif
#ifdef read
# undef read
#endif
#ifdef bind
# undef bind
#endif
#ifdef close
# undef close
#endif
#ifdef connect
# undef connect
#endif


/* Ruby 1.7 defines NUM2LL(), LL2NUM() and ULL2NUM() macros */
#ifndef NUM2LL
#define NUM2LL(x) NUM2LONG((x))
#endif
#ifndef LL2NUM
#define LL2NUM(x) INT2NUM((long) (x))
#endif
#ifndef ULL2NUM
#define ULL2NUM(x) UINT2NUM((unsigned long) (x))
#endif

/* Ruby 1.7 doesn't (yet) define NUM2ULL() */
#ifndef NUM2ULL
#ifdef HAVE_LONG_LONG
#define NUM2ULL(x) rb_num2ull((x))
#else
#define NUM2ULL(x) NUM2ULONG(x)
#endif
#endif

/* RSTRING_LEN, etc are new in Ruby 1.9, but ->ptr and ->len no longer work */
/* Define these for older versions so we can just write code the new way */
#ifndef RSTRING_LEN
# define RSTRING_LEN(x) RSTRING(x)->len
#endif
#ifndef RSTRING_PTR
# define RSTRING_PTR(x) RSTRING(x)->ptr
#endif
#ifndef RSTRING_END
# define RSTRING_END(x) (RSTRING_PTR(x) + RSTRING_LEN(x))
#endif
#ifndef RARRAY_LEN
# define RARRAY_LEN(x) RARRAY(x)->len
#endif
#ifndef RARRAY_PTR
# define RARRAY_PTR(x) RARRAY(x)->ptr
#endif
#ifndef RFLOAT_VALUE
# define RFLOAT_VALUE(x) RFLOAT(x)->value
#endif
#ifndef DOUBLE2NUM
# define DOUBLE2NUM(x) rb_float_new(x)
#endif
#ifndef RHASH_TBL
# define RHASH_TBL(x) (RHASH(x)->tbl)
#endif
#ifndef RHASH_ITER_LEV
# define RHASH_ITER_LEV(x) (RHASH(x)->iter_lev)
#endif
#ifndef RHASH_IFNONE
# define RHASH_IFNONE(x) (RHASH(x)->ifnone)
#endif
#ifndef RHASH_SIZE
# define RHASH_SIZE(x) (RHASH(x)->tbl->num_entries)
#endif
#ifndef RHASH_EMPTY_P
# define RHASH_EMPTY_P(x) (RHASH_SIZE(x) == 0)
#endif
#ifndef RSTRUCT_LEN
# define RSTRUCT_LEN(x) RSTRUCT(x)->len
#endif
#ifndef RSTRUCT_PTR
# define RSTRUCT_PTR(x) RSTRUCT(x)->ptr
#endif
#ifndef RTYPEDDATA_P
# define RTYPEDDATA_P(x) (TYPE(x) != T_DATA)
#endif



/*
 * The following macros are used for providing the correct type of a
 * function pointer to the Ruby C API.
 *
 * Starting with Ruby 2.7 these macros act transparently due to Ruby's moving
 * moving away from ANYARGS and instead employing strict function signatures.
 *
 * Note: In case of C (not C++) the macros are transparent even before
 * Ruby 2.7 due to the fact that the Ruby C API used function declarators
 * with empty parentheses, which allows for an unspecified number of
 * arguments.
 *
 * PROTECTFUNC(f) is used for the function pointer argument of the Ruby
 * C API function rb_protect().
 *
 * VALUEFUNC(f) is used for the function pointer argument(s) of Ruby C API
 * functions like rb_define_method() and rb_define_singleton_method().
 *
 * VOIDFUNC(f) is used to typecast a C function that implements either
 * the "mark" or "free" stuff for a Ruby Data object, so that it can be
 * passed as an argument to Ruby C API functions like Data_Wrap_Struct()
 * and Data_Make_Struct().
 *
 * SWIG_RUBY_VOID_ANYARGS_FUNC(f) is used for the function pointer
 * argument(s) of Ruby C API functions like rb_define_virtual_variable().
 *
 * SWIG_RUBY_INT_ANYARGS_FUNC(f) is used for the function pointer
 * argument(s) of Ruby C API functions like st_foreach().
 */
#if defined(__cplusplus) && RUBY_API_VERSION_CODE < 20700
#  define PROTECTFUNC(f) ((VALUE (*)(VALUE)) f)
#  define VALUEFUNC(f) ((VALUE (*)(ANYARGS)) f)
#  define VOIDFUNC(f) ((RUBY_DATA_FUNC) f)
#  define SWIG_RUBY_VOID_ANYARGS_FUNC(f) ((void (*)(ANYARGS))(f))
#  define SWIG_RUBY_INT_ANYARGS_FUNC(f) ((int (*)(ANYARGS))(f))
#else
#  define PROTECTFUNC(f) (f)
#  define VALUEFUNC(f) (f)
#  define VOIDFUNC(f) (f)
#  define SWIG_RUBY_VOID_ANYARGS_FUNC(f) (f)
#  define SWIG_RUBY_INT_ANYARGS_FUNC(f) (f)
#endif

/* Don't use for expressions have side effect */
#ifndef RB_STRING_VALUE
#define RB_STRING_VALUE(s) (TYPE(s) == T_STRING ? (s) : (*(volatile VALUE *)&(s) = rb_str_to_str(s)))
#endif
#ifndef StringValue
#define StringValue(s) RB_STRING_VALUE(s)
#endif
#ifndef StringValuePtr
#define StringValuePtr(s) RSTRING_PTR(RB_STRING_VALUE(s))
#endif
#ifndef StringValueLen
#define StringValueLen(s) RSTRING_LEN(RB_STRING_VALUE(s))
#endif
#ifndef SafeStringValue
#define SafeStringValue(v) do {\
    StringValue(v);\
    rb_check_safe_str(v);\
} while (0)
#endif

#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
#define rb_define_alloc_func(klass, func) rb_define_singleton_method((klass), "new", VALUEFUNC((func)), -1)
#define rb_undef_alloc_func(klass) rb_undef_method(CLASS_OF((klass)), "new")
#endif

static VALUE _mSWIG = Qnil;

/* -----------------------------------------------------------------------------
 * error manipulation
 * ----------------------------------------------------------------------------- */


/* Define some additional error types */
#define SWIG_ObjectPreviouslyDeletedError  -100


/* Define custom exceptions for errors that do not map to existing Ruby
   exceptions. */

SWIGINTERN VALUE 
getNullReferenceError(void) {
  static VALUE rb_eNullReferenceError = Qnil;
  if (rb_eNullReferenceError == Qnil) {
    rb_eNullReferenceError = rb_define_class("NullReferenceError", rb_eRuntimeError);
  }
  return rb_eNullReferenceError;
} 

SWIGINTERN VALUE 
getObjectPreviouslyDeletedError(void) {
  static VALUE rb_eObjectPreviouslyDeleted = Qnil;
  if (rb_eObjectPreviouslyDeleted == Qnil) {
    rb_eObjectPreviouslyDeleted = rb_define_class("ObjectPreviouslyDeleted", rb_eRuntimeError);
  }
  return rb_eObjectPreviouslyDeleted;
} 


SWIGINTERN VALUE
SWIG_Ruby_ErrorType(int SWIG_code) {
  VALUE type;
  switch (SWIG_code) {
  case SWIG_MemoryError:
    type = rb_eNoMemError;
    break;
  case SWIG_IOError:
    type = rb_eIOError;
    break;
  case SWIG_RuntimeError:
    type = rb_eRuntimeError;
    break;
  case SWIG_IndexError:
    type = rb_eIndexError;
    break;
  case SWIG_TypeError:
    type = rb_eTypeError;
    break;
  case SWIG_DivisionByZero:
    type = rb_eZeroDivError;
    break;
  case SWIG_OverflowError:
    type = rb_eRangeError;
    break;
  case SWIG_SyntaxError:
    type = rb_eSyntaxError;
    break;
  case SWIG_ValueError:
    type = rb_eArgError;
    break;
  case SWIG_SystemError:
    type = rb_eFatal;
    break;
  case SWIG_AttributeError:
    type = rb_eRuntimeError;
    break;
  case SWIG_NullReferenceError:
    type = getNullReferenceError();
    break;
  case SWIG_ObjectPreviouslyDeletedError:
    type = getObjectPreviouslyDeletedError();
    break;
  case SWIG_UnknownError:
    type = rb_eRuntimeError;
    break;
  default:
    type = rb_eRuntimeError;
  }
  return type;
}


/* This function is called when a user inputs a wrong argument to
   a method.
 */
SWIGINTERN 
const char* Ruby_Format_TypeError( const char* msg,
				   const char* type, 
				   const char* name, 
				   const int argn,
				   VALUE input )
{
  char buf[128];
  VALUE str;
  VALUE asStr;
  if ( msg && *msg )
    {
      str = rb_str_new2(msg);
    }
  else
    {
      str = rb_str_new(NULL, 0);
    }

  str = rb_str_cat2( str, "Expected argument " );
  SWIG_snprintf( buf, sizeof( buf), "%d of type ", argn-1 );
  str = rb_str_cat2( str, buf );
  str = rb_str_cat2( str, type );
  str = rb_str_cat2( str, ", but got " );
  str = rb_str_cat2( str, rb_obj_classname(input) );
  str = rb_str_cat2( str, " " );
  asStr = rb_inspect(input);
  if ( RSTRING_LEN(asStr) > 30 )
    {
      str = rb_str_cat( str, StringValuePtr(asStr), 30 );
      str = rb_str_cat2( str, "..." );
    }
  else
    {
      str = rb_str_append( str, asStr );
    }

  if ( name )
    {
      str = rb_str_cat2( str, "\n\tin SWIG method '" );
      str = rb_str_cat2( str, name );
      str = rb_str_cat2( str, "'" );
    }

  return StringValuePtr( str );
}

/* This function is called when an overloaded method fails */
SWIGINTERN 
void Ruby_Format_OverloadedError(
				 const int argc,
				 const int maxargs,
				 const char* method, 
				 const char* prototypes 
				 )
{
  const char* msg = "Wrong # of arguments";
  if ( argc <= maxargs ) msg = "Wrong arguments";
  rb_raise(rb_eArgError,"%s for overloaded method '%s'.\n"  
	   "Possible C/C++ prototypes are:\n%s",
	   msg, method, prototypes);
}

/* -----------------------------------------------------------------------------
 * rubytracking.swg
 *
 * This file contains support for tracking mappings from
 * Ruby objects to C++ objects.  This functionality is needed
 * to implement mark functions for Ruby's mark and sweep
 * garbage collector.
 * ----------------------------------------------------------------------------- */

#ifdef __cplusplus
extern "C" {
#endif

#if !defined(ST_DATA_T_DEFINED)
/* Needs to be explicitly included for Ruby 1.8 and earlier */
#include <st.h>
#endif

/* Ruby 1.8 actually assumes the first case. */
#if SIZEOF_VOIDP == SIZEOF_LONG
#  define SWIG2NUM(v) LONG2NUM((unsigned long)v)
#  define NUM2SWIG(x) (unsigned long)NUM2LONG(x)
#elif SIZEOF_VOIDP == SIZEOF_LONG_LONG
#  define SWIG2NUM(v) LL2NUM((unsigned long long)v)
#  define NUM2SWIG(x) (unsigned long long)NUM2LL(x)
#else
#  error sizeof(void*) is not the same as long or long long
#endif

/* Global hash table to store Trackings from C/C++
   structs to Ruby Objects.
*/
static st_table* swig_ruby_trackings = NULL;

static VALUE swig_ruby_trackings_count(ID id, VALUE *var) {
  return SWIG2NUM(swig_ruby_trackings->num_entries);
}


/* Setup a hash table to store Trackings */
SWIGRUNTIME void SWIG_RubyInitializeTrackings(void) {
  /* Create a hash table to store Trackings from C++
     objects to Ruby objects. */

  /* Try to see if some other .so has already created a
     tracking hash table, which we keep hidden in an instance var
     in the SWIG module.
     This is done to allow multiple DSOs to share the same
     tracking table.
  */
  VALUE trackings_value = Qnil;
  /* change the variable name so that we can mix modules
     compiled with older SWIG's - this used to be called "@__trackings__" */
  ID trackings_id = rb_intern( "@__safetrackings__" );
  VALUE verbose = rb_gv_get("VERBOSE");
  rb_gv_set("VERBOSE", Qfalse);
  trackings_value = rb_ivar_get( _mSWIG, trackings_id );
  rb_gv_set("VERBOSE", verbose);

  /* The trick here is that we have to store the hash table
  pointer in a Ruby variable. We do not want Ruby's GC to
  treat this pointer as a Ruby object, so we convert it to
  a Ruby numeric value. */
  if (trackings_value == Qnil) {
    /* No, it hasn't.  Create one ourselves */
    swig_ruby_trackings = st_init_numtable();
    rb_ivar_set( _mSWIG, trackings_id, SWIG2NUM(swig_ruby_trackings) );
  } else {
    swig_ruby_trackings = (st_table*)NUM2SWIG(trackings_value);
  }

  rb_define_virtual_variable("SWIG_TRACKINGS_COUNT",
                             VALUEFUNC(swig_ruby_trackings_count),
                             SWIG_RUBY_VOID_ANYARGS_FUNC((rb_gvar_setter_t*)NULL));
}

/* Add a Tracking from a C/C++ struct to a Ruby object */
SWIGRUNTIME void SWIG_RubyAddTracking(void* ptr, VALUE object) {
  /* Store the mapping to the global hash table. */
  st_insert(swig_ruby_trackings, (st_data_t)ptr, object);
}

/* Get the Ruby object that owns the specified C/C++ struct */
SWIGRUNTIME VALUE SWIG_RubyInstanceFor(void* ptr) {
  /* Now lookup the value stored in the global hash table */
  VALUE value;

  if (st_lookup(swig_ruby_trackings, (st_data_t)ptr, &value)) {
    return value;
  } else {
    return Qnil;
  }
}

/* Remove a Tracking from a C/C++ struct to a Ruby object.  It
   is very important to remove objects once they are destroyed
   since the same memory address may be reused later to create
   a new object. */
SWIGRUNTIME void SWIG_RubyRemoveTracking(void* ptr) {
  /* Delete the object from the hash table */
  st_delete(swig_ruby_trackings, (st_data_t *)&ptr, NULL);
}

/* This is a helper method that unlinks a Ruby object from its
   underlying C++ object.  This is needed if the lifetime of the
   Ruby object is longer than the C++ object. */
SWIGRUNTIME void SWIG_RubyUnlinkObjects(void* ptr) {
  VALUE object = SWIG_RubyInstanceFor(ptr);

  if (object != Qnil) {
    // object might have the T_ZOMBIE type, but that's just
    // because the GC has flagged it as such for a deferred
    // destruction. Until then, it's still a T_DATA object.
    DATA_PTR(object) = 0;
  }
}

/* This is a helper method that iterates over all the trackings
   passing the C++ object pointer and its related Ruby object
   to the passed callback function. */

/* Proxy method to abstract the internal trackings datatype */
static int swig_ruby_internal_iterate_callback(st_data_t ptr, st_data_t obj, st_data_t meth) {
  ((void (*) (void *, VALUE))meth)((void *)ptr, (VALUE)obj);
  return ST_CONTINUE;
}

SWIGRUNTIME void SWIG_RubyIterateTrackings( void(*meth)(void* ptr, VALUE obj) ) {
  st_foreach(swig_ruby_trackings,
             SWIG_RUBY_INT_ANYARGS_FUNC(swig_ruby_internal_iterate_callback),
             (st_data_t)meth);
}

#ifdef __cplusplus
}
#endif

/* -----------------------------------------------------------------------------
 * Ruby API portion that goes into the runtime
 * ----------------------------------------------------------------------------- */

#ifdef __cplusplus
extern "C" {
#endif

SWIGINTERN VALUE
SWIG_Ruby_AppendOutput(VALUE target, VALUE o, int is_void) {
  if (NIL_P(target) && is_void) {
    target = o;
  } else {
    if (TYPE(target) != T_ARRAY) {
      VALUE o2 = target;
      target = rb_ary_new();
      rb_ary_push(target, o2);
    }
    rb_ary_push(target, o);
  }
  return target;
}

/* For ruby1.8.4 and earlier. */
#ifndef RUBY_INIT_STACK
   RUBY_EXTERN void Init_stack(VALUE* addr);
#  define RUBY_INIT_STACK \
   VALUE variable_in_this_stack_frame; \
   Init_stack(&variable_in_this_stack_frame);
#endif


#ifdef __cplusplus
}
#endif


/* -----------------------------------------------------------------------------
 * rubyrun.swg
 *
 * This file contains the runtime support for Ruby modules
 * and includes code for managing global variables and pointer
 * type checking.
 * ----------------------------------------------------------------------------- */

/* For backward compatibility only */
#define SWIG_POINTER_EXCEPTION  0

/* for raw pointers */
#define SWIG_ConvertPtr(obj, pptr, type, flags)         SWIG_Ruby_ConvertPtrAndOwn(obj, pptr, type, flags, 0)
#define SWIG_ConvertPtrAndOwn(obj,pptr,type,flags,own)  SWIG_Ruby_ConvertPtrAndOwn(obj, pptr, type, flags, own)
#define SWIG_NewPointerObj(ptr, type, flags)            SWIG_Ruby_NewPointerObj(ptr, type, flags)
#define SWIG_AcquirePtr(ptr, own)                       SWIG_Ruby_AcquirePtr(ptr, own)
#define swig_owntype                                    swig_ruby_owntype

/* for raw packed data */
#define SWIG_ConvertPacked(obj, ptr, sz, ty)            SWIG_Ruby_ConvertPacked(obj, ptr, sz, ty, flags)
#define SWIG_NewPackedObj(ptr, sz, type)                SWIG_Ruby_NewPackedObj(ptr, sz, type)

/* for class or struct pointers */
#define SWIG_ConvertInstance(obj, pptr, type, flags)    SWIG_ConvertPtr(obj, pptr, type, flags)
#define SWIG_NewInstanceObj(ptr, type, flags)           SWIG_NewPointerObj(ptr, type, flags)

/* for C or C++ function pointers */
#define SWIG_ConvertFunctionPtr(obj, pptr, type)        SWIG_ConvertPtr(obj, pptr, type, 0)
#define SWIG_NewFunctionPtrObj(ptr, type)               SWIG_NewPointerObj(ptr, type, 0)

/* for C++ member pointers, ie, member methods */
#define SWIG_ConvertMember(obj, ptr, sz, ty)            SWIG_Ruby_ConvertPacked(obj, ptr, sz, ty)
#define SWIG_NewMemberObj(ptr, sz, type)                SWIG_Ruby_NewPackedObj(ptr, sz, type)


/* Runtime API */

#define SWIG_GetModule(clientdata)                      SWIG_Ruby_GetModule(clientdata)
#define SWIG_SetModule(clientdata, pointer) 		SWIG_Ruby_SetModule(pointer)


/* Error manipulation */

#define SWIG_ErrorType(code)                            SWIG_Ruby_ErrorType(code)               
#define SWIG_Error(code, msg)            		rb_raise(SWIG_Ruby_ErrorType(code), "%s", msg)
#define SWIG_fail                        		goto fail				 


/* Ruby-specific SWIG API */

#define SWIG_InitRuntime()                              SWIG_Ruby_InitRuntime()              
#define SWIG_define_class(ty)                        	SWIG_Ruby_define_class(ty)
#define SWIG_NewClassInstance(value, ty)             	SWIG_Ruby_NewClassInstance(value, ty)
#define SWIG_MangleStr(value)                        	SWIG_Ruby_MangleStr(value)		  
#define SWIG_CheckConvert(value, ty)                 	SWIG_Ruby_CheckConvert(value, ty)	  

#include "assert.h"

/* -----------------------------------------------------------------------------
 * pointers/data manipulation
 * ----------------------------------------------------------------------------- */

#ifdef __cplusplus
extern "C" {
#endif

typedef struct {
  VALUE klass;
  VALUE mImpl;
  void  (*mark)(void *);
  void  (*destroy)(void *);
  int trackObjects;
} swig_class;


/* Global pointer used to keep some internal SWIG stuff */
static VALUE _cSWIG_Pointer = Qnil;
static VALUE swig_runtime_data_type_pointer = Qnil;

/* Global IDs used to keep some internal SWIG stuff */
static ID swig_arity_id = 0;
static ID swig_call_id  = 0;
static ID swig_lowerthan_id = 0;

/*
  If your swig extension is to be run within an embedded ruby and has
  director callbacks, you should set -DRUBY_EMBEDDED during compilation.  
  This will reset ruby's stack frame on each entry point from the main 
  program the first time a virtual director function is invoked (in a 
  non-recursive way).
  If this is not done, you run the risk of Ruby trashing the stack.
*/

#ifdef RUBY_EMBEDDED

#  define SWIG_INIT_STACK                            \
      if ( !swig_virtual_calls ) { RUBY_INIT_STACK } \
      ++swig_virtual_calls;
#  define SWIG_RELEASE_STACK --swig_virtual_calls;
#  define Ruby_DirectorTypeMismatchException(x) \
          rb_raise( rb_eTypeError, "%s", x ); return c_result;

      static unsigned int swig_virtual_calls = 0;

#else  /* normal non-embedded extension */

#  define SWIG_INIT_STACK
#  define SWIG_RELEASE_STACK
#  define Ruby_DirectorTypeMismatchException(x) \
          throw Swig::DirectorTypeMismatchException( x );

#endif  /* RUBY_EMBEDDED */


SWIGRUNTIME VALUE 
getExceptionClass(void) {
  static VALUE rubyExceptionClass = Qnil;
  if (rubyExceptionClass == Qnil) {
    rubyExceptionClass = rb_const_get(_mSWIG, rb_intern("Exception"));
  }
  return rubyExceptionClass;
} 

/* This code checks to see if the Ruby object being raised as part
   of an exception inherits from the Ruby class Exception.  If so,
   the object is simply returned.  If not, then a new Ruby exception
   object is created and that will be returned to Ruby.*/
SWIGRUNTIME VALUE
SWIG_Ruby_ExceptionType(swig_type_info *desc, VALUE obj) {
  VALUE exceptionClass = getExceptionClass();
  if (rb_obj_is_kind_of(obj, exceptionClass)) {
    return obj;
  } else {
    return rb_exc_new3(rb_eRuntimeError, rb_obj_as_string(obj));
  }
}

/* Initialize Ruby runtime support */
SWIGRUNTIME void
SWIG_Ruby_InitRuntime(void)
{
  if (_mSWIG == Qnil) {
    _mSWIG = rb_define_module("SWIG");
    swig_call_id  = rb_intern("call");
    swig_arity_id = rb_intern("arity");
    swig_lowerthan_id = rb_intern("<");
  }
}

/* Define Ruby class for C type */
SWIGRUNTIME void
SWIG_Ruby_define_class(swig_type_info *type)
{
  size_t klass_len = 4 + strlen(type->name) + 1;
  char *klass_name = (char *) malloc(klass_len);
  SWIG_snprintf(klass_name, klass_len, "TYPE%s", type->name);
  if (NIL_P(_cSWIG_Pointer)) {
    _cSWIG_Pointer = rb_define_class_under(_mSWIG, "Pointer", rb_cObject);
    rb_undef_method(CLASS_OF(_cSWIG_Pointer), "new");
  }
  rb_undef_alloc_func(rb_define_class_under(_mSWIG, klass_name, _cSWIG_Pointer));
  free((void *) klass_name);
}

/* Create a new pointer object */
SWIGRUNTIME VALUE
SWIG_Ruby_NewPointerObj(void *ptr, swig_type_info *type, int flags)
{
  int own =  flags & SWIG_POINTER_OWN; 
  int track;
  char *klass_name;
  swig_class *sklass;
  VALUE klass;
  VALUE obj;

  if (!ptr)
    return Qnil;

  assert(type);
  if (type->clientdata) {
    sklass = (swig_class *) type->clientdata;
		
    /* Are we tracking this class and have we already returned this Ruby object? */
    track = sklass->trackObjects;
    if (track) {
      obj = SWIG_RubyInstanceFor(ptr);

      /* Check the object's type and make sure it has the correct type.
        It might not in cases where methods do things like 
        downcast methods. */
      if (obj != Qnil) {
        VALUE value = rb_iv_get(obj, "@__swigtype__");
        const char* type_name = RSTRING_PTR(value);
				
        if (strcmp(type->name, type_name) == 0) {
          return obj;
        }
      }
    }

    /* Create a new Ruby object */
    obj = Data_Wrap_Struct(sklass->klass, VOIDFUNC(sklass->mark), 
			   ( own ? VOIDFUNC(sklass->destroy) : 
			     (track ? VOIDFUNC(SWIG_RubyRemoveTracking) : 0 )
			     ), ptr);

    /* If tracking is on for this class then track this object. */
    if (track) {
      SWIG_RubyAddTracking(ptr, obj);
    }
  } else {
    size_t klass_len = 4 + strlen(type->name) + 1;
    klass_name = (char *) malloc(klass_len);
    SWIG_snprintf(klass_name, klass_len, "TYPE%s", type->name);
    klass = rb_const_get(_mSWIG, rb_intern(klass_name));
    free((void *) klass_name);
    obj = Data_Wrap_Struct(klass, 0, 0, ptr);
  }
  rb_iv_set(obj, "@__swigtype__", rb_str_new2(type->name));

  return obj;
}

/* Create a new class instance (always owned) */
SWIGRUNTIME VALUE
SWIG_Ruby_NewClassInstance(VALUE klass, swig_type_info *type)
{
  VALUE obj;
  swig_class *sklass = (swig_class *) type->clientdata;
  obj = Data_Wrap_Struct(klass, VOIDFUNC(sklass->mark), VOIDFUNC(sklass->destroy), 0);
  rb_iv_set(obj, "@__swigtype__", rb_str_new2(type->name));
  return obj;
}

/* Get type mangle from class name */
SWIGRUNTIMEINLINE char *
SWIG_Ruby_MangleStr(VALUE obj)
{
  VALUE stype = rb_iv_get(obj, "@__swigtype__");
  if (NIL_P(stype))
    return NULL;
  return StringValuePtr(stype);
}

/* Acquire a pointer value */
typedef struct {
  void (*datafree)(void *);
  int own;
} swig_ruby_owntype;

SWIGRUNTIME swig_ruby_owntype
SWIG_Ruby_AcquirePtr(VALUE obj, swig_ruby_owntype own) {
  swig_ruby_owntype oldown = {0, 0};
  if (TYPE(obj) == T_DATA && !RTYPEDDATA_P(obj)) {
    oldown.datafree = RDATA(obj)->dfree;
    RDATA(obj)->dfree = own.datafree;
  }
  return oldown;
}

/* Convert a pointer value */
SWIGRUNTIME int
SWIG_Ruby_ConvertPtrAndOwn(VALUE obj, void **ptr, swig_type_info *ty, int flags, swig_ruby_owntype *own)
{
  char *c;
  swig_cast_info *tc;
  void *vptr = 0;

  /* Grab the pointer */
  if (NIL_P(obj)) {
    if (ptr)
      *ptr = 0;
    return (flags & SWIG_POINTER_NO_NULL) ? SWIG_NullReferenceError : SWIG_OK;
  } else {
    if (TYPE(obj) != T_DATA || (TYPE(obj) == T_DATA && RTYPEDDATA_P(obj))) {
      return SWIG_ERROR;
    }
    Data_Get_Struct(obj, void, vptr);
  }
  
  if (own) {
    own->datafree = RDATA(obj)->dfree;
    own->own = 0;
  }
    
  if (((flags & SWIG_POINTER_RELEASE) == SWIG_POINTER_RELEASE)) {
    if (!RDATA(obj)->dfree)
      return SWIG_ERROR_RELEASE_NOT_OWNED;
  }

  /* Check to see if the input object is giving up ownership
     of the underlying C struct or C++ object.  If so then we
     need to reset the destructor since the Ruby object no 
     longer owns the underlying C++ object.*/ 
  if (flags & SWIG_POINTER_DISOWN) {
    /* Is tracking on for this class? */
    int track = 0;
    if (ty && ty->clientdata) {
      swig_class *sklass = (swig_class *) ty->clientdata;
      track = sklass->trackObjects;
    }

    if (track) {
      /* We are tracking objects for this class.  Thus we change the destructor
       * to SWIG_RubyRemoveTracking.  This allows us to
       * remove the mapping from the C++ to Ruby object
       * when the Ruby object is garbage collected.  If we don't
       * do this, then it is possible we will return a reference 
       * to a Ruby object that no longer exists thereby crashing Ruby. */
      RDATA(obj)->dfree = SWIG_RubyRemoveTracking;
    } else {    
      RDATA(obj)->dfree = 0;
    }
  }

  if (flags & SWIG_POINTER_CLEAR) {
    DATA_PTR(obj) = 0;
  }

  /* Do type-checking if type info was provided */
  if (ty) {
    if (ty->clientdata) {
      if (rb_obj_is_kind_of(obj, ((swig_class *) (ty->clientdata))->klass)) {
        if (vptr == 0) {
          /* The object has already been deleted */
          return SWIG_ObjectPreviouslyDeletedError;
        }
      }
    }
    if ((c = SWIG_MangleStr(obj)) == NULL) {
      return SWIG_ERROR;
    }
    tc = SWIG_TypeCheck(c, ty);
    if (!tc) {
      return SWIG_ERROR;
    } else {
      if (ptr) {
        if (tc->type == ty) {
          *ptr = vptr;
        } else {
          int newmemory = 0;
          *ptr = SWIG_TypeCast(tc, vptr, &newmemory);
          if (newmemory == SWIG_CAST_NEW_MEMORY) {
            assert(own); /* badly formed typemap which will lead to a memory leak - it must set and use own to delete *ptr */
            if (own)
              own->own = own->own | SWIG_CAST_NEW_MEMORY;
          }
        }
      }
    }
  } else {
    if (ptr)
      *ptr = vptr;
  }
  
  return SWIG_OK;
}

/* Check convert */
SWIGRUNTIMEINLINE int
SWIG_Ruby_CheckConvert(VALUE obj, swig_type_info *ty)
{
  char *c = SWIG_MangleStr(obj);
  if (!c) return 0;
  return SWIG_TypeCheck(c,ty) != 0;
}

SWIGRUNTIME VALUE
SWIG_Ruby_NewPackedObj(void *ptr, int sz, swig_type_info *type) {
  char result[1024];
  char *r = result;
  if ((2*sz + 1 + strlen(type->name)) > 1000) return 0;
  *(r++) = '_';
  r = SWIG_PackData(r, ptr, sz);
  strcpy(r, type->name);
  return rb_str_new2(result);
}

/* Convert a packed pointer value */
SWIGRUNTIME int
SWIG_Ruby_ConvertPacked(VALUE obj, void *ptr, int sz, swig_type_info *ty) {
  swig_cast_info *tc;
  const char  *c;

  if (TYPE(obj) != T_STRING) goto type_error;
  c = StringValuePtr(obj);
  /* Pointer values must start with leading underscore */
  if (*c != '_') goto type_error;
  c++;
  c = SWIG_UnpackData(c, ptr, sz);
  if (ty) {
    tc = SWIG_TypeCheck(c, ty);
    if (!tc) goto type_error;
  }
  return SWIG_OK;

 type_error:
  return SWIG_ERROR;
}

SWIGRUNTIME swig_module_info *
SWIG_Ruby_GetModule(void *SWIGUNUSEDPARM(clientdata))
{
  VALUE pointer;
  swig_module_info *ret = 0;
  VALUE verbose = rb_gv_get("VERBOSE");

 /* temporarily disable warnings, since the pointer check causes warnings with 'ruby -w' */
  rb_gv_set("VERBOSE", Qfalse);
  
  /* first check if pointer already created */
  pointer = rb_gv_get("$swig_runtime_data_type_pointer" SWIG_RUNTIME_VERSION SWIG_TYPE_TABLE_NAME);
  if (pointer != Qnil) {
    Data_Get_Struct(pointer, swig_module_info, ret);
  }

  /* reinstate warnings */
  rb_gv_set("VERBOSE", verbose);
  return ret;
}

SWIGRUNTIME void 
SWIG_Ruby_SetModule(swig_module_info *pointer)
{
  /* register a new class */
  VALUE cl = rb_define_class("swig_runtime_data", rb_cObject);
  rb_undef_alloc_func(cl);
  /* create and store the structure pointer to a global variable */
  swig_runtime_data_type_pointer = Data_Wrap_Struct(cl, 0, 0, pointer);
  rb_define_readonly_variable("$swig_runtime_data_type_pointer" SWIG_RUNTIME_VERSION SWIG_TYPE_TABLE_NAME, &swig_runtime_data_type_pointer);
}

/* This function can be used to check whether a proc or method or similarly
   callable function has been passed.  Usually used in a %typecheck, like:

   %typecheck(c_callback_t, precedence=SWIG_TYPECHECK_POINTER) {
        $result = SWIG_Ruby_isCallable( $input );
   }
 */
SWIGINTERN
int SWIG_Ruby_isCallable( VALUE proc )
{
  if ( rb_respond_to( proc, swig_call_id ) )
    return 1;
  return 0;
}

/* This function can be used to check the arity (number of arguments)
   a proc or method can take.  Usually used in a %typecheck.
   Valid arities will be that equal to minimal or those < 0
   which indicate a variable number of parameters at the end.
 */
SWIGINTERN
int SWIG_Ruby_arity( VALUE proc, int minimal )
{
  if ( rb_respond_to( proc, swig_arity_id ) )
    {
      VALUE num = rb_funcall2( proc, swig_arity_id, 0, 0 );
      int arity = NUM2INT(num);
      if ( arity < 0 && (arity+1) < -minimal ) return 1;
      if ( arity == minimal ) return 1;
      return 1;
    }
  return 0;
}


#ifdef __cplusplus
}
#endif



#define SWIG_exception_fail(code, msg) do { SWIG_Error(code, msg); SWIG_fail; } while(0) 

#define SWIG_contract_assert(expr, msg) do { if (!(expr)) { SWIG_Error(SWIG_RuntimeError, msg); SWIG_fail; } } while (0) 



  #define SWIG_exception(code, msg) do { SWIG_Error(code, msg);; } while(0) 


/* -------- TYPES TABLE (BEGIN) -------- */

#define SWIGTYPE_p_apr_array_header_t swig_types[0]
#define SWIGTYPE_p_apr_file_t swig_types[1]
#define SWIGTYPE_p_apr_hash_t swig_types[2]
#define SWIGTYPE_p_apr_int32_t swig_types[3]
#define SWIGTYPE_p_apr_int64_t swig_types[4]
#define SWIGTYPE_p_apr_off_t swig_types[5]
#define SWIGTYPE_p_apr_pool_t swig_types[6]
#define SWIGTYPE_p_char swig_types[7]
#define SWIGTYPE_p_f_p_apr_getopt_t_p_void_p_apr_pool_t__p_svn_error_t swig_types[8]
#define SWIGTYPE_p_f_p_p_q_const__char_p_p_q_const__char_p_void_p_q_const__char_p_apr_pool_t_p_apr_pool_t__p_svn_error_t swig_types[9]
#define SWIGTYPE_p_f_p_p_q_const__svn_string_t_p_svn_stream_t_p_void_p_apr_pool_t__p_svn_error_t swig_types[10]
#define SWIGTYPE_p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description2_t_p_void_p_apr_pool_t_p_apr_pool_t__p_svn_error_t swig_types[11]
#define SWIGTYPE_p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description_t_p_void_p_apr_pool_t__p_svn_error_t swig_types[12]
#define SWIGTYPE_p_f_p_q_const__char_p_q_const__svn_wc_entry_t_p_void_p_apr_pool_t__p_svn_error_t swig_types[13]
#define SWIGTYPE_p_f_p_q_const__char_p_svn_error_t_p_void_p_apr_pool_t__p_svn_error_t swig_types[14]
#define SWIGTYPE_p_f_p_svn_boolean_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_void_p_apr_pool_t__p_svn_error_t swig_types[15]
#define SWIGTYPE_p_f_p_svn_boolean_t_p_svn_boolean_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_void_p_apr_pool_t__p_svn_error_t swig_types[16]
#define SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_void__p_svn_error_t swig_types[17]
#define SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t swig_types[18]
#define SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_hash_t_p_void__p_svn_error_t swig_types[19]
#define SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_void__p_svn_error_t swig_types[20]
#define SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_void__p_svn_error_t swig_types[21]
#define SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_void__p_svn_error_t swig_types[22]
#define SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_svn_revnum_t_p_void__p_svn_error_t swig_types[23]
#define SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t swig_types[24]
#define SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_hash_t_p_void__p_svn_error_t swig_types[25]
#define SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_void__p_svn_error_t swig_types[26]
#define SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_void__p_svn_error_t swig_types[27]
#define SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t swig_types[28]
#define SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t swig_types[29]
#define SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_void__p_svn_error_t swig_types[30]
#define SWIGTYPE_p_f_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_hash_t_p_void_p_apr_pool_t__p_svn_error_t swig_types[31]
#define SWIGTYPE_p_f_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_void_p_apr_pool_t__p_svn_error_t swig_types[32]
#define SWIGTYPE_p_f_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_svn_boolean_t_p_q_const__apr_array_header_t_p_apr_hash_t_p_void_p_apr_pool_t__p_svn_error_t swig_types[33]
#define SWIGTYPE_p_f_p_svn_wc_notify_state_t_p_svn_boolean_t_p_svn_boolean_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_q_const__char_svn_revnum_t_p_void_p_apr_pool_t__p_svn_error_t swig_types[34]
#define SWIGTYPE_p_f_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void_p_apr_pool_t__p_svn_error_t swig_types[35]
#define SWIGTYPE_p_f_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_p_q_const__apr_array_header_t_p_apr_hash_t_p_void_p_apr_pool_t__p_svn_error_t swig_types[36]
#define SWIGTYPE_p_f_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_svn_boolean_t_p_void_p_apr_pool_t__p_svn_error_t swig_types[37]
#define SWIGTYPE_p_f_p_void__p_svn_error_t swig_types[38]
#define SWIGTYPE_p_f_p_void_p_p_apr_hash_t_p_q_const__char_p_q_const__char_p_apr_pool_t_p_apr_pool_t__p_svn_error_t swig_types[39]
#define SWIGTYPE_p_f_p_void_p_q_const__char_enum_svn_wc_notify_action_t_enum_svn_node_kind_t_p_q_const__char_enum_svn_wc_notify_state_t_enum_svn_wc_notify_state_t_long__void swig_types[40]
#define SWIGTYPE_p_f_p_void_p_q_const__char_p_q_const__char__p_svn_error_t swig_types[41]
#define SWIGTYPE_p_f_p_void_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t swig_types[42]
#define SWIGTYPE_p_f_p_void_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t swig_types[43]
#define SWIGTYPE_p_f_p_void_p_q_const__char_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t swig_types[44]
#define SWIGTYPE_p_f_p_void_p_q_const__char_p_q_const__svn_string_t_p_q_const__svn_string_t_svn_depth_t_p_apr_pool_t__p_svn_error_t swig_types[45]
#define SWIGTYPE_p_f_p_void_p_q_const__char_p_q_const__svn_wc_status3_t_p_apr_pool_t__p_svn_error_t swig_types[46]
#define SWIGTYPE_p_f_p_void_p_q_const__char_p_struct_svn_wc_status2_t__void swig_types[47]
#define SWIGTYPE_p_f_p_void_p_q_const__char_p_struct_svn_wc_status_t__void swig_types[48]
#define SWIGTYPE_p_f_p_void_p_q_const__char_p_svn_wc_status2_t_p_apr_pool_t__p_svn_error_t swig_types[49]
#define SWIGTYPE_p_f_p_void_p_q_const__char_svn_revnum_t_p_svn_stream_t_p_svn_revnum_t_p_p_apr_hash_t_p_apr_pool_t__p_svn_error_t swig_types[50]
#define SWIGTYPE_p_f_p_void_p_q_const__struct_svn_wc_notify_t_p_apr_pool_t__void swig_types[51]
#define SWIGTYPE_p_int swig_types[52]
#define SWIGTYPE_p_long swig_types[53]
#define SWIGTYPE_p_p_apr_array_header_t swig_types[54]
#define SWIGTYPE_p_p_apr_file_t swig_types[55]
#define SWIGTYPE_p_p_apr_hash_t swig_types[56]
#define SWIGTYPE_p_p_char swig_types[57]
#define SWIGTYPE_p_p_svn_checksum_t swig_types[58]
#define SWIGTYPE_p_p_svn_delta_editor_t swig_types[59]
#define SWIGTYPE_p_p_svn_stream_t swig_types[60]
#define SWIGTYPE_p_p_svn_string_t swig_types[61]
#define SWIGTYPE_p_p_svn_wc_adm_access_t swig_types[62]
#define SWIGTYPE_p_p_svn_wc_committed_queue_t swig_types[63]
#define SWIGTYPE_p_p_svn_wc_conflict_result_t swig_types[64]
#define SWIGTYPE_p_p_svn_wc_context_t swig_types[65]
#define SWIGTYPE_p_p_svn_wc_revision_status_t swig_types[66]
#define SWIGTYPE_p_p_svn_wc_status2_t swig_types[67]
#define SWIGTYPE_p_p_svn_wc_status3_t swig_types[68]
#define SWIGTYPE_p_p_svn_wc_status_t swig_types[69]
#define SWIGTYPE_p_p_void swig_types[70]
#define SWIGTYPE_p_svn_auth_baton_t swig_types[71]
#define SWIGTYPE_p_svn_auth_cred_simple_t swig_types[72]
#define SWIGTYPE_p_svn_auth_cred_ssl_client_cert_pw_t swig_types[73]
#define SWIGTYPE_p_svn_auth_cred_ssl_client_cert_t swig_types[74]
#define SWIGTYPE_p_svn_auth_cred_ssl_server_trust_t swig_types[75]
#define SWIGTYPE_p_svn_auth_cred_username_t swig_types[76]
#define SWIGTYPE_p_svn_auth_iterstate_t swig_types[77]
#define SWIGTYPE_p_svn_auth_provider_object_t swig_types[78]
#define SWIGTYPE_p_svn_auth_provider_t swig_types[79]
#define SWIGTYPE_p_svn_auth_ssl_server_cert_info_t swig_types[80]
#define SWIGTYPE_p_svn_cache_config_t swig_types[81]
#define SWIGTYPE_p_svn_checksum_ctx_t swig_types[82]
#define SWIGTYPE_p_svn_checksum_kind_t swig_types[83]
#define SWIGTYPE_p_svn_checksum_t swig_types[84]
#define SWIGTYPE_p_svn_commit_info_t swig_types[85]
#define SWIGTYPE_p_svn_config_t swig_types[86]
#define SWIGTYPE_p_svn_delta_editor_t swig_types[87]
#define SWIGTYPE_p_svn_delta_path_driver_state_t swig_types[88]
#define SWIGTYPE_p_svn_depth_t swig_types[89]
#define SWIGTYPE_p_svn_diff_binary_patch_t swig_types[90]
#define SWIGTYPE_p_svn_diff_conflict_display_style_t swig_types[91]
#define SWIGTYPE_p_svn_diff_datasource_e swig_types[92]
#define SWIGTYPE_p_svn_diff_file_ignore_space_t swig_types[93]
#define SWIGTYPE_p_svn_diff_file_options_t swig_types[94]
#define SWIGTYPE_p_svn_diff_fns2_t swig_types[95]
#define SWIGTYPE_p_svn_diff_fns_t swig_types[96]
#define SWIGTYPE_p_svn_diff_hunk_t swig_types[97]
#define SWIGTYPE_p_svn_diff_operation_kind_e swig_types[98]
#define SWIGTYPE_p_svn_diff_output_fns_t swig_types[99]
#define SWIGTYPE_p_svn_diff_t swig_types[100]
#define SWIGTYPE_p_svn_dirent_t swig_types[101]
#define SWIGTYPE_p_svn_errno_t swig_types[102]
#define SWIGTYPE_p_svn_error_t swig_types[103]
#define SWIGTYPE_p_svn_io_dirent2_t swig_types[104]
#define SWIGTYPE_p_svn_io_dirent_t swig_types[105]
#define SWIGTYPE_p_svn_io_file_del_t swig_types[106]
#define SWIGTYPE_p_svn_location_segment_t swig_types[107]
#define SWIGTYPE_p_svn_lock_t swig_types[108]
#define SWIGTYPE_p_svn_log_changed_path2_t swig_types[109]
#define SWIGTYPE_p_svn_log_changed_path_t swig_types[110]
#define SWIGTYPE_p_svn_log_entry_t swig_types[111]
#define SWIGTYPE_p_svn_merge_range_t swig_types[112]
#define SWIGTYPE_p_svn_mergeinfo_inheritance_t swig_types[113]
#define SWIGTYPE_p_svn_node_kind_t swig_types[114]
#define SWIGTYPE_p_svn_opt_revision_range_t swig_types[115]
#define SWIGTYPE_p_svn_opt_revision_t swig_types[116]
#define SWIGTYPE_p_svn_opt_revision_value_t swig_types[117]
#define SWIGTYPE_p_svn_opt_subcommand_desc2_t swig_types[118]
#define SWIGTYPE_p_svn_opt_subcommand_desc3_t swig_types[119]
#define SWIGTYPE_p_svn_opt_subcommand_desc_t swig_types[120]
#define SWIGTYPE_p_svn_patch_file_t swig_types[121]
#define SWIGTYPE_p_svn_patch_t swig_types[122]
#define SWIGTYPE_p_svn_prop_inherited_item_t swig_types[123]
#define SWIGTYPE_p_svn_prop_kind swig_types[124]
#define SWIGTYPE_p_svn_prop_patch_t swig_types[125]
#define SWIGTYPE_p_svn_ra_callbacks2_t swig_types[126]
#define SWIGTYPE_p_svn_ra_callbacks_t swig_types[127]
#define SWIGTYPE_p_svn_ra_plugin_t swig_types[128]
#define SWIGTYPE_p_svn_ra_reporter2_t swig_types[129]
#define SWIGTYPE_p_svn_ra_reporter3_t swig_types[130]
#define SWIGTYPE_p_svn_ra_reporter_t swig_types[131]
#define SWIGTYPE_p_svn_ra_session_t swig_types[132]
#define SWIGTYPE_p_svn_stream_mark_t swig_types[133]
#define SWIGTYPE_p_svn_stream_t swig_types[134]
#define SWIGTYPE_p_svn_string_t swig_types[135]
#define SWIGTYPE_p_svn_stringbuf_t swig_types[136]
#define SWIGTYPE_p_svn_tristate_t swig_types[137]
#define SWIGTYPE_p_svn_txdelta_op_t swig_types[138]
#define SWIGTYPE_p_svn_txdelta_stream_t swig_types[139]
#define SWIGTYPE_p_svn_txdelta_window_t swig_types[140]
#define SWIGTYPE_p_svn_version_checklist_t swig_types[141]
#define SWIGTYPE_p_svn_version_ext_linked_lib_t swig_types[142]
#define SWIGTYPE_p_svn_version_ext_loaded_lib_t swig_types[143]
#define SWIGTYPE_p_svn_version_extended_t swig_types[144]
#define SWIGTYPE_p_svn_version_t swig_types[145]
#define SWIGTYPE_p_svn_wc_adm_access_t swig_types[146]
#define SWIGTYPE_p_svn_wc_committed_queue_t swig_types[147]
#define SWIGTYPE_p_svn_wc_conflict_action_t swig_types[148]
#define SWIGTYPE_p_svn_wc_conflict_choice_t swig_types[149]
#define SWIGTYPE_p_svn_wc_conflict_description2_t swig_types[150]
#define SWIGTYPE_p_svn_wc_conflict_description_t swig_types[151]
#define SWIGTYPE_p_svn_wc_conflict_kind_t swig_types[152]
#define SWIGTYPE_p_svn_wc_conflict_reason_t swig_types[153]
#define SWIGTYPE_p_svn_wc_conflict_result_t swig_types[154]
#define SWIGTYPE_p_svn_wc_conflict_version_t swig_types[155]
#define SWIGTYPE_p_svn_wc_context_t swig_types[156]
#define SWIGTYPE_p_svn_wc_diff_callbacks2_t swig_types[157]
#define SWIGTYPE_p_svn_wc_diff_callbacks3_t swig_types[158]
#define SWIGTYPE_p_svn_wc_diff_callbacks4_t swig_types[159]
#define SWIGTYPE_p_svn_wc_diff_callbacks_t swig_types[160]
#define SWIGTYPE_p_svn_wc_entry_callbacks2_t swig_types[161]
#define SWIGTYPE_p_svn_wc_entry_callbacks_t swig_types[162]
#define SWIGTYPE_p_svn_wc_entry_t swig_types[163]
#define SWIGTYPE_p_svn_wc_external_item2_t swig_types[164]
#define SWIGTYPE_p_svn_wc_external_item_t swig_types[165]
#define SWIGTYPE_p_svn_wc_info_t swig_types[166]
#define SWIGTYPE_p_svn_wc_merge_outcome_t swig_types[167]
#define SWIGTYPE_p_svn_wc_notify_action_t swig_types[168]
#define SWIGTYPE_p_svn_wc_notify_lock_state_t swig_types[169]
#define SWIGTYPE_p_svn_wc_notify_state_t swig_types[170]
#define SWIGTYPE_p_svn_wc_notify_t swig_types[171]
#define SWIGTYPE_p_svn_wc_operation_t swig_types[172]
#define SWIGTYPE_p_svn_wc_revision_status_t swig_types[173]
#define SWIGTYPE_p_svn_wc_schedule_t swig_types[174]
#define SWIGTYPE_p_svn_wc_status2_t swig_types[175]
#define SWIGTYPE_p_svn_wc_status3_t swig_types[176]
#define SWIGTYPE_p_svn_wc_status_t swig_types[177]
#define SWIGTYPE_p_svn_wc_traversal_info_t swig_types[178]
#define SWIGTYPE_p_unsigned_char swig_types[179]
#define SWIGTYPE_p_unsigned_long swig_types[180]
#define SWIGTYPE_p_void swig_types[181]
static swig_type_info *swig_types[183];
static swig_module_info swig_module = {swig_types, 182, 0, 0, 0, 0};
#define SWIG_TypeQuery(name) SWIG_TypeQueryModule(&swig_module, &swig_module, name)
#define SWIG_MangledTypeQuery(name) SWIG_MangledTypeQueryModule(&swig_module, &swig_module, name)

/* -------- TYPES TABLE (END) -------- */

#define SWIG_init    Init_wc
#define SWIG_name    "Svn::Ext::Wc"

static VALUE mWc;

#define SWIG_RUBY_THREAD_BEGIN_BLOCK
#define SWIG_RUBY_THREAD_END_BLOCK


#define SWIG_as_voidptr(a) (void *)((const void *)(a)) 
#define SWIG_as_voidptrptr(a) ((void)SWIG_as_voidptr(*a),(void**)(a)) 


#ifdef __cplusplus
extern "C" {
#endif

/* Ruby 1.9 changed the file name of this header */
#ifdef HAVE_RUBY_IO_H
#include "ruby/io.h"
#else
#include "rubyio.h"
#endif

#ifdef __cplusplus
}
#endif


#ifdef __cplusplus
extern "C" {
#endif
#ifdef HAVE_SYS_TIME_H
# include <sys/time.h>
struct timeval rb_time_timeval(VALUE);
#endif
#ifdef __cplusplus
}
#endif


#include "svn_time.h"
#include "svn_pools.h"


#include "swigutil_rb.h"


static VALUE _global_svn_swig_rb_pool = Qnil;
static apr_pool_t *_global_pool = NULL;
static VALUE vresult = Qnil;
static VALUE *_global_vresult_address = &vresult;


#include <apr_md5.h>
#include "svn_md5.h"


#include "svn_wc.h"


#include <limits.h>
#if !defined(SWIG_NO_LLONG_MAX)
# if !defined(LLONG_MAX) && defined(__GNUC__) && defined (__LONG_LONG_MAX__)
#   define LLONG_MAX __LONG_LONG_MAX__
#   define LLONG_MIN (-LLONG_MAX - 1LL)
#   define ULLONG_MAX (LLONG_MAX * 2ULL + 1ULL)
# endif
#endif


  #define SWIG_From_long   LONG2NUM 


SWIGINTERNINLINE VALUE
SWIG_From_int  (int value)
{    
  return SWIG_From_long  (value);
}


SWIGINTERN swig_type_info*
SWIG_pchar_descriptor(void)
{
  static swig_type_info* info = 0;
  if (!info) {
    info = SWIG_TypeQuery("_p_char");
  }
  return info;
}


SWIGINTERN int
SWIG_AsCharPtrAndSize(VALUE obj, char** cptr, size_t* psize, int *alloc)
{
  if (TYPE(obj) == T_STRING) {
    char *cstr = StringValuePtr(obj); 
    size_t size = RSTRING_LEN(obj) + 1;
    if (cptr)  {
      if (alloc) {
	if (*alloc == SWIG_NEWOBJ) {
	  *cptr = (char *)memcpy(malloc((size)*sizeof(char)), cstr, sizeof(char)*(size));
	} else {
	  *cptr = cstr;
	  *alloc = SWIG_OLDOBJ;
	}
      }
    }
    if (psize) *psize = size;
    return SWIG_OK;
  } else {
    swig_type_info* pchar_descriptor = SWIG_pchar_descriptor();
    if (pchar_descriptor) {
      void* vptr = 0;
      if (SWIG_ConvertPtr(obj, &vptr, pchar_descriptor, 0) == SWIG_OK) {
	if (cptr) *cptr = (char *)vptr;
	if (psize) *psize = vptr ? (strlen((char*)vptr) + 1) : 0;
	if (alloc) *alloc = SWIG_OLDOBJ;
	return SWIG_OK;
      }
    }
  }  
  return SWIG_TypeError;
}





SWIGINTERN VALUE
SWIG_ruby_failed(VALUE SWIGUNUSEDPARM(arg1), VALUE SWIGUNUSEDPARM(arg2))
{
  return Qnil;
} 


SWIGINTERN VALUE SWIG_AUX_NUM2LONG(VALUE arg)
{
  VALUE *args = (VALUE *)arg;
  VALUE obj = args[0];
  VALUE type = TYPE(obj);
  long *res = (long *)(args[1]);
  *res = type == T_FIXNUM ? NUM2LONG(obj) : rb_big2long(obj);
  return obj;
}

SWIGINTERN int
SWIG_AsVal_long (VALUE obj, long* val)
{
  VALUE type = TYPE(obj);
  if ((type == T_FIXNUM) || (type == T_BIGNUM)) {
    long v;
    VALUE a[2];
    a[0] = obj;
    a[1] = (VALUE)(&v);
    if (rb_rescue(VALUEFUNC(SWIG_AUX_NUM2LONG), (VALUE)a, VALUEFUNC(SWIG_ruby_failed), 0) != Qnil) {
      if (val) *val = v;
      return SWIG_OK;
    }
  }
  return SWIG_TypeError;
}


SWIGINTERN int
SWIG_AsVal_int (VALUE obj, int *val)
{
  long v;
  int res = SWIG_AsVal_long (obj, &v);
  if (SWIG_IsOK(res)) {
    if ((v < INT_MIN || v > INT_MAX)) {
      return SWIG_OverflowError;
    } else {
      if (val) *val = (int)(v);
    }
  }  
  return res;
}


SWIGINTERNINLINE VALUE 
SWIG_FromCharPtrAndSize(const char* carray, size_t size)
{
  if (carray) {
    if (size > LONG_MAX) {
      swig_type_info* pchar_descriptor = SWIG_pchar_descriptor();
      return pchar_descriptor ? 
	SWIG_NewPointerObj((char *)(carray), pchar_descriptor, 0) : Qnil;
    } else {
      return rb_str_new(carray, (long)(size));
    }
  } else {
    return Qnil;
  }
}


SWIGINTERNINLINE VALUE 
SWIG_FromCharPtr(const char *cptr)
{ 
  return SWIG_FromCharPtrAndSize(cptr, (cptr ? strlen(cptr) : 0));
}

SWIGINTERN struct svn_wc_external_item2_t *new_svn_wc_external_item2_t(apr_pool_t *pool){
    svn_error_t *err;
    svn_wc_external_item2_t *self;
    err = svn_wc_external_item2_create(&self, pool);
    if (err)
      svn_swig_rb_handle_svn_error(err);
    return self;
  }
SWIGINTERN svn_wc_external_item2_t *svn_wc_external_item2_t_dup(struct svn_wc_external_item2_t *self,apr_pool_t *pool){
    return svn_wc_external_item2_dup(self, pool);
  }

SWIGINTERN VALUE SWIG_AUX_NUM2ULONG(VALUE arg)
{
  VALUE *args = (VALUE *)arg;
  VALUE obj = args[0];
  VALUE type = TYPE(obj);
  unsigned long *res = (unsigned long *)(args[1]);
  *res = type == T_FIXNUM ? NUM2ULONG(obj) : rb_big2ulong(obj);
  return obj;
}

SWIGINTERN int
SWIG_AsVal_unsigned_SS_long (VALUE obj, unsigned long *val) 
{
  VALUE type = TYPE(obj);
  if ((type == T_FIXNUM) || (type == T_BIGNUM)) {
    unsigned long v;
    VALUE a[2];
    a[0] = obj;
    a[1] = (VALUE)(&v);
    if (rb_rescue(VALUEFUNC(SWIG_AUX_NUM2ULONG), (VALUE)a, VALUEFUNC(SWIG_ruby_failed), 0) != Qnil) {
      if (rb_funcall(obj, swig_lowerthan_id, 1, INT2FIX(0)) != Qfalse)
        return SWIG_OverflowError;
      if (val) *val = v;
      return SWIG_OK;
    }
  }
  return SWIG_TypeError;
}


SWIGINTERNINLINE VALUE
SWIG_From_unsigned_SS_long  (unsigned long value)
{
  return ULONG2NUM(value); 
}

SWIGINTERN struct svn_wc_notify_t *new_svn_wc_notify_t(char const *path,svn_wc_notify_action_t action,apr_pool_t *pool){
    return svn_wc_create_notify(path, action, pool);
  }
SWIGINTERN struct svn_wc_entry_t *new_svn_wc_entry_t(char const *path,svn_wc_adm_access_t *adm_access,svn_boolean_t show_hidden,apr_pool_t *pool){
    const svn_wc_entry_t *self;
    svn_wc_entry(&self, path, adm_access, show_hidden, pool);
    return (svn_wc_entry_t *)self;
  }
SWIGINTERN struct svn_wc_revision_status_t *new_svn_wc_revision_status_t(char const *wc_path,char const *trail_url,svn_boolean_t committed,svn_cancel_func_t cancel_func,void *cancel_baton,apr_pool_t *pool){
    svn_error_t *err;
    svn_wc_revision_status_t *self;
    err = svn_wc_revision_status(&self, wc_path, trail_url, committed,
                                 cancel_func, cancel_baton, pool);
    if (err)
      svn_swig_rb_handle_svn_error(err);
    return self;
  }

static svn_error_t * svn_wc_diff_callbacks4_invoke_file_opened(
  svn_wc_diff_callbacks4_t * _obj, svn_boolean_t *tree_conflicted, svn_boolean_t *skip, const char *path, svn_revnum_t rev, void *diff_baton, apr_pool_t *scratch_pool) {
  return (_obj->file_opened)(tree_conflicted, skip, path, rev, diff_baton, scratch_pool);
}

static svn_error_t * svn_wc_diff_callbacks4_invoke_file_changed(
  svn_wc_diff_callbacks4_t * _obj, svn_wc_notify_state_t *contentstate, svn_wc_notify_state_t *propstate, svn_boolean_t *tree_conflicted, const char *path, const char *tmpfile1, const char *tmpfile2, svn_revnum_t rev1, svn_revnum_t rev2, const char *mimetype1, const char *mimetype2, const apr_array_header_t *propchanges, apr_hash_t *originalprops, void *diff_baton, apr_pool_t *scratch_pool) {
  return (_obj->file_changed)(contentstate, propstate, tree_conflicted, path, tmpfile1, tmpfile2, rev1, rev2, mimetype1, mimetype2, propchanges, originalprops, diff_baton, scratch_pool);
}

static svn_error_t * svn_wc_diff_callbacks4_invoke_file_added(
  svn_wc_diff_callbacks4_t * _obj, svn_wc_notify_state_t *contentstate, svn_wc_notify_state_t *propstate, svn_boolean_t *tree_conflicted, const char *path, const char *tmpfile1, const char *tmpfile2, svn_revnum_t rev1, svn_revnum_t rev2, const char *mimetype1, const char *mimetype2, const char *copyfrom_path, svn_revnum_t copyfrom_revision, const apr_array_header_t *propchanges, apr_hash_t *originalprops, void *diff_baton, apr_pool_t *scratch_pool) {
  return (_obj->file_added)(contentstate, propstate, tree_conflicted, path, tmpfile1, tmpfile2, rev1, rev2, mimetype1, mimetype2, copyfrom_path, copyfrom_revision, propchanges, originalprops, diff_baton, scratch_pool);
}

static svn_error_t * svn_wc_diff_callbacks4_invoke_file_deleted(
  svn_wc_diff_callbacks4_t * _obj, svn_wc_notify_state_t *state, svn_boolean_t *tree_conflicted, const char *path, const char *tmpfile1, const char *tmpfile2, const char *mimetype1, const char *mimetype2, apr_hash_t *originalprops, void *diff_baton, apr_pool_t *scratch_pool) {
  return (_obj->file_deleted)(state, tree_conflicted, path, tmpfile1, tmpfile2, mimetype1, mimetype2, originalprops, diff_baton, scratch_pool);
}

static svn_error_t * svn_wc_diff_callbacks4_invoke_dir_deleted(
  svn_wc_diff_callbacks4_t * _obj, svn_wc_notify_state_t *state, svn_boolean_t *tree_conflicted, const char *path, void *diff_baton, apr_pool_t *scratch_pool) {
  return (_obj->dir_deleted)(state, tree_conflicted, path, diff_baton, scratch_pool);
}

static svn_error_t * svn_wc_diff_callbacks4_invoke_dir_opened(
  svn_wc_diff_callbacks4_t * _obj, svn_boolean_t *tree_conflicted, svn_boolean_t *skip, svn_boolean_t *skip_children, const char *path, svn_revnum_t rev, void *diff_baton, apr_pool_t *scratch_pool) {
  return (_obj->dir_opened)(tree_conflicted, skip, skip_children, path, rev, diff_baton, scratch_pool);
}

static svn_error_t * svn_wc_diff_callbacks4_invoke_dir_added(
  svn_wc_diff_callbacks4_t * _obj, svn_wc_notify_state_t *state, svn_boolean_t *tree_conflicted, svn_boolean_t *skip, svn_boolean_t *skip_children, const char *path, svn_revnum_t rev, const char *copyfrom_path, svn_revnum_t copyfrom_revision, void *diff_baton, apr_pool_t *scratch_pool) {
  return (_obj->dir_added)(state, tree_conflicted, skip, skip_children, path, rev, copyfrom_path, copyfrom_revision, diff_baton, scratch_pool);
}

static svn_error_t * svn_wc_diff_callbacks4_invoke_dir_props_changed(
  svn_wc_diff_callbacks4_t * _obj, svn_wc_notify_state_t *propstate, svn_boolean_t *tree_conflicted, const char *path, svn_boolean_t dir_was_added, const apr_array_header_t *propchanges, apr_hash_t *original_props, void *diff_baton, apr_pool_t *scratch_pool) {
  return (_obj->dir_props_changed)(propstate, tree_conflicted, path, dir_was_added, propchanges, original_props, diff_baton, scratch_pool);
}

static svn_error_t * svn_wc_diff_callbacks4_invoke_dir_closed(
  svn_wc_diff_callbacks4_t * _obj, svn_wc_notify_state_t *contentstate, svn_wc_notify_state_t *propstate, svn_boolean_t *tree_conflicted, const char *path, svn_boolean_t dir_was_added, void *diff_baton, apr_pool_t *scratch_pool) {
  return (_obj->dir_closed)(contentstate, propstate, tree_conflicted, path, dir_was_added, diff_baton, scratch_pool);
}

static svn_error_t * svn_wc_diff_callbacks3_invoke_file_changed(
  svn_wc_diff_callbacks3_t * _obj, svn_wc_adm_access_t *adm_access, svn_wc_notify_state_t *contentstate, svn_wc_notify_state_t *propstate, svn_boolean_t *tree_conflicted, const char *path, const char *tmpfile1, const char *tmpfile2, svn_revnum_t rev1, svn_revnum_t rev2, const char *mimetype1, const char *mimetype2, const apr_array_header_t *propchanges, apr_hash_t *originalprops, void *diff_baton) {
  return (_obj->file_changed)(adm_access, contentstate, propstate, tree_conflicted, path, tmpfile1, tmpfile2, rev1, rev2, mimetype1, mimetype2, propchanges, originalprops, diff_baton);
}

static svn_error_t * svn_wc_diff_callbacks3_invoke_file_added(
  svn_wc_diff_callbacks3_t * _obj, svn_wc_adm_access_t *adm_access, svn_wc_notify_state_t *contentstate, svn_wc_notify_state_t *propstate, svn_boolean_t *tree_conflicted, const char *path, const char *tmpfile1, const char *tmpfile2, svn_revnum_t rev1, svn_revnum_t rev2, const char *mimetype1, const char *mimetype2, const apr_array_header_t *propchanges, apr_hash_t *originalprops, void *diff_baton) {
  return (_obj->file_added)(adm_access, contentstate, propstate, tree_conflicted, path, tmpfile1, tmpfile2, rev1, rev2, mimetype1, mimetype2, propchanges, originalprops, diff_baton);
}

static svn_error_t * svn_wc_diff_callbacks3_invoke_file_deleted(
  svn_wc_diff_callbacks3_t * _obj, svn_wc_adm_access_t *adm_access, svn_wc_notify_state_t *state, svn_boolean_t *tree_conflicted, const char *path, const char *tmpfile1, const char *tmpfile2, const char *mimetype1, const char *mimetype2, apr_hash_t *originalprops, void *diff_baton) {
  return (_obj->file_deleted)(adm_access, state, tree_conflicted, path, tmpfile1, tmpfile2, mimetype1, mimetype2, originalprops, diff_baton);
}

static svn_error_t * svn_wc_diff_callbacks3_invoke_dir_added(
  svn_wc_diff_callbacks3_t * _obj, svn_wc_adm_access_t *adm_access, svn_wc_notify_state_t *state, svn_boolean_t *tree_conflicted, const char *path, svn_revnum_t rev, void *diff_baton) {
  return (_obj->dir_added)(adm_access, state, tree_conflicted, path, rev, diff_baton);
}

static svn_error_t * svn_wc_diff_callbacks3_invoke_dir_deleted(
  svn_wc_diff_callbacks3_t * _obj, svn_wc_adm_access_t *adm_access, svn_wc_notify_state_t *state, svn_boolean_t *tree_conflicted, const char *path, void *diff_baton) {
  return (_obj->dir_deleted)(adm_access, state, tree_conflicted, path, diff_baton);
}

static svn_error_t * svn_wc_diff_callbacks3_invoke_dir_props_changed(
  svn_wc_diff_callbacks3_t * _obj, svn_wc_adm_access_t *adm_access, svn_wc_notify_state_t *propstate, svn_boolean_t *tree_conflicted, const char *path, const apr_array_header_t *propchanges, apr_hash_t *original_props, void *diff_baton) {
  return (_obj->dir_props_changed)(adm_access, propstate, tree_conflicted, path, propchanges, original_props, diff_baton);
}

static svn_error_t * svn_wc_diff_callbacks3_invoke_dir_opened(
  svn_wc_diff_callbacks3_t * _obj, svn_wc_adm_access_t *adm_access, svn_boolean_t *tree_conflicted, const char *path, svn_revnum_t rev, void *diff_baton) {
  return (_obj->dir_opened)(adm_access, tree_conflicted, path, rev, diff_baton);
}

static svn_error_t * svn_wc_diff_callbacks3_invoke_dir_closed(
  svn_wc_diff_callbacks3_t * _obj, svn_wc_adm_access_t *adm_access, svn_wc_notify_state_t *contentstate, svn_wc_notify_state_t *propstate, svn_boolean_t *tree_conflicted, const char *path, void *diff_baton) {
  return (_obj->dir_closed)(adm_access, contentstate, propstate, tree_conflicted, path, diff_baton);
}

static svn_error_t * svn_wc_diff_callbacks2_invoke_file_changed(
  svn_wc_diff_callbacks2_t * _obj, svn_wc_adm_access_t *adm_access, svn_wc_notify_state_t *contentstate, svn_wc_notify_state_t *propstate, const char *path, const char *tmpfile1, const char *tmpfile2, svn_revnum_t rev1, svn_revnum_t rev2, const char *mimetype1, const char *mimetype2, const apr_array_header_t *propchanges, apr_hash_t *originalprops, void *diff_baton) {
  return (_obj->file_changed)(adm_access, contentstate, propstate, path, tmpfile1, tmpfile2, rev1, rev2, mimetype1, mimetype2, propchanges, originalprops, diff_baton);
}

static svn_error_t * svn_wc_diff_callbacks2_invoke_file_added(
  svn_wc_diff_callbacks2_t * _obj, svn_wc_adm_access_t *adm_access, svn_wc_notify_state_t *contentstate, svn_wc_notify_state_t *propstate, const char *path, const char *tmpfile1, const char *tmpfile2, svn_revnum_t rev1, svn_revnum_t rev2, const char *mimetype1, const char *mimetype2, const apr_array_header_t *propchanges, apr_hash_t *originalprops, void *diff_baton) {
  return (_obj->file_added)(adm_access, contentstate, propstate, path, tmpfile1, tmpfile2, rev1, rev2, mimetype1, mimetype2, propchanges, originalprops, diff_baton);
}

static svn_error_t * svn_wc_diff_callbacks2_invoke_file_deleted(
  svn_wc_diff_callbacks2_t * _obj, svn_wc_adm_access_t *adm_access, svn_wc_notify_state_t *state, const char *path, const char *tmpfile1, const char *tmpfile2, const char *mimetype1, const char *mimetype2, apr_hash_t *originalprops, void *diff_baton) {
  return (_obj->file_deleted)(adm_access, state, path, tmpfile1, tmpfile2, mimetype1, mimetype2, originalprops, diff_baton);
}

static svn_error_t * svn_wc_diff_callbacks2_invoke_dir_added(
  svn_wc_diff_callbacks2_t * _obj, svn_wc_adm_access_t *adm_access, svn_wc_notify_state_t *state, const char *path, svn_revnum_t rev, void *diff_baton) {
  return (_obj->dir_added)(adm_access, state, path, rev, diff_baton);
}

static svn_error_t * svn_wc_diff_callbacks2_invoke_dir_deleted(
  svn_wc_diff_callbacks2_t * _obj, svn_wc_adm_access_t *adm_access, svn_wc_notify_state_t *state, const char *path, void *diff_baton) {
  return (_obj->dir_deleted)(adm_access, state, path, diff_baton);
}

static svn_error_t * svn_wc_diff_callbacks2_invoke_dir_props_changed(
  svn_wc_diff_callbacks2_t * _obj, svn_wc_adm_access_t *adm_access, svn_wc_notify_state_t *state, const char *path, const apr_array_header_t *propchanges, apr_hash_t *original_props, void *diff_baton) {
  return (_obj->dir_props_changed)(adm_access, state, path, propchanges, original_props, diff_baton);
}

static svn_error_t * svn_wc_diff_callbacks_invoke_file_changed(
  svn_wc_diff_callbacks_t * _obj, svn_wc_adm_access_t *adm_access, svn_wc_notify_state_t *state, const char *path, const char *tmpfile1, const char *tmpfile2, svn_revnum_t rev1, svn_revnum_t rev2, const char *mimetype1, const char *mimetype2, void *diff_baton) {
  return (_obj->file_changed)(adm_access, state, path, tmpfile1, tmpfile2, rev1, rev2, mimetype1, mimetype2, diff_baton);
}

static svn_error_t * svn_wc_diff_callbacks_invoke_file_added(
  svn_wc_diff_callbacks_t * _obj, svn_wc_adm_access_t *adm_access, svn_wc_notify_state_t *state, const char *path, const char *tmpfile1, const char *tmpfile2, svn_revnum_t rev1, svn_revnum_t rev2, const char *mimetype1, const char *mimetype2, void *diff_baton) {
  return (_obj->file_added)(adm_access, state, path, tmpfile1, tmpfile2, rev1, rev2, mimetype1, mimetype2, diff_baton);
}

static svn_error_t * svn_wc_diff_callbacks_invoke_file_deleted(
  svn_wc_diff_callbacks_t * _obj, svn_wc_adm_access_t *adm_access, svn_wc_notify_state_t *state, const char *path, const char *tmpfile1, const char *tmpfile2, const char *mimetype1, const char *mimetype2, void *diff_baton) {
  return (_obj->file_deleted)(adm_access, state, path, tmpfile1, tmpfile2, mimetype1, mimetype2, diff_baton);
}

static svn_error_t * svn_wc_diff_callbacks_invoke_dir_added(
  svn_wc_diff_callbacks_t * _obj, svn_wc_adm_access_t *adm_access, svn_wc_notify_state_t *state, const char *path, svn_revnum_t rev, void *diff_baton) {
  return (_obj->dir_added)(adm_access, state, path, rev, diff_baton);
}

static svn_error_t * svn_wc_diff_callbacks_invoke_dir_deleted(
  svn_wc_diff_callbacks_t * _obj, svn_wc_adm_access_t *adm_access, svn_wc_notify_state_t *state, const char *path, void *diff_baton) {
  return (_obj->dir_deleted)(adm_access, state, path, diff_baton);
}

static svn_error_t * svn_wc_diff_callbacks_invoke_props_changed(
  svn_wc_diff_callbacks_t * _obj, svn_wc_adm_access_t *adm_access, svn_wc_notify_state_t *state, const char *path, const apr_array_header_t *propchanges, apr_hash_t *original_props, void *diff_baton) {
  return (_obj->props_changed)(adm_access, state, path, propchanges, original_props, diff_baton);
}

static svn_error_t * svn_wc_entry_callbacks2_invoke_found_entry(
  svn_wc_entry_callbacks2_t * _obj, const char *path, const svn_wc_entry_t *entry, void *walk_baton, apr_pool_t *pool) {
  return (_obj->found_entry)(path, entry, walk_baton, pool);
}

static svn_error_t * svn_wc_entry_callbacks2_invoke_handle_error(
  svn_wc_entry_callbacks2_t * _obj, const char *path, svn_error_t *err, void *walk_baton, apr_pool_t *pool) {
  return (_obj->handle_error)(path, err, walk_baton, pool);
}

static svn_error_t * svn_wc_entry_callbacks_invoke_found_entry(
  svn_wc_entry_callbacks_t * _obj, const char *path, const svn_wc_entry_t *entry, void *walk_baton, apr_pool_t *pool) {
  return (_obj->found_entry)(path, entry, walk_baton, pool);
}

static svn_error_t * svn_wc_invoke_external_update(
  svn_wc_external_update_t _obj, void *baton, const char *local_abspath, const svn_string_t *old_val, const svn_string_t *new_val, svn_depth_t depth, apr_pool_t *scratch_pool) {
  return _obj(baton, local_abspath, old_val, new_val, depth, scratch_pool);
}

static void svn_wc_invoke_notify_func2(
  svn_wc_notify_func2_t _obj, void *baton, const svn_wc_notify_t *notify, apr_pool_t *pool) {
  _obj(baton, notify, pool);
}

static void svn_wc_invoke_notify_func(
  svn_wc_notify_func_t _obj, void *baton, const char *path, svn_wc_notify_action_t action, svn_node_kind_t kind, const char *mime_type, svn_wc_notify_state_t content_state, svn_wc_notify_state_t prop_state, svn_revnum_t revision) {
  _obj(baton, path, action, kind, mime_type, content_state, prop_state, revision);
}

static svn_error_t * svn_wc_invoke_conflict_resolver_func2(
  svn_wc_conflict_resolver_func2_t _obj, svn_wc_conflict_result_t **result, const svn_wc_conflict_description2_t *description, void *baton, apr_pool_t *result_pool, apr_pool_t *scratch_pool) {
  return _obj(result, description, baton, result_pool, scratch_pool);
}

static svn_error_t * svn_wc_invoke_conflict_resolver_func(
  svn_wc_conflict_resolver_func_t _obj, svn_wc_conflict_result_t **result, const svn_wc_conflict_description_t *description, void *baton, apr_pool_t *pool) {
  return _obj(result, description, baton, pool);
}

static svn_error_t * svn_wc_invoke_status_func4(
  svn_wc_status_func4_t _obj, void *baton, const char *local_abspath, const svn_wc_status3_t *status, apr_pool_t *scratch_pool) {
  return _obj(baton, local_abspath, status, scratch_pool);
}

static svn_error_t * svn_wc_invoke_status_func3(
  svn_wc_status_func3_t _obj, void *baton, const char *path, svn_wc_status2_t *status, apr_pool_t *pool) {
  return _obj(baton, path, status, pool);
}

static void svn_wc_invoke_status_func2(
  svn_wc_status_func2_t _obj, void *baton, const char *path, svn_wc_status2_t *status) {
  _obj(baton, path, status);
}

static void svn_wc_invoke_status_func(
  svn_wc_status_func_t _obj, void *baton, const char *path, svn_wc_status_t *status) {
  _obj(baton, path, status);
}

static svn_error_t * svn_wc_invoke_get_file(
  svn_wc_get_file_t _obj, void *baton, const char *path, svn_revnum_t revision, svn_stream_t *stream, svn_revnum_t *fetched_rev, apr_hash_t **props, apr_pool_t *pool) {
  return _obj(baton, path, revision, stream, fetched_rev, props, pool);
}

static svn_error_t * svn_wc_invoke_dirents_func(
  svn_wc_dirents_func_t _obj, void *baton, apr_hash_t **dirents, const char *repos_root_url, const char *repos_relpath, apr_pool_t *result_pool, apr_pool_t *scratch_pool) {
  return _obj(baton, dirents, repos_root_url, repos_relpath, result_pool, scratch_pool);
}

static svn_error_t * svn_wc_invoke_canonicalize_svn_prop_get_file(
  svn_wc_canonicalize_svn_prop_get_file_t _obj, const svn_string_t **mime_type, svn_stream_t *stream, void *baton, apr_pool_t *pool) {
  return _obj(mime_type, stream, baton, pool);
}

static svn_error_t * svn_wc_invoke_upgrade_get_repos_info(
  svn_wc_upgrade_get_repos_info_t _obj, const char **repos_root, const char **repos_uuid, void *baton, const char *url, apr_pool_t *result_pool, apr_pool_t *scratch_pool) {
  return _obj(repos_root, repos_uuid, baton, url, result_pool, scratch_pool);
}

static svn_error_t * svn_wc_invoke_relocation_validator3(
  svn_wc_relocation_validator3_t _obj, void *baton, const char *uuid, const char *url, const char *root_url, apr_pool_t *pool) {
  return _obj(baton, uuid, url, root_url, pool);
}

static svn_error_t * svn_wc_invoke_relocation_validator2(
  svn_wc_relocation_validator2_t _obj, void *baton, const char *uuid, const char *url, svn_boolean_t root, apr_pool_t *pool) {
  return _obj(baton, uuid, url, root, pool);
}

static svn_error_t * svn_wc_invoke_relocation_validator(
  svn_wc_relocation_validator_t _obj, void *baton, const char *uuid, const char *url) {
  return _obj(baton, uuid, url);
}

static svn_error_t * svn_changelist_invoke_receiver(
  svn_changelist_receiver_t _obj, void *baton, const char *path, const char *changelist, apr_pool_t *pool) {
  return _obj(baton, path, changelist, pool);
}



static svn_error_t *
svn_wc_swig_init_asp_dot_net_hack (apr_pool_t *pool)
{
#if defined(WIN32) || defined(__CYGWIN__)
  if (getenv ("SVN_ASP_DOT_NET_HACK"))
    SVN_ERR (svn_wc_set_adm_dir("_svn", pool));
#endif /* WIN32 */
  return SVN_NO_ERROR;
}

SWIGINTERN struct svn_wc_committed_queue_t *new_svn_wc_committed_queue_t(apr_pool_t *pool){
    return svn_wc_committed_queue_create(pool);
  }
SWIGINTERN struct svn_wc_traversal_info_t *new_svn_wc_traversal_info_t(apr_pool_t *pool){
    return svn_wc_init_traversal_info(pool);
  }
SWIGINTERN VALUE
_wrap_svn_wc_version(int argc, VALUE *argv, VALUE self) {
  svn_version_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  {
    result = (svn_version_t *)svn_wc_version();
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_version_t, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_context_create(int argc, VALUE *argv, VALUE self) {
  svn_wc_context_t **arg1 = 0 ;
  svn_config_t *arg2 = 0 ;
  apr_pool_t *arg3 = 0 ;
  apr_pool_t *arg4 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_wc_context_t *temp1 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 1) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_svn_config_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_config_t const *","svn_wc_context_create", 2, argv[0] )); 
  }
  arg2 = (svn_config_t *)(argp2);
  if (argc > 1) {
    
  }
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_wc_context_create(arg1,(struct svn_config_t const *)arg2,arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_wc_context_t, 0), 0);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_context_destroy(int argc, VALUE *argv, VALUE self) {
  svn_wc_context_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_context_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_context_t *","svn_wc_context_destroy", 1, argv[0] )); 
  }
  arg1 = (svn_wc_context_t *)(argp1);
  {
    result = (svn_error_t *)svn_wc_context_destroy(arg1);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_adm_open3(int argc, VALUE *argv, VALUE self) {
  svn_wc_adm_access_t **arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  char *arg3 = 0 ;
  svn_boolean_t arg4 ;
  int arg5 ;
  svn_cancel_func_t arg6 = 0 ;
  void *arg7 = 0 ;
  apr_pool_t *arg8 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_wc_adm_access_t *temp1 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int val5 ;
  int ecode5 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg8);
    _global_pool = arg8;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 5) || (argc > 6)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 5)",argc); SWIG_fail;
  }
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_adm_open3", 2, argv[0] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_wc_adm_open3", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  arg4 = RTEST(argv[2]);
  ecode5 = SWIG_AsVal_int(argv[3], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), Ruby_Format_TypeError( "", "int","svn_wc_adm_open3", 5, argv[3] ));
  } 
  arg5 = (int)(val5);
  {
    arg6 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg7 = (void *)svn_swig_rb_make_baton(argv[4], _global_svn_swig_rb_pool);
  }
  if (argc > 5) {
    
  }
  {
    result = (svn_error_t *)svn_wc_adm_open3(arg1,arg2,(char const *)arg3,arg4,arg5,arg6,arg7,arg8);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_wc_adm_access_t, 0), 0);
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg7);
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_adm_open2(int argc, VALUE *argv, VALUE self) {
  svn_wc_adm_access_t **arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  char *arg3 = 0 ;
  svn_boolean_t arg4 ;
  int arg5 ;
  apr_pool_t *arg6 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_wc_adm_access_t *temp1 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int val5 ;
  int ecode5 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 4) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_adm_open2", 2, argv[0] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_wc_adm_open2", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  arg4 = RTEST(argv[2]);
  ecode5 = SWIG_AsVal_int(argv[3], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), Ruby_Format_TypeError( "", "int","svn_wc_adm_open2", 5, argv[3] ));
  } 
  arg5 = (int)(val5);
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_wc_adm_open2(arg1,arg2,(char const *)arg3,arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_wc_adm_access_t, 0), 0);
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_adm_open(int argc, VALUE *argv, VALUE self) {
  svn_wc_adm_access_t **arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  char *arg3 = 0 ;
  svn_boolean_t arg4 ;
  svn_boolean_t arg5 ;
  apr_pool_t *arg6 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_wc_adm_access_t *temp1 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 4) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_adm_open", 2, argv[0] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_wc_adm_open", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  arg4 = RTEST(argv[2]);
  arg5 = RTEST(argv[3]);
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_wc_adm_open(arg1,arg2,(char const *)arg3,arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_wc_adm_access_t, 0), 0);
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_adm_probe_open3(int argc, VALUE *argv, VALUE self) {
  svn_wc_adm_access_t **arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  char *arg3 = 0 ;
  svn_boolean_t arg4 ;
  int arg5 ;
  svn_cancel_func_t arg6 = 0 ;
  void *arg7 = 0 ;
  apr_pool_t *arg8 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_wc_adm_access_t *temp1 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int val5 ;
  int ecode5 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg8);
    _global_pool = arg8;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 5) || (argc > 6)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 5)",argc); SWIG_fail;
  }
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_adm_probe_open3", 2, argv[0] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_wc_adm_probe_open3", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  arg4 = RTEST(argv[2]);
  ecode5 = SWIG_AsVal_int(argv[3], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), Ruby_Format_TypeError( "", "int","svn_wc_adm_probe_open3", 5, argv[3] ));
  } 
  arg5 = (int)(val5);
  {
    arg6 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg7 = (void *)svn_swig_rb_make_baton(argv[4], _global_svn_swig_rb_pool);
  }
  if (argc > 5) {
    
  }
  {
    result = (svn_error_t *)svn_wc_adm_probe_open3(arg1,arg2,(char const *)arg3,arg4,arg5,arg6,arg7,arg8);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_wc_adm_access_t, 0), 0);
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg7);
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_adm_probe_open2(int argc, VALUE *argv, VALUE self) {
  svn_wc_adm_access_t **arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  char *arg3 = 0 ;
  svn_boolean_t arg4 ;
  int arg5 ;
  apr_pool_t *arg6 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_wc_adm_access_t *temp1 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int val5 ;
  int ecode5 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 4) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_adm_probe_open2", 2, argv[0] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_wc_adm_probe_open2", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  arg4 = RTEST(argv[2]);
  ecode5 = SWIG_AsVal_int(argv[3], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), Ruby_Format_TypeError( "", "int","svn_wc_adm_probe_open2", 5, argv[3] ));
  } 
  arg5 = (int)(val5);
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_wc_adm_probe_open2(arg1,arg2,(char const *)arg3,arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_wc_adm_access_t, 0), 0);
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_adm_probe_open(int argc, VALUE *argv, VALUE self) {
  svn_wc_adm_access_t **arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  char *arg3 = 0 ;
  svn_boolean_t arg4 ;
  svn_boolean_t arg5 ;
  apr_pool_t *arg6 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_wc_adm_access_t *temp1 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 4) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_adm_probe_open", 2, argv[0] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_wc_adm_probe_open", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  arg4 = RTEST(argv[2]);
  arg5 = RTEST(argv[3]);
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_wc_adm_probe_open(arg1,arg2,(char const *)arg3,arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_wc_adm_access_t, 0), 0);
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_adm_open_anchor(int argc, VALUE *argv, VALUE self) {
  svn_wc_adm_access_t **arg1 = 0 ;
  svn_wc_adm_access_t **arg2 = 0 ;
  char **arg3 = 0 ;
  char *arg4 = 0 ;
  svn_boolean_t arg5 ;
  int arg6 ;
  svn_cancel_func_t arg7 = 0 ;
  void *arg8 = 0 ;
  apr_pool_t *arg9 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_wc_adm_access_t *temp1 ;
  svn_wc_adm_access_t *temp2 ;
  char *temp3 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  int val6 ;
  int ecode6 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg9);
    _global_pool = arg9;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  arg2 = &temp2;
  arg3 = &temp3;
  if ((argc < 4) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  res4 = SWIG_AsCharPtrAndSize(argv[0], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_wc_adm_open_anchor", 4, argv[0] ));
  }
  arg4 = (char *)(buf4);
  arg5 = RTEST(argv[1]);
  ecode6 = SWIG_AsVal_int(argv[2], &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), Ruby_Format_TypeError( "", "int","svn_wc_adm_open_anchor", 6, argv[2] ));
  } 
  arg6 = (int)(val6);
  {
    arg7 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg8 = (void *)svn_swig_rb_make_baton(argv[3], _global_svn_swig_rb_pool);
  }
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_wc_adm_open_anchor(arg1,arg2,(char const **)arg3,(char const *)arg4,arg5,arg6,arg7,arg8,arg9);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_wc_adm_access_t, 0), 0);
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg2, SWIGTYPE_p_svn_wc_adm_access_t, 0), 0);
  }
  {
    if (*arg3) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new2(*arg3), 0);
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil, 0);
    }
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg8);
  }
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_adm_retrieve(int argc, VALUE *argv, VALUE self) {
  svn_wc_adm_access_t **arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  char *arg3 = 0 ;
  apr_pool_t *arg4 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_wc_adm_access_t *temp1 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_adm_retrieve", 2, argv[0] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_wc_adm_retrieve", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_wc_adm_retrieve(arg1,arg2,(char const *)arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_wc_adm_access_t, 0), 0);
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_adm_probe_retrieve(int argc, VALUE *argv, VALUE self) {
  svn_wc_adm_access_t **arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  char *arg3 = 0 ;
  apr_pool_t *arg4 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_wc_adm_access_t *temp1 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_adm_probe_retrieve", 2, argv[0] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_wc_adm_probe_retrieve", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_wc_adm_probe_retrieve(arg1,arg2,(char const *)arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_wc_adm_access_t, 0), 0);
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_adm_probe_try3(int argc, VALUE *argv, VALUE self) {
  svn_wc_adm_access_t **arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  char *arg3 = 0 ;
  svn_boolean_t arg4 ;
  int arg5 ;
  svn_cancel_func_t arg6 = 0 ;
  void *arg7 = 0 ;
  apr_pool_t *arg8 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_wc_adm_access_t *temp1 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int val5 ;
  int ecode5 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg8);
    _global_pool = arg8;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 5) || (argc > 6)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 5)",argc); SWIG_fail;
  }
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_adm_probe_try3", 2, argv[0] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_wc_adm_probe_try3", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  arg4 = RTEST(argv[2]);
  ecode5 = SWIG_AsVal_int(argv[3], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), Ruby_Format_TypeError( "", "int","svn_wc_adm_probe_try3", 5, argv[3] ));
  } 
  arg5 = (int)(val5);
  {
    arg6 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg7 = (void *)svn_swig_rb_make_baton(argv[4], _global_svn_swig_rb_pool);
  }
  if (argc > 5) {
    
  }
  {
    result = (svn_error_t *)svn_wc_adm_probe_try3(arg1,arg2,(char const *)arg3,arg4,arg5,arg6,arg7,arg8);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_wc_adm_access_t, 0), 0);
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg7);
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_adm_probe_try2(int argc, VALUE *argv, VALUE self) {
  svn_wc_adm_access_t **arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  char *arg3 = 0 ;
  svn_boolean_t arg4 ;
  int arg5 ;
  apr_pool_t *arg6 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_wc_adm_access_t *temp1 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int val5 ;
  int ecode5 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 4) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_adm_probe_try2", 2, argv[0] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_wc_adm_probe_try2", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  arg4 = RTEST(argv[2]);
  ecode5 = SWIG_AsVal_int(argv[3], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), Ruby_Format_TypeError( "", "int","svn_wc_adm_probe_try2", 5, argv[3] ));
  } 
  arg5 = (int)(val5);
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_wc_adm_probe_try2(arg1,arg2,(char const *)arg3,arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_wc_adm_access_t, 0), 0);
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_adm_probe_try(int argc, VALUE *argv, VALUE self) {
  svn_wc_adm_access_t **arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  char *arg3 = 0 ;
  svn_boolean_t arg4 ;
  svn_boolean_t arg5 ;
  apr_pool_t *arg6 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_wc_adm_access_t *temp1 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 4) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_adm_probe_try", 2, argv[0] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_wc_adm_probe_try", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  arg4 = RTEST(argv[2]);
  arg5 = RTEST(argv[3]);
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_wc_adm_probe_try(arg1,arg2,(char const *)arg3,arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_wc_adm_access_t, 0), 0);
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_adm_close2(int argc, VALUE *argv, VALUE self) {
  svn_wc_adm_access_t *arg1 = 0 ;
  apr_pool_t *arg2 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_adm_close2", 1, argv[0] )); 
  }
  arg1 = (svn_wc_adm_access_t *)(argp1);
  if (argc > 1) {
    
  }
  {
    result = (svn_error_t *)svn_wc_adm_close2(arg1,arg2);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_adm_close(int argc, VALUE *argv, VALUE self) {
  svn_wc_adm_access_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_adm_close", 1, argv[0] )); 
  }
  arg1 = (svn_wc_adm_access_t *)(argp1);
  {
    result = (svn_error_t *)svn_wc_adm_close(arg1);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_adm_access_path(int argc, VALUE *argv, VALUE self) {
  svn_wc_adm_access_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_adm_access_t const *","svn_wc_adm_access_path", 1, argv[0] )); 
  }
  arg1 = (svn_wc_adm_access_t *)(argp1);
  {
    result = (char *)svn_wc_adm_access_path((struct svn_wc_adm_access_t const *)arg1);
    
    
    
  }
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_adm_access_pool(int argc, VALUE *argv, VALUE self) {
  svn_wc_adm_access_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  apr_pool_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_adm_access_t const *","svn_wc_adm_access_pool", 1, argv[0] )); 
  }
  arg1 = (svn_wc_adm_access_t *)(argp1);
  {
    result = (apr_pool_t *)svn_wc_adm_access_pool((struct svn_wc_adm_access_t const *)arg1);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_apr_pool_t, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_adm_locked(int argc, VALUE *argv, VALUE self) {
  svn_wc_adm_access_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_adm_access_t const *","svn_wc_adm_locked", 1, argv[0] )); 
  }
  arg1 = (svn_wc_adm_access_t *)(argp1);
  {
    result = (svn_boolean_t)svn_wc_adm_locked((struct svn_wc_adm_access_t const *)arg1);
    
    
    
  }
  vresult = result ? Qtrue : Qfalse;
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_locked2(int argc, VALUE *argv, VALUE self) {
  svn_boolean_t *arg1 = 0 ;
  svn_boolean_t *arg2 = 0 ;
  svn_wc_context_t *arg3 = 0 ;
  char *arg4 = 0 ;
  apr_pool_t *arg5 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_boolean_t temp1 ;
  svn_boolean_t temp2 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  arg2 = &temp2;
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res3 = SWIG_ConvertPtr(argv[0], &argp3,SWIGTYPE_p_svn_wc_context_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_wc_context_t *","svn_wc_locked2", 3, argv[0] )); 
  }
  arg3 = (svn_wc_context_t *)(argp3);
  res4 = SWIG_AsCharPtrAndSize(argv[1], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_wc_locked2", 4, argv[1] ));
  }
  arg4 = (char *)(buf4);
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_wc_locked2(arg1,arg2,arg3,(char const *)arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, *arg1 ? Qtrue : Qfalse, 0);
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, *arg2 ? Qtrue : Qfalse, 0);
  }
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_locked(int argc, VALUE *argv, VALUE self) {
  svn_boolean_t *arg1 = 0 ;
  char *arg2 = 0 ;
  apr_pool_t *arg3 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_boolean_t temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_wc_locked", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  if (argc > 1) {
    
  }
  {
    result = (svn_error_t *)svn_wc_locked(arg1,(char const *)arg2,arg3);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, *arg1 ? Qtrue : Qfalse, 0);
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_is_adm_dir(int argc, VALUE *argv, VALUE self) {
  char *arg1 = 0 ;
  apr_pool_t *arg2 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_wc_is_adm_dir", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  if (argc > 1) {
    
  }
  {
    result = (svn_boolean_t)svn_wc_is_adm_dir((char const *)arg1,arg2);
    
    
    
  }
  vresult = result ? Qtrue : Qfalse;
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_get_adm_dir(int argc, VALUE *argv, VALUE self) {
  apr_pool_t *arg1 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg1);
    _global_pool = arg1;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 0) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  if (argc > 0) {
    
  }
  {
    result = (char *)svn_wc_get_adm_dir(arg1);
    
    
    
  }
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_set_adm_dir(int argc, VALUE *argv, VALUE self) {
  char *arg1 = 0 ;
  apr_pool_t *arg2 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_wc_set_adm_dir", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  if (argc > 1) {
    
  }
  {
    result = (svn_error_t *)svn_wc_set_adm_dir((char const *)arg1,arg2);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_edited_externals(int argc, VALUE *argv, VALUE self) {
  apr_hash_t **arg1 = 0 ;
  apr_hash_t **arg2 = 0 ;
  svn_wc_traversal_info_t *arg3 = 0 ;
  apr_hash_t *temp1 ;
  apr_hash_t *temp2 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  VALUE vresult = Qnil;
  
  arg1 = &temp1;
  arg2 = &temp2;
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res3 = SWIG_ConvertPtr(argv[0], &argp3,SWIGTYPE_p_svn_wc_traversal_info_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_wc_traversal_info_t *","svn_wc_edited_externals", 3, argv[0] )); 
  }
  arg3 = (svn_wc_traversal_info_t *)(argp3);
  {
    svn_wc_edited_externals(arg1,arg2,arg3);
    
    
    
  }
  vresult = rb_ary_new();
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, svn_swig_rb_apr_hash_to_hash_string(*arg1), 1);
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, svn_swig_rb_apr_hash_to_hash_string(*arg2), 1);
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_traversed_depths(int argc, VALUE *argv, VALUE self) {
  apr_hash_t **arg1 = 0 ;
  svn_wc_traversal_info_t *arg2 = 0 ;
  apr_hash_t *temp1 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  VALUE vresult = Qnil;
  
  arg1 = &temp1;
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_svn_wc_traversal_info_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_traversal_info_t *","svn_wc_traversed_depths", 2, argv[0] )); 
  }
  arg2 = (svn_wc_traversal_info_t *)(argp2);
  {
    svn_wc_traversed_depths(arg1,arg2);
    
    
    
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, svn_swig_rb_apr_hash_to_hash_string(*arg1), 1);
  }
  return vresult;
fail:
  return Qnil;
}


static swig_class SwigClassSvn_wc_external_item2_t;

SWIGINTERN VALUE
_wrap_svn_wc_external_item2_t_target_dir_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_external_item2_t *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_external_item2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_external_item2_t *","target_dir", 1, self )); 
  }
  arg1 = (struct svn_wc_external_item2_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","target_dir", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->target_dir) free((char *)arg1->target_dir);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->target_dir = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_external_item2_t_target_dir_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_external_item2_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_external_item2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_external_item2_t *","target_dir", 1, self )); 
  }
  arg1 = (struct svn_wc_external_item2_t *)(argp1);
  result = (char *) ((arg1)->target_dir);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_external_item2_t_url_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_external_item2_t *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_external_item2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_external_item2_t *","url", 1, self )); 
  }
  arg1 = (struct svn_wc_external_item2_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","url", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->url) free((char *)arg1->url);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->url = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_external_item2_t_url_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_external_item2_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_external_item2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_external_item2_t *","url", 1, self )); 
  }
  arg1 = (struct svn_wc_external_item2_t *)(argp1);
  result = (char *) ((arg1)->url);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_external_item2_t_revision_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_external_item2_t *arg1 = 0 ;
  svn_opt_revision_t *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_opt_revision_t rev2 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_external_item2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_external_item2_t *","revision", 1, self )); 
  }
  arg1 = (struct svn_wc_external_item2_t *)(argp1);
  {
    arg2 = &rev2;
    svn_swig_rb_set_revision(&rev2, argv[0]);
  }
  if (arg1) (arg1)->revision = *arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_external_item2_t_revision_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_external_item2_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_opt_revision_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_external_item2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_external_item2_t *","revision", 1, self )); 
  }
  arg1 = (struct svn_wc_external_item2_t *)(argp1);
  result = (svn_opt_revision_t *)& ((arg1)->revision);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_opt_revision_t, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_external_item2_t_peg_revision_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_external_item2_t *arg1 = 0 ;
  svn_opt_revision_t *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_opt_revision_t rev2 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_external_item2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_external_item2_t *","peg_revision", 1, self )); 
  }
  arg1 = (struct svn_wc_external_item2_t *)(argp1);
  {
    arg2 = &rev2;
    svn_swig_rb_set_revision(&rev2, argv[0]);
  }
  if (arg1) (arg1)->peg_revision = *arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_external_item2_t_peg_revision_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_external_item2_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_opt_revision_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_external_item2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_external_item2_t *","peg_revision", 1, self )); 
  }
  arg1 = (struct svn_wc_external_item2_t *)(argp1);
  result = (svn_opt_revision_t *)& ((arg1)->peg_revision);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_opt_revision_t, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
_wrap_svn_wc_external_item2_t_allocate(VALUE self)
#else
_wrap_svn_wc_external_item2_t_allocate(int argc, VALUE *argv, VALUE self)
#endif
{
  VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_svn_wc_external_item2_t);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
  rb_obj_call_init(vresult, argc, argv);
#endif
  return vresult;
}


SWIGINTERN VALUE
_wrap_new_svn_wc_external_item2_t(int argc, VALUE *argv, VALUE self) {
  apr_pool_t *arg1 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  struct svn_wc_external_item2_t *result = 0 ;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg1);
    _global_pool = arg1;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 0) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  if (argc > 0) {
    
  }
  {
    result = (struct svn_wc_external_item2_t *)new_svn_wc_external_item2_t(arg1);
    DATA_PTR(self) = result;
    
    
    
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return self;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN void delete_svn_wc_external_item2_t(struct svn_wc_external_item2_t *self){
  }
SWIGINTERN void
free_svn_wc_external_item2_t(void *self) {
    struct svn_wc_external_item2_t *arg1 = (struct svn_wc_external_item2_t *)self;
    delete_svn_wc_external_item2_t(arg1);
}

/*
  Document-method: Svn::Ext::Wc::svn_wc_external_item2_t.dup

  call-seq:
    dup(pool) -> svn_wc_external_item2_t

Create a duplicate of the class and unfreeze it if needed.
*/
SWIGINTERN VALUE
_wrap_svn_wc_external_item2_t_dup(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_external_item2_t *arg1 = 0 ;
  apr_pool_t *arg2 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_wc_external_item2_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 0) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_external_item2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_external_item2_t *","dup", 1, self )); 
  }
  arg1 = (struct svn_wc_external_item2_t *)(argp1);
  if (argc > 0) {
    
  }
  {
    result = (svn_wc_external_item2_t *)svn_wc_external_item2_t_dup(arg1,arg2);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_wc_external_item2_t, 0 |  0 );
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


static swig_class SwigClassSvn_wc_external_item_t;

SWIGINTERN VALUE
_wrap_svn_wc_external_item_t_target_dir_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_external_item_t *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_external_item_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_external_item_t *","target_dir", 1, self )); 
  }
  arg1 = (struct svn_wc_external_item_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","target_dir", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->target_dir) free((char *)arg1->target_dir);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->target_dir = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_external_item_t_target_dir_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_external_item_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_external_item_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_external_item_t *","target_dir", 1, self )); 
  }
  arg1 = (struct svn_wc_external_item_t *)(argp1);
  result = (char *) ((arg1)->target_dir);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_external_item_t_url_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_external_item_t *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_external_item_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_external_item_t *","url", 1, self )); 
  }
  arg1 = (struct svn_wc_external_item_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","url", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->url) free((char *)arg1->url);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->url = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_external_item_t_url_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_external_item_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_external_item_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_external_item_t *","url", 1, self )); 
  }
  arg1 = (struct svn_wc_external_item_t *)(argp1);
  result = (char *) ((arg1)->url);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_external_item_t_revision_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_external_item_t *arg1 = 0 ;
  svn_opt_revision_t *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_opt_revision_t rev2 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_external_item_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_external_item_t *","revision", 1, self )); 
  }
  arg1 = (struct svn_wc_external_item_t *)(argp1);
  {
    arg2 = &rev2;
    svn_swig_rb_set_revision(&rev2, argv[0]);
  }
  if (arg1) (arg1)->revision = *arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_external_item_t_revision_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_external_item_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_opt_revision_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_external_item_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_external_item_t *","revision", 1, self )); 
  }
  arg1 = (struct svn_wc_external_item_t *)(argp1);
  result = (svn_opt_revision_t *)& ((arg1)->revision);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_opt_revision_t, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
_wrap_svn_wc_external_item_t_allocate(VALUE self)
#else
_wrap_svn_wc_external_item_t_allocate(int argc, VALUE *argv, VALUE self)
#endif
{
  VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_svn_wc_external_item_t);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
  rb_obj_call_init(vresult, argc, argv);
#endif
  return vresult;
}


SWIGINTERN VALUE
_wrap_new_svn_wc_external_item_t(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_external_item_t *result = 0 ;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  {
    result = (struct svn_wc_external_item_t *)calloc(1, sizeof(struct svn_wc_external_item_t));
    DATA_PTR(self) = result;
    
    
    
  }
  return self;
fail:
  return Qnil;
}


SWIGINTERN void
free_svn_wc_external_item_t(void *self) {
    struct svn_wc_external_item_t *arg1 = (struct svn_wc_external_item_t *)self;
    free((char *) arg1);
}

SWIGINTERN VALUE
_wrap_svn_wc_parse_externals_description3(int argc, VALUE *argv, VALUE self) {
  apr_array_header_t **arg1 = 0 ;
  char *arg2 = 0 ;
  char *arg3 = 0 ;
  svn_boolean_t arg4 ;
  apr_pool_t *arg5 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  apr_array_header_t *temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 3) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_wc_parse_externals_description3", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_wc_parse_externals_description3", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  arg4 = RTEST(argv[2]);
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_wc_parse_externals_description3(arg1,(char const *)arg2,(char const *)arg3,arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, svn_swig_rb_apr_array_to_array_external_item2(*arg1), 0);
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_parse_externals_description2(int argc, VALUE *argv, VALUE self) {
  apr_array_header_t **arg1 = 0 ;
  char *arg2 = 0 ;
  char *arg3 = 0 ;
  apr_pool_t *arg4 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  apr_array_header_t *temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_wc_parse_externals_description2", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_wc_parse_externals_description2", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_wc_parse_externals_description2(arg1,(char const *)arg2,(char const *)arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, svn_swig_rb_apr_array_to_array_external_item2(*arg1), 0);
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_parse_externals_description(int argc, VALUE *argv, VALUE self) {
  apr_hash_t **arg1 = 0 ;
  char *arg2 = 0 ;
  char *arg3 = 0 ;
  apr_pool_t *arg4 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  apr_hash_t *temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_wc_parse_externals_description", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_wc_parse_externals_description", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_wc_parse_externals_description(arg1,(char const *)arg2,(char const *)arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, svn_swig_rb_apr_hash_to_hash_swig_type(*arg1,
        "svn_wc_external_item_t *"), 0)
    
    ;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


static swig_class SwigClassSvn_wc_notify_t;

SWIGINTERN VALUE
_wrap_svn_wc_notify_t_path_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_notify_t *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_notify_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_notify_t *","path", 1, self )); 
  }
  arg1 = (struct svn_wc_notify_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","path", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->path) free((char *)arg1->path);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->path = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_notify_t_path_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_notify_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_notify_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_notify_t *","path", 1, self )); 
  }
  arg1 = (struct svn_wc_notify_t *)(argp1);
  result = (char *) ((arg1)->path);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_notify_t_action_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_notify_t *arg1 = 0 ;
  svn_wc_notify_action_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_notify_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_notify_t *","action", 1, self )); 
  }
  arg1 = (struct svn_wc_notify_t *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_wc_notify_action_t","action", 2, argv[0] ));
  } 
  arg2 = (svn_wc_notify_action_t)(val2);
  if (arg1) (arg1)->action = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_notify_t_action_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_notify_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_wc_notify_action_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_notify_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_notify_t *","action", 1, self )); 
  }
  arg1 = (struct svn_wc_notify_t *)(argp1);
  result = (svn_wc_notify_action_t) ((arg1)->action);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_notify_t_kind_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_notify_t *arg1 = 0 ;
  svn_node_kind_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_notify_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_notify_t *","kind", 1, self )); 
  }
  arg1 = (struct svn_wc_notify_t *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_node_kind_t","kind", 2, argv[0] ));
  } 
  arg2 = (svn_node_kind_t)(val2);
  if (arg1) (arg1)->kind = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_notify_t_kind_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_notify_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_node_kind_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_notify_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_notify_t *","kind", 1, self )); 
  }
  arg1 = (struct svn_wc_notify_t *)(argp1);
  result = (svn_node_kind_t) ((arg1)->kind);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_notify_t_mime_type_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_notify_t *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_notify_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_notify_t *","mime_type", 1, self )); 
  }
  arg1 = (struct svn_wc_notify_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","mime_type", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->mime_type) free((char *)arg1->mime_type);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->mime_type = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_notify_t_mime_type_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_notify_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_notify_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_notify_t *","mime_type", 1, self )); 
  }
  arg1 = (struct svn_wc_notify_t *)(argp1);
  result = (char *) ((arg1)->mime_type);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_notify_t_lock_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_notify_t *arg1 = 0 ;
  svn_lock_t *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_notify_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_notify_t *","lock", 1, self )); 
  }
  arg1 = (struct svn_wc_notify_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_svn_lock_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_lock_t const *","lock", 2, argv[0] )); 
  }
  arg2 = (svn_lock_t *)(argp2);
  if (arg1) (arg1)->lock = (svn_lock_t const *)arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_notify_t_lock_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_notify_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_lock_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_notify_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_notify_t *","lock", 1, self )); 
  }
  arg1 = (struct svn_wc_notify_t *)(argp1);
  result = (svn_lock_t *) ((arg1)->lock);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_lock_t, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_notify_t_err_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_notify_t *arg1 = 0 ;
  svn_error_t *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_notify_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_notify_t *","err", 1, self )); 
  }
  arg1 = (struct svn_wc_notify_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_svn_error_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_error_t *","err", 2, argv[0] )); 
  }
  arg2 = (svn_error_t *)(argp2);
  if (arg1) (arg1)->err = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_notify_t_err_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_notify_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_notify_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_notify_t *","err", 1, self )); 
  }
  arg1 = (struct svn_wc_notify_t *)(argp1);
  result = (svn_error_t *) ((arg1)->err);
  {
    vresult = result ? svn_swig_rb_svn_error_to_rb_error(result) : Qnil;
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_notify_t_content_state_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_notify_t *arg1 = 0 ;
  svn_wc_notify_state_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_notify_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_notify_t *","content_state", 1, self )); 
  }
  arg1 = (struct svn_wc_notify_t *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_wc_notify_state_t","content_state", 2, argv[0] ));
  } 
  arg2 = (svn_wc_notify_state_t)(val2);
  if (arg1) (arg1)->content_state = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_notify_t_content_state_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_notify_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_wc_notify_state_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_notify_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_notify_t *","content_state", 1, self )); 
  }
  arg1 = (struct svn_wc_notify_t *)(argp1);
  result = (svn_wc_notify_state_t) ((arg1)->content_state);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_notify_t_prop_state_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_notify_t *arg1 = 0 ;
  svn_wc_notify_state_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_notify_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_notify_t *","prop_state", 1, self )); 
  }
  arg1 = (struct svn_wc_notify_t *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_wc_notify_state_t","prop_state", 2, argv[0] ));
  } 
  arg2 = (svn_wc_notify_state_t)(val2);
  if (arg1) (arg1)->prop_state = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_notify_t_prop_state_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_notify_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_wc_notify_state_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_notify_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_notify_t *","prop_state", 1, self )); 
  }
  arg1 = (struct svn_wc_notify_t *)(argp1);
  result = (svn_wc_notify_state_t) ((arg1)->prop_state);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_notify_t_lock_state_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_notify_t *arg1 = 0 ;
  svn_wc_notify_lock_state_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_notify_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_notify_t *","lock_state", 1, self )); 
  }
  arg1 = (struct svn_wc_notify_t *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_wc_notify_lock_state_t","lock_state", 2, argv[0] ));
  } 
  arg2 = (svn_wc_notify_lock_state_t)(val2);
  if (arg1) (arg1)->lock_state = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_notify_t_lock_state_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_notify_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_wc_notify_lock_state_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_notify_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_notify_t *","lock_state", 1, self )); 
  }
  arg1 = (struct svn_wc_notify_t *)(argp1);
  result = (svn_wc_notify_lock_state_t) ((arg1)->lock_state);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_notify_t_revision_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_notify_t *arg1 = 0 ;
  svn_revnum_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_notify_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_notify_t *","revision", 1, self )); 
  }
  arg1 = (struct svn_wc_notify_t *)(argp1);
  ecode2 = SWIG_AsVal_long(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_revnum_t","revision", 2, argv[0] ));
  } 
  arg2 = (svn_revnum_t)(val2);
  if (arg1) (arg1)->revision = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_notify_t_revision_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_notify_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_revnum_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_notify_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_notify_t *","revision", 1, self )); 
  }
  arg1 = (struct svn_wc_notify_t *)(argp1);
  result = (svn_revnum_t) ((arg1)->revision);
  vresult = SWIG_From_long((long)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_notify_t_changelist_name_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_notify_t *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_notify_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_notify_t *","changelist_name", 1, self )); 
  }
  arg1 = (struct svn_wc_notify_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","changelist_name", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->changelist_name) free((char *)arg1->changelist_name);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->changelist_name = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_notify_t_changelist_name_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_notify_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_notify_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_notify_t *","changelist_name", 1, self )); 
  }
  arg1 = (struct svn_wc_notify_t *)(argp1);
  result = (char *) ((arg1)->changelist_name);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_notify_t_merge_range_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_notify_t *arg1 = 0 ;
  svn_merge_range_t *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_notify_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_notify_t *","merge_range", 1, self )); 
  }
  arg1 = (struct svn_wc_notify_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_svn_merge_range_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_merge_range_t *","merge_range", 2, argv[0] )); 
  }
  arg2 = (svn_merge_range_t *)(argp2);
  if (arg1) (arg1)->merge_range = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_notify_t_merge_range_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_notify_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_merge_range_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_notify_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_notify_t *","merge_range", 1, self )); 
  }
  arg1 = (struct svn_wc_notify_t *)(argp1);
  result = (svn_merge_range_t *) ((arg1)->merge_range);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_merge_range_t, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_notify_t_url_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_notify_t *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_notify_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_notify_t *","url", 1, self )); 
  }
  arg1 = (struct svn_wc_notify_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","url", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->url) free((char *)arg1->url);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->url = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_notify_t_url_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_notify_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_notify_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_notify_t *","url", 1, self )); 
  }
  arg1 = (struct svn_wc_notify_t *)(argp1);
  result = (char *) ((arg1)->url);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_notify_t_path_prefix_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_notify_t *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_notify_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_notify_t *","path_prefix", 1, self )); 
  }
  arg1 = (struct svn_wc_notify_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","path_prefix", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->path_prefix) free((char *)arg1->path_prefix);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->path_prefix = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_notify_t_path_prefix_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_notify_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_notify_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_notify_t *","path_prefix", 1, self )); 
  }
  arg1 = (struct svn_wc_notify_t *)(argp1);
  result = (char *) ((arg1)->path_prefix);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_notify_t_prop_name_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_notify_t *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_notify_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_notify_t *","prop_name", 1, self )); 
  }
  arg1 = (struct svn_wc_notify_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","prop_name", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->prop_name) free((char *)arg1->prop_name);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->prop_name = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_notify_t_prop_name_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_notify_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_notify_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_notify_t *","prop_name", 1, self )); 
  }
  arg1 = (struct svn_wc_notify_t *)(argp1);
  result = (char *) ((arg1)->prop_name);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_notify_t_rev_props_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_notify_t *arg1 = 0 ;
  apr_hash_t *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_notify_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_notify_t *","rev_props", 1, self )); 
  }
  arg1 = (struct svn_wc_notify_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_apr_hash_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "apr_hash_t *","rev_props", 2, argv[0] )); 
  }
  arg2 = (apr_hash_t *)(argp2);
  if (arg1) (arg1)->rev_props = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_notify_t_rev_props_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_notify_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  apr_hash_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_notify_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_notify_t *","rev_props", 1, self )); 
  }
  arg1 = (struct svn_wc_notify_t *)(argp1);
  result = (apr_hash_t *) ((arg1)->rev_props);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_apr_hash_t, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_notify_t_old_revision_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_notify_t *arg1 = 0 ;
  svn_revnum_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_notify_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_notify_t *","old_revision", 1, self )); 
  }
  arg1 = (struct svn_wc_notify_t *)(argp1);
  ecode2 = SWIG_AsVal_long(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_revnum_t","old_revision", 2, argv[0] ));
  } 
  arg2 = (svn_revnum_t)(val2);
  if (arg1) (arg1)->old_revision = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_notify_t_old_revision_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_notify_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_revnum_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_notify_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_notify_t *","old_revision", 1, self )); 
  }
  arg1 = (struct svn_wc_notify_t *)(argp1);
  result = (svn_revnum_t) ((arg1)->old_revision);
  vresult = SWIG_From_long((long)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_notify_t_hunk_original_start_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_notify_t *arg1 = 0 ;
  svn_linenum_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_notify_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_notify_t *","hunk_original_start", 1, self )); 
  }
  arg1 = (struct svn_wc_notify_t *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_linenum_t","hunk_original_start", 2, argv[0] ));
  } 
  arg2 = (svn_linenum_t)(val2);
  if (arg1) (arg1)->hunk_original_start = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_notify_t_hunk_original_start_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_notify_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_linenum_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_notify_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_notify_t *","hunk_original_start", 1, self )); 
  }
  arg1 = (struct svn_wc_notify_t *)(argp1);
  result = (svn_linenum_t) ((arg1)->hunk_original_start);
  vresult = SWIG_From_unsigned_SS_long((unsigned long)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_notify_t_hunk_original_length_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_notify_t *arg1 = 0 ;
  svn_linenum_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_notify_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_notify_t *","hunk_original_length", 1, self )); 
  }
  arg1 = (struct svn_wc_notify_t *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_linenum_t","hunk_original_length", 2, argv[0] ));
  } 
  arg2 = (svn_linenum_t)(val2);
  if (arg1) (arg1)->hunk_original_length = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_notify_t_hunk_original_length_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_notify_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_linenum_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_notify_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_notify_t *","hunk_original_length", 1, self )); 
  }
  arg1 = (struct svn_wc_notify_t *)(argp1);
  result = (svn_linenum_t) ((arg1)->hunk_original_length);
  vresult = SWIG_From_unsigned_SS_long((unsigned long)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_notify_t_hunk_modified_start_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_notify_t *arg1 = 0 ;
  svn_linenum_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_notify_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_notify_t *","hunk_modified_start", 1, self )); 
  }
  arg1 = (struct svn_wc_notify_t *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_linenum_t","hunk_modified_start", 2, argv[0] ));
  } 
  arg2 = (svn_linenum_t)(val2);
  if (arg1) (arg1)->hunk_modified_start = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_notify_t_hunk_modified_start_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_notify_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_linenum_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_notify_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_notify_t *","hunk_modified_start", 1, self )); 
  }
  arg1 = (struct svn_wc_notify_t *)(argp1);
  result = (svn_linenum_t) ((arg1)->hunk_modified_start);
  vresult = SWIG_From_unsigned_SS_long((unsigned long)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_notify_t_hunk_modified_length_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_notify_t *arg1 = 0 ;
  svn_linenum_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_notify_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_notify_t *","hunk_modified_length", 1, self )); 
  }
  arg1 = (struct svn_wc_notify_t *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_linenum_t","hunk_modified_length", 2, argv[0] ));
  } 
  arg2 = (svn_linenum_t)(val2);
  if (arg1) (arg1)->hunk_modified_length = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_notify_t_hunk_modified_length_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_notify_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_linenum_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_notify_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_notify_t *","hunk_modified_length", 1, self )); 
  }
  arg1 = (struct svn_wc_notify_t *)(argp1);
  result = (svn_linenum_t) ((arg1)->hunk_modified_length);
  vresult = SWIG_From_unsigned_SS_long((unsigned long)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_notify_t_hunk_matched_line_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_notify_t *arg1 = 0 ;
  svn_linenum_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_notify_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_notify_t *","hunk_matched_line", 1, self )); 
  }
  arg1 = (struct svn_wc_notify_t *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_linenum_t","hunk_matched_line", 2, argv[0] ));
  } 
  arg2 = (svn_linenum_t)(val2);
  if (arg1) (arg1)->hunk_matched_line = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_notify_t_hunk_matched_line_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_notify_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_linenum_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_notify_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_notify_t *","hunk_matched_line", 1, self )); 
  }
  arg1 = (struct svn_wc_notify_t *)(argp1);
  result = (svn_linenum_t) ((arg1)->hunk_matched_line);
  vresult = SWIG_From_unsigned_SS_long((unsigned long)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_notify_t_hunk_fuzz_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_notify_t *arg1 = 0 ;
  svn_linenum_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_notify_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_notify_t *","hunk_fuzz", 1, self )); 
  }
  arg1 = (struct svn_wc_notify_t *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_linenum_t","hunk_fuzz", 2, argv[0] ));
  } 
  arg2 = (svn_linenum_t)(val2);
  if (arg1) (arg1)->hunk_fuzz = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_notify_t_hunk_fuzz_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_notify_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_linenum_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_notify_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_notify_t *","hunk_fuzz", 1, self )); 
  }
  arg1 = (struct svn_wc_notify_t *)(argp1);
  result = (svn_linenum_t) ((arg1)->hunk_fuzz);
  vresult = SWIG_From_unsigned_SS_long((unsigned long)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
_wrap_svn_wc_notify_t_allocate(VALUE self)
#else
_wrap_svn_wc_notify_t_allocate(int argc, VALUE *argv, VALUE self)
#endif
{
  VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_svn_wc_notify_t);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
  rb_obj_call_init(vresult, argc, argv);
#endif
  return vresult;
}


SWIGINTERN VALUE
_wrap_new_svn_wc_notify_t(int argc, VALUE *argv, VALUE self) {
  char *arg1 = 0 ;
  svn_wc_notify_action_t arg2 ;
  apr_pool_t *arg3 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  struct svn_wc_notify_t *result = 0 ;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_wc_notify_t", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  ecode2 = SWIG_AsVal_int(argv[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_wc_notify_action_t","svn_wc_notify_t", 2, argv[1] ));
  } 
  arg2 = (svn_wc_notify_action_t)(val2);
  if (argc > 2) {
    
  }
  {
    result = (struct svn_wc_notify_t *)new_svn_wc_notify_t((char const *)arg1,arg2,arg3);
    DATA_PTR(self) = result;
    
    
    
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return self;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN void delete_svn_wc_notify_t(struct svn_wc_notify_t *self){
  }
SWIGINTERN void
free_svn_wc_notify_t(void *self) {
    struct svn_wc_notify_t *arg1 = (struct svn_wc_notify_t *)self;
    delete_svn_wc_notify_t(arg1);
}

SWIGINTERN VALUE
_wrap_svn_wc_create_notify(int argc, VALUE *argv, VALUE self) {
  char *arg1 = 0 ;
  svn_wc_notify_action_t arg2 ;
  apr_pool_t *arg3 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  svn_wc_notify_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_wc_create_notify", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  ecode2 = SWIG_AsVal_int(argv[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_wc_notify_action_t","svn_wc_create_notify", 2, argv[1] ));
  } 
  arg2 = (svn_wc_notify_action_t)(val2);
  if (argc > 2) {
    
  }
  {
    result = (svn_wc_notify_t *)svn_wc_create_notify((char const *)arg1,arg2,arg3);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_wc_notify_t, 0 |  0 );
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_create_notify_url(int argc, VALUE *argv, VALUE self) {
  char *arg1 = 0 ;
  svn_wc_notify_action_t arg2 ;
  apr_pool_t *arg3 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  svn_wc_notify_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_wc_create_notify_url", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  ecode2 = SWIG_AsVal_int(argv[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_wc_notify_action_t","svn_wc_create_notify_url", 2, argv[1] ));
  } 
  arg2 = (svn_wc_notify_action_t)(val2);
  if (argc > 2) {
    
  }
  {
    result = (svn_wc_notify_t *)svn_wc_create_notify_url((char const *)arg1,arg2,arg3);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_wc_notify_t, 0 |  0 );
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_dup_notify(int argc, VALUE *argv, VALUE self) {
  svn_wc_notify_t *arg1 = 0 ;
  apr_pool_t *arg2 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_wc_notify_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_notify_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_notify_t const *","svn_wc_dup_notify", 1, argv[0] )); 
  }
  arg1 = (svn_wc_notify_t *)(argp1);
  if (argc > 1) {
    
  }
  {
    result = (svn_wc_notify_t *)svn_wc_dup_notify((struct svn_wc_notify_t const *)arg1,arg2);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_wc_notify_t, 0 |  0 );
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


static swig_class SwigClassSvn_wc_conflict_version_t;

SWIGINTERN VALUE
_wrap_svn_wc_conflict_version_t_repos_url_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_conflict_version_t *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_conflict_version_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_conflict_version_t *","repos_url", 1, self )); 
  }
  arg1 = (struct svn_wc_conflict_version_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","repos_url", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->repos_url) free((char *)arg1->repos_url);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->repos_url = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_conflict_version_t_repos_url_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_conflict_version_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_conflict_version_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_conflict_version_t *","repos_url", 1, self )); 
  }
  arg1 = (struct svn_wc_conflict_version_t *)(argp1);
  result = (char *) ((arg1)->repos_url);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_conflict_version_t_peg_rev_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_conflict_version_t *arg1 = 0 ;
  svn_revnum_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_conflict_version_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_conflict_version_t *","peg_rev", 1, self )); 
  }
  arg1 = (struct svn_wc_conflict_version_t *)(argp1);
  ecode2 = SWIG_AsVal_long(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_revnum_t","peg_rev", 2, argv[0] ));
  } 
  arg2 = (svn_revnum_t)(val2);
  if (arg1) (arg1)->peg_rev = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_conflict_version_t_peg_rev_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_conflict_version_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_revnum_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_conflict_version_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_conflict_version_t *","peg_rev", 1, self )); 
  }
  arg1 = (struct svn_wc_conflict_version_t *)(argp1);
  result = (svn_revnum_t) ((arg1)->peg_rev);
  vresult = SWIG_From_long((long)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_conflict_version_t_path_in_repos_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_conflict_version_t *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_conflict_version_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_conflict_version_t *","path_in_repos", 1, self )); 
  }
  arg1 = (struct svn_wc_conflict_version_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","path_in_repos", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->path_in_repos) free((char *)arg1->path_in_repos);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->path_in_repos = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_conflict_version_t_path_in_repos_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_conflict_version_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_conflict_version_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_conflict_version_t *","path_in_repos", 1, self )); 
  }
  arg1 = (struct svn_wc_conflict_version_t *)(argp1);
  result = (char *) ((arg1)->path_in_repos);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_conflict_version_t_node_kind_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_conflict_version_t *arg1 = 0 ;
  svn_node_kind_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_conflict_version_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_conflict_version_t *","node_kind", 1, self )); 
  }
  arg1 = (struct svn_wc_conflict_version_t *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_node_kind_t","node_kind", 2, argv[0] ));
  } 
  arg2 = (svn_node_kind_t)(val2);
  if (arg1) (arg1)->node_kind = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_conflict_version_t_node_kind_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_conflict_version_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_node_kind_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_conflict_version_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_conflict_version_t *","node_kind", 1, self )); 
  }
  arg1 = (struct svn_wc_conflict_version_t *)(argp1);
  result = (svn_node_kind_t) ((arg1)->node_kind);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_conflict_version_t_repos_uuid_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_conflict_version_t *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_conflict_version_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_conflict_version_t *","repos_uuid", 1, self )); 
  }
  arg1 = (struct svn_wc_conflict_version_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","repos_uuid", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->repos_uuid) free((char *)arg1->repos_uuid);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->repos_uuid = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_conflict_version_t_repos_uuid_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_conflict_version_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_conflict_version_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_conflict_version_t *","repos_uuid", 1, self )); 
  }
  arg1 = (struct svn_wc_conflict_version_t *)(argp1);
  result = (char *) ((arg1)->repos_uuid);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
_wrap_svn_wc_conflict_version_t_allocate(VALUE self)
#else
_wrap_svn_wc_conflict_version_t_allocate(int argc, VALUE *argv, VALUE self)
#endif
{
  VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_svn_wc_conflict_version_t);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
  rb_obj_call_init(vresult, argc, argv);
#endif
  return vresult;
}


SWIGINTERN VALUE
_wrap_new_svn_wc_conflict_version_t(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_conflict_version_t *result = 0 ;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  {
    result = (struct svn_wc_conflict_version_t *)calloc(1, sizeof(struct svn_wc_conflict_version_t));
    DATA_PTR(self) = result;
    
    
    
  }
  return self;
fail:
  return Qnil;
}


SWIGINTERN void
free_svn_wc_conflict_version_t(void *self) {
    struct svn_wc_conflict_version_t *arg1 = (struct svn_wc_conflict_version_t *)self;
    free((char *) arg1);
}

SWIGINTERN VALUE
_wrap_svn_wc_conflict_version_create2(int argc, VALUE *argv, VALUE self) {
  char *arg1 = 0 ;
  char *arg2 = 0 ;
  char *arg3 = 0 ;
  svn_revnum_t arg4 ;
  svn_node_kind_t arg5 ;
  apr_pool_t *arg6 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  long val4 ;
  int ecode4 = 0 ;
  int val5 ;
  int ecode5 = 0 ;
  svn_wc_conflict_version_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 5) || (argc > 6)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 5)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_wc_conflict_version_create2", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_wc_conflict_version_create2", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_wc_conflict_version_create2", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  ecode4 = SWIG_AsVal_long(argv[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), Ruby_Format_TypeError( "", "svn_revnum_t","svn_wc_conflict_version_create2", 4, argv[3] ));
  } 
  arg4 = (svn_revnum_t)(val4);
  ecode5 = SWIG_AsVal_int(argv[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), Ruby_Format_TypeError( "", "svn_node_kind_t","svn_wc_conflict_version_create2", 5, argv[4] ));
  } 
  arg5 = (svn_node_kind_t)(val5);
  if (argc > 5) {
    
  }
  {
    result = (svn_wc_conflict_version_t *)svn_wc_conflict_version_create2((char const *)arg1,(char const *)arg2,(char const *)arg3,arg4,arg5,arg6);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_wc_conflict_version_t, 0 |  0 );
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_conflict_version_create(int argc, VALUE *argv, VALUE self) {
  char *arg1 = 0 ;
  char *arg2 = 0 ;
  svn_revnum_t arg3 ;
  svn_node_kind_t arg4 ;
  apr_pool_t *arg5 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  long val3 ;
  int ecode3 = 0 ;
  int val4 ;
  int ecode4 = 0 ;
  svn_wc_conflict_version_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 4) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_wc_conflict_version_create", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_wc_conflict_version_create", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  ecode3 = SWIG_AsVal_long(argv[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), Ruby_Format_TypeError( "", "svn_revnum_t","svn_wc_conflict_version_create", 3, argv[2] ));
  } 
  arg3 = (svn_revnum_t)(val3);
  ecode4 = SWIG_AsVal_int(argv[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), Ruby_Format_TypeError( "", "svn_node_kind_t","svn_wc_conflict_version_create", 4, argv[3] ));
  } 
  arg4 = (svn_node_kind_t)(val4);
  if (argc > 4) {
    
  }
  {
    result = (svn_wc_conflict_version_t *)svn_wc_conflict_version_create((char const *)arg1,(char const *)arg2,arg3,arg4,arg5);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_wc_conflict_version_t, 0 |  0 );
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_conflict_version_dup(int argc, VALUE *argv, VALUE self) {
  svn_wc_conflict_version_t *arg1 = 0 ;
  apr_pool_t *arg2 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_wc_conflict_version_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_conflict_version_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_conflict_version_t const *","svn_wc_conflict_version_dup", 1, argv[0] )); 
  }
  arg1 = (svn_wc_conflict_version_t *)(argp1);
  if (argc > 1) {
    
  }
  {
    result = (svn_wc_conflict_version_t *)svn_wc_conflict_version_dup((struct svn_wc_conflict_version_t const *)arg1,arg2);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_wc_conflict_version_t, 0 |  0 );
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


static swig_class SwigClassSvn_wc_conflict_description_t;

SWIGINTERN VALUE
_wrap_svn_wc_conflict_description_t_path_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_conflict_description_t *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_conflict_description_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_conflict_description_t *","path", 1, self )); 
  }
  arg1 = (struct svn_wc_conflict_description_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","path", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->path) free((char *)arg1->path);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->path = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_conflict_description_t_path_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_conflict_description_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_conflict_description_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_conflict_description_t *","path", 1, self )); 
  }
  arg1 = (struct svn_wc_conflict_description_t *)(argp1);
  result = (char *) ((arg1)->path);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_conflict_description_t_node_kind_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_conflict_description_t *arg1 = 0 ;
  svn_node_kind_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_conflict_description_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_conflict_description_t *","node_kind", 1, self )); 
  }
  arg1 = (struct svn_wc_conflict_description_t *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_node_kind_t","node_kind", 2, argv[0] ));
  } 
  arg2 = (svn_node_kind_t)(val2);
  if (arg1) (arg1)->node_kind = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_conflict_description_t_node_kind_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_conflict_description_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_node_kind_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_conflict_description_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_conflict_description_t *","node_kind", 1, self )); 
  }
  arg1 = (struct svn_wc_conflict_description_t *)(argp1);
  result = (svn_node_kind_t) ((arg1)->node_kind);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_conflict_description_t_kind_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_conflict_description_t *arg1 = 0 ;
  svn_wc_conflict_kind_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_conflict_description_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_conflict_description_t *","kind", 1, self )); 
  }
  arg1 = (struct svn_wc_conflict_description_t *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_wc_conflict_kind_t","kind", 2, argv[0] ));
  } 
  arg2 = (svn_wc_conflict_kind_t)(val2);
  if (arg1) (arg1)->kind = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_conflict_description_t_kind_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_conflict_description_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_wc_conflict_kind_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_conflict_description_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_conflict_description_t *","kind", 1, self )); 
  }
  arg1 = (struct svn_wc_conflict_description_t *)(argp1);
  result = (svn_wc_conflict_kind_t) ((arg1)->kind);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_conflict_description_t_property_name_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_conflict_description_t *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_conflict_description_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_conflict_description_t *","property_name", 1, self )); 
  }
  arg1 = (struct svn_wc_conflict_description_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","property_name", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->property_name) free((char *)arg1->property_name);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->property_name = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_conflict_description_t_property_name_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_conflict_description_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_conflict_description_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_conflict_description_t *","property_name", 1, self )); 
  }
  arg1 = (struct svn_wc_conflict_description_t *)(argp1);
  result = (char *) ((arg1)->property_name);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_conflict_description_t_is_binary_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_conflict_description_t *arg1 = 0 ;
  svn_boolean_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_conflict_description_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_conflict_description_t *","is_binary", 1, self )); 
  }
  arg1 = (struct svn_wc_conflict_description_t *)(argp1);
  arg2 = RTEST(argv[0]);
  if (arg1) (arg1)->is_binary = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_conflict_description_t_is_binary_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_conflict_description_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_conflict_description_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_conflict_description_t *","is_binary", 1, self )); 
  }
  arg1 = (struct svn_wc_conflict_description_t *)(argp1);
  result = (svn_boolean_t) ((arg1)->is_binary);
  vresult = result ? Qtrue : Qfalse;
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_conflict_description_t_mime_type_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_conflict_description_t *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_conflict_description_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_conflict_description_t *","mime_type", 1, self )); 
  }
  arg1 = (struct svn_wc_conflict_description_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","mime_type", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->mime_type) free((char *)arg1->mime_type);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->mime_type = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_conflict_description_t_mime_type_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_conflict_description_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_conflict_description_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_conflict_description_t *","mime_type", 1, self )); 
  }
  arg1 = (struct svn_wc_conflict_description_t *)(argp1);
  result = (char *) ((arg1)->mime_type);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_conflict_description_t_access_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_conflict_description_t *arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_conflict_description_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_conflict_description_t *","access", 1, self )); 
  }
  arg1 = (struct svn_wc_conflict_description_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","access", 2, argv[0] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  if (arg1) (arg1)->access = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_conflict_description_t_access_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_conflict_description_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_wc_adm_access_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_conflict_description_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_conflict_description_t *","access", 1, self )); 
  }
  arg1 = (struct svn_wc_conflict_description_t *)(argp1);
  result = (svn_wc_adm_access_t *) ((arg1)->access);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_conflict_description_t_action_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_conflict_description_t *arg1 = 0 ;
  svn_wc_conflict_action_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_conflict_description_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_conflict_description_t *","action", 1, self )); 
  }
  arg1 = (struct svn_wc_conflict_description_t *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_wc_conflict_action_t","action", 2, argv[0] ));
  } 
  arg2 = (svn_wc_conflict_action_t)(val2);
  if (arg1) (arg1)->action = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_conflict_description_t_action_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_conflict_description_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_wc_conflict_action_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_conflict_description_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_conflict_description_t *","action", 1, self )); 
  }
  arg1 = (struct svn_wc_conflict_description_t *)(argp1);
  result = (svn_wc_conflict_action_t) ((arg1)->action);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_conflict_description_t_reason_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_conflict_description_t *arg1 = 0 ;
  svn_wc_conflict_reason_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_conflict_description_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_conflict_description_t *","reason", 1, self )); 
  }
  arg1 = (struct svn_wc_conflict_description_t *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_wc_conflict_reason_t","reason", 2, argv[0] ));
  } 
  arg2 = (svn_wc_conflict_reason_t)(val2);
  if (arg1) (arg1)->reason = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_conflict_description_t_reason_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_conflict_description_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_wc_conflict_reason_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_conflict_description_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_conflict_description_t *","reason", 1, self )); 
  }
  arg1 = (struct svn_wc_conflict_description_t *)(argp1);
  result = (svn_wc_conflict_reason_t) ((arg1)->reason);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_conflict_description_t_base_file_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_conflict_description_t *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_conflict_description_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_conflict_description_t *","base_file", 1, self )); 
  }
  arg1 = (struct svn_wc_conflict_description_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","base_file", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->base_file) free((char *)arg1->base_file);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->base_file = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_conflict_description_t_base_file_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_conflict_description_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_conflict_description_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_conflict_description_t *","base_file", 1, self )); 
  }
  arg1 = (struct svn_wc_conflict_description_t *)(argp1);
  result = (char *) ((arg1)->base_file);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_conflict_description_t_their_file_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_conflict_description_t *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_conflict_description_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_conflict_description_t *","their_file", 1, self )); 
  }
  arg1 = (struct svn_wc_conflict_description_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","their_file", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->their_file) free((char *)arg1->their_file);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->their_file = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_conflict_description_t_their_file_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_conflict_description_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_conflict_description_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_conflict_description_t *","their_file", 1, self )); 
  }
  arg1 = (struct svn_wc_conflict_description_t *)(argp1);
  result = (char *) ((arg1)->their_file);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_conflict_description_t_my_file_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_conflict_description_t *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_conflict_description_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_conflict_description_t *","my_file", 1, self )); 
  }
  arg1 = (struct svn_wc_conflict_description_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","my_file", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->my_file) free((char *)arg1->my_file);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->my_file = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_conflict_description_t_my_file_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_conflict_description_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_conflict_description_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_conflict_description_t *","my_file", 1, self )); 
  }
  arg1 = (struct svn_wc_conflict_description_t *)(argp1);
  result = (char *) ((arg1)->my_file);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_conflict_description_t_merged_file_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_conflict_description_t *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_conflict_description_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_conflict_description_t *","merged_file", 1, self )); 
  }
  arg1 = (struct svn_wc_conflict_description_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","merged_file", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->merged_file) free((char *)arg1->merged_file);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->merged_file = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_conflict_description_t_merged_file_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_conflict_description_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_conflict_description_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_conflict_description_t *","merged_file", 1, self )); 
  }
  arg1 = (struct svn_wc_conflict_description_t *)(argp1);
  result = (char *) ((arg1)->merged_file);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_conflict_description_t_operation_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_conflict_description_t *arg1 = 0 ;
  svn_wc_operation_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_conflict_description_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_conflict_description_t *","operation", 1, self )); 
  }
  arg1 = (struct svn_wc_conflict_description_t *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_wc_operation_t","operation", 2, argv[0] ));
  } 
  arg2 = (svn_wc_operation_t)(val2);
  if (arg1) (arg1)->operation = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_conflict_description_t_operation_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_conflict_description_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_wc_operation_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_conflict_description_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_conflict_description_t *","operation", 1, self )); 
  }
  arg1 = (struct svn_wc_conflict_description_t *)(argp1);
  result = (svn_wc_operation_t) ((arg1)->operation);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_conflict_description_t_src_left_version_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_conflict_description_t *arg1 = 0 ;
  svn_wc_conflict_version_t *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_conflict_description_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_conflict_description_t *","src_left_version", 1, self )); 
  }
  arg1 = (struct svn_wc_conflict_description_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_svn_wc_conflict_version_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_conflict_version_t *","src_left_version", 2, argv[0] )); 
  }
  arg2 = (svn_wc_conflict_version_t *)(argp2);
  if (arg1) (arg1)->src_left_version = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_conflict_description_t_src_left_version_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_conflict_description_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_wc_conflict_version_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_conflict_description_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_conflict_description_t *","src_left_version", 1, self )); 
  }
  arg1 = (struct svn_wc_conflict_description_t *)(argp1);
  result = (svn_wc_conflict_version_t *) ((arg1)->src_left_version);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_wc_conflict_version_t, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_conflict_description_t_src_right_version_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_conflict_description_t *arg1 = 0 ;
  svn_wc_conflict_version_t *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_conflict_description_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_conflict_description_t *","src_right_version", 1, self )); 
  }
  arg1 = (struct svn_wc_conflict_description_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_svn_wc_conflict_version_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_conflict_version_t *","src_right_version", 2, argv[0] )); 
  }
  arg2 = (svn_wc_conflict_version_t *)(argp2);
  if (arg1) (arg1)->src_right_version = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_conflict_description_t_src_right_version_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_conflict_description_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_wc_conflict_version_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_conflict_description_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_conflict_description_t *","src_right_version", 1, self )); 
  }
  arg1 = (struct svn_wc_conflict_description_t *)(argp1);
  result = (svn_wc_conflict_version_t *) ((arg1)->src_right_version);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_wc_conflict_version_t, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
_wrap_svn_wc_conflict_description_t_allocate(VALUE self)
#else
_wrap_svn_wc_conflict_description_t_allocate(int argc, VALUE *argv, VALUE self)
#endif
{
  VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_svn_wc_conflict_description_t);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
  rb_obj_call_init(vresult, argc, argv);
#endif
  return vresult;
}


SWIGINTERN VALUE
_wrap_new_svn_wc_conflict_description_t(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_conflict_description_t *result = 0 ;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  {
    result = (struct svn_wc_conflict_description_t *)calloc(1, sizeof(struct svn_wc_conflict_description_t));
    DATA_PTR(self) = result;
    
    
    
  }
  return self;
fail:
  return Qnil;
}


SWIGINTERN void
free_svn_wc_conflict_description_t(void *self) {
    struct svn_wc_conflict_description_t *arg1 = (struct svn_wc_conflict_description_t *)self;
    free((char *) arg1);
}

SWIGINTERN VALUE
_wrap_svn_wc_conflict_description_create_text2(int argc, VALUE *argv, VALUE self) {
  char *arg1 = 0 ;
  apr_pool_t *arg2 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  svn_wc_conflict_description2_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_wc_conflict_description_create_text2", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  if (argc > 1) {
    
  }
  {
    result = (svn_wc_conflict_description2_t *)svn_wc_conflict_description_create_text2((char const *)arg1,arg2);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_wc_conflict_description2_t, 0 |  0 );
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_conflict_description_create_text(int argc, VALUE *argv, VALUE self) {
  char *arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  apr_pool_t *arg3 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  svn_wc_conflict_description_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_wc_conflict_description_create_text", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_conflict_description_create_text", 2, argv[1] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  if (argc > 2) {
    
  }
  {
    result = (svn_wc_conflict_description_t *)svn_wc_conflict_description_create_text((char const *)arg1,arg2,arg3);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_wc_conflict_description_t, 0 |  0 );
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_conflict_description_create_prop2(int argc, VALUE *argv, VALUE self) {
  char *arg1 = 0 ;
  svn_node_kind_t arg2 ;
  char *arg3 = 0 ;
  apr_pool_t *arg4 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  svn_wc_conflict_description2_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 3) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_wc_conflict_description_create_prop2", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  ecode2 = SWIG_AsVal_int(argv[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_node_kind_t","svn_wc_conflict_description_create_prop2", 2, argv[1] ));
  } 
  arg2 = (svn_node_kind_t)(val2);
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_wc_conflict_description_create_prop2", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  if (argc > 3) {
    
  }
  {
    result = (svn_wc_conflict_description2_t *)svn_wc_conflict_description_create_prop2((char const *)arg1,arg2,(char const *)arg3,arg4);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_wc_conflict_description2_t, 0 |  0 );
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_conflict_description_create_prop(int argc, VALUE *argv, VALUE self) {
  char *arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  svn_node_kind_t arg3 ;
  char *arg4 = 0 ;
  apr_pool_t *arg5 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  svn_wc_conflict_description_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 4) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_wc_conflict_description_create_prop", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_conflict_description_create_prop", 2, argv[1] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  ecode3 = SWIG_AsVal_int(argv[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), Ruby_Format_TypeError( "", "svn_node_kind_t","svn_wc_conflict_description_create_prop", 3, argv[2] ));
  } 
  arg3 = (svn_node_kind_t)(val3);
  res4 = SWIG_AsCharPtrAndSize(argv[3], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_wc_conflict_description_create_prop", 4, argv[3] ));
  }
  arg4 = (char *)(buf4);
  if (argc > 4) {
    
  }
  {
    result = (svn_wc_conflict_description_t *)svn_wc_conflict_description_create_prop((char const *)arg1,arg2,arg3,(char const *)arg4,arg5);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_wc_conflict_description_t, 0 |  0 );
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_conflict_description_create_tree2(int argc, VALUE *argv, VALUE self) {
  char *arg1 = 0 ;
  svn_node_kind_t arg2 ;
  svn_wc_operation_t arg3 ;
  svn_wc_conflict_version_t *arg4 = 0 ;
  svn_wc_conflict_version_t *arg5 = 0 ;
  apr_pool_t *arg6 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  svn_wc_conflict_description2_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 5) || (argc > 6)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 5)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_wc_conflict_description_create_tree2", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  ecode2 = SWIG_AsVal_int(argv[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_node_kind_t","svn_wc_conflict_description_create_tree2", 2, argv[1] ));
  } 
  arg2 = (svn_node_kind_t)(val2);
  ecode3 = SWIG_AsVal_int(argv[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), Ruby_Format_TypeError( "", "svn_wc_operation_t","svn_wc_conflict_description_create_tree2", 3, argv[2] ));
  } 
  arg3 = (svn_wc_operation_t)(val3);
  res4 = SWIG_ConvertPtr(argv[3], &argp4,SWIGTYPE_p_svn_wc_conflict_version_t, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "svn_wc_conflict_version_t const *","svn_wc_conflict_description_create_tree2", 4, argv[3] )); 
  }
  arg4 = (svn_wc_conflict_version_t *)(argp4);
  res5 = SWIG_ConvertPtr(argv[4], &argp5,SWIGTYPE_p_svn_wc_conflict_version_t, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "svn_wc_conflict_version_t const *","svn_wc_conflict_description_create_tree2", 5, argv[4] )); 
  }
  arg5 = (svn_wc_conflict_version_t *)(argp5);
  if (argc > 5) {
    
  }
  {
    result = (svn_wc_conflict_description2_t *)svn_wc_conflict_description_create_tree2((char const *)arg1,arg2,arg3,(struct svn_wc_conflict_version_t const *)arg4,(struct svn_wc_conflict_version_t const *)arg5,arg6);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_wc_conflict_description2_t, 0 |  0 );
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_conflict_description_create_tree(int argc, VALUE *argv, VALUE self) {
  char *arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  svn_node_kind_t arg3 ;
  svn_wc_operation_t arg4 ;
  svn_wc_conflict_version_t *arg5 = 0 ;
  svn_wc_conflict_version_t *arg6 = 0 ;
  apr_pool_t *arg7 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  int val4 ;
  int ecode4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  svn_wc_conflict_description_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg7);
    _global_pool = arg7;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 6) || (argc > 7)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 6)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_wc_conflict_description_create_tree", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_conflict_description_create_tree", 2, argv[1] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  ecode3 = SWIG_AsVal_int(argv[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), Ruby_Format_TypeError( "", "svn_node_kind_t","svn_wc_conflict_description_create_tree", 3, argv[2] ));
  } 
  arg3 = (svn_node_kind_t)(val3);
  ecode4 = SWIG_AsVal_int(argv[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), Ruby_Format_TypeError( "", "svn_wc_operation_t","svn_wc_conflict_description_create_tree", 4, argv[3] ));
  } 
  arg4 = (svn_wc_operation_t)(val4);
  res5 = SWIG_ConvertPtr(argv[4], &argp5,SWIGTYPE_p_svn_wc_conflict_version_t, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "svn_wc_conflict_version_t *","svn_wc_conflict_description_create_tree", 5, argv[4] )); 
  }
  arg5 = (svn_wc_conflict_version_t *)(argp5);
  res6 = SWIG_ConvertPtr(argv[5], &argp6,SWIGTYPE_p_svn_wc_conflict_version_t, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "svn_wc_conflict_version_t *","svn_wc_conflict_description_create_tree", 6, argv[5] )); 
  }
  arg6 = (svn_wc_conflict_version_t *)(argp6);
  if (argc > 6) {
    
  }
  {
    result = (svn_wc_conflict_description_t *)svn_wc_conflict_description_create_tree((char const *)arg1,arg2,arg3,arg4,arg5,arg6,arg7);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_wc_conflict_description_t, 0 |  0 );
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_conflict_description2_dup(int argc, VALUE *argv, VALUE self) {
  svn_wc_conflict_description2_t *arg1 = 0 ;
  apr_pool_t *arg2 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_wc_conflict_description2_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_conflict_description2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_conflict_description2_t const *","svn_wc_conflict_description2_dup", 1, argv[0] )); 
  }
  arg1 = (svn_wc_conflict_description2_t *)(argp1);
  if (argc > 1) {
    
  }
  {
    result = (svn_wc_conflict_description2_t *)svn_wc_conflict_description2_dup((struct svn_wc_conflict_description2_t const *)arg1,arg2);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_wc_conflict_description2_t, 0 |  0 );
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc__conflict_description2_dup(int argc, VALUE *argv, VALUE self) {
  svn_wc_conflict_description2_t *arg1 = 0 ;
  apr_pool_t *arg2 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_wc_conflict_description2_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_conflict_description2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_conflict_description2_t const *","svn_wc__conflict_description2_dup", 1, argv[0] )); 
  }
  arg1 = (svn_wc_conflict_description2_t *)(argp1);
  if (argc > 1) {
    
  }
  {
    result = (svn_wc_conflict_description2_t *)svn_wc__conflict_description2_dup((struct svn_wc_conflict_description2_t const *)arg1,arg2);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_wc_conflict_description2_t, 0 |  0 );
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_create_conflict_result(int argc, VALUE *argv, VALUE self) {
  svn_wc_conflict_choice_t arg1 ;
  char *arg2 = 0 ;
  apr_pool_t *arg3 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int val1 ;
  int ecode1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  svn_wc_conflict_result_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  ecode1 = SWIG_AsVal_int(argv[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), Ruby_Format_TypeError( "", "svn_wc_conflict_choice_t","svn_wc_create_conflict_result", 1, argv[0] ));
  } 
  arg1 = (svn_wc_conflict_choice_t)(val1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_wc_create_conflict_result", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  if (argc > 2) {
    
  }
  {
    result = (svn_wc_conflict_result_t *)svn_wc_create_conflict_result(arg1,(char const *)arg2,arg3);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_wc_conflict_result_t, 0 |  0 );
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


static swig_class SwigClassSvn_wc_diff_callbacks4_t;

SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks4_t_file_opened_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_diff_callbacks4_t *arg1 = 0 ;
  svn_error_t *(*arg2)(svn_boolean_t *,svn_boolean_t *,char const *,svn_revnum_t,void *,apr_pool_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_diff_callbacks4_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_diff_callbacks4_t *","file_opened", 1, self )); 
  }
  arg1 = (struct svn_wc_diff_callbacks4_t *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg2), SWIGTYPE_p_f_p_svn_boolean_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_void_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_error_t *(*)(svn_boolean_t *,svn_boolean_t *,char const *,svn_revnum_t,void *,apr_pool_t *)","file_opened", 2, argv[0] )); 
    }
  }
  if (arg1) (arg1)->file_opened = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks4_t_file_opened_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_diff_callbacks4_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_error_t *(*result)(svn_boolean_t *,svn_boolean_t *,char const *,svn_revnum_t,void *,apr_pool_t *) = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_diff_callbacks4_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_diff_callbacks4_t *","file_opened", 1, self )); 
  }
  arg1 = (struct svn_wc_diff_callbacks4_t *)(argp1);
  result = (svn_error_t *(*)(svn_boolean_t *,svn_boolean_t *,char const *,svn_revnum_t,void *,apr_pool_t *)) ((arg1)->file_opened);
  vresult = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_svn_boolean_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_void_p_apr_pool_t__p_svn_error_t);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks4_t_file_changed_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_diff_callbacks4_t *arg1 = 0 ;
  svn_error_t *(*arg2)(svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,apr_array_header_t const *,apr_hash_t *,void *,apr_pool_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_diff_callbacks4_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_diff_callbacks4_t *","file_changed", 1, self )); 
  }
  arg1 = (struct svn_wc_diff_callbacks4_t *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg2), SWIGTYPE_p_f_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_error_t *(*)(svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,apr_array_header_t const *,apr_hash_t *,void *,apr_pool_t *)","file_changed", 2, argv[0] )); 
    }
  }
  if (arg1) (arg1)->file_changed = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks4_t_file_changed_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_diff_callbacks4_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_error_t *(*result)(svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,apr_array_header_t const *,apr_hash_t *,void *,apr_pool_t *) = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_diff_callbacks4_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_diff_callbacks4_t *","file_changed", 1, self )); 
  }
  arg1 = (struct svn_wc_diff_callbacks4_t *)(argp1);
  result = (svn_error_t *(*)(svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,apr_array_header_t const *,apr_hash_t *,void *,apr_pool_t *)) ((arg1)->file_changed);
  vresult = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void_p_apr_pool_t__p_svn_error_t);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks4_t_file_added_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_diff_callbacks4_t *arg1 = 0 ;
  svn_error_t *(*arg2)(svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,char const *,svn_revnum_t,apr_array_header_t const *,apr_hash_t *,void *,apr_pool_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_diff_callbacks4_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_diff_callbacks4_t *","file_added", 1, self )); 
  }
  arg1 = (struct svn_wc_diff_callbacks4_t *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg2), SWIGTYPE_p_f_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_p_q_const__apr_array_header_t_p_apr_hash_t_p_void_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_error_t *(*)(svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,char const *,svn_revnum_t,apr_array_header_t const *,apr_hash_t *,void *,apr_pool_t *)","file_added", 2, argv[0] )); 
    }
  }
  if (arg1) (arg1)->file_added = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks4_t_file_added_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_diff_callbacks4_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_error_t *(*result)(svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,char const *,svn_revnum_t,apr_array_header_t const *,apr_hash_t *,void *,apr_pool_t *) = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_diff_callbacks4_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_diff_callbacks4_t *","file_added", 1, self )); 
  }
  arg1 = (struct svn_wc_diff_callbacks4_t *)(argp1);
  result = (svn_error_t *(*)(svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,char const *,svn_revnum_t,apr_array_header_t const *,apr_hash_t *,void *,apr_pool_t *)) ((arg1)->file_added);
  vresult = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_p_q_const__apr_array_header_t_p_apr_hash_t_p_void_p_apr_pool_t__p_svn_error_t);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks4_t_file_deleted_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_diff_callbacks4_t *arg1 = 0 ;
  svn_error_t *(*arg2)(svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,char const *,char const *,apr_hash_t *,void *,apr_pool_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_diff_callbacks4_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_diff_callbacks4_t *","file_deleted", 1, self )); 
  }
  arg1 = (struct svn_wc_diff_callbacks4_t *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg2), SWIGTYPE_p_f_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_hash_t_p_void_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_error_t *(*)(svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,char const *,char const *,apr_hash_t *,void *,apr_pool_t *)","file_deleted", 2, argv[0] )); 
    }
  }
  if (arg1) (arg1)->file_deleted = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks4_t_file_deleted_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_diff_callbacks4_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_error_t *(*result)(svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,char const *,char const *,apr_hash_t *,void *,apr_pool_t *) = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_diff_callbacks4_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_diff_callbacks4_t *","file_deleted", 1, self )); 
  }
  arg1 = (struct svn_wc_diff_callbacks4_t *)(argp1);
  result = (svn_error_t *(*)(svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,char const *,char const *,apr_hash_t *,void *,apr_pool_t *)) ((arg1)->file_deleted);
  vresult = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_hash_t_p_void_p_apr_pool_t__p_svn_error_t);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks4_t_dir_deleted_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_diff_callbacks4_t *arg1 = 0 ;
  svn_error_t *(*arg2)(svn_wc_notify_state_t *,svn_boolean_t *,char const *,void *,apr_pool_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_diff_callbacks4_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_diff_callbacks4_t *","dir_deleted", 1, self )); 
  }
  arg1 = (struct svn_wc_diff_callbacks4_t *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg2), SWIGTYPE_p_f_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_void_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_error_t *(*)(svn_wc_notify_state_t *,svn_boolean_t *,char const *,void *,apr_pool_t *)","dir_deleted", 2, argv[0] )); 
    }
  }
  if (arg1) (arg1)->dir_deleted = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks4_t_dir_deleted_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_diff_callbacks4_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_error_t *(*result)(svn_wc_notify_state_t *,svn_boolean_t *,char const *,void *,apr_pool_t *) = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_diff_callbacks4_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_diff_callbacks4_t *","dir_deleted", 1, self )); 
  }
  arg1 = (struct svn_wc_diff_callbacks4_t *)(argp1);
  result = (svn_error_t *(*)(svn_wc_notify_state_t *,svn_boolean_t *,char const *,void *,apr_pool_t *)) ((arg1)->dir_deleted);
  vresult = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_void_p_apr_pool_t__p_svn_error_t);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks4_t_dir_opened_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_diff_callbacks4_t *arg1 = 0 ;
  svn_error_t *(*arg2)(svn_boolean_t *,svn_boolean_t *,svn_boolean_t *,char const *,svn_revnum_t,void *,apr_pool_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_diff_callbacks4_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_diff_callbacks4_t *","dir_opened", 1, self )); 
  }
  arg1 = (struct svn_wc_diff_callbacks4_t *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg2), SWIGTYPE_p_f_p_svn_boolean_t_p_svn_boolean_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_void_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_error_t *(*)(svn_boolean_t *,svn_boolean_t *,svn_boolean_t *,char const *,svn_revnum_t,void *,apr_pool_t *)","dir_opened", 2, argv[0] )); 
    }
  }
  if (arg1) (arg1)->dir_opened = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks4_t_dir_opened_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_diff_callbacks4_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_error_t *(*result)(svn_boolean_t *,svn_boolean_t *,svn_boolean_t *,char const *,svn_revnum_t,void *,apr_pool_t *) = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_diff_callbacks4_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_diff_callbacks4_t *","dir_opened", 1, self )); 
  }
  arg1 = (struct svn_wc_diff_callbacks4_t *)(argp1);
  result = (svn_error_t *(*)(svn_boolean_t *,svn_boolean_t *,svn_boolean_t *,char const *,svn_revnum_t,void *,apr_pool_t *)) ((arg1)->dir_opened);
  vresult = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_svn_boolean_t_p_svn_boolean_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_void_p_apr_pool_t__p_svn_error_t);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks4_t_dir_added_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_diff_callbacks4_t *arg1 = 0 ;
  svn_error_t *(*arg2)(svn_wc_notify_state_t *,svn_boolean_t *,svn_boolean_t *,svn_boolean_t *,char const *,svn_revnum_t,char const *,svn_revnum_t,void *,apr_pool_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_diff_callbacks4_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_diff_callbacks4_t *","dir_added", 1, self )); 
  }
  arg1 = (struct svn_wc_diff_callbacks4_t *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg2), SWIGTYPE_p_f_p_svn_wc_notify_state_t_p_svn_boolean_t_p_svn_boolean_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_q_const__char_svn_revnum_t_p_void_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_error_t *(*)(svn_wc_notify_state_t *,svn_boolean_t *,svn_boolean_t *,svn_boolean_t *,char const *,svn_revnum_t,char const *,svn_revnum_t,void *,apr_pool_t *)","dir_added", 2, argv[0] )); 
    }
  }
  if (arg1) (arg1)->dir_added = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks4_t_dir_added_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_diff_callbacks4_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_error_t *(*result)(svn_wc_notify_state_t *,svn_boolean_t *,svn_boolean_t *,svn_boolean_t *,char const *,svn_revnum_t,char const *,svn_revnum_t,void *,apr_pool_t *) = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_diff_callbacks4_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_diff_callbacks4_t *","dir_added", 1, self )); 
  }
  arg1 = (struct svn_wc_diff_callbacks4_t *)(argp1);
  result = (svn_error_t *(*)(svn_wc_notify_state_t *,svn_boolean_t *,svn_boolean_t *,svn_boolean_t *,char const *,svn_revnum_t,char const *,svn_revnum_t,void *,apr_pool_t *)) ((arg1)->dir_added);
  vresult = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_svn_wc_notify_state_t_p_svn_boolean_t_p_svn_boolean_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_q_const__char_svn_revnum_t_p_void_p_apr_pool_t__p_svn_error_t);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks4_t_dir_props_changed_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_diff_callbacks4_t *arg1 = 0 ;
  svn_error_t *(*arg2)(svn_wc_notify_state_t *,svn_boolean_t *,char const *,svn_boolean_t,apr_array_header_t const *,apr_hash_t *,void *,apr_pool_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_diff_callbacks4_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_diff_callbacks4_t *","dir_props_changed", 1, self )); 
  }
  arg1 = (struct svn_wc_diff_callbacks4_t *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg2), SWIGTYPE_p_f_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_svn_boolean_t_p_q_const__apr_array_header_t_p_apr_hash_t_p_void_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_error_t *(*)(svn_wc_notify_state_t *,svn_boolean_t *,char const *,svn_boolean_t,apr_array_header_t const *,apr_hash_t *,void *,apr_pool_t *)","dir_props_changed", 2, argv[0] )); 
    }
  }
  if (arg1) (arg1)->dir_props_changed = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks4_t_dir_props_changed_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_diff_callbacks4_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_error_t *(*result)(svn_wc_notify_state_t *,svn_boolean_t *,char const *,svn_boolean_t,apr_array_header_t const *,apr_hash_t *,void *,apr_pool_t *) = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_diff_callbacks4_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_diff_callbacks4_t *","dir_props_changed", 1, self )); 
  }
  arg1 = (struct svn_wc_diff_callbacks4_t *)(argp1);
  result = (svn_error_t *(*)(svn_wc_notify_state_t *,svn_boolean_t *,char const *,svn_boolean_t,apr_array_header_t const *,apr_hash_t *,void *,apr_pool_t *)) ((arg1)->dir_props_changed);
  vresult = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_svn_boolean_t_p_q_const__apr_array_header_t_p_apr_hash_t_p_void_p_apr_pool_t__p_svn_error_t);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks4_t_dir_closed_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_diff_callbacks4_t *arg1 = 0 ;
  svn_error_t *(*arg2)(svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,svn_boolean_t,void *,apr_pool_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_diff_callbacks4_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_diff_callbacks4_t *","dir_closed", 1, self )); 
  }
  arg1 = (struct svn_wc_diff_callbacks4_t *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg2), SWIGTYPE_p_f_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_svn_boolean_t_p_void_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_error_t *(*)(svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,svn_boolean_t,void *,apr_pool_t *)","dir_closed", 2, argv[0] )); 
    }
  }
  if (arg1) (arg1)->dir_closed = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks4_t_dir_closed_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_diff_callbacks4_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_error_t *(*result)(svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,svn_boolean_t,void *,apr_pool_t *) = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_diff_callbacks4_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_diff_callbacks4_t *","dir_closed", 1, self )); 
  }
  arg1 = (struct svn_wc_diff_callbacks4_t *)(argp1);
  result = (svn_error_t *(*)(svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,svn_boolean_t,void *,apr_pool_t *)) ((arg1)->dir_closed);
  vresult = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_svn_boolean_t_p_void_p_apr_pool_t__p_svn_error_t);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
_wrap_svn_wc_diff_callbacks4_t_allocate(VALUE self)
#else
_wrap_svn_wc_diff_callbacks4_t_allocate(int argc, VALUE *argv, VALUE self)
#endif
{
  VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_svn_wc_diff_callbacks4_t);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
  rb_obj_call_init(vresult, argc, argv);
#endif
  return vresult;
}


SWIGINTERN VALUE
_wrap_new_svn_wc_diff_callbacks4_t(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_diff_callbacks4_t *result = 0 ;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  {
    result = (struct svn_wc_diff_callbacks4_t *)calloc(1, sizeof(struct svn_wc_diff_callbacks4_t));
    DATA_PTR(self) = result;
    
    
    
  }
  return self;
fail:
  return Qnil;
}


SWIGINTERN void
free_svn_wc_diff_callbacks4_t(void *self) {
    struct svn_wc_diff_callbacks4_t *arg1 = (struct svn_wc_diff_callbacks4_t *)self;
    free((char *) arg1);
}

static swig_class SwigClassSvn_wc_diff_callbacks3_t;

SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks3_t_file_changed_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_diff_callbacks3_t *arg1 = 0 ;
  svn_error_t *(*arg2)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,apr_array_header_t const *,apr_hash_t *,void *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_diff_callbacks3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_diff_callbacks3_t *","file_changed", 1, self )); 
  }
  arg1 = (struct svn_wc_diff_callbacks3_t *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg2), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,apr_array_header_t const *,apr_hash_t *,void *)","file_changed", 2, argv[0] )); 
    }
  }
  if (arg1) (arg1)->file_changed = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks3_t_file_changed_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_diff_callbacks3_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_error_t *(*result)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,apr_array_header_t const *,apr_hash_t *,void *) = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_diff_callbacks3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_diff_callbacks3_t *","file_changed", 1, self )); 
  }
  arg1 = (struct svn_wc_diff_callbacks3_t *)(argp1);
  result = (svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,apr_array_header_t const *,apr_hash_t *,void *)) ((arg1)->file_changed);
  vresult = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks3_t_file_added_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_diff_callbacks3_t *arg1 = 0 ;
  svn_error_t *(*arg2)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,apr_array_header_t const *,apr_hash_t *,void *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_diff_callbacks3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_diff_callbacks3_t *","file_added", 1, self )); 
  }
  arg1 = (struct svn_wc_diff_callbacks3_t *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg2), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,apr_array_header_t const *,apr_hash_t *,void *)","file_added", 2, argv[0] )); 
    }
  }
  if (arg1) (arg1)->file_added = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks3_t_file_added_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_diff_callbacks3_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_error_t *(*result)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,apr_array_header_t const *,apr_hash_t *,void *) = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_diff_callbacks3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_diff_callbacks3_t *","file_added", 1, self )); 
  }
  arg1 = (struct svn_wc_diff_callbacks3_t *)(argp1);
  result = (svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,apr_array_header_t const *,apr_hash_t *,void *)) ((arg1)->file_added);
  vresult = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks3_t_file_deleted_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_diff_callbacks3_t *arg1 = 0 ;
  svn_error_t *(*arg2)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,char const *,char const *,apr_hash_t *,void *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_diff_callbacks3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_diff_callbacks3_t *","file_deleted", 1, self )); 
  }
  arg1 = (struct svn_wc_diff_callbacks3_t *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg2), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_hash_t_p_void__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,char const *,char const *,apr_hash_t *,void *)","file_deleted", 2, argv[0] )); 
    }
  }
  if (arg1) (arg1)->file_deleted = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks3_t_file_deleted_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_diff_callbacks3_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_error_t *(*result)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,char const *,char const *,apr_hash_t *,void *) = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_diff_callbacks3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_diff_callbacks3_t *","file_deleted", 1, self )); 
  }
  arg1 = (struct svn_wc_diff_callbacks3_t *)(argp1);
  result = (svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,char const *,char const *,apr_hash_t *,void *)) ((arg1)->file_deleted);
  vresult = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_hash_t_p_void__p_svn_error_t);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks3_t_dir_added_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_diff_callbacks3_t *arg1 = 0 ;
  svn_error_t *(*arg2)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,svn_revnum_t,void *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_diff_callbacks3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_diff_callbacks3_t *","dir_added", 1, self )); 
  }
  arg1 = (struct svn_wc_diff_callbacks3_t *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg2), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_void__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,svn_revnum_t,void *)","dir_added", 2, argv[0] )); 
    }
  }
  if (arg1) (arg1)->dir_added = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks3_t_dir_added_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_diff_callbacks3_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_error_t *(*result)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,svn_revnum_t,void *) = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_diff_callbacks3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_diff_callbacks3_t *","dir_added", 1, self )); 
  }
  arg1 = (struct svn_wc_diff_callbacks3_t *)(argp1);
  result = (svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,svn_revnum_t,void *)) ((arg1)->dir_added);
  vresult = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_void__p_svn_error_t);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks3_t_dir_deleted_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_diff_callbacks3_t *arg1 = 0 ;
  svn_error_t *(*arg2)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,void *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_diff_callbacks3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_diff_callbacks3_t *","dir_deleted", 1, self )); 
  }
  arg1 = (struct svn_wc_diff_callbacks3_t *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg2), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_void__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,void *)","dir_deleted", 2, argv[0] )); 
    }
  }
  if (arg1) (arg1)->dir_deleted = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks3_t_dir_deleted_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_diff_callbacks3_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_error_t *(*result)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,void *) = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_diff_callbacks3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_diff_callbacks3_t *","dir_deleted", 1, self )); 
  }
  arg1 = (struct svn_wc_diff_callbacks3_t *)(argp1);
  result = (svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,void *)) ((arg1)->dir_deleted);
  vresult = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_void__p_svn_error_t);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks3_t_dir_props_changed_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_diff_callbacks3_t *arg1 = 0 ;
  svn_error_t *(*arg2)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,apr_array_header_t const *,apr_hash_t *,void *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_diff_callbacks3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_diff_callbacks3_t *","dir_props_changed", 1, self )); 
  }
  arg1 = (struct svn_wc_diff_callbacks3_t *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg2), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,apr_array_header_t const *,apr_hash_t *,void *)","dir_props_changed", 2, argv[0] )); 
    }
  }
  if (arg1) (arg1)->dir_props_changed = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks3_t_dir_props_changed_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_diff_callbacks3_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_error_t *(*result)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,apr_array_header_t const *,apr_hash_t *,void *) = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_diff_callbacks3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_diff_callbacks3_t *","dir_props_changed", 1, self )); 
  }
  arg1 = (struct svn_wc_diff_callbacks3_t *)(argp1);
  result = (svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,apr_array_header_t const *,apr_hash_t *,void *)) ((arg1)->dir_props_changed);
  vresult = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks3_t_dir_opened_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_diff_callbacks3_t *arg1 = 0 ;
  svn_error_t *(*arg2)(svn_wc_adm_access_t *,svn_boolean_t *,char const *,svn_revnum_t,void *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_diff_callbacks3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_diff_callbacks3_t *","dir_opened", 1, self )); 
  }
  arg1 = (struct svn_wc_diff_callbacks3_t *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg2), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_void__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_error_t *(*)(svn_wc_adm_access_t *,svn_boolean_t *,char const *,svn_revnum_t,void *)","dir_opened", 2, argv[0] )); 
    }
  }
  if (arg1) (arg1)->dir_opened = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks3_t_dir_opened_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_diff_callbacks3_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_error_t *(*result)(svn_wc_adm_access_t *,svn_boolean_t *,char const *,svn_revnum_t,void *) = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_diff_callbacks3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_diff_callbacks3_t *","dir_opened", 1, self )); 
  }
  arg1 = (struct svn_wc_diff_callbacks3_t *)(argp1);
  result = (svn_error_t *(*)(svn_wc_adm_access_t *,svn_boolean_t *,char const *,svn_revnum_t,void *)) ((arg1)->dir_opened);
  vresult = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_void__p_svn_error_t);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks3_t_dir_closed_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_diff_callbacks3_t *arg1 = 0 ;
  svn_error_t *(*arg2)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,void *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_diff_callbacks3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_diff_callbacks3_t *","dir_closed", 1, self )); 
  }
  arg1 = (struct svn_wc_diff_callbacks3_t *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg2), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_void__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,void *)","dir_closed", 2, argv[0] )); 
    }
  }
  if (arg1) (arg1)->dir_closed = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks3_t_dir_closed_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_diff_callbacks3_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_error_t *(*result)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,void *) = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_diff_callbacks3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_diff_callbacks3_t *","dir_closed", 1, self )); 
  }
  arg1 = (struct svn_wc_diff_callbacks3_t *)(argp1);
  result = (svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,void *)) ((arg1)->dir_closed);
  vresult = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_void__p_svn_error_t);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
_wrap_svn_wc_diff_callbacks3_t_allocate(VALUE self)
#else
_wrap_svn_wc_diff_callbacks3_t_allocate(int argc, VALUE *argv, VALUE self)
#endif
{
  VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_svn_wc_diff_callbacks3_t);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
  rb_obj_call_init(vresult, argc, argv);
#endif
  return vresult;
}


SWIGINTERN VALUE
_wrap_new_svn_wc_diff_callbacks3_t(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_diff_callbacks3_t *result = 0 ;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  {
    result = (struct svn_wc_diff_callbacks3_t *)calloc(1, sizeof(struct svn_wc_diff_callbacks3_t));
    DATA_PTR(self) = result;
    
    
    
  }
  return self;
fail:
  return Qnil;
}


SWIGINTERN void
free_svn_wc_diff_callbacks3_t(void *self) {
    struct svn_wc_diff_callbacks3_t *arg1 = (struct svn_wc_diff_callbacks3_t *)self;
    free((char *) arg1);
}

static swig_class SwigClassSvn_wc_diff_callbacks2_t;

SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks2_t_file_changed_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_diff_callbacks2_t *arg1 = 0 ;
  svn_error_t *(*arg2)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,apr_array_header_t const *,apr_hash_t *,void *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_diff_callbacks2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_diff_callbacks2_t *","file_changed", 1, self )); 
  }
  arg1 = (struct svn_wc_diff_callbacks2_t *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg2), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,apr_array_header_t const *,apr_hash_t *,void *)","file_changed", 2, argv[0] )); 
    }
  }
  if (arg1) (arg1)->file_changed = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks2_t_file_changed_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_diff_callbacks2_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_error_t *(*result)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,apr_array_header_t const *,apr_hash_t *,void *) = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_diff_callbacks2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_diff_callbacks2_t *","file_changed", 1, self )); 
  }
  arg1 = (struct svn_wc_diff_callbacks2_t *)(argp1);
  result = (svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,apr_array_header_t const *,apr_hash_t *,void *)) ((arg1)->file_changed);
  vresult = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks2_t_file_added_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_diff_callbacks2_t *arg1 = 0 ;
  svn_error_t *(*arg2)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,apr_array_header_t const *,apr_hash_t *,void *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_diff_callbacks2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_diff_callbacks2_t *","file_added", 1, self )); 
  }
  arg1 = (struct svn_wc_diff_callbacks2_t *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg2), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,apr_array_header_t const *,apr_hash_t *,void *)","file_added", 2, argv[0] )); 
    }
  }
  if (arg1) (arg1)->file_added = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks2_t_file_added_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_diff_callbacks2_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_error_t *(*result)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,apr_array_header_t const *,apr_hash_t *,void *) = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_diff_callbacks2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_diff_callbacks2_t *","file_added", 1, self )); 
  }
  arg1 = (struct svn_wc_diff_callbacks2_t *)(argp1);
  result = (svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,apr_array_header_t const *,apr_hash_t *,void *)) ((arg1)->file_added);
  vresult = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks2_t_file_deleted_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_diff_callbacks2_t *arg1 = 0 ;
  svn_error_t *(*arg2)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,char const *,char const *,apr_hash_t *,void *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_diff_callbacks2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_diff_callbacks2_t *","file_deleted", 1, self )); 
  }
  arg1 = (struct svn_wc_diff_callbacks2_t *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg2), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_hash_t_p_void__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,char const *,char const *,apr_hash_t *,void *)","file_deleted", 2, argv[0] )); 
    }
  }
  if (arg1) (arg1)->file_deleted = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks2_t_file_deleted_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_diff_callbacks2_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_error_t *(*result)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,char const *,char const *,apr_hash_t *,void *) = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_diff_callbacks2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_diff_callbacks2_t *","file_deleted", 1, self )); 
  }
  arg1 = (struct svn_wc_diff_callbacks2_t *)(argp1);
  result = (svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,char const *,char const *,apr_hash_t *,void *)) ((arg1)->file_deleted);
  vresult = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_hash_t_p_void__p_svn_error_t);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks2_t_dir_added_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_diff_callbacks2_t *arg1 = 0 ;
  svn_error_t *(*arg2)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,svn_revnum_t,void *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_diff_callbacks2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_diff_callbacks2_t *","dir_added", 1, self )); 
  }
  arg1 = (struct svn_wc_diff_callbacks2_t *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg2), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_svn_revnum_t_p_void__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,svn_revnum_t,void *)","dir_added", 2, argv[0] )); 
    }
  }
  if (arg1) (arg1)->dir_added = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks2_t_dir_added_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_diff_callbacks2_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_error_t *(*result)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,svn_revnum_t,void *) = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_diff_callbacks2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_diff_callbacks2_t *","dir_added", 1, self )); 
  }
  arg1 = (struct svn_wc_diff_callbacks2_t *)(argp1);
  result = (svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,svn_revnum_t,void *)) ((arg1)->dir_added);
  vresult = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_svn_revnum_t_p_void__p_svn_error_t);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks2_t_dir_deleted_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_diff_callbacks2_t *arg1 = 0 ;
  svn_error_t *(*arg2)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,void *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_diff_callbacks2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_diff_callbacks2_t *","dir_deleted", 1, self )); 
  }
  arg1 = (struct svn_wc_diff_callbacks2_t *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg2), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_void__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,void *)","dir_deleted", 2, argv[0] )); 
    }
  }
  if (arg1) (arg1)->dir_deleted = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks2_t_dir_deleted_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_diff_callbacks2_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_error_t *(*result)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,void *) = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_diff_callbacks2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_diff_callbacks2_t *","dir_deleted", 1, self )); 
  }
  arg1 = (struct svn_wc_diff_callbacks2_t *)(argp1);
  result = (svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,void *)) ((arg1)->dir_deleted);
  vresult = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_void__p_svn_error_t);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks2_t_dir_props_changed_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_diff_callbacks2_t *arg1 = 0 ;
  svn_error_t *(*arg2)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,apr_array_header_t const *,apr_hash_t *,void *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_diff_callbacks2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_diff_callbacks2_t *","dir_props_changed", 1, self )); 
  }
  arg1 = (struct svn_wc_diff_callbacks2_t *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg2), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,apr_array_header_t const *,apr_hash_t *,void *)","dir_props_changed", 2, argv[0] )); 
    }
  }
  if (arg1) (arg1)->dir_props_changed = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks2_t_dir_props_changed_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_diff_callbacks2_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_error_t *(*result)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,apr_array_header_t const *,apr_hash_t *,void *) = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_diff_callbacks2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_diff_callbacks2_t *","dir_props_changed", 1, self )); 
  }
  arg1 = (struct svn_wc_diff_callbacks2_t *)(argp1);
  result = (svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,apr_array_header_t const *,apr_hash_t *,void *)) ((arg1)->dir_props_changed);
  vresult = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
_wrap_svn_wc_diff_callbacks2_t_allocate(VALUE self)
#else
_wrap_svn_wc_diff_callbacks2_t_allocate(int argc, VALUE *argv, VALUE self)
#endif
{
  VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_svn_wc_diff_callbacks2_t);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
  rb_obj_call_init(vresult, argc, argv);
#endif
  return vresult;
}


SWIGINTERN VALUE
_wrap_new_svn_wc_diff_callbacks2_t(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_diff_callbacks2_t *result = 0 ;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  {
    result = (struct svn_wc_diff_callbacks2_t *)calloc(1, sizeof(struct svn_wc_diff_callbacks2_t));
    DATA_PTR(self) = result;
    
    
    
  }
  return self;
fail:
  return Qnil;
}


SWIGINTERN void
free_svn_wc_diff_callbacks2_t(void *self) {
    struct svn_wc_diff_callbacks2_t *arg1 = (struct svn_wc_diff_callbacks2_t *)self;
    free((char *) arg1);
}

static swig_class SwigClassSvn_wc_diff_callbacks_t;

SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks_t_file_changed_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_diff_callbacks_t *arg1 = 0 ;
  svn_error_t *(*arg2)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,void *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_diff_callbacks_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_diff_callbacks_t *","file_changed", 1, self )); 
  }
  arg1 = (struct svn_wc_diff_callbacks_t *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg2), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_void__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,void *)","file_changed", 2, argv[0] )); 
    }
  }
  if (arg1) (arg1)->file_changed = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks_t_file_changed_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_diff_callbacks_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_error_t *(*result)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,void *) = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_diff_callbacks_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_diff_callbacks_t *","file_changed", 1, self )); 
  }
  arg1 = (struct svn_wc_diff_callbacks_t *)(argp1);
  result = (svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,void *)) ((arg1)->file_changed);
  vresult = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_void__p_svn_error_t);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks_t_file_added_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_diff_callbacks_t *arg1 = 0 ;
  svn_error_t *(*arg2)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,void *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_diff_callbacks_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_diff_callbacks_t *","file_added", 1, self )); 
  }
  arg1 = (struct svn_wc_diff_callbacks_t *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg2), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_void__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,void *)","file_added", 2, argv[0] )); 
    }
  }
  if (arg1) (arg1)->file_added = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks_t_file_added_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_diff_callbacks_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_error_t *(*result)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,void *) = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_diff_callbacks_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_diff_callbacks_t *","file_added", 1, self )); 
  }
  arg1 = (struct svn_wc_diff_callbacks_t *)(argp1);
  result = (svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,void *)) ((arg1)->file_added);
  vresult = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_void__p_svn_error_t);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks_t_file_deleted_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_diff_callbacks_t *arg1 = 0 ;
  svn_error_t *(*arg2)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,char const *,char const *,void *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_diff_callbacks_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_diff_callbacks_t *","file_deleted", 1, self )); 
  }
  arg1 = (struct svn_wc_diff_callbacks_t *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg2), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_void__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,char const *,char const *,void *)","file_deleted", 2, argv[0] )); 
    }
  }
  if (arg1) (arg1)->file_deleted = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks_t_file_deleted_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_diff_callbacks_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_error_t *(*result)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,char const *,char const *,void *) = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_diff_callbacks_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_diff_callbacks_t *","file_deleted", 1, self )); 
  }
  arg1 = (struct svn_wc_diff_callbacks_t *)(argp1);
  result = (svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,char const *,char const *,void *)) ((arg1)->file_deleted);
  vresult = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_void__p_svn_error_t);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks_t_dir_added_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_diff_callbacks_t *arg1 = 0 ;
  svn_error_t *(*arg2)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,svn_revnum_t,void *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_diff_callbacks_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_diff_callbacks_t *","dir_added", 1, self )); 
  }
  arg1 = (struct svn_wc_diff_callbacks_t *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg2), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_svn_revnum_t_p_void__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,svn_revnum_t,void *)","dir_added", 2, argv[0] )); 
    }
  }
  if (arg1) (arg1)->dir_added = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks_t_dir_added_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_diff_callbacks_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_error_t *(*result)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,svn_revnum_t,void *) = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_diff_callbacks_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_diff_callbacks_t *","dir_added", 1, self )); 
  }
  arg1 = (struct svn_wc_diff_callbacks_t *)(argp1);
  result = (svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,svn_revnum_t,void *)) ((arg1)->dir_added);
  vresult = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_svn_revnum_t_p_void__p_svn_error_t);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks_t_dir_deleted_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_diff_callbacks_t *arg1 = 0 ;
  svn_error_t *(*arg2)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,void *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_diff_callbacks_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_diff_callbacks_t *","dir_deleted", 1, self )); 
  }
  arg1 = (struct svn_wc_diff_callbacks_t *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg2), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_void__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,void *)","dir_deleted", 2, argv[0] )); 
    }
  }
  if (arg1) (arg1)->dir_deleted = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks_t_dir_deleted_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_diff_callbacks_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_error_t *(*result)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,void *) = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_diff_callbacks_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_diff_callbacks_t *","dir_deleted", 1, self )); 
  }
  arg1 = (struct svn_wc_diff_callbacks_t *)(argp1);
  result = (svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,void *)) ((arg1)->dir_deleted);
  vresult = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_void__p_svn_error_t);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks_t_props_changed_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_diff_callbacks_t *arg1 = 0 ;
  svn_error_t *(*arg2)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,apr_array_header_t const *,apr_hash_t *,void *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_diff_callbacks_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_diff_callbacks_t *","props_changed", 1, self )); 
  }
  arg1 = (struct svn_wc_diff_callbacks_t *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg2), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,apr_array_header_t const *,apr_hash_t *,void *)","props_changed", 2, argv[0] )); 
    }
  }
  if (arg1) (arg1)->props_changed = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks_t_props_changed_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_diff_callbacks_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_error_t *(*result)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,apr_array_header_t const *,apr_hash_t *,void *) = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_diff_callbacks_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_diff_callbacks_t *","props_changed", 1, self )); 
  }
  arg1 = (struct svn_wc_diff_callbacks_t *)(argp1);
  result = (svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,apr_array_header_t const *,apr_hash_t *,void *)) ((arg1)->props_changed);
  vresult = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
_wrap_svn_wc_diff_callbacks_t_allocate(VALUE self)
#else
_wrap_svn_wc_diff_callbacks_t_allocate(int argc, VALUE *argv, VALUE self)
#endif
{
  VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_svn_wc_diff_callbacks_t);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
  rb_obj_call_init(vresult, argc, argv);
#endif
  return vresult;
}


SWIGINTERN VALUE
_wrap_new_svn_wc_diff_callbacks_t(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_diff_callbacks_t *result = 0 ;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  {
    result = (struct svn_wc_diff_callbacks_t *)calloc(1, sizeof(struct svn_wc_diff_callbacks_t));
    DATA_PTR(self) = result;
    
    
    
  }
  return self;
fail:
  return Qnil;
}


SWIGINTERN void
free_svn_wc_diff_callbacks_t(void *self) {
    struct svn_wc_diff_callbacks_t *arg1 = (struct svn_wc_diff_callbacks_t *)self;
    free((char *) arg1);
}

SWIGINTERN VALUE
_wrap_svn_wc_check_wc2(int argc, VALUE *argv, VALUE self) {
  int *arg1 = 0 ;
  svn_wc_context_t *arg2 = 0 ;
  char *arg3 = 0 ;
  apr_pool_t *arg4 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int temp1 ;
  int res1 = SWIG_TMPOBJ ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_svn_wc_context_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_context_t *","svn_wc_check_wc2", 2, argv[0] )); 
  }
  arg2 = (svn_wc_context_t *)(argp2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_wc_check_wc2", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_wc_check_wc2(arg1,arg2,(char const *)arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (SWIG_IsTmpObj(res1)) {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_From_int((*arg1)), 0);
  } else {
    int new_flags = SWIG_IsNewObj(res1) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj((void*)(arg1), SWIGTYPE_p_int, new_flags), 0);
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_check_wc(int argc, VALUE *argv, VALUE self) {
  char *arg1 = 0 ;
  int *arg2 = 0 ;
  apr_pool_t *arg3 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  int temp2 ;
  int res2 = SWIG_TMPOBJ ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg2 = &temp2;
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_wc_check_wc", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  if (argc > 1) {
    
  }
  {
    result = (svn_error_t *)svn_wc_check_wc((char const *)arg1,arg2,arg3);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (SWIG_IsTmpObj(res2)) {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_From_int((*arg2)), 0);
  } else {
    int new_flags = SWIG_IsNewObj(res2) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj((void*)(arg2), SWIGTYPE_p_int, new_flags), 0);
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_has_binary_prop(int argc, VALUE *argv, VALUE self) {
  svn_boolean_t *arg1 = 0 ;
  char *arg2 = 0 ;
  svn_wc_adm_access_t *arg3 = 0 ;
  apr_pool_t *arg4 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_boolean_t temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_wc_has_binary_prop", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_ConvertPtr(argv[1], &argp3,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_has_binary_prop", 3, argv[1] )); 
  }
  arg3 = (svn_wc_adm_access_t *)(argp3);
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_wc_has_binary_prop(arg1,(char const *)arg2,arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, *arg1 ? Qtrue : Qfalse, 0);
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_text_modified_p2(int argc, VALUE *argv, VALUE self) {
  svn_boolean_t *arg1 = 0 ;
  svn_wc_context_t *arg2 = 0 ;
  char *arg3 = 0 ;
  svn_boolean_t arg4 ;
  apr_pool_t *arg5 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_boolean_t temp1 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 3) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_svn_wc_context_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_context_t *","svn_wc_text_modified_p2", 2, argv[0] )); 
  }
  arg2 = (svn_wc_context_t *)(argp2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_wc_text_modified_p2", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  arg4 = RTEST(argv[2]);
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_wc_text_modified_p2(arg1,arg2,(char const *)arg3,arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, *arg1 ? Qtrue : Qfalse, 0);
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_text_modified_p(int argc, VALUE *argv, VALUE self) {
  svn_boolean_t *arg1 = 0 ;
  char *arg2 = 0 ;
  svn_boolean_t arg3 ;
  svn_wc_adm_access_t *arg4 = 0 ;
  apr_pool_t *arg5 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_boolean_t temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 3) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_wc_text_modified_p", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  arg3 = RTEST(argv[1]);
  res4 = SWIG_ConvertPtr(argv[2], &argp4,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_text_modified_p", 4, argv[2] )); 
  }
  arg4 = (svn_wc_adm_access_t *)(argp4);
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_wc_text_modified_p(arg1,(char const *)arg2,arg3,arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, *arg1 ? Qtrue : Qfalse, 0);
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_props_modified_p2(int argc, VALUE *argv, VALUE self) {
  svn_boolean_t *arg1 = 0 ;
  svn_wc_context_t *arg2 = 0 ;
  char *arg3 = 0 ;
  apr_pool_t *arg4 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_boolean_t temp1 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_svn_wc_context_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_context_t *","svn_wc_props_modified_p2", 2, argv[0] )); 
  }
  arg2 = (svn_wc_context_t *)(argp2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_wc_props_modified_p2", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_wc_props_modified_p2(arg1,arg2,(char const *)arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, *arg1 ? Qtrue : Qfalse, 0);
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_props_modified_p(int argc, VALUE *argv, VALUE self) {
  svn_boolean_t *arg1 = 0 ;
  char *arg2 = 0 ;
  svn_wc_adm_access_t *arg3 = 0 ;
  apr_pool_t *arg4 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_boolean_t temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_wc_props_modified_p", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_ConvertPtr(argv[1], &argp3,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_props_modified_p", 3, argv[1] )); 
  }
  arg3 = (svn_wc_adm_access_t *)(argp3);
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_wc_props_modified_p(arg1,(char const *)arg2,arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, *arg1 ? Qtrue : Qfalse, 0);
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


static swig_class SwigClassSvn_wc_entry_t;

SWIGINTERN VALUE
_wrap_svn_wc_entry_t_name_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","name", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","name", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->name) free((char *)arg1->name);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->name = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_name_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","name", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  result = (char *) ((arg1)->name);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_revision_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  svn_revnum_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","revision", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  ecode2 = SWIG_AsVal_long(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_revnum_t","revision", 2, argv[0] ));
  } 
  arg2 = (svn_revnum_t)(val2);
  if (arg1) (arg1)->revision = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_revision_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_revnum_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","revision", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  result = (svn_revnum_t) ((arg1)->revision);
  vresult = SWIG_From_long((long)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_url_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","url", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","url", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->url) free((char *)arg1->url);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->url = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_url_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","url", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  result = (char *) ((arg1)->url);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_repos_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","repos", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  {
    if (NIL_P(argv[0])) {
      arg2 = NULL;
    } else {
      arg2 = StringValuePtr(argv[0]);
    }
  }
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->repos) free((char *)arg1->repos);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->repos = copied;
  }
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_repos_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","repos", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  result = (char *) ((arg1)->repos);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_uuid_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","uuid", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  {
    if (NIL_P(argv[0])) {
      arg2 = NULL;
    } else {
      arg2 = StringValuePtr(argv[0]);
    }
  }
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->uuid) free((char *)arg1->uuid);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->uuid = copied;
  }
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_uuid_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","uuid", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  result = (char *) ((arg1)->uuid);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_kind_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  svn_node_kind_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","kind", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_node_kind_t","kind", 2, argv[0] ));
  } 
  arg2 = (svn_node_kind_t)(val2);
  if (arg1) (arg1)->kind = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_kind_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_node_kind_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","kind", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  result = (svn_node_kind_t) ((arg1)->kind);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_schedule_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  svn_wc_schedule_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","schedule", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_wc_schedule_t","schedule", 2, argv[0] ));
  } 
  arg2 = (svn_wc_schedule_t)(val2);
  if (arg1) (arg1)->schedule = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_schedule_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_wc_schedule_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","schedule", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  result = (svn_wc_schedule_t) ((arg1)->schedule);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_copied_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  svn_boolean_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","copied", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  arg2 = RTEST(argv[0]);
  if (arg1) (arg1)->copied = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_copied_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","copied", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  result = (svn_boolean_t) ((arg1)->copied);
  vresult = result ? Qtrue : Qfalse;
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_deleted_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  svn_boolean_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","deleted", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  arg2 = RTEST(argv[0]);
  if (arg1) (arg1)->deleted = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_deleted_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","deleted", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  result = (svn_boolean_t) ((arg1)->deleted);
  vresult = result ? Qtrue : Qfalse;
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_absent_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  svn_boolean_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","absent", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  arg2 = RTEST(argv[0]);
  if (arg1) (arg1)->absent = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_absent_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","absent", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  result = (svn_boolean_t) ((arg1)->absent);
  vresult = result ? Qtrue : Qfalse;
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_incomplete_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  svn_boolean_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","incomplete", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  arg2 = RTEST(argv[0]);
  if (arg1) (arg1)->incomplete = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_incomplete_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","incomplete", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  result = (svn_boolean_t) ((arg1)->incomplete);
  vresult = result ? Qtrue : Qfalse;
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_copyfrom_url_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","copyfrom_url", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  {
    if (NIL_P(argv[0])) {
      arg2 = NULL;
    } else {
      arg2 = StringValuePtr(argv[0]);
    }
  }
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->copyfrom_url) free((char *)arg1->copyfrom_url);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->copyfrom_url = copied;
  }
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_copyfrom_url_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","copyfrom_url", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  result = (char *) ((arg1)->copyfrom_url);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_copyfrom_rev_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  svn_revnum_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","copyfrom_rev", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  ecode2 = SWIG_AsVal_long(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_revnum_t","copyfrom_rev", 2, argv[0] ));
  } 
  arg2 = (svn_revnum_t)(val2);
  if (arg1) (arg1)->copyfrom_rev = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_copyfrom_rev_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_revnum_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","copyfrom_rev", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  result = (svn_revnum_t) ((arg1)->copyfrom_rev);
  vresult = SWIG_From_long((long)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_conflict_old_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","conflict_old", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","conflict_old", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->conflict_old) free((char *)arg1->conflict_old);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->conflict_old = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_conflict_old_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","conflict_old", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  result = (char *) ((arg1)->conflict_old);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_conflict_new_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","conflict_new", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","conflict_new", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->conflict_new) free((char *)arg1->conflict_new);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->conflict_new = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_conflict_new_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","conflict_new", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  result = (char *) ((arg1)->conflict_new);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_conflict_wrk_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","conflict_wrk", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","conflict_wrk", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->conflict_wrk) free((char *)arg1->conflict_wrk);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->conflict_wrk = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_conflict_wrk_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","conflict_wrk", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  result = (char *) ((arg1)->conflict_wrk);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_prejfile_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","prejfile", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","prejfile", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->prejfile) free((char *)arg1->prejfile);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->prejfile = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_prejfile_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","prejfile", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  result = (char *) ((arg1)->prejfile);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_text_time_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  apr_time_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","text_time", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  {
    arg2 = (apr_time_t)NUM2LL(argv[0]);
  }
  if (arg1) (arg1)->text_time = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_text_time_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  apr_time_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","text_time", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  result =  ((arg1)->text_time);
  {
    vresult = LL2NUM((apr_time_t)(result));
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_prop_time_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  apr_time_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","prop_time", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  {
    arg2 = (apr_time_t)NUM2LL(argv[0]);
  }
  if (arg1) (arg1)->prop_time = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_prop_time_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  apr_time_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","prop_time", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  result =  ((arg1)->prop_time);
  {
    vresult = LL2NUM((apr_time_t)(result));
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_checksum_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","checksum", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","checksum", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->checksum) free((char *)arg1->checksum);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->checksum = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_checksum_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","checksum", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  result = (char *) ((arg1)->checksum);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_cmt_rev_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  svn_revnum_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","cmt_rev", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  ecode2 = SWIG_AsVal_long(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_revnum_t","cmt_rev", 2, argv[0] ));
  } 
  arg2 = (svn_revnum_t)(val2);
  if (arg1) (arg1)->cmt_rev = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_cmt_rev_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_revnum_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","cmt_rev", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  result = (svn_revnum_t) ((arg1)->cmt_rev);
  vresult = SWIG_From_long((long)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_cmt_date_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  apr_time_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","cmt_date", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  {
    arg2 = (apr_time_t)NUM2LL(argv[0]);
  }
  if (arg1) (arg1)->cmt_date = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_cmt_date_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  apr_time_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","cmt_date", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  result =  ((arg1)->cmt_date);
  {
    vresult = LL2NUM((apr_time_t)(result));
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_cmt_author_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","cmt_author", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","cmt_author", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->cmt_author) free((char *)arg1->cmt_author);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->cmt_author = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_cmt_author_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","cmt_author", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  result = (char *) ((arg1)->cmt_author);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_lock_token_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","lock_token", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  {
    if (NIL_P(argv[0])) {
      arg2 = NULL;
    } else {
      arg2 = StringValuePtr(argv[0]);
    }
  }
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->lock_token) free((char *)arg1->lock_token);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->lock_token = copied;
  }
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_lock_token_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","lock_token", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  result = (char *) ((arg1)->lock_token);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_lock_owner_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","lock_owner", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","lock_owner", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->lock_owner) free((char *)arg1->lock_owner);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->lock_owner = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_lock_owner_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","lock_owner", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  result = (char *) ((arg1)->lock_owner);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_lock_comment_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","lock_comment", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","lock_comment", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->lock_comment) free((char *)arg1->lock_comment);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->lock_comment = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_lock_comment_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","lock_comment", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  result = (char *) ((arg1)->lock_comment);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_lock_creation_date_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  apr_time_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","lock_creation_date", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  {
    arg2 = (apr_time_t)NUM2LL(argv[0]);
  }
  if (arg1) (arg1)->lock_creation_date = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_lock_creation_date_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  apr_time_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","lock_creation_date", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  result =  ((arg1)->lock_creation_date);
  {
    vresult = LL2NUM((apr_time_t)(result));
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_has_props_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  svn_boolean_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","has_props", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  arg2 = RTEST(argv[0]);
  if (arg1) (arg1)->has_props = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_has_props_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","has_props", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  result = (svn_boolean_t) ((arg1)->has_props);
  vresult = result ? Qtrue : Qfalse;
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_has_prop_mods_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  svn_boolean_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","has_prop_mods", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  arg2 = RTEST(argv[0]);
  if (arg1) (arg1)->has_prop_mods = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_has_prop_mods_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","has_prop_mods", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  result = (svn_boolean_t) ((arg1)->has_prop_mods);
  vresult = result ? Qtrue : Qfalse;
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_cachable_props_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","cachable_props", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","cachable_props", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->cachable_props) free((char *)arg1->cachable_props);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->cachable_props = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_cachable_props_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","cachable_props", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  result = (char *) ((arg1)->cachable_props);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_present_props_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","present_props", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","present_props", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->present_props) free((char *)arg1->present_props);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->present_props = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_present_props_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","present_props", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  result = (char *) ((arg1)->present_props);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_changelist_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","changelist", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  {
    arg2 = StringValueCStr(argv[0]);
  }
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->changelist) free((char *)arg1->changelist);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->changelist = copied;
  }
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_changelist_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","changelist", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  result = (char *) ((arg1)->changelist);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_working_size_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  apr_off_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","working_size", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  {
    arg2 = (apr_off_t)NUM2LL(argv[0]);
  }
  if (arg1) (arg1)->working_size = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_working_size_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  apr_off_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","working_size", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  result =  ((arg1)->working_size);
  {
    vresult = LL2NUM((apr_off_t)(result));
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_keep_local_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  svn_boolean_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","keep_local", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  arg2 = RTEST(argv[0]);
  if (arg1) (arg1)->keep_local = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_keep_local_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","keep_local", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  result = (svn_boolean_t) ((arg1)->keep_local);
  vresult = result ? Qtrue : Qfalse;
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_depth_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  svn_depth_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","depth", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  {
    arg2 = svn_swig_rb_to_depth(argv[0]);
  }
  if (arg1) (arg1)->depth = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_depth_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_depth_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","depth", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  result = (svn_depth_t) ((arg1)->depth);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_tree_conflict_data_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","tree_conflict_data", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","tree_conflict_data", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->tree_conflict_data) free((char *)arg1->tree_conflict_data);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->tree_conflict_data = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_tree_conflict_data_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","tree_conflict_data", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  result = (char *) ((arg1)->tree_conflict_data);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_file_external_path_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","file_external_path", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","file_external_path", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->file_external_path) free((char *)arg1->file_external_path);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->file_external_path = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_file_external_path_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","file_external_path", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  result = (char *) ((arg1)->file_external_path);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_file_external_peg_rev_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  svn_opt_revision_t *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_opt_revision_t rev2 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","file_external_peg_rev", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  {
    arg2 = &rev2;
    svn_swig_rb_set_revision(&rev2, argv[0]);
  }
  if (arg1) (arg1)->file_external_peg_rev = *arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_file_external_peg_rev_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_opt_revision_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","file_external_peg_rev", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  result = (svn_opt_revision_t *)& ((arg1)->file_external_peg_rev);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_opt_revision_t, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_file_external_rev_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  svn_opt_revision_t *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_opt_revision_t rev2 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","file_external_rev", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  {
    arg2 = &rev2;
    svn_swig_rb_set_revision(&rev2, argv[0]);
  }
  if (arg1) (arg1)->file_external_rev = *arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_t_file_external_rev_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_opt_revision_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_t *","file_external_rev", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_t *)(argp1);
  result = (svn_opt_revision_t *)& ((arg1)->file_external_rev);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_opt_revision_t, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
_wrap_svn_wc_entry_t_allocate(VALUE self)
#else
_wrap_svn_wc_entry_t_allocate(int argc, VALUE *argv, VALUE self)
#endif
{
  VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_svn_wc_entry_t);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
  rb_obj_call_init(vresult, argc, argv);
#endif
  return vresult;
}


SWIGINTERN VALUE
_wrap_new_svn_wc_entry_t(int argc, VALUE *argv, VALUE self) {
  char *arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  svn_boolean_t arg3 ;
  apr_pool_t *arg4 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  struct svn_wc_entry_t *result = 0 ;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 3) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_wc_entry_t", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_entry_t", 2, argv[1] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  arg3 = RTEST(argv[2]);
  if (argc > 3) {
    
  }
  {
    result = (struct svn_wc_entry_t *)new_svn_wc_entry_t((char const *)arg1,arg2,arg3,arg4);
    DATA_PTR(self) = result;
    
    
    
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return self;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN void delete_svn_wc_entry_t(struct svn_wc_entry_t *self){
  }
SWIGINTERN void
free_svn_wc_entry_t(void *self) {
    struct svn_wc_entry_t *arg1 = (struct svn_wc_entry_t *)self;
    delete_svn_wc_entry_t(arg1);
}

SWIGINTERN VALUE
_wrap_svn_wc_entries_read(int argc, VALUE *argv, VALUE self) {
  apr_hash_t **arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  svn_boolean_t arg3 ;
  apr_pool_t *arg4 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  apr_hash_t *temp1 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_entries_read", 2, argv[0] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  arg3 = RTEST(argv[1]);
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_wc_entries_read(arg1,arg2,arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, svn_swig_rb_apr_hash_to_hash_swig_type(*arg1,
        "svn_wc_entry_t *"), 0)
    
    ;
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_dup(int argc, VALUE *argv, VALUE self) {
  svn_wc_entry_t *arg1 = 0 ;
  apr_pool_t *arg2 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_wc_entry_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_entry_t const *","svn_wc_entry_dup", 1, argv[0] )); 
  }
  arg1 = (svn_wc_entry_t *)(argp1);
  if (argc > 1) {
    
  }
  {
    result = (svn_wc_entry_t *)svn_wc_entry_dup((struct svn_wc_entry_t const *)arg1,arg2);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


static swig_class SwigClassSvn_wc_info_t;

SWIGINTERN VALUE
_wrap_svn_wc_info_t_schedule_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_info_t *arg1 = 0 ;
  svn_wc_schedule_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_info_t *","schedule", 1, self )); 
  }
  arg1 = (struct svn_wc_info_t *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_wc_schedule_t","schedule", 2, argv[0] ));
  } 
  arg2 = (svn_wc_schedule_t)(val2);
  if (arg1) (arg1)->schedule = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_info_t_schedule_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_info_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_wc_schedule_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_info_t *","schedule", 1, self )); 
  }
  arg1 = (struct svn_wc_info_t *)(argp1);
  result = (svn_wc_schedule_t) ((arg1)->schedule);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_info_t_copyfrom_url_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_info_t *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_info_t *","copyfrom_url", 1, self )); 
  }
  arg1 = (struct svn_wc_info_t *)(argp1);
  {
    if (NIL_P(argv[0])) {
      arg2 = NULL;
    } else {
      arg2 = StringValuePtr(argv[0]);
    }
  }
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->copyfrom_url) free((char *)arg1->copyfrom_url);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->copyfrom_url = copied;
  }
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_info_t_copyfrom_url_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_info_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_info_t *","copyfrom_url", 1, self )); 
  }
  arg1 = (struct svn_wc_info_t *)(argp1);
  result = (char *) ((arg1)->copyfrom_url);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_info_t_copyfrom_rev_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_info_t *arg1 = 0 ;
  svn_revnum_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_info_t *","copyfrom_rev", 1, self )); 
  }
  arg1 = (struct svn_wc_info_t *)(argp1);
  ecode2 = SWIG_AsVal_long(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_revnum_t","copyfrom_rev", 2, argv[0] ));
  } 
  arg2 = (svn_revnum_t)(val2);
  if (arg1) (arg1)->copyfrom_rev = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_info_t_copyfrom_rev_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_info_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_revnum_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_info_t *","copyfrom_rev", 1, self )); 
  }
  arg1 = (struct svn_wc_info_t *)(argp1);
  result = (svn_revnum_t) ((arg1)->copyfrom_rev);
  vresult = SWIG_From_long((long)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_info_t_checksum_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_info_t *arg1 = 0 ;
  svn_checksum_t *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_info_t *","checksum", 1, self )); 
  }
  arg1 = (struct svn_wc_info_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_svn_checksum_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_checksum_t const *","checksum", 2, argv[0] )); 
  }
  arg2 = (svn_checksum_t *)(argp2);
  if (arg1) (arg1)->checksum = (svn_checksum_t const *)arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_info_t_checksum_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_info_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_checksum_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_info_t *","checksum", 1, self )); 
  }
  arg1 = (struct svn_wc_info_t *)(argp1);
  result = (svn_checksum_t *) ((arg1)->checksum);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_checksum_t, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_info_t_changelist_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_info_t *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_info_t *","changelist", 1, self )); 
  }
  arg1 = (struct svn_wc_info_t *)(argp1);
  {
    arg2 = StringValueCStr(argv[0]);
  }
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->changelist) free((char *)arg1->changelist);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->changelist = copied;
  }
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_info_t_changelist_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_info_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_info_t *","changelist", 1, self )); 
  }
  arg1 = (struct svn_wc_info_t *)(argp1);
  result = (char *) ((arg1)->changelist);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_info_t_depth_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_info_t *arg1 = 0 ;
  svn_depth_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_info_t *","depth", 1, self )); 
  }
  arg1 = (struct svn_wc_info_t *)(argp1);
  {
    arg2 = svn_swig_rb_to_depth(argv[0]);
  }
  if (arg1) (arg1)->depth = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_info_t_depth_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_info_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_depth_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_info_t *","depth", 1, self )); 
  }
  arg1 = (struct svn_wc_info_t *)(argp1);
  result = (svn_depth_t) ((arg1)->depth);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_info_t_recorded_size_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_info_t *arg1 = 0 ;
  svn_filesize_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_info_t *","recorded_size", 1, self )); 
  }
  arg1 = (struct svn_wc_info_t *)(argp1);
  {
    arg2 = (svn_filesize_t)NUM2LL(argv[0]);
  }
  if (arg1) (arg1)->recorded_size = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_info_t_recorded_size_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_info_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_filesize_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_info_t *","recorded_size", 1, self )); 
  }
  arg1 = (struct svn_wc_info_t *)(argp1);
  result =  ((arg1)->recorded_size);
  {
    vresult = LL2NUM((svn_filesize_t)(result));
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_info_t_recorded_time_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_info_t *arg1 = 0 ;
  apr_time_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_info_t *","recorded_time", 1, self )); 
  }
  arg1 = (struct svn_wc_info_t *)(argp1);
  {
    arg2 = (apr_time_t)NUM2LL(argv[0]);
  }
  if (arg1) (arg1)->recorded_time = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_info_t_recorded_time_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_info_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  apr_time_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_info_t *","recorded_time", 1, self )); 
  }
  arg1 = (struct svn_wc_info_t *)(argp1);
  result =  ((arg1)->recorded_time);
  {
    vresult = LL2NUM((apr_time_t)(result));
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_info_t_conflicts_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_info_t *arg1 = 0 ;
  apr_array_header_t *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_info_t *","conflicts", 1, self )); 
  }
  arg1 = (struct svn_wc_info_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_apr_array_header_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "apr_array_header_t const *","conflicts", 2, argv[0] )); 
  }
  arg2 = (apr_array_header_t *)(argp2);
  if (arg1) (arg1)->conflicts = (apr_array_header_t const *)arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_info_t_conflicts_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_info_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  apr_array_header_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_info_t *","conflicts", 1, self )); 
  }
  arg1 = (struct svn_wc_info_t *)(argp1);
  result = (apr_array_header_t *) ((arg1)->conflicts);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_apr_array_header_t, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_info_t_wcroot_abspath_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_info_t *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_info_t *","wcroot_abspath", 1, self )); 
  }
  arg1 = (struct svn_wc_info_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","wcroot_abspath", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->wcroot_abspath) free((char *)arg1->wcroot_abspath);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->wcroot_abspath = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_info_t_wcroot_abspath_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_info_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_info_t *","wcroot_abspath", 1, self )); 
  }
  arg1 = (struct svn_wc_info_t *)(argp1);
  result = (char *) ((arg1)->wcroot_abspath);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_info_t_moved_from_abspath_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_info_t *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_info_t *","moved_from_abspath", 1, self )); 
  }
  arg1 = (struct svn_wc_info_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","moved_from_abspath", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->moved_from_abspath) free((char *)arg1->moved_from_abspath);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->moved_from_abspath = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_info_t_moved_from_abspath_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_info_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_info_t *","moved_from_abspath", 1, self )); 
  }
  arg1 = (struct svn_wc_info_t *)(argp1);
  result = (char *) ((arg1)->moved_from_abspath);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_info_t_moved_to_abspath_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_info_t *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_info_t *","moved_to_abspath", 1, self )); 
  }
  arg1 = (struct svn_wc_info_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","moved_to_abspath", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->moved_to_abspath) free((char *)arg1->moved_to_abspath);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->moved_to_abspath = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_info_t_moved_to_abspath_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_info_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_info_t *","moved_to_abspath", 1, self )); 
  }
  arg1 = (struct svn_wc_info_t *)(argp1);
  result = (char *) ((arg1)->moved_to_abspath);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
_wrap_svn_wc_info_t_allocate(VALUE self)
#else
_wrap_svn_wc_info_t_allocate(int argc, VALUE *argv, VALUE self)
#endif
{
  VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_svn_wc_info_t);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
  rb_obj_call_init(vresult, argc, argv);
#endif
  return vresult;
}


SWIGINTERN VALUE
_wrap_new_svn_wc_info_t(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_info_t *result = 0 ;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  {
    result = (struct svn_wc_info_t *)calloc(1, sizeof(struct svn_wc_info_t));
    DATA_PTR(self) = result;
    
    
    
  }
  return self;
fail:
  return Qnil;
}


SWIGINTERN void
free_svn_wc_info_t(void *self) {
    struct svn_wc_info_t *arg1 = (struct svn_wc_info_t *)self;
    free((char *) arg1);
}

SWIGINTERN VALUE
_wrap_svn_wc_info_dup(int argc, VALUE *argv, VALUE self) {
  svn_wc_info_t *arg1 = 0 ;
  apr_pool_t *arg2 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_wc_info_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_info_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_info_t const *","svn_wc_info_dup", 1, argv[0] )); 
  }
  arg1 = (svn_wc_info_t *)(argp1);
  if (argc > 1) {
    
  }
  {
    result = (svn_wc_info_t *)svn_wc_info_dup((struct svn_wc_info_t const *)arg1,arg2);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_wc_info_t, 0 |  0 );
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_conflicted_p3(int argc, VALUE *argv, VALUE self) {
  svn_boolean_t *arg1 = 0 ;
  svn_boolean_t *arg2 = 0 ;
  svn_boolean_t *arg3 = 0 ;
  svn_wc_context_t *arg4 = 0 ;
  char *arg5 = 0 ;
  apr_pool_t *arg6 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_boolean_t temp1 ;
  svn_boolean_t temp2 ;
  svn_boolean_t temp3 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  int res5 ;
  char *buf5 = 0 ;
  int alloc5 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  arg2 = &temp2;
  arg3 = &temp3;
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res4 = SWIG_ConvertPtr(argv[0], &argp4,SWIGTYPE_p_svn_wc_context_t, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "svn_wc_context_t *","svn_wc_conflicted_p3", 4, argv[0] )); 
  }
  arg4 = (svn_wc_context_t *)(argp4);
  res5 = SWIG_AsCharPtrAndSize(argv[1], &buf5, NULL, &alloc5);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "char const *","svn_wc_conflicted_p3", 5, argv[1] ));
  }
  arg5 = (char *)(buf5);
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_wc_conflicted_p3(arg1,arg2,arg3,arg4,(char const *)arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, *arg1 ? Qtrue : Qfalse, 0);
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, *arg2 ? Qtrue : Qfalse, 0);
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, *arg3 ? Qtrue : Qfalse, 0);
  }
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_conflicted_p2(int argc, VALUE *argv, VALUE self) {
  svn_boolean_t *arg1 = 0 ;
  svn_boolean_t *arg2 = 0 ;
  svn_boolean_t *arg3 = 0 ;
  char *arg4 = 0 ;
  svn_wc_adm_access_t *arg5 = 0 ;
  apr_pool_t *arg6 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_boolean_t temp1 ;
  svn_boolean_t temp2 ;
  svn_boolean_t temp3 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  arg2 = &temp2;
  arg3 = &temp3;
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res4 = SWIG_AsCharPtrAndSize(argv[0], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_wc_conflicted_p2", 4, argv[0] ));
  }
  arg4 = (char *)(buf4);
  res5 = SWIG_ConvertPtr(argv[1], &argp5,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_conflicted_p2", 5, argv[1] )); 
  }
  arg5 = (svn_wc_adm_access_t *)(argp5);
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_wc_conflicted_p2(arg1,arg2,arg3,(char const *)arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, *arg1 ? Qtrue : Qfalse, 0);
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, *arg2 ? Qtrue : Qfalse, 0);
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, *arg3 ? Qtrue : Qfalse, 0);
  }
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_conflicted_p(int argc, VALUE *argv, VALUE self) {
  svn_boolean_t *arg1 = 0 ;
  svn_boolean_t *arg2 = 0 ;
  char *arg3 = 0 ;
  svn_wc_entry_t *arg4 = 0 ;
  apr_pool_t *arg5 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_boolean_t temp1 ;
  svn_boolean_t temp2 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  arg2 = &temp2;
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res3 = SWIG_AsCharPtrAndSize(argv[0], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_wc_conflicted_p", 3, argv[0] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_ConvertPtr(argv[1], &argp4,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "svn_wc_entry_t const *","svn_wc_conflicted_p", 4, argv[1] )); 
  }
  arg4 = (svn_wc_entry_t *)(argp4);
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_wc_conflicted_p(arg1,arg2,(char const *)arg3,(struct svn_wc_entry_t const *)arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, *arg1 ? Qtrue : Qfalse, 0);
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, *arg2 ? Qtrue : Qfalse, 0);
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_get_ancestry(int argc, VALUE *argv, VALUE self) {
  char **arg1 = 0 ;
  svn_revnum_t *arg2 = 0 ;
  char *arg3 = 0 ;
  svn_wc_adm_access_t *arg4 = 0 ;
  apr_pool_t *arg5 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  char *temp1 ;
  svn_revnum_t temp2 ;
  int res2 = SWIG_TMPOBJ ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  arg2 = &temp2;
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res3 = SWIG_AsCharPtrAndSize(argv[0], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_wc_get_ancestry", 3, argv[0] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_ConvertPtr(argv[1], &argp4,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_get_ancestry", 4, argv[1] )); 
  }
  arg4 = (svn_wc_adm_access_t *)(argp4);
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_wc_get_ancestry(arg1,arg2,(char const *)arg3,arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    if (*arg1) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new2(*arg1), 0);
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil, 0);
    }
  }
  if (SWIG_IsTmpObj(res2)) {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_From_long((*arg2)), 0);
  } else {
    int new_flags = SWIG_IsNewObj(res2) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj((void*)(arg2), SWIGTYPE_p_long, new_flags), 0);
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


static swig_class SwigClassSvn_wc_entry_callbacks2_t;

SWIGINTERN VALUE
_wrap_svn_wc_entry_callbacks2_t_found_entry_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_callbacks2_t *arg1 = 0 ;
  svn_error_t *(*arg2)(char const *,svn_wc_entry_t const *,void *,apr_pool_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_callbacks2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_callbacks2_t *","found_entry", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_callbacks2_t *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg2), SWIGTYPE_p_f_p_q_const__char_p_q_const__svn_wc_entry_t_p_void_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_error_t *(*)(char const *,svn_wc_entry_t const *,void *,apr_pool_t *)","found_entry", 2, argv[0] )); 
    }
  }
  if (arg1) (arg1)->found_entry = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_callbacks2_t_found_entry_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_callbacks2_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_error_t *(*result)(char const *,svn_wc_entry_t const *,void *,apr_pool_t *) = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_callbacks2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_callbacks2_t *","found_entry", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_callbacks2_t *)(argp1);
  result = (svn_error_t *(*)(char const *,svn_wc_entry_t const *,void *,apr_pool_t *)) ((arg1)->found_entry);
  vresult = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_q_const__char_p_q_const__svn_wc_entry_t_p_void_p_apr_pool_t__p_svn_error_t);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_callbacks2_t_handle_error_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_callbacks2_t *arg1 = 0 ;
  svn_error_t *(*arg2)(char const *,svn_error_t *,void *,apr_pool_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_callbacks2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_callbacks2_t *","handle_error", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_callbacks2_t *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg2), SWIGTYPE_p_f_p_q_const__char_p_svn_error_t_p_void_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_error_t *(*)(char const *,svn_error_t *,void *,apr_pool_t *)","handle_error", 2, argv[0] )); 
    }
  }
  if (arg1) (arg1)->handle_error = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_callbacks2_t_handle_error_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_callbacks2_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_error_t *(*result)(char const *,svn_error_t *,void *,apr_pool_t *) = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_callbacks2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_callbacks2_t *","handle_error", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_callbacks2_t *)(argp1);
  result = (svn_error_t *(*)(char const *,svn_error_t *,void *,apr_pool_t *)) ((arg1)->handle_error);
  vresult = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_q_const__char_p_svn_error_t_p_void_p_apr_pool_t__p_svn_error_t);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
_wrap_svn_wc_entry_callbacks2_t_allocate(VALUE self)
#else
_wrap_svn_wc_entry_callbacks2_t_allocate(int argc, VALUE *argv, VALUE self)
#endif
{
  VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_svn_wc_entry_callbacks2_t);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
  rb_obj_call_init(vresult, argc, argv);
#endif
  return vresult;
}


SWIGINTERN VALUE
_wrap_new_svn_wc_entry_callbacks2_t(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_callbacks2_t *result = 0 ;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  {
    result = (struct svn_wc_entry_callbacks2_t *)calloc(1, sizeof(struct svn_wc_entry_callbacks2_t));
    DATA_PTR(self) = result;
    
    
    
  }
  return self;
fail:
  return Qnil;
}


SWIGINTERN void
free_svn_wc_entry_callbacks2_t(void *self) {
    struct svn_wc_entry_callbacks2_t *arg1 = (struct svn_wc_entry_callbacks2_t *)self;
    free((char *) arg1);
}

static swig_class SwigClassSvn_wc_entry_callbacks_t;

SWIGINTERN VALUE
_wrap_svn_wc_entry_callbacks_t_found_entry_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_callbacks_t *arg1 = 0 ;
  svn_error_t *(*arg2)(char const *,svn_wc_entry_t const *,void *,apr_pool_t *) = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_callbacks_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_callbacks_t *","found_entry", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_callbacks_t *)(argp1);
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg2), SWIGTYPE_p_f_p_q_const__char_p_q_const__svn_wc_entry_t_p_void_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_error_t *(*)(char const *,svn_wc_entry_t const *,void *,apr_pool_t *)","found_entry", 2, argv[0] )); 
    }
  }
  if (arg1) (arg1)->found_entry = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_callbacks_t_found_entry_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_callbacks_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_error_t *(*result)(char const *,svn_wc_entry_t const *,void *,apr_pool_t *) = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_entry_callbacks_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_entry_callbacks_t *","found_entry", 1, self )); 
  }
  arg1 = (struct svn_wc_entry_callbacks_t *)(argp1);
  result = (svn_error_t *(*)(char const *,svn_wc_entry_t const *,void *,apr_pool_t *)) ((arg1)->found_entry);
  vresult = SWIG_NewFunctionPtrObj((void *)(result), SWIGTYPE_p_f_p_q_const__char_p_q_const__svn_wc_entry_t_p_void_p_apr_pool_t__p_svn_error_t);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
_wrap_svn_wc_entry_callbacks_t_allocate(VALUE self)
#else
_wrap_svn_wc_entry_callbacks_t_allocate(int argc, VALUE *argv, VALUE self)
#endif
{
  VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_svn_wc_entry_callbacks_t);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
  rb_obj_call_init(vresult, argc, argv);
#endif
  return vresult;
}


SWIGINTERN VALUE
_wrap_new_svn_wc_entry_callbacks_t(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_entry_callbacks_t *result = 0 ;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  {
    result = (struct svn_wc_entry_callbacks_t *)calloc(1, sizeof(struct svn_wc_entry_callbacks_t));
    DATA_PTR(self) = result;
    
    
    
  }
  return self;
fail:
  return Qnil;
}


SWIGINTERN void
free_svn_wc_entry_callbacks_t(void *self) {
    struct svn_wc_entry_callbacks_t *arg1 = (struct svn_wc_entry_callbacks_t *)self;
    free((char *) arg1);
}

SWIGINTERN VALUE
_wrap_svn_wc_walk_entries3(int argc, VALUE *argv, VALUE self) {
  char *arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  svn_wc_entry_callbacks2_t *arg3 = 0 ;
  void *arg4 = 0 ;
  svn_depth_t arg5 ;
  svn_boolean_t arg6 ;
  svn_cancel_func_t arg7 = 0 ;
  void *arg8 = 0 ;
  apr_pool_t *arg9 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg9);
    _global_pool = arg9;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 6) || (argc > 7)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 6)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_wc_walk_entries3", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_walk_entries3", 2, argv[1] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  {
    arg3 = (svn_wc_entry_callbacks2_t *) svn_swig_rb_wc_entry_callbacks2();
    arg4 = (void *)svn_swig_rb_make_baton(argv[2], _global_svn_swig_rb_pool);
  }
  {
    arg5 = svn_swig_rb_to_depth(argv[3]);
  }
  arg6 = RTEST(argv[4]);
  {
    arg7 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg8 = (void *)svn_swig_rb_make_baton(argv[5], _global_svn_swig_rb_pool);
  }
  if (argc > 6) {
    
  }
  {
    result = (svn_error_t *)svn_wc_walk_entries3((char const *)arg1,arg2,(struct svn_wc_entry_callbacks2_t const *)arg3,arg4,arg5,arg6,arg7,arg8,arg9);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg8);
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_walk_entries2(int argc, VALUE *argv, VALUE self) {
  char *arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  svn_wc_entry_callbacks_t *arg3 = 0 ;
  void *arg4 = 0 ;
  svn_boolean_t arg5 ;
  svn_cancel_func_t arg6 = 0 ;
  void *arg7 = 0 ;
  apr_pool_t *arg8 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  int res4 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg8);
    _global_pool = arg8;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 6) || (argc > 7)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 6)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_wc_walk_entries2", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_walk_entries2", 2, argv[1] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  res3 = SWIG_ConvertPtr(argv[2], &argp3,SWIGTYPE_p_svn_wc_entry_callbacks_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_wc_entry_callbacks_t const *","svn_wc_walk_entries2", 3, argv[2] )); 
  }
  arg3 = (svn_wc_entry_callbacks_t *)(argp3);
  res4 = SWIG_ConvertPtr(argv[3],SWIG_as_voidptrptr(&arg4), 0, 0);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "void *","svn_wc_walk_entries2", 4, argv[3] )); 
  }
  arg5 = RTEST(argv[4]);
  {
    arg6 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg7 = (void *)svn_swig_rb_make_baton(argv[5], _global_svn_swig_rb_pool);
  }
  if (argc > 6) {
    
  }
  {
    result = (svn_error_t *)svn_wc_walk_entries2((char const *)arg1,arg2,(struct svn_wc_entry_callbacks_t const *)arg3,arg4,arg5,arg6,arg7,arg8);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg7);
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_walk_entries(int argc, VALUE *argv, VALUE self) {
  char *arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  svn_wc_entry_callbacks_t *arg3 = 0 ;
  void *arg4 = 0 ;
  svn_boolean_t arg5 ;
  apr_pool_t *arg6 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  int res4 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 5) || (argc > 6)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 5)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_wc_walk_entries", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_walk_entries", 2, argv[1] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  res3 = SWIG_ConvertPtr(argv[2], &argp3,SWIGTYPE_p_svn_wc_entry_callbacks_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_wc_entry_callbacks_t const *","svn_wc_walk_entries", 3, argv[2] )); 
  }
  arg3 = (svn_wc_entry_callbacks_t *)(argp3);
  res4 = SWIG_ConvertPtr(argv[3],SWIG_as_voidptrptr(&arg4), 0, 0);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "void *","svn_wc_walk_entries", 4, argv[3] )); 
  }
  arg5 = RTEST(argv[4]);
  if (argc > 5) {
    
  }
  {
    result = (svn_error_t *)svn_wc_walk_entries((char const *)arg1,arg2,(struct svn_wc_entry_callbacks_t const *)arg3,arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_mark_missing_deleted(int argc, VALUE *argv, VALUE self) {
  char *arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  apr_pool_t *arg3 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_wc_mark_missing_deleted", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_mark_missing_deleted", 2, argv[1] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_wc_mark_missing_deleted((char const *)arg1,arg2,arg3);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_ensure_adm4(int argc, VALUE *argv, VALUE self) {
  svn_wc_context_t *arg1 = 0 ;
  char *arg2 = 0 ;
  char *arg3 = 0 ;
  char *arg4 = 0 ;
  char *arg5 = 0 ;
  svn_revnum_t arg6 ;
  svn_depth_t arg7 ;
  apr_pool_t *arg8 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  int res5 ;
  char *buf5 = 0 ;
  int alloc5 = 0 ;
  long val6 ;
  int ecode6 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg8);
    _global_pool = arg8;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 7) || (argc > 8)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 7)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_context_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_context_t *","svn_wc_ensure_adm4", 1, argv[0] )); 
  }
  arg1 = (svn_wc_context_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_wc_ensure_adm4", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_wc_ensure_adm4", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_AsCharPtrAndSize(argv[3], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_wc_ensure_adm4", 4, argv[3] ));
  }
  arg4 = (char *)(buf4);
  res5 = SWIG_AsCharPtrAndSize(argv[4], &buf5, NULL, &alloc5);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "char const *","svn_wc_ensure_adm4", 5, argv[4] ));
  }
  arg5 = (char *)(buf5);
  ecode6 = SWIG_AsVal_long(argv[5], &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), Ruby_Format_TypeError( "", "svn_revnum_t","svn_wc_ensure_adm4", 6, argv[5] ));
  } 
  arg6 = (svn_revnum_t)(val6);
  {
    arg7 = svn_swig_rb_to_depth(argv[6]);
  }
  if (argc > 7) {
    
  }
  {
    result = (svn_error_t *)svn_wc_ensure_adm4(arg1,(char const *)arg2,(char const *)arg3,(char const *)arg4,(char const *)arg5,arg6,arg7,arg8);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_ensure_adm3(int argc, VALUE *argv, VALUE self) {
  char *arg1 = 0 ;
  char *arg2 = 0 ;
  char *arg3 = 0 ;
  char *arg4 = 0 ;
  svn_revnum_t arg5 ;
  svn_depth_t arg6 ;
  apr_pool_t *arg7 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  long val5 ;
  int ecode5 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg7);
    _global_pool = arg7;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 6) || (argc > 7)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 6)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_wc_ensure_adm3", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    if (NIL_P(argv[1])) {
      arg2 = NULL;
    } else {
      arg2 = StringValuePtr(argv[1]);
    }
  }
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_wc_ensure_adm3", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  {
    if (NIL_P(argv[3])) {
      arg4 = NULL;
    } else {
      arg4 = StringValuePtr(argv[3]);
    }
  }
  ecode5 = SWIG_AsVal_long(argv[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), Ruby_Format_TypeError( "", "svn_revnum_t","svn_wc_ensure_adm3", 5, argv[4] ));
  } 
  arg5 = (svn_revnum_t)(val5);
  {
    arg6 = svn_swig_rb_to_depth(argv[5]);
  }
  if (argc > 6) {
    
  }
  {
    result = (svn_error_t *)svn_wc_ensure_adm3((char const *)arg1,(char const *)arg2,(char const *)arg3,(char const *)arg4,arg5,arg6,arg7);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_ensure_adm2(int argc, VALUE *argv, VALUE self) {
  char *arg1 = 0 ;
  char *arg2 = 0 ;
  char *arg3 = 0 ;
  char *arg4 = 0 ;
  svn_revnum_t arg5 ;
  apr_pool_t *arg6 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  long val5 ;
  int ecode5 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 5) || (argc > 6)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 5)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_wc_ensure_adm2", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    if (NIL_P(argv[1])) {
      arg2 = NULL;
    } else {
      arg2 = StringValuePtr(argv[1]);
    }
  }
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_wc_ensure_adm2", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  {
    if (NIL_P(argv[3])) {
      arg4 = NULL;
    } else {
      arg4 = StringValuePtr(argv[3]);
    }
  }
  ecode5 = SWIG_AsVal_long(argv[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), Ruby_Format_TypeError( "", "svn_revnum_t","svn_wc_ensure_adm2", 5, argv[4] ));
  } 
  arg5 = (svn_revnum_t)(val5);
  if (argc > 5) {
    
  }
  {
    result = (svn_error_t *)svn_wc_ensure_adm2((char const *)arg1,(char const *)arg2,(char const *)arg3,(char const *)arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_ensure_adm(int argc, VALUE *argv, VALUE self) {
  char *arg1 = 0 ;
  char *arg2 = 0 ;
  char *arg3 = 0 ;
  svn_revnum_t arg4 ;
  apr_pool_t *arg5 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  long val4 ;
  int ecode4 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 4) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_wc_ensure_adm", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    if (NIL_P(argv[1])) {
      arg2 = NULL;
    } else {
      arg2 = StringValuePtr(argv[1]);
    }
  }
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_wc_ensure_adm", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  ecode4 = SWIG_AsVal_long(argv[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), Ruby_Format_TypeError( "", "svn_revnum_t","svn_wc_ensure_adm", 4, argv[3] ));
  } 
  arg4 = (svn_revnum_t)(val4);
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_wc_ensure_adm((char const *)arg1,(char const *)arg2,(char const *)arg3,arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_maybe_set_repos_root(int argc, VALUE *argv, VALUE self) {
  svn_wc_adm_access_t *arg1 = 0 ;
  char *arg2 = 0 ;
  char *arg3 = 0 ;
  apr_pool_t *arg4 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 3) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_maybe_set_repos_root", 1, argv[0] )); 
  }
  arg1 = (svn_wc_adm_access_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_wc_maybe_set_repos_root", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  {
    if (NIL_P(argv[2])) {
      arg3 = NULL;
    } else {
      arg3 = StringValuePtr(argv[2]);
    }
  }
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_wc_maybe_set_repos_root(arg1,(char const *)arg2,(char const *)arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


static swig_class SwigClassSvn_wc_status3_t;

SWIGINTERN VALUE
_wrap_svn_wc_status3_t_kind_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status3_t *arg1 = 0 ;
  svn_node_kind_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status3_t *","kind", 1, self )); 
  }
  arg1 = (struct svn_wc_status3_t *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_node_kind_t","kind", 2, argv[0] ));
  } 
  arg2 = (svn_node_kind_t)(val2);
  if (arg1) (arg1)->kind = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status3_t_kind_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status3_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_node_kind_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status3_t *","kind", 1, self )); 
  }
  arg1 = (struct svn_wc_status3_t *)(argp1);
  result = (svn_node_kind_t) ((arg1)->kind);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status3_t_depth_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status3_t *arg1 = 0 ;
  svn_depth_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status3_t *","depth", 1, self )); 
  }
  arg1 = (struct svn_wc_status3_t *)(argp1);
  {
    arg2 = svn_swig_rb_to_depth(argv[0]);
  }
  if (arg1) (arg1)->depth = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status3_t_depth_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status3_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_depth_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status3_t *","depth", 1, self )); 
  }
  arg1 = (struct svn_wc_status3_t *)(argp1);
  result = (svn_depth_t) ((arg1)->depth);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status3_t_filesize_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status3_t *arg1 = 0 ;
  svn_filesize_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status3_t *","filesize", 1, self )); 
  }
  arg1 = (struct svn_wc_status3_t *)(argp1);
  {
    arg2 = (svn_filesize_t)NUM2LL(argv[0]);
  }
  if (arg1) (arg1)->filesize = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status3_t_filesize_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status3_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_filesize_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status3_t *","filesize", 1, self )); 
  }
  arg1 = (struct svn_wc_status3_t *)(argp1);
  result =  ((arg1)->filesize);
  {
    vresult = LL2NUM((svn_filesize_t)(result));
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status3_t_versioned_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status3_t *arg1 = 0 ;
  svn_boolean_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status3_t *","versioned", 1, self )); 
  }
  arg1 = (struct svn_wc_status3_t *)(argp1);
  arg2 = RTEST(argv[0]);
  if (arg1) (arg1)->versioned = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status3_t_versioned_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status3_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status3_t *","versioned", 1, self )); 
  }
  arg1 = (struct svn_wc_status3_t *)(argp1);
  result = (svn_boolean_t) ((arg1)->versioned);
  vresult = result ? Qtrue : Qfalse;
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status3_t_conflicted_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status3_t *arg1 = 0 ;
  svn_boolean_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status3_t *","conflicted", 1, self )); 
  }
  arg1 = (struct svn_wc_status3_t *)(argp1);
  arg2 = RTEST(argv[0]);
  if (arg1) (arg1)->conflicted = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status3_t_conflicted_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status3_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status3_t *","conflicted", 1, self )); 
  }
  arg1 = (struct svn_wc_status3_t *)(argp1);
  result = (svn_boolean_t) ((arg1)->conflicted);
  vresult = result ? Qtrue : Qfalse;
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status3_t_node_status_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status3_t *arg1 = 0 ;
  enum svn_wc_status_kind arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status3_t *","node_status", 1, self )); 
  }
  arg1 = (struct svn_wc_status3_t *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "enum svn_wc_status_kind","node_status", 2, argv[0] ));
  } 
  arg2 = (enum svn_wc_status_kind)(val2);
  if (arg1) (arg1)->node_status = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status3_t_node_status_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status3_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  enum svn_wc_status_kind result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status3_t *","node_status", 1, self )); 
  }
  arg1 = (struct svn_wc_status3_t *)(argp1);
  result = (enum svn_wc_status_kind) ((arg1)->node_status);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status3_t_text_status_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status3_t *arg1 = 0 ;
  enum svn_wc_status_kind arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status3_t *","text_status", 1, self )); 
  }
  arg1 = (struct svn_wc_status3_t *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "enum svn_wc_status_kind","text_status", 2, argv[0] ));
  } 
  arg2 = (enum svn_wc_status_kind)(val2);
  if (arg1) (arg1)->text_status = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status3_t_text_status_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status3_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  enum svn_wc_status_kind result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status3_t *","text_status", 1, self )); 
  }
  arg1 = (struct svn_wc_status3_t *)(argp1);
  result = (enum svn_wc_status_kind) ((arg1)->text_status);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status3_t_prop_status_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status3_t *arg1 = 0 ;
  enum svn_wc_status_kind arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status3_t *","prop_status", 1, self )); 
  }
  arg1 = (struct svn_wc_status3_t *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "enum svn_wc_status_kind","prop_status", 2, argv[0] ));
  } 
  arg2 = (enum svn_wc_status_kind)(val2);
  if (arg1) (arg1)->prop_status = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status3_t_prop_status_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status3_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  enum svn_wc_status_kind result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status3_t *","prop_status", 1, self )); 
  }
  arg1 = (struct svn_wc_status3_t *)(argp1);
  result = (enum svn_wc_status_kind) ((arg1)->prop_status);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status3_t_copied_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status3_t *arg1 = 0 ;
  svn_boolean_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status3_t *","copied", 1, self )); 
  }
  arg1 = (struct svn_wc_status3_t *)(argp1);
  arg2 = RTEST(argv[0]);
  if (arg1) (arg1)->copied = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status3_t_copied_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status3_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status3_t *","copied", 1, self )); 
  }
  arg1 = (struct svn_wc_status3_t *)(argp1);
  result = (svn_boolean_t) ((arg1)->copied);
  vresult = result ? Qtrue : Qfalse;
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status3_t_revision_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status3_t *arg1 = 0 ;
  svn_revnum_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status3_t *","revision", 1, self )); 
  }
  arg1 = (struct svn_wc_status3_t *)(argp1);
  ecode2 = SWIG_AsVal_long(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_revnum_t","revision", 2, argv[0] ));
  } 
  arg2 = (svn_revnum_t)(val2);
  if (arg1) (arg1)->revision = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status3_t_revision_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status3_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_revnum_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status3_t *","revision", 1, self )); 
  }
  arg1 = (struct svn_wc_status3_t *)(argp1);
  result = (svn_revnum_t) ((arg1)->revision);
  vresult = SWIG_From_long((long)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status3_t_changed_rev_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status3_t *arg1 = 0 ;
  svn_revnum_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status3_t *","changed_rev", 1, self )); 
  }
  arg1 = (struct svn_wc_status3_t *)(argp1);
  ecode2 = SWIG_AsVal_long(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_revnum_t","changed_rev", 2, argv[0] ));
  } 
  arg2 = (svn_revnum_t)(val2);
  if (arg1) (arg1)->changed_rev = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status3_t_changed_rev_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status3_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_revnum_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status3_t *","changed_rev", 1, self )); 
  }
  arg1 = (struct svn_wc_status3_t *)(argp1);
  result = (svn_revnum_t) ((arg1)->changed_rev);
  vresult = SWIG_From_long((long)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status3_t_changed_date_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status3_t *arg1 = 0 ;
  apr_time_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status3_t *","changed_date", 1, self )); 
  }
  arg1 = (struct svn_wc_status3_t *)(argp1);
  {
    arg2 = (apr_time_t)NUM2LL(argv[0]);
  }
  if (arg1) (arg1)->changed_date = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status3_t_changed_date_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status3_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  apr_time_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status3_t *","changed_date", 1, self )); 
  }
  arg1 = (struct svn_wc_status3_t *)(argp1);
  result =  ((arg1)->changed_date);
  {
    vresult = LL2NUM((apr_time_t)(result));
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status3_t_changed_author_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status3_t *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status3_t *","changed_author", 1, self )); 
  }
  arg1 = (struct svn_wc_status3_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","changed_author", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->changed_author) free((char *)arg1->changed_author);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->changed_author = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status3_t_changed_author_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status3_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status3_t *","changed_author", 1, self )); 
  }
  arg1 = (struct svn_wc_status3_t *)(argp1);
  result = (char *) ((arg1)->changed_author);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status3_t_repos_root_url_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status3_t *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status3_t *","repos_root_url", 1, self )); 
  }
  arg1 = (struct svn_wc_status3_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","repos_root_url", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->repos_root_url) free((char *)arg1->repos_root_url);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->repos_root_url = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status3_t_repos_root_url_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status3_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status3_t *","repos_root_url", 1, self )); 
  }
  arg1 = (struct svn_wc_status3_t *)(argp1);
  result = (char *) ((arg1)->repos_root_url);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status3_t_repos_uuid_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status3_t *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status3_t *","repos_uuid", 1, self )); 
  }
  arg1 = (struct svn_wc_status3_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","repos_uuid", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->repos_uuid) free((char *)arg1->repos_uuid);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->repos_uuid = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status3_t_repos_uuid_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status3_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status3_t *","repos_uuid", 1, self )); 
  }
  arg1 = (struct svn_wc_status3_t *)(argp1);
  result = (char *) ((arg1)->repos_uuid);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status3_t_repos_relpath_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status3_t *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status3_t *","repos_relpath", 1, self )); 
  }
  arg1 = (struct svn_wc_status3_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","repos_relpath", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->repos_relpath) free((char *)arg1->repos_relpath);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->repos_relpath = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status3_t_repos_relpath_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status3_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status3_t *","repos_relpath", 1, self )); 
  }
  arg1 = (struct svn_wc_status3_t *)(argp1);
  result = (char *) ((arg1)->repos_relpath);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status3_t_switched_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status3_t *arg1 = 0 ;
  svn_boolean_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status3_t *","switched", 1, self )); 
  }
  arg1 = (struct svn_wc_status3_t *)(argp1);
  arg2 = RTEST(argv[0]);
  if (arg1) (arg1)->switched = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status3_t_switched_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status3_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status3_t *","switched", 1, self )); 
  }
  arg1 = (struct svn_wc_status3_t *)(argp1);
  result = (svn_boolean_t) ((arg1)->switched);
  vresult = result ? Qtrue : Qfalse;
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status3_t_locked_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status3_t *arg1 = 0 ;
  svn_boolean_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status3_t *","locked", 1, self )); 
  }
  arg1 = (struct svn_wc_status3_t *)(argp1);
  arg2 = RTEST(argv[0]);
  if (arg1) (arg1)->locked = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status3_t_locked_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status3_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status3_t *","locked", 1, self )); 
  }
  arg1 = (struct svn_wc_status3_t *)(argp1);
  result = (svn_boolean_t) ((arg1)->locked);
  vresult = result ? Qtrue : Qfalse;
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status3_t_lock_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status3_t *arg1 = 0 ;
  svn_lock_t *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status3_t *","lock", 1, self )); 
  }
  arg1 = (struct svn_wc_status3_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_svn_lock_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_lock_t const *","lock", 2, argv[0] )); 
  }
  arg2 = (svn_lock_t *)(argp2);
  if (arg1) (arg1)->lock = (svn_lock_t const *)arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status3_t_lock_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status3_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_lock_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status3_t *","lock", 1, self )); 
  }
  arg1 = (struct svn_wc_status3_t *)(argp1);
  result = (svn_lock_t *) ((arg1)->lock);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_lock_t, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status3_t_changelist_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status3_t *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status3_t *","changelist", 1, self )); 
  }
  arg1 = (struct svn_wc_status3_t *)(argp1);
  {
    arg2 = StringValueCStr(argv[0]);
  }
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->changelist) free((char *)arg1->changelist);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->changelist = copied;
  }
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status3_t_changelist_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status3_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status3_t *","changelist", 1, self )); 
  }
  arg1 = (struct svn_wc_status3_t *)(argp1);
  result = (char *) ((arg1)->changelist);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status3_t_ood_kind_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status3_t *arg1 = 0 ;
  svn_node_kind_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status3_t *","ood_kind", 1, self )); 
  }
  arg1 = (struct svn_wc_status3_t *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_node_kind_t","ood_kind", 2, argv[0] ));
  } 
  arg2 = (svn_node_kind_t)(val2);
  if (arg1) (arg1)->ood_kind = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status3_t_ood_kind_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status3_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_node_kind_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status3_t *","ood_kind", 1, self )); 
  }
  arg1 = (struct svn_wc_status3_t *)(argp1);
  result = (svn_node_kind_t) ((arg1)->ood_kind);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status3_t_repos_node_status_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status3_t *arg1 = 0 ;
  enum svn_wc_status_kind arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status3_t *","repos_node_status", 1, self )); 
  }
  arg1 = (struct svn_wc_status3_t *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "enum svn_wc_status_kind","repos_node_status", 2, argv[0] ));
  } 
  arg2 = (enum svn_wc_status_kind)(val2);
  if (arg1) (arg1)->repos_node_status = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status3_t_repos_node_status_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status3_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  enum svn_wc_status_kind result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status3_t *","repos_node_status", 1, self )); 
  }
  arg1 = (struct svn_wc_status3_t *)(argp1);
  result = (enum svn_wc_status_kind) ((arg1)->repos_node_status);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status3_t_repos_text_status_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status3_t *arg1 = 0 ;
  enum svn_wc_status_kind arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status3_t *","repos_text_status", 1, self )); 
  }
  arg1 = (struct svn_wc_status3_t *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "enum svn_wc_status_kind","repos_text_status", 2, argv[0] ));
  } 
  arg2 = (enum svn_wc_status_kind)(val2);
  if (arg1) (arg1)->repos_text_status = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status3_t_repos_text_status_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status3_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  enum svn_wc_status_kind result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status3_t *","repos_text_status", 1, self )); 
  }
  arg1 = (struct svn_wc_status3_t *)(argp1);
  result = (enum svn_wc_status_kind) ((arg1)->repos_text_status);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status3_t_repos_prop_status_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status3_t *arg1 = 0 ;
  enum svn_wc_status_kind arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status3_t *","repos_prop_status", 1, self )); 
  }
  arg1 = (struct svn_wc_status3_t *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "enum svn_wc_status_kind","repos_prop_status", 2, argv[0] ));
  } 
  arg2 = (enum svn_wc_status_kind)(val2);
  if (arg1) (arg1)->repos_prop_status = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status3_t_repos_prop_status_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status3_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  enum svn_wc_status_kind result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status3_t *","repos_prop_status", 1, self )); 
  }
  arg1 = (struct svn_wc_status3_t *)(argp1);
  result = (enum svn_wc_status_kind) ((arg1)->repos_prop_status);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status3_t_repos_lock_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status3_t *arg1 = 0 ;
  svn_lock_t *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status3_t *","repos_lock", 1, self )); 
  }
  arg1 = (struct svn_wc_status3_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_svn_lock_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_lock_t const *","repos_lock", 2, argv[0] )); 
  }
  arg2 = (svn_lock_t *)(argp2);
  if (arg1) (arg1)->repos_lock = (svn_lock_t const *)arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status3_t_repos_lock_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status3_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_lock_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status3_t *","repos_lock", 1, self )); 
  }
  arg1 = (struct svn_wc_status3_t *)(argp1);
  result = (svn_lock_t *) ((arg1)->repos_lock);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_lock_t, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status3_t_ood_changed_rev_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status3_t *arg1 = 0 ;
  svn_revnum_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status3_t *","ood_changed_rev", 1, self )); 
  }
  arg1 = (struct svn_wc_status3_t *)(argp1);
  ecode2 = SWIG_AsVal_long(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_revnum_t","ood_changed_rev", 2, argv[0] ));
  } 
  arg2 = (svn_revnum_t)(val2);
  if (arg1) (arg1)->ood_changed_rev = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status3_t_ood_changed_rev_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status3_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_revnum_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status3_t *","ood_changed_rev", 1, self )); 
  }
  arg1 = (struct svn_wc_status3_t *)(argp1);
  result = (svn_revnum_t) ((arg1)->ood_changed_rev);
  vresult = SWIG_From_long((long)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status3_t_ood_changed_date_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status3_t *arg1 = 0 ;
  apr_time_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status3_t *","ood_changed_date", 1, self )); 
  }
  arg1 = (struct svn_wc_status3_t *)(argp1);
  {
    arg2 = (apr_time_t)NUM2LL(argv[0]);
  }
  if (arg1) (arg1)->ood_changed_date = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status3_t_ood_changed_date_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status3_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  apr_time_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status3_t *","ood_changed_date", 1, self )); 
  }
  arg1 = (struct svn_wc_status3_t *)(argp1);
  result =  ((arg1)->ood_changed_date);
  {
    vresult = LL2NUM((apr_time_t)(result));
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status3_t_ood_changed_author_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status3_t *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status3_t *","ood_changed_author", 1, self )); 
  }
  arg1 = (struct svn_wc_status3_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","ood_changed_author", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->ood_changed_author) free((char *)arg1->ood_changed_author);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->ood_changed_author = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status3_t_ood_changed_author_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status3_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status3_t *","ood_changed_author", 1, self )); 
  }
  arg1 = (struct svn_wc_status3_t *)(argp1);
  result = (char *) ((arg1)->ood_changed_author);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status3_t_moved_from_abspath_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status3_t *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status3_t *","moved_from_abspath", 1, self )); 
  }
  arg1 = (struct svn_wc_status3_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","moved_from_abspath", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->moved_from_abspath) free((char *)arg1->moved_from_abspath);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->moved_from_abspath = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status3_t_moved_from_abspath_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status3_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status3_t *","moved_from_abspath", 1, self )); 
  }
  arg1 = (struct svn_wc_status3_t *)(argp1);
  result = (char *) ((arg1)->moved_from_abspath);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status3_t_moved_to_abspath_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status3_t *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status3_t *","moved_to_abspath", 1, self )); 
  }
  arg1 = (struct svn_wc_status3_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","moved_to_abspath", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->moved_to_abspath) free((char *)arg1->moved_to_abspath);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->moved_to_abspath = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status3_t_moved_to_abspath_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status3_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status3_t *","moved_to_abspath", 1, self )); 
  }
  arg1 = (struct svn_wc_status3_t *)(argp1);
  result = (char *) ((arg1)->moved_to_abspath);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status3_t_file_external_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status3_t *arg1 = 0 ;
  svn_boolean_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status3_t *","file_external", 1, self )); 
  }
  arg1 = (struct svn_wc_status3_t *)(argp1);
  arg2 = RTEST(argv[0]);
  if (arg1) (arg1)->file_external = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status3_t_file_external_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status3_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status3_t *","file_external", 1, self )); 
  }
  arg1 = (struct svn_wc_status3_t *)(argp1);
  result = (svn_boolean_t) ((arg1)->file_external);
  vresult = result ? Qtrue : Qfalse;
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status3_t_actual_kind_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status3_t *arg1 = 0 ;
  svn_node_kind_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status3_t *","actual_kind", 1, self )); 
  }
  arg1 = (struct svn_wc_status3_t *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_node_kind_t","actual_kind", 2, argv[0] ));
  } 
  arg2 = (svn_node_kind_t)(val2);
  if (arg1) (arg1)->actual_kind = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status3_t_actual_kind_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status3_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_node_kind_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status3_t *","actual_kind", 1, self )); 
  }
  arg1 = (struct svn_wc_status3_t *)(argp1);
  result = (svn_node_kind_t) ((arg1)->actual_kind);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
_wrap_svn_wc_status3_t_allocate(VALUE self)
#else
_wrap_svn_wc_status3_t_allocate(int argc, VALUE *argv, VALUE self)
#endif
{
  VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_svn_wc_status3_t);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
  rb_obj_call_init(vresult, argc, argv);
#endif
  return vresult;
}


SWIGINTERN VALUE
_wrap_new_svn_wc_status3_t(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status3_t *result = 0 ;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  {
    result = (struct svn_wc_status3_t *)calloc(1, sizeof(struct svn_wc_status3_t));
    DATA_PTR(self) = result;
    
    
    
  }
  return self;
fail:
  return Qnil;
}


SWIGINTERN void
free_svn_wc_status3_t(void *self) {
    struct svn_wc_status3_t *arg1 = (struct svn_wc_status3_t *)self;
    free((char *) arg1);
}

static swig_class SwigClassSvn_wc_status2_t;

SWIGINTERN VALUE
_wrap_svn_wc_status2_t_entry_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status2_t *arg1 = 0 ;
  svn_wc_entry_t *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status2_t *","entry", 1, self )); 
  }
  arg1 = (struct svn_wc_status2_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_svn_wc_entry_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_entry_t const *","entry", 2, argv[0] )); 
  }
  arg2 = (svn_wc_entry_t *)(argp2);
  if (arg1) (arg1)->entry = (svn_wc_entry_t const *)arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status2_t_entry_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status2_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_wc_entry_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status2_t *","entry", 1, self )); 
  }
  arg1 = (struct svn_wc_status2_t *)(argp1);
  result = (svn_wc_entry_t *) ((arg1)->entry);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status2_t_text_status_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status2_t *arg1 = 0 ;
  enum svn_wc_status_kind arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status2_t *","text_status", 1, self )); 
  }
  arg1 = (struct svn_wc_status2_t *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "enum svn_wc_status_kind","text_status", 2, argv[0] ));
  } 
  arg2 = (enum svn_wc_status_kind)(val2);
  if (arg1) (arg1)->text_status = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status2_t_text_status_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status2_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  enum svn_wc_status_kind result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status2_t *","text_status", 1, self )); 
  }
  arg1 = (struct svn_wc_status2_t *)(argp1);
  result = (enum svn_wc_status_kind) ((arg1)->text_status);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status2_t_prop_status_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status2_t *arg1 = 0 ;
  enum svn_wc_status_kind arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status2_t *","prop_status", 1, self )); 
  }
  arg1 = (struct svn_wc_status2_t *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "enum svn_wc_status_kind","prop_status", 2, argv[0] ));
  } 
  arg2 = (enum svn_wc_status_kind)(val2);
  if (arg1) (arg1)->prop_status = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status2_t_prop_status_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status2_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  enum svn_wc_status_kind result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status2_t *","prop_status", 1, self )); 
  }
  arg1 = (struct svn_wc_status2_t *)(argp1);
  result = (enum svn_wc_status_kind) ((arg1)->prop_status);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status2_t_locked_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status2_t *arg1 = 0 ;
  svn_boolean_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status2_t *","locked", 1, self )); 
  }
  arg1 = (struct svn_wc_status2_t *)(argp1);
  arg2 = RTEST(argv[0]);
  if (arg1) (arg1)->locked = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status2_t_locked_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status2_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status2_t *","locked", 1, self )); 
  }
  arg1 = (struct svn_wc_status2_t *)(argp1);
  result = (svn_boolean_t) ((arg1)->locked);
  vresult = result ? Qtrue : Qfalse;
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status2_t_copied_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status2_t *arg1 = 0 ;
  svn_boolean_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status2_t *","copied", 1, self )); 
  }
  arg1 = (struct svn_wc_status2_t *)(argp1);
  arg2 = RTEST(argv[0]);
  if (arg1) (arg1)->copied = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status2_t_copied_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status2_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status2_t *","copied", 1, self )); 
  }
  arg1 = (struct svn_wc_status2_t *)(argp1);
  result = (svn_boolean_t) ((arg1)->copied);
  vresult = result ? Qtrue : Qfalse;
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status2_t_switched_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status2_t *arg1 = 0 ;
  svn_boolean_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status2_t *","switched", 1, self )); 
  }
  arg1 = (struct svn_wc_status2_t *)(argp1);
  arg2 = RTEST(argv[0]);
  if (arg1) (arg1)->switched = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status2_t_switched_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status2_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status2_t *","switched", 1, self )); 
  }
  arg1 = (struct svn_wc_status2_t *)(argp1);
  result = (svn_boolean_t) ((arg1)->switched);
  vresult = result ? Qtrue : Qfalse;
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status2_t_repos_text_status_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status2_t *arg1 = 0 ;
  enum svn_wc_status_kind arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status2_t *","repos_text_status", 1, self )); 
  }
  arg1 = (struct svn_wc_status2_t *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "enum svn_wc_status_kind","repos_text_status", 2, argv[0] ));
  } 
  arg2 = (enum svn_wc_status_kind)(val2);
  if (arg1) (arg1)->repos_text_status = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status2_t_repos_text_status_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status2_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  enum svn_wc_status_kind result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status2_t *","repos_text_status", 1, self )); 
  }
  arg1 = (struct svn_wc_status2_t *)(argp1);
  result = (enum svn_wc_status_kind) ((arg1)->repos_text_status);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status2_t_repos_prop_status_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status2_t *arg1 = 0 ;
  enum svn_wc_status_kind arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status2_t *","repos_prop_status", 1, self )); 
  }
  arg1 = (struct svn_wc_status2_t *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "enum svn_wc_status_kind","repos_prop_status", 2, argv[0] ));
  } 
  arg2 = (enum svn_wc_status_kind)(val2);
  if (arg1) (arg1)->repos_prop_status = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status2_t_repos_prop_status_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status2_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  enum svn_wc_status_kind result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status2_t *","repos_prop_status", 1, self )); 
  }
  arg1 = (struct svn_wc_status2_t *)(argp1);
  result = (enum svn_wc_status_kind) ((arg1)->repos_prop_status);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status2_t_repos_lock_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status2_t *arg1 = 0 ;
  svn_lock_t *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status2_t *","repos_lock", 1, self )); 
  }
  arg1 = (struct svn_wc_status2_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_svn_lock_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_lock_t *","repos_lock", 2, argv[0] )); 
  }
  arg2 = (svn_lock_t *)(argp2);
  if (arg1) (arg1)->repos_lock = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status2_t_repos_lock_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status2_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_lock_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status2_t *","repos_lock", 1, self )); 
  }
  arg1 = (struct svn_wc_status2_t *)(argp1);
  result = (svn_lock_t *) ((arg1)->repos_lock);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_lock_t, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status2_t_url_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status2_t *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status2_t *","url", 1, self )); 
  }
  arg1 = (struct svn_wc_status2_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","url", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->url) free((char *)arg1->url);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->url = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status2_t_url_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status2_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status2_t *","url", 1, self )); 
  }
  arg1 = (struct svn_wc_status2_t *)(argp1);
  result = (char *) ((arg1)->url);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status2_t_ood_last_cmt_rev_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status2_t *arg1 = 0 ;
  svn_revnum_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status2_t *","ood_last_cmt_rev", 1, self )); 
  }
  arg1 = (struct svn_wc_status2_t *)(argp1);
  ecode2 = SWIG_AsVal_long(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_revnum_t","ood_last_cmt_rev", 2, argv[0] ));
  } 
  arg2 = (svn_revnum_t)(val2);
  if (arg1) (arg1)->ood_last_cmt_rev = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status2_t_ood_last_cmt_rev_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status2_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_revnum_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status2_t *","ood_last_cmt_rev", 1, self )); 
  }
  arg1 = (struct svn_wc_status2_t *)(argp1);
  result = (svn_revnum_t) ((arg1)->ood_last_cmt_rev);
  vresult = SWIG_From_long((long)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status2_t_ood_last_cmt_date_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status2_t *arg1 = 0 ;
  apr_time_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status2_t *","ood_last_cmt_date", 1, self )); 
  }
  arg1 = (struct svn_wc_status2_t *)(argp1);
  {
    arg2 = (apr_time_t)NUM2LL(argv[0]);
  }
  if (arg1) (arg1)->ood_last_cmt_date = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status2_t_ood_last_cmt_date_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status2_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  apr_time_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status2_t *","ood_last_cmt_date", 1, self )); 
  }
  arg1 = (struct svn_wc_status2_t *)(argp1);
  result =  ((arg1)->ood_last_cmt_date);
  {
    vresult = LL2NUM((apr_time_t)(result));
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status2_t_ood_kind_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status2_t *arg1 = 0 ;
  svn_node_kind_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status2_t *","ood_kind", 1, self )); 
  }
  arg1 = (struct svn_wc_status2_t *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_node_kind_t","ood_kind", 2, argv[0] ));
  } 
  arg2 = (svn_node_kind_t)(val2);
  if (arg1) (arg1)->ood_kind = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status2_t_ood_kind_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status2_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_node_kind_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status2_t *","ood_kind", 1, self )); 
  }
  arg1 = (struct svn_wc_status2_t *)(argp1);
  result = (svn_node_kind_t) ((arg1)->ood_kind);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status2_t_ood_last_cmt_author_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status2_t *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status2_t *","ood_last_cmt_author", 1, self )); 
  }
  arg1 = (struct svn_wc_status2_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","ood_last_cmt_author", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    apr_size_t len = strlen(arg2) + 1;
    char *copied;
    if (arg1->ood_last_cmt_author) free((char *)arg1->ood_last_cmt_author);
    copied = malloc(len);
    memcpy(copied, arg2, len);
    arg1->ood_last_cmt_author = copied;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status2_t_ood_last_cmt_author_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status2_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status2_t *","ood_last_cmt_author", 1, self )); 
  }
  arg1 = (struct svn_wc_status2_t *)(argp1);
  result = (char *) ((arg1)->ood_last_cmt_author);
  {
    if (result) {
      vresult = rb_str_new2(result);
    } else {
      vresult = Qnil;
    }
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status2_t_tree_conflict_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status2_t *arg1 = 0 ;
  svn_wc_conflict_description_t *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status2_t *","tree_conflict", 1, self )); 
  }
  arg1 = (struct svn_wc_status2_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_svn_wc_conflict_description_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_conflict_description_t *","tree_conflict", 2, argv[0] )); 
  }
  arg2 = (svn_wc_conflict_description_t *)(argp2);
  if (arg1) (arg1)->tree_conflict = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status2_t_tree_conflict_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status2_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_wc_conflict_description_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status2_t *","tree_conflict", 1, self )); 
  }
  arg1 = (struct svn_wc_status2_t *)(argp1);
  result = (svn_wc_conflict_description_t *) ((arg1)->tree_conflict);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_wc_conflict_description_t, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status2_t_file_external_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status2_t *arg1 = 0 ;
  svn_boolean_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status2_t *","file_external", 1, self )); 
  }
  arg1 = (struct svn_wc_status2_t *)(argp1);
  arg2 = RTEST(argv[0]);
  if (arg1) (arg1)->file_external = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status2_t_file_external_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status2_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status2_t *","file_external", 1, self )); 
  }
  arg1 = (struct svn_wc_status2_t *)(argp1);
  result = (svn_boolean_t) ((arg1)->file_external);
  vresult = result ? Qtrue : Qfalse;
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status2_t_pristine_text_status_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status2_t *arg1 = 0 ;
  enum svn_wc_status_kind arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status2_t *","pristine_text_status", 1, self )); 
  }
  arg1 = (struct svn_wc_status2_t *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "enum svn_wc_status_kind","pristine_text_status", 2, argv[0] ));
  } 
  arg2 = (enum svn_wc_status_kind)(val2);
  if (arg1) (arg1)->pristine_text_status = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status2_t_pristine_text_status_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status2_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  enum svn_wc_status_kind result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status2_t *","pristine_text_status", 1, self )); 
  }
  arg1 = (struct svn_wc_status2_t *)(argp1);
  result = (enum svn_wc_status_kind) ((arg1)->pristine_text_status);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status2_t_pristine_prop_status_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status2_t *arg1 = 0 ;
  enum svn_wc_status_kind arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status2_t *","pristine_prop_status", 1, self )); 
  }
  arg1 = (struct svn_wc_status2_t *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "enum svn_wc_status_kind","pristine_prop_status", 2, argv[0] ));
  } 
  arg2 = (enum svn_wc_status_kind)(val2);
  if (arg1) (arg1)->pristine_prop_status = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status2_t_pristine_prop_status_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status2_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  enum svn_wc_status_kind result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status2_t *","pristine_prop_status", 1, self )); 
  }
  arg1 = (struct svn_wc_status2_t *)(argp1);
  result = (enum svn_wc_status_kind) ((arg1)->pristine_prop_status);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
_wrap_svn_wc_status2_t_allocate(VALUE self)
#else
_wrap_svn_wc_status2_t_allocate(int argc, VALUE *argv, VALUE self)
#endif
{
  VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_svn_wc_status2_t);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
  rb_obj_call_init(vresult, argc, argv);
#endif
  return vresult;
}


SWIGINTERN VALUE
_wrap_new_svn_wc_status2_t(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status2_t *result = 0 ;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  {
    result = (struct svn_wc_status2_t *)calloc(1, sizeof(struct svn_wc_status2_t));
    DATA_PTR(self) = result;
    
    
    
  }
  return self;
fail:
  return Qnil;
}


SWIGINTERN void
free_svn_wc_status2_t(void *self) {
    struct svn_wc_status2_t *arg1 = (struct svn_wc_status2_t *)self;
    free((char *) arg1);
}

static swig_class SwigClassSvn_wc_status_t;

SWIGINTERN VALUE
_wrap_svn_wc_status_t_entry_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status_t *arg1 = 0 ;
  svn_wc_entry_t *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status_t *","entry", 1, self )); 
  }
  arg1 = (struct svn_wc_status_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_svn_wc_entry_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_entry_t const *","entry", 2, argv[0] )); 
  }
  arg2 = (svn_wc_entry_t *)(argp2);
  if (arg1) (arg1)->entry = (svn_wc_entry_t const *)arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status_t_entry_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_wc_entry_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status_t *","entry", 1, self )); 
  }
  arg1 = (struct svn_wc_status_t *)(argp1);
  result = (svn_wc_entry_t *) ((arg1)->entry);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status_t_text_status_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status_t *arg1 = 0 ;
  enum svn_wc_status_kind arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status_t *","text_status", 1, self )); 
  }
  arg1 = (struct svn_wc_status_t *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "enum svn_wc_status_kind","text_status", 2, argv[0] ));
  } 
  arg2 = (enum svn_wc_status_kind)(val2);
  if (arg1) (arg1)->text_status = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status_t_text_status_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  enum svn_wc_status_kind result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status_t *","text_status", 1, self )); 
  }
  arg1 = (struct svn_wc_status_t *)(argp1);
  result = (enum svn_wc_status_kind) ((arg1)->text_status);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status_t_prop_status_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status_t *arg1 = 0 ;
  enum svn_wc_status_kind arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status_t *","prop_status", 1, self )); 
  }
  arg1 = (struct svn_wc_status_t *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "enum svn_wc_status_kind","prop_status", 2, argv[0] ));
  } 
  arg2 = (enum svn_wc_status_kind)(val2);
  if (arg1) (arg1)->prop_status = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status_t_prop_status_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  enum svn_wc_status_kind result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status_t *","prop_status", 1, self )); 
  }
  arg1 = (struct svn_wc_status_t *)(argp1);
  result = (enum svn_wc_status_kind) ((arg1)->prop_status);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status_t_locked_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status_t *arg1 = 0 ;
  svn_boolean_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status_t *","locked", 1, self )); 
  }
  arg1 = (struct svn_wc_status_t *)(argp1);
  arg2 = RTEST(argv[0]);
  if (arg1) (arg1)->locked = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status_t_locked_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status_t *","locked", 1, self )); 
  }
  arg1 = (struct svn_wc_status_t *)(argp1);
  result = (svn_boolean_t) ((arg1)->locked);
  vresult = result ? Qtrue : Qfalse;
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status_t_copied_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status_t *arg1 = 0 ;
  svn_boolean_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status_t *","copied", 1, self )); 
  }
  arg1 = (struct svn_wc_status_t *)(argp1);
  arg2 = RTEST(argv[0]);
  if (arg1) (arg1)->copied = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status_t_copied_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status_t *","copied", 1, self )); 
  }
  arg1 = (struct svn_wc_status_t *)(argp1);
  result = (svn_boolean_t) ((arg1)->copied);
  vresult = result ? Qtrue : Qfalse;
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status_t_switched_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status_t *arg1 = 0 ;
  svn_boolean_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status_t *","switched", 1, self )); 
  }
  arg1 = (struct svn_wc_status_t *)(argp1);
  arg2 = RTEST(argv[0]);
  if (arg1) (arg1)->switched = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status_t_switched_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status_t *","switched", 1, self )); 
  }
  arg1 = (struct svn_wc_status_t *)(argp1);
  result = (svn_boolean_t) ((arg1)->switched);
  vresult = result ? Qtrue : Qfalse;
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status_t_repos_text_status_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status_t *arg1 = 0 ;
  enum svn_wc_status_kind arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status_t *","repos_text_status", 1, self )); 
  }
  arg1 = (struct svn_wc_status_t *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "enum svn_wc_status_kind","repos_text_status", 2, argv[0] ));
  } 
  arg2 = (enum svn_wc_status_kind)(val2);
  if (arg1) (arg1)->repos_text_status = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status_t_repos_text_status_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  enum svn_wc_status_kind result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status_t *","repos_text_status", 1, self )); 
  }
  arg1 = (struct svn_wc_status_t *)(argp1);
  result = (enum svn_wc_status_kind) ((arg1)->repos_text_status);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status_t_repos_prop_status_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status_t *arg1 = 0 ;
  enum svn_wc_status_kind arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status_t *","repos_prop_status", 1, self )); 
  }
  arg1 = (struct svn_wc_status_t *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "enum svn_wc_status_kind","repos_prop_status", 2, argv[0] ));
  } 
  arg2 = (enum svn_wc_status_kind)(val2);
  if (arg1) (arg1)->repos_prop_status = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status_t_repos_prop_status_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  enum svn_wc_status_kind result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_status_t *","repos_prop_status", 1, self )); 
  }
  arg1 = (struct svn_wc_status_t *)(argp1);
  result = (enum svn_wc_status_kind) ((arg1)->repos_prop_status);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
_wrap_svn_wc_status_t_allocate(VALUE self)
#else
_wrap_svn_wc_status_t_allocate(int argc, VALUE *argv, VALUE self)
#endif
{
  VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_svn_wc_status_t);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
  rb_obj_call_init(vresult, argc, argv);
#endif
  return vresult;
}


SWIGINTERN VALUE
_wrap_new_svn_wc_status_t(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_status_t *result = 0 ;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  {
    result = (struct svn_wc_status_t *)calloc(1, sizeof(struct svn_wc_status_t));
    DATA_PTR(self) = result;
    
    
    
  }
  return self;
fail:
  return Qnil;
}


SWIGINTERN void
free_svn_wc_status_t(void *self) {
    struct svn_wc_status_t *arg1 = (struct svn_wc_status_t *)self;
    free((char *) arg1);
}

SWIGINTERN VALUE
_wrap_svn_wc_dup_status3(int argc, VALUE *argv, VALUE self) {
  svn_wc_status3_t *arg1 = 0 ;
  apr_pool_t *arg2 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_wc_status3_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_status3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_status3_t const *","svn_wc_dup_status3", 1, argv[0] )); 
  }
  arg1 = (svn_wc_status3_t *)(argp1);
  if (argc > 1) {
    
  }
  {
    result = (svn_wc_status3_t *)svn_wc_dup_status3((struct svn_wc_status3_t const *)arg1,arg2);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_wc_status3_t, 0 |  0 );
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_dup_status2(int argc, VALUE *argv, VALUE self) {
  svn_wc_status2_t *arg1 = 0 ;
  apr_pool_t *arg2 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_wc_status2_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_status2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_status2_t const *","svn_wc_dup_status2", 1, argv[0] )); 
  }
  arg1 = (svn_wc_status2_t *)(argp1);
  if (argc > 1) {
    
  }
  {
    result = (svn_wc_status2_t *)svn_wc_dup_status2((struct svn_wc_status2_t const *)arg1,arg2);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_wc_status2_t, 0 |  0 );
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_dup_status(int argc, VALUE *argv, VALUE self) {
  svn_wc_status_t *arg1 = 0 ;
  apr_pool_t *arg2 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_wc_status_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg2);
    _global_pool = arg2;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_status_t const *","svn_wc_dup_status", 1, argv[0] )); 
  }
  arg1 = (svn_wc_status_t *)(argp1);
  if (argc > 1) {
    
  }
  {
    result = (svn_wc_status_t *)svn_wc_dup_status((struct svn_wc_status_t const *)arg1,arg2);
    
    
    
  }
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_svn_wc_status_t, 0 |  0 );
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status3(int argc, VALUE *argv, VALUE self) {
  svn_wc_status3_t **arg1 = 0 ;
  svn_wc_context_t *arg2 = 0 ;
  char *arg3 = 0 ;
  apr_pool_t *arg4 = 0 ;
  apr_pool_t *arg5 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_wc_status3_t *temp1 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 2) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_svn_wc_context_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_context_t *","svn_wc_status3", 2, argv[0] )); 
  }
  arg2 = (svn_wc_context_t *)(argp2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_wc_status3", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  if (argc > 2) {
    
  }
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_wc_status3(arg1,arg2,(char const *)arg3,arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    /* FIXME: Missing argout typemap: svn_wc_status3 arg 1 (svn_wc_status3_t **) */
    
    
    
    
    SWIG_exception(SWIG_ValueError, "svn_wc_status3 is not implemented yet");
    
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status2(int argc, VALUE *argv, VALUE self) {
  svn_wc_status2_t **arg1 = 0 ;
  char *arg2 = 0 ;
  svn_wc_adm_access_t *arg3 = 0 ;
  apr_pool_t *arg4 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_wc_status2_t *temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_wc_status2", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_ConvertPtr(argv[1], &argp3,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_status2", 3, argv[1] )); 
  }
  arg3 = (svn_wc_adm_access_t *)(argp3);
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_wc_status2(arg1,(char const *)arg2,arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_wc_status2_t, 0), 0);
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status(int argc, VALUE *argv, VALUE self) {
  svn_wc_status_t **arg1 = 0 ;
  char *arg2 = 0 ;
  svn_wc_adm_access_t *arg3 = 0 ;
  apr_pool_t *arg4 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_wc_status_t *temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_wc_status", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_ConvertPtr(argv[1], &argp3,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_status", 3, argv[1] )); 
  }
  arg3 = (svn_wc_adm_access_t *)(argp3);
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_wc_status(arg1,(char const *)arg2,arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_wc_status_t, 0), 0);
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_walk_status(int argc, VALUE *argv, VALUE self) {
  svn_wc_context_t *arg1 = 0 ;
  char *arg2 = 0 ;
  svn_depth_t arg3 ;
  svn_boolean_t arg4 ;
  svn_boolean_t arg5 ;
  svn_boolean_t arg6 ;
  apr_array_header_t *arg7 = 0 ;
  svn_wc_status_func4_t arg8 = 0 ;
  void *arg9 = 0 ;
  svn_cancel_func_t arg10 = 0 ;
  void *arg11 = 0 ;
  apr_pool_t *arg12 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  int res9 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg12);
    _global_pool = arg12;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 10) || (argc > 11)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 10)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_context_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_context_t *","svn_wc_walk_status", 1, argv[0] )); 
  }
  arg1 = (svn_wc_context_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_wc_walk_status", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  {
    arg3 = svn_swig_rb_to_depth(argv[2]);
  }
  arg4 = RTEST(argv[3]);
  arg5 = RTEST(argv[4]);
  arg6 = RTEST(argv[5]);
  res7 = SWIG_ConvertPtr(argv[6], &argp7,SWIGTYPE_p_apr_array_header_t, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), Ruby_Format_TypeError( "", "apr_array_header_t const *","svn_wc_walk_status", 7, argv[6] )); 
  }
  arg7 = (apr_array_header_t *)(argp7);
  {
    int res = SWIG_ConvertFunctionPtr(argv[7], (void**)(&arg8), SWIGTYPE_p_f_p_void_p_q_const__char_p_q_const__svn_wc_status3_t_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_wc_status_func4_t","svn_wc_walk_status", 8, argv[7] )); 
    }
  }
  res9 = SWIG_ConvertPtr(argv[8],SWIG_as_voidptrptr(&arg9), 0, 0);
  if (!SWIG_IsOK(res9)) {
    SWIG_exception_fail(SWIG_ArgError(res9), Ruby_Format_TypeError( "", "void *","svn_wc_walk_status", 9, argv[8] )); 
  }
  {
    arg10 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg11 = (void *)svn_swig_rb_make_baton(argv[9], _global_svn_swig_rb_pool);
  }
  if (argc > 10) {
    
  }
  {
    result = (svn_error_t *)svn_wc_walk_status(arg1,(char const *)arg2,arg3,arg4,arg5,arg6,(apr_array_header_t const *)arg7,arg8,arg9,arg10,arg11,arg12);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg11);
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_get_status_editor5(int argc, VALUE *argv, VALUE self) {
  svn_delta_editor_t **arg1 = 0 ;
  void **arg2 = 0 ;
  void **arg3 = 0 ;
  svn_revnum_t *arg4 = 0 ;
  svn_wc_context_t *arg5 = 0 ;
  char *arg6 = 0 ;
  char *arg7 = 0 ;
  svn_depth_t arg8 ;
  svn_boolean_t arg9 ;
  svn_boolean_t arg10 ;
  svn_boolean_t arg11 ;
  svn_boolean_t arg12 ;
  apr_array_header_t *arg13 = 0 ;
  svn_wc_status_func4_t arg14 = 0 ;
  void *arg15 = 0 ;
  svn_cancel_func_t arg16 = 0 ;
  void *arg17 = 0 ;
  apr_pool_t *arg18 = 0 ;
  apr_pool_t *arg19 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_delta_editor_t *temp1 ;
  void *temp2 ;
  void *temp3 ;
  svn_revnum_t temp4 ;
  int res4 = SWIG_TMPOBJ ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  int res6 ;
  char *buf6 = 0 ;
  int alloc6 = 0 ;
  int res7 ;
  char *buf7 = 0 ;
  int alloc7 = 0 ;
  void *argp13 = 0 ;
  int res13 = 0 ;
  int res15 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg18);
    _global_pool = arg18;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg19);
    _global_pool = arg19;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  arg2 = &temp2;
  arg3 = &temp3;
  arg4 = &temp4;
  if ((argc < 12) || (argc > 14)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 12)",argc); SWIG_fail;
  }
  res5 = SWIG_ConvertPtr(argv[0], &argp5,SWIGTYPE_p_svn_wc_context_t, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "svn_wc_context_t *","svn_wc_get_status_editor5", 5, argv[0] )); 
  }
  arg5 = (svn_wc_context_t *)(argp5);
  res6 = SWIG_AsCharPtrAndSize(argv[1], &buf6, NULL, &alloc6);
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "char const *","svn_wc_get_status_editor5", 6, argv[1] ));
  }
  arg6 = (char *)(buf6);
  res7 = SWIG_AsCharPtrAndSize(argv[2], &buf7, NULL, &alloc7);
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), Ruby_Format_TypeError( "", "char const *","svn_wc_get_status_editor5", 7, argv[2] ));
  }
  arg7 = (char *)(buf7);
  {
    arg8 = svn_swig_rb_to_depth(argv[3]);
  }
  arg9 = RTEST(argv[4]);
  arg10 = RTEST(argv[5]);
  arg11 = RTEST(argv[6]);
  arg12 = RTEST(argv[7]);
  res13 = SWIG_ConvertPtr(argv[8], &argp13,SWIGTYPE_p_apr_array_header_t, 0 |  0 );
  if (!SWIG_IsOK(res13)) {
    SWIG_exception_fail(SWIG_ArgError(res13), Ruby_Format_TypeError( "", "apr_array_header_t const *","svn_wc_get_status_editor5", 13, argv[8] )); 
  }
  arg13 = (apr_array_header_t *)(argp13);
  {
    int res = SWIG_ConvertFunctionPtr(argv[9], (void**)(&arg14), SWIGTYPE_p_f_p_void_p_q_const__char_p_q_const__svn_wc_status3_t_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_wc_status_func4_t","svn_wc_get_status_editor5", 14, argv[9] )); 
    }
  }
  res15 = SWIG_ConvertPtr(argv[10],SWIG_as_voidptrptr(&arg15), 0, 0);
  if (!SWIG_IsOK(res15)) {
    SWIG_exception_fail(SWIG_ArgError(res15), Ruby_Format_TypeError( "", "void *","svn_wc_get_status_editor5", 15, argv[10] )); 
  }
  {
    arg16 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg17 = (void *)svn_swig_rb_make_baton(argv[11], _global_svn_swig_rb_pool);
  }
  if (argc > 12) {
    
  }
  if (argc > 13) {
    
  }
  {
    result = (svn_error_t *)svn_wc_get_status_editor5((struct svn_delta_editor_t const **)arg1,arg2,arg3,arg4,arg5,(char const *)arg6,(char const *)arg7,arg8,arg9,arg10,arg11,arg12,(apr_array_header_t const *)arg13,arg14,arg15,arg16,arg17,arg18,arg19);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_delta_editor_t, 0), 0);
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg2, SWIGTYPE_p_void, 0), 0);
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg3, SWIGTYPE_p_void, 0), 0);
  }
  if (SWIG_IsTmpObj(res4)) {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_From_long((*arg4)), 0);
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_long, new_flags), 0);
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg17);
  }
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  if (alloc7 == SWIG_NEWOBJ) free((char*)buf7);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  if (alloc7 == SWIG_NEWOBJ) free((char*)buf7);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_get_status_editor4(int argc, VALUE *argv, VALUE self) {
  svn_delta_editor_t **arg1 = 0 ;
  void **arg2 = 0 ;
  void **arg3 = 0 ;
  svn_revnum_t *arg4 = 0 ;
  svn_wc_adm_access_t *arg5 = 0 ;
  char *arg6 = 0 ;
  svn_depth_t arg7 ;
  svn_boolean_t arg8 ;
  svn_boolean_t arg9 ;
  apr_array_header_t *arg10 = 0 ;
  svn_wc_status_func3_t arg11 = 0 ;
  void *arg12 = 0 ;
  svn_cancel_func_t arg13 = 0 ;
  void *arg14 = 0 ;
  svn_wc_traversal_info_t *arg15 = 0 ;
  apr_pool_t *arg16 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_delta_editor_t *temp1 ;
  void *temp2 ;
  void *temp3 ;
  svn_revnum_t temp4 ;
  int res4 = SWIG_TMPOBJ ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  int res6 ;
  char *buf6 = 0 ;
  int alloc6 = 0 ;
  void *argp10 = 0 ;
  int res10 = 0 ;
  int res12 ;
  void *argp15 = 0 ;
  int res15 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg16);
    _global_pool = arg16;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  arg2 = &temp2;
  arg3 = &temp3;
  arg4 = &temp4;
  if ((argc < 10) || (argc > 11)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 10)",argc); SWIG_fail;
  }
  res5 = SWIG_ConvertPtr(argv[0], &argp5,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_get_status_editor4", 5, argv[0] )); 
  }
  arg5 = (svn_wc_adm_access_t *)(argp5);
  res6 = SWIG_AsCharPtrAndSize(argv[1], &buf6, NULL, &alloc6);
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "char const *","svn_wc_get_status_editor4", 6, argv[1] ));
  }
  arg6 = (char *)(buf6);
  {
    arg7 = svn_swig_rb_to_depth(argv[2]);
  }
  arg8 = RTEST(argv[3]);
  arg9 = RTEST(argv[4]);
  res10 = SWIG_ConvertPtr(argv[5], &argp10,SWIGTYPE_p_apr_array_header_t, 0 |  0 );
  if (!SWIG_IsOK(res10)) {
    SWIG_exception_fail(SWIG_ArgError(res10), Ruby_Format_TypeError( "", "apr_array_header_t const *","svn_wc_get_status_editor4", 10, argv[5] )); 
  }
  arg10 = (apr_array_header_t *)(argp10);
  {
    int res = SWIG_ConvertFunctionPtr(argv[6], (void**)(&arg11), SWIGTYPE_p_f_p_void_p_q_const__char_p_svn_wc_status2_t_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_wc_status_func3_t","svn_wc_get_status_editor4", 11, argv[6] )); 
    }
  }
  res12 = SWIG_ConvertPtr(argv[7],SWIG_as_voidptrptr(&arg12), 0, 0);
  if (!SWIG_IsOK(res12)) {
    SWIG_exception_fail(SWIG_ArgError(res12), Ruby_Format_TypeError( "", "void *","svn_wc_get_status_editor4", 12, argv[7] )); 
  }
  {
    arg13 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg14 = (void *)svn_swig_rb_make_baton(argv[8], _global_svn_swig_rb_pool);
  }
  res15 = SWIG_ConvertPtr(argv[9], &argp15,SWIGTYPE_p_svn_wc_traversal_info_t, 0 |  0 );
  if (!SWIG_IsOK(res15)) {
    SWIG_exception_fail(SWIG_ArgError(res15), Ruby_Format_TypeError( "", "svn_wc_traversal_info_t *","svn_wc_get_status_editor4", 15, argv[9] )); 
  }
  arg15 = (svn_wc_traversal_info_t *)(argp15);
  if (argc > 10) {
    
  }
  {
    result = (svn_error_t *)svn_wc_get_status_editor4((struct svn_delta_editor_t const **)arg1,arg2,arg3,arg4,arg5,(char const *)arg6,arg7,arg8,arg9,(apr_array_header_t const *)arg10,arg11,arg12,arg13,arg14,arg15,arg16);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_delta_editor_t, 0), 0);
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg2, SWIGTYPE_p_void, 0), 0);
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg3, SWIGTYPE_p_void, 0), 0);
  }
  if (SWIG_IsTmpObj(res4)) {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_From_long((*arg4)), 0);
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_long, new_flags), 0);
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg14);
  }
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_get_status_editor3(int argc, VALUE *argv, VALUE self) {
  svn_delta_editor_t **arg1 = 0 ;
  void **arg2 = 0 ;
  void **arg3 = 0 ;
  svn_revnum_t *arg4 = 0 ;
  svn_wc_adm_access_t *arg5 = 0 ;
  char *arg6 = 0 ;
  svn_depth_t arg7 ;
  svn_boolean_t arg8 ;
  svn_boolean_t arg9 ;
  apr_array_header_t *arg10 = 0 ;
  svn_wc_status_func2_t arg11 = 0 ;
  void *arg12 = 0 ;
  svn_cancel_func_t arg13 = 0 ;
  void *arg14 = 0 ;
  svn_wc_traversal_info_t *arg15 = 0 ;
  apr_pool_t *arg16 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_delta_editor_t *temp1 ;
  void *temp2 ;
  void *temp3 ;
  svn_revnum_t temp4 ;
  int res4 = SWIG_TMPOBJ ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  int res6 ;
  char *buf6 = 0 ;
  int alloc6 = 0 ;
  void *argp10 = 0 ;
  int res10 = 0 ;
  void *argp15 = 0 ;
  int res15 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg16);
    _global_pool = arg16;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  arg2 = &temp2;
  arg3 = &temp3;
  arg4 = &temp4;
  if ((argc < 9) || (argc > 10)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 9)",argc); SWIG_fail;
  }
  res5 = SWIG_ConvertPtr(argv[0], &argp5,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_get_status_editor3", 5, argv[0] )); 
  }
  arg5 = (svn_wc_adm_access_t *)(argp5);
  res6 = SWIG_AsCharPtrAndSize(argv[1], &buf6, NULL, &alloc6);
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "char const *","svn_wc_get_status_editor3", 6, argv[1] ));
  }
  arg6 = (char *)(buf6);
  {
    arg7 = svn_swig_rb_to_depth(argv[2]);
  }
  arg8 = RTEST(argv[3]);
  arg9 = RTEST(argv[4]);
  res10 = SWIG_ConvertPtr(argv[5], &argp10,SWIGTYPE_p_apr_array_header_t, 0 |  0 );
  if (!SWIG_IsOK(res10)) {
    SWIG_exception_fail(SWIG_ArgError(res10), Ruby_Format_TypeError( "", "apr_array_header_t const *","svn_wc_get_status_editor3", 10, argv[5] )); 
  }
  arg10 = (apr_array_header_t *)(argp10);
  {
    arg11 = (svn_wc_status_func2_t) svn_swig_rb_wc_status_func;
    arg12 = (void *)svn_swig_rb_make_baton(argv[6], _global_svn_swig_rb_pool);
  }
  {
    arg13 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg14 = (void *)svn_swig_rb_make_baton(argv[7], _global_svn_swig_rb_pool);
  }
  res15 = SWIG_ConvertPtr(argv[8], &argp15,SWIGTYPE_p_svn_wc_traversal_info_t, 0 |  0 );
  if (!SWIG_IsOK(res15)) {
    SWIG_exception_fail(SWIG_ArgError(res15), Ruby_Format_TypeError( "", "svn_wc_traversal_info_t *","svn_wc_get_status_editor3", 15, argv[8] )); 
  }
  arg15 = (svn_wc_traversal_info_t *)(argp15);
  if (argc > 9) {
    
  }
  {
    result = (svn_error_t *)svn_wc_get_status_editor3((struct svn_delta_editor_t const **)arg1,arg2,arg3,arg4,arg5,(char const *)arg6,arg7,arg8,arg9,(apr_array_header_t const *)arg10,arg11,arg12,arg13,arg14,arg15,arg16);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_delta_editor_t, 0), 0);
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg2, SWIGTYPE_p_void, 0), 0);
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg3, SWIGTYPE_p_void, 0), 0);
  }
  if (SWIG_IsTmpObj(res4)) {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_From_long((*arg4)), 0);
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_long, new_flags), 0);
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg14);
  }
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_get_status_editor2(int argc, VALUE *argv, VALUE self) {
  svn_delta_editor_t **arg1 = 0 ;
  void **arg2 = 0 ;
  void **arg3 = 0 ;
  svn_revnum_t *arg4 = 0 ;
  svn_wc_adm_access_t *arg5 = 0 ;
  char *arg6 = 0 ;
  apr_hash_t *arg7 = 0 ;
  svn_boolean_t arg8 ;
  svn_boolean_t arg9 ;
  svn_boolean_t arg10 ;
  svn_wc_status_func2_t arg11 = 0 ;
  void *arg12 = 0 ;
  svn_cancel_func_t arg13 = 0 ;
  void *arg14 = 0 ;
  svn_wc_traversal_info_t *arg15 = 0 ;
  apr_pool_t *arg16 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_delta_editor_t *temp1 ;
  void *temp2 ;
  void *temp3 ;
  svn_revnum_t temp4 ;
  int res4 = SWIG_TMPOBJ ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  int res6 ;
  char *buf6 = 0 ;
  int alloc6 = 0 ;
  void *argp15 = 0 ;
  int res15 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg16);
    _global_pool = arg16;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  arg2 = &temp2;
  arg3 = &temp3;
  arg4 = &temp4;
  if ((argc < 9) || (argc > 10)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 9)",argc); SWIG_fail;
  }
  res5 = SWIG_ConvertPtr(argv[0], &argp5,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_get_status_editor2", 5, argv[0] )); 
  }
  arg5 = (svn_wc_adm_access_t *)(argp5);
  res6 = SWIG_AsCharPtrAndSize(argv[1], &buf6, NULL, &alloc6);
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "char const *","svn_wc_get_status_editor2", 6, argv[1] ));
  }
  arg6 = (char *)(buf6);
  {
    if (NIL_P(argv[2])) {
      arg7 = NULL;
    } else {
      arg7 = svn_swig_rb_hash_to_apr_hash_swig_type(argv[2], "svn_config_t *", _global_pool);
    }
  }
  arg8 = RTEST(argv[3]);
  arg9 = RTEST(argv[4]);
  arg10 = RTEST(argv[5]);
  {
    arg11 = (svn_wc_status_func2_t) svn_swig_rb_wc_status_func;
    arg12 = (void *)svn_swig_rb_make_baton(argv[6], _global_svn_swig_rb_pool);
  }
  {
    arg13 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg14 = (void *)svn_swig_rb_make_baton(argv[7], _global_svn_swig_rb_pool);
  }
  res15 = SWIG_ConvertPtr(argv[8], &argp15,SWIGTYPE_p_svn_wc_traversal_info_t, 0 |  0 );
  if (!SWIG_IsOK(res15)) {
    SWIG_exception_fail(SWIG_ArgError(res15), Ruby_Format_TypeError( "", "svn_wc_traversal_info_t *","svn_wc_get_status_editor2", 15, argv[8] )); 
  }
  arg15 = (svn_wc_traversal_info_t *)(argp15);
  if (argc > 9) {
    
  }
  {
    result = (svn_error_t *)svn_wc_get_status_editor2((struct svn_delta_editor_t const **)arg1,arg2,arg3,arg4,arg5,(char const *)arg6,arg7,arg8,arg9,arg10,arg11,arg12,arg13,arg14,arg15,arg16);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_delta_editor_t, 0), 0);
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg2, SWIGTYPE_p_void, 0), 0);
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg3, SWIGTYPE_p_void, 0), 0);
  }
  if (SWIG_IsTmpObj(res4)) {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_From_long((*arg4)), 0);
  } else {
    int new_flags = SWIG_IsNewObj(res4) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj((void*)(arg4), SWIGTYPE_p_long, new_flags), 0);
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg14);
  }
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_get_status_editor(int argc, VALUE *argv, VALUE self) {
  svn_delta_editor_t **arg1 = 0 ;
  void **arg2 = 0 ;
  svn_revnum_t *arg3 = 0 ;
  svn_wc_adm_access_t *arg4 = 0 ;
  char *arg5 = 0 ;
  apr_hash_t *arg6 = 0 ;
  svn_boolean_t arg7 ;
  svn_boolean_t arg8 ;
  svn_boolean_t arg9 ;
  svn_wc_status_func_t arg10 = 0 ;
  void *arg11 = 0 ;
  svn_cancel_func_t arg12 = 0 ;
  void *arg13 = 0 ;
  svn_wc_traversal_info_t *arg14 = 0 ;
  apr_pool_t *arg15 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_delta_editor_t *temp1 ;
  void *temp2 ;
  svn_revnum_t temp3 ;
  int res3 = SWIG_TMPOBJ ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  int res5 ;
  char *buf5 = 0 ;
  int alloc5 = 0 ;
  int res11 ;
  void *argp14 = 0 ;
  int res14 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg15);
    _global_pool = arg15;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  arg2 = &temp2;
  arg3 = &temp3;
  if ((argc < 10) || (argc > 11)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 10)",argc); SWIG_fail;
  }
  res4 = SWIG_ConvertPtr(argv[0], &argp4,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_get_status_editor", 4, argv[0] )); 
  }
  arg4 = (svn_wc_adm_access_t *)(argp4);
  res5 = SWIG_AsCharPtrAndSize(argv[1], &buf5, NULL, &alloc5);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "char const *","svn_wc_get_status_editor", 5, argv[1] ));
  }
  arg5 = (char *)(buf5);
  {
    if (NIL_P(argv[2])) {
      arg6 = NULL;
    } else {
      arg6 = svn_swig_rb_hash_to_apr_hash_swig_type(argv[2], "svn_config_t *", _global_pool);
    }
  }
  arg7 = RTEST(argv[3]);
  arg8 = RTEST(argv[4]);
  arg9 = RTEST(argv[5]);
  {
    int res = SWIG_ConvertFunctionPtr(argv[6], (void**)(&arg10), SWIGTYPE_p_f_p_void_p_q_const__char_p_struct_svn_wc_status_t__void);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_wc_status_func_t","svn_wc_get_status_editor", 10, argv[6] )); 
    }
  }
  res11 = SWIG_ConvertPtr(argv[7],SWIG_as_voidptrptr(&arg11), 0, 0);
  if (!SWIG_IsOK(res11)) {
    SWIG_exception_fail(SWIG_ArgError(res11), Ruby_Format_TypeError( "", "void *","svn_wc_get_status_editor", 11, argv[7] )); 
  }
  {
    arg12 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg13 = (void *)svn_swig_rb_make_baton(argv[8], _global_svn_swig_rb_pool);
  }
  res14 = SWIG_ConvertPtr(argv[9], &argp14,SWIGTYPE_p_svn_wc_traversal_info_t, 0 |  0 );
  if (!SWIG_IsOK(res14)) {
    SWIG_exception_fail(SWIG_ArgError(res14), Ruby_Format_TypeError( "", "svn_wc_traversal_info_t *","svn_wc_get_status_editor", 14, argv[9] )); 
  }
  arg14 = (svn_wc_traversal_info_t *)(argp14);
  if (argc > 10) {
    
  }
  {
    result = (svn_error_t *)svn_wc_get_status_editor((struct svn_delta_editor_t const **)arg1,arg2,arg3,arg4,(char const *)arg5,arg6,arg7,arg8,arg9,arg10,arg11,arg12,arg13,arg14,arg15);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_delta_editor_t, 0), 0);
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg2, SWIGTYPE_p_void, 0), 0);
  }
  if (SWIG_IsTmpObj(res3)) {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_From_long((*arg3)), 0);
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_long, new_flags), 0);
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg13);
  }
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_status_set_repos_locks(int argc, VALUE *argv, VALUE self) {
  void *arg1 = 0 ;
  apr_hash_t *arg2 = 0 ;
  char *arg3 = 0 ;
  apr_pool_t *arg4 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 3) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0],SWIG_as_voidptrptr(&arg1), 0, 0);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "void *","svn_wc_status_set_repos_locks", 1, argv[0] )); 
  }
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_apr_hash_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "apr_hash_t *","svn_wc_status_set_repos_locks", 2, argv[1] )); 
  }
  arg2 = (apr_hash_t *)(argp2);
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_wc_status_set_repos_locks", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_wc_status_set_repos_locks(arg1,arg2,(char const *)arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_copy3(int argc, VALUE *argv, VALUE self) {
  svn_wc_context_t *arg1 = 0 ;
  char *arg2 = 0 ;
  char *arg3 = 0 ;
  svn_boolean_t arg4 ;
  svn_cancel_func_t arg5 = 0 ;
  void *arg6 = 0 ;
  svn_wc_notify_func2_t arg7 = 0 ;
  void *arg8 = 0 ;
  apr_pool_t *arg9 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg9);
    _global_pool = arg9;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 6) || (argc > 7)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 6)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_context_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_context_t *","svn_wc_copy3", 1, argv[0] )); 
  }
  arg1 = (svn_wc_context_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_wc_copy3", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_wc_copy3", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  arg4 = RTEST(argv[3]);
  {
    arg5 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg6 = (void *)svn_swig_rb_make_baton(argv[4], _global_svn_swig_rb_pool);
  }
  {
    arg7 = (svn_wc_notify_func2_t) svn_swig_rb_notify_func2;
    arg8 = (void *)svn_swig_rb_make_baton(argv[5], _global_svn_swig_rb_pool);
  }
  if (argc > 6) {
    
  }
  {
    result = (svn_error_t *)svn_wc_copy3(arg1,(char const *)arg2,(char const *)arg3,arg4,arg5,arg6,arg7,arg8,arg9);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg6);
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_copy2(int argc, VALUE *argv, VALUE self) {
  char *arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  char *arg3 = 0 ;
  svn_cancel_func_t arg4 = 0 ;
  void *arg5 = 0 ;
  svn_wc_notify_func2_t arg6 = 0 ;
  void *arg7 = 0 ;
  apr_pool_t *arg8 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg8);
    _global_pool = arg8;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 5) || (argc > 6)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 5)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_wc_copy2", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_copy2", 2, argv[1] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_wc_copy2", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  {
    arg4 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg5 = (void *)svn_swig_rb_make_baton(argv[3], _global_svn_swig_rb_pool);
  }
  {
    arg6 = (svn_wc_notify_func2_t) svn_swig_rb_notify_func2;
    arg7 = (void *)svn_swig_rb_make_baton(argv[4], _global_svn_swig_rb_pool);
  }
  if (argc > 5) {
    
  }
  {
    result = (svn_error_t *)svn_wc_copy2((char const *)arg1,arg2,(char const *)arg3,arg4,arg5,arg6,arg7,arg8);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg5);
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_copy(int argc, VALUE *argv, VALUE self) {
  char *arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  char *arg3 = 0 ;
  svn_cancel_func_t arg4 = 0 ;
  void *arg5 = 0 ;
  svn_wc_notify_func_t arg6 = 0 ;
  void *arg7 = 0 ;
  apr_pool_t *arg8 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int res7 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg8);
    _global_pool = arg8;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 6) || (argc > 7)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 6)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_wc_copy", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_copy", 2, argv[1] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_wc_copy", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  {
    arg4 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg5 = (void *)svn_swig_rb_make_baton(argv[3], _global_svn_swig_rb_pool);
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[4], (void**)(&arg6), SWIGTYPE_p_f_p_void_p_q_const__char_enum_svn_wc_notify_action_t_enum_svn_node_kind_t_p_q_const__char_enum_svn_wc_notify_state_t_enum_svn_wc_notify_state_t_long__void);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_wc_notify_func_t","svn_wc_copy", 6, argv[4] )); 
    }
  }
  res7 = SWIG_ConvertPtr(argv[5],SWIG_as_voidptrptr(&arg7), 0, 0);
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), Ruby_Format_TypeError( "", "void *","svn_wc_copy", 7, argv[5] )); 
  }
  if (argc > 6) {
    
  }
  {
    result = (svn_error_t *)svn_wc_copy((char const *)arg1,arg2,(char const *)arg3,arg4,arg5,arg6,arg7,arg8);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg5);
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_move(int argc, VALUE *argv, VALUE self) {
  svn_wc_context_t *arg1 = 0 ;
  char *arg2 = 0 ;
  char *arg3 = 0 ;
  svn_boolean_t arg4 ;
  svn_cancel_func_t arg5 = 0 ;
  void *arg6 = 0 ;
  svn_wc_notify_func2_t arg7 = 0 ;
  void *arg8 = 0 ;
  apr_pool_t *arg9 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg9);
    _global_pool = arg9;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 6) || (argc > 7)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 6)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_context_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_context_t *","svn_wc_move", 1, argv[0] )); 
  }
  arg1 = (svn_wc_context_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_wc_move", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_wc_move", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  arg4 = RTEST(argv[3]);
  {
    arg5 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg6 = (void *)svn_swig_rb_make_baton(argv[4], _global_svn_swig_rb_pool);
  }
  {
    arg7 = (svn_wc_notify_func2_t) svn_swig_rb_notify_func2;
    arg8 = (void *)svn_swig_rb_make_baton(argv[5], _global_svn_swig_rb_pool);
  }
  if (argc > 6) {
    
  }
  {
    result = (svn_error_t *)svn_wc_move(arg1,(char const *)arg2,(char const *)arg3,arg4,arg5,arg6,arg7,arg8,arg9);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg6);
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_delete4(int argc, VALUE *argv, VALUE self) {
  svn_wc_context_t *arg1 = 0 ;
  char *arg2 = 0 ;
  svn_boolean_t arg3 ;
  svn_boolean_t arg4 ;
  svn_cancel_func_t arg5 = 0 ;
  void *arg6 = 0 ;
  svn_wc_notify_func2_t arg7 = 0 ;
  void *arg8 = 0 ;
  apr_pool_t *arg9 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg9);
    _global_pool = arg9;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 6) || (argc > 7)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 6)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_context_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_context_t *","svn_wc_delete4", 1, argv[0] )); 
  }
  arg1 = (svn_wc_context_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_wc_delete4", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  arg3 = RTEST(argv[2]);
  arg4 = RTEST(argv[3]);
  {
    arg5 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg6 = (void *)svn_swig_rb_make_baton(argv[4], _global_svn_swig_rb_pool);
  }
  {
    arg7 = (svn_wc_notify_func2_t) svn_swig_rb_notify_func2;
    arg8 = (void *)svn_swig_rb_make_baton(argv[5], _global_svn_swig_rb_pool);
  }
  if (argc > 6) {
    
  }
  {
    result = (svn_error_t *)svn_wc_delete4(arg1,(char const *)arg2,arg3,arg4,arg5,arg6,arg7,arg8,arg9);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg6);
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_delete3(int argc, VALUE *argv, VALUE self) {
  char *arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  svn_cancel_func_t arg3 = 0 ;
  void *arg4 = 0 ;
  svn_wc_notify_func2_t arg5 = 0 ;
  void *arg6 = 0 ;
  svn_boolean_t arg7 ;
  apr_pool_t *arg8 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg8);
    _global_pool = arg8;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 5) || (argc > 6)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 5)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_wc_delete3", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_delete3", 2, argv[1] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  {
    arg3 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg4 = (void *)svn_swig_rb_make_baton(argv[2], _global_svn_swig_rb_pool);
  }
  {
    arg5 = (svn_wc_notify_func2_t) svn_swig_rb_notify_func2;
    arg6 = (void *)svn_swig_rb_make_baton(argv[3], _global_svn_swig_rb_pool);
  }
  arg7 = RTEST(argv[4]);
  if (argc > 5) {
    
  }
  {
    result = (svn_error_t *)svn_wc_delete3((char const *)arg1,arg2,arg3,arg4,arg5,arg6,arg7,arg8);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg4);
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_delete2(int argc, VALUE *argv, VALUE self) {
  char *arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  svn_cancel_func_t arg3 = 0 ;
  void *arg4 = 0 ;
  svn_wc_notify_func2_t arg5 = 0 ;
  void *arg6 = 0 ;
  apr_pool_t *arg7 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg7);
    _global_pool = arg7;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 4) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_wc_delete2", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_delete2", 2, argv[1] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  {
    arg3 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg4 = (void *)svn_swig_rb_make_baton(argv[2], _global_svn_swig_rb_pool);
  }
  {
    arg5 = (svn_wc_notify_func2_t) svn_swig_rb_notify_func2;
    arg6 = (void *)svn_swig_rb_make_baton(argv[3], _global_svn_swig_rb_pool);
  }
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_wc_delete2((char const *)arg1,arg2,arg3,arg4,arg5,arg6,arg7);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg4);
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_delete(int argc, VALUE *argv, VALUE self) {
  char *arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  svn_cancel_func_t arg3 = 0 ;
  void *arg4 = 0 ;
  svn_wc_notify_func_t arg5 = 0 ;
  void *arg6 = 0 ;
  apr_pool_t *arg7 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res6 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg7);
    _global_pool = arg7;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 5) || (argc > 6)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 5)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_wc_delete", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_delete", 2, argv[1] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  {
    arg3 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg4 = (void *)svn_swig_rb_make_baton(argv[2], _global_svn_swig_rb_pool);
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[3], (void**)(&arg5), SWIGTYPE_p_f_p_void_p_q_const__char_enum_svn_wc_notify_action_t_enum_svn_node_kind_t_p_q_const__char_enum_svn_wc_notify_state_t_enum_svn_wc_notify_state_t_long__void);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_wc_notify_func_t","svn_wc_delete", 5, argv[3] )); 
    }
  }
  res6 = SWIG_ConvertPtr(argv[4],SWIG_as_voidptrptr(&arg6), 0, 0);
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "void *","svn_wc_delete", 6, argv[4] )); 
  }
  if (argc > 5) {
    
  }
  {
    result = (svn_error_t *)svn_wc_delete((char const *)arg1,arg2,arg3,arg4,arg5,arg6,arg7);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg4);
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_add_from_disk3(int argc, VALUE *argv, VALUE self) {
  svn_wc_context_t *arg1 = 0 ;
  char *arg2 = 0 ;
  apr_hash_t *arg3 = 0 ;
  svn_boolean_t arg4 ;
  svn_wc_notify_func2_t arg5 = 0 ;
  void *arg6 = 0 ;
  apr_pool_t *arg7 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg7);
    _global_pool = arg7;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 5) || (argc > 6)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 5)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_context_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_context_t *","svn_wc_add_from_disk3", 1, argv[0] )); 
  }
  arg1 = (svn_wc_context_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_wc_add_from_disk3", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  {
    VALUE rb_pool = Qnil;
    if (!_global_pool) {
      svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &_global_pool);
      svn_swig_rb_push_pool(rb_pool);
    }
    arg3 = svn_swig_rb_hash_to_apr_hash_svn_string(argv[2], _global_pool);
    _global_pool = NULL;
    if (!NIL_P(rb_pool)) {
      if (arg3 == NULL) {
        svn_swig_rb_destroy_pool(rb_pool);
      } else {
        svn_swig_rb_set_pool_for_no_swig_type(argv[2], rb_pool);
      }
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
    }
  }
  arg4 = RTEST(argv[3]);
  {
    arg5 = (svn_wc_notify_func2_t) svn_swig_rb_notify_func2;
    arg6 = (void *)svn_swig_rb_make_baton(argv[4], _global_svn_swig_rb_pool);
  }
  if (argc > 5) {
    
  }
  {
    result = (svn_error_t *)svn_wc_add_from_disk3(arg1,(char const *)arg2,(apr_hash_t const *)arg3,arg4,arg5,arg6,arg7);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_add_from_disk2(int argc, VALUE *argv, VALUE self) {
  svn_wc_context_t *arg1 = 0 ;
  char *arg2 = 0 ;
  apr_hash_t *arg3 = 0 ;
  svn_wc_notify_func2_t arg4 = 0 ;
  void *arg5 = 0 ;
  apr_pool_t *arg6 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 4) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_context_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_context_t *","svn_wc_add_from_disk2", 1, argv[0] )); 
  }
  arg1 = (svn_wc_context_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_wc_add_from_disk2", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  {
    VALUE rb_pool = Qnil;
    if (!_global_pool) {
      svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &_global_pool);
      svn_swig_rb_push_pool(rb_pool);
    }
    arg3 = svn_swig_rb_hash_to_apr_hash_svn_string(argv[2], _global_pool);
    _global_pool = NULL;
    if (!NIL_P(rb_pool)) {
      if (arg3 == NULL) {
        svn_swig_rb_destroy_pool(rb_pool);
      } else {
        svn_swig_rb_set_pool_for_no_swig_type(argv[2], rb_pool);
      }
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
    }
  }
  {
    arg4 = (svn_wc_notify_func2_t) svn_swig_rb_notify_func2;
    arg5 = (void *)svn_swig_rb_make_baton(argv[3], _global_svn_swig_rb_pool);
  }
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_wc_add_from_disk2(arg1,(char const *)arg2,(apr_hash_t const *)arg3,arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_add_from_disk(int argc, VALUE *argv, VALUE self) {
  svn_wc_context_t *arg1 = 0 ;
  char *arg2 = 0 ;
  svn_wc_notify_func2_t arg3 = 0 ;
  void *arg4 = 0 ;
  apr_pool_t *arg5 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 3) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_context_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_context_t *","svn_wc_add_from_disk", 1, argv[0] )); 
  }
  arg1 = (svn_wc_context_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_wc_add_from_disk", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  {
    arg3 = (svn_wc_notify_func2_t) svn_swig_rb_notify_func2;
    arg4 = (void *)svn_swig_rb_make_baton(argv[2], _global_svn_swig_rb_pool);
  }
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_wc_add_from_disk(arg1,(char const *)arg2,arg3,arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_add4(int argc, VALUE *argv, VALUE self) {
  svn_wc_context_t *arg1 = 0 ;
  char *arg2 = 0 ;
  svn_depth_t arg3 ;
  char *arg4 = 0 ;
  svn_revnum_t arg5 ;
  svn_cancel_func_t arg6 = 0 ;
  void *arg7 = 0 ;
  svn_wc_notify_func2_t arg8 = 0 ;
  void *arg9 = 0 ;
  apr_pool_t *arg10 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  long val5 ;
  int ecode5 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg10);
    _global_pool = arg10;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 7) || (argc > 8)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 7)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_context_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_context_t *","svn_wc_add4", 1, argv[0] )); 
  }
  arg1 = (svn_wc_context_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_wc_add4", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  {
    arg3 = svn_swig_rb_to_depth(argv[2]);
  }
  {
    if (NIL_P(argv[3])) {
      arg4 = NULL;
    } else {
      arg4 = StringValuePtr(argv[3]);
    }
  }
  ecode5 = SWIG_AsVal_long(argv[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), Ruby_Format_TypeError( "", "svn_revnum_t","svn_wc_add4", 5, argv[4] ));
  } 
  arg5 = (svn_revnum_t)(val5);
  {
    arg6 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg7 = (void *)svn_swig_rb_make_baton(argv[5], _global_svn_swig_rb_pool);
  }
  {
    arg8 = (svn_wc_notify_func2_t) svn_swig_rb_notify_func2;
    arg9 = (void *)svn_swig_rb_make_baton(argv[6], _global_svn_swig_rb_pool);
  }
  if (argc > 7) {
    
  }
  {
    result = (svn_error_t *)svn_wc_add4(arg1,(char const *)arg2,arg3,(char const *)arg4,arg5,arg6,arg7,arg8,arg9,arg10);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg7);
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_add3(int argc, VALUE *argv, VALUE self) {
  char *arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  svn_depth_t arg3 ;
  char *arg4 = 0 ;
  svn_revnum_t arg5 ;
  svn_cancel_func_t arg6 = 0 ;
  void *arg7 = 0 ;
  svn_wc_notify_func2_t arg8 = 0 ;
  void *arg9 = 0 ;
  apr_pool_t *arg10 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  long val5 ;
  int ecode5 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg10);
    _global_pool = arg10;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 7) || (argc > 8)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 7)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_wc_add3", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_add3", 2, argv[1] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  {
    arg3 = svn_swig_rb_to_depth(argv[2]);
  }
  {
    if (NIL_P(argv[3])) {
      arg4 = NULL;
    } else {
      arg4 = StringValuePtr(argv[3]);
    }
  }
  ecode5 = SWIG_AsVal_long(argv[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), Ruby_Format_TypeError( "", "svn_revnum_t","svn_wc_add3", 5, argv[4] ));
  } 
  arg5 = (svn_revnum_t)(val5);
  {
    arg6 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg7 = (void *)svn_swig_rb_make_baton(argv[5], _global_svn_swig_rb_pool);
  }
  {
    arg8 = (svn_wc_notify_func2_t) svn_swig_rb_notify_func2;
    arg9 = (void *)svn_swig_rb_make_baton(argv[6], _global_svn_swig_rb_pool);
  }
  if (argc > 7) {
    
  }
  {
    result = (svn_error_t *)svn_wc_add3((char const *)arg1,arg2,arg3,(char const *)arg4,arg5,arg6,arg7,arg8,arg9,arg10);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg7);
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_add2(int argc, VALUE *argv, VALUE self) {
  char *arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  char *arg3 = 0 ;
  svn_revnum_t arg4 ;
  svn_cancel_func_t arg5 = 0 ;
  void *arg6 = 0 ;
  svn_wc_notify_func2_t arg7 = 0 ;
  void *arg8 = 0 ;
  apr_pool_t *arg9 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  long val4 ;
  int ecode4 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg9);
    _global_pool = arg9;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 6) || (argc > 7)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 6)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_wc_add2", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_add2", 2, argv[1] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  {
    if (NIL_P(argv[2])) {
      arg3 = NULL;
    } else {
      arg3 = StringValuePtr(argv[2]);
    }
  }
  ecode4 = SWIG_AsVal_long(argv[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), Ruby_Format_TypeError( "", "svn_revnum_t","svn_wc_add2", 4, argv[3] ));
  } 
  arg4 = (svn_revnum_t)(val4);
  {
    arg5 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg6 = (void *)svn_swig_rb_make_baton(argv[4], _global_svn_swig_rb_pool);
  }
  {
    arg7 = (svn_wc_notify_func2_t) svn_swig_rb_notify_func2;
    arg8 = (void *)svn_swig_rb_make_baton(argv[5], _global_svn_swig_rb_pool);
  }
  if (argc > 6) {
    
  }
  {
    result = (svn_error_t *)svn_wc_add2((char const *)arg1,arg2,(char const *)arg3,arg4,arg5,arg6,arg7,arg8,arg9);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg6);
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_add(int argc, VALUE *argv, VALUE self) {
  char *arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  char *arg3 = 0 ;
  svn_revnum_t arg4 ;
  svn_cancel_func_t arg5 = 0 ;
  void *arg6 = 0 ;
  svn_wc_notify_func_t arg7 = 0 ;
  void *arg8 = 0 ;
  apr_pool_t *arg9 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  long val4 ;
  int ecode4 = 0 ;
  int res8 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg9);
    _global_pool = arg9;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 7) || (argc > 8)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 7)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_wc_add", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_add", 2, argv[1] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  {
    if (NIL_P(argv[2])) {
      arg3 = NULL;
    } else {
      arg3 = StringValuePtr(argv[2]);
    }
  }
  ecode4 = SWIG_AsVal_long(argv[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), Ruby_Format_TypeError( "", "svn_revnum_t","svn_wc_add", 4, argv[3] ));
  } 
  arg4 = (svn_revnum_t)(val4);
  {
    arg5 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg6 = (void *)svn_swig_rb_make_baton(argv[4], _global_svn_swig_rb_pool);
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[5], (void**)(&arg7), SWIGTYPE_p_f_p_void_p_q_const__char_enum_svn_wc_notify_action_t_enum_svn_node_kind_t_p_q_const__char_enum_svn_wc_notify_state_t_enum_svn_wc_notify_state_t_long__void);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_wc_notify_func_t","svn_wc_add", 7, argv[5] )); 
    }
  }
  res8 = SWIG_ConvertPtr(argv[6],SWIG_as_voidptrptr(&arg8), 0, 0);
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), Ruby_Format_TypeError( "", "void *","svn_wc_add", 8, argv[6] )); 
  }
  if (argc > 7) {
    
  }
  {
    result = (svn_error_t *)svn_wc_add((char const *)arg1,arg2,(char const *)arg3,arg4,arg5,arg6,arg7,arg8,arg9);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg6);
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_add_repos_file4(int argc, VALUE *argv, VALUE self) {
  svn_wc_context_t *arg1 = 0 ;
  char *arg2 = 0 ;
  svn_stream_t *arg3 = 0 ;
  svn_stream_t *arg4 = 0 ;
  apr_hash_t *arg5 = 0 ;
  apr_hash_t *arg6 = 0 ;
  char *arg7 = 0 ;
  svn_revnum_t arg8 ;
  svn_cancel_func_t arg9 = 0 ;
  void *arg10 = 0 ;
  apr_pool_t *arg11 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  long val8 ;
  int ecode8 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg11);
    _global_pool = arg11;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 9) || (argc > 10)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 9)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_context_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_context_t *","svn_wc_add_repos_file4", 1, argv[0] )); 
  }
  arg1 = (svn_wc_context_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_wc_add_repos_file4", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  {
    arg3 = svn_swig_rb_make_stream(argv[2]);
  }
  {
    arg4 = svn_swig_rb_make_stream(argv[3]);
  }
  {
    VALUE rb_pool = Qnil;
    if (!_global_pool) {
      svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &_global_pool);
      svn_swig_rb_push_pool(rb_pool);
    }
    arg5 = svn_swig_rb_hash_to_apr_hash_svn_string(argv[4], _global_pool);
    _global_pool = NULL;
    if (!NIL_P(rb_pool)) {
      if (arg5 == NULL) {
        svn_swig_rb_destroy_pool(rb_pool);
      } else {
        svn_swig_rb_set_pool_for_no_swig_type(argv[4], rb_pool);
      }
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
    }
  }
  {
    VALUE rb_pool = Qnil;
    if (!_global_pool) {
      svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &_global_pool);
      svn_swig_rb_push_pool(rb_pool);
    }
    arg6 = svn_swig_rb_hash_to_apr_hash_svn_string(argv[5], _global_pool);
    _global_pool = NULL;
    if (!NIL_P(rb_pool)) {
      if (arg6 == NULL) {
        svn_swig_rb_destroy_pool(rb_pool);
      } else {
        svn_swig_rb_set_pool_for_no_swig_type(argv[5], rb_pool);
      }
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
    }
  }
  {
    if (NIL_P(argv[6])) {
      arg7 = NULL;
    } else {
      arg7 = StringValuePtr(argv[6]);
    }
  }
  ecode8 = SWIG_AsVal_long(argv[7], &val8);
  if (!SWIG_IsOK(ecode8)) {
    SWIG_exception_fail(SWIG_ArgError(ecode8), Ruby_Format_TypeError( "", "svn_revnum_t","svn_wc_add_repos_file4", 8, argv[7] ));
  } 
  arg8 = (svn_revnum_t)(val8);
  {
    arg9 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg10 = (void *)svn_swig_rb_make_baton(argv[8], _global_svn_swig_rb_pool);
  }
  if (argc > 9) {
    
  }
  {
    result = (svn_error_t *)svn_wc_add_repos_file4(arg1,(char const *)arg2,arg3,arg4,arg5,arg6,(char const *)arg7,arg8,arg9,arg10,arg11);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg10);
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_add_repos_file3(int argc, VALUE *argv, VALUE self) {
  char *arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  svn_stream_t *arg3 = 0 ;
  svn_stream_t *arg4 = 0 ;
  apr_hash_t *arg5 = 0 ;
  apr_hash_t *arg6 = 0 ;
  char *arg7 = 0 ;
  svn_revnum_t arg8 ;
  svn_cancel_func_t arg9 = 0 ;
  void *arg10 = 0 ;
  svn_wc_notify_func2_t arg11 = 0 ;
  void *arg12 = 0 ;
  apr_pool_t *arg13 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  long val8 ;
  int ecode8 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg13);
    _global_pool = arg13;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 10) || (argc > 11)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 10)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_wc_add_repos_file3", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_add_repos_file3", 2, argv[1] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  {
    arg3 = svn_swig_rb_make_stream(argv[2]);
  }
  {
    arg4 = svn_swig_rb_make_stream(argv[3]);
  }
  {
    VALUE rb_pool = Qnil;
    if (!_global_pool) {
      svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &_global_pool);
      svn_swig_rb_push_pool(rb_pool);
    }
    arg5 = svn_swig_rb_hash_to_apr_hash_svn_string(argv[4], _global_pool);
    _global_pool = NULL;
    if (!NIL_P(rb_pool)) {
      if (arg5 == NULL) {
        svn_swig_rb_destroy_pool(rb_pool);
      } else {
        svn_swig_rb_set_pool_for_no_swig_type(argv[4], rb_pool);
      }
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
    }
  }
  {
    VALUE rb_pool = Qnil;
    if (!_global_pool) {
      svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &_global_pool);
      svn_swig_rb_push_pool(rb_pool);
    }
    arg6 = svn_swig_rb_hash_to_apr_hash_svn_string(argv[5], _global_pool);
    _global_pool = NULL;
    if (!NIL_P(rb_pool)) {
      if (arg6 == NULL) {
        svn_swig_rb_destroy_pool(rb_pool);
      } else {
        svn_swig_rb_set_pool_for_no_swig_type(argv[5], rb_pool);
      }
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
    }
  }
  {
    if (NIL_P(argv[6])) {
      arg7 = NULL;
    } else {
      arg7 = StringValuePtr(argv[6]);
    }
  }
  ecode8 = SWIG_AsVal_long(argv[7], &val8);
  if (!SWIG_IsOK(ecode8)) {
    SWIG_exception_fail(SWIG_ArgError(ecode8), Ruby_Format_TypeError( "", "svn_revnum_t","svn_wc_add_repos_file3", 8, argv[7] ));
  } 
  arg8 = (svn_revnum_t)(val8);
  {
    arg9 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg10 = (void *)svn_swig_rb_make_baton(argv[8], _global_svn_swig_rb_pool);
  }
  {
    arg11 = (svn_wc_notify_func2_t) svn_swig_rb_notify_func2;
    arg12 = (void *)svn_swig_rb_make_baton(argv[9], _global_svn_swig_rb_pool);
  }
  if (argc > 10) {
    
  }
  {
    result = (svn_error_t *)svn_wc_add_repos_file3((char const *)arg1,arg2,arg3,arg4,arg5,arg6,(char const *)arg7,arg8,arg9,arg10,arg11,arg12,arg13);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg10);
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_add_repos_file2(int argc, VALUE *argv, VALUE self) {
  char *arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  char *arg3 = 0 ;
  char *arg4 = 0 ;
  apr_hash_t *arg5 = 0 ;
  apr_hash_t *arg6 = 0 ;
  char *arg7 = 0 ;
  svn_revnum_t arg8 ;
  apr_pool_t *arg9 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  long val8 ;
  int ecode8 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg9);
    _global_pool = arg9;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 8) || (argc > 9)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 8)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_wc_add_repos_file2", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_add_repos_file2", 2, argv[1] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_wc_add_repos_file2", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  {
    if (NIL_P(argv[3])) {
      arg4 = NULL;
    } else {
      arg4 = StringValuePtr(argv[3]);
    }
  }
  {
    VALUE rb_pool = Qnil;
    if (!_global_pool) {
      svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &_global_pool);
      svn_swig_rb_push_pool(rb_pool);
    }
    arg5 = svn_swig_rb_hash_to_apr_hash_svn_string(argv[4], _global_pool);
    _global_pool = NULL;
    if (!NIL_P(rb_pool)) {
      if (arg5 == NULL) {
        svn_swig_rb_destroy_pool(rb_pool);
      } else {
        svn_swig_rb_set_pool_for_no_swig_type(argv[4], rb_pool);
      }
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
    }
  }
  {
    VALUE rb_pool = Qnil;
    if (!_global_pool) {
      svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &_global_pool);
      svn_swig_rb_push_pool(rb_pool);
    }
    arg6 = svn_swig_rb_hash_to_apr_hash_svn_string(argv[5], _global_pool);
    _global_pool = NULL;
    if (!NIL_P(rb_pool)) {
      if (arg6 == NULL) {
        svn_swig_rb_destroy_pool(rb_pool);
      } else {
        svn_swig_rb_set_pool_for_no_swig_type(argv[5], rb_pool);
      }
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
    }
  }
  {
    if (NIL_P(argv[6])) {
      arg7 = NULL;
    } else {
      arg7 = StringValuePtr(argv[6]);
    }
  }
  ecode8 = SWIG_AsVal_long(argv[7], &val8);
  if (!SWIG_IsOK(ecode8)) {
    SWIG_exception_fail(SWIG_ArgError(ecode8), Ruby_Format_TypeError( "", "svn_revnum_t","svn_wc_add_repos_file2", 8, argv[7] ));
  } 
  arg8 = (svn_revnum_t)(val8);
  if (argc > 8) {
    
  }
  {
    result = (svn_error_t *)svn_wc_add_repos_file2((char const *)arg1,arg2,(char const *)arg3,(char const *)arg4,arg5,arg6,(char const *)arg7,arg8,arg9);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_add_repos_file(int argc, VALUE *argv, VALUE self) {
  char *arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  char *arg3 = 0 ;
  apr_hash_t *arg4 = 0 ;
  char *arg5 = 0 ;
  svn_revnum_t arg6 ;
  apr_pool_t *arg7 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  long val6 ;
  int ecode6 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg7);
    _global_pool = arg7;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 6) || (argc > 7)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 6)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_wc_add_repos_file", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_add_repos_file", 2, argv[1] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  {
    if (NIL_P(argv[2])) {
      arg3 = NULL;
    } else {
      arg3 = StringValuePtr(argv[2]);
    }
  }
  {
    VALUE rb_pool = Qnil;
    if (!_global_pool) {
      svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &_global_pool);
      svn_swig_rb_push_pool(rb_pool);
    }
    arg4 = svn_swig_rb_hash_to_apr_hash_svn_string(argv[3], _global_pool);
    _global_pool = NULL;
    if (!NIL_P(rb_pool)) {
      if (arg4 == NULL) {
        svn_swig_rb_destroy_pool(rb_pool);
      } else {
        svn_swig_rb_set_pool_for_no_swig_type(argv[3], rb_pool);
      }
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
    }
  }
  {
    if (NIL_P(argv[4])) {
      arg5 = NULL;
    } else {
      arg5 = StringValuePtr(argv[4]);
    }
  }
  ecode6 = SWIG_AsVal_long(argv[5], &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), Ruby_Format_TypeError( "", "svn_revnum_t","svn_wc_add_repos_file", 6, argv[5] ));
  } 
  arg6 = (svn_revnum_t)(val6);
  if (argc > 6) {
    
  }
  {
    result = (svn_error_t *)svn_wc_add_repos_file((char const *)arg1,arg2,(char const *)arg3,arg4,(char const *)arg5,arg6,arg7);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_remove_from_revision_control2(int argc, VALUE *argv, VALUE self) {
  svn_wc_context_t *arg1 = 0 ;
  char *arg2 = 0 ;
  svn_boolean_t arg3 ;
  svn_boolean_t arg4 ;
  svn_cancel_func_t arg5 = 0 ;
  void *arg6 = 0 ;
  apr_pool_t *arg7 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg7);
    _global_pool = arg7;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 5) || (argc > 6)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 5)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_context_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_context_t *","svn_wc_remove_from_revision_control2", 1, argv[0] )); 
  }
  arg1 = (svn_wc_context_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_wc_remove_from_revision_control2", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  arg3 = RTEST(argv[2]);
  arg4 = RTEST(argv[3]);
  {
    arg5 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg6 = (void *)svn_swig_rb_make_baton(argv[4], _global_svn_swig_rb_pool);
  }
  if (argc > 5) {
    
  }
  {
    result = (svn_error_t *)svn_wc_remove_from_revision_control2(arg1,(char const *)arg2,arg3,arg4,arg5,arg6,arg7);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg6);
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_remove_from_revision_control(int argc, VALUE *argv, VALUE self) {
  svn_wc_adm_access_t *arg1 = 0 ;
  char *arg2 = 0 ;
  svn_boolean_t arg3 ;
  svn_boolean_t arg4 ;
  svn_cancel_func_t arg5 = 0 ;
  void *arg6 = 0 ;
  apr_pool_t *arg7 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg7);
    _global_pool = arg7;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 5) || (argc > 6)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 5)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_remove_from_revision_control", 1, argv[0] )); 
  }
  arg1 = (svn_wc_adm_access_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_wc_remove_from_revision_control", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  arg3 = RTEST(argv[2]);
  arg4 = RTEST(argv[3]);
  {
    arg5 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg6 = (void *)svn_swig_rb_make_baton(argv[4], _global_svn_swig_rb_pool);
  }
  if (argc > 5) {
    
  }
  {
    result = (svn_error_t *)svn_wc_remove_from_revision_control(arg1,(char const *)arg2,arg3,arg4,arg5,arg6,arg7);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg6);
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_resolved_conflict5(int argc, VALUE *argv, VALUE self) {
  svn_wc_context_t *arg1 = 0 ;
  char *arg2 = 0 ;
  svn_depth_t arg3 ;
  svn_boolean_t arg4 ;
  char *arg5 = 0 ;
  svn_boolean_t arg6 ;
  svn_wc_conflict_choice_t arg7 ;
  svn_cancel_func_t arg8 = 0 ;
  void *arg9 = 0 ;
  svn_wc_notify_func2_t arg10 = 0 ;
  void *arg11 = 0 ;
  apr_pool_t *arg12 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res5 ;
  char *buf5 = 0 ;
  int alloc5 = 0 ;
  int val7 ;
  int ecode7 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg12);
    _global_pool = arg12;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 9) || (argc > 10)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 9)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_context_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_context_t *","svn_wc_resolved_conflict5", 1, argv[0] )); 
  }
  arg1 = (svn_wc_context_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_wc_resolved_conflict5", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  {
    arg3 = svn_swig_rb_to_depth(argv[2]);
  }
  arg4 = RTEST(argv[3]);
  res5 = SWIG_AsCharPtrAndSize(argv[4], &buf5, NULL, &alloc5);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "char const *","svn_wc_resolved_conflict5", 5, argv[4] ));
  }
  arg5 = (char *)(buf5);
  arg6 = RTEST(argv[5]);
  ecode7 = SWIG_AsVal_int(argv[6], &val7);
  if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), Ruby_Format_TypeError( "", "svn_wc_conflict_choice_t","svn_wc_resolved_conflict5", 7, argv[6] ));
  } 
  arg7 = (svn_wc_conflict_choice_t)(val7);
  {
    arg8 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg9 = (void *)svn_swig_rb_make_baton(argv[7], _global_svn_swig_rb_pool);
  }
  {
    arg10 = (svn_wc_notify_func2_t) svn_swig_rb_notify_func2;
    arg11 = (void *)svn_swig_rb_make_baton(argv[8], _global_svn_swig_rb_pool);
  }
  if (argc > 9) {
    
  }
  {
    result = (svn_error_t *)svn_wc_resolved_conflict5(arg1,(char const *)arg2,arg3,arg4,(char const *)arg5,arg6,arg7,arg8,arg9,arg10,arg11,arg12);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg9);
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_resolved_conflict4(int argc, VALUE *argv, VALUE self) {
  char *arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  svn_boolean_t arg3 ;
  svn_boolean_t arg4 ;
  svn_boolean_t arg5 ;
  svn_depth_t arg6 ;
  svn_wc_conflict_choice_t arg7 ;
  svn_wc_notify_func2_t arg8 = 0 ;
  void *arg9 = 0 ;
  svn_cancel_func_t arg10 = 0 ;
  void *arg11 = 0 ;
  apr_pool_t *arg12 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int val7 ;
  int ecode7 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg12);
    _global_pool = arg12;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 9) || (argc > 10)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 9)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_wc_resolved_conflict4", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_resolved_conflict4", 2, argv[1] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  arg3 = RTEST(argv[2]);
  arg4 = RTEST(argv[3]);
  arg5 = RTEST(argv[4]);
  {
    arg6 = svn_swig_rb_to_depth(argv[5]);
  }
  ecode7 = SWIG_AsVal_int(argv[6], &val7);
  if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), Ruby_Format_TypeError( "", "svn_wc_conflict_choice_t","svn_wc_resolved_conflict4", 7, argv[6] ));
  } 
  arg7 = (svn_wc_conflict_choice_t)(val7);
  {
    arg8 = (svn_wc_notify_func2_t) svn_swig_rb_notify_func2;
    arg9 = (void *)svn_swig_rb_make_baton(argv[7], _global_svn_swig_rb_pool);
  }
  {
    arg10 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg11 = (void *)svn_swig_rb_make_baton(argv[8], _global_svn_swig_rb_pool);
  }
  if (argc > 9) {
    
  }
  {
    result = (svn_error_t *)svn_wc_resolved_conflict4((char const *)arg1,arg2,arg3,arg4,arg5,arg6,arg7,arg8,arg9,arg10,arg11,arg12);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg11);
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_resolved_conflict3(int argc, VALUE *argv, VALUE self) {
  char *arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  svn_boolean_t arg3 ;
  svn_boolean_t arg4 ;
  svn_depth_t arg5 ;
  svn_wc_conflict_choice_t arg6 ;
  svn_wc_notify_func2_t arg7 = 0 ;
  void *arg8 = 0 ;
  svn_cancel_func_t arg9 = 0 ;
  void *arg10 = 0 ;
  apr_pool_t *arg11 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int val6 ;
  int ecode6 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg11);
    _global_pool = arg11;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 8) || (argc > 9)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 8)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_wc_resolved_conflict3", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_resolved_conflict3", 2, argv[1] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  arg3 = RTEST(argv[2]);
  arg4 = RTEST(argv[3]);
  {
    arg5 = svn_swig_rb_to_depth(argv[4]);
  }
  ecode6 = SWIG_AsVal_int(argv[5], &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), Ruby_Format_TypeError( "", "svn_wc_conflict_choice_t","svn_wc_resolved_conflict3", 6, argv[5] ));
  } 
  arg6 = (svn_wc_conflict_choice_t)(val6);
  {
    arg7 = (svn_wc_notify_func2_t) svn_swig_rb_notify_func2;
    arg8 = (void *)svn_swig_rb_make_baton(argv[6], _global_svn_swig_rb_pool);
  }
  {
    arg9 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg10 = (void *)svn_swig_rb_make_baton(argv[7], _global_svn_swig_rb_pool);
  }
  if (argc > 8) {
    
  }
  {
    result = (svn_error_t *)svn_wc_resolved_conflict3((char const *)arg1,arg2,arg3,arg4,arg5,arg6,arg7,arg8,arg9,arg10,arg11);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg10);
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_resolved_conflict2(int argc, VALUE *argv, VALUE self) {
  char *arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  svn_boolean_t arg3 ;
  svn_boolean_t arg4 ;
  svn_boolean_t arg5 ;
  svn_wc_notify_func2_t arg6 = 0 ;
  void *arg7 = 0 ;
  svn_cancel_func_t arg8 = 0 ;
  void *arg9 = 0 ;
  apr_pool_t *arg10 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg10);
    _global_pool = arg10;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 7) || (argc > 8)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 7)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_wc_resolved_conflict2", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_resolved_conflict2", 2, argv[1] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  arg3 = RTEST(argv[2]);
  arg4 = RTEST(argv[3]);
  arg5 = RTEST(argv[4]);
  {
    arg6 = (svn_wc_notify_func2_t) svn_swig_rb_notify_func2;
    arg7 = (void *)svn_swig_rb_make_baton(argv[5], _global_svn_swig_rb_pool);
  }
  {
    arg8 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg9 = (void *)svn_swig_rb_make_baton(argv[6], _global_svn_swig_rb_pool);
  }
  if (argc > 7) {
    
  }
  {
    result = (svn_error_t *)svn_wc_resolved_conflict2((char const *)arg1,arg2,arg3,arg4,arg5,arg6,arg7,arg8,arg9,arg10);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg9);
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_resolved_conflict(int argc, VALUE *argv, VALUE self) {
  char *arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  svn_boolean_t arg3 ;
  svn_boolean_t arg4 ;
  svn_boolean_t arg5 ;
  svn_wc_notify_func_t arg6 = 0 ;
  void *arg7 = 0 ;
  apr_pool_t *arg8 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res7 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg8);
    _global_pool = arg8;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 7) || (argc > 8)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 7)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_wc_resolved_conflict", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_resolved_conflict", 2, argv[1] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  arg3 = RTEST(argv[2]);
  arg4 = RTEST(argv[3]);
  arg5 = RTEST(argv[4]);
  {
    int res = SWIG_ConvertFunctionPtr(argv[5], (void**)(&arg6), SWIGTYPE_p_f_p_void_p_q_const__char_enum_svn_wc_notify_action_t_enum_svn_node_kind_t_p_q_const__char_enum_svn_wc_notify_state_t_enum_svn_wc_notify_state_t_long__void);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_wc_notify_func_t","svn_wc_resolved_conflict", 6, argv[5] )); 
    }
  }
  res7 = SWIG_ConvertPtr(argv[6],SWIG_as_voidptrptr(&arg7), 0, 0);
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), Ruby_Format_TypeError( "", "void *","svn_wc_resolved_conflict", 7, argv[6] )); 
  }
  if (argc > 7) {
    
  }
  {
    result = (svn_error_t *)svn_wc_resolved_conflict((char const *)arg1,arg2,arg3,arg4,arg5,arg6,arg7,arg8);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_queue_committed4(int argc, VALUE *argv, VALUE self) {
  svn_wc_committed_queue_t *arg1 = 0 ;
  svn_wc_context_t *arg2 = 0 ;
  char *arg3 = 0 ;
  svn_boolean_t arg4 ;
  svn_boolean_t arg5 ;
  apr_array_header_t *arg6 = 0 ;
  svn_boolean_t arg7 ;
  svn_boolean_t arg8 ;
  svn_checksum_t *arg9 = 0 ;
  apr_pool_t *arg10 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  void *argp9 = 0 ;
  int res9 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg10);
    _global_pool = arg10;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 9) || (argc > 10)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 9)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_committed_queue_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_committed_queue_t *","svn_wc_queue_committed4", 1, argv[0] )); 
  }
  arg1 = (svn_wc_committed_queue_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_context_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_context_t *","svn_wc_queue_committed4", 2, argv[1] )); 
  }
  arg2 = (svn_wc_context_t *)(argp2);
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_wc_queue_committed4", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  arg4 = RTEST(argv[3]);
  arg5 = RTEST(argv[4]);
  {
    VALUE rb_pool;
    apr_pool_t *pool;
    
    svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &pool);
    
    arg6 = svn_swig_rb_to_apr_array_prop(argv[5], pool);
  }
  arg7 = RTEST(argv[6]);
  arg8 = RTEST(argv[7]);
  res9 = SWIG_ConvertPtr(argv[8], &argp9,SWIGTYPE_p_svn_checksum_t, 0 |  0 );
  if (!SWIG_IsOK(res9)) {
    SWIG_exception_fail(SWIG_ArgError(res9), Ruby_Format_TypeError( "", "svn_checksum_t const *","svn_wc_queue_committed4", 9, argv[8] )); 
  }
  arg9 = (svn_checksum_t *)(argp9);
  if (argc > 9) {
    
  }
  {
    result = (svn_error_t *)svn_wc_queue_committed4(arg1,arg2,(char const *)arg3,arg4,arg5,(apr_array_header_t const *)arg6,arg7,arg8,(struct svn_checksum_t const *)arg9,arg10);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_queue_committed3(int argc, VALUE *argv, VALUE self) {
  svn_wc_committed_queue_t *arg1 = 0 ;
  svn_wc_context_t *arg2 = 0 ;
  char *arg3 = 0 ;
  svn_boolean_t arg4 ;
  apr_array_header_t *arg5 = 0 ;
  svn_boolean_t arg6 ;
  svn_boolean_t arg7 ;
  svn_checksum_t *arg8 = 0 ;
  apr_pool_t *arg9 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg9);
    _global_pool = arg9;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 8) || (argc > 9)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 8)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_committed_queue_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_committed_queue_t *","svn_wc_queue_committed3", 1, argv[0] )); 
  }
  arg1 = (svn_wc_committed_queue_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_context_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_context_t *","svn_wc_queue_committed3", 2, argv[1] )); 
  }
  arg2 = (svn_wc_context_t *)(argp2);
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_wc_queue_committed3", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  arg4 = RTEST(argv[3]);
  {
    VALUE rb_pool;
    apr_pool_t *pool;
    
    svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &pool);
    
    arg5 = svn_swig_rb_to_apr_array_prop(argv[4], pool);
  }
  arg6 = RTEST(argv[5]);
  arg7 = RTEST(argv[6]);
  res8 = SWIG_ConvertPtr(argv[7], &argp8,SWIGTYPE_p_svn_checksum_t, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), Ruby_Format_TypeError( "", "svn_checksum_t const *","svn_wc_queue_committed3", 8, argv[7] )); 
  }
  arg8 = (svn_checksum_t *)(argp8);
  if (argc > 8) {
    
  }
  {
    result = (svn_error_t *)svn_wc_queue_committed3(arg1,arg2,(char const *)arg3,arg4,(apr_array_header_t const *)arg5,arg6,arg7,(struct svn_checksum_t const *)arg8,arg9);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_queue_committed2(int argc, VALUE *argv, VALUE self) {
  svn_wc_committed_queue_t *arg1 = 0 ;
  char *arg2 = 0 ;
  svn_wc_adm_access_t *arg3 = 0 ;
  svn_boolean_t arg4 ;
  apr_array_header_t *arg5 = 0 ;
  svn_boolean_t arg6 ;
  svn_boolean_t arg7 ;
  svn_checksum_t *arg8 = 0 ;
  apr_pool_t *arg9 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  void *argp8 = 0 ;
  int res8 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg9);
    _global_pool = arg9;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 8) || (argc > 9)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 8)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_committed_queue_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_committed_queue_t *","svn_wc_queue_committed2", 1, argv[0] )); 
  }
  arg1 = (svn_wc_committed_queue_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_wc_queue_committed2", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_ConvertPtr(argv[2], &argp3,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_queue_committed2", 3, argv[2] )); 
  }
  arg3 = (svn_wc_adm_access_t *)(argp3);
  arg4 = RTEST(argv[3]);
  {
    VALUE rb_pool;
    apr_pool_t *pool;
    
    svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &pool);
    
    arg5 = svn_swig_rb_to_apr_array_prop(argv[4], pool);
  }
  arg6 = RTEST(argv[5]);
  arg7 = RTEST(argv[6]);
  res8 = SWIG_ConvertPtr(argv[7], &argp8,SWIGTYPE_p_svn_checksum_t, 0 |  0 );
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), Ruby_Format_TypeError( "", "svn_checksum_t const *","svn_wc_queue_committed2", 8, argv[7] )); 
  }
  arg8 = (svn_checksum_t *)(argp8);
  if (argc > 8) {
    
  }
  {
    result = (svn_error_t *)svn_wc_queue_committed2(arg1,(char const *)arg2,arg3,arg4,(apr_array_header_t const *)arg5,arg6,arg7,(struct svn_checksum_t const *)arg8,arg9);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_queue_committed(int argc, VALUE *argv, VALUE self) {
  svn_wc_committed_queue_t **arg1 = 0 ;
  char *arg2 = 0 ;
  svn_wc_adm_access_t *arg3 = 0 ;
  svn_boolean_t arg4 ;
  apr_array_header_t *arg5 = 0 ;
  svn_boolean_t arg6 ;
  svn_boolean_t arg7 ;
  unsigned char *arg8 = 0 ;
  apr_pool_t *arg9 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *tempp1 = NULL ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg9);
    _global_pool = arg9;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 8) || (argc > 9)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 8)",argc); SWIG_fail;
  }
  {
    SWIG_ConvertPtr(argv[0], &tempp1, SWIGTYPE_p_svn_wc_committed_queue_t, 0);
    arg1 = (svn_wc_committed_queue_t **)&tempp1;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_wc_queue_committed", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_ConvertPtr(argv[2], &argp3,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_queue_committed", 3, argv[2] )); 
  }
  arg3 = (svn_wc_adm_access_t *)(argp3);
  arg4 = RTEST(argv[3]);
  {
    VALUE rb_pool;
    apr_pool_t *pool;
    
    svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &pool);
    
    arg5 = svn_swig_rb_to_apr_array_prop(argv[4], pool);
  }
  arg6 = RTEST(argv[5]);
  arg7 = RTEST(argv[6]);
  
  if (argc > 8) {
    
  }
  {
    result = (svn_error_t *)svn_wc_queue_committed(arg1,(char const *)arg2,arg3,arg4,(apr_array_header_t const *)arg5,arg6,arg7,(unsigned char const *)arg8,arg9);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, argv[0], 0);
  }
  
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_process_committed_queue2(int argc, VALUE *argv, VALUE self) {
  svn_wc_committed_queue_t *arg1 = 0 ;
  svn_wc_context_t *arg2 = 0 ;
  svn_revnum_t arg3 ;
  char *arg4 = 0 ;
  char *arg5 = 0 ;
  svn_cancel_func_t arg6 = 0 ;
  void *arg7 = 0 ;
  apr_pool_t *arg8 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  long val3 ;
  int ecode3 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg8);
    _global_pool = arg8;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 6) || (argc > 7)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 6)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_committed_queue_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_committed_queue_t *","svn_wc_process_committed_queue2", 1, argv[0] )); 
  }
  arg1 = (svn_wc_committed_queue_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_context_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_context_t *","svn_wc_process_committed_queue2", 2, argv[1] )); 
  }
  arg2 = (svn_wc_context_t *)(argp2);
  ecode3 = SWIG_AsVal_long(argv[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), Ruby_Format_TypeError( "", "svn_revnum_t","svn_wc_process_committed_queue2", 3, argv[2] ));
  } 
  arg3 = (svn_revnum_t)(val3);
  {
    if (NIL_P(argv[3])) {
      arg4 = NULL;
    } else {
      arg4 = StringValuePtr(argv[3]);
    }
  }
  {
    if (NIL_P(argv[4])) {
      arg5 = NULL;
    } else {
      arg5 = StringValuePtr(argv[4]);
    }
  }
  {
    arg6 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg7 = (void *)svn_swig_rb_make_baton(argv[5], _global_svn_swig_rb_pool);
  }
  if (argc > 6) {
    
  }
  {
    result = (svn_error_t *)svn_wc_process_committed_queue2(arg1,arg2,arg3,(char const *)arg4,(char const *)arg5,arg6,arg7,arg8);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg7);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_process_committed_queue(int argc, VALUE *argv, VALUE self) {
  svn_wc_committed_queue_t *arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  svn_revnum_t arg3 ;
  char *arg4 = 0 ;
  char *arg5 = 0 ;
  apr_pool_t *arg6 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  long val3 ;
  int ecode3 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 5) || (argc > 6)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 5)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_committed_queue_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_committed_queue_t *","svn_wc_process_committed_queue", 1, argv[0] )); 
  }
  arg1 = (svn_wc_committed_queue_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_process_committed_queue", 2, argv[1] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  ecode3 = SWIG_AsVal_long(argv[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), Ruby_Format_TypeError( "", "svn_revnum_t","svn_wc_process_committed_queue", 3, argv[2] ));
  } 
  arg3 = (svn_revnum_t)(val3);
  {
    if (NIL_P(argv[3])) {
      arg4 = NULL;
    } else {
      arg4 = StringValuePtr(argv[3]);
    }
  }
  {
    if (NIL_P(argv[4])) {
      arg5 = NULL;
    } else {
      arg5 = StringValuePtr(argv[4]);
    }
  }
  if (argc > 5) {
    
  }
  {
    result = (svn_error_t *)svn_wc_process_committed_queue(arg1,arg2,arg3,(char const *)arg4,(char const *)arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_process_committed4(int argc, VALUE *argv, VALUE self) {
  char *arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  svn_boolean_t arg3 ;
  svn_revnum_t arg4 ;
  char *arg5 = 0 ;
  char *arg6 = 0 ;
  apr_array_header_t *arg7 = 0 ;
  svn_boolean_t arg8 ;
  svn_boolean_t arg9 ;
  unsigned char *arg10 = 0 ;
  apr_pool_t *arg11 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  long val4 ;
  int ecode4 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg11);
    _global_pool = arg11;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 10) || (argc > 11)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 10)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_wc_process_committed4", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_process_committed4", 2, argv[1] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  arg3 = RTEST(argv[2]);
  ecode4 = SWIG_AsVal_long(argv[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), Ruby_Format_TypeError( "", "svn_revnum_t","svn_wc_process_committed4", 4, argv[3] ));
  } 
  arg4 = (svn_revnum_t)(val4);
  {
    if (NIL_P(argv[4])) {
      arg5 = NULL;
    } else {
      arg5 = StringValuePtr(argv[4]);
    }
  }
  {
    if (NIL_P(argv[5])) {
      arg6 = NULL;
    } else {
      arg6 = StringValuePtr(argv[5]);
    }
  }
  {
    VALUE rb_pool;
    apr_pool_t *pool;
    
    svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &pool);
    
    arg7 = svn_swig_rb_to_apr_array_prop(argv[6], pool);
  }
  arg8 = RTEST(argv[7]);
  arg9 = RTEST(argv[8]);
  
  if (argc > 10) {
    
  }
  {
    result = (svn_error_t *)svn_wc_process_committed4((char const *)arg1,arg2,arg3,arg4,(char const *)arg5,(char const *)arg6,(apr_array_header_t const *)arg7,arg8,arg9,(unsigned char const *)arg10,arg11);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_process_committed3(int argc, VALUE *argv, VALUE self) {
  char *arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  svn_boolean_t arg3 ;
  svn_revnum_t arg4 ;
  char *arg5 = 0 ;
  char *arg6 = 0 ;
  apr_array_header_t *arg7 = 0 ;
  svn_boolean_t arg8 ;
  unsigned char *arg9 = 0 ;
  apr_pool_t *arg10 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  long val4 ;
  int ecode4 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg10);
    _global_pool = arg10;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 9) || (argc > 10)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 9)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_wc_process_committed3", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_process_committed3", 2, argv[1] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  arg3 = RTEST(argv[2]);
  ecode4 = SWIG_AsVal_long(argv[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), Ruby_Format_TypeError( "", "svn_revnum_t","svn_wc_process_committed3", 4, argv[3] ));
  } 
  arg4 = (svn_revnum_t)(val4);
  {
    if (NIL_P(argv[4])) {
      arg5 = NULL;
    } else {
      arg5 = StringValuePtr(argv[4]);
    }
  }
  {
    if (NIL_P(argv[5])) {
      arg6 = NULL;
    } else {
      arg6 = StringValuePtr(argv[5]);
    }
  }
  {
    VALUE rb_pool;
    apr_pool_t *pool;
    
    svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &pool);
    
    arg7 = svn_swig_rb_to_apr_array_prop(argv[6], pool);
  }
  arg8 = RTEST(argv[7]);
  
  if (argc > 9) {
    
  }
  {
    result = (svn_error_t *)svn_wc_process_committed3((char const *)arg1,arg2,arg3,arg4,(char const *)arg5,(char const *)arg6,(apr_array_header_t const *)arg7,arg8,(unsigned char const *)arg9,arg10);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_process_committed2(int argc, VALUE *argv, VALUE self) {
  char *arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  svn_boolean_t arg3 ;
  svn_revnum_t arg4 ;
  char *arg5 = 0 ;
  char *arg6 = 0 ;
  apr_array_header_t *arg7 = 0 ;
  svn_boolean_t arg8 ;
  apr_pool_t *arg9 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  long val4 ;
  int ecode4 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg9);
    _global_pool = arg9;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 8) || (argc > 9)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 8)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_wc_process_committed2", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_process_committed2", 2, argv[1] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  arg3 = RTEST(argv[2]);
  ecode4 = SWIG_AsVal_long(argv[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), Ruby_Format_TypeError( "", "svn_revnum_t","svn_wc_process_committed2", 4, argv[3] ));
  } 
  arg4 = (svn_revnum_t)(val4);
  {
    if (NIL_P(argv[4])) {
      arg5 = NULL;
    } else {
      arg5 = StringValuePtr(argv[4]);
    }
  }
  {
    if (NIL_P(argv[5])) {
      arg6 = NULL;
    } else {
      arg6 = StringValuePtr(argv[5]);
    }
  }
  {
    VALUE rb_pool;
    apr_pool_t *pool;
    
    svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &pool);
    
    arg7 = svn_swig_rb_to_apr_array_prop(argv[6], pool);
  }
  arg8 = RTEST(argv[7]);
  if (argc > 8) {
    
  }
  {
    result = (svn_error_t *)svn_wc_process_committed2((char const *)arg1,arg2,arg3,arg4,(char const *)arg5,(char const *)arg6,(apr_array_header_t const *)arg7,arg8,arg9);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_process_committed(int argc, VALUE *argv, VALUE self) {
  char *arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  svn_boolean_t arg3 ;
  svn_revnum_t arg4 ;
  char *arg5 = 0 ;
  char *arg6 = 0 ;
  apr_array_header_t *arg7 = 0 ;
  apr_pool_t *arg8 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  long val4 ;
  int ecode4 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg8);
    _global_pool = arg8;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 7) || (argc > 8)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 7)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_wc_process_committed", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_process_committed", 2, argv[1] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  arg3 = RTEST(argv[2]);
  ecode4 = SWIG_AsVal_long(argv[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), Ruby_Format_TypeError( "", "svn_revnum_t","svn_wc_process_committed", 4, argv[3] ));
  } 
  arg4 = (svn_revnum_t)(val4);
  {
    if (NIL_P(argv[4])) {
      arg5 = NULL;
    } else {
      arg5 = StringValuePtr(argv[4]);
    }
  }
  {
    if (NIL_P(argv[5])) {
      arg6 = NULL;
    } else {
      arg6 = StringValuePtr(argv[5]);
    }
  }
  {
    VALUE rb_pool;
    apr_pool_t *pool;
    
    svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &pool);
    
    arg7 = svn_swig_rb_to_apr_array_prop(argv[6], pool);
  }
  if (argc > 7) {
    
  }
  {
    result = (svn_error_t *)svn_wc_process_committed((char const *)arg1,arg2,arg3,arg4,(char const *)arg5,(char const *)arg6,(apr_array_header_t const *)arg7,arg8);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_crawl_revisions5(int argc, VALUE *argv, VALUE self) {
  svn_wc_context_t *arg1 = 0 ;
  char *arg2 = 0 ;
  svn_ra_reporter3_t *arg3 = 0 ;
  void *arg4 = 0 ;
  svn_boolean_t arg5 ;
  svn_depth_t arg6 ;
  svn_boolean_t arg7 ;
  svn_boolean_t arg8 ;
  svn_boolean_t arg9 ;
  svn_cancel_func_t arg10 = 0 ;
  void *arg11 = 0 ;
  svn_wc_notify_func2_t arg12 = 0 ;
  void *arg13 = 0 ;
  apr_pool_t *arg14 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg14);
    _global_pool = arg14;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 10) || (argc > 11)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 10)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_context_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_context_t *","svn_wc_crawl_revisions5", 1, argv[0] )); 
  }
  arg1 = (svn_wc_context_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_wc_crawl_revisions5", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  {
    arg3 = (svn_ra_reporter3_t *) svn_swig_rb_get_ra_reporter3();
    arg4 = (void *)svn_swig_rb_make_baton(argv[2], _global_svn_swig_rb_pool);
  }
  arg5 = RTEST(argv[3]);
  {
    arg6 = svn_swig_rb_to_depth(argv[4]);
  }
  arg7 = RTEST(argv[5]);
  arg8 = RTEST(argv[6]);
  arg9 = RTEST(argv[7]);
  {
    arg10 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg11 = (void *)svn_swig_rb_make_baton(argv[8], _global_svn_swig_rb_pool);
  }
  {
    arg12 = (svn_wc_notify_func2_t) svn_swig_rb_notify_func2;
    arg13 = (void *)svn_swig_rb_make_baton(argv[9], _global_svn_swig_rb_pool);
  }
  if (argc > 10) {
    
  }
  {
    result = (svn_error_t *)svn_wc_crawl_revisions5(arg1,(char const *)arg2,(struct svn_ra_reporter3_t const *)arg3,arg4,arg5,arg6,arg7,arg8,arg9,arg10,arg11,arg12,arg13,arg14);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg11);
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_crawl_revisions4(int argc, VALUE *argv, VALUE self) {
  char *arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  svn_ra_reporter3_t *arg3 = 0 ;
  void *arg4 = 0 ;
  svn_boolean_t arg5 ;
  svn_depth_t arg6 ;
  svn_boolean_t arg7 ;
  svn_boolean_t arg8 ;
  svn_boolean_t arg9 ;
  svn_wc_notify_func2_t arg10 = 0 ;
  void *arg11 = 0 ;
  svn_wc_traversal_info_t *arg12 = 0 ;
  apr_pool_t *arg13 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp12 = 0 ;
  int res12 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg13);
    _global_pool = arg13;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 10) || (argc > 11)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 10)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_wc_crawl_revisions4", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_crawl_revisions4", 2, argv[1] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  {
    arg3 = (svn_ra_reporter3_t *) svn_swig_rb_get_ra_reporter3();
    arg4 = (void *)svn_swig_rb_make_baton(argv[2], _global_svn_swig_rb_pool);
  }
  arg5 = RTEST(argv[3]);
  {
    arg6 = svn_swig_rb_to_depth(argv[4]);
  }
  arg7 = RTEST(argv[5]);
  arg8 = RTEST(argv[6]);
  arg9 = RTEST(argv[7]);
  {
    arg10 = (svn_wc_notify_func2_t) svn_swig_rb_notify_func2;
    arg11 = (void *)svn_swig_rb_make_baton(argv[8], _global_svn_swig_rb_pool);
  }
  res12 = SWIG_ConvertPtr(argv[9], &argp12,SWIGTYPE_p_svn_wc_traversal_info_t, 0 |  0 );
  if (!SWIG_IsOK(res12)) {
    SWIG_exception_fail(SWIG_ArgError(res12), Ruby_Format_TypeError( "", "svn_wc_traversal_info_t *","svn_wc_crawl_revisions4", 12, argv[9] )); 
  }
  arg12 = (svn_wc_traversal_info_t *)(argp12);
  if (argc > 10) {
    
  }
  {
    result = (svn_error_t *)svn_wc_crawl_revisions4((char const *)arg1,arg2,(struct svn_ra_reporter3_t const *)arg3,arg4,arg5,arg6,arg7,arg8,arg9,arg10,arg11,arg12,arg13);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_crawl_revisions3(int argc, VALUE *argv, VALUE self) {
  char *arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  svn_ra_reporter3_t *arg3 = 0 ;
  void *arg4 = 0 ;
  svn_boolean_t arg5 ;
  svn_depth_t arg6 ;
  svn_boolean_t arg7 ;
  svn_boolean_t arg8 ;
  svn_wc_notify_func2_t arg9 = 0 ;
  void *arg10 = 0 ;
  svn_wc_traversal_info_t *arg11 = 0 ;
  apr_pool_t *arg12 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp11 = 0 ;
  int res11 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg12);
    _global_pool = arg12;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 9) || (argc > 10)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 9)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_wc_crawl_revisions3", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_crawl_revisions3", 2, argv[1] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  {
    arg3 = (svn_ra_reporter3_t *) svn_swig_rb_get_ra_reporter3();
    arg4 = (void *)svn_swig_rb_make_baton(argv[2], _global_svn_swig_rb_pool);
  }
  arg5 = RTEST(argv[3]);
  {
    arg6 = svn_swig_rb_to_depth(argv[4]);
  }
  arg7 = RTEST(argv[5]);
  arg8 = RTEST(argv[6]);
  {
    arg9 = (svn_wc_notify_func2_t) svn_swig_rb_notify_func2;
    arg10 = (void *)svn_swig_rb_make_baton(argv[7], _global_svn_swig_rb_pool);
  }
  res11 = SWIG_ConvertPtr(argv[8], &argp11,SWIGTYPE_p_svn_wc_traversal_info_t, 0 |  0 );
  if (!SWIG_IsOK(res11)) {
    SWIG_exception_fail(SWIG_ArgError(res11), Ruby_Format_TypeError( "", "svn_wc_traversal_info_t *","svn_wc_crawl_revisions3", 11, argv[8] )); 
  }
  arg11 = (svn_wc_traversal_info_t *)(argp11);
  if (argc > 9) {
    
  }
  {
    result = (svn_error_t *)svn_wc_crawl_revisions3((char const *)arg1,arg2,(struct svn_ra_reporter3_t const *)arg3,arg4,arg5,arg6,arg7,arg8,arg9,arg10,arg11,arg12);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_crawl_revisions2(int argc, VALUE *argv, VALUE self) {
  char *arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  svn_ra_reporter2_t *arg3 = 0 ;
  void *arg4 = 0 ;
  svn_boolean_t arg5 ;
  svn_boolean_t arg6 ;
  svn_boolean_t arg7 ;
  svn_wc_notify_func2_t arg8 = 0 ;
  void *arg9 = 0 ;
  svn_wc_traversal_info_t *arg10 = 0 ;
  apr_pool_t *arg11 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  int res4 ;
  void *argp10 = 0 ;
  int res10 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg11);
    _global_pool = arg11;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 9) || (argc > 10)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 9)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_wc_crawl_revisions2", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_crawl_revisions2", 2, argv[1] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  res3 = SWIG_ConvertPtr(argv[2], &argp3,SWIGTYPE_p_svn_ra_reporter2_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_ra_reporter2_t const *","svn_wc_crawl_revisions2", 3, argv[2] )); 
  }
  arg3 = (svn_ra_reporter2_t *)(argp3);
  res4 = SWIG_ConvertPtr(argv[3],SWIG_as_voidptrptr(&arg4), 0, 0);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "void *","svn_wc_crawl_revisions2", 4, argv[3] )); 
  }
  arg5 = RTEST(argv[4]);
  arg6 = RTEST(argv[5]);
  arg7 = RTEST(argv[6]);
  {
    arg8 = (svn_wc_notify_func2_t) svn_swig_rb_notify_func2;
    arg9 = (void *)svn_swig_rb_make_baton(argv[7], _global_svn_swig_rb_pool);
  }
  res10 = SWIG_ConvertPtr(argv[8], &argp10,SWIGTYPE_p_svn_wc_traversal_info_t, 0 |  0 );
  if (!SWIG_IsOK(res10)) {
    SWIG_exception_fail(SWIG_ArgError(res10), Ruby_Format_TypeError( "", "svn_wc_traversal_info_t *","svn_wc_crawl_revisions2", 10, argv[8] )); 
  }
  arg10 = (svn_wc_traversal_info_t *)(argp10);
  if (argc > 9) {
    
  }
  {
    result = (svn_error_t *)svn_wc_crawl_revisions2((char const *)arg1,arg2,(struct svn_ra_reporter2_t const *)arg3,arg4,arg5,arg6,arg7,arg8,arg9,arg10,arg11);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_crawl_revisions(int argc, VALUE *argv, VALUE self) {
  char *arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  svn_ra_reporter_t *arg3 = 0 ;
  void *arg4 = 0 ;
  svn_boolean_t arg5 ;
  svn_boolean_t arg6 ;
  svn_boolean_t arg7 ;
  svn_wc_notify_func_t arg8 = 0 ;
  void *arg9 = 0 ;
  svn_wc_traversal_info_t *arg10 = 0 ;
  apr_pool_t *arg11 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  int res4 ;
  int res9 ;
  void *argp10 = 0 ;
  int res10 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg11);
    _global_pool = arg11;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 10) || (argc > 11)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 10)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_wc_crawl_revisions", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_crawl_revisions", 2, argv[1] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  res3 = SWIG_ConvertPtr(argv[2], &argp3,SWIGTYPE_p_svn_ra_reporter_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_ra_reporter_t const *","svn_wc_crawl_revisions", 3, argv[2] )); 
  }
  arg3 = (svn_ra_reporter_t *)(argp3);
  res4 = SWIG_ConvertPtr(argv[3],SWIG_as_voidptrptr(&arg4), 0, 0);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "void *","svn_wc_crawl_revisions", 4, argv[3] )); 
  }
  arg5 = RTEST(argv[4]);
  arg6 = RTEST(argv[5]);
  arg7 = RTEST(argv[6]);
  {
    int res = SWIG_ConvertFunctionPtr(argv[7], (void**)(&arg8), SWIGTYPE_p_f_p_void_p_q_const__char_enum_svn_wc_notify_action_t_enum_svn_node_kind_t_p_q_const__char_enum_svn_wc_notify_state_t_enum_svn_wc_notify_state_t_long__void);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_wc_notify_func_t","svn_wc_crawl_revisions", 8, argv[7] )); 
    }
  }
  res9 = SWIG_ConvertPtr(argv[8],SWIG_as_voidptrptr(&arg9), 0, 0);
  if (!SWIG_IsOK(res9)) {
    SWIG_exception_fail(SWIG_ArgError(res9), Ruby_Format_TypeError( "", "void *","svn_wc_crawl_revisions", 9, argv[8] )); 
  }
  res10 = SWIG_ConvertPtr(argv[9], &argp10,SWIGTYPE_p_svn_wc_traversal_info_t, 0 |  0 );
  if (!SWIG_IsOK(res10)) {
    SWIG_exception_fail(SWIG_ArgError(res10), Ruby_Format_TypeError( "", "svn_wc_traversal_info_t *","svn_wc_crawl_revisions", 10, argv[9] )); 
  }
  arg10 = (svn_wc_traversal_info_t *)(argp10);
  if (argc > 10) {
    
  }
  {
    result = (svn_error_t *)svn_wc_crawl_revisions((char const *)arg1,arg2,(struct svn_ra_reporter_t const *)arg3,arg4,arg5,arg6,arg7,arg8,arg9,arg10,arg11);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_check_root(int argc, VALUE *argv, VALUE self) {
  svn_boolean_t *arg1 = 0 ;
  svn_boolean_t *arg2 = 0 ;
  svn_node_kind_t *arg3 = 0 ;
  svn_wc_context_t *arg4 = 0 ;
  char *arg5 = 0 ;
  apr_pool_t *arg6 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_boolean_t temp1 ;
  svn_boolean_t temp2 ;
  svn_node_kind_t temp3 ;
  int res3 = SWIG_TMPOBJ ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  int res5 ;
  char *buf5 = 0 ;
  int alloc5 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  arg2 = &temp2;
  arg3 = &temp3;
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res4 = SWIG_ConvertPtr(argv[0], &argp4,SWIGTYPE_p_svn_wc_context_t, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "svn_wc_context_t *","svn_wc_check_root", 4, argv[0] )); 
  }
  arg4 = (svn_wc_context_t *)(argp4);
  res5 = SWIG_AsCharPtrAndSize(argv[1], &buf5, NULL, &alloc5);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "char const *","svn_wc_check_root", 5, argv[1] ));
  }
  arg5 = (char *)(buf5);
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_wc_check_root(arg1,arg2,arg3,arg4,(char const *)arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, *arg1 ? Qtrue : Qfalse, 0);
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, *arg2 ? Qtrue : Qfalse, 0);
  }
  if (SWIG_IsTmpObj(res3)) {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_From_long((*arg3)), 0);
  } else {
    int new_flags = SWIG_IsNewObj(res3) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj((void*)(arg3), SWIGTYPE_p_svn_node_kind_t, new_flags), 0);
  }
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_is_wc_root2(int argc, VALUE *argv, VALUE self) {
  svn_boolean_t *arg1 = 0 ;
  svn_wc_context_t *arg2 = 0 ;
  char *arg3 = 0 ;
  apr_pool_t *arg4 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_boolean_t temp1 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_svn_wc_context_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_context_t *","svn_wc_is_wc_root2", 2, argv[0] )); 
  }
  arg2 = (svn_wc_context_t *)(argp2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_wc_is_wc_root2", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_wc_is_wc_root2(arg1,arg2,(char const *)arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, *arg1 ? Qtrue : Qfalse, 0);
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_is_wc_root(int argc, VALUE *argv, VALUE self) {
  svn_boolean_t *arg1 = 0 ;
  char *arg2 = 0 ;
  svn_wc_adm_access_t *arg3 = 0 ;
  apr_pool_t *arg4 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_boolean_t temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_wc_is_wc_root", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_ConvertPtr(argv[1], &argp3,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_is_wc_root", 3, argv[1] )); 
  }
  arg3 = (svn_wc_adm_access_t *)(argp3);
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_wc_is_wc_root(arg1,(char const *)arg2,arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, *arg1 ? Qtrue : Qfalse, 0);
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_get_actual_target2(int argc, VALUE *argv, VALUE self) {
  char **arg1 = 0 ;
  char **arg2 = 0 ;
  svn_wc_context_t *arg3 = 0 ;
  char *arg4 = 0 ;
  apr_pool_t *arg5 = 0 ;
  apr_pool_t *arg6 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  char *temp1 ;
  char *temp2 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  arg2 = &temp2;
  if ((argc < 2) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res3 = SWIG_ConvertPtr(argv[0], &argp3,SWIGTYPE_p_svn_wc_context_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_wc_context_t *","svn_wc_get_actual_target2", 3, argv[0] )); 
  }
  arg3 = (svn_wc_context_t *)(argp3);
  res4 = SWIG_AsCharPtrAndSize(argv[1], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_wc_get_actual_target2", 4, argv[1] ));
  }
  arg4 = (char *)(buf4);
  if (argc > 2) {
    
  }
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_wc_get_actual_target2((char const **)arg1,(char const **)arg2,arg3,(char const *)arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    if (*arg1) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new2(*arg1), 0);
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil, 0);
    }
  }
  {
    if (*arg2) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new2(*arg2), 0);
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil, 0);
    }
  }
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_get_actual_target(int argc, VALUE *argv, VALUE self) {
  char *arg1 = 0 ;
  char **arg2 = 0 ;
  char **arg3 = 0 ;
  apr_pool_t *arg4 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  char *temp2 ;
  char *temp3 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg2 = &temp2;
  arg3 = &temp3;
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_wc_get_actual_target", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  if (argc > 1) {
    
  }
  {
    result = (svn_error_t *)svn_wc_get_actual_target((char const *)arg1,(char const **)arg2,(char const **)arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    if (*arg2) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new2(*arg2), 0);
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil, 0);
    }
  }
  {
    if (*arg3) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new2(*arg3), 0);
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil, 0);
    }
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_get_update_editor4(int argc, VALUE *argv, VALUE self) {
  svn_delta_editor_t **arg1 = 0 ;
  void **arg2 = 0 ;
  svn_revnum_t *arg3 = 0 ;
  svn_wc_context_t *arg4 = 0 ;
  char *arg5 = 0 ;
  char *arg6 = 0 ;
  svn_boolean_t arg7 ;
  svn_depth_t arg8 ;
  svn_boolean_t arg9 ;
  svn_boolean_t arg10 ;
  svn_boolean_t arg11 ;
  svn_boolean_t arg12 ;
  svn_boolean_t arg13 ;
  char *arg14 = 0 ;
  apr_array_header_t *arg15 = 0 ;
  svn_wc_dirents_func_t arg16 = 0 ;
  void *arg17 = 0 ;
  svn_wc_conflict_resolver_func2_t arg18 = 0 ;
  void *arg19 = 0 ;
  svn_wc_external_update_t arg20 = 0 ;
  void *arg21 = 0 ;
  svn_cancel_func_t arg22 = 0 ;
  void *arg23 = 0 ;
  svn_wc_notify_func2_t arg24 = 0 ;
  void *arg25 = 0 ;
  apr_pool_t *arg26 = 0 ;
  apr_pool_t *arg27 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_delta_editor_t *temp1 ;
  void *temp2 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  int res5 ;
  char *buf5 = 0 ;
  int alloc5 = 0 ;
  int res6 ;
  char *buf6 = 0 ;
  int alloc6 = 0 ;
  int res17 ;
  int res19 ;
  int res21 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg26);
    _global_pool = arg26;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg27);
    _global_pool = arg27;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  arg2 = &temp2;
  if ((argc < 21) || (argc > 23)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 21)",argc); SWIG_fail;
  }
  {
    arg3 = apr_palloc(_global_pool, sizeof(svn_revnum_t));
    if (NIL_P(argv[0])) {
      *arg3 = ((svn_revnum_t) -1);
    } else {
      *arg3 = NUM2LONG(argv[0]);
    }
  }
  res4 = SWIG_ConvertPtr(argv[1], &argp4,SWIGTYPE_p_svn_wc_context_t, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "svn_wc_context_t *","svn_wc_get_update_editor4", 4, argv[1] )); 
  }
  arg4 = (svn_wc_context_t *)(argp4);
  res5 = SWIG_AsCharPtrAndSize(argv[2], &buf5, NULL, &alloc5);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "char const *","svn_wc_get_update_editor4", 5, argv[2] ));
  }
  arg5 = (char *)(buf5);
  res6 = SWIG_AsCharPtrAndSize(argv[3], &buf6, NULL, &alloc6);
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "char const *","svn_wc_get_update_editor4", 6, argv[3] ));
  }
  arg6 = (char *)(buf6);
  arg7 = RTEST(argv[4]);
  {
    arg8 = svn_swig_rb_to_depth(argv[5]);
  }
  arg9 = RTEST(argv[6]);
  arg10 = RTEST(argv[7]);
  arg11 = RTEST(argv[8]);
  arg12 = RTEST(argv[9]);
  arg13 = RTEST(argv[10]);
  {
    if (NIL_P(argv[11])) {
      arg14 = NULL;
    } else {
      arg14 = StringValuePtr(argv[11]);
    }
  }
  {
    arg15 = svn_swig_rb_strings_to_apr_array(argv[12], _global_pool);
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[13], (void**)(&arg16), SWIGTYPE_p_f_p_void_p_p_apr_hash_t_p_q_const__char_p_q_const__char_p_apr_pool_t_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_wc_dirents_func_t","svn_wc_get_update_editor4", 16, argv[13] )); 
    }
  }
  res17 = SWIG_ConvertPtr(argv[14],SWIG_as_voidptrptr(&arg17), 0, 0);
  if (!SWIG_IsOK(res17)) {
    SWIG_exception_fail(SWIG_ArgError(res17), Ruby_Format_TypeError( "", "void *","svn_wc_get_update_editor4", 17, argv[14] )); 
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[15], (void**)(&arg18), SWIGTYPE_p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description2_t_p_void_p_apr_pool_t_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_wc_conflict_resolver_func2_t","svn_wc_get_update_editor4", 18, argv[15] )); 
    }
  }
  res19 = SWIG_ConvertPtr(argv[16],SWIG_as_voidptrptr(&arg19), 0, 0);
  if (!SWIG_IsOK(res19)) {
    SWIG_exception_fail(SWIG_ArgError(res19), Ruby_Format_TypeError( "", "void *","svn_wc_get_update_editor4", 19, argv[16] )); 
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[17], (void**)(&arg20), SWIGTYPE_p_f_p_void_p_q_const__char_p_q_const__svn_string_t_p_q_const__svn_string_t_svn_depth_t_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_wc_external_update_t","svn_wc_get_update_editor4", 20, argv[17] )); 
    }
  }
  res21 = SWIG_ConvertPtr(argv[18],SWIG_as_voidptrptr(&arg21), 0, 0);
  if (!SWIG_IsOK(res21)) {
    SWIG_exception_fail(SWIG_ArgError(res21), Ruby_Format_TypeError( "", "void *","svn_wc_get_update_editor4", 21, argv[18] )); 
  }
  {
    arg22 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg23 = (void *)svn_swig_rb_make_baton(argv[19], _global_svn_swig_rb_pool);
  }
  {
    arg24 = (svn_wc_notify_func2_t) svn_swig_rb_notify_func2;
    arg25 = (void *)svn_swig_rb_make_baton(argv[20], _global_svn_swig_rb_pool);
  }
  if (argc > 21) {
    
  }
  if (argc > 22) {
    
  }
  {
    result = (svn_error_t *)svn_wc_get_update_editor4((struct svn_delta_editor_t const **)arg1,arg2,arg3,arg4,(char const *)arg5,(char const *)arg6,arg7,arg8,arg9,arg10,arg11,arg12,arg13,(char const *)arg14,(apr_array_header_t const *)arg15,arg16,arg17,arg18,arg19,arg20,arg21,arg22,arg23,arg24,arg25,arg26,arg27);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_delta_editor_t, 0), 0);
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg2, SWIGTYPE_p_void, 0), 0);
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, LONG2NUM((long)arg3), 0);
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg23);
  }
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_get_update_editor3(int argc, VALUE *argv, VALUE self) {
  svn_revnum_t *arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  char *arg3 = 0 ;
  svn_boolean_t arg4 ;
  svn_depth_t arg5 ;
  svn_boolean_t arg6 ;
  svn_boolean_t arg7 ;
  svn_wc_notify_func2_t arg8 = 0 ;
  void *arg9 = 0 ;
  svn_cancel_func_t arg10 = 0 ;
  void *arg11 = 0 ;
  svn_wc_conflict_resolver_func_t arg12 = 0 ;
  void *arg13 = 0 ;
  svn_wc_get_file_t arg14 = 0 ;
  void *arg15 = 0 ;
  char *arg16 = 0 ;
  apr_array_header_t *arg17 = 0 ;
  svn_delta_editor_t **arg18 = 0 ;
  void **arg19 = 0 ;
  svn_wc_traversal_info_t *arg20 = 0 ;
  apr_pool_t *arg21 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  svn_delta_editor_t *temp18 ;
  void *temp19 ;
  void *argp20 = 0 ;
  int res20 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg21);
    _global_pool = arg21;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg18 = &temp18;
  arg19 = &temp19;
  if ((argc < 14) || (argc > 15)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 14)",argc); SWIG_fail;
  }
  {
    arg1 = apr_palloc(_global_pool, sizeof(svn_revnum_t));
    if (NIL_P(argv[0])) {
      *arg1 = ((svn_revnum_t) -1);
    } else {
      *arg1 = NUM2LONG(argv[0]);
    }
  }
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_get_update_editor3", 2, argv[1] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_wc_get_update_editor3", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  arg4 = RTEST(argv[3]);
  {
    arg5 = svn_swig_rb_to_depth(argv[4]);
  }
  arg6 = RTEST(argv[5]);
  arg7 = RTEST(argv[6]);
  {
    arg8 = (svn_wc_notify_func2_t) svn_swig_rb_notify_func2;
    arg9 = (void *)svn_swig_rb_make_baton(argv[7], _global_svn_swig_rb_pool);
  }
  {
    arg10 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg11 = (void *)svn_swig_rb_make_baton(argv[8], _global_svn_swig_rb_pool);
  }
  {
    arg12 = svn_swig_rb_conflict_resolver_func;
    arg13 = (void *)svn_swig_rb_make_baton(argv[9], _global_svn_swig_rb_pool);
  }
  {
    /* ### TODO(rb support fetch_fun): implement
      arg14 = svn_swig_rb_fetch_func;
      arg15 = (void *)svn_swig_rb_make_baton(argv[10], _global_svn_swig_rb_pool);
      */
    arg14 = NULL;
    arg15 = NULL;
  }
  {
    if (NIL_P(argv[11])) {
      arg16 = NULL;
    } else {
      arg16 = StringValuePtr(argv[11]);
    }
  }
  {
    arg17 = svn_swig_rb_strings_to_apr_array(argv[12], _global_pool);
  }
  res20 = SWIG_ConvertPtr(argv[13], &argp20,SWIGTYPE_p_svn_wc_traversal_info_t, 0 |  0 );
  if (!SWIG_IsOK(res20)) {
    SWIG_exception_fail(SWIG_ArgError(res20), Ruby_Format_TypeError( "", "svn_wc_traversal_info_t *","svn_wc_get_update_editor3", 20, argv[13] )); 
  }
  arg20 = (svn_wc_traversal_info_t *)(argp20);
  if (argc > 14) {
    
  }
  {
    result = (svn_error_t *)svn_wc_get_update_editor3(arg1,arg2,(char const *)arg3,arg4,arg5,arg6,arg7,arg8,arg9,arg10,arg11,arg12,arg13,arg14,arg15,(char const *)arg16,(apr_array_header_t const *)arg17,(struct svn_delta_editor_t const **)arg18,arg19,arg20,arg21);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, LONG2NUM((long)arg1), 0);
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg11);
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg13);
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg15);
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg18, SWIGTYPE_p_svn_delta_editor_t, 0), 0);
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg19, SWIGTYPE_p_void, 0), 0);
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_get_update_editor2(int argc, VALUE *argv, VALUE self) {
  svn_revnum_t *arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  char *arg3 = 0 ;
  svn_boolean_t arg4 ;
  svn_boolean_t arg5 ;
  svn_wc_notify_func2_t arg6 = 0 ;
  void *arg7 = 0 ;
  svn_cancel_func_t arg8 = 0 ;
  void *arg9 = 0 ;
  char *arg10 = 0 ;
  svn_delta_editor_t **arg11 = 0 ;
  void **arg12 = 0 ;
  svn_wc_traversal_info_t *arg13 = 0 ;
  apr_pool_t *arg14 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  svn_delta_editor_t *temp11 ;
  void *temp12 ;
  void *argp13 = 0 ;
  int res13 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg14);
    _global_pool = arg14;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg11 = &temp11;
  arg12 = &temp12;
  if ((argc < 9) || (argc > 10)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 9)",argc); SWIG_fail;
  }
  {
    arg1 = apr_palloc(_global_pool, sizeof(svn_revnum_t));
    if (NIL_P(argv[0])) {
      *arg1 = ((svn_revnum_t) -1);
    } else {
      *arg1 = NUM2LONG(argv[0]);
    }
  }
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_get_update_editor2", 2, argv[1] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_wc_get_update_editor2", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  arg4 = RTEST(argv[3]);
  arg5 = RTEST(argv[4]);
  {
    arg6 = (svn_wc_notify_func2_t) svn_swig_rb_notify_func2;
    arg7 = (void *)svn_swig_rb_make_baton(argv[5], _global_svn_swig_rb_pool);
  }
  {
    arg8 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg9 = (void *)svn_swig_rb_make_baton(argv[6], _global_svn_swig_rb_pool);
  }
  {
    if (NIL_P(argv[7])) {
      arg10 = NULL;
    } else {
      arg10 = StringValuePtr(argv[7]);
    }
  }
  res13 = SWIG_ConvertPtr(argv[8], &argp13,SWIGTYPE_p_svn_wc_traversal_info_t, 0 |  0 );
  if (!SWIG_IsOK(res13)) {
    SWIG_exception_fail(SWIG_ArgError(res13), Ruby_Format_TypeError( "", "svn_wc_traversal_info_t *","svn_wc_get_update_editor2", 13, argv[8] )); 
  }
  arg13 = (svn_wc_traversal_info_t *)(argp13);
  if (argc > 9) {
    
  }
  {
    result = (svn_error_t *)svn_wc_get_update_editor2(arg1,arg2,(char const *)arg3,arg4,arg5,arg6,arg7,arg8,arg9,(char const *)arg10,(struct svn_delta_editor_t const **)arg11,arg12,arg13,arg14);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, LONG2NUM((long)arg1), 0);
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg9);
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg11, SWIGTYPE_p_svn_delta_editor_t, 0), 0);
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg12, SWIGTYPE_p_void, 0), 0);
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_get_update_editor(int argc, VALUE *argv, VALUE self) {
  svn_revnum_t *arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  char *arg3 = 0 ;
  svn_boolean_t arg4 ;
  svn_boolean_t arg5 ;
  svn_wc_notify_func_t arg6 = 0 ;
  void *arg7 = 0 ;
  svn_cancel_func_t arg8 = 0 ;
  void *arg9 = 0 ;
  char *arg10 = 0 ;
  svn_delta_editor_t **arg11 = 0 ;
  void **arg12 = 0 ;
  svn_wc_traversal_info_t *arg13 = 0 ;
  apr_pool_t *arg14 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int res7 ;
  svn_delta_editor_t *temp11 ;
  void *temp12 ;
  void *argp13 = 0 ;
  int res13 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg14);
    _global_pool = arg14;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg11 = &temp11;
  arg12 = &temp12;
  if ((argc < 10) || (argc > 11)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 10)",argc); SWIG_fail;
  }
  {
    arg1 = apr_palloc(_global_pool, sizeof(svn_revnum_t));
    if (NIL_P(argv[0])) {
      *arg1 = ((svn_revnum_t) -1);
    } else {
      *arg1 = NUM2LONG(argv[0]);
    }
  }
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_get_update_editor", 2, argv[1] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_wc_get_update_editor", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  arg4 = RTEST(argv[3]);
  arg5 = RTEST(argv[4]);
  {
    int res = SWIG_ConvertFunctionPtr(argv[5], (void**)(&arg6), SWIGTYPE_p_f_p_void_p_q_const__char_enum_svn_wc_notify_action_t_enum_svn_node_kind_t_p_q_const__char_enum_svn_wc_notify_state_t_enum_svn_wc_notify_state_t_long__void);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_wc_notify_func_t","svn_wc_get_update_editor", 6, argv[5] )); 
    }
  }
  res7 = SWIG_ConvertPtr(argv[6],SWIG_as_voidptrptr(&arg7), 0, 0);
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), Ruby_Format_TypeError( "", "void *","svn_wc_get_update_editor", 7, argv[6] )); 
  }
  {
    arg8 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg9 = (void *)svn_swig_rb_make_baton(argv[7], _global_svn_swig_rb_pool);
  }
  {
    if (NIL_P(argv[8])) {
      arg10 = NULL;
    } else {
      arg10 = StringValuePtr(argv[8]);
    }
  }
  res13 = SWIG_ConvertPtr(argv[9], &argp13,SWIGTYPE_p_svn_wc_traversal_info_t, 0 |  0 );
  if (!SWIG_IsOK(res13)) {
    SWIG_exception_fail(SWIG_ArgError(res13), Ruby_Format_TypeError( "", "svn_wc_traversal_info_t *","svn_wc_get_update_editor", 13, argv[9] )); 
  }
  arg13 = (svn_wc_traversal_info_t *)(argp13);
  if (argc > 10) {
    
  }
  {
    result = (svn_error_t *)svn_wc_get_update_editor(arg1,arg2,(char const *)arg3,arg4,arg5,arg6,arg7,arg8,arg9,(char const *)arg10,(struct svn_delta_editor_t const **)arg11,arg12,arg13,arg14);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, LONG2NUM((long)arg1), 0);
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg9);
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg11, SWIGTYPE_p_svn_delta_editor_t, 0), 0);
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg12, SWIGTYPE_p_void, 0), 0);
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_get_switch_editor4(int argc, VALUE *argv, VALUE self) {
  svn_delta_editor_t **arg1 = 0 ;
  void **arg2 = 0 ;
  svn_revnum_t *arg3 = 0 ;
  svn_wc_context_t *arg4 = 0 ;
  char *arg5 = 0 ;
  char *arg6 = 0 ;
  char *arg7 = 0 ;
  svn_boolean_t arg8 ;
  svn_depth_t arg9 ;
  svn_boolean_t arg10 ;
  svn_boolean_t arg11 ;
  svn_boolean_t arg12 ;
  char *arg13 = 0 ;
  apr_array_header_t *arg14 = 0 ;
  svn_wc_dirents_func_t arg15 = 0 ;
  void *arg16 = 0 ;
  svn_wc_conflict_resolver_func2_t arg17 = 0 ;
  void *arg18 = 0 ;
  svn_wc_external_update_t arg19 = 0 ;
  void *arg20 = 0 ;
  svn_cancel_func_t arg21 = 0 ;
  void *arg22 = 0 ;
  svn_wc_notify_func2_t arg23 = 0 ;
  void *arg24 = 0 ;
  apr_pool_t *arg25 = 0 ;
  apr_pool_t *arg26 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_delta_editor_t *temp1 ;
  void *temp2 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  int res5 ;
  char *buf5 = 0 ;
  int alloc5 = 0 ;
  int res6 ;
  char *buf6 = 0 ;
  int alloc6 = 0 ;
  int res7 ;
  char *buf7 = 0 ;
  int alloc7 = 0 ;
  int res16 ;
  int res18 ;
  int res20 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg25);
    _global_pool = arg25;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg26);
    _global_pool = arg26;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  arg2 = &temp2;
  if ((argc < 20) || (argc > 22)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 20)",argc); SWIG_fail;
  }
  {
    arg3 = apr_palloc(_global_pool, sizeof(svn_revnum_t));
    if (NIL_P(argv[0])) {
      *arg3 = ((svn_revnum_t) -1);
    } else {
      *arg3 = NUM2LONG(argv[0]);
    }
  }
  res4 = SWIG_ConvertPtr(argv[1], &argp4,SWIGTYPE_p_svn_wc_context_t, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "svn_wc_context_t *","svn_wc_get_switch_editor4", 4, argv[1] )); 
  }
  arg4 = (svn_wc_context_t *)(argp4);
  res5 = SWIG_AsCharPtrAndSize(argv[2], &buf5, NULL, &alloc5);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "char const *","svn_wc_get_switch_editor4", 5, argv[2] ));
  }
  arg5 = (char *)(buf5);
  res6 = SWIG_AsCharPtrAndSize(argv[3], &buf6, NULL, &alloc6);
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "char const *","svn_wc_get_switch_editor4", 6, argv[3] ));
  }
  arg6 = (char *)(buf6);
  res7 = SWIG_AsCharPtrAndSize(argv[4], &buf7, NULL, &alloc7);
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), Ruby_Format_TypeError( "", "char const *","svn_wc_get_switch_editor4", 7, argv[4] ));
  }
  arg7 = (char *)(buf7);
  arg8 = RTEST(argv[5]);
  {
    arg9 = svn_swig_rb_to_depth(argv[6]);
  }
  arg10 = RTEST(argv[7]);
  arg11 = RTEST(argv[8]);
  arg12 = RTEST(argv[9]);
  {
    if (NIL_P(argv[10])) {
      arg13 = NULL;
    } else {
      arg13 = StringValuePtr(argv[10]);
    }
  }
  {
    arg14 = svn_swig_rb_strings_to_apr_array(argv[11], _global_pool);
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[12], (void**)(&arg15), SWIGTYPE_p_f_p_void_p_p_apr_hash_t_p_q_const__char_p_q_const__char_p_apr_pool_t_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_wc_dirents_func_t","svn_wc_get_switch_editor4", 15, argv[12] )); 
    }
  }
  res16 = SWIG_ConvertPtr(argv[13],SWIG_as_voidptrptr(&arg16), 0, 0);
  if (!SWIG_IsOK(res16)) {
    SWIG_exception_fail(SWIG_ArgError(res16), Ruby_Format_TypeError( "", "void *","svn_wc_get_switch_editor4", 16, argv[13] )); 
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[14], (void**)(&arg17), SWIGTYPE_p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description2_t_p_void_p_apr_pool_t_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_wc_conflict_resolver_func2_t","svn_wc_get_switch_editor4", 17, argv[14] )); 
    }
  }
  res18 = SWIG_ConvertPtr(argv[15],SWIG_as_voidptrptr(&arg18), 0, 0);
  if (!SWIG_IsOK(res18)) {
    SWIG_exception_fail(SWIG_ArgError(res18), Ruby_Format_TypeError( "", "void *","svn_wc_get_switch_editor4", 18, argv[15] )); 
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[16], (void**)(&arg19), SWIGTYPE_p_f_p_void_p_q_const__char_p_q_const__svn_string_t_p_q_const__svn_string_t_svn_depth_t_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_wc_external_update_t","svn_wc_get_switch_editor4", 19, argv[16] )); 
    }
  }
  res20 = SWIG_ConvertPtr(argv[17],SWIG_as_voidptrptr(&arg20), 0, 0);
  if (!SWIG_IsOK(res20)) {
    SWIG_exception_fail(SWIG_ArgError(res20), Ruby_Format_TypeError( "", "void *","svn_wc_get_switch_editor4", 20, argv[17] )); 
  }
  {
    arg21 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg22 = (void *)svn_swig_rb_make_baton(argv[18], _global_svn_swig_rb_pool);
  }
  {
    arg23 = (svn_wc_notify_func2_t) svn_swig_rb_notify_func2;
    arg24 = (void *)svn_swig_rb_make_baton(argv[19], _global_svn_swig_rb_pool);
  }
  if (argc > 20) {
    
  }
  if (argc > 21) {
    
  }
  {
    result = (svn_error_t *)svn_wc_get_switch_editor4((struct svn_delta_editor_t const **)arg1,arg2,arg3,arg4,(char const *)arg5,(char const *)arg6,(char const *)arg7,arg8,arg9,arg10,arg11,arg12,(char const *)arg13,(apr_array_header_t const *)arg14,arg15,arg16,arg17,arg18,arg19,arg20,arg21,arg22,arg23,arg24,arg25,arg26);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_delta_editor_t, 0), 0);
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg2, SWIGTYPE_p_void, 0), 0);
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, LONG2NUM((long)arg3), 0);
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg22);
  }
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  if (alloc7 == SWIG_NEWOBJ) free((char*)buf7);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  if (alloc7 == SWIG_NEWOBJ) free((char*)buf7);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_get_switch_editor3(int argc, VALUE *argv, VALUE self) {
  svn_revnum_t *arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  char *arg3 = 0 ;
  char *arg4 = 0 ;
  svn_boolean_t arg5 ;
  svn_depth_t arg6 ;
  svn_boolean_t arg7 ;
  svn_boolean_t arg8 ;
  svn_wc_notify_func2_t arg9 = 0 ;
  void *arg10 = 0 ;
  svn_cancel_func_t arg11 = 0 ;
  void *arg12 = 0 ;
  svn_wc_conflict_resolver_func_t arg13 = 0 ;
  void *arg14 = 0 ;
  char *arg15 = 0 ;
  apr_array_header_t *arg16 = 0 ;
  svn_delta_editor_t **arg17 = 0 ;
  void **arg18 = 0 ;
  svn_wc_traversal_info_t *arg19 = 0 ;
  apr_pool_t *arg20 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  svn_delta_editor_t *temp17 ;
  void *temp18 ;
  void *argp19 = 0 ;
  int res19 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg20);
    _global_pool = arg20;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg17 = &temp17;
  arg18 = &temp18;
  if ((argc < 14) || (argc > 15)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 14)",argc); SWIG_fail;
  }
  {
    arg1 = apr_palloc(_global_pool, sizeof(svn_revnum_t));
    if (NIL_P(argv[0])) {
      *arg1 = ((svn_revnum_t) -1);
    } else {
      *arg1 = NUM2LONG(argv[0]);
    }
  }
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_get_switch_editor3", 2, argv[1] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_wc_get_switch_editor3", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_AsCharPtrAndSize(argv[3], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_wc_get_switch_editor3", 4, argv[3] ));
  }
  arg4 = (char *)(buf4);
  arg5 = RTEST(argv[4]);
  {
    arg6 = svn_swig_rb_to_depth(argv[5]);
  }
  arg7 = RTEST(argv[6]);
  arg8 = RTEST(argv[7]);
  {
    arg9 = (svn_wc_notify_func2_t) svn_swig_rb_notify_func2;
    arg10 = (void *)svn_swig_rb_make_baton(argv[8], _global_svn_swig_rb_pool);
  }
  {
    arg11 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg12 = (void *)svn_swig_rb_make_baton(argv[9], _global_svn_swig_rb_pool);
  }
  {
    arg13 = svn_swig_rb_conflict_resolver_func;
    arg14 = (void *)svn_swig_rb_make_baton(argv[10], _global_svn_swig_rb_pool);
  }
  {
    if (NIL_P(argv[11])) {
      arg15 = NULL;
    } else {
      arg15 = StringValuePtr(argv[11]);
    }
  }
  {
    arg16 = svn_swig_rb_strings_to_apr_array(argv[12], _global_pool);
  }
  res19 = SWIG_ConvertPtr(argv[13], &argp19,SWIGTYPE_p_svn_wc_traversal_info_t, 0 |  0 );
  if (!SWIG_IsOK(res19)) {
    SWIG_exception_fail(SWIG_ArgError(res19), Ruby_Format_TypeError( "", "svn_wc_traversal_info_t *","svn_wc_get_switch_editor3", 19, argv[13] )); 
  }
  arg19 = (svn_wc_traversal_info_t *)(argp19);
  if (argc > 14) {
    
  }
  {
    result = (svn_error_t *)svn_wc_get_switch_editor3(arg1,arg2,(char const *)arg3,(char const *)arg4,arg5,arg6,arg7,arg8,arg9,arg10,arg11,arg12,arg13,arg14,(char const *)arg15,(apr_array_header_t const *)arg16,(struct svn_delta_editor_t const **)arg17,arg18,arg19,arg20);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, LONG2NUM((long)arg1), 0);
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg12);
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg14);
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg17, SWIGTYPE_p_svn_delta_editor_t, 0), 0);
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg18, SWIGTYPE_p_void, 0), 0);
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_get_switch_editor2(int argc, VALUE *argv, VALUE self) {
  svn_revnum_t *arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  char *arg3 = 0 ;
  char *arg4 = 0 ;
  svn_boolean_t arg5 ;
  svn_boolean_t arg6 ;
  svn_wc_notify_func2_t arg7 = 0 ;
  void *arg8 = 0 ;
  svn_cancel_func_t arg9 = 0 ;
  void *arg10 = 0 ;
  char *arg11 = 0 ;
  svn_delta_editor_t **arg12 = 0 ;
  void **arg13 = 0 ;
  svn_wc_traversal_info_t *arg14 = 0 ;
  apr_pool_t *arg15 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  svn_delta_editor_t *temp12 ;
  void *temp13 ;
  void *argp14 = 0 ;
  int res14 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg15);
    _global_pool = arg15;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg12 = &temp12;
  arg13 = &temp13;
  if ((argc < 10) || (argc > 11)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 10)",argc); SWIG_fail;
  }
  {
    arg1 = apr_palloc(_global_pool, sizeof(svn_revnum_t));
    if (NIL_P(argv[0])) {
      *arg1 = ((svn_revnum_t) -1);
    } else {
      *arg1 = NUM2LONG(argv[0]);
    }
  }
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_get_switch_editor2", 2, argv[1] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_wc_get_switch_editor2", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_AsCharPtrAndSize(argv[3], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_wc_get_switch_editor2", 4, argv[3] ));
  }
  arg4 = (char *)(buf4);
  arg5 = RTEST(argv[4]);
  arg6 = RTEST(argv[5]);
  {
    arg7 = (svn_wc_notify_func2_t) svn_swig_rb_notify_func2;
    arg8 = (void *)svn_swig_rb_make_baton(argv[6], _global_svn_swig_rb_pool);
  }
  {
    arg9 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg10 = (void *)svn_swig_rb_make_baton(argv[7], _global_svn_swig_rb_pool);
  }
  {
    if (NIL_P(argv[8])) {
      arg11 = NULL;
    } else {
      arg11 = StringValuePtr(argv[8]);
    }
  }
  res14 = SWIG_ConvertPtr(argv[9], &argp14,SWIGTYPE_p_svn_wc_traversal_info_t, 0 |  0 );
  if (!SWIG_IsOK(res14)) {
    SWIG_exception_fail(SWIG_ArgError(res14), Ruby_Format_TypeError( "", "svn_wc_traversal_info_t *","svn_wc_get_switch_editor2", 14, argv[9] )); 
  }
  arg14 = (svn_wc_traversal_info_t *)(argp14);
  if (argc > 10) {
    
  }
  {
    result = (svn_error_t *)svn_wc_get_switch_editor2(arg1,arg2,(char const *)arg3,(char const *)arg4,arg5,arg6,arg7,arg8,arg9,arg10,(char const *)arg11,(struct svn_delta_editor_t const **)arg12,arg13,arg14,arg15);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, LONG2NUM((long)arg1), 0);
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg10);
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg12, SWIGTYPE_p_svn_delta_editor_t, 0), 0);
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg13, SWIGTYPE_p_void, 0), 0);
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_get_switch_editor(int argc, VALUE *argv, VALUE self) {
  svn_revnum_t *arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  char *arg3 = 0 ;
  char *arg4 = 0 ;
  svn_boolean_t arg5 ;
  svn_boolean_t arg6 ;
  svn_wc_notify_func_t arg7 = 0 ;
  void *arg8 = 0 ;
  svn_cancel_func_t arg9 = 0 ;
  void *arg10 = 0 ;
  char *arg11 = 0 ;
  svn_delta_editor_t **arg12 = 0 ;
  void **arg13 = 0 ;
  svn_wc_traversal_info_t *arg14 = 0 ;
  apr_pool_t *arg15 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  int res8 ;
  svn_delta_editor_t *temp12 ;
  void *temp13 ;
  void *argp14 = 0 ;
  int res14 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg15);
    _global_pool = arg15;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg12 = &temp12;
  arg13 = &temp13;
  if ((argc < 11) || (argc > 12)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 11)",argc); SWIG_fail;
  }
  {
    arg1 = apr_palloc(_global_pool, sizeof(svn_revnum_t));
    if (NIL_P(argv[0])) {
      *arg1 = ((svn_revnum_t) -1);
    } else {
      *arg1 = NUM2LONG(argv[0]);
    }
  }
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_get_switch_editor", 2, argv[1] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_wc_get_switch_editor", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_AsCharPtrAndSize(argv[3], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_wc_get_switch_editor", 4, argv[3] ));
  }
  arg4 = (char *)(buf4);
  arg5 = RTEST(argv[4]);
  arg6 = RTEST(argv[5]);
  {
    int res = SWIG_ConvertFunctionPtr(argv[6], (void**)(&arg7), SWIGTYPE_p_f_p_void_p_q_const__char_enum_svn_wc_notify_action_t_enum_svn_node_kind_t_p_q_const__char_enum_svn_wc_notify_state_t_enum_svn_wc_notify_state_t_long__void);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_wc_notify_func_t","svn_wc_get_switch_editor", 7, argv[6] )); 
    }
  }
  res8 = SWIG_ConvertPtr(argv[7],SWIG_as_voidptrptr(&arg8), 0, 0);
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), Ruby_Format_TypeError( "", "void *","svn_wc_get_switch_editor", 8, argv[7] )); 
  }
  {
    arg9 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg10 = (void *)svn_swig_rb_make_baton(argv[8], _global_svn_swig_rb_pool);
  }
  {
    if (NIL_P(argv[9])) {
      arg11 = NULL;
    } else {
      arg11 = StringValuePtr(argv[9]);
    }
  }
  res14 = SWIG_ConvertPtr(argv[10], &argp14,SWIGTYPE_p_svn_wc_traversal_info_t, 0 |  0 );
  if (!SWIG_IsOK(res14)) {
    SWIG_exception_fail(SWIG_ArgError(res14), Ruby_Format_TypeError( "", "svn_wc_traversal_info_t *","svn_wc_get_switch_editor", 14, argv[10] )); 
  }
  arg14 = (svn_wc_traversal_info_t *)(argp14);
  if (argc > 11) {
    
  }
  {
    result = (svn_error_t *)svn_wc_get_switch_editor(arg1,arg2,(char const *)arg3,(char const *)arg4,arg5,arg6,arg7,arg8,arg9,arg10,(char const *)arg11,(struct svn_delta_editor_t const **)arg12,arg13,arg14,arg15);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, LONG2NUM((long)arg1), 0);
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg10);
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg12, SWIGTYPE_p_svn_delta_editor_t, 0), 0);
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg13, SWIGTYPE_p_void, 0), 0);
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_prop_list2(int argc, VALUE *argv, VALUE self) {
  apr_hash_t **arg1 = 0 ;
  svn_wc_context_t *arg2 = 0 ;
  char *arg3 = 0 ;
  apr_pool_t *arg4 = 0 ;
  apr_pool_t *arg5 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  apr_hash_t *temp1 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 2) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_svn_wc_context_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_context_t *","svn_wc_prop_list2", 2, argv[0] )); 
  }
  arg2 = (svn_wc_context_t *)(argp2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_wc_prop_list2", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  if (argc > 2) {
    
  }
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_wc_prop_list2(arg1,arg2,(char const *)arg3,arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, svn_swig_rb_apr_hash_to_hash_svn_string(*arg1), 0);
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_prop_list(int argc, VALUE *argv, VALUE self) {
  apr_hash_t **arg1 = 0 ;
  char *arg2 = 0 ;
  svn_wc_adm_access_t *arg3 = 0 ;
  apr_pool_t *arg4 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  apr_hash_t *temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_wc_prop_list", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_ConvertPtr(argv[1], &argp3,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_prop_list", 3, argv[1] )); 
  }
  arg3 = (svn_wc_adm_access_t *)(argp3);
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_wc_prop_list(arg1,(char const *)arg2,arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, svn_swig_rb_apr_hash_to_hash_svn_string(*arg1), 0);
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_get_pristine_props(int argc, VALUE *argv, VALUE self) {
  apr_hash_t **arg1 = 0 ;
  svn_wc_context_t *arg2 = 0 ;
  char *arg3 = 0 ;
  apr_pool_t *arg4 = 0 ;
  apr_pool_t *arg5 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  apr_hash_t *temp1 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 2) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_svn_wc_context_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_context_t *","svn_wc_get_pristine_props", 2, argv[0] )); 
  }
  arg2 = (svn_wc_context_t *)(argp2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_wc_get_pristine_props", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  if (argc > 2) {
    
  }
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_wc_get_pristine_props(arg1,arg2,(char const *)arg3,arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, svn_swig_rb_apr_hash_to_hash_svn_string(*arg1), 0);
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_prop_get2(int argc, VALUE *argv, VALUE self) {
  svn_string_t **arg1 = 0 ;
  svn_wc_context_t *arg2 = 0 ;
  char *arg3 = 0 ;
  char *arg4 = 0 ;
  apr_pool_t *arg5 = 0 ;
  apr_pool_t *arg6 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_string_t *temp1 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 3) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_svn_wc_context_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_context_t *","svn_wc_prop_get2", 2, argv[0] )); 
  }
  arg2 = (svn_wc_context_t *)(argp2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_wc_prop_get2", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_AsCharPtrAndSize(argv[2], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_wc_prop_get2", 4, argv[2] ));
  }
  arg4 = (char *)(buf4);
  if (argc > 3) {
    
  }
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_wc_prop_get2((struct svn_string_t const **)arg1,arg2,(char const *)arg3,(char const *)arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    if (*arg1) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new((*arg1)->data, (*arg1)->len), 0);
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil, 0);
    }
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_prop_get(int argc, VALUE *argv, VALUE self) {
  svn_string_t **arg1 = 0 ;
  char *arg2 = 0 ;
  char *arg3 = 0 ;
  svn_wc_adm_access_t *arg4 = 0 ;
  apr_pool_t *arg5 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_string_t *temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 3) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_wc_prop_get", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_wc_prop_get", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_ConvertPtr(argv[2], &argp4,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_prop_get", 4, argv[2] )); 
  }
  arg4 = (svn_wc_adm_access_t *)(argp4);
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_wc_prop_get((struct svn_string_t const **)arg1,(char const *)arg2,(char const *)arg3,arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    if (*arg1) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new((*arg1)->data, (*arg1)->len), 0);
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil, 0);
    }
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_prop_set4(int argc, VALUE *argv, VALUE self) {
  svn_wc_context_t *arg1 = 0 ;
  char *arg2 = 0 ;
  char *arg3 = 0 ;
  svn_string_t *arg4 = 0 ;
  svn_depth_t arg5 ;
  svn_boolean_t arg6 ;
  apr_array_header_t *arg7 = 0 ;
  svn_cancel_func_t arg8 = 0 ;
  void *arg9 = 0 ;
  svn_wc_notify_func2_t arg10 = 0 ;
  void *arg11 = 0 ;
  apr_pool_t *arg12 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  svn_string_t value4 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg12);
    _global_pool = arg12;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 9) || (argc > 10)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 9)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_context_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_context_t *","svn_wc_prop_set4", 1, argv[0] )); 
  }
  arg1 = (svn_wc_context_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_wc_prop_set4", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_wc_prop_set4", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  {
    if (NIL_P(argv[3])) {
      arg4 = NULL;
    } else {
      value4.data = StringValuePtr(argv[3]);
      value4.len = RSTRING_LEN(argv[3]);
      arg4 = &value4;
    }
  }
  {
    arg5 = svn_swig_rb_to_depth(argv[4]);
  }
  arg6 = RTEST(argv[5]);
  res7 = SWIG_ConvertPtr(argv[6], &argp7,SWIGTYPE_p_apr_array_header_t, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), Ruby_Format_TypeError( "", "apr_array_header_t const *","svn_wc_prop_set4", 7, argv[6] )); 
  }
  arg7 = (apr_array_header_t *)(argp7);
  {
    arg8 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg9 = (void *)svn_swig_rb_make_baton(argv[7], _global_svn_swig_rb_pool);
  }
  {
    arg10 = (svn_wc_notify_func2_t) svn_swig_rb_notify_func2;
    arg11 = (void *)svn_swig_rb_make_baton(argv[8], _global_svn_swig_rb_pool);
  }
  if (argc > 9) {
    
  }
  {
    result = (svn_error_t *)svn_wc_prop_set4(arg1,(char const *)arg2,(char const *)arg3,(struct svn_string_t const *)arg4,arg5,arg6,(apr_array_header_t const *)arg7,arg8,arg9,arg10,arg11,arg12);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg9);
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_prop_set3(int argc, VALUE *argv, VALUE self) {
  char *arg1 = 0 ;
  svn_string_t *arg2 = 0 ;
  char *arg3 = 0 ;
  svn_wc_adm_access_t *arg4 = 0 ;
  svn_boolean_t arg5 ;
  svn_wc_notify_func2_t arg6 = 0 ;
  void *arg7 = 0 ;
  apr_pool_t *arg8 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  svn_string_t value2 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg8);
    _global_pool = arg8;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 6) || (argc > 7)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 6)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_wc_prop_set3", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    if (NIL_P(argv[1])) {
      arg2 = NULL;
    } else {
      value2.data = StringValuePtr(argv[1]);
      value2.len = RSTRING_LEN(argv[1]);
      arg2 = &value2;
    }
  }
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_wc_prop_set3", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_ConvertPtr(argv[3], &argp4,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_prop_set3", 4, argv[3] )); 
  }
  arg4 = (svn_wc_adm_access_t *)(argp4);
  arg5 = RTEST(argv[4]);
  {
    arg6 = (svn_wc_notify_func2_t) svn_swig_rb_notify_func2;
    arg7 = (void *)svn_swig_rb_make_baton(argv[5], _global_svn_swig_rb_pool);
  }
  if (argc > 6) {
    
  }
  {
    result = (svn_error_t *)svn_wc_prop_set3((char const *)arg1,(struct svn_string_t const *)arg2,(char const *)arg3,arg4,arg5,arg6,arg7,arg8);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_prop_set2(int argc, VALUE *argv, VALUE self) {
  char *arg1 = 0 ;
  svn_string_t *arg2 = 0 ;
  char *arg3 = 0 ;
  svn_wc_adm_access_t *arg4 = 0 ;
  svn_boolean_t arg5 ;
  apr_pool_t *arg6 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  svn_string_t value2 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 5) || (argc > 6)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 5)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_wc_prop_set2", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    if (NIL_P(argv[1])) {
      arg2 = NULL;
    } else {
      value2.data = StringValuePtr(argv[1]);
      value2.len = RSTRING_LEN(argv[1]);
      arg2 = &value2;
    }
  }
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_wc_prop_set2", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_ConvertPtr(argv[3], &argp4,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_prop_set2", 4, argv[3] )); 
  }
  arg4 = (svn_wc_adm_access_t *)(argp4);
  arg5 = RTEST(argv[4]);
  if (argc > 5) {
    
  }
  {
    result = (svn_error_t *)svn_wc_prop_set2((char const *)arg1,(struct svn_string_t const *)arg2,(char const *)arg3,arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_prop_set(int argc, VALUE *argv, VALUE self) {
  char *arg1 = 0 ;
  svn_string_t *arg2 = 0 ;
  char *arg3 = 0 ;
  svn_wc_adm_access_t *arg4 = 0 ;
  apr_pool_t *arg5 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  svn_string_t value2 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 4) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_wc_prop_set", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    if (NIL_P(argv[1])) {
      arg2 = NULL;
    } else {
      value2.data = StringValuePtr(argv[1]);
      value2.len = RSTRING_LEN(argv[1]);
      arg2 = &value2;
    }
  }
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_wc_prop_set", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_ConvertPtr(argv[3], &argp4,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_prop_set", 4, argv[3] )); 
  }
  arg4 = (svn_wc_adm_access_t *)(argp4);
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_wc_prop_set((char const *)arg1,(struct svn_string_t const *)arg2,(char const *)arg3,arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_is_normal_prop(int argc, VALUE *argv, VALUE self) {
  char *arg1 = 0 ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_wc_is_normal_prop", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    result = (svn_boolean_t)svn_wc_is_normal_prop((char const *)arg1);
    
    
    
  }
  vresult = result ? Qtrue : Qfalse;
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_is_wc_prop(int argc, VALUE *argv, VALUE self) {
  char *arg1 = 0 ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_wc_is_wc_prop", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    result = (svn_boolean_t)svn_wc_is_wc_prop((char const *)arg1);
    
    
    
  }
  vresult = result ? Qtrue : Qfalse;
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_is_entry_prop(int argc, VALUE *argv, VALUE self) {
  char *arg1 = 0 ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_wc_is_entry_prop", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    result = (svn_boolean_t)svn_wc_is_entry_prop((char const *)arg1);
    
    
    
  }
  vresult = result ? Qtrue : Qfalse;
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_canonicalize_svn_prop(int argc, VALUE *argv, VALUE self) {
  svn_string_t **arg1 = 0 ;
  char *arg2 = 0 ;
  svn_string_t *arg3 = 0 ;
  char *arg4 = 0 ;
  svn_node_kind_t arg5 ;
  svn_boolean_t arg6 ;
  svn_wc_canonicalize_svn_prop_get_file_t arg7 = 0 ;
  void *arg8 = 0 ;
  apr_pool_t *arg9 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_string_t *temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  svn_string_t value3 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  int val5 ;
  int ecode5 = 0 ;
  int res8 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg9);
    _global_pool = arg9;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 7) || (argc > 8)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 7)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_wc_canonicalize_svn_prop", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  {
    if (NIL_P(argv[1])) {
      arg3 = NULL;
    } else {
      value3.data = StringValuePtr(argv[1]);
      value3.len = RSTRING_LEN(argv[1]);
      arg3 = &value3;
    }
  }
  res4 = SWIG_AsCharPtrAndSize(argv[2], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_wc_canonicalize_svn_prop", 4, argv[2] ));
  }
  arg4 = (char *)(buf4);
  ecode5 = SWIG_AsVal_int(argv[3], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), Ruby_Format_TypeError( "", "svn_node_kind_t","svn_wc_canonicalize_svn_prop", 5, argv[3] ));
  } 
  arg5 = (svn_node_kind_t)(val5);
  arg6 = RTEST(argv[4]);
  {
    int res = SWIG_ConvertFunctionPtr(argv[5], (void**)(&arg7), SWIGTYPE_p_f_p_p_q_const__svn_string_t_p_svn_stream_t_p_void_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_wc_canonicalize_svn_prop_get_file_t","svn_wc_canonicalize_svn_prop", 7, argv[5] )); 
    }
  }
  res8 = SWIG_ConvertPtr(argv[6],SWIG_as_voidptrptr(&arg8), 0, 0);
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), Ruby_Format_TypeError( "", "void *","svn_wc_canonicalize_svn_prop", 8, argv[6] )); 
  }
  if (argc > 7) {
    
  }
  {
    result = (svn_error_t *)svn_wc_canonicalize_svn_prop((struct svn_string_t const **)arg1,(char const *)arg2,(struct svn_string_t const *)arg3,(char const *)arg4,arg5,arg6,arg7,arg8,arg9);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    if (*arg1) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new((*arg1)->data, (*arg1)->len), 0);
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil, 0);
    }
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_get_diff_editor6(int argc, VALUE *argv, VALUE self) {
  svn_delta_editor_t **arg1 = 0 ;
  void **arg2 = 0 ;
  svn_wc_context_t *arg3 = 0 ;
  char *arg4 = 0 ;
  char *arg5 = 0 ;
  svn_depth_t arg6 ;
  svn_boolean_t arg7 ;
  svn_boolean_t arg8 ;
  svn_boolean_t arg9 ;
  svn_boolean_t arg10 ;
  svn_boolean_t arg11 ;
  svn_boolean_t arg12 ;
  apr_array_header_t *arg13 = 0 ;
  svn_wc_diff_callbacks4_t *arg14 = 0 ;
  void *arg15 = 0 ;
  svn_cancel_func_t arg16 = 0 ;
  void *arg17 = 0 ;
  apr_pool_t *arg18 = 0 ;
  apr_pool_t *arg19 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_delta_editor_t *temp1 ;
  void *temp2 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  int res5 ;
  char *buf5 = 0 ;
  int alloc5 = 0 ;
  void *argp13 = 0 ;
  int res13 = 0 ;
  void *argp14 = 0 ;
  int res14 = 0 ;
  int res15 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg18);
    _global_pool = arg18;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg19);
    _global_pool = arg19;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  arg2 = &temp2;
  if ((argc < 14) || (argc > 16)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 14)",argc); SWIG_fail;
  }
  res3 = SWIG_ConvertPtr(argv[0], &argp3,SWIGTYPE_p_svn_wc_context_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_wc_context_t *","svn_wc_get_diff_editor6", 3, argv[0] )); 
  }
  arg3 = (svn_wc_context_t *)(argp3);
  res4 = SWIG_AsCharPtrAndSize(argv[1], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_wc_get_diff_editor6", 4, argv[1] ));
  }
  arg4 = (char *)(buf4);
  res5 = SWIG_AsCharPtrAndSize(argv[2], &buf5, NULL, &alloc5);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "char const *","svn_wc_get_diff_editor6", 5, argv[2] ));
  }
  arg5 = (char *)(buf5);
  {
    arg6 = svn_swig_rb_to_depth(argv[3]);
  }
  arg7 = RTEST(argv[4]);
  arg8 = RTEST(argv[5]);
  arg9 = RTEST(argv[6]);
  arg10 = RTEST(argv[7]);
  arg11 = RTEST(argv[8]);
  arg12 = RTEST(argv[9]);
  res13 = SWIG_ConvertPtr(argv[10], &argp13,SWIGTYPE_p_apr_array_header_t, 0 |  0 );
  if (!SWIG_IsOK(res13)) {
    SWIG_exception_fail(SWIG_ArgError(res13), Ruby_Format_TypeError( "", "apr_array_header_t const *","svn_wc_get_diff_editor6", 13, argv[10] )); 
  }
  arg13 = (apr_array_header_t *)(argp13);
  res14 = SWIG_ConvertPtr(argv[11], &argp14,SWIGTYPE_p_svn_wc_diff_callbacks4_t, 0 |  0 );
  if (!SWIG_IsOK(res14)) {
    SWIG_exception_fail(SWIG_ArgError(res14), Ruby_Format_TypeError( "", "svn_wc_diff_callbacks4_t const *","svn_wc_get_diff_editor6", 14, argv[11] )); 
  }
  arg14 = (svn_wc_diff_callbacks4_t *)(argp14);
  res15 = SWIG_ConvertPtr(argv[12],SWIG_as_voidptrptr(&arg15), 0, 0);
  if (!SWIG_IsOK(res15)) {
    SWIG_exception_fail(SWIG_ArgError(res15), Ruby_Format_TypeError( "", "void *","svn_wc_get_diff_editor6", 15, argv[12] )); 
  }
  {
    arg16 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg17 = (void *)svn_swig_rb_make_baton(argv[13], _global_svn_swig_rb_pool);
  }
  if (argc > 14) {
    
  }
  if (argc > 15) {
    
  }
  {
    result = (svn_error_t *)svn_wc_get_diff_editor6((struct svn_delta_editor_t const **)arg1,arg2,arg3,(char const *)arg4,(char const *)arg5,arg6,arg7,arg8,arg9,arg10,arg11,arg12,(apr_array_header_t const *)arg13,(struct svn_wc_diff_callbacks4_t const *)arg14,arg15,arg16,arg17,arg18,arg19);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_delta_editor_t, 0), 0);
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg2, SWIGTYPE_p_void, 0), 0);
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg17);
  }
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_get_diff_editor5(int argc, VALUE *argv, VALUE self) {
  svn_wc_adm_access_t *arg1 = 0 ;
  char *arg2 = 0 ;
  svn_wc_diff_callbacks3_t *arg3 = 0 ;
  void *arg4 = 0 ;
  svn_depth_t arg5 ;
  svn_boolean_t arg6 ;
  svn_boolean_t arg7 ;
  svn_boolean_t arg8 ;
  svn_cancel_func_t arg9 = 0 ;
  void *arg10 = 0 ;
  apr_array_header_t *arg11 = 0 ;
  svn_delta_editor_t **arg12 = 0 ;
  void **arg13 = 0 ;
  apr_pool_t *arg14 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  int res4 ;
  void *argp11 = 0 ;
  int res11 = 0 ;
  svn_delta_editor_t *temp12 ;
  void *temp13 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg14);
    _global_pool = arg14;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg12 = &temp12;
  arg13 = &temp13;
  if ((argc < 10) || (argc > 11)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 10)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_get_diff_editor5", 1, argv[0] )); 
  }
  arg1 = (svn_wc_adm_access_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_wc_get_diff_editor5", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_ConvertPtr(argv[2], &argp3,SWIGTYPE_p_svn_wc_diff_callbacks3_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_wc_diff_callbacks3_t const *","svn_wc_get_diff_editor5", 3, argv[2] )); 
  }
  arg3 = (svn_wc_diff_callbacks3_t *)(argp3);
  res4 = SWIG_ConvertPtr(argv[3],SWIG_as_voidptrptr(&arg4), 0, 0);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "void *","svn_wc_get_diff_editor5", 4, argv[3] )); 
  }
  {
    arg5 = svn_swig_rb_to_depth(argv[4]);
  }
  arg6 = RTEST(argv[5]);
  arg7 = RTEST(argv[6]);
  arg8 = RTEST(argv[7]);
  {
    arg9 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg10 = (void *)svn_swig_rb_make_baton(argv[8], _global_svn_swig_rb_pool);
  }
  res11 = SWIG_ConvertPtr(argv[9], &argp11,SWIGTYPE_p_apr_array_header_t, 0 |  0 );
  if (!SWIG_IsOK(res11)) {
    SWIG_exception_fail(SWIG_ArgError(res11), Ruby_Format_TypeError( "", "apr_array_header_t const *","svn_wc_get_diff_editor5", 11, argv[9] )); 
  }
  arg11 = (apr_array_header_t *)(argp11);
  if (argc > 10) {
    
  }
  {
    result = (svn_error_t *)svn_wc_get_diff_editor5(arg1,(char const *)arg2,(struct svn_wc_diff_callbacks3_t const *)arg3,arg4,arg5,arg6,arg7,arg8,arg9,arg10,(apr_array_header_t const *)arg11,(struct svn_delta_editor_t const **)arg12,arg13,arg14);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg10);
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg12, SWIGTYPE_p_svn_delta_editor_t, 0), 0);
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg13, SWIGTYPE_p_void, 0), 0);
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_get_diff_editor4(int argc, VALUE *argv, VALUE self) {
  svn_wc_adm_access_t *arg1 = 0 ;
  char *arg2 = 0 ;
  svn_wc_diff_callbacks2_t *arg3 = 0 ;
  void *arg4 = 0 ;
  svn_depth_t arg5 ;
  svn_boolean_t arg6 ;
  svn_boolean_t arg7 ;
  svn_boolean_t arg8 ;
  svn_cancel_func_t arg9 = 0 ;
  void *arg10 = 0 ;
  apr_array_header_t *arg11 = 0 ;
  svn_delta_editor_t **arg12 = 0 ;
  void **arg13 = 0 ;
  apr_pool_t *arg14 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  void *argp11 = 0 ;
  int res11 = 0 ;
  svn_delta_editor_t *temp12 ;
  void *temp13 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg14);
    _global_pool = arg14;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg12 = &temp12;
  arg13 = &temp13;
  if ((argc < 9) || (argc > 10)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 9)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_get_diff_editor4", 1, argv[0] )); 
  }
  arg1 = (svn_wc_adm_access_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_wc_get_diff_editor4", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  {
    arg3 = (svn_wc_diff_callbacks2_t *) svn_swig_rb_wc_diff_callbacks2();
    arg4 = (void *)svn_swig_rb_make_baton(argv[2], _global_svn_swig_rb_pool);
  }
  {
    arg5 = svn_swig_rb_to_depth(argv[3]);
  }
  arg6 = RTEST(argv[4]);
  arg7 = RTEST(argv[5]);
  arg8 = RTEST(argv[6]);
  {
    arg9 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg10 = (void *)svn_swig_rb_make_baton(argv[7], _global_svn_swig_rb_pool);
  }
  res11 = SWIG_ConvertPtr(argv[8], &argp11,SWIGTYPE_p_apr_array_header_t, 0 |  0 );
  if (!SWIG_IsOK(res11)) {
    SWIG_exception_fail(SWIG_ArgError(res11), Ruby_Format_TypeError( "", "apr_array_header_t const *","svn_wc_get_diff_editor4", 11, argv[8] )); 
  }
  arg11 = (apr_array_header_t *)(argp11);
  if (argc > 9) {
    
  }
  {
    result = (svn_error_t *)svn_wc_get_diff_editor4(arg1,(char const *)arg2,(struct svn_wc_diff_callbacks2_t const *)arg3,arg4,arg5,arg6,arg7,arg8,arg9,arg10,(apr_array_header_t const *)arg11,(struct svn_delta_editor_t const **)arg12,arg13,arg14);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg10);
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg12, SWIGTYPE_p_svn_delta_editor_t, 0), 0);
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg13, SWIGTYPE_p_void, 0), 0);
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_get_diff_editor3(int argc, VALUE *argv, VALUE self) {
  svn_wc_adm_access_t *arg1 = 0 ;
  char *arg2 = 0 ;
  svn_wc_diff_callbacks2_t *arg3 = 0 ;
  void *arg4 = 0 ;
  svn_boolean_t arg5 ;
  svn_boolean_t arg6 ;
  svn_boolean_t arg7 ;
  svn_boolean_t arg8 ;
  svn_cancel_func_t arg9 = 0 ;
  void *arg10 = 0 ;
  svn_delta_editor_t **arg11 = 0 ;
  void **arg12 = 0 ;
  apr_pool_t *arg13 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  svn_delta_editor_t *temp11 ;
  void *temp12 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg13);
    _global_pool = arg13;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg11 = &temp11;
  arg12 = &temp12;
  if ((argc < 8) || (argc > 9)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 8)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_get_diff_editor3", 1, argv[0] )); 
  }
  arg1 = (svn_wc_adm_access_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_wc_get_diff_editor3", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  {
    arg3 = (svn_wc_diff_callbacks2_t *) svn_swig_rb_wc_diff_callbacks2();
    arg4 = (void *)svn_swig_rb_make_baton(argv[2], _global_svn_swig_rb_pool);
  }
  arg5 = RTEST(argv[3]);
  arg6 = RTEST(argv[4]);
  arg7 = RTEST(argv[5]);
  arg8 = RTEST(argv[6]);
  {
    arg9 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg10 = (void *)svn_swig_rb_make_baton(argv[7], _global_svn_swig_rb_pool);
  }
  if (argc > 8) {
    
  }
  {
    result = (svn_error_t *)svn_wc_get_diff_editor3(arg1,(char const *)arg2,(struct svn_wc_diff_callbacks2_t const *)arg3,arg4,arg5,arg6,arg7,arg8,arg9,arg10,(struct svn_delta_editor_t const **)arg11,arg12,arg13);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg10);
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg11, SWIGTYPE_p_svn_delta_editor_t, 0), 0);
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg12, SWIGTYPE_p_void, 0), 0);
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_get_diff_editor2(int argc, VALUE *argv, VALUE self) {
  svn_wc_adm_access_t *arg1 = 0 ;
  char *arg2 = 0 ;
  svn_wc_diff_callbacks_t *arg3 = 0 ;
  void *arg4 = 0 ;
  svn_boolean_t arg5 ;
  svn_boolean_t arg6 ;
  svn_boolean_t arg7 ;
  svn_boolean_t arg8 ;
  svn_cancel_func_t arg9 = 0 ;
  void *arg10 = 0 ;
  svn_delta_editor_t **arg11 = 0 ;
  void **arg12 = 0 ;
  apr_pool_t *arg13 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  int res4 ;
  svn_delta_editor_t *temp11 ;
  void *temp12 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg13);
    _global_pool = arg13;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg11 = &temp11;
  arg12 = &temp12;
  if ((argc < 9) || (argc > 10)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 9)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_get_diff_editor2", 1, argv[0] )); 
  }
  arg1 = (svn_wc_adm_access_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_wc_get_diff_editor2", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_ConvertPtr(argv[2], &argp3,SWIGTYPE_p_svn_wc_diff_callbacks_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_wc_diff_callbacks_t const *","svn_wc_get_diff_editor2", 3, argv[2] )); 
  }
  arg3 = (svn_wc_diff_callbacks_t *)(argp3);
  res4 = SWIG_ConvertPtr(argv[3],SWIG_as_voidptrptr(&arg4), 0, 0);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "void *","svn_wc_get_diff_editor2", 4, argv[3] )); 
  }
  arg5 = RTEST(argv[4]);
  arg6 = RTEST(argv[5]);
  arg7 = RTEST(argv[6]);
  arg8 = RTEST(argv[7]);
  {
    arg9 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg10 = (void *)svn_swig_rb_make_baton(argv[8], _global_svn_swig_rb_pool);
  }
  if (argc > 9) {
    
  }
  {
    result = (svn_error_t *)svn_wc_get_diff_editor2(arg1,(char const *)arg2,(struct svn_wc_diff_callbacks_t const *)arg3,arg4,arg5,arg6,arg7,arg8,arg9,arg10,(struct svn_delta_editor_t const **)arg11,arg12,arg13);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg10);
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg11, SWIGTYPE_p_svn_delta_editor_t, 0), 0);
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg12, SWIGTYPE_p_void, 0), 0);
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_get_diff_editor(int argc, VALUE *argv, VALUE self) {
  svn_wc_adm_access_t *arg1 = 0 ;
  char *arg2 = 0 ;
  svn_wc_diff_callbacks_t *arg3 = 0 ;
  void *arg4 = 0 ;
  svn_boolean_t arg5 ;
  svn_boolean_t arg6 ;
  svn_boolean_t arg7 ;
  svn_cancel_func_t arg8 = 0 ;
  void *arg9 = 0 ;
  svn_delta_editor_t **arg10 = 0 ;
  void **arg11 = 0 ;
  apr_pool_t *arg12 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  int res4 ;
  svn_delta_editor_t *temp10 ;
  void *temp11 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg12);
    _global_pool = arg12;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg10 = &temp10;
  arg11 = &temp11;
  if ((argc < 8) || (argc > 9)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 8)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_get_diff_editor", 1, argv[0] )); 
  }
  arg1 = (svn_wc_adm_access_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_wc_get_diff_editor", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_ConvertPtr(argv[2], &argp3,SWIGTYPE_p_svn_wc_diff_callbacks_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_wc_diff_callbacks_t const *","svn_wc_get_diff_editor", 3, argv[2] )); 
  }
  arg3 = (svn_wc_diff_callbacks_t *)(argp3);
  res4 = SWIG_ConvertPtr(argv[3],SWIG_as_voidptrptr(&arg4), 0, 0);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "void *","svn_wc_get_diff_editor", 4, argv[3] )); 
  }
  arg5 = RTEST(argv[4]);
  arg6 = RTEST(argv[5]);
  arg7 = RTEST(argv[6]);
  {
    arg8 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg9 = (void *)svn_swig_rb_make_baton(argv[7], _global_svn_swig_rb_pool);
  }
  if (argc > 8) {
    
  }
  {
    result = (svn_error_t *)svn_wc_get_diff_editor(arg1,(char const *)arg2,(struct svn_wc_diff_callbacks_t const *)arg3,arg4,arg5,arg6,arg7,arg8,arg9,(struct svn_delta_editor_t const **)arg10,arg11,arg12);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg9);
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg10, SWIGTYPE_p_svn_delta_editor_t, 0), 0);
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg11, SWIGTYPE_p_void, 0), 0);
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff6(int argc, VALUE *argv, VALUE self) {
  svn_wc_context_t *arg1 = 0 ;
  char *arg2 = 0 ;
  svn_wc_diff_callbacks4_t *arg3 = 0 ;
  void *arg4 = 0 ;
  svn_depth_t arg5 ;
  svn_boolean_t arg6 ;
  svn_boolean_t arg7 ;
  svn_boolean_t arg8 ;
  apr_array_header_t *arg9 = 0 ;
  svn_cancel_func_t arg10 = 0 ;
  void *arg11 = 0 ;
  apr_pool_t *arg12 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  int res4 ;
  void *argp9 = 0 ;
  int res9 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg12);
    _global_pool = arg12;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 10) || (argc > 11)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 10)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_context_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_context_t *","svn_wc_diff6", 1, argv[0] )); 
  }
  arg1 = (svn_wc_context_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_wc_diff6", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_ConvertPtr(argv[2], &argp3,SWIGTYPE_p_svn_wc_diff_callbacks4_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_wc_diff_callbacks4_t const *","svn_wc_diff6", 3, argv[2] )); 
  }
  arg3 = (svn_wc_diff_callbacks4_t *)(argp3);
  res4 = SWIG_ConvertPtr(argv[3],SWIG_as_voidptrptr(&arg4), 0, 0);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "void *","svn_wc_diff6", 4, argv[3] )); 
  }
  {
    arg5 = svn_swig_rb_to_depth(argv[4]);
  }
  arg6 = RTEST(argv[5]);
  arg7 = RTEST(argv[6]);
  arg8 = RTEST(argv[7]);
  res9 = SWIG_ConvertPtr(argv[8], &argp9,SWIGTYPE_p_apr_array_header_t, 0 |  0 );
  if (!SWIG_IsOK(res9)) {
    SWIG_exception_fail(SWIG_ArgError(res9), Ruby_Format_TypeError( "", "apr_array_header_t const *","svn_wc_diff6", 9, argv[8] )); 
  }
  arg9 = (apr_array_header_t *)(argp9);
  {
    arg10 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg11 = (void *)svn_swig_rb_make_baton(argv[9], _global_svn_swig_rb_pool);
  }
  if (argc > 10) {
    
  }
  {
    result = (svn_error_t *)svn_wc_diff6(arg1,(char const *)arg2,(struct svn_wc_diff_callbacks4_t const *)arg3,arg4,arg5,arg6,arg7,arg8,(apr_array_header_t const *)arg9,arg10,arg11,arg12);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg11);
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff5(int argc, VALUE *argv, VALUE self) {
  svn_wc_adm_access_t *arg1 = 0 ;
  char *arg2 = 0 ;
  svn_wc_diff_callbacks3_t *arg3 = 0 ;
  void *arg4 = 0 ;
  svn_depth_t arg5 ;
  svn_boolean_t arg6 ;
  apr_array_header_t *arg7 = 0 ;
  apr_pool_t *arg8 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  int res4 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg8);
    _global_pool = arg8;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 7) || (argc > 8)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 7)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_diff5", 1, argv[0] )); 
  }
  arg1 = (svn_wc_adm_access_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_wc_diff5", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_ConvertPtr(argv[2], &argp3,SWIGTYPE_p_svn_wc_diff_callbacks3_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_wc_diff_callbacks3_t const *","svn_wc_diff5", 3, argv[2] )); 
  }
  arg3 = (svn_wc_diff_callbacks3_t *)(argp3);
  res4 = SWIG_ConvertPtr(argv[3],SWIG_as_voidptrptr(&arg4), 0, 0);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "void *","svn_wc_diff5", 4, argv[3] )); 
  }
  {
    arg5 = svn_swig_rb_to_depth(argv[4]);
  }
  arg6 = RTEST(argv[5]);
  res7 = SWIG_ConvertPtr(argv[6], &argp7,SWIGTYPE_p_apr_array_header_t, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), Ruby_Format_TypeError( "", "apr_array_header_t const *","svn_wc_diff5", 7, argv[6] )); 
  }
  arg7 = (apr_array_header_t *)(argp7);
  if (argc > 7) {
    
  }
  {
    result = (svn_error_t *)svn_wc_diff5(arg1,(char const *)arg2,(struct svn_wc_diff_callbacks3_t const *)arg3,arg4,arg5,arg6,(apr_array_header_t const *)arg7,arg8);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff4(int argc, VALUE *argv, VALUE self) {
  svn_wc_adm_access_t *arg1 = 0 ;
  char *arg2 = 0 ;
  svn_wc_diff_callbacks2_t *arg3 = 0 ;
  void *arg4 = 0 ;
  svn_depth_t arg5 ;
  svn_boolean_t arg6 ;
  apr_array_header_t *arg7 = 0 ;
  apr_pool_t *arg8 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg8);
    _global_pool = arg8;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 6) || (argc > 7)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 6)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_diff4", 1, argv[0] )); 
  }
  arg1 = (svn_wc_adm_access_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_wc_diff4", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  {
    arg3 = (svn_wc_diff_callbacks2_t *) svn_swig_rb_wc_diff_callbacks2();
    arg4 = (void *)svn_swig_rb_make_baton(argv[2], _global_svn_swig_rb_pool);
  }
  {
    arg5 = svn_swig_rb_to_depth(argv[3]);
  }
  arg6 = RTEST(argv[4]);
  res7 = SWIG_ConvertPtr(argv[5], &argp7,SWIGTYPE_p_apr_array_header_t, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), Ruby_Format_TypeError( "", "apr_array_header_t const *","svn_wc_diff4", 7, argv[5] )); 
  }
  arg7 = (apr_array_header_t *)(argp7);
  if (argc > 6) {
    
  }
  {
    result = (svn_error_t *)svn_wc_diff4(arg1,(char const *)arg2,(struct svn_wc_diff_callbacks2_t const *)arg3,arg4,arg5,arg6,(apr_array_header_t const *)arg7,arg8);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff3(int argc, VALUE *argv, VALUE self) {
  svn_wc_adm_access_t *arg1 = 0 ;
  char *arg2 = 0 ;
  svn_wc_diff_callbacks2_t *arg3 = 0 ;
  void *arg4 = 0 ;
  svn_boolean_t arg5 ;
  svn_boolean_t arg6 ;
  apr_pool_t *arg7 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg7);
    _global_pool = arg7;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 5) || (argc > 6)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 5)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_diff3", 1, argv[0] )); 
  }
  arg1 = (svn_wc_adm_access_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_wc_diff3", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  {
    arg3 = (svn_wc_diff_callbacks2_t *) svn_swig_rb_wc_diff_callbacks2();
    arg4 = (void *)svn_swig_rb_make_baton(argv[2], _global_svn_swig_rb_pool);
  }
  arg5 = RTEST(argv[3]);
  arg6 = RTEST(argv[4]);
  if (argc > 5) {
    
  }
  {
    result = (svn_error_t *)svn_wc_diff3(arg1,(char const *)arg2,(struct svn_wc_diff_callbacks2_t const *)arg3,arg4,arg5,arg6,arg7);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff2(int argc, VALUE *argv, VALUE self) {
  svn_wc_adm_access_t *arg1 = 0 ;
  char *arg2 = 0 ;
  svn_wc_diff_callbacks_t *arg3 = 0 ;
  void *arg4 = 0 ;
  svn_boolean_t arg5 ;
  svn_boolean_t arg6 ;
  apr_pool_t *arg7 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  int res4 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg7);
    _global_pool = arg7;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 6) || (argc > 7)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 6)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_diff2", 1, argv[0] )); 
  }
  arg1 = (svn_wc_adm_access_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_wc_diff2", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_ConvertPtr(argv[2], &argp3,SWIGTYPE_p_svn_wc_diff_callbacks_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_wc_diff_callbacks_t const *","svn_wc_diff2", 3, argv[2] )); 
  }
  arg3 = (svn_wc_diff_callbacks_t *)(argp3);
  res4 = SWIG_ConvertPtr(argv[3],SWIG_as_voidptrptr(&arg4), 0, 0);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "void *","svn_wc_diff2", 4, argv[3] )); 
  }
  arg5 = RTEST(argv[4]);
  arg6 = RTEST(argv[5]);
  if (argc > 6) {
    
  }
  {
    result = (svn_error_t *)svn_wc_diff2(arg1,(char const *)arg2,(struct svn_wc_diff_callbacks_t const *)arg3,arg4,arg5,arg6,arg7);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff(int argc, VALUE *argv, VALUE self) {
  svn_wc_adm_access_t *arg1 = 0 ;
  char *arg2 = 0 ;
  svn_wc_diff_callbacks_t *arg3 = 0 ;
  void *arg4 = 0 ;
  svn_boolean_t arg5 ;
  apr_pool_t *arg6 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  int res4 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 5) || (argc > 6)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 5)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_diff", 1, argv[0] )); 
  }
  arg1 = (svn_wc_adm_access_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_wc_diff", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_ConvertPtr(argv[2], &argp3,SWIGTYPE_p_svn_wc_diff_callbacks_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_wc_diff_callbacks_t const *","svn_wc_diff", 3, argv[2] )); 
  }
  arg3 = (svn_wc_diff_callbacks_t *)(argp3);
  res4 = SWIG_ConvertPtr(argv[3],SWIG_as_voidptrptr(&arg4), 0, 0);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "void *","svn_wc_diff", 4, argv[3] )); 
  }
  arg5 = RTEST(argv[4]);
  if (argc > 5) {
    
  }
  {
    result = (svn_error_t *)svn_wc_diff(arg1,(char const *)arg2,(struct svn_wc_diff_callbacks_t const *)arg3,arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_get_prop_diffs2(int argc, VALUE *argv, VALUE self) {
  apr_array_header_t **arg1 = 0 ;
  apr_hash_t **arg2 = 0 ;
  svn_wc_context_t *arg3 = 0 ;
  char *arg4 = 0 ;
  apr_pool_t *arg5 = 0 ;
  apr_pool_t *arg6 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  apr_array_header_t *temp1 ;
  apr_hash_t *temp2 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  arg2 = &temp2;
  if ((argc < 2) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res3 = SWIG_ConvertPtr(argv[0], &argp3,SWIGTYPE_p_svn_wc_context_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_wc_context_t *","svn_wc_get_prop_diffs2", 3, argv[0] )); 
  }
  arg3 = (svn_wc_context_t *)(argp3);
  res4 = SWIG_AsCharPtrAndSize(argv[1], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_wc_get_prop_diffs2", 4, argv[1] ));
  }
  arg4 = (char *)(buf4);
  if (argc > 2) {
    
  }
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_wc_get_prop_diffs2(arg1,arg2,arg3,(char const *)arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, svn_swig_rb_prop_apr_array_to_hash_prop(*arg1), 0);
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, svn_swig_rb_apr_hash_to_hash_svn_string(*arg2), 0);
  }
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_get_prop_diffs(int argc, VALUE *argv, VALUE self) {
  apr_array_header_t **arg1 = 0 ;
  apr_hash_t **arg2 = 0 ;
  char *arg3 = 0 ;
  svn_wc_adm_access_t *arg4 = 0 ;
  apr_pool_t *arg5 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  apr_array_header_t *temp1 ;
  apr_hash_t *temp2 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  arg2 = &temp2;
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res3 = SWIG_AsCharPtrAndSize(argv[0], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_wc_get_prop_diffs", 3, argv[0] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_ConvertPtr(argv[1], &argp4,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_get_prop_diffs", 4, argv[1] )); 
  }
  arg4 = (svn_wc_adm_access_t *)(argp4);
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_wc_get_prop_diffs(arg1,arg2,(char const *)arg3,arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, svn_swig_rb_prop_apr_array_to_hash_prop(*arg1), 0);
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, svn_swig_rb_apr_hash_to_hash_svn_string(*arg2), 0);
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_merge5(int argc, VALUE *argv, VALUE self) {
  enum svn_wc_merge_outcome_t *arg1 = 0 ;
  enum svn_wc_notify_state_t *arg2 = 0 ;
  svn_wc_context_t *arg3 = 0 ;
  char *arg4 = 0 ;
  char *arg5 = 0 ;
  char *arg6 = 0 ;
  char *arg7 = 0 ;
  char *arg8 = 0 ;
  char *arg9 = 0 ;
  svn_wc_conflict_version_t *arg10 = 0 ;
  svn_wc_conflict_version_t *arg11 = 0 ;
  svn_boolean_t arg12 ;
  char *arg13 = 0 ;
  apr_array_header_t *arg14 = 0 ;
  apr_hash_t *arg15 = 0 ;
  apr_array_header_t *arg16 = 0 ;
  svn_wc_conflict_resolver_func2_t arg17 = 0 ;
  void *arg18 = 0 ;
  svn_cancel_func_t arg19 = 0 ;
  void *arg20 = 0 ;
  apr_pool_t *arg21 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  enum svn_wc_merge_outcome_t temp1 ;
  int res1 = SWIG_TMPOBJ ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  int res5 ;
  char *buf5 = 0 ;
  int alloc5 = 0 ;
  int res6 ;
  char *buf6 = 0 ;
  int alloc6 = 0 ;
  int res7 ;
  char *buf7 = 0 ;
  int alloc7 = 0 ;
  int res8 ;
  char *buf8 = 0 ;
  int alloc8 = 0 ;
  int res9 ;
  char *buf9 = 0 ;
  int alloc9 = 0 ;
  void *argp10 = 0 ;
  int res10 = 0 ;
  void *argp11 = 0 ;
  int res11 = 0 ;
  void *argp15 = 0 ;
  int res15 = 0 ;
  void *argp16 = 0 ;
  int res16 = 0 ;
  int res18 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg21);
    _global_pool = arg21;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 18) || (argc > 19)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 18)",argc); SWIG_fail;
  }
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_svn_wc_notify_state_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "enum svn_wc_notify_state_t *","svn_wc_merge5", 2, argv[0] )); 
  }
  arg2 = (enum svn_wc_notify_state_t *)(argp2);
  res3 = SWIG_ConvertPtr(argv[1], &argp3,SWIGTYPE_p_svn_wc_context_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_wc_context_t *","svn_wc_merge5", 3, argv[1] )); 
  }
  arg3 = (svn_wc_context_t *)(argp3);
  res4 = SWIG_AsCharPtrAndSize(argv[2], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_wc_merge5", 4, argv[2] ));
  }
  arg4 = (char *)(buf4);
  res5 = SWIG_AsCharPtrAndSize(argv[3], &buf5, NULL, &alloc5);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "char const *","svn_wc_merge5", 5, argv[3] ));
  }
  arg5 = (char *)(buf5);
  res6 = SWIG_AsCharPtrAndSize(argv[4], &buf6, NULL, &alloc6);
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "char const *","svn_wc_merge5", 6, argv[4] ));
  }
  arg6 = (char *)(buf6);
  res7 = SWIG_AsCharPtrAndSize(argv[5], &buf7, NULL, &alloc7);
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), Ruby_Format_TypeError( "", "char const *","svn_wc_merge5", 7, argv[5] ));
  }
  arg7 = (char *)(buf7);
  res8 = SWIG_AsCharPtrAndSize(argv[6], &buf8, NULL, &alloc8);
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), Ruby_Format_TypeError( "", "char const *","svn_wc_merge5", 8, argv[6] ));
  }
  arg8 = (char *)(buf8);
  res9 = SWIG_AsCharPtrAndSize(argv[7], &buf9, NULL, &alloc9);
  if (!SWIG_IsOK(res9)) {
    SWIG_exception_fail(SWIG_ArgError(res9), Ruby_Format_TypeError( "", "char const *","svn_wc_merge5", 9, argv[7] ));
  }
  arg9 = (char *)(buf9);
  res10 = SWIG_ConvertPtr(argv[8], &argp10,SWIGTYPE_p_svn_wc_conflict_version_t, 0 |  0 );
  if (!SWIG_IsOK(res10)) {
    SWIG_exception_fail(SWIG_ArgError(res10), Ruby_Format_TypeError( "", "svn_wc_conflict_version_t const *","svn_wc_merge5", 10, argv[8] )); 
  }
  arg10 = (svn_wc_conflict_version_t *)(argp10);
  res11 = SWIG_ConvertPtr(argv[9], &argp11,SWIGTYPE_p_svn_wc_conflict_version_t, 0 |  0 );
  if (!SWIG_IsOK(res11)) {
    SWIG_exception_fail(SWIG_ArgError(res11), Ruby_Format_TypeError( "", "svn_wc_conflict_version_t const *","svn_wc_merge5", 11, argv[9] )); 
  }
  arg11 = (svn_wc_conflict_version_t *)(argp11);
  arg12 = RTEST(argv[10]);
  {
    if (NIL_P(argv[11])) {
      arg13 = NULL;
    } else {
      arg13 = StringValuePtr(argv[11]);
    }
  }
  {
    if (NIL_P(argv[12])) {
      arg14 = NULL;
    } else {
      VALUE rb_pool;
      apr_pool_t *pool;
      
      svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &pool);
      arg14 = svn_swig_rb_to_apr_array_prop(argv[12], pool);
    }
  }
  res15 = SWIG_ConvertPtr(argv[13], &argp15,SWIGTYPE_p_apr_hash_t, 0 |  0 );
  if (!SWIG_IsOK(res15)) {
    SWIG_exception_fail(SWIG_ArgError(res15), Ruby_Format_TypeError( "", "apr_hash_t *","svn_wc_merge5", 15, argv[13] )); 
  }
  arg15 = (apr_hash_t *)(argp15);
  res16 = SWIG_ConvertPtr(argv[14], &argp16,SWIGTYPE_p_apr_array_header_t, 0 |  0 );
  if (!SWIG_IsOK(res16)) {
    SWIG_exception_fail(SWIG_ArgError(res16), Ruby_Format_TypeError( "", "apr_array_header_t const *","svn_wc_merge5", 16, argv[14] )); 
  }
  arg16 = (apr_array_header_t *)(argp16);
  {
    int res = SWIG_ConvertFunctionPtr(argv[15], (void**)(&arg17), SWIGTYPE_p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description2_t_p_void_p_apr_pool_t_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_wc_conflict_resolver_func2_t","svn_wc_merge5", 17, argv[15] )); 
    }
  }
  res18 = SWIG_ConvertPtr(argv[16],SWIG_as_voidptrptr(&arg18), 0, 0);
  if (!SWIG_IsOK(res18)) {
    SWIG_exception_fail(SWIG_ArgError(res18), Ruby_Format_TypeError( "", "void *","svn_wc_merge5", 18, argv[16] )); 
  }
  {
    arg19 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg20 = (void *)svn_swig_rb_make_baton(argv[17], _global_svn_swig_rb_pool);
  }
  if (argc > 18) {
    
  }
  {
    result = (svn_error_t *)svn_wc_merge5(arg1,arg2,arg3,(char const *)arg4,(char const *)arg5,(char const *)arg6,(char const *)arg7,(char const *)arg8,(char const *)arg9,(struct svn_wc_conflict_version_t const *)arg10,(struct svn_wc_conflict_version_t const *)arg11,arg12,(char const *)arg13,(apr_array_header_t const *)arg14,arg15,(apr_array_header_t const *)arg16,arg17,arg18,arg19,arg20,arg21);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (SWIG_IsTmpObj(res1)) {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_From_int((*arg1)), 0);
  } else {
    int new_flags = SWIG_IsNewObj(res1) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj((void*)(arg1), SWIGTYPE_p_svn_wc_merge_outcome_t, new_flags), 0);
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg20);
  }
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  if (alloc7 == SWIG_NEWOBJ) free((char*)buf7);
  if (alloc8 == SWIG_NEWOBJ) free((char*)buf8);
  if (alloc9 == SWIG_NEWOBJ) free((char*)buf9);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  if (alloc7 == SWIG_NEWOBJ) free((char*)buf7);
  if (alloc8 == SWIG_NEWOBJ) free((char*)buf8);
  if (alloc9 == SWIG_NEWOBJ) free((char*)buf9);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_merge4(int argc, VALUE *argv, VALUE self) {
  enum svn_wc_merge_outcome_t *arg1 = 0 ;
  svn_wc_context_t *arg2 = 0 ;
  char *arg3 = 0 ;
  char *arg4 = 0 ;
  char *arg5 = 0 ;
  char *arg6 = 0 ;
  char *arg7 = 0 ;
  char *arg8 = 0 ;
  svn_wc_conflict_version_t *arg9 = 0 ;
  svn_wc_conflict_version_t *arg10 = 0 ;
  svn_boolean_t arg11 ;
  char *arg12 = 0 ;
  apr_array_header_t *arg13 = 0 ;
  apr_array_header_t *arg14 = 0 ;
  svn_wc_conflict_resolver_func2_t arg15 = 0 ;
  void *arg16 = 0 ;
  svn_cancel_func_t arg17 = 0 ;
  void *arg18 = 0 ;
  apr_pool_t *arg19 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  enum svn_wc_merge_outcome_t temp1 ;
  int res1 = SWIG_TMPOBJ ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  int res5 ;
  char *buf5 = 0 ;
  int alloc5 = 0 ;
  int res6 ;
  char *buf6 = 0 ;
  int alloc6 = 0 ;
  int res7 ;
  char *buf7 = 0 ;
  int alloc7 = 0 ;
  int res8 ;
  char *buf8 = 0 ;
  int alloc8 = 0 ;
  void *argp9 = 0 ;
  int res9 = 0 ;
  void *argp10 = 0 ;
  int res10 = 0 ;
  void *argp14 = 0 ;
  int res14 = 0 ;
  int res16 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg19);
    _global_pool = arg19;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 16) || (argc > 17)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 16)",argc); SWIG_fail;
  }
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_svn_wc_context_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_context_t *","svn_wc_merge4", 2, argv[0] )); 
  }
  arg2 = (svn_wc_context_t *)(argp2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_wc_merge4", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_AsCharPtrAndSize(argv[2], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_wc_merge4", 4, argv[2] ));
  }
  arg4 = (char *)(buf4);
  res5 = SWIG_AsCharPtrAndSize(argv[3], &buf5, NULL, &alloc5);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "char const *","svn_wc_merge4", 5, argv[3] ));
  }
  arg5 = (char *)(buf5);
  res6 = SWIG_AsCharPtrAndSize(argv[4], &buf6, NULL, &alloc6);
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "char const *","svn_wc_merge4", 6, argv[4] ));
  }
  arg6 = (char *)(buf6);
  res7 = SWIG_AsCharPtrAndSize(argv[5], &buf7, NULL, &alloc7);
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), Ruby_Format_TypeError( "", "char const *","svn_wc_merge4", 7, argv[5] ));
  }
  arg7 = (char *)(buf7);
  res8 = SWIG_AsCharPtrAndSize(argv[6], &buf8, NULL, &alloc8);
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), Ruby_Format_TypeError( "", "char const *","svn_wc_merge4", 8, argv[6] ));
  }
  arg8 = (char *)(buf8);
  res9 = SWIG_ConvertPtr(argv[7], &argp9,SWIGTYPE_p_svn_wc_conflict_version_t, 0 |  0 );
  if (!SWIG_IsOK(res9)) {
    SWIG_exception_fail(SWIG_ArgError(res9), Ruby_Format_TypeError( "", "svn_wc_conflict_version_t const *","svn_wc_merge4", 9, argv[7] )); 
  }
  arg9 = (svn_wc_conflict_version_t *)(argp9);
  res10 = SWIG_ConvertPtr(argv[8], &argp10,SWIGTYPE_p_svn_wc_conflict_version_t, 0 |  0 );
  if (!SWIG_IsOK(res10)) {
    SWIG_exception_fail(SWIG_ArgError(res10), Ruby_Format_TypeError( "", "svn_wc_conflict_version_t const *","svn_wc_merge4", 10, argv[8] )); 
  }
  arg10 = (svn_wc_conflict_version_t *)(argp10);
  arg11 = RTEST(argv[9]);
  {
    if (NIL_P(argv[10])) {
      arg12 = NULL;
    } else {
      arg12 = StringValuePtr(argv[10]);
    }
  }
  {
    if (NIL_P(argv[11])) {
      arg13 = NULL;
    } else {
      VALUE rb_pool;
      apr_pool_t *pool;
      
      svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &pool);
      arg13 = svn_swig_rb_to_apr_array_prop(argv[11], pool);
    }
  }
  res14 = SWIG_ConvertPtr(argv[12], &argp14,SWIGTYPE_p_apr_array_header_t, 0 |  0 );
  if (!SWIG_IsOK(res14)) {
    SWIG_exception_fail(SWIG_ArgError(res14), Ruby_Format_TypeError( "", "apr_array_header_t const *","svn_wc_merge4", 14, argv[12] )); 
  }
  arg14 = (apr_array_header_t *)(argp14);
  {
    int res = SWIG_ConvertFunctionPtr(argv[13], (void**)(&arg15), SWIGTYPE_p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description2_t_p_void_p_apr_pool_t_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_wc_conflict_resolver_func2_t","svn_wc_merge4", 15, argv[13] )); 
    }
  }
  res16 = SWIG_ConvertPtr(argv[14],SWIG_as_voidptrptr(&arg16), 0, 0);
  if (!SWIG_IsOK(res16)) {
    SWIG_exception_fail(SWIG_ArgError(res16), Ruby_Format_TypeError( "", "void *","svn_wc_merge4", 16, argv[14] )); 
  }
  {
    arg17 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg18 = (void *)svn_swig_rb_make_baton(argv[15], _global_svn_swig_rb_pool);
  }
  if (argc > 16) {
    
  }
  {
    result = (svn_error_t *)svn_wc_merge4(arg1,arg2,(char const *)arg3,(char const *)arg4,(char const *)arg5,(char const *)arg6,(char const *)arg7,(char const *)arg8,(struct svn_wc_conflict_version_t const *)arg9,(struct svn_wc_conflict_version_t const *)arg10,arg11,(char const *)arg12,(apr_array_header_t const *)arg13,(apr_array_header_t const *)arg14,arg15,arg16,arg17,arg18,arg19);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (SWIG_IsTmpObj(res1)) {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_From_int((*arg1)), 0);
  } else {
    int new_flags = SWIG_IsNewObj(res1) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj((void*)(arg1), SWIGTYPE_p_svn_wc_merge_outcome_t, new_flags), 0);
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg18);
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  if (alloc7 == SWIG_NEWOBJ) free((char*)buf7);
  if (alloc8 == SWIG_NEWOBJ) free((char*)buf8);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  if (alloc7 == SWIG_NEWOBJ) free((char*)buf7);
  if (alloc8 == SWIG_NEWOBJ) free((char*)buf8);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_merge3(int argc, VALUE *argv, VALUE self) {
  enum svn_wc_merge_outcome_t *arg1 = 0 ;
  char *arg2 = 0 ;
  char *arg3 = 0 ;
  char *arg4 = 0 ;
  svn_wc_adm_access_t *arg5 = 0 ;
  char *arg6 = 0 ;
  char *arg7 = 0 ;
  char *arg8 = 0 ;
  svn_boolean_t arg9 ;
  char *arg10 = 0 ;
  apr_array_header_t *arg11 = 0 ;
  apr_array_header_t *arg12 = 0 ;
  svn_wc_conflict_resolver_func_t arg13 = 0 ;
  void *arg14 = 0 ;
  apr_pool_t *arg15 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  enum svn_wc_merge_outcome_t temp1 ;
  int res1 = SWIG_TMPOBJ ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  int res6 ;
  char *buf6 = 0 ;
  int alloc6 = 0 ;
  int res7 ;
  char *buf7 = 0 ;
  int alloc7 = 0 ;
  int res8 ;
  char *buf8 = 0 ;
  int alloc8 = 0 ;
  void *argp12 = 0 ;
  int res12 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg15);
    _global_pool = arg15;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 12) || (argc > 13)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 12)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_wc_merge3", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_wc_merge3", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_AsCharPtrAndSize(argv[2], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_wc_merge3", 4, argv[2] ));
  }
  arg4 = (char *)(buf4);
  res5 = SWIG_ConvertPtr(argv[3], &argp5,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_merge3", 5, argv[3] )); 
  }
  arg5 = (svn_wc_adm_access_t *)(argp5);
  res6 = SWIG_AsCharPtrAndSize(argv[4], &buf6, NULL, &alloc6);
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "char const *","svn_wc_merge3", 6, argv[4] ));
  }
  arg6 = (char *)(buf6);
  res7 = SWIG_AsCharPtrAndSize(argv[5], &buf7, NULL, &alloc7);
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), Ruby_Format_TypeError( "", "char const *","svn_wc_merge3", 7, argv[5] ));
  }
  arg7 = (char *)(buf7);
  res8 = SWIG_AsCharPtrAndSize(argv[6], &buf8, NULL, &alloc8);
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), Ruby_Format_TypeError( "", "char const *","svn_wc_merge3", 8, argv[6] ));
  }
  arg8 = (char *)(buf8);
  arg9 = RTEST(argv[7]);
  {
    if (NIL_P(argv[8])) {
      arg10 = NULL;
    } else {
      arg10 = StringValuePtr(argv[8]);
    }
  }
  {
    if (NIL_P(argv[9])) {
      arg11 = NULL;
    } else {
      VALUE rb_pool;
      apr_pool_t *pool;
      
      svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &pool);
      arg11 = svn_swig_rb_to_apr_array_prop(argv[9], pool);
    }
  }
  res12 = SWIG_ConvertPtr(argv[10], &argp12,SWIGTYPE_p_apr_array_header_t, 0 |  0 );
  if (!SWIG_IsOK(res12)) {
    SWIG_exception_fail(SWIG_ArgError(res12), Ruby_Format_TypeError( "", "apr_array_header_t const *","svn_wc_merge3", 12, argv[10] )); 
  }
  arg12 = (apr_array_header_t *)(argp12);
  {
    arg13 = svn_swig_rb_conflict_resolver_func;
    arg14 = (void *)svn_swig_rb_make_baton(argv[11], _global_svn_swig_rb_pool);
  }
  if (argc > 12) {
    
  }
  {
    result = (svn_error_t *)svn_wc_merge3(arg1,(char const *)arg2,(char const *)arg3,(char const *)arg4,arg5,(char const *)arg6,(char const *)arg7,(char const *)arg8,arg9,(char const *)arg10,(apr_array_header_t const *)arg11,(apr_array_header_t const *)arg12,arg13,arg14,arg15);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (SWIG_IsTmpObj(res1)) {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_From_int((*arg1)), 0);
  } else {
    int new_flags = SWIG_IsNewObj(res1) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj((void*)(arg1), SWIGTYPE_p_svn_wc_merge_outcome_t, new_flags), 0);
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg14);
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  if (alloc7 == SWIG_NEWOBJ) free((char*)buf7);
  if (alloc8 == SWIG_NEWOBJ) free((char*)buf8);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  if (alloc7 == SWIG_NEWOBJ) free((char*)buf7);
  if (alloc8 == SWIG_NEWOBJ) free((char*)buf8);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_merge2(int argc, VALUE *argv, VALUE self) {
  enum svn_wc_merge_outcome_t *arg1 = 0 ;
  char *arg2 = 0 ;
  char *arg3 = 0 ;
  char *arg4 = 0 ;
  svn_wc_adm_access_t *arg5 = 0 ;
  char *arg6 = 0 ;
  char *arg7 = 0 ;
  char *arg8 = 0 ;
  svn_boolean_t arg9 ;
  char *arg10 = 0 ;
  apr_array_header_t *arg11 = 0 ;
  apr_pool_t *arg12 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  enum svn_wc_merge_outcome_t temp1 ;
  int res1 = SWIG_TMPOBJ ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  int res6 ;
  char *buf6 = 0 ;
  int alloc6 = 0 ;
  int res7 ;
  char *buf7 = 0 ;
  int alloc7 = 0 ;
  int res8 ;
  char *buf8 = 0 ;
  int alloc8 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg12);
    _global_pool = arg12;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 10) || (argc > 11)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 10)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_wc_merge2", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_wc_merge2", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_AsCharPtrAndSize(argv[2], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_wc_merge2", 4, argv[2] ));
  }
  arg4 = (char *)(buf4);
  res5 = SWIG_ConvertPtr(argv[3], &argp5,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_merge2", 5, argv[3] )); 
  }
  arg5 = (svn_wc_adm_access_t *)(argp5);
  res6 = SWIG_AsCharPtrAndSize(argv[4], &buf6, NULL, &alloc6);
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "char const *","svn_wc_merge2", 6, argv[4] ));
  }
  arg6 = (char *)(buf6);
  res7 = SWIG_AsCharPtrAndSize(argv[5], &buf7, NULL, &alloc7);
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), Ruby_Format_TypeError( "", "char const *","svn_wc_merge2", 7, argv[5] ));
  }
  arg7 = (char *)(buf7);
  res8 = SWIG_AsCharPtrAndSize(argv[6], &buf8, NULL, &alloc8);
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), Ruby_Format_TypeError( "", "char const *","svn_wc_merge2", 8, argv[6] ));
  }
  arg8 = (char *)(buf8);
  arg9 = RTEST(argv[7]);
  {
    if (NIL_P(argv[8])) {
      arg10 = NULL;
    } else {
      arg10 = StringValuePtr(argv[8]);
    }
  }
  {
    if (NIL_P(argv[9])) {
      arg11 = NULL;
    } else {
      VALUE rb_pool;
      apr_pool_t *pool;
      
      svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &pool);
      arg11 = svn_swig_rb_to_apr_array_prop(argv[9], pool);
    }
  }
  if (argc > 10) {
    
  }
  {
    result = (svn_error_t *)svn_wc_merge2(arg1,(char const *)arg2,(char const *)arg3,(char const *)arg4,arg5,(char const *)arg6,(char const *)arg7,(char const *)arg8,arg9,(char const *)arg10,(apr_array_header_t const *)arg11,arg12);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (SWIG_IsTmpObj(res1)) {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_From_int((*arg1)), 0);
  } else {
    int new_flags = SWIG_IsNewObj(res1) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj((void*)(arg1), SWIGTYPE_p_svn_wc_merge_outcome_t, new_flags), 0);
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  if (alloc7 == SWIG_NEWOBJ) free((char*)buf7);
  if (alloc8 == SWIG_NEWOBJ) free((char*)buf8);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  if (alloc7 == SWIG_NEWOBJ) free((char*)buf7);
  if (alloc8 == SWIG_NEWOBJ) free((char*)buf8);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_merge(int argc, VALUE *argv, VALUE self) {
  char *arg1 = 0 ;
  char *arg2 = 0 ;
  char *arg3 = 0 ;
  svn_wc_adm_access_t *arg4 = 0 ;
  char *arg5 = 0 ;
  char *arg6 = 0 ;
  char *arg7 = 0 ;
  svn_boolean_t arg8 ;
  enum svn_wc_merge_outcome_t *arg9 = 0 ;
  char *arg10 = 0 ;
  apr_pool_t *arg11 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  int res5 ;
  char *buf5 = 0 ;
  int alloc5 = 0 ;
  int res6 ;
  char *buf6 = 0 ;
  int alloc6 = 0 ;
  int res7 ;
  char *buf7 = 0 ;
  int alloc7 = 0 ;
  enum svn_wc_merge_outcome_t temp9 ;
  int res9 = SWIG_TMPOBJ ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg11);
    _global_pool = arg11;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg9 = &temp9;
  if ((argc < 9) || (argc > 10)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 9)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_wc_merge", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_wc_merge", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_wc_merge", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_ConvertPtr(argv[3], &argp4,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_merge", 4, argv[3] )); 
  }
  arg4 = (svn_wc_adm_access_t *)(argp4);
  res5 = SWIG_AsCharPtrAndSize(argv[4], &buf5, NULL, &alloc5);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "char const *","svn_wc_merge", 5, argv[4] ));
  }
  arg5 = (char *)(buf5);
  res6 = SWIG_AsCharPtrAndSize(argv[5], &buf6, NULL, &alloc6);
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "char const *","svn_wc_merge", 6, argv[5] ));
  }
  arg6 = (char *)(buf6);
  res7 = SWIG_AsCharPtrAndSize(argv[6], &buf7, NULL, &alloc7);
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), Ruby_Format_TypeError( "", "char const *","svn_wc_merge", 7, argv[6] ));
  }
  arg7 = (char *)(buf7);
  arg8 = RTEST(argv[7]);
  {
    if (NIL_P(argv[8])) {
      arg10 = NULL;
    } else {
      arg10 = StringValuePtr(argv[8]);
    }
  }
  if (argc > 9) {
    
  }
  {
    result = (svn_error_t *)svn_wc_merge((char const *)arg1,(char const *)arg2,(char const *)arg3,arg4,(char const *)arg5,(char const *)arg6,(char const *)arg7,arg8,arg9,(char const *)arg10,arg11);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (SWIG_IsTmpObj(res9)) {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_From_int((*arg9)), 0);
  } else {
    int new_flags = SWIG_IsNewObj(res9) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj((void*)(arg9), SWIGTYPE_p_svn_wc_merge_outcome_t, new_flags), 0);
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  if (alloc7 == SWIG_NEWOBJ) free((char*)buf7);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  if (alloc7 == SWIG_NEWOBJ) free((char*)buf7);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_merge_props3(int argc, VALUE *argv, VALUE self) {
  svn_wc_notify_state_t *arg1 = 0 ;
  svn_wc_context_t *arg2 = 0 ;
  char *arg3 = 0 ;
  svn_wc_conflict_version_t *arg4 = 0 ;
  svn_wc_conflict_version_t *arg5 = 0 ;
  apr_hash_t *arg6 = 0 ;
  apr_array_header_t *arg7 = 0 ;
  svn_boolean_t arg8 ;
  svn_wc_conflict_resolver_func2_t arg9 = 0 ;
  void *arg10 = 0 ;
  svn_cancel_func_t arg11 = 0 ;
  void *arg12 = 0 ;
  apr_pool_t *arg13 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  int res10 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg13);
    _global_pool = arg13;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 11) || (argc > 12)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 11)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_notify_state_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_notify_state_t *","svn_wc_merge_props3", 1, argv[0] )); 
  }
  arg1 = (svn_wc_notify_state_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_context_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_context_t *","svn_wc_merge_props3", 2, argv[1] )); 
  }
  arg2 = (svn_wc_context_t *)(argp2);
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_wc_merge_props3", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_ConvertPtr(argv[3], &argp4,SWIGTYPE_p_svn_wc_conflict_version_t, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "svn_wc_conflict_version_t const *","svn_wc_merge_props3", 4, argv[3] )); 
  }
  arg4 = (svn_wc_conflict_version_t *)(argp4);
  res5 = SWIG_ConvertPtr(argv[4], &argp5,SWIGTYPE_p_svn_wc_conflict_version_t, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "svn_wc_conflict_version_t const *","svn_wc_merge_props3", 5, argv[4] )); 
  }
  arg5 = (svn_wc_conflict_version_t *)(argp5);
  {
    VALUE rb_pool = Qnil;
    if (!_global_pool) {
      svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &_global_pool);
      svn_swig_rb_push_pool(rb_pool);
    }
    arg6 = svn_swig_rb_hash_to_apr_hash_svn_string(argv[5], _global_pool);
    _global_pool = NULL;
    if (!NIL_P(rb_pool)) {
      if (arg6 == NULL) {
        svn_swig_rb_destroy_pool(rb_pool);
      } else {
        svn_swig_rb_set_pool_for_no_swig_type(argv[5], rb_pool);
      }
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
    }
  }
  {
    VALUE rb_pool;
    apr_pool_t *pool;
    
    svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &pool);
    
    arg7 = svn_swig_rb_to_apr_array_prop(argv[6], pool);
  }
  arg8 = RTEST(argv[7]);
  {
    int res = SWIG_ConvertFunctionPtr(argv[8], (void**)(&arg9), SWIGTYPE_p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description2_t_p_void_p_apr_pool_t_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_wc_conflict_resolver_func2_t","svn_wc_merge_props3", 9, argv[8] )); 
    }
  }
  res10 = SWIG_ConvertPtr(argv[9],SWIG_as_voidptrptr(&arg10), 0, 0);
  if (!SWIG_IsOK(res10)) {
    SWIG_exception_fail(SWIG_ArgError(res10), Ruby_Format_TypeError( "", "void *","svn_wc_merge_props3", 10, argv[9] )); 
  }
  {
    arg11 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg12 = (void *)svn_swig_rb_make_baton(argv[10], _global_svn_swig_rb_pool);
  }
  if (argc > 11) {
    
  }
  {
    result = (svn_error_t *)svn_wc_merge_props3(arg1,arg2,(char const *)arg3,(struct svn_wc_conflict_version_t const *)arg4,(struct svn_wc_conflict_version_t const *)arg5,arg6,(apr_array_header_t const *)arg7,arg8,arg9,arg10,arg11,arg12,arg13);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg12);
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_merge_props2(int argc, VALUE *argv, VALUE self) {
  svn_wc_notify_state_t *arg1 = 0 ;
  char *arg2 = 0 ;
  svn_wc_adm_access_t *arg3 = 0 ;
  apr_hash_t *arg4 = 0 ;
  apr_array_header_t *arg5 = 0 ;
  svn_boolean_t arg6 ;
  svn_boolean_t arg7 ;
  svn_wc_conflict_resolver_func_t arg8 = 0 ;
  void *arg9 = 0 ;
  apr_pool_t *arg10 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg10);
    _global_pool = arg10;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 8) || (argc > 9)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 8)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_notify_state_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_notify_state_t *","svn_wc_merge_props2", 1, argv[0] )); 
  }
  arg1 = (svn_wc_notify_state_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_wc_merge_props2", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_ConvertPtr(argv[2], &argp3,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_merge_props2", 3, argv[2] )); 
  }
  arg3 = (svn_wc_adm_access_t *)(argp3);
  {
    VALUE rb_pool = Qnil;
    if (!_global_pool) {
      svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &_global_pool);
      svn_swig_rb_push_pool(rb_pool);
    }
    arg4 = svn_swig_rb_hash_to_apr_hash_svn_string(argv[3], _global_pool);
    _global_pool = NULL;
    if (!NIL_P(rb_pool)) {
      if (arg4 == NULL) {
        svn_swig_rb_destroy_pool(rb_pool);
      } else {
        svn_swig_rb_set_pool_for_no_swig_type(argv[3], rb_pool);
      }
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
    }
  }
  {
    VALUE rb_pool;
    apr_pool_t *pool;
    
    svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &pool);
    
    arg5 = svn_swig_rb_to_apr_array_prop(argv[4], pool);
  }
  arg6 = RTEST(argv[5]);
  arg7 = RTEST(argv[6]);
  {
    arg8 = svn_swig_rb_conflict_resolver_func;
    arg9 = (void *)svn_swig_rb_make_baton(argv[7], _global_svn_swig_rb_pool);
  }
  if (argc > 8) {
    
  }
  {
    result = (svn_error_t *)svn_wc_merge_props2(arg1,(char const *)arg2,arg3,arg4,(apr_array_header_t const *)arg5,arg6,arg7,arg8,arg9,arg10);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg9);
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_merge_props(int argc, VALUE *argv, VALUE self) {
  svn_wc_notify_state_t *arg1 = 0 ;
  char *arg2 = 0 ;
  svn_wc_adm_access_t *arg3 = 0 ;
  apr_hash_t *arg4 = 0 ;
  apr_array_header_t *arg5 = 0 ;
  svn_boolean_t arg6 ;
  svn_boolean_t arg7 ;
  apr_pool_t *arg8 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg8);
    _global_pool = arg8;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 7) || (argc > 8)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 7)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_notify_state_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_notify_state_t *","svn_wc_merge_props", 1, argv[0] )); 
  }
  arg1 = (svn_wc_notify_state_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_wc_merge_props", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_ConvertPtr(argv[2], &argp3,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_merge_props", 3, argv[2] )); 
  }
  arg3 = (svn_wc_adm_access_t *)(argp3);
  {
    VALUE rb_pool = Qnil;
    if (!_global_pool) {
      svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &_global_pool);
      svn_swig_rb_push_pool(rb_pool);
    }
    arg4 = svn_swig_rb_hash_to_apr_hash_svn_string(argv[3], _global_pool);
    _global_pool = NULL;
    if (!NIL_P(rb_pool)) {
      if (arg4 == NULL) {
        svn_swig_rb_destroy_pool(rb_pool);
      } else {
        svn_swig_rb_set_pool_for_no_swig_type(argv[3], rb_pool);
      }
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
    }
  }
  {
    VALUE rb_pool;
    apr_pool_t *pool;
    
    svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &pool);
    
    arg5 = svn_swig_rb_to_apr_array_prop(argv[4], pool);
  }
  arg6 = RTEST(argv[5]);
  arg7 = RTEST(argv[6]);
  if (argc > 7) {
    
  }
  {
    result = (svn_error_t *)svn_wc_merge_props(arg1,(char const *)arg2,arg3,arg4,(apr_array_header_t const *)arg5,arg6,arg7,arg8);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_merge_prop_diffs(int argc, VALUE *argv, VALUE self) {
  svn_wc_notify_state_t *arg1 = 0 ;
  char *arg2 = 0 ;
  svn_wc_adm_access_t *arg3 = 0 ;
  apr_array_header_t *arg4 = 0 ;
  svn_boolean_t arg5 ;
  svn_boolean_t arg6 ;
  apr_pool_t *arg7 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg7);
    _global_pool = arg7;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 6) || (argc > 7)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 6)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_notify_state_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_notify_state_t *","svn_wc_merge_prop_diffs", 1, argv[0] )); 
  }
  arg1 = (svn_wc_notify_state_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_wc_merge_prop_diffs", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_ConvertPtr(argv[2], &argp3,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_merge_prop_diffs", 3, argv[2] )); 
  }
  arg3 = (svn_wc_adm_access_t *)(argp3);
  {
    VALUE rb_pool;
    apr_pool_t *pool;
    
    svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &pool);
    
    arg4 = svn_swig_rb_to_apr_array_prop(argv[3], pool);
  }
  arg5 = RTEST(argv[4]);
  arg6 = RTEST(argv[5]);
  if (argc > 6) {
    
  }
  {
    result = (svn_error_t *)svn_wc_merge_prop_diffs(arg1,(char const *)arg2,arg3,(apr_array_header_t const *)arg4,arg5,arg6,arg7);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_get_pristine_contents2(int argc, VALUE *argv, VALUE self) {
  svn_stream_t **arg1 = 0 ;
  svn_wc_context_t *arg2 = 0 ;
  char *arg3 = 0 ;
  apr_pool_t *arg4 = 0 ;
  apr_pool_t *arg5 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_stream_t *temp1 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 2) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_svn_wc_context_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_context_t *","svn_wc_get_pristine_contents2", 2, argv[0] )); 
  }
  arg2 = (svn_wc_context_t *)(argp2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_wc_get_pristine_contents2", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  if (argc > 2) {
    
  }
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_wc_get_pristine_contents2(arg1,arg2,(char const *)arg3,arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_stream_t, 0), 0);
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_get_pristine_contents(int argc, VALUE *argv, VALUE self) {
  svn_stream_t **arg1 = 0 ;
  char *arg2 = 0 ;
  apr_pool_t *arg3 = 0 ;
  apr_pool_t *arg4 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_stream_t *temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 1) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_wc_get_pristine_contents", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  if (argc > 1) {
    
  }
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_wc_get_pristine_contents(arg1,(char const *)arg2,arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_stream_t, 0), 0);
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_get_pristine_copy_path(int argc, VALUE *argv, VALUE self) {
  char *arg1 = 0 ;
  char **arg2 = 0 ;
  apr_pool_t *arg3 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  char *temp2 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg2 = &temp2;
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_wc_get_pristine_copy_path", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  if (argc > 1) {
    
  }
  {
    result = (svn_error_t *)svn_wc_get_pristine_copy_path((char const *)arg1,(char const **)arg2,arg3);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    if (*arg2) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new2(*arg2), 0);
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil, 0);
    }
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_cleanup4(int argc, VALUE *argv, VALUE self) {
  svn_wc_context_t *arg1 = 0 ;
  char *arg2 = 0 ;
  svn_boolean_t arg3 ;
  svn_boolean_t arg4 ;
  svn_boolean_t arg5 ;
  svn_boolean_t arg6 ;
  svn_cancel_func_t arg7 = 0 ;
  void *arg8 = 0 ;
  svn_wc_notify_func2_t arg9 = 0 ;
  void *arg10 = 0 ;
  apr_pool_t *arg11 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg11);
    _global_pool = arg11;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 8) || (argc > 9)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 8)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_context_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_context_t *","svn_wc_cleanup4", 1, argv[0] )); 
  }
  arg1 = (svn_wc_context_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_wc_cleanup4", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  arg3 = RTEST(argv[2]);
  arg4 = RTEST(argv[3]);
  arg5 = RTEST(argv[4]);
  arg6 = RTEST(argv[5]);
  {
    arg7 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg8 = (void *)svn_swig_rb_make_baton(argv[6], _global_svn_swig_rb_pool);
  }
  {
    arg9 = (svn_wc_notify_func2_t) svn_swig_rb_notify_func2;
    arg10 = (void *)svn_swig_rb_make_baton(argv[7], _global_svn_swig_rb_pool);
  }
  if (argc > 8) {
    
  }
  {
    result = (svn_error_t *)svn_wc_cleanup4(arg1,(char const *)arg2,arg3,arg4,arg5,arg6,arg7,arg8,arg9,arg10,arg11);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg8);
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_cleanup3(int argc, VALUE *argv, VALUE self) {
  svn_wc_context_t *arg1 = 0 ;
  char *arg2 = 0 ;
  svn_cancel_func_t arg3 = 0 ;
  void *arg4 = 0 ;
  apr_pool_t *arg5 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 3) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_context_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_context_t *","svn_wc_cleanup3", 1, argv[0] )); 
  }
  arg1 = (svn_wc_context_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_wc_cleanup3", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  {
    arg3 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg4 = (void *)svn_swig_rb_make_baton(argv[2], _global_svn_swig_rb_pool);
  }
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_wc_cleanup3(arg1,(char const *)arg2,arg3,arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg4);
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_cleanup2(int argc, VALUE *argv, VALUE self) {
  char *arg1 = 0 ;
  char *arg2 = 0 ;
  svn_cancel_func_t arg3 = 0 ;
  void *arg4 = 0 ;
  apr_pool_t *arg5 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 3) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_wc_cleanup2", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    if (NIL_P(argv[1])) {
      arg2 = NULL;
    } else {
      arg2 = StringValuePtr(argv[1]);
    }
  }
  {
    arg3 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg4 = (void *)svn_swig_rb_make_baton(argv[2], _global_svn_swig_rb_pool);
  }
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_wc_cleanup2((char const *)arg1,(char const *)arg2,arg3,arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg4);
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_cleanup(int argc, VALUE *argv, VALUE self) {
  char *arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  char *arg3 = 0 ;
  svn_cancel_func_t arg4 = 0 ;
  void *arg5 = 0 ;
  apr_pool_t *arg6 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 4) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_wc_cleanup", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_cleanup", 2, argv[1] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  {
    if (NIL_P(argv[2])) {
      arg3 = NULL;
    } else {
      arg3 = StringValuePtr(argv[2]);
    }
  }
  {
    arg4 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg5 = (void *)svn_swig_rb_make_baton(argv[3], _global_svn_swig_rb_pool);
  }
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_wc_cleanup((char const *)arg1,arg2,(char const *)arg3,arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg5);
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_upgrade(int argc, VALUE *argv, VALUE self) {
  svn_wc_context_t *arg1 = 0 ;
  char *arg2 = 0 ;
  svn_wc_upgrade_get_repos_info_t arg3 = 0 ;
  void *arg4 = 0 ;
  svn_cancel_func_t arg5 = 0 ;
  void *arg6 = 0 ;
  svn_wc_notify_func2_t arg7 = 0 ;
  void *arg8 = 0 ;
  apr_pool_t *arg9 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res4 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg9);
    _global_pool = arg9;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 6) || (argc > 7)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 6)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_context_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_context_t *","svn_wc_upgrade", 1, argv[0] )); 
  }
  arg1 = (svn_wc_context_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_wc_upgrade", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  {
    int res = SWIG_ConvertFunctionPtr(argv[2], (void**)(&arg3), SWIGTYPE_p_f_p_p_q_const__char_p_p_q_const__char_p_void_p_q_const__char_p_apr_pool_t_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_wc_upgrade_get_repos_info_t","svn_wc_upgrade", 3, argv[2] )); 
    }
  }
  res4 = SWIG_ConvertPtr(argv[3],SWIG_as_voidptrptr(&arg4), 0, 0);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "void *","svn_wc_upgrade", 4, argv[3] )); 
  }
  {
    arg5 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg6 = (void *)svn_swig_rb_make_baton(argv[4], _global_svn_swig_rb_pool);
  }
  {
    arg7 = (svn_wc_notify_func2_t) svn_swig_rb_notify_func2;
    arg8 = (void *)svn_swig_rb_make_baton(argv[5], _global_svn_swig_rb_pool);
  }
  if (argc > 6) {
    
  }
  {
    result = (svn_error_t *)svn_wc_upgrade(arg1,(char const *)arg2,arg3,arg4,arg5,arg6,arg7,arg8,arg9);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg6);
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_relocate4(int argc, VALUE *argv, VALUE self) {
  svn_wc_context_t *arg1 = 0 ;
  char *arg2 = 0 ;
  char *arg3 = 0 ;
  char *arg4 = 0 ;
  svn_wc_relocation_validator3_t arg5 = 0 ;
  void *arg6 = 0 ;
  apr_pool_t *arg7 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg7);
    _global_pool = arg7;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 5) || (argc > 6)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 5)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_context_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_context_t *","svn_wc_relocate4", 1, argv[0] )); 
  }
  arg1 = (svn_wc_context_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_wc_relocate4", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_wc_relocate4", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_AsCharPtrAndSize(argv[3], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_wc_relocate4", 4, argv[3] ));
  }
  arg4 = (char *)(buf4);
  {
    arg5 = (svn_wc_relocation_validator3_t) svn_swig_rb_wc_relocation_validator3;
    arg6 = (void *)svn_swig_rb_make_baton(argv[4], _global_svn_swig_rb_pool);
  }
  if (argc > 5) {
    
  }
  {
    result = (svn_error_t *)svn_wc_relocate4(arg1,(char const *)arg2,(char const *)arg3,(char const *)arg4,arg5,arg6,arg7);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_relocate3(int argc, VALUE *argv, VALUE self) {
  char *arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  char *arg3 = 0 ;
  char *arg4 = 0 ;
  svn_boolean_t arg5 ;
  svn_wc_relocation_validator3_t arg6 = 0 ;
  void *arg7 = 0 ;
  apr_pool_t *arg8 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg8);
    _global_pool = arg8;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 6) || (argc > 7)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 6)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_wc_relocate3", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_relocate3", 2, argv[1] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_wc_relocate3", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_AsCharPtrAndSize(argv[3], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_wc_relocate3", 4, argv[3] ));
  }
  arg4 = (char *)(buf4);
  arg5 = RTEST(argv[4]);
  {
    arg6 = (svn_wc_relocation_validator3_t) svn_swig_rb_wc_relocation_validator3;
    arg7 = (void *)svn_swig_rb_make_baton(argv[5], _global_svn_swig_rb_pool);
  }
  if (argc > 6) {
    
  }
  {
    result = (svn_error_t *)svn_wc_relocate3((char const *)arg1,arg2,(char const *)arg3,(char const *)arg4,arg5,arg6,arg7,arg8);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_relocate2(int argc, VALUE *argv, VALUE self) {
  char *arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  char *arg3 = 0 ;
  char *arg4 = 0 ;
  svn_boolean_t arg5 ;
  svn_wc_relocation_validator2_t arg6 = 0 ;
  void *arg7 = 0 ;
  apr_pool_t *arg8 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  int res7 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg8);
    _global_pool = arg8;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 7) || (argc > 8)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 7)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_wc_relocate2", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_relocate2", 2, argv[1] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_wc_relocate2", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_AsCharPtrAndSize(argv[3], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_wc_relocate2", 4, argv[3] ));
  }
  arg4 = (char *)(buf4);
  arg5 = RTEST(argv[4]);
  {
    int res = SWIG_ConvertFunctionPtr(argv[5], (void**)(&arg6), SWIGTYPE_p_f_p_void_p_q_const__char_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_wc_relocation_validator2_t","svn_wc_relocate2", 6, argv[5] )); 
    }
  }
  res7 = SWIG_ConvertPtr(argv[6],SWIG_as_voidptrptr(&arg7), 0, 0);
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), Ruby_Format_TypeError( "", "void *","svn_wc_relocate2", 7, argv[6] )); 
  }
  if (argc > 7) {
    
  }
  {
    result = (svn_error_t *)svn_wc_relocate2((char const *)arg1,arg2,(char const *)arg3,(char const *)arg4,arg5,arg6,arg7,arg8);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_relocate(int argc, VALUE *argv, VALUE self) {
  char *arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  char *arg3 = 0 ;
  char *arg4 = 0 ;
  svn_boolean_t arg5 ;
  svn_wc_relocation_validator_t arg6 = 0 ;
  void *arg7 = 0 ;
  apr_pool_t *arg8 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  int res7 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg8);
    _global_pool = arg8;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 7) || (argc > 8)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 7)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_wc_relocate", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_relocate", 2, argv[1] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_wc_relocate", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_AsCharPtrAndSize(argv[3], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_wc_relocate", 4, argv[3] ));
  }
  arg4 = (char *)(buf4);
  arg5 = RTEST(argv[4]);
  {
    int res = SWIG_ConvertFunctionPtr(argv[5], (void**)(&arg6), SWIGTYPE_p_f_p_void_p_q_const__char_p_q_const__char__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_wc_relocation_validator_t","svn_wc_relocate", 6, argv[5] )); 
    }
  }
  res7 = SWIG_ConvertPtr(argv[6],SWIG_as_voidptrptr(&arg7), 0, 0);
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), Ruby_Format_TypeError( "", "void *","svn_wc_relocate", 7, argv[6] )); 
  }
  if (argc > 7) {
    
  }
  {
    result = (svn_error_t *)svn_wc_relocate((char const *)arg1,arg2,(char const *)arg3,(char const *)arg4,arg5,arg6,arg7,arg8);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_revert6(int argc, VALUE *argv, VALUE self) {
  svn_wc_context_t *arg1 = 0 ;
  char *arg2 = 0 ;
  svn_depth_t arg3 ;
  svn_boolean_t arg4 ;
  apr_array_header_t *arg5 = 0 ;
  svn_boolean_t arg6 ;
  svn_boolean_t arg7 ;
  svn_boolean_t arg8 ;
  svn_cancel_func_t arg9 = 0 ;
  void *arg10 = 0 ;
  svn_wc_notify_func2_t arg11 = 0 ;
  void *arg12 = 0 ;
  apr_pool_t *arg13 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg13);
    _global_pool = arg13;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 10) || (argc > 11)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 10)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_context_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_context_t *","svn_wc_revert6", 1, argv[0] )); 
  }
  arg1 = (svn_wc_context_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_wc_revert6", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  {
    arg3 = svn_swig_rb_to_depth(argv[2]);
  }
  arg4 = RTEST(argv[3]);
  res5 = SWIG_ConvertPtr(argv[4], &argp5,SWIGTYPE_p_apr_array_header_t, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "apr_array_header_t const *","svn_wc_revert6", 5, argv[4] )); 
  }
  arg5 = (apr_array_header_t *)(argp5);
  arg6 = RTEST(argv[5]);
  arg7 = RTEST(argv[6]);
  arg8 = RTEST(argv[7]);
  {
    arg9 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg10 = (void *)svn_swig_rb_make_baton(argv[8], _global_svn_swig_rb_pool);
  }
  {
    arg11 = (svn_wc_notify_func2_t) svn_swig_rb_notify_func2;
    arg12 = (void *)svn_swig_rb_make_baton(argv[9], _global_svn_swig_rb_pool);
  }
  if (argc > 10) {
    
  }
  {
    result = (svn_error_t *)svn_wc_revert6(arg1,(char const *)arg2,arg3,arg4,(apr_array_header_t const *)arg5,arg6,arg7,arg8,arg9,arg10,arg11,arg12,arg13);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg10);
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_revert5(int argc, VALUE *argv, VALUE self) {
  svn_wc_context_t *arg1 = 0 ;
  char *arg2 = 0 ;
  svn_depth_t arg3 ;
  svn_boolean_t arg4 ;
  apr_array_header_t *arg5 = 0 ;
  svn_boolean_t arg6 ;
  svn_boolean_t arg7 ;
  svn_cancel_func_t arg8 = 0 ;
  void *arg9 = 0 ;
  svn_wc_notify_func2_t arg10 = 0 ;
  void *arg11 = 0 ;
  apr_pool_t *arg12 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg12);
    _global_pool = arg12;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 9) || (argc > 10)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 9)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_context_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_context_t *","svn_wc_revert5", 1, argv[0] )); 
  }
  arg1 = (svn_wc_context_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_wc_revert5", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  {
    arg3 = svn_swig_rb_to_depth(argv[2]);
  }
  arg4 = RTEST(argv[3]);
  res5 = SWIG_ConvertPtr(argv[4], &argp5,SWIGTYPE_p_apr_array_header_t, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "apr_array_header_t const *","svn_wc_revert5", 5, argv[4] )); 
  }
  arg5 = (apr_array_header_t *)(argp5);
  arg6 = RTEST(argv[5]);
  arg7 = RTEST(argv[6]);
  {
    arg8 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg9 = (void *)svn_swig_rb_make_baton(argv[7], _global_svn_swig_rb_pool);
  }
  {
    arg10 = (svn_wc_notify_func2_t) svn_swig_rb_notify_func2;
    arg11 = (void *)svn_swig_rb_make_baton(argv[8], _global_svn_swig_rb_pool);
  }
  if (argc > 9) {
    
  }
  {
    result = (svn_error_t *)svn_wc_revert5(arg1,(char const *)arg2,arg3,arg4,(apr_array_header_t const *)arg5,arg6,arg7,arg8,arg9,arg10,arg11,arg12);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg9);
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_revert4(int argc, VALUE *argv, VALUE self) {
  svn_wc_context_t *arg1 = 0 ;
  char *arg2 = 0 ;
  svn_depth_t arg3 ;
  svn_boolean_t arg4 ;
  apr_array_header_t *arg5 = 0 ;
  svn_cancel_func_t arg6 = 0 ;
  void *arg7 = 0 ;
  svn_wc_notify_func2_t arg8 = 0 ;
  void *arg9 = 0 ;
  apr_pool_t *arg10 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg10);
    _global_pool = arg10;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 7) || (argc > 8)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 7)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_context_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_context_t *","svn_wc_revert4", 1, argv[0] )); 
  }
  arg1 = (svn_wc_context_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_wc_revert4", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  {
    arg3 = svn_swig_rb_to_depth(argv[2]);
  }
  arg4 = RTEST(argv[3]);
  res5 = SWIG_ConvertPtr(argv[4], &argp5,SWIGTYPE_p_apr_array_header_t, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "apr_array_header_t const *","svn_wc_revert4", 5, argv[4] )); 
  }
  arg5 = (apr_array_header_t *)(argp5);
  {
    arg6 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg7 = (void *)svn_swig_rb_make_baton(argv[5], _global_svn_swig_rb_pool);
  }
  {
    arg8 = (svn_wc_notify_func2_t) svn_swig_rb_notify_func2;
    arg9 = (void *)svn_swig_rb_make_baton(argv[6], _global_svn_swig_rb_pool);
  }
  if (argc > 7) {
    
  }
  {
    result = (svn_error_t *)svn_wc_revert4(arg1,(char const *)arg2,arg3,arg4,(apr_array_header_t const *)arg5,arg6,arg7,arg8,arg9,arg10);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg7);
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_revert3(int argc, VALUE *argv, VALUE self) {
  char *arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  svn_depth_t arg3 ;
  svn_boolean_t arg4 ;
  apr_array_header_t *arg5 = 0 ;
  svn_cancel_func_t arg6 = 0 ;
  void *arg7 = 0 ;
  svn_wc_notify_func2_t arg8 = 0 ;
  void *arg9 = 0 ;
  apr_pool_t *arg10 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg10);
    _global_pool = arg10;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 7) || (argc > 8)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 7)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_wc_revert3", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_revert3", 2, argv[1] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  {
    arg3 = svn_swig_rb_to_depth(argv[2]);
  }
  arg4 = RTEST(argv[3]);
  res5 = SWIG_ConvertPtr(argv[4], &argp5,SWIGTYPE_p_apr_array_header_t, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "apr_array_header_t const *","svn_wc_revert3", 5, argv[4] )); 
  }
  arg5 = (apr_array_header_t *)(argp5);
  {
    arg6 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg7 = (void *)svn_swig_rb_make_baton(argv[5], _global_svn_swig_rb_pool);
  }
  {
    arg8 = (svn_wc_notify_func2_t) svn_swig_rb_notify_func2;
    arg9 = (void *)svn_swig_rb_make_baton(argv[6], _global_svn_swig_rb_pool);
  }
  if (argc > 7) {
    
  }
  {
    result = (svn_error_t *)svn_wc_revert3((char const *)arg1,arg2,arg3,arg4,(apr_array_header_t const *)arg5,arg6,arg7,arg8,arg9,arg10);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg7);
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_revert2(int argc, VALUE *argv, VALUE self) {
  char *arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  svn_boolean_t arg3 ;
  svn_boolean_t arg4 ;
  svn_cancel_func_t arg5 = 0 ;
  void *arg6 = 0 ;
  svn_wc_notify_func2_t arg7 = 0 ;
  void *arg8 = 0 ;
  apr_pool_t *arg9 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg9);
    _global_pool = arg9;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 6) || (argc > 7)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 6)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_wc_revert2", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_revert2", 2, argv[1] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  arg3 = RTEST(argv[2]);
  arg4 = RTEST(argv[3]);
  {
    arg5 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg6 = (void *)svn_swig_rb_make_baton(argv[4], _global_svn_swig_rb_pool);
  }
  {
    arg7 = (svn_wc_notify_func2_t) svn_swig_rb_notify_func2;
    arg8 = (void *)svn_swig_rb_make_baton(argv[5], _global_svn_swig_rb_pool);
  }
  if (argc > 6) {
    
  }
  {
    result = (svn_error_t *)svn_wc_revert2((char const *)arg1,arg2,arg3,arg4,arg5,arg6,arg7,arg8,arg9);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg6);
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_revert(int argc, VALUE *argv, VALUE self) {
  char *arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  svn_boolean_t arg3 ;
  svn_boolean_t arg4 ;
  svn_cancel_func_t arg5 = 0 ;
  void *arg6 = 0 ;
  svn_wc_notify_func_t arg7 = 0 ;
  void *arg8 = 0 ;
  apr_pool_t *arg9 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res8 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg9);
    _global_pool = arg9;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 7) || (argc > 8)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 7)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_wc_revert", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_revert", 2, argv[1] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  arg3 = RTEST(argv[2]);
  arg4 = RTEST(argv[3]);
  {
    arg5 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg6 = (void *)svn_swig_rb_make_baton(argv[4], _global_svn_swig_rb_pool);
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[5], (void**)(&arg7), SWIGTYPE_p_f_p_void_p_q_const__char_enum_svn_wc_notify_action_t_enum_svn_node_kind_t_p_q_const__char_enum_svn_wc_notify_state_t_enum_svn_wc_notify_state_t_long__void);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_wc_notify_func_t","svn_wc_revert", 7, argv[5] )); 
    }
  }
  res8 = SWIG_ConvertPtr(argv[6],SWIG_as_voidptrptr(&arg8), 0, 0);
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), Ruby_Format_TypeError( "", "void *","svn_wc_revert", 8, argv[6] )); 
  }
  if (argc > 7) {
    
  }
  {
    result = (svn_error_t *)svn_wc_revert((char const *)arg1,arg2,arg3,arg4,arg5,arg6,arg7,arg8,arg9);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg6);
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_restore(int argc, VALUE *argv, VALUE self) {
  svn_wc_context_t *arg1 = 0 ;
  char *arg2 = 0 ;
  svn_boolean_t arg3 ;
  apr_pool_t *arg4 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 3) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_context_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_context_t *","svn_wc_restore", 1, argv[0] )); 
  }
  arg1 = (svn_wc_context_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_wc_restore", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  arg3 = RTEST(argv[2]);
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_wc_restore(arg1,(char const *)arg2,arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_create_tmp_file2(int argc, VALUE *argv, VALUE self) {
  apr_file_t **arg1 = 0 ;
  char **arg2 = 0 ;
  char *arg3 = 0 ;
  svn_io_file_del_t arg4 ;
  apr_pool_t *arg5 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  apr_file_t *temp1 ;
  char *temp2 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int val4 ;
  int ecode4 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  arg2 = &temp2;
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res3 = SWIG_AsCharPtrAndSize(argv[0], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_wc_create_tmp_file2", 3, argv[0] ));
  }
  arg3 = (char *)(buf3);
  ecode4 = SWIG_AsVal_int(argv[1], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), Ruby_Format_TypeError( "", "svn_io_file_del_t","svn_wc_create_tmp_file2", 4, argv[1] ));
  } 
  arg4 = (svn_io_file_del_t)(val4);
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_wc_create_tmp_file2(arg1,(char const **)arg2,(char const *)arg3,arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_apr_file_t, 0), 0);
  }
  {
    if (*arg2) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new2(*arg2), 0);
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil, 0);
    }
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_create_tmp_file(int argc, VALUE *argv, VALUE self) {
  apr_file_t **arg1 = 0 ;
  char *arg2 = 0 ;
  svn_boolean_t arg3 ;
  apr_pool_t *arg4 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  apr_file_t *temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_wc_create_tmp_file", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  arg3 = RTEST(argv[1]);
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_wc_create_tmp_file(arg1,(char const *)arg2,arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_apr_file_t, 0), 0);
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_translated_file2(int argc, VALUE *argv, VALUE self) {
  char **arg1 = 0 ;
  char *arg2 = 0 ;
  char *arg3 = 0 ;
  svn_wc_adm_access_t *arg4 = 0 ;
  apr_uint32_t arg5 ;
  apr_pool_t *arg6 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  char *temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  unsigned long val5 ;
  int ecode5 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 4) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_wc_translated_file2", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_wc_translated_file2", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_ConvertPtr(argv[2], &argp4,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_translated_file2", 4, argv[2] )); 
  }
  arg4 = (svn_wc_adm_access_t *)(argp4);
  ecode5 = SWIG_AsVal_unsigned_SS_long(argv[3], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), Ruby_Format_TypeError( "", "apr_uint32_t","svn_wc_translated_file2", 5, argv[3] ));
  } 
  arg5 = (apr_uint32_t)(val5);
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_wc_translated_file2((char const **)arg1,(char const *)arg2,(char const *)arg3,arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, svn_swig_rb_filename_to_temp_file(*arg1), 0);
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_translated_file(int argc, VALUE *argv, VALUE self) {
  char **arg1 = 0 ;
  char *arg2 = 0 ;
  svn_wc_adm_access_t *arg3 = 0 ;
  svn_boolean_t arg4 ;
  apr_pool_t *arg5 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  char *temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 3) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_wc_translated_file", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_ConvertPtr(argv[1], &argp3,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_translated_file", 3, argv[1] )); 
  }
  arg3 = (svn_wc_adm_access_t *)(argp3);
  arg4 = RTEST(argv[2]);
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_wc_translated_file((char const **)arg1,(char const *)arg2,arg3,arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    if (*arg1) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new2(*arg1), 0);
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil, 0);
    }
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_translated_stream(int argc, VALUE *argv, VALUE self) {
  svn_stream_t **arg1 = 0 ;
  char *arg2 = 0 ;
  char *arg3 = 0 ;
  svn_wc_adm_access_t *arg4 = 0 ;
  apr_uint32_t arg5 ;
  apr_pool_t *arg6 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_stream_t *temp1 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  unsigned long val5 ;
  int ecode5 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 4) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_wc_translated_stream", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_wc_translated_stream", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_ConvertPtr(argv[2], &argp4,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_translated_stream", 4, argv[2] )); 
  }
  arg4 = (svn_wc_adm_access_t *)(argp4);
  ecode5 = SWIG_AsVal_unsigned_SS_long(argv[3], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), Ruby_Format_TypeError( "", "apr_uint32_t","svn_wc_translated_stream", 5, argv[3] ));
  } 
  arg5 = (apr_uint32_t)(val5);
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_wc_translated_stream(arg1,(char const *)arg2,(char const *)arg3,arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_stream_t, 0), 0);
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_transmit_text_deltas3(int argc, VALUE *argv, VALUE self) {
  svn_checksum_t **arg1 = 0 ;
  svn_checksum_t **arg2 = 0 ;
  svn_wc_context_t *arg3 = 0 ;
  char *arg4 = 0 ;
  svn_boolean_t arg5 ;
  svn_delta_editor_t *arg6 = 0 ;
  void *arg7 = 0 ;
  apr_pool_t *arg8 = 0 ;
  apr_pool_t *arg9 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_checksum_t *temp1 ;
  svn_checksum_t *temp2 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg8);
    _global_pool = arg8;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg9);
    _global_pool = arg9;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  arg2 = &temp2;
  if ((argc < 4) || (argc > 6)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  res3 = SWIG_ConvertPtr(argv[0], &argp3,SWIGTYPE_p_svn_wc_context_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_wc_context_t *","svn_wc_transmit_text_deltas3", 3, argv[0] )); 
  }
  arg3 = (svn_wc_context_t *)(argp3);
  res4 = SWIG_AsCharPtrAndSize(argv[1], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_wc_transmit_text_deltas3", 4, argv[1] ));
  }
  arg4 = (char *)(buf4);
  arg5 = RTEST(argv[2]);
  {
    if (RTEST(rb_obj_is_kind_of(argv[3],
          svn_swig_rb_svn_delta_editor()))) {
      arg6 = svn_swig_rb_to_swig_type(argv[3],
        "svn_delta_editor_t *",
        _global_pool);
      arg7 = svn_swig_rb_to_swig_type(rb_funcall(argv[3], rb_intern("baton"), 0),
        "void *", _global_pool);
    } else {
      svn_swig_rb_make_delta_editor(&arg6, &arg7, argv[3], _global_pool);
    }
  }
  if (argc > 4) {
    
  }
  if (argc > 5) {
    
  }
  {
    result = (svn_error_t *)svn_wc_transmit_text_deltas3((struct svn_checksum_t const **)arg1,(struct svn_checksum_t const **)arg2,arg3,(char const *)arg4,arg5,(struct svn_delta_editor_t const *)arg6,arg7,arg8,arg9);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_checksum_t, 0), 0);
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg2, SWIGTYPE_p_svn_checksum_t, 0), 0);
  }
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_transmit_text_deltas2(int argc, VALUE *argv, VALUE self) {
  char **arg1 = 0 ;
  unsigned char *arg2 = (unsigned char *)0 ;
  char *arg3 = 0 ;
  svn_wc_adm_access_t *arg4 = 0 ;
  svn_boolean_t arg5 ;
  svn_delta_editor_t *arg6 = 0 ;
  void *arg7 = 0 ;
  apr_pool_t *arg8 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  char *temp1 ;
  unsigned char temp2[APR_MD5_DIGESTSIZE] ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg8);
    _global_pool = arg8;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  arg2 = temp2;
  if ((argc < 4) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  res3 = SWIG_AsCharPtrAndSize(argv[0], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_wc_transmit_text_deltas2", 3, argv[0] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_ConvertPtr(argv[1], &argp4,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_transmit_text_deltas2", 4, argv[1] )); 
  }
  arg4 = (svn_wc_adm_access_t *)(argp4);
  arg5 = RTEST(argv[2]);
  {
    if (RTEST(rb_obj_is_kind_of(argv[3],
          svn_swig_rb_svn_delta_editor()))) {
      arg6 = svn_swig_rb_to_swig_type(argv[3],
        "svn_delta_editor_t *",
        _global_pool);
      arg7 = svn_swig_rb_to_swig_type(rb_funcall(argv[3], rb_intern("baton"), 0),
        "void *", _global_pool);
    } else {
      svn_swig_rb_make_delta_editor(&arg6, &arg7, argv[3], _global_pool);
    }
  }
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_wc_transmit_text_deltas2((char const **)arg1,arg2,(char const *)arg3,arg4,arg5,(struct svn_delta_editor_t const *)arg6,arg7,arg8);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    if (*arg1) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new2(*arg1), 0);
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil, 0);
    }
  }
  {
    char *digest_string = (char *)svn_md5_digest_to_cstring(arg2, _global_pool);
    vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new2(digest_string ? digest_string : ""), 0);
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_transmit_text_deltas(int argc, VALUE *argv, VALUE self) {
  char *arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  svn_boolean_t arg3 ;
  svn_delta_editor_t *arg4 = 0 ;
  void *arg5 = 0 ;
  char **arg6 = 0 ;
  apr_pool_t *arg7 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  char *temp6 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg7);
    _global_pool = arg7;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg6 = &temp6;
  if ((argc < 4) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_wc_transmit_text_deltas", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_transmit_text_deltas", 2, argv[1] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  arg3 = RTEST(argv[2]);
  {
    if (RTEST(rb_obj_is_kind_of(argv[3],
          svn_swig_rb_svn_delta_editor()))) {
      arg4 = svn_swig_rb_to_swig_type(argv[3],
        "svn_delta_editor_t *",
        _global_pool);
      arg5 = svn_swig_rb_to_swig_type(rb_funcall(argv[3], rb_intern("baton"), 0),
        "void *", _global_pool);
    } else {
      svn_swig_rb_make_delta_editor(&arg4, &arg5, argv[3], _global_pool);
    }
  }
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_wc_transmit_text_deltas((char const *)arg1,arg2,arg3,(struct svn_delta_editor_t const *)arg4,arg5,(char const **)arg6,arg7);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    if (*arg6) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new2(*arg6), 0);
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil, 0);
    }
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_transmit_prop_deltas2(int argc, VALUE *argv, VALUE self) {
  svn_wc_context_t *arg1 = 0 ;
  char *arg2 = 0 ;
  svn_delta_editor_t *arg3 = 0 ;
  void *arg4 = 0 ;
  apr_pool_t *arg5 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 3) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_context_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_context_t *","svn_wc_transmit_prop_deltas2", 1, argv[0] )); 
  }
  arg1 = (svn_wc_context_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_wc_transmit_prop_deltas2", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  {
    if (RTEST(rb_obj_is_kind_of(argv[2],
          svn_swig_rb_svn_delta_editor()))) {
      arg3 = svn_swig_rb_to_swig_type(argv[2],
        "svn_delta_editor_t *",
        _global_pool);
      arg4 = svn_swig_rb_to_swig_type(rb_funcall(argv[2], rb_intern("baton"), 0),
        "void *", _global_pool);
    } else {
      svn_swig_rb_make_delta_editor(&arg3, &arg4, argv[2], _global_pool);
    }
  }
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_wc_transmit_prop_deltas2(arg1,(char const *)arg2,(struct svn_delta_editor_t const *)arg3,arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_transmit_prop_deltas(int argc, VALUE *argv, VALUE self) {
  char *arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  svn_wc_entry_t *arg3 = 0 ;
  svn_delta_editor_t *arg4 = 0 ;
  void *arg5 = 0 ;
  char **arg6 = 0 ;
  apr_pool_t *arg7 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  char *temp6 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg7);
    _global_pool = arg7;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg6 = &temp6;
  if ((argc < 4) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_wc_transmit_prop_deltas", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_transmit_prop_deltas", 2, argv[1] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  res3 = SWIG_ConvertPtr(argv[2], &argp3,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_wc_entry_t const *","svn_wc_transmit_prop_deltas", 3, argv[2] )); 
  }
  arg3 = (svn_wc_entry_t *)(argp3);
  {
    if (RTEST(rb_obj_is_kind_of(argv[3],
          svn_swig_rb_svn_delta_editor()))) {
      arg4 = svn_swig_rb_to_swig_type(argv[3],
        "svn_delta_editor_t *",
        _global_pool);
      arg5 = svn_swig_rb_to_swig_type(rb_funcall(argv[3], rb_intern("baton"), 0),
        "void *", _global_pool);
    } else {
      svn_swig_rb_make_delta_editor(&arg4, &arg5, argv[3], _global_pool);
    }
  }
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_wc_transmit_prop_deltas((char const *)arg1,arg2,(struct svn_wc_entry_t const *)arg3,(struct svn_delta_editor_t const *)arg4,arg5,(char const **)arg6,arg7);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    if (*arg6) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new2(*arg6), 0);
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil, 0);
    }
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_get_default_ignores(int argc, VALUE *argv, VALUE self) {
  apr_array_header_t **arg1 = 0 ;
  apr_hash_t *arg2 = 0 ;
  apr_pool_t *arg3 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  apr_array_header_t *temp1 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  {
    if (NIL_P(argv[0])) {
      arg2 = NULL;
    } else {
      arg2 = svn_swig_rb_hash_to_apr_hash_swig_type(argv[0], "svn_config_t *", _global_pool);
    }
  }
  if (argc > 1) {
    
  }
  {
    result = (svn_error_t *)svn_wc_get_default_ignores(arg1,arg2,arg3);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, svn_swig_rb_apr_array_to_array_string(*arg1), 0);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_get_ignores2(int argc, VALUE *argv, VALUE self) {
  apr_array_header_t **arg1 = 0 ;
  svn_wc_context_t *arg2 = 0 ;
  char *arg3 = 0 ;
  apr_hash_t *arg4 = 0 ;
  apr_pool_t *arg5 = 0 ;
  apr_pool_t *arg6 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  apr_array_header_t *temp1 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 3) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_svn_wc_context_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_context_t *","svn_wc_get_ignores2", 2, argv[0] )); 
  }
  arg2 = (svn_wc_context_t *)(argp2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_wc_get_ignores2", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  {
    if (NIL_P(argv[2])) {
      arg4 = NULL;
    } else {
      arg4 = svn_swig_rb_hash_to_apr_hash_swig_type(argv[2], "svn_config_t *", _global_pool);
    }
  }
  if (argc > 3) {
    
  }
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_wc_get_ignores2(arg1,arg2,(char const *)arg3,arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, svn_swig_rb_apr_array_to_array_string(*arg1), 0);
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_get_ignores(int argc, VALUE *argv, VALUE self) {
  apr_array_header_t **arg1 = 0 ;
  apr_hash_t *arg2 = 0 ;
  svn_wc_adm_access_t *arg3 = 0 ;
  apr_pool_t *arg4 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  apr_array_header_t *temp1 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  {
    if (NIL_P(argv[0])) {
      arg2 = NULL;
    } else {
      arg2 = svn_swig_rb_hash_to_apr_hash_swig_type(argv[0], "svn_config_t *", _global_pool);
    }
  }
  res3 = SWIG_ConvertPtr(argv[1], &argp3,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_get_ignores", 3, argv[1] )); 
  }
  arg3 = (svn_wc_adm_access_t *)(argp3);
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_wc_get_ignores(arg1,arg2,arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, svn_swig_rb_apr_array_to_array_string(*arg1), 0);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_match_ignore_list(int argc, VALUE *argv, VALUE self) {
  char *arg1 = 0 ;
  apr_array_header_t *arg2 = 0 ;
  apr_pool_t *arg3 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_wc_match_ignore_list", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    arg2 = svn_swig_rb_strings_to_apr_array(argv[1], _global_pool);
  }
  if (argc > 2) {
    
  }
  {
    result = (svn_boolean_t)svn_wc_match_ignore_list((char const *)arg1,(apr_array_header_t const *)arg2,arg3);
    
    
    
  }
  vresult = result ? Qtrue : Qfalse;
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_add_lock2(int argc, VALUE *argv, VALUE self) {
  svn_wc_context_t *arg1 = 0 ;
  char *arg2 = 0 ;
  svn_lock_t *arg3 = 0 ;
  apr_pool_t *arg4 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 3) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_context_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_context_t *","svn_wc_add_lock2", 1, argv[0] )); 
  }
  arg1 = (svn_wc_context_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_wc_add_lock2", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_ConvertPtr(argv[2], &argp3,SWIGTYPE_p_svn_lock_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_lock_t const *","svn_wc_add_lock2", 3, argv[2] )); 
  }
  arg3 = (svn_lock_t *)(argp3);
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_wc_add_lock2(arg1,(char const *)arg2,(struct svn_lock_t const *)arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_add_lock(int argc, VALUE *argv, VALUE self) {
  char *arg1 = 0 ;
  svn_lock_t *arg2 = 0 ;
  svn_wc_adm_access_t *arg3 = 0 ;
  apr_pool_t *arg4 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 3) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_wc_add_lock", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_lock_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_lock_t const *","svn_wc_add_lock", 2, argv[1] )); 
  }
  arg2 = (svn_lock_t *)(argp2);
  res3 = SWIG_ConvertPtr(argv[2], &argp3,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_add_lock", 3, argv[2] )); 
  }
  arg3 = (svn_wc_adm_access_t *)(argp3);
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_wc_add_lock((char const *)arg1,(struct svn_lock_t const *)arg2,arg3,arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_remove_lock2(int argc, VALUE *argv, VALUE self) {
  svn_wc_context_t *arg1 = 0 ;
  char *arg2 = 0 ;
  apr_pool_t *arg3 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_context_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_context_t *","svn_wc_remove_lock2", 1, argv[0] )); 
  }
  arg1 = (svn_wc_context_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_wc_remove_lock2", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_wc_remove_lock2(arg1,(char const *)arg2,arg3);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_remove_lock(int argc, VALUE *argv, VALUE self) {
  char *arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  apr_pool_t *arg3 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg3);
    _global_pool = arg3;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_wc_remove_lock", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_remove_lock", 2, argv[1] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  if (argc > 2) {
    
  }
  {
    result = (svn_error_t *)svn_wc_remove_lock((char const *)arg1,arg2,arg3);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


static swig_class SwigClassSvn_wc_revision_status_t;

SWIGINTERN VALUE
_wrap_svn_wc_revision_status_t_min_rev_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_revision_status_t *arg1 = 0 ;
  svn_revnum_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_revision_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_revision_status_t *","min_rev", 1, self )); 
  }
  arg1 = (struct svn_wc_revision_status_t *)(argp1);
  ecode2 = SWIG_AsVal_long(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_revnum_t","min_rev", 2, argv[0] ));
  } 
  arg2 = (svn_revnum_t)(val2);
  if (arg1) (arg1)->min_rev = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_revision_status_t_min_rev_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_revision_status_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_revnum_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_revision_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_revision_status_t *","min_rev", 1, self )); 
  }
  arg1 = (struct svn_wc_revision_status_t *)(argp1);
  result = (svn_revnum_t) ((arg1)->min_rev);
  vresult = SWIG_From_long((long)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_revision_status_t_max_rev_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_revision_status_t *arg1 = 0 ;
  svn_revnum_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_revision_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_revision_status_t *","max_rev", 1, self )); 
  }
  arg1 = (struct svn_wc_revision_status_t *)(argp1);
  ecode2 = SWIG_AsVal_long(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "svn_revnum_t","max_rev", 2, argv[0] ));
  } 
  arg2 = (svn_revnum_t)(val2);
  if (arg1) (arg1)->max_rev = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_revision_status_t_max_rev_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_revision_status_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_revnum_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_revision_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_revision_status_t *","max_rev", 1, self )); 
  }
  arg1 = (struct svn_wc_revision_status_t *)(argp1);
  result = (svn_revnum_t) ((arg1)->max_rev);
  vresult = SWIG_From_long((long)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_revision_status_t_switched_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_revision_status_t *arg1 = 0 ;
  svn_boolean_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_revision_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_revision_status_t *","switched", 1, self )); 
  }
  arg1 = (struct svn_wc_revision_status_t *)(argp1);
  arg2 = RTEST(argv[0]);
  if (arg1) (arg1)->switched = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_revision_status_t_switched_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_revision_status_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_revision_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_revision_status_t *","switched", 1, self )); 
  }
  arg1 = (struct svn_wc_revision_status_t *)(argp1);
  result = (svn_boolean_t) ((arg1)->switched);
  vresult = result ? Qtrue : Qfalse;
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_revision_status_t_modified_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_revision_status_t *arg1 = 0 ;
  svn_boolean_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_revision_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_revision_status_t *","modified", 1, self )); 
  }
  arg1 = (struct svn_wc_revision_status_t *)(argp1);
  arg2 = RTEST(argv[0]);
  if (arg1) (arg1)->modified = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_revision_status_t_modified_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_revision_status_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_revision_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_revision_status_t *","modified", 1, self )); 
  }
  arg1 = (struct svn_wc_revision_status_t *)(argp1);
  result = (svn_boolean_t) ((arg1)->modified);
  vresult = result ? Qtrue : Qfalse;
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_revision_status_t_sparse_checkout_set(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_revision_status_t *arg1 = 0 ;
  svn_boolean_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_revision_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_revision_status_t *","sparse_checkout", 1, self )); 
  }
  arg1 = (struct svn_wc_revision_status_t *)(argp1);
  arg2 = RTEST(argv[0]);
  if (arg1) (arg1)->sparse_checkout = arg2;
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_revision_status_t_sparse_checkout_get(int argc, VALUE *argv, VALUE self) {
  struct svn_wc_revision_status_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_boolean_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_svn_wc_revision_status_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct svn_wc_revision_status_t *","sparse_checkout", 1, self )); 
  }
  arg1 = (struct svn_wc_revision_status_t *)(argp1);
  result = (svn_boolean_t) ((arg1)->sparse_checkout);
  vresult = result ? Qtrue : Qfalse;
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
_wrap_svn_wc_revision_status_t_allocate(VALUE self)
#else
_wrap_svn_wc_revision_status_t_allocate(int argc, VALUE *argv, VALUE self)
#endif
{
  VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_svn_wc_revision_status_t);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
  rb_obj_call_init(vresult, argc, argv);
#endif
  return vresult;
}


SWIGINTERN VALUE
_wrap_new_svn_wc_revision_status_t(int argc, VALUE *argv, VALUE self) {
  char *arg1 = 0 ;
  char *arg2 = 0 ;
  svn_boolean_t arg3 ;
  svn_cancel_func_t arg4 = 0 ;
  void *arg5 = 0 ;
  apr_pool_t *arg6 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  struct svn_wc_revision_status_t *result = 0 ;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 4) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_wc_revision_status_t", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    if (NIL_P(argv[1])) {
      arg2 = NULL;
    } else {
      arg2 = StringValuePtr(argv[1]);
    }
  }
  arg3 = RTEST(argv[2]);
  {
    arg4 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg5 = (void *)svn_swig_rb_make_baton(argv[3], _global_svn_swig_rb_pool);
  }
  if (argc > 4) {
    
  }
  {
    result = (struct svn_wc_revision_status_t *)new_svn_wc_revision_status_t((char const *)arg1,(char const *)arg2,arg3,arg4,arg5,arg6);
    DATA_PTR(self) = result;
    
    
    
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg5);
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return self;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN void delete_svn_wc_revision_status_t(struct svn_wc_revision_status_t *self){
  }
SWIGINTERN void
free_svn_wc_revision_status_t(void *self) {
    struct svn_wc_revision_status_t *arg1 = (struct svn_wc_revision_status_t *)self;
    delete_svn_wc_revision_status_t(arg1);
}

SWIGINTERN VALUE
_wrap_svn_wc_revision_status2(int argc, VALUE *argv, VALUE self) {
  svn_wc_revision_status_t **arg1 = 0 ;
  svn_wc_context_t *arg2 = 0 ;
  char *arg3 = 0 ;
  char *arg4 = 0 ;
  svn_boolean_t arg5 ;
  svn_cancel_func_t arg6 = 0 ;
  void *arg7 = 0 ;
  apr_pool_t *arg8 = 0 ;
  apr_pool_t *arg9 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_wc_revision_status_t *temp1 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg8);
    _global_pool = arg8;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg9);
    _global_pool = arg9;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 5) || (argc > 7)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 5)",argc); SWIG_fail;
  }
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_svn_wc_context_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_context_t *","svn_wc_revision_status2", 2, argv[0] )); 
  }
  arg2 = (svn_wc_context_t *)(argp2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_wc_revision_status2", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  {
    if (NIL_P(argv[2])) {
      arg4 = NULL;
    } else {
      arg4 = StringValuePtr(argv[2]);
    }
  }
  arg5 = RTEST(argv[3]);
  {
    arg6 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg7 = (void *)svn_swig_rb_make_baton(argv[4], _global_svn_swig_rb_pool);
  }
  if (argc > 5) {
    
  }
  if (argc > 6) {
    
  }
  {
    result = (svn_error_t *)svn_wc_revision_status2(arg1,arg2,(char const *)arg3,(char const *)arg4,arg5,arg6,arg7,arg8,arg9);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj(*arg1, SWIGTYPE_p_svn_wc_revision_status_t, 0), 0);
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg7);
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_set_changelist2(int argc, VALUE *argv, VALUE self) {
  svn_wc_context_t *arg1 = 0 ;
  char *arg2 = 0 ;
  char *arg3 = 0 ;
  svn_depth_t arg4 ;
  apr_array_header_t *arg5 = 0 ;
  svn_cancel_func_t arg6 = 0 ;
  void *arg7 = 0 ;
  svn_wc_notify_func2_t arg8 = 0 ;
  void *arg9 = 0 ;
  apr_pool_t *arg10 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg10);
    _global_pool = arg10;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 7) || (argc > 8)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 7)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_context_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_context_t *","svn_wc_set_changelist2", 1, argv[0] )); 
  }
  arg1 = (svn_wc_context_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_wc_set_changelist2", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  {
    arg3 = StringValueCStr(argv[2]);
  }
  {
    arg4 = svn_swig_rb_to_depth(argv[3]);
  }
  res5 = SWIG_ConvertPtr(argv[4], &argp5,SWIGTYPE_p_apr_array_header_t, 0 |  0 );
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "apr_array_header_t const *","svn_wc_set_changelist2", 5, argv[4] )); 
  }
  arg5 = (apr_array_header_t *)(argp5);
  {
    arg6 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg7 = (void *)svn_swig_rb_make_baton(argv[5], _global_svn_swig_rb_pool);
  }
  {
    arg8 = (svn_wc_notify_func2_t) svn_swig_rb_notify_func2;
    arg9 = (void *)svn_swig_rb_make_baton(argv[6], _global_svn_swig_rb_pool);
  }
  if (argc > 7) {
    
  }
  {
    result = (svn_error_t *)svn_wc_set_changelist2(arg1,(char const *)arg2,(char const *)arg3,arg4,(apr_array_header_t const *)arg5,arg6,arg7,arg8,arg9,arg10);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg7);
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_set_changelist(int argc, VALUE *argv, VALUE self) {
  char *arg1 = 0 ;
  char *arg2 = 0 ;
  svn_wc_adm_access_t *arg3 = 0 ;
  svn_cancel_func_t arg4 = 0 ;
  void *arg5 = 0 ;
  svn_wc_notify_func2_t arg6 = 0 ;
  void *arg7 = 0 ;
  apr_pool_t *arg8 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg8);
    _global_pool = arg8;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 5) || (argc > 6)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 5)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","svn_wc_set_changelist", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  {
    arg2 = StringValueCStr(argv[1]);
  }
  res3 = SWIG_ConvertPtr(argv[2], &argp3,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_set_changelist", 3, argv[2] )); 
  }
  arg3 = (svn_wc_adm_access_t *)(argp3);
  {
    arg4 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg5 = (void *)svn_swig_rb_make_baton(argv[3], _global_svn_swig_rb_pool);
  }
  {
    arg6 = (svn_wc_notify_func2_t) svn_swig_rb_notify_func2;
    arg7 = (void *)svn_swig_rb_make_baton(argv[4], _global_svn_swig_rb_pool);
  }
  if (argc > 5) {
    
  }
  {
    result = (svn_error_t *)svn_wc_set_changelist((char const *)arg1,(char const *)arg2,arg3,arg4,arg5,arg6,arg7,arg8);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg5);
  }
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_get_changelists(int argc, VALUE *argv, VALUE self) {
  svn_wc_context_t *arg1 = 0 ;
  char *arg2 = 0 ;
  svn_depth_t arg3 ;
  apr_array_header_t *arg4 = 0 ;
  svn_changelist_receiver_t arg5 = 0 ;
  void *arg6 = 0 ;
  svn_cancel_func_t arg7 = 0 ;
  void *arg8 = 0 ;
  apr_pool_t *arg9 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  int res6 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg9);
    _global_pool = arg9;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 7) || (argc > 8)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 7)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_context_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_context_t *","svn_wc_get_changelists", 1, argv[0] )); 
  }
  arg1 = (svn_wc_context_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_wc_get_changelists", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  {
    arg3 = svn_swig_rb_to_depth(argv[2]);
  }
  res4 = SWIG_ConvertPtr(argv[3], &argp4,SWIGTYPE_p_apr_array_header_t, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "apr_array_header_t const *","svn_wc_get_changelists", 4, argv[3] )); 
  }
  arg4 = (apr_array_header_t *)(argp4);
  {
    int res = SWIG_ConvertFunctionPtr(argv[4], (void**)(&arg5), SWIGTYPE_p_f_p_void_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_changelist_receiver_t","svn_wc_get_changelists", 5, argv[4] )); 
    }
  }
  res6 = SWIG_ConvertPtr(argv[5],SWIG_as_voidptrptr(&arg6), 0, 0);
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "void *","svn_wc_get_changelists", 6, argv[5] )); 
  }
  {
    arg7 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg8 = (void *)svn_swig_rb_make_baton(argv[6], _global_svn_swig_rb_pool);
  }
  if (argc > 7) {
    
  }
  {
    result = (svn_error_t *)svn_wc_get_changelists(arg1,(char const *)arg2,arg3,(apr_array_header_t const *)arg4,arg5,arg6,arg7,arg8,arg9);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg8);
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_crop_tree2(int argc, VALUE *argv, VALUE self) {
  svn_wc_context_t *arg1 = 0 ;
  char *arg2 = 0 ;
  svn_depth_t arg3 ;
  svn_cancel_func_t arg4 = 0 ;
  void *arg5 = 0 ;
  svn_wc_notify_func2_t arg6 = 0 ;
  void *arg7 = 0 ;
  apr_pool_t *arg8 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg8);
    _global_pool = arg8;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 5) || (argc > 6)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 5)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_context_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_context_t *","svn_wc_crop_tree2", 1, argv[0] )); 
  }
  arg1 = (svn_wc_context_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_wc_crop_tree2", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  {
    arg3 = svn_swig_rb_to_depth(argv[2]);
  }
  {
    arg4 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg5 = (void *)svn_swig_rb_make_baton(argv[3], _global_svn_swig_rb_pool);
  }
  {
    arg6 = (svn_wc_notify_func2_t) svn_swig_rb_notify_func2;
    arg7 = (void *)svn_swig_rb_make_baton(argv[4], _global_svn_swig_rb_pool);
  }
  if (argc > 5) {
    
  }
  {
    result = (svn_error_t *)svn_wc_crop_tree2(arg1,(char const *)arg2,arg3,arg4,arg5,arg6,arg7,arg8);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg5);
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_crop_tree(int argc, VALUE *argv, VALUE self) {
  svn_wc_adm_access_t *arg1 = 0 ;
  char *arg2 = 0 ;
  svn_depth_t arg3 ;
  svn_wc_notify_func2_t arg4 = 0 ;
  void *arg5 = 0 ;
  svn_cancel_func_t arg6 = 0 ;
  void *arg7 = 0 ;
  apr_pool_t *arg8 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg8);
    _global_pool = arg8;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 5) || (argc > 6)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 5)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_crop_tree", 1, argv[0] )); 
  }
  arg1 = (svn_wc_adm_access_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_wc_crop_tree", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  {
    arg3 = svn_swig_rb_to_depth(argv[2]);
  }
  {
    arg4 = (svn_wc_notify_func2_t) svn_swig_rb_notify_func2;
    arg5 = (void *)svn_swig_rb_make_baton(argv[3], _global_svn_swig_rb_pool);
  }
  {
    arg6 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg7 = (void *)svn_swig_rb_make_baton(argv[4], _global_svn_swig_rb_pool);
  }
  if (argc > 5) {
    
  }
  {
    result = (svn_error_t *)svn_wc_crop_tree(arg1,(char const *)arg2,arg3,arg4,arg5,arg6,arg7,arg8);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg7);
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_exclude(int argc, VALUE *argv, VALUE self) {
  svn_wc_context_t *arg1 = 0 ;
  char *arg2 = 0 ;
  svn_cancel_func_t arg3 = 0 ;
  void *arg4 = 0 ;
  svn_wc_notify_func2_t arg5 = 0 ;
  void *arg6 = 0 ;
  apr_pool_t *arg7 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg7);
    _global_pool = arg7;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 4) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_context_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_context_t *","svn_wc_exclude", 1, argv[0] )); 
  }
  arg1 = (svn_wc_context_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_wc_exclude", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  {
    arg3 = (svn_cancel_func_t) svn_swig_rb_cancel_func;
    arg4 = (void *)svn_swig_rb_make_baton(argv[2], _global_svn_swig_rb_pool);
  }
  {
    arg5 = (svn_wc_notify_func2_t) svn_swig_rb_notify_func2;
    arg6 = (void *)svn_swig_rb_make_baton(argv[3], _global_svn_swig_rb_pool);
  }
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_wc_exclude(arg1,(char const *)arg2,arg3,arg4,arg5,arg6,arg7);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    svn_swig_rb_set_baton(vresult, (VALUE)arg4);
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_read_kind2(int argc, VALUE *argv, VALUE self) {
  svn_node_kind_t *arg1 = 0 ;
  svn_wc_context_t *arg2 = 0 ;
  char *arg3 = 0 ;
  svn_boolean_t arg4 ;
  svn_boolean_t arg5 ;
  apr_pool_t *arg6 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_node_kind_t temp1 ;
  int res1 = SWIG_TMPOBJ ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 4) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_svn_wc_context_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_context_t *","svn_wc_read_kind2", 2, argv[0] )); 
  }
  arg2 = (svn_wc_context_t *)(argp2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_wc_read_kind2", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  arg4 = RTEST(argv[2]);
  arg5 = RTEST(argv[3]);
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_wc_read_kind2(arg1,arg2,(char const *)arg3,arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (SWIG_IsTmpObj(res1)) {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_From_long((*arg1)), 0);
  } else {
    int new_flags = SWIG_IsNewObj(res1) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj((void*)(arg1), SWIGTYPE_p_svn_node_kind_t, new_flags), 0);
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_read_kind(int argc, VALUE *argv, VALUE self) {
  svn_node_kind_t *arg1 = 0 ;
  svn_wc_context_t *arg2 = 0 ;
  char *arg3 = 0 ;
  svn_boolean_t arg4 ;
  apr_pool_t *arg5 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_node_kind_t temp1 ;
  int res1 = SWIG_TMPOBJ ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg1 = &temp1;
  if ((argc < 3) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_svn_wc_context_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_context_t *","svn_wc_read_kind", 2, argv[0] )); 
  }
  arg2 = (svn_wc_context_t *)(argp2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_wc_read_kind", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  arg4 = RTEST(argv[2]);
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_wc_read_kind(arg1,arg2,(char const *)arg3,arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (SWIG_IsTmpObj(res1)) {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_From_long((*arg1)), 0);
  } else {
    int new_flags = SWIG_IsNewObj(res1) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj((void*)(arg1), SWIGTYPE_p_svn_node_kind_t, new_flags), 0);
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks4_invoke_file_opened(int argc, VALUE *argv, VALUE self) {
  svn_wc_diff_callbacks4_t *arg1 = 0 ;
  svn_boolean_t *arg2 = 0 ;
  svn_boolean_t *arg3 = 0 ;
  char *arg4 = 0 ;
  svn_revnum_t arg5 ;
  void *arg6 = 0 ;
  apr_pool_t *arg7 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_boolean_t temp2 ;
  svn_boolean_t temp3 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  long val5 ;
  int ecode5 = 0 ;
  int res6 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg7);
    _global_pool = arg7;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg2 = &temp2;
  arg3 = &temp3;
  if ((argc < 4) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_diff_callbacks4_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_diff_callbacks4_t *","svn_wc_diff_callbacks4_invoke_file_opened", 1, argv[0] )); 
  }
  arg1 = (svn_wc_diff_callbacks4_t *)(argp1);
  res4 = SWIG_AsCharPtrAndSize(argv[1], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks4_invoke_file_opened", 4, argv[1] ));
  }
  arg4 = (char *)(buf4);
  ecode5 = SWIG_AsVal_long(argv[2], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), Ruby_Format_TypeError( "", "svn_revnum_t","svn_wc_diff_callbacks4_invoke_file_opened", 5, argv[2] ));
  } 
  arg5 = (svn_revnum_t)(val5);
  res6 = SWIG_ConvertPtr(argv[3],SWIG_as_voidptrptr(&arg6), 0, 0);
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "void *","svn_wc_diff_callbacks4_invoke_file_opened", 6, argv[3] )); 
  }
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_wc_diff_callbacks4_invoke_file_opened(arg1,arg2,arg3,(char const *)arg4,arg5,arg6,arg7);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, *arg2 ? Qtrue : Qfalse, 0);
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, *arg3 ? Qtrue : Qfalse, 0);
  }
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks4_invoke_file_changed(int argc, VALUE *argv, VALUE self) {
  svn_wc_diff_callbacks4_t *arg1 = 0 ;
  svn_wc_notify_state_t *arg2 = 0 ;
  svn_wc_notify_state_t *arg3 = 0 ;
  svn_boolean_t *arg4 = 0 ;
  char *arg5 = 0 ;
  char *arg6 = 0 ;
  char *arg7 = 0 ;
  svn_revnum_t arg8 ;
  svn_revnum_t arg9 ;
  char *arg10 = 0 ;
  char *arg11 = 0 ;
  apr_array_header_t *arg12 = 0 ;
  apr_hash_t *arg13 = 0 ;
  void *arg14 = 0 ;
  apr_pool_t *arg15 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  svn_boolean_t temp4 ;
  int res5 ;
  char *buf5 = 0 ;
  int alloc5 = 0 ;
  int res6 ;
  char *buf6 = 0 ;
  int alloc6 = 0 ;
  int res7 ;
  char *buf7 = 0 ;
  int alloc7 = 0 ;
  long val8 ;
  int ecode8 = 0 ;
  long val9 ;
  int ecode9 = 0 ;
  int res10 ;
  char *buf10 = 0 ;
  int alloc10 = 0 ;
  int res11 ;
  char *buf11 = 0 ;
  int alloc11 = 0 ;
  void *argp13 = 0 ;
  int res13 = 0 ;
  int res14 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg15);
    _global_pool = arg15;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg4 = &temp4;
  if ((argc < 13) || (argc > 14)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 13)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_diff_callbacks4_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_diff_callbacks4_t *","svn_wc_diff_callbacks4_invoke_file_changed", 1, argv[0] )); 
  }
  arg1 = (svn_wc_diff_callbacks4_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_notify_state_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_notify_state_t *","svn_wc_diff_callbacks4_invoke_file_changed", 2, argv[1] )); 
  }
  arg2 = (svn_wc_notify_state_t *)(argp2);
  res3 = SWIG_ConvertPtr(argv[2], &argp3,SWIGTYPE_p_svn_wc_notify_state_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_wc_notify_state_t *","svn_wc_diff_callbacks4_invoke_file_changed", 3, argv[2] )); 
  }
  arg3 = (svn_wc_notify_state_t *)(argp3);
  res5 = SWIG_AsCharPtrAndSize(argv[3], &buf5, NULL, &alloc5);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks4_invoke_file_changed", 5, argv[3] ));
  }
  arg5 = (char *)(buf5);
  res6 = SWIG_AsCharPtrAndSize(argv[4], &buf6, NULL, &alloc6);
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks4_invoke_file_changed", 6, argv[4] ));
  }
  arg6 = (char *)(buf6);
  res7 = SWIG_AsCharPtrAndSize(argv[5], &buf7, NULL, &alloc7);
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks4_invoke_file_changed", 7, argv[5] ));
  }
  arg7 = (char *)(buf7);
  ecode8 = SWIG_AsVal_long(argv[6], &val8);
  if (!SWIG_IsOK(ecode8)) {
    SWIG_exception_fail(SWIG_ArgError(ecode8), Ruby_Format_TypeError( "", "svn_revnum_t","svn_wc_diff_callbacks4_invoke_file_changed", 8, argv[6] ));
  } 
  arg8 = (svn_revnum_t)(val8);
  ecode9 = SWIG_AsVal_long(argv[7], &val9);
  if (!SWIG_IsOK(ecode9)) {
    SWIG_exception_fail(SWIG_ArgError(ecode9), Ruby_Format_TypeError( "", "svn_revnum_t","svn_wc_diff_callbacks4_invoke_file_changed", 9, argv[7] ));
  } 
  arg9 = (svn_revnum_t)(val9);
  res10 = SWIG_AsCharPtrAndSize(argv[8], &buf10, NULL, &alloc10);
  if (!SWIG_IsOK(res10)) {
    SWIG_exception_fail(SWIG_ArgError(res10), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks4_invoke_file_changed", 10, argv[8] ));
  }
  arg10 = (char *)(buf10);
  res11 = SWIG_AsCharPtrAndSize(argv[9], &buf11, NULL, &alloc11);
  if (!SWIG_IsOK(res11)) {
    SWIG_exception_fail(SWIG_ArgError(res11), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks4_invoke_file_changed", 11, argv[9] ));
  }
  arg11 = (char *)(buf11);
  {
    VALUE rb_pool;
    apr_pool_t *pool;
    
    svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &pool);
    
    arg12 = svn_swig_rb_to_apr_array_prop(argv[10], pool);
  }
  res13 = SWIG_ConvertPtr(argv[11], &argp13,SWIGTYPE_p_apr_hash_t, 0 |  0 );
  if (!SWIG_IsOK(res13)) {
    SWIG_exception_fail(SWIG_ArgError(res13), Ruby_Format_TypeError( "", "apr_hash_t *","svn_wc_diff_callbacks4_invoke_file_changed", 13, argv[11] )); 
  }
  arg13 = (apr_hash_t *)(argp13);
  res14 = SWIG_ConvertPtr(argv[12],SWIG_as_voidptrptr(&arg14), 0, 0);
  if (!SWIG_IsOK(res14)) {
    SWIG_exception_fail(SWIG_ArgError(res14), Ruby_Format_TypeError( "", "void *","svn_wc_diff_callbacks4_invoke_file_changed", 14, argv[12] )); 
  }
  if (argc > 13) {
    
  }
  {
    result = (svn_error_t *)svn_wc_diff_callbacks4_invoke_file_changed(arg1,arg2,arg3,arg4,(char const *)arg5,(char const *)arg6,(char const *)arg7,arg8,arg9,(char const *)arg10,(char const *)arg11,(apr_array_header_t const *)arg12,arg13,arg14,arg15);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, *arg4 ? Qtrue : Qfalse, 0);
  }
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  if (alloc7 == SWIG_NEWOBJ) free((char*)buf7);
  if (alloc10 == SWIG_NEWOBJ) free((char*)buf10);
  if (alloc11 == SWIG_NEWOBJ) free((char*)buf11);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  if (alloc7 == SWIG_NEWOBJ) free((char*)buf7);
  if (alloc10 == SWIG_NEWOBJ) free((char*)buf10);
  if (alloc11 == SWIG_NEWOBJ) free((char*)buf11);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks4_invoke_file_added(int argc, VALUE *argv, VALUE self) {
  svn_wc_diff_callbacks4_t *arg1 = 0 ;
  svn_wc_notify_state_t *arg2 = 0 ;
  svn_wc_notify_state_t *arg3 = 0 ;
  svn_boolean_t *arg4 = 0 ;
  char *arg5 = 0 ;
  char *arg6 = 0 ;
  char *arg7 = 0 ;
  svn_revnum_t arg8 ;
  svn_revnum_t arg9 ;
  char *arg10 = 0 ;
  char *arg11 = 0 ;
  char *arg12 = 0 ;
  svn_revnum_t arg13 ;
  apr_array_header_t *arg14 = 0 ;
  apr_hash_t *arg15 = 0 ;
  void *arg16 = 0 ;
  apr_pool_t *arg17 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  svn_boolean_t temp4 ;
  int res5 ;
  char *buf5 = 0 ;
  int alloc5 = 0 ;
  int res6 ;
  char *buf6 = 0 ;
  int alloc6 = 0 ;
  int res7 ;
  char *buf7 = 0 ;
  int alloc7 = 0 ;
  long val8 ;
  int ecode8 = 0 ;
  long val9 ;
  int ecode9 = 0 ;
  int res10 ;
  char *buf10 = 0 ;
  int alloc10 = 0 ;
  int res11 ;
  char *buf11 = 0 ;
  int alloc11 = 0 ;
  long val13 ;
  int ecode13 = 0 ;
  void *argp15 = 0 ;
  int res15 = 0 ;
  int res16 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg17);
    _global_pool = arg17;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg4 = &temp4;
  if ((argc < 15) || (argc > 16)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 15)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_diff_callbacks4_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_diff_callbacks4_t *","svn_wc_diff_callbacks4_invoke_file_added", 1, argv[0] )); 
  }
  arg1 = (svn_wc_diff_callbacks4_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_notify_state_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_notify_state_t *","svn_wc_diff_callbacks4_invoke_file_added", 2, argv[1] )); 
  }
  arg2 = (svn_wc_notify_state_t *)(argp2);
  res3 = SWIG_ConvertPtr(argv[2], &argp3,SWIGTYPE_p_svn_wc_notify_state_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_wc_notify_state_t *","svn_wc_diff_callbacks4_invoke_file_added", 3, argv[2] )); 
  }
  arg3 = (svn_wc_notify_state_t *)(argp3);
  res5 = SWIG_AsCharPtrAndSize(argv[3], &buf5, NULL, &alloc5);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks4_invoke_file_added", 5, argv[3] ));
  }
  arg5 = (char *)(buf5);
  res6 = SWIG_AsCharPtrAndSize(argv[4], &buf6, NULL, &alloc6);
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks4_invoke_file_added", 6, argv[4] ));
  }
  arg6 = (char *)(buf6);
  res7 = SWIG_AsCharPtrAndSize(argv[5], &buf7, NULL, &alloc7);
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks4_invoke_file_added", 7, argv[5] ));
  }
  arg7 = (char *)(buf7);
  ecode8 = SWIG_AsVal_long(argv[6], &val8);
  if (!SWIG_IsOK(ecode8)) {
    SWIG_exception_fail(SWIG_ArgError(ecode8), Ruby_Format_TypeError( "", "svn_revnum_t","svn_wc_diff_callbacks4_invoke_file_added", 8, argv[6] ));
  } 
  arg8 = (svn_revnum_t)(val8);
  ecode9 = SWIG_AsVal_long(argv[7], &val9);
  if (!SWIG_IsOK(ecode9)) {
    SWIG_exception_fail(SWIG_ArgError(ecode9), Ruby_Format_TypeError( "", "svn_revnum_t","svn_wc_diff_callbacks4_invoke_file_added", 9, argv[7] ));
  } 
  arg9 = (svn_revnum_t)(val9);
  res10 = SWIG_AsCharPtrAndSize(argv[8], &buf10, NULL, &alloc10);
  if (!SWIG_IsOK(res10)) {
    SWIG_exception_fail(SWIG_ArgError(res10), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks4_invoke_file_added", 10, argv[8] ));
  }
  arg10 = (char *)(buf10);
  res11 = SWIG_AsCharPtrAndSize(argv[9], &buf11, NULL, &alloc11);
  if (!SWIG_IsOK(res11)) {
    SWIG_exception_fail(SWIG_ArgError(res11), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks4_invoke_file_added", 11, argv[9] ));
  }
  arg11 = (char *)(buf11);
  {
    if (NIL_P(argv[10])) {
      arg12 = NULL;
    } else {
      arg12 = StringValuePtr(argv[10]);
    }
  }
  ecode13 = SWIG_AsVal_long(argv[11], &val13);
  if (!SWIG_IsOK(ecode13)) {
    SWIG_exception_fail(SWIG_ArgError(ecode13), Ruby_Format_TypeError( "", "svn_revnum_t","svn_wc_diff_callbacks4_invoke_file_added", 13, argv[11] ));
  } 
  arg13 = (svn_revnum_t)(val13);
  {
    VALUE rb_pool;
    apr_pool_t *pool;
    
    svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &pool);
    
    arg14 = svn_swig_rb_to_apr_array_prop(argv[12], pool);
  }
  res15 = SWIG_ConvertPtr(argv[13], &argp15,SWIGTYPE_p_apr_hash_t, 0 |  0 );
  if (!SWIG_IsOK(res15)) {
    SWIG_exception_fail(SWIG_ArgError(res15), Ruby_Format_TypeError( "", "apr_hash_t *","svn_wc_diff_callbacks4_invoke_file_added", 15, argv[13] )); 
  }
  arg15 = (apr_hash_t *)(argp15);
  res16 = SWIG_ConvertPtr(argv[14],SWIG_as_voidptrptr(&arg16), 0, 0);
  if (!SWIG_IsOK(res16)) {
    SWIG_exception_fail(SWIG_ArgError(res16), Ruby_Format_TypeError( "", "void *","svn_wc_diff_callbacks4_invoke_file_added", 16, argv[14] )); 
  }
  if (argc > 15) {
    
  }
  {
    result = (svn_error_t *)svn_wc_diff_callbacks4_invoke_file_added(arg1,arg2,arg3,arg4,(char const *)arg5,(char const *)arg6,(char const *)arg7,arg8,arg9,(char const *)arg10,(char const *)arg11,(char const *)arg12,arg13,(apr_array_header_t const *)arg14,arg15,arg16,arg17);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, *arg4 ? Qtrue : Qfalse, 0);
  }
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  if (alloc7 == SWIG_NEWOBJ) free((char*)buf7);
  if (alloc10 == SWIG_NEWOBJ) free((char*)buf10);
  if (alloc11 == SWIG_NEWOBJ) free((char*)buf11);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  if (alloc7 == SWIG_NEWOBJ) free((char*)buf7);
  if (alloc10 == SWIG_NEWOBJ) free((char*)buf10);
  if (alloc11 == SWIG_NEWOBJ) free((char*)buf11);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks4_invoke_file_deleted(int argc, VALUE *argv, VALUE self) {
  svn_wc_diff_callbacks4_t *arg1 = 0 ;
  svn_wc_notify_state_t *arg2 = 0 ;
  svn_boolean_t *arg3 = 0 ;
  char *arg4 = 0 ;
  char *arg5 = 0 ;
  char *arg6 = 0 ;
  char *arg7 = 0 ;
  char *arg8 = 0 ;
  apr_hash_t *arg9 = 0 ;
  void *arg10 = 0 ;
  apr_pool_t *arg11 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  svn_boolean_t temp3 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  int res5 ;
  char *buf5 = 0 ;
  int alloc5 = 0 ;
  int res6 ;
  char *buf6 = 0 ;
  int alloc6 = 0 ;
  int res7 ;
  char *buf7 = 0 ;
  int alloc7 = 0 ;
  int res8 ;
  char *buf8 = 0 ;
  int alloc8 = 0 ;
  void *argp9 = 0 ;
  int res9 = 0 ;
  int res10 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg11);
    _global_pool = arg11;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg3 = &temp3;
  if ((argc < 9) || (argc > 10)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 9)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_diff_callbacks4_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_diff_callbacks4_t *","svn_wc_diff_callbacks4_invoke_file_deleted", 1, argv[0] )); 
  }
  arg1 = (svn_wc_diff_callbacks4_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_notify_state_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_notify_state_t *","svn_wc_diff_callbacks4_invoke_file_deleted", 2, argv[1] )); 
  }
  arg2 = (svn_wc_notify_state_t *)(argp2);
  res4 = SWIG_AsCharPtrAndSize(argv[2], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks4_invoke_file_deleted", 4, argv[2] ));
  }
  arg4 = (char *)(buf4);
  res5 = SWIG_AsCharPtrAndSize(argv[3], &buf5, NULL, &alloc5);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks4_invoke_file_deleted", 5, argv[3] ));
  }
  arg5 = (char *)(buf5);
  res6 = SWIG_AsCharPtrAndSize(argv[4], &buf6, NULL, &alloc6);
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks4_invoke_file_deleted", 6, argv[4] ));
  }
  arg6 = (char *)(buf6);
  res7 = SWIG_AsCharPtrAndSize(argv[5], &buf7, NULL, &alloc7);
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks4_invoke_file_deleted", 7, argv[5] ));
  }
  arg7 = (char *)(buf7);
  res8 = SWIG_AsCharPtrAndSize(argv[6], &buf8, NULL, &alloc8);
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks4_invoke_file_deleted", 8, argv[6] ));
  }
  arg8 = (char *)(buf8);
  res9 = SWIG_ConvertPtr(argv[7], &argp9,SWIGTYPE_p_apr_hash_t, 0 |  0 );
  if (!SWIG_IsOK(res9)) {
    SWIG_exception_fail(SWIG_ArgError(res9), Ruby_Format_TypeError( "", "apr_hash_t *","svn_wc_diff_callbacks4_invoke_file_deleted", 9, argv[7] )); 
  }
  arg9 = (apr_hash_t *)(argp9);
  res10 = SWIG_ConvertPtr(argv[8],SWIG_as_voidptrptr(&arg10), 0, 0);
  if (!SWIG_IsOK(res10)) {
    SWIG_exception_fail(SWIG_ArgError(res10), Ruby_Format_TypeError( "", "void *","svn_wc_diff_callbacks4_invoke_file_deleted", 10, argv[8] )); 
  }
  if (argc > 9) {
    
  }
  {
    result = (svn_error_t *)svn_wc_diff_callbacks4_invoke_file_deleted(arg1,arg2,arg3,(char const *)arg4,(char const *)arg5,(char const *)arg6,(char const *)arg7,(char const *)arg8,arg9,arg10,arg11);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, *arg3 ? Qtrue : Qfalse, 0);
  }
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  if (alloc7 == SWIG_NEWOBJ) free((char*)buf7);
  if (alloc8 == SWIG_NEWOBJ) free((char*)buf8);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  if (alloc7 == SWIG_NEWOBJ) free((char*)buf7);
  if (alloc8 == SWIG_NEWOBJ) free((char*)buf8);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks4_invoke_dir_deleted(int argc, VALUE *argv, VALUE self) {
  svn_wc_diff_callbacks4_t *arg1 = 0 ;
  svn_wc_notify_state_t *arg2 = 0 ;
  svn_boolean_t *arg3 = 0 ;
  char *arg4 = 0 ;
  void *arg5 = 0 ;
  apr_pool_t *arg6 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  svn_boolean_t temp3 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  int res5 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg3 = &temp3;
  if ((argc < 4) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_diff_callbacks4_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_diff_callbacks4_t *","svn_wc_diff_callbacks4_invoke_dir_deleted", 1, argv[0] )); 
  }
  arg1 = (svn_wc_diff_callbacks4_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_notify_state_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_notify_state_t *","svn_wc_diff_callbacks4_invoke_dir_deleted", 2, argv[1] )); 
  }
  arg2 = (svn_wc_notify_state_t *)(argp2);
  res4 = SWIG_AsCharPtrAndSize(argv[2], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks4_invoke_dir_deleted", 4, argv[2] ));
  }
  arg4 = (char *)(buf4);
  res5 = SWIG_ConvertPtr(argv[3],SWIG_as_voidptrptr(&arg5), 0, 0);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "void *","svn_wc_diff_callbacks4_invoke_dir_deleted", 5, argv[3] )); 
  }
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_wc_diff_callbacks4_invoke_dir_deleted(arg1,arg2,arg3,(char const *)arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, *arg3 ? Qtrue : Qfalse, 0);
  }
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks4_invoke_dir_opened(int argc, VALUE *argv, VALUE self) {
  svn_wc_diff_callbacks4_t *arg1 = 0 ;
  svn_boolean_t *arg2 = 0 ;
  svn_boolean_t *arg3 = 0 ;
  svn_boolean_t *arg4 = 0 ;
  char *arg5 = 0 ;
  svn_revnum_t arg6 ;
  void *arg7 = 0 ;
  apr_pool_t *arg8 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  svn_boolean_t temp2 ;
  svn_boolean_t temp3 ;
  svn_boolean_t temp4 ;
  int res5 ;
  char *buf5 = 0 ;
  int alloc5 = 0 ;
  long val6 ;
  int ecode6 = 0 ;
  int res7 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg8);
    _global_pool = arg8;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg2 = &temp2;
  arg3 = &temp3;
  arg4 = &temp4;
  if ((argc < 4) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_diff_callbacks4_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_diff_callbacks4_t *","svn_wc_diff_callbacks4_invoke_dir_opened", 1, argv[0] )); 
  }
  arg1 = (svn_wc_diff_callbacks4_t *)(argp1);
  res5 = SWIG_AsCharPtrAndSize(argv[1], &buf5, NULL, &alloc5);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks4_invoke_dir_opened", 5, argv[1] ));
  }
  arg5 = (char *)(buf5);
  ecode6 = SWIG_AsVal_long(argv[2], &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), Ruby_Format_TypeError( "", "svn_revnum_t","svn_wc_diff_callbacks4_invoke_dir_opened", 6, argv[2] ));
  } 
  arg6 = (svn_revnum_t)(val6);
  res7 = SWIG_ConvertPtr(argv[3],SWIG_as_voidptrptr(&arg7), 0, 0);
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), Ruby_Format_TypeError( "", "void *","svn_wc_diff_callbacks4_invoke_dir_opened", 7, argv[3] )); 
  }
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_wc_diff_callbacks4_invoke_dir_opened(arg1,arg2,arg3,arg4,(char const *)arg5,arg6,arg7,arg8);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, *arg2 ? Qtrue : Qfalse, 0);
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, *arg3 ? Qtrue : Qfalse, 0);
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, *arg4 ? Qtrue : Qfalse, 0);
  }
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks4_invoke_dir_added(int argc, VALUE *argv, VALUE self) {
  svn_wc_diff_callbacks4_t *arg1 = 0 ;
  svn_wc_notify_state_t *arg2 = 0 ;
  svn_boolean_t *arg3 = 0 ;
  svn_boolean_t *arg4 = 0 ;
  svn_boolean_t *arg5 = 0 ;
  char *arg6 = 0 ;
  svn_revnum_t arg7 ;
  char *arg8 = 0 ;
  svn_revnum_t arg9 ;
  void *arg10 = 0 ;
  apr_pool_t *arg11 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  svn_boolean_t temp3 ;
  svn_boolean_t temp4 ;
  svn_boolean_t temp5 ;
  int res6 ;
  char *buf6 = 0 ;
  int alloc6 = 0 ;
  long val7 ;
  int ecode7 = 0 ;
  long val9 ;
  int ecode9 = 0 ;
  int res10 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg11);
    _global_pool = arg11;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg3 = &temp3;
  arg4 = &temp4;
  arg5 = &temp5;
  if ((argc < 7) || (argc > 8)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 7)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_diff_callbacks4_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_diff_callbacks4_t *","svn_wc_diff_callbacks4_invoke_dir_added", 1, argv[0] )); 
  }
  arg1 = (svn_wc_diff_callbacks4_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_notify_state_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_notify_state_t *","svn_wc_diff_callbacks4_invoke_dir_added", 2, argv[1] )); 
  }
  arg2 = (svn_wc_notify_state_t *)(argp2);
  res6 = SWIG_AsCharPtrAndSize(argv[2], &buf6, NULL, &alloc6);
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks4_invoke_dir_added", 6, argv[2] ));
  }
  arg6 = (char *)(buf6);
  ecode7 = SWIG_AsVal_long(argv[3], &val7);
  if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), Ruby_Format_TypeError( "", "svn_revnum_t","svn_wc_diff_callbacks4_invoke_dir_added", 7, argv[3] ));
  } 
  arg7 = (svn_revnum_t)(val7);
  {
    if (NIL_P(argv[4])) {
      arg8 = NULL;
    } else {
      arg8 = StringValuePtr(argv[4]);
    }
  }
  ecode9 = SWIG_AsVal_long(argv[5], &val9);
  if (!SWIG_IsOK(ecode9)) {
    SWIG_exception_fail(SWIG_ArgError(ecode9), Ruby_Format_TypeError( "", "svn_revnum_t","svn_wc_diff_callbacks4_invoke_dir_added", 9, argv[5] ));
  } 
  arg9 = (svn_revnum_t)(val9);
  res10 = SWIG_ConvertPtr(argv[6],SWIG_as_voidptrptr(&arg10), 0, 0);
  if (!SWIG_IsOK(res10)) {
    SWIG_exception_fail(SWIG_ArgError(res10), Ruby_Format_TypeError( "", "void *","svn_wc_diff_callbacks4_invoke_dir_added", 10, argv[6] )); 
  }
  if (argc > 7) {
    
  }
  {
    result = (svn_error_t *)svn_wc_diff_callbacks4_invoke_dir_added(arg1,arg2,arg3,arg4,arg5,(char const *)arg6,arg7,(char const *)arg8,arg9,arg10,arg11);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, *arg3 ? Qtrue : Qfalse, 0);
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, *arg4 ? Qtrue : Qfalse, 0);
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, *arg5 ? Qtrue : Qfalse, 0);
  }
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks4_invoke_dir_props_changed(int argc, VALUE *argv, VALUE self) {
  svn_wc_diff_callbacks4_t *arg1 = 0 ;
  svn_wc_notify_state_t *arg2 = 0 ;
  svn_boolean_t *arg3 = 0 ;
  char *arg4 = 0 ;
  svn_boolean_t arg5 ;
  apr_array_header_t *arg6 = 0 ;
  apr_hash_t *arg7 = 0 ;
  void *arg8 = 0 ;
  apr_pool_t *arg9 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  svn_boolean_t temp3 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  int res8 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg9);
    _global_pool = arg9;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg3 = &temp3;
  if ((argc < 7) || (argc > 8)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 7)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_diff_callbacks4_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_diff_callbacks4_t *","svn_wc_diff_callbacks4_invoke_dir_props_changed", 1, argv[0] )); 
  }
  arg1 = (svn_wc_diff_callbacks4_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_notify_state_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_notify_state_t *","svn_wc_diff_callbacks4_invoke_dir_props_changed", 2, argv[1] )); 
  }
  arg2 = (svn_wc_notify_state_t *)(argp2);
  res4 = SWIG_AsCharPtrAndSize(argv[2], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks4_invoke_dir_props_changed", 4, argv[2] ));
  }
  arg4 = (char *)(buf4);
  arg5 = RTEST(argv[3]);
  {
    VALUE rb_pool;
    apr_pool_t *pool;
    
    svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &pool);
    
    arg6 = svn_swig_rb_to_apr_array_prop(argv[4], pool);
  }
  res7 = SWIG_ConvertPtr(argv[5], &argp7,SWIGTYPE_p_apr_hash_t, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), Ruby_Format_TypeError( "", "apr_hash_t *","svn_wc_diff_callbacks4_invoke_dir_props_changed", 7, argv[5] )); 
  }
  arg7 = (apr_hash_t *)(argp7);
  res8 = SWIG_ConvertPtr(argv[6],SWIG_as_voidptrptr(&arg8), 0, 0);
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), Ruby_Format_TypeError( "", "void *","svn_wc_diff_callbacks4_invoke_dir_props_changed", 8, argv[6] )); 
  }
  if (argc > 7) {
    
  }
  {
    result = (svn_error_t *)svn_wc_diff_callbacks4_invoke_dir_props_changed(arg1,arg2,arg3,(char const *)arg4,arg5,(apr_array_header_t const *)arg6,arg7,arg8,arg9);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, *arg3 ? Qtrue : Qfalse, 0);
  }
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks4_invoke_dir_closed(int argc, VALUE *argv, VALUE self) {
  svn_wc_diff_callbacks4_t *arg1 = 0 ;
  svn_wc_notify_state_t *arg2 = 0 ;
  svn_wc_notify_state_t *arg3 = 0 ;
  svn_boolean_t *arg4 = 0 ;
  char *arg5 = 0 ;
  svn_boolean_t arg6 ;
  void *arg7 = 0 ;
  apr_pool_t *arg8 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  svn_boolean_t temp4 ;
  int res5 ;
  char *buf5 = 0 ;
  int alloc5 = 0 ;
  int res7 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg8);
    _global_pool = arg8;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg4 = &temp4;
  if ((argc < 6) || (argc > 7)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 6)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_diff_callbacks4_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_diff_callbacks4_t *","svn_wc_diff_callbacks4_invoke_dir_closed", 1, argv[0] )); 
  }
  arg1 = (svn_wc_diff_callbacks4_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_notify_state_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_notify_state_t *","svn_wc_diff_callbacks4_invoke_dir_closed", 2, argv[1] )); 
  }
  arg2 = (svn_wc_notify_state_t *)(argp2);
  res3 = SWIG_ConvertPtr(argv[2], &argp3,SWIGTYPE_p_svn_wc_notify_state_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_wc_notify_state_t *","svn_wc_diff_callbacks4_invoke_dir_closed", 3, argv[2] )); 
  }
  arg3 = (svn_wc_notify_state_t *)(argp3);
  res5 = SWIG_AsCharPtrAndSize(argv[3], &buf5, NULL, &alloc5);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks4_invoke_dir_closed", 5, argv[3] ));
  }
  arg5 = (char *)(buf5);
  arg6 = RTEST(argv[4]);
  res7 = SWIG_ConvertPtr(argv[5],SWIG_as_voidptrptr(&arg7), 0, 0);
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), Ruby_Format_TypeError( "", "void *","svn_wc_diff_callbacks4_invoke_dir_closed", 7, argv[5] )); 
  }
  if (argc > 6) {
    
  }
  {
    result = (svn_error_t *)svn_wc_diff_callbacks4_invoke_dir_closed(arg1,arg2,arg3,arg4,(char const *)arg5,arg6,arg7,arg8);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, *arg4 ? Qtrue : Qfalse, 0);
  }
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks3_invoke_file_changed(int argc, VALUE *argv, VALUE self) {
  svn_wc_diff_callbacks3_t *arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  svn_wc_notify_state_t *arg3 = 0 ;
  svn_wc_notify_state_t *arg4 = 0 ;
  svn_boolean_t *arg5 = 0 ;
  char *arg6 = 0 ;
  char *arg7 = 0 ;
  char *arg8 = 0 ;
  svn_revnum_t arg9 ;
  svn_revnum_t arg10 ;
  char *arg11 = 0 ;
  char *arg12 = 0 ;
  apr_array_header_t *arg13 = 0 ;
  apr_hash_t *arg14 = 0 ;
  void *arg15 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  svn_boolean_t temp5 ;
  int res6 ;
  char *buf6 = 0 ;
  int alloc6 = 0 ;
  int res7 ;
  char *buf7 = 0 ;
  int alloc7 = 0 ;
  int res8 ;
  char *buf8 = 0 ;
  int alloc8 = 0 ;
  long val9 ;
  int ecode9 = 0 ;
  long val10 ;
  int ecode10 = 0 ;
  int res11 ;
  char *buf11 = 0 ;
  int alloc11 = 0 ;
  int res12 ;
  char *buf12 = 0 ;
  int alloc12 = 0 ;
  void *argp14 = 0 ;
  int res14 = 0 ;
  int res15 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  arg5 = &temp5;
  if ((argc < 14) || (argc > 14)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 14)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_diff_callbacks3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_diff_callbacks3_t *","svn_wc_diff_callbacks3_invoke_file_changed", 1, argv[0] )); 
  }
  arg1 = (svn_wc_diff_callbacks3_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_diff_callbacks3_invoke_file_changed", 2, argv[1] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  res3 = SWIG_ConvertPtr(argv[2], &argp3,SWIGTYPE_p_svn_wc_notify_state_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_wc_notify_state_t *","svn_wc_diff_callbacks3_invoke_file_changed", 3, argv[2] )); 
  }
  arg3 = (svn_wc_notify_state_t *)(argp3);
  res4 = SWIG_ConvertPtr(argv[3], &argp4,SWIGTYPE_p_svn_wc_notify_state_t, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "svn_wc_notify_state_t *","svn_wc_diff_callbacks3_invoke_file_changed", 4, argv[3] )); 
  }
  arg4 = (svn_wc_notify_state_t *)(argp4);
  res6 = SWIG_AsCharPtrAndSize(argv[4], &buf6, NULL, &alloc6);
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks3_invoke_file_changed", 6, argv[4] ));
  }
  arg6 = (char *)(buf6);
  res7 = SWIG_AsCharPtrAndSize(argv[5], &buf7, NULL, &alloc7);
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks3_invoke_file_changed", 7, argv[5] ));
  }
  arg7 = (char *)(buf7);
  res8 = SWIG_AsCharPtrAndSize(argv[6], &buf8, NULL, &alloc8);
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks3_invoke_file_changed", 8, argv[6] ));
  }
  arg8 = (char *)(buf8);
  ecode9 = SWIG_AsVal_long(argv[7], &val9);
  if (!SWIG_IsOK(ecode9)) {
    SWIG_exception_fail(SWIG_ArgError(ecode9), Ruby_Format_TypeError( "", "svn_revnum_t","svn_wc_diff_callbacks3_invoke_file_changed", 9, argv[7] ));
  } 
  arg9 = (svn_revnum_t)(val9);
  ecode10 = SWIG_AsVal_long(argv[8], &val10);
  if (!SWIG_IsOK(ecode10)) {
    SWIG_exception_fail(SWIG_ArgError(ecode10), Ruby_Format_TypeError( "", "svn_revnum_t","svn_wc_diff_callbacks3_invoke_file_changed", 10, argv[8] ));
  } 
  arg10 = (svn_revnum_t)(val10);
  res11 = SWIG_AsCharPtrAndSize(argv[9], &buf11, NULL, &alloc11);
  if (!SWIG_IsOK(res11)) {
    SWIG_exception_fail(SWIG_ArgError(res11), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks3_invoke_file_changed", 11, argv[9] ));
  }
  arg11 = (char *)(buf11);
  res12 = SWIG_AsCharPtrAndSize(argv[10], &buf12, NULL, &alloc12);
  if (!SWIG_IsOK(res12)) {
    SWIG_exception_fail(SWIG_ArgError(res12), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks3_invoke_file_changed", 12, argv[10] ));
  }
  arg12 = (char *)(buf12);
  {
    VALUE rb_pool;
    apr_pool_t *pool;
    
    svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &pool);
    
    arg13 = svn_swig_rb_to_apr_array_prop(argv[11], pool);
  }
  res14 = SWIG_ConvertPtr(argv[12], &argp14,SWIGTYPE_p_apr_hash_t, 0 |  0 );
  if (!SWIG_IsOK(res14)) {
    SWIG_exception_fail(SWIG_ArgError(res14), Ruby_Format_TypeError( "", "apr_hash_t *","svn_wc_diff_callbacks3_invoke_file_changed", 14, argv[12] )); 
  }
  arg14 = (apr_hash_t *)(argp14);
  res15 = SWIG_ConvertPtr(argv[13],SWIG_as_voidptrptr(&arg15), 0, 0);
  if (!SWIG_IsOK(res15)) {
    SWIG_exception_fail(SWIG_ArgError(res15), Ruby_Format_TypeError( "", "void *","svn_wc_diff_callbacks3_invoke_file_changed", 15, argv[13] )); 
  }
  {
    result = (svn_error_t *)svn_wc_diff_callbacks3_invoke_file_changed(arg1,arg2,arg3,arg4,arg5,(char const *)arg6,(char const *)arg7,(char const *)arg8,arg9,arg10,(char const *)arg11,(char const *)arg12,(apr_array_header_t const *)arg13,arg14,arg15);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, *arg5 ? Qtrue : Qfalse, 0);
  }
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  if (alloc7 == SWIG_NEWOBJ) free((char*)buf7);
  if (alloc8 == SWIG_NEWOBJ) free((char*)buf8);
  if (alloc11 == SWIG_NEWOBJ) free((char*)buf11);
  if (alloc12 == SWIG_NEWOBJ) free((char*)buf12);
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  if (alloc7 == SWIG_NEWOBJ) free((char*)buf7);
  if (alloc8 == SWIG_NEWOBJ) free((char*)buf8);
  if (alloc11 == SWIG_NEWOBJ) free((char*)buf11);
  if (alloc12 == SWIG_NEWOBJ) free((char*)buf12);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks3_invoke_file_added(int argc, VALUE *argv, VALUE self) {
  svn_wc_diff_callbacks3_t *arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  svn_wc_notify_state_t *arg3 = 0 ;
  svn_wc_notify_state_t *arg4 = 0 ;
  svn_boolean_t *arg5 = 0 ;
  char *arg6 = 0 ;
  char *arg7 = 0 ;
  char *arg8 = 0 ;
  svn_revnum_t arg9 ;
  svn_revnum_t arg10 ;
  char *arg11 = 0 ;
  char *arg12 = 0 ;
  apr_array_header_t *arg13 = 0 ;
  apr_hash_t *arg14 = 0 ;
  void *arg15 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  svn_boolean_t temp5 ;
  int res6 ;
  char *buf6 = 0 ;
  int alloc6 = 0 ;
  int res7 ;
  char *buf7 = 0 ;
  int alloc7 = 0 ;
  int res8 ;
  char *buf8 = 0 ;
  int alloc8 = 0 ;
  long val9 ;
  int ecode9 = 0 ;
  long val10 ;
  int ecode10 = 0 ;
  int res11 ;
  char *buf11 = 0 ;
  int alloc11 = 0 ;
  int res12 ;
  char *buf12 = 0 ;
  int alloc12 = 0 ;
  void *argp14 = 0 ;
  int res14 = 0 ;
  int res15 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  arg5 = &temp5;
  if ((argc < 14) || (argc > 14)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 14)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_diff_callbacks3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_diff_callbacks3_t *","svn_wc_diff_callbacks3_invoke_file_added", 1, argv[0] )); 
  }
  arg1 = (svn_wc_diff_callbacks3_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_diff_callbacks3_invoke_file_added", 2, argv[1] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  res3 = SWIG_ConvertPtr(argv[2], &argp3,SWIGTYPE_p_svn_wc_notify_state_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_wc_notify_state_t *","svn_wc_diff_callbacks3_invoke_file_added", 3, argv[2] )); 
  }
  arg3 = (svn_wc_notify_state_t *)(argp3);
  res4 = SWIG_ConvertPtr(argv[3], &argp4,SWIGTYPE_p_svn_wc_notify_state_t, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "svn_wc_notify_state_t *","svn_wc_diff_callbacks3_invoke_file_added", 4, argv[3] )); 
  }
  arg4 = (svn_wc_notify_state_t *)(argp4);
  res6 = SWIG_AsCharPtrAndSize(argv[4], &buf6, NULL, &alloc6);
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks3_invoke_file_added", 6, argv[4] ));
  }
  arg6 = (char *)(buf6);
  res7 = SWIG_AsCharPtrAndSize(argv[5], &buf7, NULL, &alloc7);
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks3_invoke_file_added", 7, argv[5] ));
  }
  arg7 = (char *)(buf7);
  res8 = SWIG_AsCharPtrAndSize(argv[6], &buf8, NULL, &alloc8);
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks3_invoke_file_added", 8, argv[6] ));
  }
  arg8 = (char *)(buf8);
  ecode9 = SWIG_AsVal_long(argv[7], &val9);
  if (!SWIG_IsOK(ecode9)) {
    SWIG_exception_fail(SWIG_ArgError(ecode9), Ruby_Format_TypeError( "", "svn_revnum_t","svn_wc_diff_callbacks3_invoke_file_added", 9, argv[7] ));
  } 
  arg9 = (svn_revnum_t)(val9);
  ecode10 = SWIG_AsVal_long(argv[8], &val10);
  if (!SWIG_IsOK(ecode10)) {
    SWIG_exception_fail(SWIG_ArgError(ecode10), Ruby_Format_TypeError( "", "svn_revnum_t","svn_wc_diff_callbacks3_invoke_file_added", 10, argv[8] ));
  } 
  arg10 = (svn_revnum_t)(val10);
  res11 = SWIG_AsCharPtrAndSize(argv[9], &buf11, NULL, &alloc11);
  if (!SWIG_IsOK(res11)) {
    SWIG_exception_fail(SWIG_ArgError(res11), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks3_invoke_file_added", 11, argv[9] ));
  }
  arg11 = (char *)(buf11);
  res12 = SWIG_AsCharPtrAndSize(argv[10], &buf12, NULL, &alloc12);
  if (!SWIG_IsOK(res12)) {
    SWIG_exception_fail(SWIG_ArgError(res12), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks3_invoke_file_added", 12, argv[10] ));
  }
  arg12 = (char *)(buf12);
  {
    VALUE rb_pool;
    apr_pool_t *pool;
    
    svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &pool);
    
    arg13 = svn_swig_rb_to_apr_array_prop(argv[11], pool);
  }
  res14 = SWIG_ConvertPtr(argv[12], &argp14,SWIGTYPE_p_apr_hash_t, 0 |  0 );
  if (!SWIG_IsOK(res14)) {
    SWIG_exception_fail(SWIG_ArgError(res14), Ruby_Format_TypeError( "", "apr_hash_t *","svn_wc_diff_callbacks3_invoke_file_added", 14, argv[12] )); 
  }
  arg14 = (apr_hash_t *)(argp14);
  res15 = SWIG_ConvertPtr(argv[13],SWIG_as_voidptrptr(&arg15), 0, 0);
  if (!SWIG_IsOK(res15)) {
    SWIG_exception_fail(SWIG_ArgError(res15), Ruby_Format_TypeError( "", "void *","svn_wc_diff_callbacks3_invoke_file_added", 15, argv[13] )); 
  }
  {
    result = (svn_error_t *)svn_wc_diff_callbacks3_invoke_file_added(arg1,arg2,arg3,arg4,arg5,(char const *)arg6,(char const *)arg7,(char const *)arg8,arg9,arg10,(char const *)arg11,(char const *)arg12,(apr_array_header_t const *)arg13,arg14,arg15);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, *arg5 ? Qtrue : Qfalse, 0);
  }
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  if (alloc7 == SWIG_NEWOBJ) free((char*)buf7);
  if (alloc8 == SWIG_NEWOBJ) free((char*)buf8);
  if (alloc11 == SWIG_NEWOBJ) free((char*)buf11);
  if (alloc12 == SWIG_NEWOBJ) free((char*)buf12);
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  if (alloc7 == SWIG_NEWOBJ) free((char*)buf7);
  if (alloc8 == SWIG_NEWOBJ) free((char*)buf8);
  if (alloc11 == SWIG_NEWOBJ) free((char*)buf11);
  if (alloc12 == SWIG_NEWOBJ) free((char*)buf12);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks3_invoke_file_deleted(int argc, VALUE *argv, VALUE self) {
  svn_wc_diff_callbacks3_t *arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  svn_wc_notify_state_t *arg3 = 0 ;
  svn_boolean_t *arg4 = 0 ;
  char *arg5 = 0 ;
  char *arg6 = 0 ;
  char *arg7 = 0 ;
  char *arg8 = 0 ;
  char *arg9 = 0 ;
  apr_hash_t *arg10 = 0 ;
  void *arg11 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  svn_boolean_t temp4 ;
  int res5 ;
  char *buf5 = 0 ;
  int alloc5 = 0 ;
  int res6 ;
  char *buf6 = 0 ;
  int alloc6 = 0 ;
  int res7 ;
  char *buf7 = 0 ;
  int alloc7 = 0 ;
  int res8 ;
  char *buf8 = 0 ;
  int alloc8 = 0 ;
  int res9 ;
  char *buf9 = 0 ;
  int alloc9 = 0 ;
  void *argp10 = 0 ;
  int res10 = 0 ;
  int res11 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  arg4 = &temp4;
  if ((argc < 10) || (argc > 10)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 10)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_diff_callbacks3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_diff_callbacks3_t *","svn_wc_diff_callbacks3_invoke_file_deleted", 1, argv[0] )); 
  }
  arg1 = (svn_wc_diff_callbacks3_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_diff_callbacks3_invoke_file_deleted", 2, argv[1] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  res3 = SWIG_ConvertPtr(argv[2], &argp3,SWIGTYPE_p_svn_wc_notify_state_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_wc_notify_state_t *","svn_wc_diff_callbacks3_invoke_file_deleted", 3, argv[2] )); 
  }
  arg3 = (svn_wc_notify_state_t *)(argp3);
  res5 = SWIG_AsCharPtrAndSize(argv[3], &buf5, NULL, &alloc5);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks3_invoke_file_deleted", 5, argv[3] ));
  }
  arg5 = (char *)(buf5);
  res6 = SWIG_AsCharPtrAndSize(argv[4], &buf6, NULL, &alloc6);
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks3_invoke_file_deleted", 6, argv[4] ));
  }
  arg6 = (char *)(buf6);
  res7 = SWIG_AsCharPtrAndSize(argv[5], &buf7, NULL, &alloc7);
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks3_invoke_file_deleted", 7, argv[5] ));
  }
  arg7 = (char *)(buf7);
  res8 = SWIG_AsCharPtrAndSize(argv[6], &buf8, NULL, &alloc8);
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks3_invoke_file_deleted", 8, argv[6] ));
  }
  arg8 = (char *)(buf8);
  res9 = SWIG_AsCharPtrAndSize(argv[7], &buf9, NULL, &alloc9);
  if (!SWIG_IsOK(res9)) {
    SWIG_exception_fail(SWIG_ArgError(res9), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks3_invoke_file_deleted", 9, argv[7] ));
  }
  arg9 = (char *)(buf9);
  res10 = SWIG_ConvertPtr(argv[8], &argp10,SWIGTYPE_p_apr_hash_t, 0 |  0 );
  if (!SWIG_IsOK(res10)) {
    SWIG_exception_fail(SWIG_ArgError(res10), Ruby_Format_TypeError( "", "apr_hash_t *","svn_wc_diff_callbacks3_invoke_file_deleted", 10, argv[8] )); 
  }
  arg10 = (apr_hash_t *)(argp10);
  res11 = SWIG_ConvertPtr(argv[9],SWIG_as_voidptrptr(&arg11), 0, 0);
  if (!SWIG_IsOK(res11)) {
    SWIG_exception_fail(SWIG_ArgError(res11), Ruby_Format_TypeError( "", "void *","svn_wc_diff_callbacks3_invoke_file_deleted", 11, argv[9] )); 
  }
  {
    result = (svn_error_t *)svn_wc_diff_callbacks3_invoke_file_deleted(arg1,arg2,arg3,arg4,(char const *)arg5,(char const *)arg6,(char const *)arg7,(char const *)arg8,(char const *)arg9,arg10,arg11);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, *arg4 ? Qtrue : Qfalse, 0);
  }
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  if (alloc7 == SWIG_NEWOBJ) free((char*)buf7);
  if (alloc8 == SWIG_NEWOBJ) free((char*)buf8);
  if (alloc9 == SWIG_NEWOBJ) free((char*)buf9);
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  if (alloc7 == SWIG_NEWOBJ) free((char*)buf7);
  if (alloc8 == SWIG_NEWOBJ) free((char*)buf8);
  if (alloc9 == SWIG_NEWOBJ) free((char*)buf9);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks3_invoke_dir_added(int argc, VALUE *argv, VALUE self) {
  svn_wc_diff_callbacks3_t *arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  svn_wc_notify_state_t *arg3 = 0 ;
  svn_boolean_t *arg4 = 0 ;
  char *arg5 = 0 ;
  svn_revnum_t arg6 ;
  void *arg7 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  svn_boolean_t temp4 ;
  int res5 ;
  char *buf5 = 0 ;
  int alloc5 = 0 ;
  long val6 ;
  int ecode6 = 0 ;
  int res7 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  arg4 = &temp4;
  if ((argc < 6) || (argc > 6)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 6)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_diff_callbacks3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_diff_callbacks3_t *","svn_wc_diff_callbacks3_invoke_dir_added", 1, argv[0] )); 
  }
  arg1 = (svn_wc_diff_callbacks3_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_diff_callbacks3_invoke_dir_added", 2, argv[1] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  res3 = SWIG_ConvertPtr(argv[2], &argp3,SWIGTYPE_p_svn_wc_notify_state_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_wc_notify_state_t *","svn_wc_diff_callbacks3_invoke_dir_added", 3, argv[2] )); 
  }
  arg3 = (svn_wc_notify_state_t *)(argp3);
  res5 = SWIG_AsCharPtrAndSize(argv[3], &buf5, NULL, &alloc5);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks3_invoke_dir_added", 5, argv[3] ));
  }
  arg5 = (char *)(buf5);
  ecode6 = SWIG_AsVal_long(argv[4], &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), Ruby_Format_TypeError( "", "svn_revnum_t","svn_wc_diff_callbacks3_invoke_dir_added", 6, argv[4] ));
  } 
  arg6 = (svn_revnum_t)(val6);
  res7 = SWIG_ConvertPtr(argv[5],SWIG_as_voidptrptr(&arg7), 0, 0);
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), Ruby_Format_TypeError( "", "void *","svn_wc_diff_callbacks3_invoke_dir_added", 7, argv[5] )); 
  }
  {
    result = (svn_error_t *)svn_wc_diff_callbacks3_invoke_dir_added(arg1,arg2,arg3,arg4,(char const *)arg5,arg6,arg7);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, *arg4 ? Qtrue : Qfalse, 0);
  }
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks3_invoke_dir_deleted(int argc, VALUE *argv, VALUE self) {
  svn_wc_diff_callbacks3_t *arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  svn_wc_notify_state_t *arg3 = 0 ;
  svn_boolean_t *arg4 = 0 ;
  char *arg5 = 0 ;
  void *arg6 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  svn_boolean_t temp4 ;
  int res5 ;
  char *buf5 = 0 ;
  int alloc5 = 0 ;
  int res6 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  arg4 = &temp4;
  if ((argc < 5) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 5)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_diff_callbacks3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_diff_callbacks3_t *","svn_wc_diff_callbacks3_invoke_dir_deleted", 1, argv[0] )); 
  }
  arg1 = (svn_wc_diff_callbacks3_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_diff_callbacks3_invoke_dir_deleted", 2, argv[1] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  res3 = SWIG_ConvertPtr(argv[2], &argp3,SWIGTYPE_p_svn_wc_notify_state_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_wc_notify_state_t *","svn_wc_diff_callbacks3_invoke_dir_deleted", 3, argv[2] )); 
  }
  arg3 = (svn_wc_notify_state_t *)(argp3);
  res5 = SWIG_AsCharPtrAndSize(argv[3], &buf5, NULL, &alloc5);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks3_invoke_dir_deleted", 5, argv[3] ));
  }
  arg5 = (char *)(buf5);
  res6 = SWIG_ConvertPtr(argv[4],SWIG_as_voidptrptr(&arg6), 0, 0);
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "void *","svn_wc_diff_callbacks3_invoke_dir_deleted", 6, argv[4] )); 
  }
  {
    result = (svn_error_t *)svn_wc_diff_callbacks3_invoke_dir_deleted(arg1,arg2,arg3,arg4,(char const *)arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, *arg4 ? Qtrue : Qfalse, 0);
  }
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks3_invoke_dir_props_changed(int argc, VALUE *argv, VALUE self) {
  svn_wc_diff_callbacks3_t *arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  svn_wc_notify_state_t *arg3 = 0 ;
  svn_boolean_t *arg4 = 0 ;
  char *arg5 = 0 ;
  apr_array_header_t *arg6 = 0 ;
  apr_hash_t *arg7 = 0 ;
  void *arg8 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  svn_boolean_t temp4 ;
  int res5 ;
  char *buf5 = 0 ;
  int alloc5 = 0 ;
  void *argp7 = 0 ;
  int res7 = 0 ;
  int res8 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  arg4 = &temp4;
  if ((argc < 7) || (argc > 7)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 7)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_diff_callbacks3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_diff_callbacks3_t *","svn_wc_diff_callbacks3_invoke_dir_props_changed", 1, argv[0] )); 
  }
  arg1 = (svn_wc_diff_callbacks3_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_diff_callbacks3_invoke_dir_props_changed", 2, argv[1] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  res3 = SWIG_ConvertPtr(argv[2], &argp3,SWIGTYPE_p_svn_wc_notify_state_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_wc_notify_state_t *","svn_wc_diff_callbacks3_invoke_dir_props_changed", 3, argv[2] )); 
  }
  arg3 = (svn_wc_notify_state_t *)(argp3);
  res5 = SWIG_AsCharPtrAndSize(argv[3], &buf5, NULL, &alloc5);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks3_invoke_dir_props_changed", 5, argv[3] ));
  }
  arg5 = (char *)(buf5);
  {
    VALUE rb_pool;
    apr_pool_t *pool;
    
    svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &pool);
    
    arg6 = svn_swig_rb_to_apr_array_prop(argv[4], pool);
  }
  res7 = SWIG_ConvertPtr(argv[5], &argp7,SWIGTYPE_p_apr_hash_t, 0 |  0 );
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), Ruby_Format_TypeError( "", "apr_hash_t *","svn_wc_diff_callbacks3_invoke_dir_props_changed", 7, argv[5] )); 
  }
  arg7 = (apr_hash_t *)(argp7);
  res8 = SWIG_ConvertPtr(argv[6],SWIG_as_voidptrptr(&arg8), 0, 0);
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), Ruby_Format_TypeError( "", "void *","svn_wc_diff_callbacks3_invoke_dir_props_changed", 8, argv[6] )); 
  }
  {
    result = (svn_error_t *)svn_wc_diff_callbacks3_invoke_dir_props_changed(arg1,arg2,arg3,arg4,(char const *)arg5,(apr_array_header_t const *)arg6,arg7,arg8);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, *arg4 ? Qtrue : Qfalse, 0);
  }
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks3_invoke_dir_opened(int argc, VALUE *argv, VALUE self) {
  svn_wc_diff_callbacks3_t *arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  svn_boolean_t *arg3 = 0 ;
  char *arg4 = 0 ;
  svn_revnum_t arg5 ;
  void *arg6 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  svn_boolean_t temp3 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  long val5 ;
  int ecode5 = 0 ;
  int res6 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  arg3 = &temp3;
  if ((argc < 5) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 5)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_diff_callbacks3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_diff_callbacks3_t *","svn_wc_diff_callbacks3_invoke_dir_opened", 1, argv[0] )); 
  }
  arg1 = (svn_wc_diff_callbacks3_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_diff_callbacks3_invoke_dir_opened", 2, argv[1] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  res4 = SWIG_AsCharPtrAndSize(argv[2], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks3_invoke_dir_opened", 4, argv[2] ));
  }
  arg4 = (char *)(buf4);
  ecode5 = SWIG_AsVal_long(argv[3], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), Ruby_Format_TypeError( "", "svn_revnum_t","svn_wc_diff_callbacks3_invoke_dir_opened", 5, argv[3] ));
  } 
  arg5 = (svn_revnum_t)(val5);
  res6 = SWIG_ConvertPtr(argv[4],SWIG_as_voidptrptr(&arg6), 0, 0);
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "void *","svn_wc_diff_callbacks3_invoke_dir_opened", 6, argv[4] )); 
  }
  {
    result = (svn_error_t *)svn_wc_diff_callbacks3_invoke_dir_opened(arg1,arg2,arg3,(char const *)arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, *arg3 ? Qtrue : Qfalse, 0);
  }
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks3_invoke_dir_closed(int argc, VALUE *argv, VALUE self) {
  svn_wc_diff_callbacks3_t *arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  svn_wc_notify_state_t *arg3 = 0 ;
  svn_wc_notify_state_t *arg4 = 0 ;
  svn_boolean_t *arg5 = 0 ;
  char *arg6 = 0 ;
  void *arg7 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  svn_boolean_t temp5 ;
  int res6 ;
  char *buf6 = 0 ;
  int alloc6 = 0 ;
  int res7 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  arg5 = &temp5;
  if ((argc < 6) || (argc > 6)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 6)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_diff_callbacks3_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_diff_callbacks3_t *","svn_wc_diff_callbacks3_invoke_dir_closed", 1, argv[0] )); 
  }
  arg1 = (svn_wc_diff_callbacks3_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_diff_callbacks3_invoke_dir_closed", 2, argv[1] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  res3 = SWIG_ConvertPtr(argv[2], &argp3,SWIGTYPE_p_svn_wc_notify_state_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_wc_notify_state_t *","svn_wc_diff_callbacks3_invoke_dir_closed", 3, argv[2] )); 
  }
  arg3 = (svn_wc_notify_state_t *)(argp3);
  res4 = SWIG_ConvertPtr(argv[3], &argp4,SWIGTYPE_p_svn_wc_notify_state_t, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "svn_wc_notify_state_t *","svn_wc_diff_callbacks3_invoke_dir_closed", 4, argv[3] )); 
  }
  arg4 = (svn_wc_notify_state_t *)(argp4);
  res6 = SWIG_AsCharPtrAndSize(argv[4], &buf6, NULL, &alloc6);
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks3_invoke_dir_closed", 6, argv[4] ));
  }
  arg6 = (char *)(buf6);
  res7 = SWIG_ConvertPtr(argv[5],SWIG_as_voidptrptr(&arg7), 0, 0);
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), Ruby_Format_TypeError( "", "void *","svn_wc_diff_callbacks3_invoke_dir_closed", 7, argv[5] )); 
  }
  {
    result = (svn_error_t *)svn_wc_diff_callbacks3_invoke_dir_closed(arg1,arg2,arg3,arg4,arg5,(char const *)arg6,arg7);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, *arg5 ? Qtrue : Qfalse, 0);
  }
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks2_invoke_file_changed(int argc, VALUE *argv, VALUE self) {
  svn_wc_diff_callbacks2_t *arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  svn_wc_notify_state_t *arg3 = 0 ;
  svn_wc_notify_state_t *arg4 = 0 ;
  char *arg5 = 0 ;
  char *arg6 = 0 ;
  char *arg7 = 0 ;
  svn_revnum_t arg8 ;
  svn_revnum_t arg9 ;
  char *arg10 = 0 ;
  char *arg11 = 0 ;
  apr_array_header_t *arg12 = 0 ;
  apr_hash_t *arg13 = 0 ;
  void *arg14 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  int res5 ;
  char *buf5 = 0 ;
  int alloc5 = 0 ;
  int res6 ;
  char *buf6 = 0 ;
  int alloc6 = 0 ;
  int res7 ;
  char *buf7 = 0 ;
  int alloc7 = 0 ;
  long val8 ;
  int ecode8 = 0 ;
  long val9 ;
  int ecode9 = 0 ;
  int res10 ;
  char *buf10 = 0 ;
  int alloc10 = 0 ;
  int res11 ;
  char *buf11 = 0 ;
  int alloc11 = 0 ;
  void *argp13 = 0 ;
  int res13 = 0 ;
  int res14 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 14) || (argc > 14)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 14)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_diff_callbacks2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_diff_callbacks2_t *","svn_wc_diff_callbacks2_invoke_file_changed", 1, argv[0] )); 
  }
  arg1 = (svn_wc_diff_callbacks2_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_diff_callbacks2_invoke_file_changed", 2, argv[1] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  res3 = SWIG_ConvertPtr(argv[2], &argp3,SWIGTYPE_p_svn_wc_notify_state_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_wc_notify_state_t *","svn_wc_diff_callbacks2_invoke_file_changed", 3, argv[2] )); 
  }
  arg3 = (svn_wc_notify_state_t *)(argp3);
  res4 = SWIG_ConvertPtr(argv[3], &argp4,SWIGTYPE_p_svn_wc_notify_state_t, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "svn_wc_notify_state_t *","svn_wc_diff_callbacks2_invoke_file_changed", 4, argv[3] )); 
  }
  arg4 = (svn_wc_notify_state_t *)(argp4);
  res5 = SWIG_AsCharPtrAndSize(argv[4], &buf5, NULL, &alloc5);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks2_invoke_file_changed", 5, argv[4] ));
  }
  arg5 = (char *)(buf5);
  res6 = SWIG_AsCharPtrAndSize(argv[5], &buf6, NULL, &alloc6);
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks2_invoke_file_changed", 6, argv[5] ));
  }
  arg6 = (char *)(buf6);
  res7 = SWIG_AsCharPtrAndSize(argv[6], &buf7, NULL, &alloc7);
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks2_invoke_file_changed", 7, argv[6] ));
  }
  arg7 = (char *)(buf7);
  ecode8 = SWIG_AsVal_long(argv[7], &val8);
  if (!SWIG_IsOK(ecode8)) {
    SWIG_exception_fail(SWIG_ArgError(ecode8), Ruby_Format_TypeError( "", "svn_revnum_t","svn_wc_diff_callbacks2_invoke_file_changed", 8, argv[7] ));
  } 
  arg8 = (svn_revnum_t)(val8);
  ecode9 = SWIG_AsVal_long(argv[8], &val9);
  if (!SWIG_IsOK(ecode9)) {
    SWIG_exception_fail(SWIG_ArgError(ecode9), Ruby_Format_TypeError( "", "svn_revnum_t","svn_wc_diff_callbacks2_invoke_file_changed", 9, argv[8] ));
  } 
  arg9 = (svn_revnum_t)(val9);
  res10 = SWIG_AsCharPtrAndSize(argv[9], &buf10, NULL, &alloc10);
  if (!SWIG_IsOK(res10)) {
    SWIG_exception_fail(SWIG_ArgError(res10), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks2_invoke_file_changed", 10, argv[9] ));
  }
  arg10 = (char *)(buf10);
  res11 = SWIG_AsCharPtrAndSize(argv[10], &buf11, NULL, &alloc11);
  if (!SWIG_IsOK(res11)) {
    SWIG_exception_fail(SWIG_ArgError(res11), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks2_invoke_file_changed", 11, argv[10] ));
  }
  arg11 = (char *)(buf11);
  {
    VALUE rb_pool;
    apr_pool_t *pool;
    
    svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &pool);
    
    arg12 = svn_swig_rb_to_apr_array_prop(argv[11], pool);
  }
  res13 = SWIG_ConvertPtr(argv[12], &argp13,SWIGTYPE_p_apr_hash_t, 0 |  0 );
  if (!SWIG_IsOK(res13)) {
    SWIG_exception_fail(SWIG_ArgError(res13), Ruby_Format_TypeError( "", "apr_hash_t *","svn_wc_diff_callbacks2_invoke_file_changed", 13, argv[12] )); 
  }
  arg13 = (apr_hash_t *)(argp13);
  res14 = SWIG_ConvertPtr(argv[13],SWIG_as_voidptrptr(&arg14), 0, 0);
  if (!SWIG_IsOK(res14)) {
    SWIG_exception_fail(SWIG_ArgError(res14), Ruby_Format_TypeError( "", "void *","svn_wc_diff_callbacks2_invoke_file_changed", 14, argv[13] )); 
  }
  {
    result = (svn_error_t *)svn_wc_diff_callbacks2_invoke_file_changed(arg1,arg2,arg3,arg4,(char const *)arg5,(char const *)arg6,(char const *)arg7,arg8,arg9,(char const *)arg10,(char const *)arg11,(apr_array_header_t const *)arg12,arg13,arg14);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  if (alloc7 == SWIG_NEWOBJ) free((char*)buf7);
  if (alloc10 == SWIG_NEWOBJ) free((char*)buf10);
  if (alloc11 == SWIG_NEWOBJ) free((char*)buf11);
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  if (alloc7 == SWIG_NEWOBJ) free((char*)buf7);
  if (alloc10 == SWIG_NEWOBJ) free((char*)buf10);
  if (alloc11 == SWIG_NEWOBJ) free((char*)buf11);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks2_invoke_file_added(int argc, VALUE *argv, VALUE self) {
  svn_wc_diff_callbacks2_t *arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  svn_wc_notify_state_t *arg3 = 0 ;
  svn_wc_notify_state_t *arg4 = 0 ;
  char *arg5 = 0 ;
  char *arg6 = 0 ;
  char *arg7 = 0 ;
  svn_revnum_t arg8 ;
  svn_revnum_t arg9 ;
  char *arg10 = 0 ;
  char *arg11 = 0 ;
  apr_array_header_t *arg12 = 0 ;
  apr_hash_t *arg13 = 0 ;
  void *arg14 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  int res5 ;
  char *buf5 = 0 ;
  int alloc5 = 0 ;
  int res6 ;
  char *buf6 = 0 ;
  int alloc6 = 0 ;
  int res7 ;
  char *buf7 = 0 ;
  int alloc7 = 0 ;
  long val8 ;
  int ecode8 = 0 ;
  long val9 ;
  int ecode9 = 0 ;
  int res10 ;
  char *buf10 = 0 ;
  int alloc10 = 0 ;
  int res11 ;
  char *buf11 = 0 ;
  int alloc11 = 0 ;
  void *argp13 = 0 ;
  int res13 = 0 ;
  int res14 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 14) || (argc > 14)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 14)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_diff_callbacks2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_diff_callbacks2_t *","svn_wc_diff_callbacks2_invoke_file_added", 1, argv[0] )); 
  }
  arg1 = (svn_wc_diff_callbacks2_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_diff_callbacks2_invoke_file_added", 2, argv[1] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  res3 = SWIG_ConvertPtr(argv[2], &argp3,SWIGTYPE_p_svn_wc_notify_state_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_wc_notify_state_t *","svn_wc_diff_callbacks2_invoke_file_added", 3, argv[2] )); 
  }
  arg3 = (svn_wc_notify_state_t *)(argp3);
  res4 = SWIG_ConvertPtr(argv[3], &argp4,SWIGTYPE_p_svn_wc_notify_state_t, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "svn_wc_notify_state_t *","svn_wc_diff_callbacks2_invoke_file_added", 4, argv[3] )); 
  }
  arg4 = (svn_wc_notify_state_t *)(argp4);
  res5 = SWIG_AsCharPtrAndSize(argv[4], &buf5, NULL, &alloc5);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks2_invoke_file_added", 5, argv[4] ));
  }
  arg5 = (char *)(buf5);
  res6 = SWIG_AsCharPtrAndSize(argv[5], &buf6, NULL, &alloc6);
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks2_invoke_file_added", 6, argv[5] ));
  }
  arg6 = (char *)(buf6);
  res7 = SWIG_AsCharPtrAndSize(argv[6], &buf7, NULL, &alloc7);
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks2_invoke_file_added", 7, argv[6] ));
  }
  arg7 = (char *)(buf7);
  ecode8 = SWIG_AsVal_long(argv[7], &val8);
  if (!SWIG_IsOK(ecode8)) {
    SWIG_exception_fail(SWIG_ArgError(ecode8), Ruby_Format_TypeError( "", "svn_revnum_t","svn_wc_diff_callbacks2_invoke_file_added", 8, argv[7] ));
  } 
  arg8 = (svn_revnum_t)(val8);
  ecode9 = SWIG_AsVal_long(argv[8], &val9);
  if (!SWIG_IsOK(ecode9)) {
    SWIG_exception_fail(SWIG_ArgError(ecode9), Ruby_Format_TypeError( "", "svn_revnum_t","svn_wc_diff_callbacks2_invoke_file_added", 9, argv[8] ));
  } 
  arg9 = (svn_revnum_t)(val9);
  res10 = SWIG_AsCharPtrAndSize(argv[9], &buf10, NULL, &alloc10);
  if (!SWIG_IsOK(res10)) {
    SWIG_exception_fail(SWIG_ArgError(res10), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks2_invoke_file_added", 10, argv[9] ));
  }
  arg10 = (char *)(buf10);
  res11 = SWIG_AsCharPtrAndSize(argv[10], &buf11, NULL, &alloc11);
  if (!SWIG_IsOK(res11)) {
    SWIG_exception_fail(SWIG_ArgError(res11), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks2_invoke_file_added", 11, argv[10] ));
  }
  arg11 = (char *)(buf11);
  {
    VALUE rb_pool;
    apr_pool_t *pool;
    
    svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &pool);
    
    arg12 = svn_swig_rb_to_apr_array_prop(argv[11], pool);
  }
  res13 = SWIG_ConvertPtr(argv[12], &argp13,SWIGTYPE_p_apr_hash_t, 0 |  0 );
  if (!SWIG_IsOK(res13)) {
    SWIG_exception_fail(SWIG_ArgError(res13), Ruby_Format_TypeError( "", "apr_hash_t *","svn_wc_diff_callbacks2_invoke_file_added", 13, argv[12] )); 
  }
  arg13 = (apr_hash_t *)(argp13);
  res14 = SWIG_ConvertPtr(argv[13],SWIG_as_voidptrptr(&arg14), 0, 0);
  if (!SWIG_IsOK(res14)) {
    SWIG_exception_fail(SWIG_ArgError(res14), Ruby_Format_TypeError( "", "void *","svn_wc_diff_callbacks2_invoke_file_added", 14, argv[13] )); 
  }
  {
    result = (svn_error_t *)svn_wc_diff_callbacks2_invoke_file_added(arg1,arg2,arg3,arg4,(char const *)arg5,(char const *)arg6,(char const *)arg7,arg8,arg9,(char const *)arg10,(char const *)arg11,(apr_array_header_t const *)arg12,arg13,arg14);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  if (alloc7 == SWIG_NEWOBJ) free((char*)buf7);
  if (alloc10 == SWIG_NEWOBJ) free((char*)buf10);
  if (alloc11 == SWIG_NEWOBJ) free((char*)buf11);
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  if (alloc7 == SWIG_NEWOBJ) free((char*)buf7);
  if (alloc10 == SWIG_NEWOBJ) free((char*)buf10);
  if (alloc11 == SWIG_NEWOBJ) free((char*)buf11);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks2_invoke_file_deleted(int argc, VALUE *argv, VALUE self) {
  svn_wc_diff_callbacks2_t *arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  svn_wc_notify_state_t *arg3 = 0 ;
  char *arg4 = 0 ;
  char *arg5 = 0 ;
  char *arg6 = 0 ;
  char *arg7 = 0 ;
  char *arg8 = 0 ;
  apr_hash_t *arg9 = 0 ;
  void *arg10 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  int res5 ;
  char *buf5 = 0 ;
  int alloc5 = 0 ;
  int res6 ;
  char *buf6 = 0 ;
  int alloc6 = 0 ;
  int res7 ;
  char *buf7 = 0 ;
  int alloc7 = 0 ;
  int res8 ;
  char *buf8 = 0 ;
  int alloc8 = 0 ;
  void *argp9 = 0 ;
  int res9 = 0 ;
  int res10 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 10) || (argc > 10)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 10)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_diff_callbacks2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_diff_callbacks2_t *","svn_wc_diff_callbacks2_invoke_file_deleted", 1, argv[0] )); 
  }
  arg1 = (svn_wc_diff_callbacks2_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_diff_callbacks2_invoke_file_deleted", 2, argv[1] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  res3 = SWIG_ConvertPtr(argv[2], &argp3,SWIGTYPE_p_svn_wc_notify_state_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_wc_notify_state_t *","svn_wc_diff_callbacks2_invoke_file_deleted", 3, argv[2] )); 
  }
  arg3 = (svn_wc_notify_state_t *)(argp3);
  res4 = SWIG_AsCharPtrAndSize(argv[3], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks2_invoke_file_deleted", 4, argv[3] ));
  }
  arg4 = (char *)(buf4);
  res5 = SWIG_AsCharPtrAndSize(argv[4], &buf5, NULL, &alloc5);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks2_invoke_file_deleted", 5, argv[4] ));
  }
  arg5 = (char *)(buf5);
  res6 = SWIG_AsCharPtrAndSize(argv[5], &buf6, NULL, &alloc6);
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks2_invoke_file_deleted", 6, argv[5] ));
  }
  arg6 = (char *)(buf6);
  res7 = SWIG_AsCharPtrAndSize(argv[6], &buf7, NULL, &alloc7);
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks2_invoke_file_deleted", 7, argv[6] ));
  }
  arg7 = (char *)(buf7);
  res8 = SWIG_AsCharPtrAndSize(argv[7], &buf8, NULL, &alloc8);
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks2_invoke_file_deleted", 8, argv[7] ));
  }
  arg8 = (char *)(buf8);
  res9 = SWIG_ConvertPtr(argv[8], &argp9,SWIGTYPE_p_apr_hash_t, 0 |  0 );
  if (!SWIG_IsOK(res9)) {
    SWIG_exception_fail(SWIG_ArgError(res9), Ruby_Format_TypeError( "", "apr_hash_t *","svn_wc_diff_callbacks2_invoke_file_deleted", 9, argv[8] )); 
  }
  arg9 = (apr_hash_t *)(argp9);
  res10 = SWIG_ConvertPtr(argv[9],SWIG_as_voidptrptr(&arg10), 0, 0);
  if (!SWIG_IsOK(res10)) {
    SWIG_exception_fail(SWIG_ArgError(res10), Ruby_Format_TypeError( "", "void *","svn_wc_diff_callbacks2_invoke_file_deleted", 10, argv[9] )); 
  }
  {
    result = (svn_error_t *)svn_wc_diff_callbacks2_invoke_file_deleted(arg1,arg2,arg3,(char const *)arg4,(char const *)arg5,(char const *)arg6,(char const *)arg7,(char const *)arg8,arg9,arg10);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  if (alloc7 == SWIG_NEWOBJ) free((char*)buf7);
  if (alloc8 == SWIG_NEWOBJ) free((char*)buf8);
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  if (alloc7 == SWIG_NEWOBJ) free((char*)buf7);
  if (alloc8 == SWIG_NEWOBJ) free((char*)buf8);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks2_invoke_dir_added(int argc, VALUE *argv, VALUE self) {
  svn_wc_diff_callbacks2_t *arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  svn_wc_notify_state_t *arg3 = 0 ;
  char *arg4 = 0 ;
  svn_revnum_t arg5 ;
  void *arg6 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  long val5 ;
  int ecode5 = 0 ;
  int res6 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 6) || (argc > 6)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 6)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_diff_callbacks2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_diff_callbacks2_t *","svn_wc_diff_callbacks2_invoke_dir_added", 1, argv[0] )); 
  }
  arg1 = (svn_wc_diff_callbacks2_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_diff_callbacks2_invoke_dir_added", 2, argv[1] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  res3 = SWIG_ConvertPtr(argv[2], &argp3,SWIGTYPE_p_svn_wc_notify_state_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_wc_notify_state_t *","svn_wc_diff_callbacks2_invoke_dir_added", 3, argv[2] )); 
  }
  arg3 = (svn_wc_notify_state_t *)(argp3);
  res4 = SWIG_AsCharPtrAndSize(argv[3], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks2_invoke_dir_added", 4, argv[3] ));
  }
  arg4 = (char *)(buf4);
  ecode5 = SWIG_AsVal_long(argv[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), Ruby_Format_TypeError( "", "svn_revnum_t","svn_wc_diff_callbacks2_invoke_dir_added", 5, argv[4] ));
  } 
  arg5 = (svn_revnum_t)(val5);
  res6 = SWIG_ConvertPtr(argv[5],SWIG_as_voidptrptr(&arg6), 0, 0);
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "void *","svn_wc_diff_callbacks2_invoke_dir_added", 6, argv[5] )); 
  }
  {
    result = (svn_error_t *)svn_wc_diff_callbacks2_invoke_dir_added(arg1,arg2,arg3,(char const *)arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks2_invoke_dir_deleted(int argc, VALUE *argv, VALUE self) {
  svn_wc_diff_callbacks2_t *arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  svn_wc_notify_state_t *arg3 = 0 ;
  char *arg4 = 0 ;
  void *arg5 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  int res5 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 5) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 5)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_diff_callbacks2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_diff_callbacks2_t *","svn_wc_diff_callbacks2_invoke_dir_deleted", 1, argv[0] )); 
  }
  arg1 = (svn_wc_diff_callbacks2_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_diff_callbacks2_invoke_dir_deleted", 2, argv[1] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  res3 = SWIG_ConvertPtr(argv[2], &argp3,SWIGTYPE_p_svn_wc_notify_state_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_wc_notify_state_t *","svn_wc_diff_callbacks2_invoke_dir_deleted", 3, argv[2] )); 
  }
  arg3 = (svn_wc_notify_state_t *)(argp3);
  res4 = SWIG_AsCharPtrAndSize(argv[3], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks2_invoke_dir_deleted", 4, argv[3] ));
  }
  arg4 = (char *)(buf4);
  res5 = SWIG_ConvertPtr(argv[4],SWIG_as_voidptrptr(&arg5), 0, 0);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "void *","svn_wc_diff_callbacks2_invoke_dir_deleted", 5, argv[4] )); 
  }
  {
    result = (svn_error_t *)svn_wc_diff_callbacks2_invoke_dir_deleted(arg1,arg2,arg3,(char const *)arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks2_invoke_dir_props_changed(int argc, VALUE *argv, VALUE self) {
  svn_wc_diff_callbacks2_t *arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  svn_wc_notify_state_t *arg3 = 0 ;
  char *arg4 = 0 ;
  apr_array_header_t *arg5 = 0 ;
  apr_hash_t *arg6 = 0 ;
  void *arg7 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  int res7 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 7) || (argc > 7)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 7)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_diff_callbacks2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_diff_callbacks2_t *","svn_wc_diff_callbacks2_invoke_dir_props_changed", 1, argv[0] )); 
  }
  arg1 = (svn_wc_diff_callbacks2_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_diff_callbacks2_invoke_dir_props_changed", 2, argv[1] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  res3 = SWIG_ConvertPtr(argv[2], &argp3,SWIGTYPE_p_svn_wc_notify_state_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_wc_notify_state_t *","svn_wc_diff_callbacks2_invoke_dir_props_changed", 3, argv[2] )); 
  }
  arg3 = (svn_wc_notify_state_t *)(argp3);
  res4 = SWIG_AsCharPtrAndSize(argv[3], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks2_invoke_dir_props_changed", 4, argv[3] ));
  }
  arg4 = (char *)(buf4);
  {
    VALUE rb_pool;
    apr_pool_t *pool;
    
    svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &pool);
    
    arg5 = svn_swig_rb_to_apr_array_prop(argv[4], pool);
  }
  res6 = SWIG_ConvertPtr(argv[5], &argp6,SWIGTYPE_p_apr_hash_t, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "apr_hash_t *","svn_wc_diff_callbacks2_invoke_dir_props_changed", 6, argv[5] )); 
  }
  arg6 = (apr_hash_t *)(argp6);
  res7 = SWIG_ConvertPtr(argv[6],SWIG_as_voidptrptr(&arg7), 0, 0);
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), Ruby_Format_TypeError( "", "void *","svn_wc_diff_callbacks2_invoke_dir_props_changed", 7, argv[6] )); 
  }
  {
    result = (svn_error_t *)svn_wc_diff_callbacks2_invoke_dir_props_changed(arg1,arg2,arg3,(char const *)arg4,(apr_array_header_t const *)arg5,arg6,arg7);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks_invoke_file_changed(int argc, VALUE *argv, VALUE self) {
  svn_wc_diff_callbacks_t *arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  svn_wc_notify_state_t *arg3 = 0 ;
  char *arg4 = 0 ;
  char *arg5 = 0 ;
  char *arg6 = 0 ;
  svn_revnum_t arg7 ;
  svn_revnum_t arg8 ;
  char *arg9 = 0 ;
  char *arg10 = 0 ;
  void *arg11 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  int res5 ;
  char *buf5 = 0 ;
  int alloc5 = 0 ;
  int res6 ;
  char *buf6 = 0 ;
  int alloc6 = 0 ;
  long val7 ;
  int ecode7 = 0 ;
  long val8 ;
  int ecode8 = 0 ;
  int res9 ;
  char *buf9 = 0 ;
  int alloc9 = 0 ;
  int res10 ;
  char *buf10 = 0 ;
  int alloc10 = 0 ;
  int res11 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 11) || (argc > 11)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 11)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_diff_callbacks_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_diff_callbacks_t *","svn_wc_diff_callbacks_invoke_file_changed", 1, argv[0] )); 
  }
  arg1 = (svn_wc_diff_callbacks_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_diff_callbacks_invoke_file_changed", 2, argv[1] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  res3 = SWIG_ConvertPtr(argv[2], &argp3,SWIGTYPE_p_svn_wc_notify_state_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_wc_notify_state_t *","svn_wc_diff_callbacks_invoke_file_changed", 3, argv[2] )); 
  }
  arg3 = (svn_wc_notify_state_t *)(argp3);
  res4 = SWIG_AsCharPtrAndSize(argv[3], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks_invoke_file_changed", 4, argv[3] ));
  }
  arg4 = (char *)(buf4);
  res5 = SWIG_AsCharPtrAndSize(argv[4], &buf5, NULL, &alloc5);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks_invoke_file_changed", 5, argv[4] ));
  }
  arg5 = (char *)(buf5);
  res6 = SWIG_AsCharPtrAndSize(argv[5], &buf6, NULL, &alloc6);
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks_invoke_file_changed", 6, argv[5] ));
  }
  arg6 = (char *)(buf6);
  ecode7 = SWIG_AsVal_long(argv[6], &val7);
  if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), Ruby_Format_TypeError( "", "svn_revnum_t","svn_wc_diff_callbacks_invoke_file_changed", 7, argv[6] ));
  } 
  arg7 = (svn_revnum_t)(val7);
  ecode8 = SWIG_AsVal_long(argv[7], &val8);
  if (!SWIG_IsOK(ecode8)) {
    SWIG_exception_fail(SWIG_ArgError(ecode8), Ruby_Format_TypeError( "", "svn_revnum_t","svn_wc_diff_callbacks_invoke_file_changed", 8, argv[7] ));
  } 
  arg8 = (svn_revnum_t)(val8);
  res9 = SWIG_AsCharPtrAndSize(argv[8], &buf9, NULL, &alloc9);
  if (!SWIG_IsOK(res9)) {
    SWIG_exception_fail(SWIG_ArgError(res9), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks_invoke_file_changed", 9, argv[8] ));
  }
  arg9 = (char *)(buf9);
  res10 = SWIG_AsCharPtrAndSize(argv[9], &buf10, NULL, &alloc10);
  if (!SWIG_IsOK(res10)) {
    SWIG_exception_fail(SWIG_ArgError(res10), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks_invoke_file_changed", 10, argv[9] ));
  }
  arg10 = (char *)(buf10);
  res11 = SWIG_ConvertPtr(argv[10],SWIG_as_voidptrptr(&arg11), 0, 0);
  if (!SWIG_IsOK(res11)) {
    SWIG_exception_fail(SWIG_ArgError(res11), Ruby_Format_TypeError( "", "void *","svn_wc_diff_callbacks_invoke_file_changed", 11, argv[10] )); 
  }
  {
    result = (svn_error_t *)svn_wc_diff_callbacks_invoke_file_changed(arg1,arg2,arg3,(char const *)arg4,(char const *)arg5,(char const *)arg6,arg7,arg8,(char const *)arg9,(char const *)arg10,arg11);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  if (alloc9 == SWIG_NEWOBJ) free((char*)buf9);
  if (alloc10 == SWIG_NEWOBJ) free((char*)buf10);
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  if (alloc9 == SWIG_NEWOBJ) free((char*)buf9);
  if (alloc10 == SWIG_NEWOBJ) free((char*)buf10);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks_invoke_file_added(int argc, VALUE *argv, VALUE self) {
  svn_wc_diff_callbacks_t *arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  svn_wc_notify_state_t *arg3 = 0 ;
  char *arg4 = 0 ;
  char *arg5 = 0 ;
  char *arg6 = 0 ;
  svn_revnum_t arg7 ;
  svn_revnum_t arg8 ;
  char *arg9 = 0 ;
  char *arg10 = 0 ;
  void *arg11 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  int res5 ;
  char *buf5 = 0 ;
  int alloc5 = 0 ;
  int res6 ;
  char *buf6 = 0 ;
  int alloc6 = 0 ;
  long val7 ;
  int ecode7 = 0 ;
  long val8 ;
  int ecode8 = 0 ;
  int res9 ;
  char *buf9 = 0 ;
  int alloc9 = 0 ;
  int res10 ;
  char *buf10 = 0 ;
  int alloc10 = 0 ;
  int res11 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 11) || (argc > 11)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 11)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_diff_callbacks_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_diff_callbacks_t *","svn_wc_diff_callbacks_invoke_file_added", 1, argv[0] )); 
  }
  arg1 = (svn_wc_diff_callbacks_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_diff_callbacks_invoke_file_added", 2, argv[1] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  res3 = SWIG_ConvertPtr(argv[2], &argp3,SWIGTYPE_p_svn_wc_notify_state_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_wc_notify_state_t *","svn_wc_diff_callbacks_invoke_file_added", 3, argv[2] )); 
  }
  arg3 = (svn_wc_notify_state_t *)(argp3);
  res4 = SWIG_AsCharPtrAndSize(argv[3], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks_invoke_file_added", 4, argv[3] ));
  }
  arg4 = (char *)(buf4);
  res5 = SWIG_AsCharPtrAndSize(argv[4], &buf5, NULL, &alloc5);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks_invoke_file_added", 5, argv[4] ));
  }
  arg5 = (char *)(buf5);
  res6 = SWIG_AsCharPtrAndSize(argv[5], &buf6, NULL, &alloc6);
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks_invoke_file_added", 6, argv[5] ));
  }
  arg6 = (char *)(buf6);
  ecode7 = SWIG_AsVal_long(argv[6], &val7);
  if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), Ruby_Format_TypeError( "", "svn_revnum_t","svn_wc_diff_callbacks_invoke_file_added", 7, argv[6] ));
  } 
  arg7 = (svn_revnum_t)(val7);
  ecode8 = SWIG_AsVal_long(argv[7], &val8);
  if (!SWIG_IsOK(ecode8)) {
    SWIG_exception_fail(SWIG_ArgError(ecode8), Ruby_Format_TypeError( "", "svn_revnum_t","svn_wc_diff_callbacks_invoke_file_added", 8, argv[7] ));
  } 
  arg8 = (svn_revnum_t)(val8);
  res9 = SWIG_AsCharPtrAndSize(argv[8], &buf9, NULL, &alloc9);
  if (!SWIG_IsOK(res9)) {
    SWIG_exception_fail(SWIG_ArgError(res9), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks_invoke_file_added", 9, argv[8] ));
  }
  arg9 = (char *)(buf9);
  res10 = SWIG_AsCharPtrAndSize(argv[9], &buf10, NULL, &alloc10);
  if (!SWIG_IsOK(res10)) {
    SWIG_exception_fail(SWIG_ArgError(res10), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks_invoke_file_added", 10, argv[9] ));
  }
  arg10 = (char *)(buf10);
  res11 = SWIG_ConvertPtr(argv[10],SWIG_as_voidptrptr(&arg11), 0, 0);
  if (!SWIG_IsOK(res11)) {
    SWIG_exception_fail(SWIG_ArgError(res11), Ruby_Format_TypeError( "", "void *","svn_wc_diff_callbacks_invoke_file_added", 11, argv[10] )); 
  }
  {
    result = (svn_error_t *)svn_wc_diff_callbacks_invoke_file_added(arg1,arg2,arg3,(char const *)arg4,(char const *)arg5,(char const *)arg6,arg7,arg8,(char const *)arg9,(char const *)arg10,arg11);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  if (alloc9 == SWIG_NEWOBJ) free((char*)buf9);
  if (alloc10 == SWIG_NEWOBJ) free((char*)buf10);
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  if (alloc9 == SWIG_NEWOBJ) free((char*)buf9);
  if (alloc10 == SWIG_NEWOBJ) free((char*)buf10);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks_invoke_file_deleted(int argc, VALUE *argv, VALUE self) {
  svn_wc_diff_callbacks_t *arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  svn_wc_notify_state_t *arg3 = 0 ;
  char *arg4 = 0 ;
  char *arg5 = 0 ;
  char *arg6 = 0 ;
  char *arg7 = 0 ;
  char *arg8 = 0 ;
  void *arg9 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  int res5 ;
  char *buf5 = 0 ;
  int alloc5 = 0 ;
  int res6 ;
  char *buf6 = 0 ;
  int alloc6 = 0 ;
  int res7 ;
  char *buf7 = 0 ;
  int alloc7 = 0 ;
  int res8 ;
  char *buf8 = 0 ;
  int alloc8 = 0 ;
  int res9 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 9) || (argc > 9)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 9)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_diff_callbacks_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_diff_callbacks_t *","svn_wc_diff_callbacks_invoke_file_deleted", 1, argv[0] )); 
  }
  arg1 = (svn_wc_diff_callbacks_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_diff_callbacks_invoke_file_deleted", 2, argv[1] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  res3 = SWIG_ConvertPtr(argv[2], &argp3,SWIGTYPE_p_svn_wc_notify_state_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_wc_notify_state_t *","svn_wc_diff_callbacks_invoke_file_deleted", 3, argv[2] )); 
  }
  arg3 = (svn_wc_notify_state_t *)(argp3);
  res4 = SWIG_AsCharPtrAndSize(argv[3], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks_invoke_file_deleted", 4, argv[3] ));
  }
  arg4 = (char *)(buf4);
  res5 = SWIG_AsCharPtrAndSize(argv[4], &buf5, NULL, &alloc5);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks_invoke_file_deleted", 5, argv[4] ));
  }
  arg5 = (char *)(buf5);
  res6 = SWIG_AsCharPtrAndSize(argv[5], &buf6, NULL, &alloc6);
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks_invoke_file_deleted", 6, argv[5] ));
  }
  arg6 = (char *)(buf6);
  res7 = SWIG_AsCharPtrAndSize(argv[6], &buf7, NULL, &alloc7);
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks_invoke_file_deleted", 7, argv[6] ));
  }
  arg7 = (char *)(buf7);
  res8 = SWIG_AsCharPtrAndSize(argv[7], &buf8, NULL, &alloc8);
  if (!SWIG_IsOK(res8)) {
    SWIG_exception_fail(SWIG_ArgError(res8), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks_invoke_file_deleted", 8, argv[7] ));
  }
  arg8 = (char *)(buf8);
  res9 = SWIG_ConvertPtr(argv[8],SWIG_as_voidptrptr(&arg9), 0, 0);
  if (!SWIG_IsOK(res9)) {
    SWIG_exception_fail(SWIG_ArgError(res9), Ruby_Format_TypeError( "", "void *","svn_wc_diff_callbacks_invoke_file_deleted", 9, argv[8] )); 
  }
  {
    result = (svn_error_t *)svn_wc_diff_callbacks_invoke_file_deleted(arg1,arg2,arg3,(char const *)arg4,(char const *)arg5,(char const *)arg6,(char const *)arg7,(char const *)arg8,arg9);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  if (alloc7 == SWIG_NEWOBJ) free((char*)buf7);
  if (alloc8 == SWIG_NEWOBJ) free((char*)buf8);
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  if (alloc7 == SWIG_NEWOBJ) free((char*)buf7);
  if (alloc8 == SWIG_NEWOBJ) free((char*)buf8);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks_invoke_dir_added(int argc, VALUE *argv, VALUE self) {
  svn_wc_diff_callbacks_t *arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  svn_wc_notify_state_t *arg3 = 0 ;
  char *arg4 = 0 ;
  svn_revnum_t arg5 ;
  void *arg6 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  long val5 ;
  int ecode5 = 0 ;
  int res6 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 6) || (argc > 6)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 6)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_diff_callbacks_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_diff_callbacks_t *","svn_wc_diff_callbacks_invoke_dir_added", 1, argv[0] )); 
  }
  arg1 = (svn_wc_diff_callbacks_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_diff_callbacks_invoke_dir_added", 2, argv[1] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  res3 = SWIG_ConvertPtr(argv[2], &argp3,SWIGTYPE_p_svn_wc_notify_state_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_wc_notify_state_t *","svn_wc_diff_callbacks_invoke_dir_added", 3, argv[2] )); 
  }
  arg3 = (svn_wc_notify_state_t *)(argp3);
  res4 = SWIG_AsCharPtrAndSize(argv[3], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks_invoke_dir_added", 4, argv[3] ));
  }
  arg4 = (char *)(buf4);
  ecode5 = SWIG_AsVal_long(argv[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), Ruby_Format_TypeError( "", "svn_revnum_t","svn_wc_diff_callbacks_invoke_dir_added", 5, argv[4] ));
  } 
  arg5 = (svn_revnum_t)(val5);
  res6 = SWIG_ConvertPtr(argv[5],SWIG_as_voidptrptr(&arg6), 0, 0);
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "void *","svn_wc_diff_callbacks_invoke_dir_added", 6, argv[5] )); 
  }
  {
    result = (svn_error_t *)svn_wc_diff_callbacks_invoke_dir_added(arg1,arg2,arg3,(char const *)arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks_invoke_dir_deleted(int argc, VALUE *argv, VALUE self) {
  svn_wc_diff_callbacks_t *arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  svn_wc_notify_state_t *arg3 = 0 ;
  char *arg4 = 0 ;
  void *arg5 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  int res5 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 5) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 5)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_diff_callbacks_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_diff_callbacks_t *","svn_wc_diff_callbacks_invoke_dir_deleted", 1, argv[0] )); 
  }
  arg1 = (svn_wc_diff_callbacks_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_diff_callbacks_invoke_dir_deleted", 2, argv[1] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  res3 = SWIG_ConvertPtr(argv[2], &argp3,SWIGTYPE_p_svn_wc_notify_state_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_wc_notify_state_t *","svn_wc_diff_callbacks_invoke_dir_deleted", 3, argv[2] )); 
  }
  arg3 = (svn_wc_notify_state_t *)(argp3);
  res4 = SWIG_AsCharPtrAndSize(argv[3], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks_invoke_dir_deleted", 4, argv[3] ));
  }
  arg4 = (char *)(buf4);
  res5 = SWIG_ConvertPtr(argv[4],SWIG_as_voidptrptr(&arg5), 0, 0);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "void *","svn_wc_diff_callbacks_invoke_dir_deleted", 5, argv[4] )); 
  }
  {
    result = (svn_error_t *)svn_wc_diff_callbacks_invoke_dir_deleted(arg1,arg2,arg3,(char const *)arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_diff_callbacks_invoke_props_changed(int argc, VALUE *argv, VALUE self) {
  svn_wc_diff_callbacks_t *arg1 = 0 ;
  svn_wc_adm_access_t *arg2 = 0 ;
  svn_wc_notify_state_t *arg3 = 0 ;
  char *arg4 = 0 ;
  apr_array_header_t *arg5 = 0 ;
  apr_hash_t *arg6 = 0 ;
  void *arg7 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  void *argp6 = 0 ;
  int res6 = 0 ;
  int res7 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 7) || (argc > 7)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 7)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_diff_callbacks_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_diff_callbacks_t *","svn_wc_diff_callbacks_invoke_props_changed", 1, argv[0] )); 
  }
  arg1 = (svn_wc_diff_callbacks_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p_svn_wc_adm_access_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "svn_wc_adm_access_t *","svn_wc_diff_callbacks_invoke_props_changed", 2, argv[1] )); 
  }
  arg2 = (svn_wc_adm_access_t *)(argp2);
  res3 = SWIG_ConvertPtr(argv[2], &argp3,SWIGTYPE_p_svn_wc_notify_state_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_wc_notify_state_t *","svn_wc_diff_callbacks_invoke_props_changed", 3, argv[2] )); 
  }
  arg3 = (svn_wc_notify_state_t *)(argp3);
  res4 = SWIG_AsCharPtrAndSize(argv[3], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_wc_diff_callbacks_invoke_props_changed", 4, argv[3] ));
  }
  arg4 = (char *)(buf4);
  {
    VALUE rb_pool;
    apr_pool_t *pool;
    
    svn_swig_rb_get_pool(argc, argv, self, &rb_pool, &pool);
    
    arg5 = svn_swig_rb_to_apr_array_prop(argv[4], pool);
  }
  res6 = SWIG_ConvertPtr(argv[5], &argp6,SWIGTYPE_p_apr_hash_t, 0 |  0 );
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "apr_hash_t *","svn_wc_diff_callbacks_invoke_props_changed", 6, argv[5] )); 
  }
  arg6 = (apr_hash_t *)(argp6);
  res7 = SWIG_ConvertPtr(argv[6],SWIG_as_voidptrptr(&arg7), 0, 0);
  if (!SWIG_IsOK(res7)) {
    SWIG_exception_fail(SWIG_ArgError(res7), Ruby_Format_TypeError( "", "void *","svn_wc_diff_callbacks_invoke_props_changed", 7, argv[6] )); 
  }
  {
    result = (svn_error_t *)svn_wc_diff_callbacks_invoke_props_changed(arg1,arg2,arg3,(char const *)arg4,(apr_array_header_t const *)arg5,arg6,arg7);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_callbacks2_invoke_found_entry(int argc, VALUE *argv, VALUE self) {
  svn_wc_entry_callbacks2_t *arg1 = 0 ;
  char *arg2 = 0 ;
  svn_wc_entry_t *arg3 = 0 ;
  void *arg4 = 0 ;
  apr_pool_t *arg5 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  int res4 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 4) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_entry_callbacks2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_entry_callbacks2_t *","svn_wc_entry_callbacks2_invoke_found_entry", 1, argv[0] )); 
  }
  arg1 = (svn_wc_entry_callbacks2_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_wc_entry_callbacks2_invoke_found_entry", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_ConvertPtr(argv[2], &argp3,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_wc_entry_t const *","svn_wc_entry_callbacks2_invoke_found_entry", 3, argv[2] )); 
  }
  arg3 = (svn_wc_entry_t *)(argp3);
  res4 = SWIG_ConvertPtr(argv[3],SWIG_as_voidptrptr(&arg4), 0, 0);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "void *","svn_wc_entry_callbacks2_invoke_found_entry", 4, argv[3] )); 
  }
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_wc_entry_callbacks2_invoke_found_entry(arg1,(char const *)arg2,(struct svn_wc_entry_t const *)arg3,arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_callbacks2_invoke_handle_error(int argc, VALUE *argv, VALUE self) {
  svn_wc_entry_callbacks2_t *arg1 = 0 ;
  char *arg2 = 0 ;
  svn_error_t *arg3 = 0 ;
  void *arg4 = 0 ;
  apr_pool_t *arg5 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  int res4 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 4) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_entry_callbacks2_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_entry_callbacks2_t *","svn_wc_entry_callbacks2_invoke_handle_error", 1, argv[0] )); 
  }
  arg1 = (svn_wc_entry_callbacks2_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_wc_entry_callbacks2_invoke_handle_error", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_ConvertPtr(argv[2], &argp3,SWIGTYPE_p_svn_error_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_error_t *","svn_wc_entry_callbacks2_invoke_handle_error", 3, argv[2] )); 
  }
  arg3 = (svn_error_t *)(argp3);
  res4 = SWIG_ConvertPtr(argv[3],SWIG_as_voidptrptr(&arg4), 0, 0);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "void *","svn_wc_entry_callbacks2_invoke_handle_error", 4, argv[3] )); 
  }
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_wc_entry_callbacks2_invoke_handle_error(arg1,(char const *)arg2,arg3,arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_entry_callbacks_invoke_found_entry(int argc, VALUE *argv, VALUE self) {
  svn_wc_entry_callbacks_t *arg1 = 0 ;
  char *arg2 = 0 ;
  svn_wc_entry_t *arg3 = 0 ;
  void *arg4 = 0 ;
  apr_pool_t *arg5 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  int res4 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 4) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p_svn_wc_entry_callbacks_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "svn_wc_entry_callbacks_t *","svn_wc_entry_callbacks_invoke_found_entry", 1, argv[0] )); 
  }
  arg1 = (svn_wc_entry_callbacks_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","svn_wc_entry_callbacks_invoke_found_entry", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_ConvertPtr(argv[2], &argp3,SWIGTYPE_p_svn_wc_entry_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_wc_entry_t const *","svn_wc_entry_callbacks_invoke_found_entry", 3, argv[2] )); 
  }
  arg3 = (svn_wc_entry_t *)(argp3);
  res4 = SWIG_ConvertPtr(argv[3],SWIG_as_voidptrptr(&arg4), 0, 0);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "void *","svn_wc_entry_callbacks_invoke_found_entry", 4, argv[3] )); 
  }
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_wc_entry_callbacks_invoke_found_entry(arg1,(char const *)arg2,(struct svn_wc_entry_t const *)arg3,arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_invoke_external_update(int argc, VALUE *argv, VALUE self) {
  svn_wc_external_update_t arg1 = 0 ;
  void *arg2 = 0 ;
  char *arg3 = 0 ;
  svn_string_t *arg4 = 0 ;
  svn_string_t *arg5 = 0 ;
  svn_depth_t arg6 ;
  apr_pool_t *arg7 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res2 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  svn_string_t value4 ;
  svn_string_t value5 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg7);
    _global_pool = arg7;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 6) || (argc > 7)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 6)",argc); SWIG_fail;
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg1), SWIGTYPE_p_f_p_void_p_q_const__char_p_q_const__svn_string_t_p_q_const__svn_string_t_svn_depth_t_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_wc_external_update_t","svn_wc_invoke_external_update", 1, argv[0] )); 
    }
  }
  res2 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void *","svn_wc_invoke_external_update", 2, argv[1] )); 
  }
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_wc_invoke_external_update", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  {
    if (NIL_P(argv[3])) {
      arg4 = NULL;
    } else {
      value4.data = StringValuePtr(argv[3]);
      value4.len = RSTRING_LEN(argv[3]);
      arg4 = &value4;
    }
  }
  {
    if (NIL_P(argv[4])) {
      arg5 = NULL;
    } else {
      value5.data = StringValuePtr(argv[4]);
      value5.len = RSTRING_LEN(argv[4]);
      arg5 = &value5;
    }
  }
  {
    arg6 = svn_swig_rb_to_depth(argv[5]);
  }
  if (argc > 6) {
    
  }
  {
    result = (svn_error_t *)svn_wc_invoke_external_update(arg1,arg2,(char const *)arg3,(struct svn_string_t const *)arg4,(struct svn_string_t const *)arg5,arg6,arg7);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_invoke_notify_func2(int argc, VALUE *argv, VALUE self) {
  svn_wc_notify_func2_t arg1 = 0 ;
  void *arg2 = 0 ;
  svn_wc_notify_t *arg3 = 0 ;
  apr_pool_t *arg4 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res2 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg4);
    _global_pool = arg4;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 3) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg1), SWIGTYPE_p_f_p_void_p_q_const__struct_svn_wc_notify_t_p_apr_pool_t__void);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_wc_notify_func2_t","svn_wc_invoke_notify_func2", 1, argv[0] )); 
    }
  }
  res2 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void *","svn_wc_invoke_notify_func2", 2, argv[1] )); 
  }
  res3 = SWIG_ConvertPtr(argv[2], &argp3,SWIGTYPE_p_svn_wc_notify_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_wc_notify_t const *","svn_wc_invoke_notify_func2", 3, argv[2] )); 
  }
  arg3 = (svn_wc_notify_t *)(argp3);
  if (argc > 3) {
    
  }
  {
    svn_wc_invoke_notify_func2(arg1,arg2,(struct svn_wc_notify_t const *)arg3,arg4);
    
    
    
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_invoke_notify_func(int argc, VALUE *argv, VALUE self) {
  svn_wc_notify_func_t arg1 = 0 ;
  void *arg2 = 0 ;
  char *arg3 = 0 ;
  svn_wc_notify_action_t arg4 ;
  svn_node_kind_t arg5 ;
  char *arg6 = 0 ;
  svn_wc_notify_state_t arg7 ;
  svn_wc_notify_state_t arg8 ;
  svn_revnum_t arg9 ;
  int res2 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int val4 ;
  int ecode4 = 0 ;
  int val5 ;
  int ecode5 = 0 ;
  int res6 ;
  char *buf6 = 0 ;
  int alloc6 = 0 ;
  int val7 ;
  int ecode7 = 0 ;
  int val8 ;
  int ecode8 = 0 ;
  long val9 ;
  int ecode9 = 0 ;
  
  if ((argc < 9) || (argc > 9)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 9)",argc); SWIG_fail;
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg1), SWIGTYPE_p_f_p_void_p_q_const__char_enum_svn_wc_notify_action_t_enum_svn_node_kind_t_p_q_const__char_enum_svn_wc_notify_state_t_enum_svn_wc_notify_state_t_long__void);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_wc_notify_func_t","svn_wc_invoke_notify_func", 1, argv[0] )); 
    }
  }
  res2 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void *","svn_wc_invoke_notify_func", 2, argv[1] )); 
  }
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_wc_invoke_notify_func", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  ecode4 = SWIG_AsVal_int(argv[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), Ruby_Format_TypeError( "", "svn_wc_notify_action_t","svn_wc_invoke_notify_func", 4, argv[3] ));
  } 
  arg4 = (svn_wc_notify_action_t)(val4);
  ecode5 = SWIG_AsVal_int(argv[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), Ruby_Format_TypeError( "", "svn_node_kind_t","svn_wc_invoke_notify_func", 5, argv[4] ));
  } 
  arg5 = (svn_node_kind_t)(val5);
  res6 = SWIG_AsCharPtrAndSize(argv[5], &buf6, NULL, &alloc6);
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "char const *","svn_wc_invoke_notify_func", 6, argv[5] ));
  }
  arg6 = (char *)(buf6);
  ecode7 = SWIG_AsVal_int(argv[6], &val7);
  if (!SWIG_IsOK(ecode7)) {
    SWIG_exception_fail(SWIG_ArgError(ecode7), Ruby_Format_TypeError( "", "svn_wc_notify_state_t","svn_wc_invoke_notify_func", 7, argv[6] ));
  } 
  arg7 = (svn_wc_notify_state_t)(val7);
  ecode8 = SWIG_AsVal_int(argv[7], &val8);
  if (!SWIG_IsOK(ecode8)) {
    SWIG_exception_fail(SWIG_ArgError(ecode8), Ruby_Format_TypeError( "", "svn_wc_notify_state_t","svn_wc_invoke_notify_func", 8, argv[7] ));
  } 
  arg8 = (svn_wc_notify_state_t)(val8);
  ecode9 = SWIG_AsVal_long(argv[8], &val9);
  if (!SWIG_IsOK(ecode9)) {
    SWIG_exception_fail(SWIG_ArgError(ecode9), Ruby_Format_TypeError( "", "svn_revnum_t","svn_wc_invoke_notify_func", 9, argv[8] ));
  } 
  arg9 = (svn_revnum_t)(val9);
  {
    svn_wc_invoke_notify_func(arg1,arg2,(char const *)arg3,arg4,arg5,(char const *)arg6,arg7,arg8,arg9);
    
    
    
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  return Qnil;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_invoke_conflict_resolver_func2(int argc, VALUE *argv, VALUE self) {
  svn_wc_conflict_resolver_func2_t arg1 = 0 ;
  svn_wc_conflict_result_t **arg2 = 0 ;
  svn_wc_conflict_description2_t *arg3 = 0 ;
  void *arg4 = 0 ;
  apr_pool_t *arg5 = 0 ;
  apr_pool_t *arg6 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_wc_conflict_result_t *temp2 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  int res4 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg2 = &temp2;
  if ((argc < 3) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg1), SWIGTYPE_p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description2_t_p_void_p_apr_pool_t_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_wc_conflict_resolver_func2_t","svn_wc_invoke_conflict_resolver_func2", 1, argv[0] )); 
    }
  }
  res3 = SWIG_ConvertPtr(argv[1], &argp3,SWIGTYPE_p_svn_wc_conflict_description2_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_wc_conflict_description2_t const *","svn_wc_invoke_conflict_resolver_func2", 3, argv[1] )); 
  }
  arg3 = (svn_wc_conflict_description2_t *)(argp3);
  res4 = SWIG_ConvertPtr(argv[2],SWIG_as_voidptrptr(&arg4), 0, 0);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "void *","svn_wc_invoke_conflict_resolver_func2", 4, argv[2] )); 
  }
  if (argc > 3) {
    
  }
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_wc_invoke_conflict_resolver_func2(arg1,arg2,(struct svn_wc_conflict_description2_t const *)arg3,arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    /* FIXME: Missing argout typemap: svn_wc_invoke_conflict_resolver_func2 arg 2 (svn_wc_conflict_result_t **) */
    
    
    
    
    SWIG_exception(SWIG_ValueError, "svn_wc_invoke_conflict_resolver_func2 is not implemented yet");
    
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_invoke_conflict_resolver_func(int argc, VALUE *argv, VALUE self) {
  svn_wc_conflict_resolver_func_t arg1 = 0 ;
  svn_wc_conflict_result_t **arg2 = 0 ;
  svn_wc_conflict_description_t *arg3 = 0 ;
  void *arg4 = 0 ;
  apr_pool_t *arg5 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_wc_conflict_result_t *temp2 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  int res4 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg2 = &temp2;
  if ((argc < 3) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg1), SWIGTYPE_p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description_t_p_void_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_wc_conflict_resolver_func_t","svn_wc_invoke_conflict_resolver_func", 1, argv[0] )); 
    }
  }
  res3 = SWIG_ConvertPtr(argv[1], &argp3,SWIGTYPE_p_svn_wc_conflict_description_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "svn_wc_conflict_description_t const *","svn_wc_invoke_conflict_resolver_func", 3, argv[1] )); 
  }
  arg3 = (svn_wc_conflict_description_t *)(argp3);
  res4 = SWIG_ConvertPtr(argv[2],SWIG_as_voidptrptr(&arg4), 0, 0);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "void *","svn_wc_invoke_conflict_resolver_func", 4, argv[2] )); 
  }
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_wc_invoke_conflict_resolver_func(arg1,arg2,(struct svn_wc_conflict_description_t const *)arg3,arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    /* FIXME: Missing argout typemap: svn_wc_invoke_conflict_resolver_func arg 2 (svn_wc_conflict_result_t **) */
    
    
    
    
    SWIG_exception(SWIG_ValueError, "svn_wc_invoke_conflict_resolver_func is not implemented yet");
    
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_invoke_status_func4(int argc, VALUE *argv, VALUE self) {
  svn_wc_status_func4_t arg1 = 0 ;
  void *arg2 = 0 ;
  char *arg3 = 0 ;
  svn_wc_status3_t *arg4 = 0 ;
  apr_pool_t *arg5 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res2 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 4) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg1), SWIGTYPE_p_f_p_void_p_q_const__char_p_q_const__svn_wc_status3_t_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_wc_status_func4_t","svn_wc_invoke_status_func4", 1, argv[0] )); 
    }
  }
  res2 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void *","svn_wc_invoke_status_func4", 2, argv[1] )); 
  }
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_wc_invoke_status_func4", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_ConvertPtr(argv[3], &argp4,SWIGTYPE_p_svn_wc_status3_t, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "svn_wc_status3_t const *","svn_wc_invoke_status_func4", 4, argv[3] )); 
  }
  arg4 = (svn_wc_status3_t *)(argp4);
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_wc_invoke_status_func4(arg1,arg2,(char const *)arg3,(struct svn_wc_status3_t const *)arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_invoke_status_func3(int argc, VALUE *argv, VALUE self) {
  svn_wc_status_func3_t arg1 = 0 ;
  void *arg2 = 0 ;
  char *arg3 = 0 ;
  svn_wc_status2_t *arg4 = 0 ;
  apr_pool_t *arg5 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res2 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 4) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg1), SWIGTYPE_p_f_p_void_p_q_const__char_p_svn_wc_status2_t_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_wc_status_func3_t","svn_wc_invoke_status_func3", 1, argv[0] )); 
    }
  }
  res2 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void *","svn_wc_invoke_status_func3", 2, argv[1] )); 
  }
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_wc_invoke_status_func3", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_ConvertPtr(argv[3], &argp4,SWIGTYPE_p_svn_wc_status2_t, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "svn_wc_status2_t *","svn_wc_invoke_status_func3", 4, argv[3] )); 
  }
  arg4 = (svn_wc_status2_t *)(argp4);
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_wc_invoke_status_func3(arg1,arg2,(char const *)arg3,arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_invoke_status_func2(int argc, VALUE *argv, VALUE self) {
  svn_wc_status_func2_t arg1 = 0 ;
  void *arg2 = 0 ;
  char *arg3 = 0 ;
  svn_wc_status2_t *arg4 = 0 ;
  int res2 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  
  if ((argc < 4) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg1), SWIGTYPE_p_f_p_void_p_q_const__char_p_struct_svn_wc_status2_t__void);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_wc_status_func2_t","svn_wc_invoke_status_func2", 1, argv[0] )); 
    }
  }
  res2 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void *","svn_wc_invoke_status_func2", 2, argv[1] )); 
  }
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_wc_invoke_status_func2", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_ConvertPtr(argv[3], &argp4,SWIGTYPE_p_svn_wc_status2_t, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "svn_wc_status2_t *","svn_wc_invoke_status_func2", 4, argv[3] )); 
  }
  arg4 = (svn_wc_status2_t *)(argp4);
  {
    svn_wc_invoke_status_func2(arg1,arg2,(char const *)arg3,arg4);
    
    
    
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return Qnil;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_invoke_status_func(int argc, VALUE *argv, VALUE self) {
  svn_wc_status_func_t arg1 = 0 ;
  void *arg2 = 0 ;
  char *arg3 = 0 ;
  svn_wc_status_t *arg4 = 0 ;
  int res2 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  
  if ((argc < 4) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg1), SWIGTYPE_p_f_p_void_p_q_const__char_p_struct_svn_wc_status_t__void);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_wc_status_func_t","svn_wc_invoke_status_func", 1, argv[0] )); 
    }
  }
  res2 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void *","svn_wc_invoke_status_func", 2, argv[1] )); 
  }
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_wc_invoke_status_func", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_ConvertPtr(argv[3], &argp4,SWIGTYPE_p_svn_wc_status_t, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "svn_wc_status_t *","svn_wc_invoke_status_func", 4, argv[3] )); 
  }
  arg4 = (svn_wc_status_t *)(argp4);
  {
    svn_wc_invoke_status_func(arg1,arg2,(char const *)arg3,arg4);
    
    
    
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return Qnil;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_invoke_get_file(int argc, VALUE *argv, VALUE self) {
  svn_wc_get_file_t arg1 = 0 ;
  void *arg2 = 0 ;
  char *arg3 = 0 ;
  svn_revnum_t arg4 ;
  svn_stream_t *arg5 = 0 ;
  svn_revnum_t *arg6 = 0 ;
  apr_hash_t **arg7 = 0 ;
  apr_pool_t *arg8 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res2 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  long val4 ;
  int ecode4 = 0 ;
  svn_revnum_t temp6 ;
  int res6 = SWIG_TMPOBJ ;
  apr_hash_t *temp7 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg8);
    _global_pool = arg8;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg6 = &temp6;
  arg7 = &temp7;
  if ((argc < 5) || (argc > 6)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 5)",argc); SWIG_fail;
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg1), SWIGTYPE_p_f_p_void_p_q_const__char_svn_revnum_t_p_svn_stream_t_p_svn_revnum_t_p_p_apr_hash_t_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_wc_get_file_t","svn_wc_invoke_get_file", 1, argv[0] )); 
    }
  }
  res2 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void *","svn_wc_invoke_get_file", 2, argv[1] )); 
  }
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_wc_invoke_get_file", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  ecode4 = SWIG_AsVal_long(argv[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), Ruby_Format_TypeError( "", "svn_revnum_t","svn_wc_invoke_get_file", 4, argv[3] ));
  } 
  arg4 = (svn_revnum_t)(val4);
  {
    arg5 = svn_swig_rb_make_stream(argv[4]);
  }
  if (argc > 5) {
    
  }
  {
    result = (svn_error_t *)svn_wc_invoke_get_file(arg1,arg2,(char const *)arg3,arg4,arg5,arg6,arg7,arg8);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (SWIG_IsTmpObj(res6)) {
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_From_long((*arg6)), 0);
  } else {
    int new_flags = SWIG_IsNewObj(res6) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    vresult = SWIG_Ruby_AppendOutput(vresult, SWIG_NewPointerObj((void*)(arg6), SWIGTYPE_p_long, new_flags), 0);
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, svn_swig_rb_apr_hash_to_hash_svn_string(*arg7), 0);
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_invoke_dirents_func(int argc, VALUE *argv, VALUE self) {
  svn_wc_dirents_func_t arg1 = 0 ;
  void *arg2 = 0 ;
  apr_hash_t **arg3 = 0 ;
  char *arg4 = 0 ;
  char *arg5 = 0 ;
  apr_pool_t *arg6 = 0 ;
  apr_pool_t *arg7 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res2 ;
  apr_hash_t *temp3 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  int res5 ;
  char *buf5 = 0 ;
  int alloc5 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg7);
    _global_pool = arg7;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg3 = &temp3;
  if ((argc < 4) || (argc > 6)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg1), SWIGTYPE_p_f_p_void_p_p_apr_hash_t_p_q_const__char_p_q_const__char_p_apr_pool_t_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_wc_dirents_func_t","svn_wc_invoke_dirents_func", 1, argv[0] )); 
    }
  }
  res2 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void *","svn_wc_invoke_dirents_func", 2, argv[1] )); 
  }
  res4 = SWIG_AsCharPtrAndSize(argv[2], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_wc_invoke_dirents_func", 4, argv[2] ));
  }
  arg4 = (char *)(buf4);
  res5 = SWIG_AsCharPtrAndSize(argv[3], &buf5, NULL, &alloc5);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "char const *","svn_wc_invoke_dirents_func", 5, argv[3] ));
  }
  arg5 = (char *)(buf5);
  if (argc > 4) {
    
  }
  if (argc > 5) {
    
  }
  {
    result = (svn_error_t *)svn_wc_invoke_dirents_func(arg1,arg2,arg3,(char const *)arg4,(char const *)arg5,arg6,arg7);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    vresult = SWIG_Ruby_AppendOutput(vresult, svn_swig_rb_apr_hash_to_hash_swig_type(*arg3,
        "svn_dirent_t *"), 0)
    
    ;
  }
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_invoke_canonicalize_svn_prop_get_file(int argc, VALUE *argv, VALUE self) {
  svn_wc_canonicalize_svn_prop_get_file_t arg1 = 0 ;
  svn_string_t **arg2 = 0 ;
  svn_stream_t *arg3 = 0 ;
  void *arg4 = 0 ;
  apr_pool_t *arg5 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  svn_string_t *temp2 ;
  int res4 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg2 = &temp2;
  if ((argc < 3) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg1), SWIGTYPE_p_f_p_p_q_const__svn_string_t_p_svn_stream_t_p_void_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_wc_canonicalize_svn_prop_get_file_t","svn_wc_invoke_canonicalize_svn_prop_get_file", 1, argv[0] )); 
    }
  }
  {
    arg3 = svn_swig_rb_make_stream(argv[1]);
  }
  res4 = SWIG_ConvertPtr(argv[2],SWIG_as_voidptrptr(&arg4), 0, 0);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "void *","svn_wc_invoke_canonicalize_svn_prop_get_file", 4, argv[2] )); 
  }
  if (argc > 3) {
    
  }
  {
    result = (svn_error_t *)svn_wc_invoke_canonicalize_svn_prop_get_file(arg1,(struct svn_string_t const **)arg2,arg3,arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    if (*arg2) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new((*arg2)->data, (*arg2)->len), 0);
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil, 0);
    }
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_invoke_upgrade_get_repos_info(int argc, VALUE *argv, VALUE self) {
  svn_wc_upgrade_get_repos_info_t arg1 = 0 ;
  char **arg2 = 0 ;
  char **arg3 = 0 ;
  void *arg4 = 0 ;
  char *arg5 = 0 ;
  apr_pool_t *arg6 = 0 ;
  apr_pool_t *arg7 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  char *temp2 ;
  char *temp3 ;
  int res4 ;
  int res5 ;
  char *buf5 = 0 ;
  int alloc5 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg7);
    _global_pool = arg7;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  arg2 = &temp2;
  arg3 = &temp3;
  if ((argc < 3) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg1), SWIGTYPE_p_f_p_p_q_const__char_p_p_q_const__char_p_void_p_q_const__char_p_apr_pool_t_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_wc_upgrade_get_repos_info_t","svn_wc_invoke_upgrade_get_repos_info", 1, argv[0] )); 
    }
  }
  res4 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg4), 0, 0);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "void *","svn_wc_invoke_upgrade_get_repos_info", 4, argv[1] )); 
  }
  res5 = SWIG_AsCharPtrAndSize(argv[2], &buf5, NULL, &alloc5);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "char const *","svn_wc_invoke_upgrade_get_repos_info", 5, argv[2] ));
  }
  arg5 = (char *)(buf5);
  if (argc > 3) {
    
  }
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_wc_invoke_upgrade_get_repos_info(arg1,(char const **)arg2,(char const **)arg3,arg4,(char const *)arg5,arg6,arg7);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    if (*arg2) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new2(*arg2), 0);
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil, 0);
    }
  }
  {
    if (*arg3) {
      vresult = SWIG_Ruby_AppendOutput(vresult, rb_str_new2(*arg3), 0);
    } else {
      vresult = SWIG_Ruby_AppendOutput(vresult, Qnil, 0);
    }
  }
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_invoke_relocation_validator3(int argc, VALUE *argv, VALUE self) {
  svn_wc_relocation_validator3_t arg1 = 0 ;
  void *arg2 = 0 ;
  char *arg3 = 0 ;
  char *arg4 = 0 ;
  char *arg5 = 0 ;
  apr_pool_t *arg6 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res2 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  int res5 ;
  char *buf5 = 0 ;
  int alloc5 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 5) || (argc > 6)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 5)",argc); SWIG_fail;
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg1), SWIGTYPE_p_f_p_void_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_wc_relocation_validator3_t","svn_wc_invoke_relocation_validator3", 1, argv[0] )); 
    }
  }
  res2 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void *","svn_wc_invoke_relocation_validator3", 2, argv[1] )); 
  }
  {
    if (NIL_P(argv[2])) {
      arg3 = NULL;
    } else {
      arg3 = StringValuePtr(argv[2]);
    }
  }
  res4 = SWIG_AsCharPtrAndSize(argv[3], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_wc_invoke_relocation_validator3", 4, argv[3] ));
  }
  arg4 = (char *)(buf4);
  res5 = SWIG_AsCharPtrAndSize(argv[4], &buf5, NULL, &alloc5);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "char const *","svn_wc_invoke_relocation_validator3", 5, argv[4] ));
  }
  arg5 = (char *)(buf5);
  if (argc > 5) {
    
  }
  {
    result = (svn_error_t *)svn_wc_invoke_relocation_validator3(arg1,arg2,(char const *)arg3,(char const *)arg4,(char const *)arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_invoke_relocation_validator2(int argc, VALUE *argv, VALUE self) {
  svn_wc_relocation_validator2_t arg1 = 0 ;
  void *arg2 = 0 ;
  char *arg3 = 0 ;
  char *arg4 = 0 ;
  svn_boolean_t arg5 ;
  apr_pool_t *arg6 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res2 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg6);
    _global_pool = arg6;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 5) || (argc > 6)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 5)",argc); SWIG_fail;
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg1), SWIGTYPE_p_f_p_void_p_q_const__char_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_wc_relocation_validator2_t","svn_wc_invoke_relocation_validator2", 1, argv[0] )); 
    }
  }
  res2 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void *","svn_wc_invoke_relocation_validator2", 2, argv[1] )); 
  }
  {
    if (NIL_P(argv[2])) {
      arg3 = NULL;
    } else {
      arg3 = StringValuePtr(argv[2]);
    }
  }
  res4 = SWIG_AsCharPtrAndSize(argv[3], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_wc_invoke_relocation_validator2", 4, argv[3] ));
  }
  arg4 = (char *)(buf4);
  arg5 = RTEST(argv[4]);
  if (argc > 5) {
    
  }
  {
    result = (svn_error_t *)svn_wc_invoke_relocation_validator2(arg1,arg2,(char const *)arg3,(char const *)arg4,arg5,arg6);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_invoke_relocation_validator(int argc, VALUE *argv, VALUE self) {
  svn_wc_relocation_validator_t arg1 = 0 ;
  void *arg2 = 0 ;
  char *arg3 = 0 ;
  char *arg4 = 0 ;
  int res2 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 4) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg1), SWIGTYPE_p_f_p_void_p_q_const__char_p_q_const__char__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_wc_relocation_validator_t","svn_wc_invoke_relocation_validator", 1, argv[0] )); 
    }
  }
  res2 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void *","svn_wc_invoke_relocation_validator", 2, argv[1] )); 
  }
  {
    if (NIL_P(argv[2])) {
      arg3 = NULL;
    } else {
      arg3 = StringValuePtr(argv[2]);
    }
  }
  res4 = SWIG_AsCharPtrAndSize(argv[3], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","svn_wc_invoke_relocation_validator", 4, argv[3] ));
  }
  arg4 = (char *)(buf4);
  {
    result = (svn_error_t *)svn_wc_invoke_relocation_validator(arg1,arg2,(char const *)arg3,(char const *)arg4);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_changelist_invoke_receiver(int argc, VALUE *argv, VALUE self) {
  svn_changelist_receiver_t arg1 = 0 ;
  void *arg2 = 0 ;
  char *arg3 = 0 ;
  char *arg4 = 0 ;
  apr_pool_t *arg5 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  int res2 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg5);
    _global_pool = arg5;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 4) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  {
    int res = SWIG_ConvertFunctionPtr(argv[0], (void**)(&arg1), SWIGTYPE_p_f_p_void_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t);
    if (!SWIG_IsOK(res)) {
      SWIG_exception_fail(SWIG_ArgError(res), Ruby_Format_TypeError( "", "svn_changelist_receiver_t","svn_changelist_invoke_receiver", 1, argv[0] )); 
    }
  }
  res2 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void *","svn_changelist_invoke_receiver", 2, argv[1] )); 
  }
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","svn_changelist_invoke_receiver", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  {
    arg4 = StringValueCStr(argv[3]);
  }
  if (argc > 4) {
    
  }
  {
    result = (svn_error_t *)svn_changelist_invoke_receiver(arg1,arg2,(char const *)arg3,(char const *)arg4,arg5);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN VALUE
_wrap_svn_wc_swig_init_asp_dot_net_hack(int argc, VALUE *argv, VALUE self) {
  apr_pool_t *arg1 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  VALUE *svn_presult = NULL ;
  svn_error_t *result = 0 ;
  VALUE vresult = Qnil;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg1);
    _global_pool = arg1;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 0) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  if (argc > 0) {
    
  }
  {
    result = (svn_error_t *)svn_wc_swig_init_asp_dot_net_hack(arg1);
    
    
    
  }
  {
    if (result) {
      svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
      svn_swig_rb_handle_svn_error(result);
    }
    vresult = rb_ary_new();
    svn_presult = &vresult;
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  {
    if (TYPE(*svn_presult) == T_ARRAY) {
      switch (rb_array_len(*svn_presult)) {
      case 0:
        *svn_presult = Qnil;
        break;
      case 1:
        *svn_presult = rb_ary_entry(*svn_presult, 0);
        break;
      }
    }
  }    return vresult;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


static swig_class SwigClassSvn_wc_committed_queue_t;

SWIGINTERN VALUE
#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
_wrap_svn_wc_committed_queue_t_allocate(VALUE self)
#else
_wrap_svn_wc_committed_queue_t_allocate(int argc, VALUE *argv, VALUE self)
#endif
{
  VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_svn_wc_committed_queue_t);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
  rb_obj_call_init(vresult, argc, argv);
#endif
  return vresult;
}


SWIGINTERN VALUE
_wrap_new_svn_wc_committed_queue_t(int argc, VALUE *argv, VALUE self) {
  apr_pool_t *arg1 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  struct svn_wc_committed_queue_t *result = 0 ;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg1);
    _global_pool = arg1;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 0) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  if (argc > 0) {
    
  }
  {
    result = (struct svn_wc_committed_queue_t *)new_svn_wc_committed_queue_t(arg1);
    DATA_PTR(self) = result;
    
    
    
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return self;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN void delete_svn_wc_committed_queue_t(struct svn_wc_committed_queue_t *self){
  }
SWIGINTERN void
free_svn_wc_committed_queue_t(void *self) {
    struct svn_wc_committed_queue_t *arg1 = (struct svn_wc_committed_queue_t *)self;
    delete_svn_wc_committed_queue_t(arg1);
}

static swig_class SwigClassSvn_wc_traversal_info_t;

SWIGINTERN VALUE
#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
_wrap_svn_wc_traversal_info_t_allocate(VALUE self)
#else
_wrap_svn_wc_traversal_info_t_allocate(int argc, VALUE *argv, VALUE self)
#endif
{
  VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_svn_wc_traversal_info_t);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
  rb_obj_call_init(vresult, argc, argv);
#endif
  return vresult;
}


SWIGINTERN VALUE
_wrap_new_svn_wc_traversal_info_t(int argc, VALUE *argv, VALUE self) {
  apr_pool_t *arg1 = 0 ;
  VALUE _global_svn_swig_rb_pool ;
  apr_pool_t *_global_pool ;
  struct svn_wc_traversal_info_t *result = 0 ;
  
  {
    svn_swig_rb_get_pool(argc, argv, self, &_global_svn_swig_rb_pool, &arg1);
    _global_pool = arg1;
    svn_swig_rb_push_pool(_global_svn_swig_rb_pool);
  }
  if ((argc < 0) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  if (argc > 0) {
    
  }
  {
    result = (struct svn_wc_traversal_info_t *)new_svn_wc_traversal_info_t(arg1);
    DATA_PTR(self) = result;
    
    
    
  }
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return self;
fail:
  {
    VALUE target;
    target = _global_vresult_address == &vresult ? self : vresult;
    if (!svn_swig_rb_set_pool(target, _global_svn_swig_rb_pool))
    svn_swig_rb_destroy_pool(_global_svn_swig_rb_pool);
    svn_swig_rb_pop_pool(_global_svn_swig_rb_pool);
  }
  return Qnil;
}


SWIGINTERN void delete_svn_wc_traversal_info_t(struct svn_wc_traversal_info_t *self){
  }
SWIGINTERN void
free_svn_wc_traversal_info_t(void *self) {
    struct svn_wc_traversal_info_t *arg1 = (struct svn_wc_traversal_info_t *)self;
    delete_svn_wc_traversal_info_t(arg1);
}


/* -------- TYPE CONVERSION AND EQUIVALENCE RULES (BEGIN) -------- */

static swig_type_info _swigt__p_apr_array_header_t = {"_p_apr_array_header_t", "svn_rangelist_t *|apr_array_header_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_apr_file_t = {"_p_apr_file_t", "apr_file_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_apr_hash_t = {"_p_apr_hash_t", "apr_hash_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_apr_int32_t = {"_p_apr_int32_t", "time_t *|apr_int32_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_apr_int64_t = {"_p_apr_int64_t", "apr_time_t *|svn_filesize_t *|apr_int64_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_apr_off_t = {"_p_apr_off_t", "apr_off_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_apr_pool_t = {"_p_apr_pool_t", "apr_pool_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_char = {"_p_char", "char *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_apr_getopt_t_p_void_p_apr_pool_t__p_svn_error_t = {"_p_f_p_apr_getopt_t_p_void_p_apr_pool_t__p_svn_error_t", "svn_opt_subcommand_t *|svn_error_t *(*)(apr_getopt_t *,void *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_p_q_const__char_p_p_q_const__char_p_void_p_q_const__char_p_apr_pool_t_p_apr_pool_t__p_svn_error_t = {"_p_f_p_p_q_const__char_p_p_q_const__char_p_void_p_q_const__char_p_apr_pool_t_p_apr_pool_t__p_svn_error_t", "svn_wc_upgrade_get_repos_info_t|svn_error_t *(*)(char const **,char const **,void *,char const *,apr_pool_t *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_p_q_const__svn_string_t_p_svn_stream_t_p_void_p_apr_pool_t__p_svn_error_t = {"_p_f_p_p_q_const__svn_string_t_p_svn_stream_t_p_void_p_apr_pool_t__p_svn_error_t", "svn_wc_canonicalize_svn_prop_get_file_t|svn_error_t *(*)(svn_string_t const **,svn_stream_t *,void *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description2_t_p_void_p_apr_pool_t_p_apr_pool_t__p_svn_error_t = {"_p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description2_t_p_void_p_apr_pool_t_p_apr_pool_t__p_svn_error_t", "svn_wc_conflict_resolver_func2_t|svn_error_t *(*)(svn_wc_conflict_result_t **,svn_wc_conflict_description2_t const *,void *,apr_pool_t *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description_t_p_void_p_apr_pool_t__p_svn_error_t = {"_p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description_t_p_void_p_apr_pool_t__p_svn_error_t", "svn_wc_conflict_resolver_func_t|svn_error_t *(*)(svn_wc_conflict_result_t **,svn_wc_conflict_description_t const *,void *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_q_const__char_p_q_const__svn_wc_entry_t_p_void_p_apr_pool_t__p_svn_error_t = {"_p_f_p_q_const__char_p_q_const__svn_wc_entry_t_p_void_p_apr_pool_t__p_svn_error_t", "svn_error_t *(*)(char const *,svn_wc_entry_t const *,void *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_q_const__char_p_svn_error_t_p_void_p_apr_pool_t__p_svn_error_t = {"_p_f_p_q_const__char_p_svn_error_t_p_void_p_apr_pool_t__p_svn_error_t", "svn_error_t *(*)(char const *,svn_error_t *,void *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_svn_boolean_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_void_p_apr_pool_t__p_svn_error_t = {"_p_f_p_svn_boolean_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_void_p_apr_pool_t__p_svn_error_t", "svn_error_t *(*)(svn_boolean_t *,svn_boolean_t *,char const *,svn_revnum_t,void *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_svn_boolean_t_p_svn_boolean_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_void_p_apr_pool_t__p_svn_error_t = {"_p_f_p_svn_boolean_t_p_svn_boolean_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_void_p_apr_pool_t__p_svn_error_t", "svn_error_t *(*)(svn_boolean_t *,svn_boolean_t *,svn_boolean_t *,char const *,svn_revnum_t,void *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_svn_wc_adm_access_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_void__p_svn_error_t = {"_p_f_p_svn_wc_adm_access_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_void__p_svn_error_t", "svn_error_t *(*)(svn_wc_adm_access_t *,svn_boolean_t *,char const *,svn_revnum_t,void *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t = {"_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t", "svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,apr_array_header_t const *,apr_hash_t *,void *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_hash_t_p_void__p_svn_error_t = {"_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_hash_t_p_void__p_svn_error_t", "svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,char const *,char const *,apr_hash_t *,void *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_void__p_svn_error_t = {"_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_void__p_svn_error_t", "svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,char const *,char const *,void *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_void__p_svn_error_t = {"_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_void__p_svn_error_t", "svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,void *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_void__p_svn_error_t = {"_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_void__p_svn_error_t", "svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,void *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_svn_revnum_t_p_void__p_svn_error_t = {"_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_svn_revnum_t_p_void__p_svn_error_t", "svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,char const *,svn_revnum_t,void *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t = {"_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t", "svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,apr_array_header_t const *,apr_hash_t *,void *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_hash_t_p_void__p_svn_error_t = {"_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_hash_t_p_void__p_svn_error_t", "svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,char const *,char const *,apr_hash_t *,void *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_void__p_svn_error_t = {"_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_void__p_svn_error_t", "svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,void *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_void__p_svn_error_t = {"_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_void__p_svn_error_t", "svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,svn_revnum_t,void *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t = {"_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t", "svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_wc_notify_state_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,apr_array_header_t const *,apr_hash_t *,void *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t = {"_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t", "svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,apr_array_header_t const *,apr_hash_t *,void *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_void__p_svn_error_t = {"_p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_void__p_svn_error_t", "svn_error_t *(*)(svn_wc_adm_access_t *,svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,void *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_hash_t_p_void_p_apr_pool_t__p_svn_error_t = {"_p_f_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_hash_t_p_void_p_apr_pool_t__p_svn_error_t", "svn_error_t *(*)(svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,char const *,char const *,apr_hash_t *,void *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_void_p_apr_pool_t__p_svn_error_t = {"_p_f_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_void_p_apr_pool_t__p_svn_error_t", "svn_error_t *(*)(svn_wc_notify_state_t *,svn_boolean_t *,char const *,void *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_svn_boolean_t_p_q_const__apr_array_header_t_p_apr_hash_t_p_void_p_apr_pool_t__p_svn_error_t = {"_p_f_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_svn_boolean_t_p_q_const__apr_array_header_t_p_apr_hash_t_p_void_p_apr_pool_t__p_svn_error_t", "svn_error_t *(*)(svn_wc_notify_state_t *,svn_boolean_t *,char const *,svn_boolean_t,apr_array_header_t const *,apr_hash_t *,void *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_svn_wc_notify_state_t_p_svn_boolean_t_p_svn_boolean_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_q_const__char_svn_revnum_t_p_void_p_apr_pool_t__p_svn_error_t = {"_p_f_p_svn_wc_notify_state_t_p_svn_boolean_t_p_svn_boolean_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_q_const__char_svn_revnum_t_p_void_p_apr_pool_t__p_svn_error_t", "svn_error_t *(*)(svn_wc_notify_state_t *,svn_boolean_t *,svn_boolean_t *,svn_boolean_t *,char const *,svn_revnum_t,char const *,svn_revnum_t,void *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void_p_apr_pool_t__p_svn_error_t = {"_p_f_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void_p_apr_pool_t__p_svn_error_t", "svn_error_t *(*)(svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,apr_array_header_t const *,apr_hash_t *,void *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_p_q_const__apr_array_header_t_p_apr_hash_t_p_void_p_apr_pool_t__p_svn_error_t = {"_p_f_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_p_q_const__apr_array_header_t_p_apr_hash_t_p_void_p_apr_pool_t__p_svn_error_t", "svn_error_t *(*)(svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,char const *,char const *,svn_revnum_t,svn_revnum_t,char const *,char const *,char const *,svn_revnum_t,apr_array_header_t const *,apr_hash_t *,void *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_svn_boolean_t_p_void_p_apr_pool_t__p_svn_error_t = {"_p_f_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_svn_boolean_t_p_void_p_apr_pool_t__p_svn_error_t", "svn_error_t *(*)(svn_wc_notify_state_t *,svn_wc_notify_state_t *,svn_boolean_t *,char const *,svn_boolean_t,void *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void__p_svn_error_t = {"_p_f_p_void__p_svn_error_t", "svn_cancel_func_t|svn_error_t *(*)(void *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_p_apr_hash_t_p_q_const__char_p_q_const__char_p_apr_pool_t_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_p_p_apr_hash_t_p_q_const__char_p_q_const__char_p_apr_pool_t_p_apr_pool_t__p_svn_error_t", "svn_wc_dirents_func_t|svn_error_t *(*)(void *,apr_hash_t **,char const *,char const *,apr_pool_t *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_q_const__char_enum_svn_wc_notify_action_t_enum_svn_node_kind_t_p_q_const__char_enum_svn_wc_notify_state_t_enum_svn_wc_notify_state_t_long__void = {"_p_f_p_void_p_q_const__char_enum_svn_wc_notify_action_t_enum_svn_node_kind_t_p_q_const__char_enum_svn_wc_notify_state_t_enum_svn_wc_notify_state_t_long__void", "svn_wc_notify_func_t|void (*)(void *,char const *,enum svn_wc_notify_action_t,enum svn_node_kind_t,char const *,enum svn_wc_notify_state_t,enum svn_wc_notify_state_t,long)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_q_const__char_p_q_const__char__p_svn_error_t = {"_p_f_p_void_p_q_const__char_p_q_const__char__p_svn_error_t", "svn_wc_relocation_validator_t|svn_error_t *(*)(void *,char const *,char const *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t", "svn_changelist_receiver_t|svn_error_t *(*)(void *,char const *,char const *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t", "svn_wc_relocation_validator3_t|svn_error_t *(*)(void *,char const *,char const *,char const *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_q_const__char_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_p_q_const__char_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t", "svn_wc_relocation_validator2_t|svn_error_t *(*)(void *,char const *,char const *,svn_boolean_t,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_q_const__char_p_q_const__svn_string_t_p_q_const__svn_string_t_svn_depth_t_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_p_q_const__char_p_q_const__svn_string_t_p_q_const__svn_string_t_svn_depth_t_p_apr_pool_t__p_svn_error_t", "svn_wc_external_update_t|svn_error_t *(*)(void *,char const *,svn_string_t const *,svn_string_t const *,svn_depth_t,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_q_const__char_p_q_const__svn_wc_status3_t_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_p_q_const__char_p_q_const__svn_wc_status3_t_p_apr_pool_t__p_svn_error_t", "svn_wc_status_func4_t|svn_error_t *(*)(void *,char const *,svn_wc_status3_t const *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_q_const__char_p_struct_svn_wc_status2_t__void = {"_p_f_p_void_p_q_const__char_p_struct_svn_wc_status2_t__void", "svn_wc_status_func2_t|void (*)(void *,char const *,struct svn_wc_status2_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_q_const__char_p_struct_svn_wc_status_t__void = {"_p_f_p_void_p_q_const__char_p_struct_svn_wc_status_t__void", "svn_wc_status_func_t|void (*)(void *,char const *,struct svn_wc_status_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_q_const__char_p_svn_wc_status2_t_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_p_q_const__char_p_svn_wc_status2_t_p_apr_pool_t__p_svn_error_t", "svn_wc_status_func3_t|svn_error_t *(*)(void *,char const *,svn_wc_status2_t *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_q_const__char_svn_revnum_t_p_svn_stream_t_p_svn_revnum_t_p_p_apr_hash_t_p_apr_pool_t__p_svn_error_t = {"_p_f_p_void_p_q_const__char_svn_revnum_t_p_svn_stream_t_p_svn_revnum_t_p_p_apr_hash_t_p_apr_pool_t__p_svn_error_t", "svn_wc_get_file_t|svn_error_t *(*)(void *,char const *,svn_revnum_t,svn_stream_t *,svn_revnum_t *,apr_hash_t **,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_f_p_void_p_q_const__struct_svn_wc_notify_t_p_apr_pool_t__void = {"_p_f_p_void_p_q_const__struct_svn_wc_notify_t_p_apr_pool_t__void", "svn_wc_notify_func2_t|void (*)(void *,struct svn_wc_notify_t const *,apr_pool_t *)", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_int = {"_p_int", "apr_status_t *|svn_boolean_t *|int *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_long = {"_p_long", "svn_revnum_t *|long *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_apr_array_header_t = {"_p_p_apr_array_header_t", "apr_array_header_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_apr_file_t = {"_p_p_apr_file_t", "apr_file_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_apr_hash_t = {"_p_p_apr_hash_t", "apr_hash_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_char = {"_p_p_char", "char **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_svn_checksum_t = {"_p_p_svn_checksum_t", "svn_checksum_t **|struct svn_checksum_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_svn_delta_editor_t = {"_p_p_svn_delta_editor_t", "svn_delta_editor_t **|struct svn_delta_editor_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_svn_stream_t = {"_p_p_svn_stream_t", "svn_stream_t **|struct svn_stream_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_svn_string_t = {"_p_p_svn_string_t", "svn_string_t **|struct svn_string_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_svn_wc_adm_access_t = {"_p_p_svn_wc_adm_access_t", "svn_wc_adm_access_t **|struct svn_wc_adm_access_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_svn_wc_committed_queue_t = {"_p_p_svn_wc_committed_queue_t", "svn_wc_committed_queue_t **|struct svn_wc_committed_queue_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_svn_wc_conflict_result_t = {"_p_p_svn_wc_conflict_result_t", "svn_wc_conflict_result_t **|struct svn_wc_conflict_result_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_svn_wc_context_t = {"_p_p_svn_wc_context_t", "svn_wc_context_t **|struct svn_wc_context_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_svn_wc_revision_status_t = {"_p_p_svn_wc_revision_status_t", "svn_wc_revision_status_t **|struct svn_wc_revision_status_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_svn_wc_status2_t = {"_p_p_svn_wc_status2_t", "svn_wc_status2_t **|struct svn_wc_status2_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_svn_wc_status3_t = {"_p_p_svn_wc_status3_t", "svn_wc_status3_t **|struct svn_wc_status3_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_svn_wc_status_t = {"_p_p_svn_wc_status_t", "svn_wc_status_t **|struct svn_wc_status_t **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_void = {"_p_p_void", "void **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_auth_baton_t = {"_p_svn_auth_baton_t", "svn_auth_baton_t *|struct svn_auth_baton_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_auth_cred_simple_t = {"_p_svn_auth_cred_simple_t", "svn_auth_cred_simple_t *|struct svn_auth_cred_simple_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_auth_cred_ssl_client_cert_pw_t = {"_p_svn_auth_cred_ssl_client_cert_pw_t", "svn_auth_cred_ssl_client_cert_pw_t *|struct svn_auth_cred_ssl_client_cert_pw_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_auth_cred_ssl_client_cert_t = {"_p_svn_auth_cred_ssl_client_cert_t", "svn_auth_cred_ssl_client_cert_t *|struct svn_auth_cred_ssl_client_cert_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_auth_cred_ssl_server_trust_t = {"_p_svn_auth_cred_ssl_server_trust_t", "svn_auth_cred_ssl_server_trust_t *|struct svn_auth_cred_ssl_server_trust_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_auth_cred_username_t = {"_p_svn_auth_cred_username_t", "svn_auth_cred_username_t *|struct svn_auth_cred_username_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_auth_iterstate_t = {"_p_svn_auth_iterstate_t", "svn_auth_iterstate_t *|struct svn_auth_iterstate_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_auth_provider_object_t = {"_p_svn_auth_provider_object_t", "svn_auth_provider_object_t *|struct svn_auth_provider_object_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_auth_provider_t = {"_p_svn_auth_provider_t", "svn_auth_provider_t *|struct svn_auth_provider_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_auth_ssl_server_cert_info_t = {"_p_svn_auth_ssl_server_cert_info_t", "svn_auth_ssl_server_cert_info_t *|struct svn_auth_ssl_server_cert_info_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_cache_config_t = {"_p_svn_cache_config_t", "svn_cache_config_t *|struct svn_cache_config_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_checksum_ctx_t = {"_p_svn_checksum_ctx_t", "svn_checksum_ctx_t *|struct svn_checksum_ctx_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_checksum_kind_t = {"_p_svn_checksum_kind_t", "svn_checksum_kind_t *|enum svn_checksum_kind_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_checksum_t = {"_p_svn_checksum_t", "svn_checksum_t *|struct svn_checksum_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_commit_info_t = {"_p_svn_commit_info_t", "svn_commit_info_t *|struct svn_commit_info_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_config_t = {"_p_svn_config_t", "svn_config_t *|struct svn_config_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_delta_editor_t = {"_p_svn_delta_editor_t", "svn_delta_editor_t *|struct svn_delta_editor_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_delta_path_driver_state_t = {"_p_svn_delta_path_driver_state_t", "svn_delta_path_driver_state_t *|struct svn_delta_path_driver_state_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_depth_t = {"_p_svn_depth_t", "svn_depth_t *|enum svn_depth_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_diff_binary_patch_t = {"_p_svn_diff_binary_patch_t", "svn_diff_binary_patch_t *|struct svn_diff_binary_patch_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_diff_conflict_display_style_t = {"_p_svn_diff_conflict_display_style_t", "svn_diff_conflict_display_style_t *|enum svn_diff_conflict_display_style_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_diff_datasource_e = {"_p_svn_diff_datasource_e", "svn_diff_datasource_e *|enum svn_diff_datasource_e *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_diff_file_ignore_space_t = {"_p_svn_diff_file_ignore_space_t", "svn_diff_file_ignore_space_t *|enum svn_diff_file_ignore_space_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_diff_file_options_t = {"_p_svn_diff_file_options_t", "svn_diff_file_options_t *|struct svn_diff_file_options_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_diff_fns2_t = {"_p_svn_diff_fns2_t", "svn_diff_fns2_t *|struct svn_diff_fns2_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_diff_fns_t = {"_p_svn_diff_fns_t", "svn_diff_fns_t *|struct svn_diff_fns_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_diff_hunk_t = {"_p_svn_diff_hunk_t", "svn_diff_hunk_t *|struct svn_diff_hunk_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_diff_operation_kind_e = {"_p_svn_diff_operation_kind_e", "svn_diff_operation_kind_t *|enum svn_diff_operation_kind_e *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_diff_output_fns_t = {"_p_svn_diff_output_fns_t", "svn_diff_output_fns_t *|struct svn_diff_output_fns_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_diff_t = {"_p_svn_diff_t", "svn_diff_t *|struct svn_diff_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_dirent_t = {"_p_svn_dirent_t", "svn_dirent_t *|struct svn_dirent_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_errno_t = {"_p_svn_errno_t", "svn_errno_t *|enum svn_errno_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_error_t = {"_p_svn_error_t", "svn_error_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_io_dirent2_t = {"_p_svn_io_dirent2_t", "svn_io_dirent2_t *|struct svn_io_dirent2_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_io_dirent_t = {"_p_svn_io_dirent_t", "svn_io_dirent_t *|struct svn_io_dirent_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_io_file_del_t = {"_p_svn_io_file_del_t", "svn_io_file_del_t *|enum svn_io_file_del_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_location_segment_t = {"_p_svn_location_segment_t", "svn_location_segment_t *|struct svn_location_segment_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_lock_t = {"_p_svn_lock_t", "svn_lock_t *|struct svn_lock_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_log_changed_path2_t = {"_p_svn_log_changed_path2_t", "svn_log_changed_path2_t *|struct svn_log_changed_path2_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_log_changed_path_t = {"_p_svn_log_changed_path_t", "svn_log_changed_path_t *|struct svn_log_changed_path_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_log_entry_t = {"_p_svn_log_entry_t", "svn_log_entry_t *|struct svn_log_entry_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_merge_range_t = {"_p_svn_merge_range_t", "svn_merge_range_t *|struct svn_merge_range_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_mergeinfo_inheritance_t = {"_p_svn_mergeinfo_inheritance_t", "svn_mergeinfo_inheritance_t *|enum svn_mergeinfo_inheritance_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_node_kind_t = {"_p_svn_node_kind_t", "svn_node_kind_t *|enum svn_node_kind_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_opt_revision_range_t = {"_p_svn_opt_revision_range_t", "svn_opt_revision_range_t *|struct svn_opt_revision_range_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_opt_revision_t = {"_p_svn_opt_revision_t", "svn_opt_revision_t *|struct svn_opt_revision_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_opt_revision_value_t = {"_p_svn_opt_revision_value_t", "svn_opt_revision_value_t *|union svn_opt_revision_value_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_opt_subcommand_desc2_t = {"_p_svn_opt_subcommand_desc2_t", "svn_opt_subcommand_desc2_t *|struct svn_opt_subcommand_desc2_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_opt_subcommand_desc3_t = {"_p_svn_opt_subcommand_desc3_t", "svn_opt_subcommand_desc3_t *|struct svn_opt_subcommand_desc3_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_opt_subcommand_desc_t = {"_p_svn_opt_subcommand_desc_t", "svn_opt_subcommand_desc_t *|struct svn_opt_subcommand_desc_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_patch_file_t = {"_p_svn_patch_file_t", "svn_patch_file_t *|struct svn_patch_file_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_patch_t = {"_p_svn_patch_t", "svn_patch_t *|struct svn_patch_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_prop_inherited_item_t = {"_p_svn_prop_inherited_item_t", "svn_prop_inherited_item_t *|struct svn_prop_inherited_item_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_prop_kind = {"_p_svn_prop_kind", "svn_prop_kind_t *|enum svn_prop_kind *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_prop_patch_t = {"_p_svn_prop_patch_t", "svn_prop_patch_t *|struct svn_prop_patch_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_ra_callbacks2_t = {"_p_svn_ra_callbacks2_t", "svn_ra_callbacks2_t *|struct svn_ra_callbacks2_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_ra_callbacks_t = {"_p_svn_ra_callbacks_t", "svn_ra_callbacks_t *|struct svn_ra_callbacks_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_ra_plugin_t = {"_p_svn_ra_plugin_t", "svn_ra_plugin_t *|struct svn_ra_plugin_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_ra_reporter2_t = {"_p_svn_ra_reporter2_t", "svn_ra_reporter2_t *|struct svn_ra_reporter2_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_ra_reporter3_t = {"_p_svn_ra_reporter3_t", "svn_ra_reporter3_t *|struct svn_ra_reporter3_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_ra_reporter_t = {"_p_svn_ra_reporter_t", "svn_ra_reporter_t *|struct svn_ra_reporter_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_ra_session_t = {"_p_svn_ra_session_t", "svn_ra_session_t *|struct svn_ra_session_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_stream_mark_t = {"_p_svn_stream_mark_t", "svn_stream_mark_t *|struct svn_stream_mark_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_stream_t = {"_p_svn_stream_t", "svn_stream_t *|struct svn_stream_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_string_t = {"_p_svn_string_t", "svn_string_t *|struct svn_string_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_stringbuf_t = {"_p_svn_stringbuf_t", "svn_stringbuf_t *|struct svn_stringbuf_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_tristate_t = {"_p_svn_tristate_t", "svn_tristate_t *|enum svn_tristate_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_txdelta_op_t = {"_p_svn_txdelta_op_t", "svn_txdelta_op_t *|struct svn_txdelta_op_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_txdelta_stream_t = {"_p_svn_txdelta_stream_t", "svn_txdelta_stream_t *|struct svn_txdelta_stream_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_txdelta_window_t = {"_p_svn_txdelta_window_t", "svn_txdelta_window_t *|struct svn_txdelta_window_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_version_checklist_t = {"_p_svn_version_checklist_t", "svn_version_checklist_t *|struct svn_version_checklist_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_version_ext_linked_lib_t = {"_p_svn_version_ext_linked_lib_t", "svn_version_ext_linked_lib_t *|struct svn_version_ext_linked_lib_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_version_ext_loaded_lib_t = {"_p_svn_version_ext_loaded_lib_t", "svn_version_ext_loaded_lib_t *|struct svn_version_ext_loaded_lib_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_version_extended_t = {"_p_svn_version_extended_t", "svn_version_extended_t *|struct svn_version_extended_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_version_t = {"_p_svn_version_t", "svn_version_t *|struct svn_version_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_adm_access_t = {"_p_svn_wc_adm_access_t", "svn_wc_adm_access_t *|struct svn_wc_adm_access_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_committed_queue_t = {"_p_svn_wc_committed_queue_t", "svn_wc_committed_queue_t *|struct svn_wc_committed_queue_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_conflict_action_t = {"_p_svn_wc_conflict_action_t", "svn_wc_conflict_action_t *|enum svn_wc_conflict_action_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_conflict_choice_t = {"_p_svn_wc_conflict_choice_t", "svn_wc_conflict_choice_t *|enum svn_wc_conflict_choice_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_conflict_description2_t = {"_p_svn_wc_conflict_description2_t", "svn_wc_conflict_description2_t *|struct svn_wc_conflict_description2_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_conflict_description_t = {"_p_svn_wc_conflict_description_t", "svn_wc_conflict_description_t *|struct svn_wc_conflict_description_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_conflict_kind_t = {"_p_svn_wc_conflict_kind_t", "svn_wc_conflict_kind_t *|enum svn_wc_conflict_kind_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_conflict_reason_t = {"_p_svn_wc_conflict_reason_t", "svn_wc_conflict_reason_t *|enum svn_wc_conflict_reason_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_conflict_result_t = {"_p_svn_wc_conflict_result_t", "svn_wc_conflict_result_t *|struct svn_wc_conflict_result_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_conflict_version_t = {"_p_svn_wc_conflict_version_t", "svn_wc_conflict_version_t *|struct svn_wc_conflict_version_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_context_t = {"_p_svn_wc_context_t", "svn_wc_context_t *|struct svn_wc_context_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_diff_callbacks2_t = {"_p_svn_wc_diff_callbacks2_t", "svn_wc_diff_callbacks2_t *|struct svn_wc_diff_callbacks2_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_diff_callbacks3_t = {"_p_svn_wc_diff_callbacks3_t", "svn_wc_diff_callbacks3_t *|struct svn_wc_diff_callbacks3_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_diff_callbacks4_t = {"_p_svn_wc_diff_callbacks4_t", "svn_wc_diff_callbacks4_t *|struct svn_wc_diff_callbacks4_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_diff_callbacks_t = {"_p_svn_wc_diff_callbacks_t", "svn_wc_diff_callbacks_t *|struct svn_wc_diff_callbacks_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_entry_callbacks2_t = {"_p_svn_wc_entry_callbacks2_t", "svn_wc_entry_callbacks2_t *|struct svn_wc_entry_callbacks2_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_entry_callbacks_t = {"_p_svn_wc_entry_callbacks_t", "svn_wc_entry_callbacks_t *|struct svn_wc_entry_callbacks_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_entry_t = {"_p_svn_wc_entry_t", "svn_wc_entry_t *|struct svn_wc_entry_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_external_item2_t = {"_p_svn_wc_external_item2_t", "svn_wc_external_item2_t *|struct svn_wc_external_item2_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_external_item_t = {"_p_svn_wc_external_item_t", "svn_wc_external_item_t *|struct svn_wc_external_item_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_info_t = {"_p_svn_wc_info_t", "svn_wc_info_t *|struct svn_wc_info_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_merge_outcome_t = {"_p_svn_wc_merge_outcome_t", "svn_wc_merge_outcome_t *|enum svn_wc_merge_outcome_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_notify_action_t = {"_p_svn_wc_notify_action_t", "svn_wc_notify_action_t *|enum svn_wc_notify_action_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_notify_lock_state_t = {"_p_svn_wc_notify_lock_state_t", "svn_wc_notify_lock_state_t *|enum svn_wc_notify_lock_state_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_notify_state_t = {"_p_svn_wc_notify_state_t", "svn_wc_notify_state_t *|enum svn_wc_notify_state_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_notify_t = {"_p_svn_wc_notify_t", "svn_wc_notify_t *|struct svn_wc_notify_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_operation_t = {"_p_svn_wc_operation_t", "svn_wc_operation_t *|enum svn_wc_operation_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_revision_status_t = {"_p_svn_wc_revision_status_t", "svn_wc_revision_status_t *|struct svn_wc_revision_status_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_schedule_t = {"_p_svn_wc_schedule_t", "svn_wc_schedule_t *|enum svn_wc_schedule_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_status2_t = {"_p_svn_wc_status2_t", "svn_wc_status2_t *|struct svn_wc_status2_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_status3_t = {"_p_svn_wc_status3_t", "svn_wc_status3_t *|struct svn_wc_status3_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_status_t = {"_p_svn_wc_status_t", "svn_wc_status_t *|struct svn_wc_status_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_svn_wc_traversal_info_t = {"_p_svn_wc_traversal_info_t", "svn_wc_traversal_info_t *|struct svn_wc_traversal_info_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_unsigned_char = {"_p_unsigned_char", "unsigned char *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_unsigned_long = {"_p_unsigned_long", "svn_linenum_t *|unsigned long *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_void = {"_p_void", "void *", 0, 0, (void*)0, 0};

static swig_type_info *swig_type_initial[] = {
  &_swigt__p_apr_array_header_t,
  &_swigt__p_apr_file_t,
  &_swigt__p_apr_hash_t,
  &_swigt__p_apr_int32_t,
  &_swigt__p_apr_int64_t,
  &_swigt__p_apr_off_t,
  &_swigt__p_apr_pool_t,
  &_swigt__p_char,
  &_swigt__p_f_p_apr_getopt_t_p_void_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_p_q_const__char_p_p_q_const__char_p_void_p_q_const__char_p_apr_pool_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_p_q_const__svn_string_t_p_svn_stream_t_p_void_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description2_t_p_void_p_apr_pool_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description_t_p_void_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_q_const__char_p_q_const__svn_wc_entry_t_p_void_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_q_const__char_p_svn_error_t_p_void_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_svn_boolean_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_void_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_svn_boolean_t_p_svn_boolean_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_void_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_svn_wc_adm_access_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_void__p_svn_error_t,
  &_swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t,
  &_swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_hash_t_p_void__p_svn_error_t,
  &_swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_void__p_svn_error_t,
  &_swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_void__p_svn_error_t,
  &_swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_void__p_svn_error_t,
  &_swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_svn_revnum_t_p_void__p_svn_error_t,
  &_swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t,
  &_swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_hash_t_p_void__p_svn_error_t,
  &_swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_void__p_svn_error_t,
  &_swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_void__p_svn_error_t,
  &_swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t,
  &_swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t,
  &_swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_void__p_svn_error_t,
  &_swigt__p_f_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_hash_t_p_void_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_void_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_svn_boolean_t_p_q_const__apr_array_header_t_p_apr_hash_t_p_void_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_svn_wc_notify_state_t_p_svn_boolean_t_p_svn_boolean_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_q_const__char_svn_revnum_t_p_void_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_p_q_const__apr_array_header_t_p_apr_hash_t_p_void_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_svn_boolean_t_p_void_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void__p_svn_error_t,
  &_swigt__p_f_p_void_p_p_apr_hash_t_p_q_const__char_p_q_const__char_p_apr_pool_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_q_const__char_enum_svn_wc_notify_action_t_enum_svn_node_kind_t_p_q_const__char_enum_svn_wc_notify_state_t_enum_svn_wc_notify_state_t_long__void,
  &_swigt__p_f_p_void_p_q_const__char_p_q_const__char__p_svn_error_t,
  &_swigt__p_f_p_void_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_q_const__char_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_q_const__char_p_q_const__svn_string_t_p_q_const__svn_string_t_svn_depth_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_q_const__char_p_q_const__svn_wc_status3_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_q_const__char_p_struct_svn_wc_status2_t__void,
  &_swigt__p_f_p_void_p_q_const__char_p_struct_svn_wc_status_t__void,
  &_swigt__p_f_p_void_p_q_const__char_p_svn_wc_status2_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_q_const__char_svn_revnum_t_p_svn_stream_t_p_svn_revnum_t_p_p_apr_hash_t_p_apr_pool_t__p_svn_error_t,
  &_swigt__p_f_p_void_p_q_const__struct_svn_wc_notify_t_p_apr_pool_t__void,
  &_swigt__p_int,
  &_swigt__p_long,
  &_swigt__p_p_apr_array_header_t,
  &_swigt__p_p_apr_file_t,
  &_swigt__p_p_apr_hash_t,
  &_swigt__p_p_char,
  &_swigt__p_p_svn_checksum_t,
  &_swigt__p_p_svn_delta_editor_t,
  &_swigt__p_p_svn_stream_t,
  &_swigt__p_p_svn_string_t,
  &_swigt__p_p_svn_wc_adm_access_t,
  &_swigt__p_p_svn_wc_committed_queue_t,
  &_swigt__p_p_svn_wc_conflict_result_t,
  &_swigt__p_p_svn_wc_context_t,
  &_swigt__p_p_svn_wc_revision_status_t,
  &_swigt__p_p_svn_wc_status2_t,
  &_swigt__p_p_svn_wc_status3_t,
  &_swigt__p_p_svn_wc_status_t,
  &_swigt__p_p_void,
  &_swigt__p_svn_auth_baton_t,
  &_swigt__p_svn_auth_cred_simple_t,
  &_swigt__p_svn_auth_cred_ssl_client_cert_pw_t,
  &_swigt__p_svn_auth_cred_ssl_client_cert_t,
  &_swigt__p_svn_auth_cred_ssl_server_trust_t,
  &_swigt__p_svn_auth_cred_username_t,
  &_swigt__p_svn_auth_iterstate_t,
  &_swigt__p_svn_auth_provider_object_t,
  &_swigt__p_svn_auth_provider_t,
  &_swigt__p_svn_auth_ssl_server_cert_info_t,
  &_swigt__p_svn_cache_config_t,
  &_swigt__p_svn_checksum_ctx_t,
  &_swigt__p_svn_checksum_kind_t,
  &_swigt__p_svn_checksum_t,
  &_swigt__p_svn_commit_info_t,
  &_swigt__p_svn_config_t,
  &_swigt__p_svn_delta_editor_t,
  &_swigt__p_svn_delta_path_driver_state_t,
  &_swigt__p_svn_depth_t,
  &_swigt__p_svn_diff_binary_patch_t,
  &_swigt__p_svn_diff_conflict_display_style_t,
  &_swigt__p_svn_diff_datasource_e,
  &_swigt__p_svn_diff_file_ignore_space_t,
  &_swigt__p_svn_diff_file_options_t,
  &_swigt__p_svn_diff_fns2_t,
  &_swigt__p_svn_diff_fns_t,
  &_swigt__p_svn_diff_hunk_t,
  &_swigt__p_svn_diff_operation_kind_e,
  &_swigt__p_svn_diff_output_fns_t,
  &_swigt__p_svn_diff_t,
  &_swigt__p_svn_dirent_t,
  &_swigt__p_svn_errno_t,
  &_swigt__p_svn_error_t,
  &_swigt__p_svn_io_dirent2_t,
  &_swigt__p_svn_io_dirent_t,
  &_swigt__p_svn_io_file_del_t,
  &_swigt__p_svn_location_segment_t,
  &_swigt__p_svn_lock_t,
  &_swigt__p_svn_log_changed_path2_t,
  &_swigt__p_svn_log_changed_path_t,
  &_swigt__p_svn_log_entry_t,
  &_swigt__p_svn_merge_range_t,
  &_swigt__p_svn_mergeinfo_inheritance_t,
  &_swigt__p_svn_node_kind_t,
  &_swigt__p_svn_opt_revision_range_t,
  &_swigt__p_svn_opt_revision_t,
  &_swigt__p_svn_opt_revision_value_t,
  &_swigt__p_svn_opt_subcommand_desc2_t,
  &_swigt__p_svn_opt_subcommand_desc3_t,
  &_swigt__p_svn_opt_subcommand_desc_t,
  &_swigt__p_svn_patch_file_t,
  &_swigt__p_svn_patch_t,
  &_swigt__p_svn_prop_inherited_item_t,
  &_swigt__p_svn_prop_kind,
  &_swigt__p_svn_prop_patch_t,
  &_swigt__p_svn_ra_callbacks2_t,
  &_swigt__p_svn_ra_callbacks_t,
  &_swigt__p_svn_ra_plugin_t,
  &_swigt__p_svn_ra_reporter2_t,
  &_swigt__p_svn_ra_reporter3_t,
  &_swigt__p_svn_ra_reporter_t,
  &_swigt__p_svn_ra_session_t,
  &_swigt__p_svn_stream_mark_t,
  &_swigt__p_svn_stream_t,
  &_swigt__p_svn_string_t,
  &_swigt__p_svn_stringbuf_t,
  &_swigt__p_svn_tristate_t,
  &_swigt__p_svn_txdelta_op_t,
  &_swigt__p_svn_txdelta_stream_t,
  &_swigt__p_svn_txdelta_window_t,
  &_swigt__p_svn_version_checklist_t,
  &_swigt__p_svn_version_ext_linked_lib_t,
  &_swigt__p_svn_version_ext_loaded_lib_t,
  &_swigt__p_svn_version_extended_t,
  &_swigt__p_svn_version_t,
  &_swigt__p_svn_wc_adm_access_t,
  &_swigt__p_svn_wc_committed_queue_t,
  &_swigt__p_svn_wc_conflict_action_t,
  &_swigt__p_svn_wc_conflict_choice_t,
  &_swigt__p_svn_wc_conflict_description2_t,
  &_swigt__p_svn_wc_conflict_description_t,
  &_swigt__p_svn_wc_conflict_kind_t,
  &_swigt__p_svn_wc_conflict_reason_t,
  &_swigt__p_svn_wc_conflict_result_t,
  &_swigt__p_svn_wc_conflict_version_t,
  &_swigt__p_svn_wc_context_t,
  &_swigt__p_svn_wc_diff_callbacks2_t,
  &_swigt__p_svn_wc_diff_callbacks3_t,
  &_swigt__p_svn_wc_diff_callbacks4_t,
  &_swigt__p_svn_wc_diff_callbacks_t,
  &_swigt__p_svn_wc_entry_callbacks2_t,
  &_swigt__p_svn_wc_entry_callbacks_t,
  &_swigt__p_svn_wc_entry_t,
  &_swigt__p_svn_wc_external_item2_t,
  &_swigt__p_svn_wc_external_item_t,
  &_swigt__p_svn_wc_info_t,
  &_swigt__p_svn_wc_merge_outcome_t,
  &_swigt__p_svn_wc_notify_action_t,
  &_swigt__p_svn_wc_notify_lock_state_t,
  &_swigt__p_svn_wc_notify_state_t,
  &_swigt__p_svn_wc_notify_t,
  &_swigt__p_svn_wc_operation_t,
  &_swigt__p_svn_wc_revision_status_t,
  &_swigt__p_svn_wc_schedule_t,
  &_swigt__p_svn_wc_status2_t,
  &_swigt__p_svn_wc_status3_t,
  &_swigt__p_svn_wc_status_t,
  &_swigt__p_svn_wc_traversal_info_t,
  &_swigt__p_unsigned_char,
  &_swigt__p_unsigned_long,
  &_swigt__p_void,
};

static swig_cast_info _swigc__p_apr_array_header_t[] = {  {&_swigt__p_apr_array_header_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_apr_file_t[] = {  {&_swigt__p_apr_file_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_apr_hash_t[] = {  {&_swigt__p_apr_hash_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_apr_int32_t[] = {  {&_swigt__p_apr_int32_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_apr_int64_t[] = {  {&_swigt__p_apr_int64_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_apr_off_t[] = {  {&_swigt__p_apr_off_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_apr_pool_t[] = {  {&_swigt__p_apr_pool_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_char[] = {  {&_swigt__p_char, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_apr_getopt_t_p_void_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_apr_getopt_t_p_void_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_p_q_const__char_p_p_q_const__char_p_void_p_q_const__char_p_apr_pool_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_p_q_const__char_p_p_q_const__char_p_void_p_q_const__char_p_apr_pool_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_p_q_const__svn_string_t_p_svn_stream_t_p_void_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_p_q_const__svn_string_t_p_svn_stream_t_p_void_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description2_t_p_void_p_apr_pool_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description2_t_p_void_p_apr_pool_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description_t_p_void_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description_t_p_void_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_q_const__char_p_q_const__svn_wc_entry_t_p_void_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_q_const__char_p_q_const__svn_wc_entry_t_p_void_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_q_const__char_p_svn_error_t_p_void_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_q_const__char_p_svn_error_t_p_void_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_svn_boolean_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_void_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_svn_boolean_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_void_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_svn_boolean_t_p_svn_boolean_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_void_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_svn_boolean_t_p_svn_boolean_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_void_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_svn_wc_adm_access_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_void__p_svn_error_t[] = {  {&_swigt__p_f_p_svn_wc_adm_access_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_void__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t[] = {  {&_swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_hash_t_p_void__p_svn_error_t[] = {  {&_swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_hash_t_p_void__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_void__p_svn_error_t[] = {  {&_swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_void__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_void__p_svn_error_t[] = {  {&_swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_void__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_void__p_svn_error_t[] = {  {&_swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_void__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_svn_revnum_t_p_void__p_svn_error_t[] = {  {&_swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_svn_revnum_t_p_void__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t[] = {  {&_swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_hash_t_p_void__p_svn_error_t[] = {  {&_swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_hash_t_p_void__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_void__p_svn_error_t[] = {  {&_swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_void__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_void__p_svn_error_t[] = {  {&_swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_void__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t[] = {  {&_swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t[] = {  {&_swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_void__p_svn_error_t[] = {  {&_swigt__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_void__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_hash_t_p_void_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_hash_t_p_void_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_void_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_void_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_svn_boolean_t_p_q_const__apr_array_header_t_p_apr_hash_t_p_void_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_svn_boolean_t_p_q_const__apr_array_header_t_p_apr_hash_t_p_void_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_svn_wc_notify_state_t_p_svn_boolean_t_p_svn_boolean_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_q_const__char_svn_revnum_t_p_void_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_svn_wc_notify_state_t_p_svn_boolean_t_p_svn_boolean_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_q_const__char_svn_revnum_t_p_void_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_p_q_const__apr_array_header_t_p_apr_hash_t_p_void_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_p_q_const__apr_array_header_t_p_apr_hash_t_p_void_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_svn_boolean_t_p_void_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_svn_boolean_t_p_void_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void__p_svn_error_t[] = {  {&_swigt__p_f_p_void__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_p_apr_hash_t_p_q_const__char_p_q_const__char_p_apr_pool_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_p_apr_hash_t_p_q_const__char_p_q_const__char_p_apr_pool_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_q_const__char_enum_svn_wc_notify_action_t_enum_svn_node_kind_t_p_q_const__char_enum_svn_wc_notify_state_t_enum_svn_wc_notify_state_t_long__void[] = {  {&_swigt__p_f_p_void_p_q_const__char_enum_svn_wc_notify_action_t_enum_svn_node_kind_t_p_q_const__char_enum_svn_wc_notify_state_t_enum_svn_wc_notify_state_t_long__void, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_q_const__char_p_q_const__char__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_q_const__char_p_q_const__char__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_q_const__char_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_q_const__char_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_q_const__char_p_q_const__svn_string_t_p_q_const__svn_string_t_svn_depth_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_q_const__char_p_q_const__svn_string_t_p_q_const__svn_string_t_svn_depth_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_q_const__char_p_q_const__svn_wc_status3_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_q_const__char_p_q_const__svn_wc_status3_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_q_const__char_p_struct_svn_wc_status2_t__void[] = {  {&_swigt__p_f_p_void_p_q_const__char_p_struct_svn_wc_status2_t__void, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_q_const__char_p_struct_svn_wc_status_t__void[] = {  {&_swigt__p_f_p_void_p_q_const__char_p_struct_svn_wc_status_t__void, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_q_const__char_p_svn_wc_status2_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_q_const__char_p_svn_wc_status2_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_q_const__char_svn_revnum_t_p_svn_stream_t_p_svn_revnum_t_p_p_apr_hash_t_p_apr_pool_t__p_svn_error_t[] = {  {&_swigt__p_f_p_void_p_q_const__char_svn_revnum_t_p_svn_stream_t_p_svn_revnum_t_p_p_apr_hash_t_p_apr_pool_t__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_f_p_void_p_q_const__struct_svn_wc_notify_t_p_apr_pool_t__void[] = {  {&_swigt__p_f_p_void_p_q_const__struct_svn_wc_notify_t_p_apr_pool_t__void, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_int[] = {  {&_swigt__p_int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_long[] = {  {&_swigt__p_long, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_apr_array_header_t[] = {  {&_swigt__p_p_apr_array_header_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_apr_file_t[] = {  {&_swigt__p_p_apr_file_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_apr_hash_t[] = {  {&_swigt__p_p_apr_hash_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_char[] = {  {&_swigt__p_p_char, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_svn_checksum_t[] = {  {&_swigt__p_p_svn_checksum_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_svn_delta_editor_t[] = {  {&_swigt__p_p_svn_delta_editor_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_svn_stream_t[] = {  {&_swigt__p_p_svn_stream_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_svn_string_t[] = {  {&_swigt__p_p_svn_string_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_svn_wc_adm_access_t[] = {  {&_swigt__p_p_svn_wc_adm_access_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_svn_wc_committed_queue_t[] = {  {&_swigt__p_p_svn_wc_committed_queue_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_svn_wc_conflict_result_t[] = {  {&_swigt__p_p_svn_wc_conflict_result_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_svn_wc_context_t[] = {  {&_swigt__p_p_svn_wc_context_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_svn_wc_revision_status_t[] = {  {&_swigt__p_p_svn_wc_revision_status_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_svn_wc_status2_t[] = {  {&_swigt__p_p_svn_wc_status2_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_svn_wc_status3_t[] = {  {&_swigt__p_p_svn_wc_status3_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_svn_wc_status_t[] = {  {&_swigt__p_p_svn_wc_status_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_void[] = {  {&_swigt__p_p_void, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_auth_baton_t[] = {  {&_swigt__p_svn_auth_baton_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_auth_cred_simple_t[] = {  {&_swigt__p_svn_auth_cred_simple_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_auth_cred_ssl_client_cert_pw_t[] = {  {&_swigt__p_svn_auth_cred_ssl_client_cert_pw_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_auth_cred_ssl_client_cert_t[] = {  {&_swigt__p_svn_auth_cred_ssl_client_cert_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_auth_cred_ssl_server_trust_t[] = {  {&_swigt__p_svn_auth_cred_ssl_server_trust_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_auth_cred_username_t[] = {  {&_swigt__p_svn_auth_cred_username_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_auth_iterstate_t[] = {  {&_swigt__p_svn_auth_iterstate_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_auth_provider_object_t[] = {  {&_swigt__p_svn_auth_provider_object_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_auth_provider_t[] = {  {&_swigt__p_svn_auth_provider_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_auth_ssl_server_cert_info_t[] = {  {&_swigt__p_svn_auth_ssl_server_cert_info_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_cache_config_t[] = {  {&_swigt__p_svn_cache_config_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_checksum_ctx_t[] = {  {&_swigt__p_svn_checksum_ctx_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_checksum_kind_t[] = {  {&_swigt__p_svn_checksum_kind_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_checksum_t[] = {  {&_swigt__p_svn_checksum_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_commit_info_t[] = {  {&_swigt__p_svn_commit_info_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_config_t[] = {  {&_swigt__p_svn_config_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_delta_editor_t[] = {  {&_swigt__p_svn_delta_editor_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_delta_path_driver_state_t[] = {  {&_swigt__p_svn_delta_path_driver_state_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_depth_t[] = {  {&_swigt__p_svn_depth_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_diff_binary_patch_t[] = {  {&_swigt__p_svn_diff_binary_patch_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_diff_conflict_display_style_t[] = {  {&_swigt__p_svn_diff_conflict_display_style_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_diff_datasource_e[] = {  {&_swigt__p_svn_diff_datasource_e, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_diff_file_ignore_space_t[] = {  {&_swigt__p_svn_diff_file_ignore_space_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_diff_file_options_t[] = {  {&_swigt__p_svn_diff_file_options_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_diff_fns2_t[] = {  {&_swigt__p_svn_diff_fns2_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_diff_fns_t[] = {  {&_swigt__p_svn_diff_fns_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_diff_hunk_t[] = {  {&_swigt__p_svn_diff_hunk_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_diff_operation_kind_e[] = {  {&_swigt__p_svn_diff_operation_kind_e, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_diff_output_fns_t[] = {  {&_swigt__p_svn_diff_output_fns_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_diff_t[] = {  {&_swigt__p_svn_diff_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_dirent_t[] = {  {&_swigt__p_svn_dirent_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_errno_t[] = {  {&_swigt__p_svn_errno_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_error_t[] = {  {&_swigt__p_svn_error_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_io_dirent2_t[] = {  {&_swigt__p_svn_io_dirent2_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_io_dirent_t[] = {  {&_swigt__p_svn_io_dirent_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_io_file_del_t[] = {  {&_swigt__p_svn_io_file_del_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_location_segment_t[] = {  {&_swigt__p_svn_location_segment_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_lock_t[] = {  {&_swigt__p_svn_lock_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_log_changed_path2_t[] = {  {&_swigt__p_svn_log_changed_path2_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_log_changed_path_t[] = {  {&_swigt__p_svn_log_changed_path_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_log_entry_t[] = {  {&_swigt__p_svn_log_entry_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_merge_range_t[] = {  {&_swigt__p_svn_merge_range_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_mergeinfo_inheritance_t[] = {  {&_swigt__p_svn_mergeinfo_inheritance_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_node_kind_t[] = {  {&_swigt__p_svn_node_kind_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_opt_revision_range_t[] = {  {&_swigt__p_svn_opt_revision_range_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_opt_revision_t[] = {  {&_swigt__p_svn_opt_revision_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_opt_revision_value_t[] = {  {&_swigt__p_svn_opt_revision_value_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_opt_subcommand_desc2_t[] = {  {&_swigt__p_svn_opt_subcommand_desc2_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_opt_subcommand_desc3_t[] = {  {&_swigt__p_svn_opt_subcommand_desc3_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_opt_subcommand_desc_t[] = {  {&_swigt__p_svn_opt_subcommand_desc_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_patch_file_t[] = {  {&_swigt__p_svn_patch_file_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_patch_t[] = {  {&_swigt__p_svn_patch_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_prop_inherited_item_t[] = {  {&_swigt__p_svn_prop_inherited_item_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_prop_kind[] = {  {&_swigt__p_svn_prop_kind, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_prop_patch_t[] = {  {&_swigt__p_svn_prop_patch_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_ra_callbacks2_t[] = {  {&_swigt__p_svn_ra_callbacks2_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_ra_callbacks_t[] = {  {&_swigt__p_svn_ra_callbacks_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_ra_plugin_t[] = {  {&_swigt__p_svn_ra_plugin_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_ra_reporter2_t[] = {  {&_swigt__p_svn_ra_reporter2_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_ra_reporter3_t[] = {  {&_swigt__p_svn_ra_reporter3_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_ra_reporter_t[] = {  {&_swigt__p_svn_ra_reporter_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_ra_session_t[] = {  {&_swigt__p_svn_ra_session_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_stream_mark_t[] = {  {&_swigt__p_svn_stream_mark_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_stream_t[] = {  {&_swigt__p_svn_stream_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_string_t[] = {  {&_swigt__p_svn_string_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_stringbuf_t[] = {  {&_swigt__p_svn_stringbuf_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_tristate_t[] = {  {&_swigt__p_svn_tristate_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_txdelta_op_t[] = {  {&_swigt__p_svn_txdelta_op_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_txdelta_stream_t[] = {  {&_swigt__p_svn_txdelta_stream_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_txdelta_window_t[] = {  {&_swigt__p_svn_txdelta_window_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_version_checklist_t[] = {  {&_swigt__p_svn_version_checklist_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_version_ext_linked_lib_t[] = {  {&_swigt__p_svn_version_ext_linked_lib_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_version_ext_loaded_lib_t[] = {  {&_swigt__p_svn_version_ext_loaded_lib_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_version_extended_t[] = {  {&_swigt__p_svn_version_extended_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_version_t[] = {  {&_swigt__p_svn_version_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_adm_access_t[] = {  {&_swigt__p_svn_wc_adm_access_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_committed_queue_t[] = {  {&_swigt__p_svn_wc_committed_queue_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_conflict_action_t[] = {  {&_swigt__p_svn_wc_conflict_action_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_conflict_choice_t[] = {  {&_swigt__p_svn_wc_conflict_choice_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_conflict_description2_t[] = {  {&_swigt__p_svn_wc_conflict_description2_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_conflict_description_t[] = {  {&_swigt__p_svn_wc_conflict_description_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_conflict_kind_t[] = {  {&_swigt__p_svn_wc_conflict_kind_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_conflict_reason_t[] = {  {&_swigt__p_svn_wc_conflict_reason_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_conflict_result_t[] = {  {&_swigt__p_svn_wc_conflict_result_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_conflict_version_t[] = {  {&_swigt__p_svn_wc_conflict_version_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_context_t[] = {  {&_swigt__p_svn_wc_context_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_diff_callbacks2_t[] = {  {&_swigt__p_svn_wc_diff_callbacks2_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_diff_callbacks3_t[] = {  {&_swigt__p_svn_wc_diff_callbacks3_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_diff_callbacks4_t[] = {  {&_swigt__p_svn_wc_diff_callbacks4_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_diff_callbacks_t[] = {  {&_swigt__p_svn_wc_diff_callbacks_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_entry_callbacks2_t[] = {  {&_swigt__p_svn_wc_entry_callbacks2_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_entry_callbacks_t[] = {  {&_swigt__p_svn_wc_entry_callbacks_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_entry_t[] = {  {&_swigt__p_svn_wc_entry_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_external_item2_t[] = {  {&_swigt__p_svn_wc_external_item2_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_external_item_t[] = {  {&_swigt__p_svn_wc_external_item_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_info_t[] = {  {&_swigt__p_svn_wc_info_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_merge_outcome_t[] = {  {&_swigt__p_svn_wc_merge_outcome_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_notify_action_t[] = {  {&_swigt__p_svn_wc_notify_action_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_notify_lock_state_t[] = {  {&_swigt__p_svn_wc_notify_lock_state_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_notify_state_t[] = {  {&_swigt__p_svn_wc_notify_state_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_notify_t[] = {  {&_swigt__p_svn_wc_notify_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_operation_t[] = {  {&_swigt__p_svn_wc_operation_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_revision_status_t[] = {  {&_swigt__p_svn_wc_revision_status_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_schedule_t[] = {  {&_swigt__p_svn_wc_schedule_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_status2_t[] = {  {&_swigt__p_svn_wc_status2_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_status3_t[] = {  {&_swigt__p_svn_wc_status3_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_status_t[] = {  {&_swigt__p_svn_wc_status_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_svn_wc_traversal_info_t[] = {  {&_swigt__p_svn_wc_traversal_info_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_unsigned_char[] = {  {&_swigt__p_unsigned_char, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_unsigned_long[] = {  {&_swigt__p_unsigned_long, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_void[] = {  {&_swigt__p_void, 0, 0, 0},{0, 0, 0, 0}};

static swig_cast_info *swig_cast_initial[] = {
  _swigc__p_apr_array_header_t,
  _swigc__p_apr_file_t,
  _swigc__p_apr_hash_t,
  _swigc__p_apr_int32_t,
  _swigc__p_apr_int64_t,
  _swigc__p_apr_off_t,
  _swigc__p_apr_pool_t,
  _swigc__p_char,
  _swigc__p_f_p_apr_getopt_t_p_void_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_p_q_const__char_p_p_q_const__char_p_void_p_q_const__char_p_apr_pool_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_p_q_const__svn_string_t_p_svn_stream_t_p_void_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description2_t_p_void_p_apr_pool_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_p_svn_wc_conflict_result_t_p_q_const__svn_wc_conflict_description_t_p_void_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_q_const__char_p_q_const__svn_wc_entry_t_p_void_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_q_const__char_p_svn_error_t_p_void_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_svn_boolean_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_void_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_svn_boolean_t_p_svn_boolean_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_void_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_svn_wc_adm_access_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_void__p_svn_error_t,
  _swigc__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t,
  _swigc__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_hash_t_p_void__p_svn_error_t,
  _swigc__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_void__p_svn_error_t,
  _swigc__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_void__p_svn_error_t,
  _swigc__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_p_void__p_svn_error_t,
  _swigc__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_q_const__char_svn_revnum_t_p_void__p_svn_error_t,
  _swigc__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t,
  _swigc__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_hash_t_p_void__p_svn_error_t,
  _swigc__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_void__p_svn_error_t,
  _swigc__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_void__p_svn_error_t,
  _swigc__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t,
  _swigc__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void__p_svn_error_t,
  _swigc__p_f_p_svn_wc_adm_access_t_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_void__p_svn_error_t,
  _swigc__p_f_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_hash_t_p_void_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_void_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_svn_boolean_t_p_q_const__apr_array_header_t_p_apr_hash_t_p_void_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_svn_wc_notify_state_t_p_svn_boolean_t_p_svn_boolean_t_p_svn_boolean_t_p_q_const__char_svn_revnum_t_p_q_const__char_svn_revnum_t_p_void_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__apr_array_header_t_p_apr_hash_t_p_void_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_svn_revnum_t_p_q_const__char_p_q_const__char_p_q_const__char_svn_revnum_t_p_q_const__apr_array_header_t_p_apr_hash_t_p_void_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_svn_wc_notify_state_t_p_svn_wc_notify_state_t_p_svn_boolean_t_p_q_const__char_svn_boolean_t_p_void_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void__p_svn_error_t,
  _swigc__p_f_p_void_p_p_apr_hash_t_p_q_const__char_p_q_const__char_p_apr_pool_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void_p_q_const__char_enum_svn_wc_notify_action_t_enum_svn_node_kind_t_p_q_const__char_enum_svn_wc_notify_state_t_enum_svn_wc_notify_state_t_long__void,
  _swigc__p_f_p_void_p_q_const__char_p_q_const__char__p_svn_error_t,
  _swigc__p_f_p_void_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void_p_q_const__char_p_q_const__char_p_q_const__char_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void_p_q_const__char_p_q_const__char_svn_boolean_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void_p_q_const__char_p_q_const__svn_string_t_p_q_const__svn_string_t_svn_depth_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void_p_q_const__char_p_q_const__svn_wc_status3_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void_p_q_const__char_p_struct_svn_wc_status2_t__void,
  _swigc__p_f_p_void_p_q_const__char_p_struct_svn_wc_status_t__void,
  _swigc__p_f_p_void_p_q_const__char_p_svn_wc_status2_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void_p_q_const__char_svn_revnum_t_p_svn_stream_t_p_svn_revnum_t_p_p_apr_hash_t_p_apr_pool_t__p_svn_error_t,
  _swigc__p_f_p_void_p_q_const__struct_svn_wc_notify_t_p_apr_pool_t__void,
  _swigc__p_int,
  _swigc__p_long,
  _swigc__p_p_apr_array_header_t,
  _swigc__p_p_apr_file_t,
  _swigc__p_p_apr_hash_t,
  _swigc__p_p_char,
  _swigc__p_p_svn_checksum_t,
  _swigc__p_p_svn_delta_editor_t,
  _swigc__p_p_svn_stream_t,
  _swigc__p_p_svn_string_t,
  _swigc__p_p_svn_wc_adm_access_t,
  _swigc__p_p_svn_wc_committed_queue_t,
  _swigc__p_p_svn_wc_conflict_result_t,
  _swigc__p_p_svn_wc_context_t,
  _swigc__p_p_svn_wc_revision_status_t,
  _swigc__p_p_svn_wc_status2_t,
  _swigc__p_p_svn_wc_status3_t,
  _swigc__p_p_svn_wc_status_t,
  _swigc__p_p_void,
  _swigc__p_svn_auth_baton_t,
  _swigc__p_svn_auth_cred_simple_t,
  _swigc__p_svn_auth_cred_ssl_client_cert_pw_t,
  _swigc__p_svn_auth_cred_ssl_client_cert_t,
  _swigc__p_svn_auth_cred_ssl_server_trust_t,
  _swigc__p_svn_auth_cred_username_t,
  _swigc__p_svn_auth_iterstate_t,
  _swigc__p_svn_auth_provider_object_t,
  _swigc__p_svn_auth_provider_t,
  _swigc__p_svn_auth_ssl_server_cert_info_t,
  _swigc__p_svn_cache_config_t,
  _swigc__p_svn_checksum_ctx_t,
  _swigc__p_svn_checksum_kind_t,
  _swigc__p_svn_checksum_t,
  _swigc__p_svn_commit_info_t,
  _swigc__p_svn_config_t,
  _swigc__p_svn_delta_editor_t,
  _swigc__p_svn_delta_path_driver_state_t,
  _swigc__p_svn_depth_t,
  _swigc__p_svn_diff_binary_patch_t,
  _swigc__p_svn_diff_conflict_display_style_t,
  _swigc__p_svn_diff_datasource_e,
  _swigc__p_svn_diff_file_ignore_space_t,
  _swigc__p_svn_diff_file_options_t,
  _swigc__p_svn_diff_fns2_t,
  _swigc__p_svn_diff_fns_t,
  _swigc__p_svn_diff_hunk_t,
  _swigc__p_svn_diff_operation_kind_e,
  _swigc__p_svn_diff_output_fns_t,
  _swigc__p_svn_diff_t,
  _swigc__p_svn_dirent_t,
  _swigc__p_svn_errno_t,
  _swigc__p_svn_error_t,
  _swigc__p_svn_io_dirent2_t,
  _swigc__p_svn_io_dirent_t,
  _swigc__p_svn_io_file_del_t,
  _swigc__p_svn_location_segment_t,
  _swigc__p_svn_lock_t,
  _swigc__p_svn_log_changed_path2_t,
  _swigc__p_svn_log_changed_path_t,
  _swigc__p_svn_log_entry_t,
  _swigc__p_svn_merge_range_t,
  _swigc__p_svn_mergeinfo_inheritance_t,
  _swigc__p_svn_node_kind_t,
  _swigc__p_svn_opt_revision_range_t,
  _swigc__p_svn_opt_revision_t,
  _swigc__p_svn_opt_revision_value_t,
  _swigc__p_svn_opt_subcommand_desc2_t,
  _swigc__p_svn_opt_subcommand_desc3_t,
  _swigc__p_svn_opt_subcommand_desc_t,
  _swigc__p_svn_patch_file_t,
  _swigc__p_svn_patch_t,
  _swigc__p_svn_prop_inherited_item_t,
  _swigc__p_svn_prop_kind,
  _swigc__p_svn_prop_patch_t,
  _swigc__p_svn_ra_callbacks2_t,
  _swigc__p_svn_ra_callbacks_t,
  _swigc__p_svn_ra_plugin_t,
  _swigc__p_svn_ra_reporter2_t,
  _swigc__p_svn_ra_reporter3_t,
  _swigc__p_svn_ra_reporter_t,
  _swigc__p_svn_ra_session_t,
  _swigc__p_svn_stream_mark_t,
  _swigc__p_svn_stream_t,
  _swigc__p_svn_string_t,
  _swigc__p_svn_stringbuf_t,
  _swigc__p_svn_tristate_t,
  _swigc__p_svn_txdelta_op_t,
  _swigc__p_svn_txdelta_stream_t,
  _swigc__p_svn_txdelta_window_t,
  _swigc__p_svn_version_checklist_t,
  _swigc__p_svn_version_ext_linked_lib_t,
  _swigc__p_svn_version_ext_loaded_lib_t,
  _swigc__p_svn_version_extended_t,
  _swigc__p_svn_version_t,
  _swigc__p_svn_wc_adm_access_t,
  _swigc__p_svn_wc_committed_queue_t,
  _swigc__p_svn_wc_conflict_action_t,
  _swigc__p_svn_wc_conflict_choice_t,
  _swigc__p_svn_wc_conflict_description2_t,
  _swigc__p_svn_wc_conflict_description_t,
  _swigc__p_svn_wc_conflict_kind_t,
  _swigc__p_svn_wc_conflict_reason_t,
  _swigc__p_svn_wc_conflict_result_t,
  _swigc__p_svn_wc_conflict_version_t,
  _swigc__p_svn_wc_context_t,
  _swigc__p_svn_wc_diff_callbacks2_t,
  _swigc__p_svn_wc_diff_callbacks3_t,
  _swigc__p_svn_wc_diff_callbacks4_t,
  _swigc__p_svn_wc_diff_callbacks_t,
  _swigc__p_svn_wc_entry_callbacks2_t,
  _swigc__p_svn_wc_entry_callbacks_t,
  _swigc__p_svn_wc_entry_t,
  _swigc__p_svn_wc_external_item2_t,
  _swigc__p_svn_wc_external_item_t,
  _swigc__p_svn_wc_info_t,
  _swigc__p_svn_wc_merge_outcome_t,
  _swigc__p_svn_wc_notify_action_t,
  _swigc__p_svn_wc_notify_lock_state_t,
  _swigc__p_svn_wc_notify_state_t,
  _swigc__p_svn_wc_notify_t,
  _swigc__p_svn_wc_operation_t,
  _swigc__p_svn_wc_revision_status_t,
  _swigc__p_svn_wc_schedule_t,
  _swigc__p_svn_wc_status2_t,
  _swigc__p_svn_wc_status3_t,
  _swigc__p_svn_wc_status_t,
  _swigc__p_svn_wc_traversal_info_t,
  _swigc__p_unsigned_char,
  _swigc__p_unsigned_long,
  _swigc__p_void,
};


/* -------- TYPE CONVERSION AND EQUIVALENCE RULES (END) -------- */

/* -----------------------------------------------------------------------------
 * Type initialization:
 * This problem is made tough by the requirement that no dynamic
 * memory is used. Also, since swig_type_info structures store pointers to
 * swig_cast_info structures and swig_cast_info structures store pointers back
 * to swig_type_info structures, we need some lookup code at initialization.
 * The idea is that swig generates all the structures that are needed.
 * The runtime then collects these partially filled structures.
 * The SWIG_InitializeModule function takes these initial arrays out of
 * swig_module, and does all the lookup, filling in the swig_module.types
 * array with the correct data and linking the correct swig_cast_info
 * structures together.
 *
 * The generated swig_type_info structures are assigned statically to an initial
 * array. We just loop through that array, and handle each type individually.
 * First we lookup if this type has been already loaded, and if so, use the
 * loaded structure instead of the generated one. Then we have to fill in the
 * cast dependencies. The cast data is initially stored in something like a
 * two-dimensional array. Each row corresponds to a type (there are the same
 * number of rows as there are in the swig_type_initial array). Each entry in
 * a column is one of the swig_cast_info structures for that type.
 * The cast_initial array is actually an array of arrays, because each row has
 * a variable number of columns.
 *
 * We loop through the cast[] array associated with the type and mark casts
 * which have not been defined in previously loaded modules by assigning
 * cast pointer value to cast->next. We also hash cast->type->name string
 * and store the value in the cast->value field. If we encounter swig_cast_info
 * structure that represents a cast to self we move it to the beginning
 * of the cast array. One trick we need to do is making sure the type pointer
 * in the swig_cast_info struct is correct.
 *
 * First off, we lookup the cast->type name to see if it is already loaded.
 * There are three cases to handle:
 *  1) If the cast->type has already been loaded AND the type we are adding
 *     casting info to has not been loaded (it is in this module), THEN we
 *     replace the cast->type pointer with the type pointer that has already
 *     been loaded.
 *  2) If BOTH types (the one we are adding casting info to, and the
 *     cast->type) are loaded, THEN the cast info has already been loaded by
 *     the previous module so we just ignore it.
 *  3) Finally, if cast->type has not already been loaded, then we add that
 *     swig_cast_info to the list (because the cast->type) pointer will
 *     be correct.
 *
 * Once the cast array has been set up AND it does have new casts that need
 * to be added we sort non-self cast entries to move filtered out entries
 * to the end of the array and to arrange the rest in the increasing order
 * of their type pointer values. We store the index of the last added entry
 * in the cast->value field of the entry[0] (overwriting the name hash). Then
 * we sort <next|value> fields of the remaining entries to arrange hash values
 * in the increasing order. This way cast->next->type->name field matches
 * the cast->value hash.
 *
 * Example:
 * Array of casts for type stored at 0x5000, cast to type stored at 0x3000
 * has already been loaded
 *
 *      After sweep-and-hash:   After sort-by-type:    After sort-by-hash:
 *        ________________       ________________       ________________
 *       |                |     |                |     |                |
 * Entry |  type = 0x5000 |     |  type = 0x5000 |     |  type = 0x5000 |
 *   0   |                |     |                |     |                |
 *       |  next = Entry0 |     |  next = Entry0 |     |  next = Entry0 |
 *       | value = 1212   |     | value = 3      |     | value = 3      |
 *       |                |     |                |     |                |
 *       |================|     |================|     |================|
 *       |                |     |                |     |                |
 * Entry |  type = 0x2000 |     |  type = 0x1000 |     |  type = 0x1000 |
 *   1   |                |     |                |     |                |
 *       |  next = Entry1 |     |  next = Entry1 |     |  next = Entry3 |
 *       | value = 3434   |     | value = 4545   |     | value = 2323   |
 *       |________________|     |________________|     |________________|
 *       |                |     |                |     |                |
 * Entry |  type = 0x3000 |     |  type = 0x2000 |     |  type = 0x2000 |
 *   2   |                |     |                |     |                |
 *       |  next = 0      |     |  next = Entry2 |     |  next = Entry2 |
 *       | value = 0      |     | value = 3434   |     | value = 3434   |
 *       |________________|     |________________|     |________________|
 *       |                |     |                |     |                |
 * Entry |  type = 0x1000 |     |  type = 0x4000 |     |  type = 0x4000 |
 *   3   |                |     |                |     |                |
 *       |  next = Entry3 |     |  next = Entry3 |     |  next = Entry1 |
 *       | value = 4545   |     | value = 2323   |     | value = 4545   |
 *       |________________|     |________________|     |________________|
 *       |                |     |                |     |                |
 * Entry |  type = 0x4000 |     |  type = 0x3000 |     |  type = 0x3000 |
 *   4   |                |     |                |     |                |
 *       |  next = Entry4 |     |  next = 0      |     |  next = 0      |
 *       | value = 2323   |     | value = 0      |     | value = 0      |
 *       |________________|     |________________|     |________________|
 *
 * Once the cast array has been initialized, we use cast[0]->next field to link
 * it into the list of cast arrays for the type.
 *              ____    ____    ____
 * type->cast->|next|->|next|->|next|->0
 *             |----|  |----|  |----|
 *             |----|  |----|  |----|
 *             |----|  |----|  |----|
 *
 * Subsequent cast resolution works as follows:
 *
 * 1. Check whether the type matches the first entry in the current cast array.
 * 2. If not, then do a binary search over the (0:cast->value] interval using
 *    either type address or the hash value of the type name.
 * 3. If not found, then move over to the next cast array (cast[0]->next).
 *
 * ----------------------------------------------------------------------------- */

#ifdef __cplusplus
extern "C" {
#if 0
} /* c-mode */
#endif
#endif

#if 0
#define SWIGRUNTIME_DEBUG
#endif

#ifndef SWIG_INIT_CLIENT_DATA_TYPE
#define SWIG_INIT_CLIENT_DATA_TYPE void *
#endif

/*
 * Sort function that puts cast entries with nonzero 'next' at the front
 * of the array while ordering them by addresses of their 'type' structs.
 */
SWIGINTERN int SWIG_CastCmpStruct(const void *pa, const void *pb) {
  swig_cast_info *pca = (swig_cast_info *)pa;
  swig_cast_info *pcb = (swig_cast_info *)pb;
  if (pca->type < pcb->type)
    return (pca->next || pcb->next == 0) ? -1 : 1;
  if (pca->type > pcb->type)
    return (pcb->next || pca->next == 0) ? 1 : -1;
  return 0;
}

/*
 * Shell-sort 'next' and 'value' field pairs to order them by 'value'.
 */
SWIGINTERN void SWIG_CastHashSort(swig_cast_info *cast, int size) {
  const int hmax = size/9;
  int h, i;
  for(h = 1; h <= hmax; h = 3*h+1);
  for(; h > 0; h /= 3)
  {
    for(i = h; i < size; ++i)
    {
      swig_cast_info *p = cast[i].next;
      unsigned int hash = cast[i].value;
      int j = i;
      while(j >= h && hash < cast[j-h].value)
      {
        cast[j].next = cast[j-h].next;
        cast[j].value = cast[j-h].value;
        j -= h;
      }
      cast[j].next = p;
      cast[j].value = hash;
    }
  }
}

SWIGRUNTIME void
SWIG_InitializeModule(SWIG_INIT_CLIENT_DATA_TYPE clientdata) {
  size_t i;
  swig_module_info *module_head, *iter;
  int init;

  /* check to see if the circular list has been setup, if not, set it up */
  if (swig_module.next==0) {
    /* Initialize the swig_module */
    swig_module.type_initial = swig_type_initial;
    swig_module.cast_initial = swig_cast_initial;
    swig_module.next = &swig_module;
    init = 1;
  } else {
    init = 0;
  }

  /* Try and load any already created modules */
  module_head = SWIG_GetModule(clientdata);
  if (!module_head) {
    /* This is the first module loaded for this interpreter */
    /* so set the swig module into the interpreter */
    SWIG_SetModule(clientdata, &swig_module);
  } else {
    /* the interpreter has loaded a SWIG module, but has it loaded this one? */
    iter=module_head;
    do {
      if (iter==&swig_module) {
        /* Our module is already in the list, so there's nothing more to do. */
        return;
      }
      iter=iter->next;
    } while (iter!= module_head);

    /* otherwise we must add our module into the list */
    swig_module.next = module_head->next;
    module_head->next = &swig_module;
  }

  /* When multiple interpreters are used, a module could have already been initialized in
     a different interpreter, but not yet have a pointer in this interpreter.
     In this case, we do not want to continue adding types... everything should be
     set up already */
  if (init == 0) return;

  /* Now work on filling in swig_module.types */
#ifdef SWIGRUNTIME_DEBUG
  printf("SWIG_InitializeModule: size %lu\n", (unsigned long)swig_module.size);
#endif
  for (i = 0; i < swig_module.size; ++i) {
    swig_type_info *type = 0;
    swig_type_info *target_type;
    swig_cast_info *cast, *first;
    int num_mapped = 0;

#ifdef SWIGRUNTIME_DEBUG
    printf("SWIG_InitializeModule: type %lu %s\n", (unsigned long)i, swig_module.type_initial[i]->name);
#endif

    /* if there is another module already loaded */
    if (swig_module.next != &swig_module) {
      type = SWIG_MangledTypeQueryModule(swig_module.next, &swig_module, swig_module.type_initial[i]->name);
    }
    if (type) {
      /* Overwrite clientdata field */
#ifdef SWIGRUNTIME_DEBUG
      printf("SWIG_InitializeModule: found type %s\n", type->name);
#endif
      if (swig_module.type_initial[i]->clientdata) {
	type->clientdata = swig_module.type_initial[i]->clientdata;
#ifdef SWIGRUNTIME_DEBUG
      printf("SWIG_InitializeModule: found and overwrite type %s \n", type->name);
#endif
      }
    } else {
      type = swig_module.type_initial[i];
    }

    /* Insert casting types */
    cast = first = swig_module.cast_initial[i];
    while (cast->type) {

      /* Don't need to add information already in the list */
      target_type = 0;
#ifdef SWIGRUNTIME_DEBUG
      printf("SWIG_InitializeModule: look cast %s\n", cast->type->name);
#endif
      if (swig_module.next != &swig_module) {
        target_type = SWIG_MangledTypeQueryModule(swig_module.next, &swig_module, cast->type->name);
        if (target_type) {

          /* Target type already defined in another module */
#ifdef SWIGRUNTIME_DEBUG
          printf("SWIG_InitializeModule: found cast %s\n", target_type->name);
#endif
          if (type == swig_module.type_initial[i]) {
#ifdef SWIGRUNTIME_DEBUG
            printf("SWIG_InitializeModule: skip old type %s\n", target_type->name);
#endif
            cast->type = target_type;
            target_type = 0;
          } else {
            /* Check if this cast is already in the list */
            swig_cast_info *ocast = SWIG_TypeCheck(target_type->name, type);
#ifdef SWIGRUNTIME_DEBUG
            if (ocast) printf("SWIG_InitializeModule: skip old cast %s\n", target_type->name);
#endif
            if (!ocast) target_type = 0;
          }
        }
      }

      if (!target_type) {
#ifdef SWIGRUNTIME_DEBUG
        printf("SWIG_InitializeModule: adding cast %s\n", cast->type->name);
#endif
        /* Set inclusion mark for sorting */
        cast->next = cast;
        num_mapped++;

        if (type == cast->type) {
#ifdef SWIGRUNTIME_DEBUG
          printf("%s : self cast at pos [%li]\n", type->name, cast - first);
#endif
          if (cast - first) {

            /* Move cast to itself to the first entry in the array */

            swig_cast_info tmp = *cast;
            *cast = *first;
            *first = tmp;
          }
          first++;

        } else {
          cast->value = SWIG_Hash(cast->type->name, (unsigned int)strlen(cast->type->name));
        }
      }
      cast++;
    }

    if (num_mapped) {
      if (cast - first) {
        swig_cast_info *tmp;

        /* Sort casts by type address for binary search in SWIG_TypeCheckStruct */
        qsort(first, cast - first, sizeof(swig_cast_info), SWIG_CastCmpStruct);

        /* Remap back links for added entries */
        cast = swig_module.cast_initial[i] + num_mapped;
        for (tmp = first; tmp < cast; tmp++) {
          tmp->next = tmp;
        }
      }

      /* Set the value field of the first entry to the index of the last added entry */
      cast = swig_module.cast_initial[i];
      cast->value = num_mapped - 1;

      num_mapped -= (int)(first - cast);
      if (num_mapped > 1) {
        /* Sort <'next','value'> pairs by 'value' for binary search in SWIG_TypeCheck */

        SWIG_CastHashSort(first, num_mapped);
      }

      first = type->cast;
      if (first) {
        /* Link the current set into the list of cast arrays */
        cast->next = first->next;
        first->next = cast;
      } else {
        cast->next = 0;
        type->cast = cast;
      }
    }

    /* Set entry in modules->types array equal to the type */
    swig_module.types[i] = type;
  }
  swig_module.types[i] = 0;

#ifdef SWIGRUNTIME_DEBUG
  printf("**** SWIG_InitializeModule: Cast List ******\n");
  for (i = 0; i < swig_module.size; ++i) {
    int j = 0;
    swig_cast_info *cast = swig_module.cast_initial[i];
    printf("SWIG_InitializeModule: type %lu %s\n", (unsigned long)i, swig_module.type_initial[i]->name);
    while (cast->type) {
      printf("SWIG_InitializeModule: cast type %s\n", cast->type->name);
      cast++;
      ++j;
    }
  printf("---- Total casts: %d\n",j);
  }
  printf("**** SWIG_InitializeModule: Cast List ******\n");
#endif
}

/* This function will propagate the clientdata field of type to
* any new swig_type_info structures that have been added into the list
* of equivalent types.  It is like calling
* SWIG_TypeClientData(type, clientdata) a second time.
*/
SWIGRUNTIME void
SWIG_PropagateClientData(void) {
  size_t i;
  static int init_run = 0;

  if (init_run) return;
  init_run = 1;

  for (i = 0; i < swig_module.size; i++) {
    if (swig_module.types[i]->clientdata) {
      swig_cast_info *head, *cast;
      head = swig_module.types[i]->cast;
      while (head) {
        for (cast = head; (cast - head) <= head->value; cast++) {
          if (!cast->converter) {
            if (cast->type && !cast->type->clientdata)
              SWIG_TypeClientData(cast->type, swig_module.types[i]->clientdata);
          }
        }
        head = head->next;
      }
    }
  }
}

#ifdef __cplusplus
#if 0
{ /* c-mode */
#endif
}
#endif

/*

*/
#ifdef __cplusplus
extern "C"
#endif
SWIGEXPORT void Init_wc(void) {
  size_t i;
  
  SWIG_InitRuntime();
  mWc = rb_define_module("Svn");
  mWc = rb_define_module_under(mWc, "Ext");
  mWc = rb_define_module_under(mWc, "Wc");
  
  SWIG_InitializeModule(0);
  for (i = 0; i < swig_module.size; i++) {
    SWIG_define_class(swig_module.types[i]);
  }
  
  SWIG_RubyInitializeTrackings();
  rb_require("svn/ext/core");
  rb_require("svn/ext/delta");
  rb_require("svn/ext/ra");
  rb_define_module_function(mWc, "svn_wc_version", _wrap_svn_wc_version, -1);
  rb_define_const(mWc, "SVN_WC_TRANSLATE_FROM_NF", SWIG_From_int((int)(0x00000000)));
  rb_define_const(mWc, "SVN_WC_TRANSLATE_TO_NF", SWIG_From_int((int)(0x00000001)));
  rb_define_const(mWc, "SVN_WC_TRANSLATE_FORCE_EOL_REPAIR", SWIG_From_int((int)(0x00000002)));
  rb_define_const(mWc, "SVN_WC_TRANSLATE_NO_OUTPUT_CLEANUP", SWIG_From_int((int)(0x00000004)));
  rb_define_const(mWc, "SVN_WC_TRANSLATE_FORCE_COPY", SWIG_From_int((int)(0x00000008)));
  rb_define_const(mWc, "SVN_WC_TRANSLATE_USE_GLOBAL_TMP", SWIG_From_int((int)(0x00000010)));
  rb_define_module_function(mWc, "svn_wc_context_create", _wrap_svn_wc_context_create, -1);
  rb_define_module_function(mWc, "svn_wc_context_destroy", _wrap_svn_wc_context_destroy, -1);
  rb_define_module_function(mWc, "svn_wc_adm_open3", _wrap_svn_wc_adm_open3, -1);
  rb_define_module_function(mWc, "svn_wc_adm_open2", _wrap_svn_wc_adm_open2, -1);
  rb_define_module_function(mWc, "svn_wc_adm_open", _wrap_svn_wc_adm_open, -1);
  rb_define_module_function(mWc, "svn_wc_adm_probe_open3", _wrap_svn_wc_adm_probe_open3, -1);
  rb_define_module_function(mWc, "svn_wc_adm_probe_open2", _wrap_svn_wc_adm_probe_open2, -1);
  rb_define_module_function(mWc, "svn_wc_adm_probe_open", _wrap_svn_wc_adm_probe_open, -1);
  rb_define_module_function(mWc, "svn_wc_adm_open_anchor", _wrap_svn_wc_adm_open_anchor, -1);
  rb_define_module_function(mWc, "svn_wc_adm_retrieve", _wrap_svn_wc_adm_retrieve, -1);
  rb_define_module_function(mWc, "svn_wc_adm_probe_retrieve", _wrap_svn_wc_adm_probe_retrieve, -1);
  rb_define_module_function(mWc, "svn_wc_adm_probe_try3", _wrap_svn_wc_adm_probe_try3, -1);
  rb_define_module_function(mWc, "svn_wc_adm_probe_try2", _wrap_svn_wc_adm_probe_try2, -1);
  rb_define_module_function(mWc, "svn_wc_adm_probe_try", _wrap_svn_wc_adm_probe_try, -1);
  rb_define_module_function(mWc, "svn_wc_adm_close2", _wrap_svn_wc_adm_close2, -1);
  rb_define_module_function(mWc, "svn_wc_adm_close", _wrap_svn_wc_adm_close, -1);
  rb_define_module_function(mWc, "svn_wc_adm_access_path", _wrap_svn_wc_adm_access_path, -1);
  rb_define_module_function(mWc, "svn_wc_adm_access_pool", _wrap_svn_wc_adm_access_pool, -1);
  rb_define_module_function(mWc, "svn_wc_adm_locked", _wrap_svn_wc_adm_locked, -1);
  rb_define_module_function(mWc, "svn_wc_locked2", _wrap_svn_wc_locked2, -1);
  rb_define_module_function(mWc, "svn_wc_locked", _wrap_svn_wc_locked, -1);
  rb_define_const(mWc, "SVN_WC_ADM_DIR_NAME", SWIG_FromCharPtr(".svn"));
  rb_define_module_function(mWc, "svn_wc_is_adm_dir", _wrap_svn_wc_is_adm_dir, -1);
  rb_define_module_function(mWc, "svn_wc_get_adm_dir", _wrap_svn_wc_get_adm_dir, -1);
  rb_define_module_function(mWc, "svn_wc_set_adm_dir", _wrap_svn_wc_set_adm_dir, -1);
  rb_define_module_function(mWc, "svn_wc_edited_externals", _wrap_svn_wc_edited_externals, -1);
  rb_define_module_function(mWc, "svn_wc_traversed_depths", _wrap_svn_wc_traversed_depths, -1);
  
  SwigClassSvn_wc_external_item2_t.klass = rb_define_class_under(mWc, "Svn_wc_external_item2_t", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_svn_wc_external_item2_t, (void *) &SwigClassSvn_wc_external_item2_t);
  rb_define_alloc_func(SwigClassSvn_wc_external_item2_t.klass, _wrap_svn_wc_external_item2_t_allocate);
  rb_define_method(SwigClassSvn_wc_external_item2_t.klass, "initialize", _wrap_new_svn_wc_external_item2_t, -1);
  rb_define_method(SwigClassSvn_wc_external_item2_t.klass, "target_dir=", _wrap_svn_wc_external_item2_t_target_dir_set, -1);
  rb_define_method(SwigClassSvn_wc_external_item2_t.klass, "target_dir", _wrap_svn_wc_external_item2_t_target_dir_get, -1);
  rb_define_method(SwigClassSvn_wc_external_item2_t.klass, "url=", _wrap_svn_wc_external_item2_t_url_set, -1);
  rb_define_method(SwigClassSvn_wc_external_item2_t.klass, "url", _wrap_svn_wc_external_item2_t_url_get, -1);
  rb_define_method(SwigClassSvn_wc_external_item2_t.klass, "revision=", _wrap_svn_wc_external_item2_t_revision_set, -1);
  rb_define_method(SwigClassSvn_wc_external_item2_t.klass, "revision", _wrap_svn_wc_external_item2_t_revision_get, -1);
  rb_define_method(SwigClassSvn_wc_external_item2_t.klass, "peg_revision=", _wrap_svn_wc_external_item2_t_peg_revision_set, -1);
  rb_define_method(SwigClassSvn_wc_external_item2_t.klass, "peg_revision", _wrap_svn_wc_external_item2_t_peg_revision_get, -1);
  rb_define_method(SwigClassSvn_wc_external_item2_t.klass, "dup", _wrap_svn_wc_external_item2_t_dup, -1);
  SwigClassSvn_wc_external_item2_t.mark = 0;
  SwigClassSvn_wc_external_item2_t.destroy = (void (*)(void *)) free_svn_wc_external_item2_t;
  SwigClassSvn_wc_external_item2_t.trackObjects = 0;
  
  SwigClassSvn_wc_external_item_t.klass = rb_define_class_under(mWc, "Svn_wc_external_item_t", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_svn_wc_external_item_t, (void *) &SwigClassSvn_wc_external_item_t);
  rb_define_alloc_func(SwigClassSvn_wc_external_item_t.klass, _wrap_svn_wc_external_item_t_allocate);
  rb_define_method(SwigClassSvn_wc_external_item_t.klass, "initialize", _wrap_new_svn_wc_external_item_t, -1);
  rb_define_method(SwigClassSvn_wc_external_item_t.klass, "target_dir=", _wrap_svn_wc_external_item_t_target_dir_set, -1);
  rb_define_method(SwigClassSvn_wc_external_item_t.klass, "target_dir", _wrap_svn_wc_external_item_t_target_dir_get, -1);
  rb_define_method(SwigClassSvn_wc_external_item_t.klass, "url=", _wrap_svn_wc_external_item_t_url_set, -1);
  rb_define_method(SwigClassSvn_wc_external_item_t.klass, "url", _wrap_svn_wc_external_item_t_url_get, -1);
  rb_define_method(SwigClassSvn_wc_external_item_t.klass, "revision=", _wrap_svn_wc_external_item_t_revision_set, -1);
  rb_define_method(SwigClassSvn_wc_external_item_t.klass, "revision", _wrap_svn_wc_external_item_t_revision_get, -1);
  SwigClassSvn_wc_external_item_t.mark = 0;
  SwigClassSvn_wc_external_item_t.destroy = (void (*)(void *)) free_svn_wc_external_item_t;
  SwigClassSvn_wc_external_item_t.trackObjects = 0;
  rb_define_module_function(mWc, "svn_wc_parse_externals_description3", _wrap_svn_wc_parse_externals_description3, -1);
  rb_define_module_function(mWc, "svn_wc_parse_externals_description2", _wrap_svn_wc_parse_externals_description2, -1);
  rb_define_module_function(mWc, "svn_wc_parse_externals_description", _wrap_svn_wc_parse_externals_description, -1);
  rb_define_const(mWc, "Svn_wc_notify_add", SWIG_From_int((int)(svn_wc_notify_add)));
  rb_define_const(mWc, "Svn_wc_notify_copy", SWIG_From_int((int)(svn_wc_notify_copy)));
  rb_define_const(mWc, "Svn_wc_notify_delete", SWIG_From_int((int)(svn_wc_notify_delete)));
  rb_define_const(mWc, "Svn_wc_notify_restore", SWIG_From_int((int)(svn_wc_notify_restore)));
  rb_define_const(mWc, "Svn_wc_notify_revert", SWIG_From_int((int)(svn_wc_notify_revert)));
  rb_define_const(mWc, "Svn_wc_notify_failed_revert", SWIG_From_int((int)(svn_wc_notify_failed_revert)));
  rb_define_const(mWc, "Svn_wc_notify_resolved", SWIG_From_int((int)(svn_wc_notify_resolved)));
  rb_define_const(mWc, "Svn_wc_notify_skip", SWIG_From_int((int)(svn_wc_notify_skip)));
  rb_define_const(mWc, "Svn_wc_notify_update_delete", SWIG_From_int((int)(svn_wc_notify_update_delete)));
  rb_define_const(mWc, "Svn_wc_notify_update_add", SWIG_From_int((int)(svn_wc_notify_update_add)));
  rb_define_const(mWc, "Svn_wc_notify_update_update", SWIG_From_int((int)(svn_wc_notify_update_update)));
  rb_define_const(mWc, "Svn_wc_notify_update_completed", SWIG_From_int((int)(svn_wc_notify_update_completed)));
  rb_define_const(mWc, "Svn_wc_notify_update_external", SWIG_From_int((int)(svn_wc_notify_update_external)));
  rb_define_const(mWc, "Svn_wc_notify_status_completed", SWIG_From_int((int)(svn_wc_notify_status_completed)));
  rb_define_const(mWc, "Svn_wc_notify_status_external", SWIG_From_int((int)(svn_wc_notify_status_external)));
  rb_define_const(mWc, "Svn_wc_notify_commit_modified", SWIG_From_int((int)(svn_wc_notify_commit_modified)));
  rb_define_const(mWc, "Svn_wc_notify_commit_added", SWIG_From_int((int)(svn_wc_notify_commit_added)));
  rb_define_const(mWc, "Svn_wc_notify_commit_deleted", SWIG_From_int((int)(svn_wc_notify_commit_deleted)));
  rb_define_const(mWc, "Svn_wc_notify_commit_replaced", SWIG_From_int((int)(svn_wc_notify_commit_replaced)));
  rb_define_const(mWc, "Svn_wc_notify_commit_postfix_txdelta", SWIG_From_int((int)(svn_wc_notify_commit_postfix_txdelta)));
  rb_define_const(mWc, "Svn_wc_notify_blame_revision", SWIG_From_int((int)(svn_wc_notify_blame_revision)));
  rb_define_const(mWc, "Svn_wc_notify_locked", SWIG_From_int((int)(svn_wc_notify_locked)));
  rb_define_const(mWc, "Svn_wc_notify_unlocked", SWIG_From_int((int)(svn_wc_notify_unlocked)));
  rb_define_const(mWc, "Svn_wc_notify_failed_lock", SWIG_From_int((int)(svn_wc_notify_failed_lock)));
  rb_define_const(mWc, "Svn_wc_notify_failed_unlock", SWIG_From_int((int)(svn_wc_notify_failed_unlock)));
  rb_define_const(mWc, "Svn_wc_notify_exists", SWIG_From_int((int)(svn_wc_notify_exists)));
  rb_define_const(mWc, "Svn_wc_notify_changelist_set", SWIG_From_int((int)(svn_wc_notify_changelist_set)));
  rb_define_const(mWc, "Svn_wc_notify_changelist_clear", SWIG_From_int((int)(svn_wc_notify_changelist_clear)));
  rb_define_const(mWc, "Svn_wc_notify_changelist_moved", SWIG_From_int((int)(svn_wc_notify_changelist_moved)));
  rb_define_const(mWc, "Svn_wc_notify_merge_begin", SWIG_From_int((int)(svn_wc_notify_merge_begin)));
  rb_define_const(mWc, "Svn_wc_notify_foreign_merge_begin", SWIG_From_int((int)(svn_wc_notify_foreign_merge_begin)));
  rb_define_const(mWc, "Svn_wc_notify_update_replace", SWIG_From_int((int)(svn_wc_notify_update_replace)));
  rb_define_const(mWc, "Svn_wc_notify_property_added", SWIG_From_int((int)(svn_wc_notify_property_added)));
  rb_define_const(mWc, "Svn_wc_notify_property_modified", SWIG_From_int((int)(svn_wc_notify_property_modified)));
  rb_define_const(mWc, "Svn_wc_notify_property_deleted", SWIG_From_int((int)(svn_wc_notify_property_deleted)));
  rb_define_const(mWc, "Svn_wc_notify_property_deleted_nonexistent", SWIG_From_int((int)(svn_wc_notify_property_deleted_nonexistent)));
  rb_define_const(mWc, "Svn_wc_notify_revprop_set", SWIG_From_int((int)(svn_wc_notify_revprop_set)));
  rb_define_const(mWc, "Svn_wc_notify_revprop_deleted", SWIG_From_int((int)(svn_wc_notify_revprop_deleted)));
  rb_define_const(mWc, "Svn_wc_notify_merge_completed", SWIG_From_int((int)(svn_wc_notify_merge_completed)));
  rb_define_const(mWc, "Svn_wc_notify_tree_conflict", SWIG_From_int((int)(svn_wc_notify_tree_conflict)));
  rb_define_const(mWc, "Svn_wc_notify_failed_external", SWIG_From_int((int)(svn_wc_notify_failed_external)));
  rb_define_const(mWc, "Svn_wc_notify_update_started", SWIG_From_int((int)(svn_wc_notify_update_started)));
  rb_define_const(mWc, "Svn_wc_notify_update_skip_obstruction", SWIG_From_int((int)(svn_wc_notify_update_skip_obstruction)));
  rb_define_const(mWc, "Svn_wc_notify_update_skip_working_only", SWIG_From_int((int)(svn_wc_notify_update_skip_working_only)));
  rb_define_const(mWc, "Svn_wc_notify_update_skip_access_denied", SWIG_From_int((int)(svn_wc_notify_update_skip_access_denied)));
  rb_define_const(mWc, "Svn_wc_notify_update_external_removed", SWIG_From_int((int)(svn_wc_notify_update_external_removed)));
  rb_define_const(mWc, "Svn_wc_notify_update_shadowed_add", SWIG_From_int((int)(svn_wc_notify_update_shadowed_add)));
  rb_define_const(mWc, "Svn_wc_notify_update_shadowed_update", SWIG_From_int((int)(svn_wc_notify_update_shadowed_update)));
  rb_define_const(mWc, "Svn_wc_notify_update_shadowed_delete", SWIG_From_int((int)(svn_wc_notify_update_shadowed_delete)));
  rb_define_const(mWc, "Svn_wc_notify_merge_record_info", SWIG_From_int((int)(svn_wc_notify_merge_record_info)));
  rb_define_const(mWc, "Svn_wc_notify_upgraded_path", SWIG_From_int((int)(svn_wc_notify_upgraded_path)));
  rb_define_const(mWc, "Svn_wc_notify_merge_record_info_begin", SWIG_From_int((int)(svn_wc_notify_merge_record_info_begin)));
  rb_define_const(mWc, "Svn_wc_notify_merge_elide_info", SWIG_From_int((int)(svn_wc_notify_merge_elide_info)));
  rb_define_const(mWc, "Svn_wc_notify_patch", SWIG_From_int((int)(svn_wc_notify_patch)));
  rb_define_const(mWc, "Svn_wc_notify_patch_applied_hunk", SWIG_From_int((int)(svn_wc_notify_patch_applied_hunk)));
  rb_define_const(mWc, "Svn_wc_notify_patch_rejected_hunk", SWIG_From_int((int)(svn_wc_notify_patch_rejected_hunk)));
  rb_define_const(mWc, "Svn_wc_notify_patch_hunk_already_applied", SWIG_From_int((int)(svn_wc_notify_patch_hunk_already_applied)));
  rb_define_const(mWc, "Svn_wc_notify_commit_copied", SWIG_From_int((int)(svn_wc_notify_commit_copied)));
  rb_define_const(mWc, "Svn_wc_notify_commit_copied_replaced", SWIG_From_int((int)(svn_wc_notify_commit_copied_replaced)));
  rb_define_const(mWc, "Svn_wc_notify_url_redirect", SWIG_From_int((int)(svn_wc_notify_url_redirect)));
  rb_define_const(mWc, "Svn_wc_notify_path_nonexistent", SWIG_From_int((int)(svn_wc_notify_path_nonexistent)));
  rb_define_const(mWc, "Svn_wc_notify_exclude", SWIG_From_int((int)(svn_wc_notify_exclude)));
  rb_define_const(mWc, "Svn_wc_notify_failed_conflict", SWIG_From_int((int)(svn_wc_notify_failed_conflict)));
  rb_define_const(mWc, "Svn_wc_notify_failed_missing", SWIG_From_int((int)(svn_wc_notify_failed_missing)));
  rb_define_const(mWc, "Svn_wc_notify_failed_out_of_date", SWIG_From_int((int)(svn_wc_notify_failed_out_of_date)));
  rb_define_const(mWc, "Svn_wc_notify_failed_no_parent", SWIG_From_int((int)(svn_wc_notify_failed_no_parent)));
  rb_define_const(mWc, "Svn_wc_notify_failed_locked", SWIG_From_int((int)(svn_wc_notify_failed_locked)));
  rb_define_const(mWc, "Svn_wc_notify_failed_forbidden_by_server", SWIG_From_int((int)(svn_wc_notify_failed_forbidden_by_server)));
  rb_define_const(mWc, "Svn_wc_notify_skip_conflicted", SWIG_From_int((int)(svn_wc_notify_skip_conflicted)));
  rb_define_const(mWc, "Svn_wc_notify_update_broken_lock", SWIG_From_int((int)(svn_wc_notify_update_broken_lock)));
  rb_define_const(mWc, "Svn_wc_notify_failed_obstruction", SWIG_From_int((int)(svn_wc_notify_failed_obstruction)));
  rb_define_const(mWc, "Svn_wc_notify_conflict_resolver_starting", SWIG_From_int((int)(svn_wc_notify_conflict_resolver_starting)));
  rb_define_const(mWc, "Svn_wc_notify_conflict_resolver_done", SWIG_From_int((int)(svn_wc_notify_conflict_resolver_done)));
  rb_define_const(mWc, "Svn_wc_notify_left_local_modifications", SWIG_From_int((int)(svn_wc_notify_left_local_modifications)));
  rb_define_const(mWc, "Svn_wc_notify_foreign_copy_begin", SWIG_From_int((int)(svn_wc_notify_foreign_copy_begin)));
  rb_define_const(mWc, "Svn_wc_notify_move_broken", SWIG_From_int((int)(svn_wc_notify_move_broken)));
  rb_define_const(mWc, "Svn_wc_notify_cleanup_external", SWIG_From_int((int)(svn_wc_notify_cleanup_external)));
  rb_define_const(mWc, "Svn_wc_notify_failed_requires_target", SWIG_From_int((int)(svn_wc_notify_failed_requires_target)));
  rb_define_const(mWc, "Svn_wc_notify_info_external", SWIG_From_int((int)(svn_wc_notify_info_external)));
  rb_define_const(mWc, "Svn_wc_notify_commit_finalizing", SWIG_From_int((int)(svn_wc_notify_commit_finalizing)));
  rb_define_const(mWc, "Svn_wc_notify_resolved_text", SWIG_From_int((int)(svn_wc_notify_resolved_text)));
  rb_define_const(mWc, "Svn_wc_notify_resolved_prop", SWIG_From_int((int)(svn_wc_notify_resolved_prop)));
  rb_define_const(mWc, "Svn_wc_notify_resolved_tree", SWIG_From_int((int)(svn_wc_notify_resolved_tree)));
  rb_define_const(mWc, "Svn_wc_notify_begin_search_tree_conflict_details", SWIG_From_int((int)(svn_wc_notify_begin_search_tree_conflict_details)));
  rb_define_const(mWc, "Svn_wc_notify_tree_conflict_details_progress", SWIG_From_int((int)(svn_wc_notify_tree_conflict_details_progress)));
  rb_define_const(mWc, "Svn_wc_notify_end_search_tree_conflict_details", SWIG_From_int((int)(svn_wc_notify_end_search_tree_conflict_details)));
  rb_define_const(mWc, "Svn_wc_notify_state_inapplicable", SWIG_From_int((int)(svn_wc_notify_state_inapplicable)));
  rb_define_const(mWc, "Svn_wc_notify_state_unknown", SWIG_From_int((int)(svn_wc_notify_state_unknown)));
  rb_define_const(mWc, "Svn_wc_notify_state_unchanged", SWIG_From_int((int)(svn_wc_notify_state_unchanged)));
  rb_define_const(mWc, "Svn_wc_notify_state_missing", SWIG_From_int((int)(svn_wc_notify_state_missing)));
  rb_define_const(mWc, "Svn_wc_notify_state_obstructed", SWIG_From_int((int)(svn_wc_notify_state_obstructed)));
  rb_define_const(mWc, "Svn_wc_notify_state_changed", SWIG_From_int((int)(svn_wc_notify_state_changed)));
  rb_define_const(mWc, "Svn_wc_notify_state_merged", SWIG_From_int((int)(svn_wc_notify_state_merged)));
  rb_define_const(mWc, "Svn_wc_notify_state_conflicted", SWIG_From_int((int)(svn_wc_notify_state_conflicted)));
  rb_define_const(mWc, "Svn_wc_notify_state_source_missing", SWIG_From_int((int)(svn_wc_notify_state_source_missing)));
  rb_define_const(mWc, "Svn_wc_notify_lock_state_inapplicable", SWIG_From_int((int)(svn_wc_notify_lock_state_inapplicable)));
  rb_define_const(mWc, "Svn_wc_notify_lock_state_unknown", SWIG_From_int((int)(svn_wc_notify_lock_state_unknown)));
  rb_define_const(mWc, "Svn_wc_notify_lock_state_unchanged", SWIG_From_int((int)(svn_wc_notify_lock_state_unchanged)));
  rb_define_const(mWc, "Svn_wc_notify_lock_state_locked", SWIG_From_int((int)(svn_wc_notify_lock_state_locked)));
  rb_define_const(mWc, "Svn_wc_notify_lock_state_unlocked", SWIG_From_int((int)(svn_wc_notify_lock_state_unlocked)));
  
  SwigClassSvn_wc_notify_t.klass = rb_define_class_under(mWc, "Svn_wc_notify_t", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_svn_wc_notify_t, (void *) &SwigClassSvn_wc_notify_t);
  rb_define_alloc_func(SwigClassSvn_wc_notify_t.klass, _wrap_svn_wc_notify_t_allocate);
  rb_define_method(SwigClassSvn_wc_notify_t.klass, "initialize", _wrap_new_svn_wc_notify_t, -1);
  rb_define_method(SwigClassSvn_wc_notify_t.klass, "path=", _wrap_svn_wc_notify_t_path_set, -1);
  rb_define_method(SwigClassSvn_wc_notify_t.klass, "path", _wrap_svn_wc_notify_t_path_get, -1);
  rb_define_method(SwigClassSvn_wc_notify_t.klass, "action=", _wrap_svn_wc_notify_t_action_set, -1);
  rb_define_method(SwigClassSvn_wc_notify_t.klass, "action", _wrap_svn_wc_notify_t_action_get, -1);
  rb_define_method(SwigClassSvn_wc_notify_t.klass, "kind=", _wrap_svn_wc_notify_t_kind_set, -1);
  rb_define_method(SwigClassSvn_wc_notify_t.klass, "kind", _wrap_svn_wc_notify_t_kind_get, -1);
  rb_define_method(SwigClassSvn_wc_notify_t.klass, "mime_type=", _wrap_svn_wc_notify_t_mime_type_set, -1);
  rb_define_method(SwigClassSvn_wc_notify_t.klass, "mime_type", _wrap_svn_wc_notify_t_mime_type_get, -1);
  rb_define_method(SwigClassSvn_wc_notify_t.klass, "lock=", _wrap_svn_wc_notify_t_lock_set, -1);
  rb_define_method(SwigClassSvn_wc_notify_t.klass, "lock", _wrap_svn_wc_notify_t_lock_get, -1);
  rb_define_method(SwigClassSvn_wc_notify_t.klass, "err=", _wrap_svn_wc_notify_t_err_set, -1);
  rb_define_method(SwigClassSvn_wc_notify_t.klass, "err", _wrap_svn_wc_notify_t_err_get, -1);
  rb_define_method(SwigClassSvn_wc_notify_t.klass, "content_state=", _wrap_svn_wc_notify_t_content_state_set, -1);
  rb_define_method(SwigClassSvn_wc_notify_t.klass, "content_state", _wrap_svn_wc_notify_t_content_state_get, -1);
  rb_define_method(SwigClassSvn_wc_notify_t.klass, "prop_state=", _wrap_svn_wc_notify_t_prop_state_set, -1);
  rb_define_method(SwigClassSvn_wc_notify_t.klass, "prop_state", _wrap_svn_wc_notify_t_prop_state_get, -1);
  rb_define_method(SwigClassSvn_wc_notify_t.klass, "lock_state=", _wrap_svn_wc_notify_t_lock_state_set, -1);
  rb_define_method(SwigClassSvn_wc_notify_t.klass, "lock_state", _wrap_svn_wc_notify_t_lock_state_get, -1);
  rb_define_method(SwigClassSvn_wc_notify_t.klass, "revision=", _wrap_svn_wc_notify_t_revision_set, -1);
  rb_define_method(SwigClassSvn_wc_notify_t.klass, "revision", _wrap_svn_wc_notify_t_revision_get, -1);
  rb_define_method(SwigClassSvn_wc_notify_t.klass, "changelist_name=", _wrap_svn_wc_notify_t_changelist_name_set, -1);
  rb_define_method(SwigClassSvn_wc_notify_t.klass, "changelist_name", _wrap_svn_wc_notify_t_changelist_name_get, -1);
  rb_define_method(SwigClassSvn_wc_notify_t.klass, "merge_range=", _wrap_svn_wc_notify_t_merge_range_set, -1);
  rb_define_method(SwigClassSvn_wc_notify_t.klass, "merge_range", _wrap_svn_wc_notify_t_merge_range_get, -1);
  rb_define_method(SwigClassSvn_wc_notify_t.klass, "url=", _wrap_svn_wc_notify_t_url_set, -1);
  rb_define_method(SwigClassSvn_wc_notify_t.klass, "url", _wrap_svn_wc_notify_t_url_get, -1);
  rb_define_method(SwigClassSvn_wc_notify_t.klass, "path_prefix=", _wrap_svn_wc_notify_t_path_prefix_set, -1);
  rb_define_method(SwigClassSvn_wc_notify_t.klass, "path_prefix", _wrap_svn_wc_notify_t_path_prefix_get, -1);
  rb_define_method(SwigClassSvn_wc_notify_t.klass, "prop_name=", _wrap_svn_wc_notify_t_prop_name_set, -1);
  rb_define_method(SwigClassSvn_wc_notify_t.klass, "prop_name", _wrap_svn_wc_notify_t_prop_name_get, -1);
  rb_define_method(SwigClassSvn_wc_notify_t.klass, "rev_props=", _wrap_svn_wc_notify_t_rev_props_set, -1);
  rb_define_method(SwigClassSvn_wc_notify_t.klass, "rev_props", _wrap_svn_wc_notify_t_rev_props_get, -1);
  rb_define_method(SwigClassSvn_wc_notify_t.klass, "old_revision=", _wrap_svn_wc_notify_t_old_revision_set, -1);
  rb_define_method(SwigClassSvn_wc_notify_t.klass, "old_revision", _wrap_svn_wc_notify_t_old_revision_get, -1);
  rb_define_method(SwigClassSvn_wc_notify_t.klass, "hunk_original_start=", _wrap_svn_wc_notify_t_hunk_original_start_set, -1);
  rb_define_method(SwigClassSvn_wc_notify_t.klass, "hunk_original_start", _wrap_svn_wc_notify_t_hunk_original_start_get, -1);
  rb_define_method(SwigClassSvn_wc_notify_t.klass, "hunk_original_length=", _wrap_svn_wc_notify_t_hunk_original_length_set, -1);
  rb_define_method(SwigClassSvn_wc_notify_t.klass, "hunk_original_length", _wrap_svn_wc_notify_t_hunk_original_length_get, -1);
  rb_define_method(SwigClassSvn_wc_notify_t.klass, "hunk_modified_start=", _wrap_svn_wc_notify_t_hunk_modified_start_set, -1);
  rb_define_method(SwigClassSvn_wc_notify_t.klass, "hunk_modified_start", _wrap_svn_wc_notify_t_hunk_modified_start_get, -1);
  rb_define_method(SwigClassSvn_wc_notify_t.klass, "hunk_modified_length=", _wrap_svn_wc_notify_t_hunk_modified_length_set, -1);
  rb_define_method(SwigClassSvn_wc_notify_t.klass, "hunk_modified_length", _wrap_svn_wc_notify_t_hunk_modified_length_get, -1);
  rb_define_method(SwigClassSvn_wc_notify_t.klass, "hunk_matched_line=", _wrap_svn_wc_notify_t_hunk_matched_line_set, -1);
  rb_define_method(SwigClassSvn_wc_notify_t.klass, "hunk_matched_line", _wrap_svn_wc_notify_t_hunk_matched_line_get, -1);
  rb_define_method(SwigClassSvn_wc_notify_t.klass, "hunk_fuzz=", _wrap_svn_wc_notify_t_hunk_fuzz_set, -1);
  rb_define_method(SwigClassSvn_wc_notify_t.klass, "hunk_fuzz", _wrap_svn_wc_notify_t_hunk_fuzz_get, -1);
  SwigClassSvn_wc_notify_t.mark = 0;
  SwigClassSvn_wc_notify_t.destroy = (void (*)(void *)) free_svn_wc_notify_t;
  SwigClassSvn_wc_notify_t.trackObjects = 0;
  rb_define_module_function(mWc, "svn_wc_create_notify", _wrap_svn_wc_create_notify, -1);
  rb_define_module_function(mWc, "svn_wc_create_notify_url", _wrap_svn_wc_create_notify_url, -1);
  rb_define_module_function(mWc, "svn_wc_dup_notify", _wrap_svn_wc_dup_notify, -1);
  rb_define_const(mWc, "Svn_wc_conflict_action_edit", SWIG_From_int((int)(svn_wc_conflict_action_edit)));
  rb_define_const(mWc, "Svn_wc_conflict_action_add", SWIG_From_int((int)(svn_wc_conflict_action_add)));
  rb_define_const(mWc, "Svn_wc_conflict_action_delete", SWIG_From_int((int)(svn_wc_conflict_action_delete)));
  rb_define_const(mWc, "Svn_wc_conflict_action_replace", SWIG_From_int((int)(svn_wc_conflict_action_replace)));
  rb_define_const(mWc, "Svn_wc_conflict_reason_edited", SWIG_From_int((int)(svn_wc_conflict_reason_edited)));
  rb_define_const(mWc, "Svn_wc_conflict_reason_obstructed", SWIG_From_int((int)(svn_wc_conflict_reason_obstructed)));
  rb_define_const(mWc, "Svn_wc_conflict_reason_deleted", SWIG_From_int((int)(svn_wc_conflict_reason_deleted)));
  rb_define_const(mWc, "Svn_wc_conflict_reason_missing", SWIG_From_int((int)(svn_wc_conflict_reason_missing)));
  rb_define_const(mWc, "Svn_wc_conflict_reason_unversioned", SWIG_From_int((int)(svn_wc_conflict_reason_unversioned)));
  rb_define_const(mWc, "Svn_wc_conflict_reason_added", SWIG_From_int((int)(svn_wc_conflict_reason_added)));
  rb_define_const(mWc, "Svn_wc_conflict_reason_replaced", SWIG_From_int((int)(svn_wc_conflict_reason_replaced)));
  rb_define_const(mWc, "Svn_wc_conflict_reason_moved_away", SWIG_From_int((int)(svn_wc_conflict_reason_moved_away)));
  rb_define_const(mWc, "Svn_wc_conflict_reason_moved_here", SWIG_From_int((int)(svn_wc_conflict_reason_moved_here)));
  rb_define_const(mWc, "Svn_wc_conflict_kind_text", SWIG_From_int((int)(svn_wc_conflict_kind_text)));
  rb_define_const(mWc, "Svn_wc_conflict_kind_property", SWIG_From_int((int)(svn_wc_conflict_kind_property)));
  rb_define_const(mWc, "Svn_wc_conflict_kind_tree", SWIG_From_int((int)(svn_wc_conflict_kind_tree)));
  rb_define_const(mWc, "Svn_wc_operation_none", SWIG_From_int((int)(svn_wc_operation_none)));
  rb_define_const(mWc, "Svn_wc_operation_update", SWIG_From_int((int)(svn_wc_operation_update)));
  rb_define_const(mWc, "Svn_wc_operation_switch", SWIG_From_int((int)(svn_wc_operation_switch)));
  rb_define_const(mWc, "Svn_wc_operation_merge", SWIG_From_int((int)(svn_wc_operation_merge)));
  
  SwigClassSvn_wc_conflict_version_t.klass = rb_define_class_under(mWc, "Svn_wc_conflict_version_t", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_svn_wc_conflict_version_t, (void *) &SwigClassSvn_wc_conflict_version_t);
  rb_define_alloc_func(SwigClassSvn_wc_conflict_version_t.klass, _wrap_svn_wc_conflict_version_t_allocate);
  rb_define_method(SwigClassSvn_wc_conflict_version_t.klass, "initialize", _wrap_new_svn_wc_conflict_version_t, -1);
  rb_define_method(SwigClassSvn_wc_conflict_version_t.klass, "repos_url=", _wrap_svn_wc_conflict_version_t_repos_url_set, -1);
  rb_define_method(SwigClassSvn_wc_conflict_version_t.klass, "repos_url", _wrap_svn_wc_conflict_version_t_repos_url_get, -1);
  rb_define_method(SwigClassSvn_wc_conflict_version_t.klass, "peg_rev=", _wrap_svn_wc_conflict_version_t_peg_rev_set, -1);
  rb_define_method(SwigClassSvn_wc_conflict_version_t.klass, "peg_rev", _wrap_svn_wc_conflict_version_t_peg_rev_get, -1);
  rb_define_method(SwigClassSvn_wc_conflict_version_t.klass, "path_in_repos=", _wrap_svn_wc_conflict_version_t_path_in_repos_set, -1);
  rb_define_method(SwigClassSvn_wc_conflict_version_t.klass, "path_in_repos", _wrap_svn_wc_conflict_version_t_path_in_repos_get, -1);
  rb_define_method(SwigClassSvn_wc_conflict_version_t.klass, "node_kind=", _wrap_svn_wc_conflict_version_t_node_kind_set, -1);
  rb_define_method(SwigClassSvn_wc_conflict_version_t.klass, "node_kind", _wrap_svn_wc_conflict_version_t_node_kind_get, -1);
  rb_define_method(SwigClassSvn_wc_conflict_version_t.klass, "repos_uuid=", _wrap_svn_wc_conflict_version_t_repos_uuid_set, -1);
  rb_define_method(SwigClassSvn_wc_conflict_version_t.klass, "repos_uuid", _wrap_svn_wc_conflict_version_t_repos_uuid_get, -1);
  SwigClassSvn_wc_conflict_version_t.mark = 0;
  SwigClassSvn_wc_conflict_version_t.destroy = (void (*)(void *)) free_svn_wc_conflict_version_t;
  SwigClassSvn_wc_conflict_version_t.trackObjects = 0;
  rb_define_module_function(mWc, "svn_wc_conflict_version_create2", _wrap_svn_wc_conflict_version_create2, -1);
  rb_define_module_function(mWc, "svn_wc_conflict_version_create", _wrap_svn_wc_conflict_version_create, -1);
  rb_define_module_function(mWc, "svn_wc_conflict_version_dup", _wrap_svn_wc_conflict_version_dup, -1);
  
  SwigClassSvn_wc_conflict_description_t.klass = rb_define_class_under(mWc, "Svn_wc_conflict_description_t", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_svn_wc_conflict_description_t, (void *) &SwigClassSvn_wc_conflict_description_t);
  rb_define_alloc_func(SwigClassSvn_wc_conflict_description_t.klass, _wrap_svn_wc_conflict_description_t_allocate);
  rb_define_method(SwigClassSvn_wc_conflict_description_t.klass, "initialize", _wrap_new_svn_wc_conflict_description_t, -1);
  rb_define_method(SwigClassSvn_wc_conflict_description_t.klass, "path=", _wrap_svn_wc_conflict_description_t_path_set, -1);
  rb_define_method(SwigClassSvn_wc_conflict_description_t.klass, "path", _wrap_svn_wc_conflict_description_t_path_get, -1);
  rb_define_method(SwigClassSvn_wc_conflict_description_t.klass, "node_kind=", _wrap_svn_wc_conflict_description_t_node_kind_set, -1);
  rb_define_method(SwigClassSvn_wc_conflict_description_t.klass, "node_kind", _wrap_svn_wc_conflict_description_t_node_kind_get, -1);
  rb_define_method(SwigClassSvn_wc_conflict_description_t.klass, "kind=", _wrap_svn_wc_conflict_description_t_kind_set, -1);
  rb_define_method(SwigClassSvn_wc_conflict_description_t.klass, "kind", _wrap_svn_wc_conflict_description_t_kind_get, -1);
  rb_define_method(SwigClassSvn_wc_conflict_description_t.klass, "property_name=", _wrap_svn_wc_conflict_description_t_property_name_set, -1);
  rb_define_method(SwigClassSvn_wc_conflict_description_t.klass, "property_name", _wrap_svn_wc_conflict_description_t_property_name_get, -1);
  rb_define_method(SwigClassSvn_wc_conflict_description_t.klass, "is_binary=", _wrap_svn_wc_conflict_description_t_is_binary_set, -1);
  rb_define_method(SwigClassSvn_wc_conflict_description_t.klass, "is_binary", _wrap_svn_wc_conflict_description_t_is_binary_get, -1);
  rb_define_method(SwigClassSvn_wc_conflict_description_t.klass, "mime_type=", _wrap_svn_wc_conflict_description_t_mime_type_set, -1);
  rb_define_method(SwigClassSvn_wc_conflict_description_t.klass, "mime_type", _wrap_svn_wc_conflict_description_t_mime_type_get, -1);
  rb_define_method(SwigClassSvn_wc_conflict_description_t.klass, "access=", _wrap_svn_wc_conflict_description_t_access_set, -1);
  rb_define_method(SwigClassSvn_wc_conflict_description_t.klass, "access", _wrap_svn_wc_conflict_description_t_access_get, -1);
  rb_define_method(SwigClassSvn_wc_conflict_description_t.klass, "action=", _wrap_svn_wc_conflict_description_t_action_set, -1);
  rb_define_method(SwigClassSvn_wc_conflict_description_t.klass, "action", _wrap_svn_wc_conflict_description_t_action_get, -1);
  rb_define_method(SwigClassSvn_wc_conflict_description_t.klass, "reason=", _wrap_svn_wc_conflict_description_t_reason_set, -1);
  rb_define_method(SwigClassSvn_wc_conflict_description_t.klass, "reason", _wrap_svn_wc_conflict_description_t_reason_get, -1);
  rb_define_method(SwigClassSvn_wc_conflict_description_t.klass, "base_file=", _wrap_svn_wc_conflict_description_t_base_file_set, -1);
  rb_define_method(SwigClassSvn_wc_conflict_description_t.klass, "base_file", _wrap_svn_wc_conflict_description_t_base_file_get, -1);
  rb_define_method(SwigClassSvn_wc_conflict_description_t.klass, "their_file=", _wrap_svn_wc_conflict_description_t_their_file_set, -1);
  rb_define_method(SwigClassSvn_wc_conflict_description_t.klass, "their_file", _wrap_svn_wc_conflict_description_t_their_file_get, -1);
  rb_define_method(SwigClassSvn_wc_conflict_description_t.klass, "my_file=", _wrap_svn_wc_conflict_description_t_my_file_set, -1);
  rb_define_method(SwigClassSvn_wc_conflict_description_t.klass, "my_file", _wrap_svn_wc_conflict_description_t_my_file_get, -1);
  rb_define_method(SwigClassSvn_wc_conflict_description_t.klass, "merged_file=", _wrap_svn_wc_conflict_description_t_merged_file_set, -1);
  rb_define_method(SwigClassSvn_wc_conflict_description_t.klass, "merged_file", _wrap_svn_wc_conflict_description_t_merged_file_get, -1);
  rb_define_method(SwigClassSvn_wc_conflict_description_t.klass, "operation=", _wrap_svn_wc_conflict_description_t_operation_set, -1);
  rb_define_method(SwigClassSvn_wc_conflict_description_t.klass, "operation", _wrap_svn_wc_conflict_description_t_operation_get, -1);
  rb_define_method(SwigClassSvn_wc_conflict_description_t.klass, "src_left_version=", _wrap_svn_wc_conflict_description_t_src_left_version_set, -1);
  rb_define_method(SwigClassSvn_wc_conflict_description_t.klass, "src_left_version", _wrap_svn_wc_conflict_description_t_src_left_version_get, -1);
  rb_define_method(SwigClassSvn_wc_conflict_description_t.klass, "src_right_version=", _wrap_svn_wc_conflict_description_t_src_right_version_set, -1);
  rb_define_method(SwigClassSvn_wc_conflict_description_t.klass, "src_right_version", _wrap_svn_wc_conflict_description_t_src_right_version_get, -1);
  SwigClassSvn_wc_conflict_description_t.mark = 0;
  SwigClassSvn_wc_conflict_description_t.destroy = (void (*)(void *)) free_svn_wc_conflict_description_t;
  SwigClassSvn_wc_conflict_description_t.trackObjects = 0;
  rb_define_module_function(mWc, "svn_wc_conflict_description_create_text2", _wrap_svn_wc_conflict_description_create_text2, -1);
  rb_define_module_function(mWc, "svn_wc_conflict_description_create_text", _wrap_svn_wc_conflict_description_create_text, -1);
  rb_define_module_function(mWc, "svn_wc_conflict_description_create_prop2", _wrap_svn_wc_conflict_description_create_prop2, -1);
  rb_define_module_function(mWc, "svn_wc_conflict_description_create_prop", _wrap_svn_wc_conflict_description_create_prop, -1);
  rb_define_module_function(mWc, "svn_wc_conflict_description_create_tree2", _wrap_svn_wc_conflict_description_create_tree2, -1);
  rb_define_module_function(mWc, "svn_wc_conflict_description_create_tree", _wrap_svn_wc_conflict_description_create_tree, -1);
  rb_define_module_function(mWc, "svn_wc_conflict_description2_dup", _wrap_svn_wc_conflict_description2_dup, -1);
  rb_define_module_function(mWc, "svn_wc__conflict_description2_dup", _wrap_svn_wc__conflict_description2_dup, -1);
  rb_define_const(mWc, "Svn_wc_conflict_choose_undefined", SWIG_From_int((int)(svn_wc_conflict_choose_undefined)));
  rb_define_const(mWc, "Svn_wc_conflict_choose_postpone", SWIG_From_int((int)(svn_wc_conflict_choose_postpone)));
  rb_define_const(mWc, "Svn_wc_conflict_choose_base", SWIG_From_int((int)(svn_wc_conflict_choose_base)));
  rb_define_const(mWc, "Svn_wc_conflict_choose_theirs_full", SWIG_From_int((int)(svn_wc_conflict_choose_theirs_full)));
  rb_define_const(mWc, "Svn_wc_conflict_choose_mine_full", SWIG_From_int((int)(svn_wc_conflict_choose_mine_full)));
  rb_define_const(mWc, "Svn_wc_conflict_choose_theirs_conflict", SWIG_From_int((int)(svn_wc_conflict_choose_theirs_conflict)));
  rb_define_const(mWc, "Svn_wc_conflict_choose_mine_conflict", SWIG_From_int((int)(svn_wc_conflict_choose_mine_conflict)));
  rb_define_const(mWc, "Svn_wc_conflict_choose_merged", SWIG_From_int((int)(svn_wc_conflict_choose_merged)));
  rb_define_const(mWc, "Svn_wc_conflict_choose_unspecified", SWIG_From_int((int)(svn_wc_conflict_choose_unspecified)));
  rb_define_module_function(mWc, "svn_wc_create_conflict_result", _wrap_svn_wc_create_conflict_result, -1);
  
  SwigClassSvn_wc_diff_callbacks4_t.klass = rb_define_class_under(mWc, "Svn_wc_diff_callbacks4_t", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_svn_wc_diff_callbacks4_t, (void *) &SwigClassSvn_wc_diff_callbacks4_t);
  rb_define_alloc_func(SwigClassSvn_wc_diff_callbacks4_t.klass, _wrap_svn_wc_diff_callbacks4_t_allocate);
  rb_define_method(SwigClassSvn_wc_diff_callbacks4_t.klass, "initialize", _wrap_new_svn_wc_diff_callbacks4_t, -1);
  rb_define_method(SwigClassSvn_wc_diff_callbacks4_t.klass, "file_opened=", _wrap_svn_wc_diff_callbacks4_t_file_opened_set, -1);
  rb_define_method(SwigClassSvn_wc_diff_callbacks4_t.klass, "file_opened", _wrap_svn_wc_diff_callbacks4_t_file_opened_get, -1);
  rb_define_method(SwigClassSvn_wc_diff_callbacks4_t.klass, "file_changed=", _wrap_svn_wc_diff_callbacks4_t_file_changed_set, -1);
  rb_define_method(SwigClassSvn_wc_diff_callbacks4_t.klass, "file_changed", _wrap_svn_wc_diff_callbacks4_t_file_changed_get, -1);
  rb_define_method(SwigClassSvn_wc_diff_callbacks4_t.klass, "file_added=", _wrap_svn_wc_diff_callbacks4_t_file_added_set, -1);
  rb_define_method(SwigClassSvn_wc_diff_callbacks4_t.klass, "file_added", _wrap_svn_wc_diff_callbacks4_t_file_added_get, -1);
  rb_define_method(SwigClassSvn_wc_diff_callbacks4_t.klass, "file_deleted=", _wrap_svn_wc_diff_callbacks4_t_file_deleted_set, -1);
  rb_define_method(SwigClassSvn_wc_diff_callbacks4_t.klass, "file_deleted", _wrap_svn_wc_diff_callbacks4_t_file_deleted_get, -1);
  rb_define_method(SwigClassSvn_wc_diff_callbacks4_t.klass, "dir_deleted=", _wrap_svn_wc_diff_callbacks4_t_dir_deleted_set, -1);
  rb_define_method(SwigClassSvn_wc_diff_callbacks4_t.klass, "dir_deleted", _wrap_svn_wc_diff_callbacks4_t_dir_deleted_get, -1);
  rb_define_method(SwigClassSvn_wc_diff_callbacks4_t.klass, "dir_opened=", _wrap_svn_wc_diff_callbacks4_t_dir_opened_set, -1);
  rb_define_method(SwigClassSvn_wc_diff_callbacks4_t.klass, "dir_opened", _wrap_svn_wc_diff_callbacks4_t_dir_opened_get, -1);
  rb_define_method(SwigClassSvn_wc_diff_callbacks4_t.klass, "dir_added=", _wrap_svn_wc_diff_callbacks4_t_dir_added_set, -1);
  rb_define_method(SwigClassSvn_wc_diff_callbacks4_t.klass, "dir_added", _wrap_svn_wc_diff_callbacks4_t_dir_added_get, -1);
  rb_define_method(SwigClassSvn_wc_diff_callbacks4_t.klass, "dir_props_changed=", _wrap_svn_wc_diff_callbacks4_t_dir_props_changed_set, -1);
  rb_define_method(SwigClassSvn_wc_diff_callbacks4_t.klass, "dir_props_changed", _wrap_svn_wc_diff_callbacks4_t_dir_props_changed_get, -1);
  rb_define_method(SwigClassSvn_wc_diff_callbacks4_t.klass, "dir_closed=", _wrap_svn_wc_diff_callbacks4_t_dir_closed_set, -1);
  rb_define_method(SwigClassSvn_wc_diff_callbacks4_t.klass, "dir_closed", _wrap_svn_wc_diff_callbacks4_t_dir_closed_get, -1);
  SwigClassSvn_wc_diff_callbacks4_t.mark = 0;
  SwigClassSvn_wc_diff_callbacks4_t.destroy = (void (*)(void *)) free_svn_wc_diff_callbacks4_t;
  SwigClassSvn_wc_diff_callbacks4_t.trackObjects = 0;
  
  SwigClassSvn_wc_diff_callbacks3_t.klass = rb_define_class_under(mWc, "Svn_wc_diff_callbacks3_t", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_svn_wc_diff_callbacks3_t, (void *) &SwigClassSvn_wc_diff_callbacks3_t);
  rb_define_alloc_func(SwigClassSvn_wc_diff_callbacks3_t.klass, _wrap_svn_wc_diff_callbacks3_t_allocate);
  rb_define_method(SwigClassSvn_wc_diff_callbacks3_t.klass, "initialize", _wrap_new_svn_wc_diff_callbacks3_t, -1);
  rb_define_method(SwigClassSvn_wc_diff_callbacks3_t.klass, "file_changed=", _wrap_svn_wc_diff_callbacks3_t_file_changed_set, -1);
  rb_define_method(SwigClassSvn_wc_diff_callbacks3_t.klass, "file_changed", _wrap_svn_wc_diff_callbacks3_t_file_changed_get, -1);
  rb_define_method(SwigClassSvn_wc_diff_callbacks3_t.klass, "file_added=", _wrap_svn_wc_diff_callbacks3_t_file_added_set, -1);
  rb_define_method(SwigClassSvn_wc_diff_callbacks3_t.klass, "file_added", _wrap_svn_wc_diff_callbacks3_t_file_added_get, -1);
  rb_define_method(SwigClassSvn_wc_diff_callbacks3_t.klass, "file_deleted=", _wrap_svn_wc_diff_callbacks3_t_file_deleted_set, -1);
  rb_define_method(SwigClassSvn_wc_diff_callbacks3_t.klass, "file_deleted", _wrap_svn_wc_diff_callbacks3_t_file_deleted_get, -1);
  rb_define_method(SwigClassSvn_wc_diff_callbacks3_t.klass, "dir_added=", _wrap_svn_wc_diff_callbacks3_t_dir_added_set, -1);
  rb_define_method(SwigClassSvn_wc_diff_callbacks3_t.klass, "dir_added", _wrap_svn_wc_diff_callbacks3_t_dir_added_get, -1);
  rb_define_method(SwigClassSvn_wc_diff_callbacks3_t.klass, "dir_deleted=", _wrap_svn_wc_diff_callbacks3_t_dir_deleted_set, -1);
  rb_define_method(SwigClassSvn_wc_diff_callbacks3_t.klass, "dir_deleted", _wrap_svn_wc_diff_callbacks3_t_dir_deleted_get, -1);
  rb_define_method(SwigClassSvn_wc_diff_callbacks3_t.klass, "dir_props_changed=", _wrap_svn_wc_diff_callbacks3_t_dir_props_changed_set, -1);
  rb_define_method(SwigClassSvn_wc_diff_callbacks3_t.klass, "dir_props_changed", _wrap_svn_wc_diff_callbacks3_t_dir_props_changed_get, -1);
  rb_define_method(SwigClassSvn_wc_diff_callbacks3_t.klass, "dir_opened=", _wrap_svn_wc_diff_callbacks3_t_dir_opened_set, -1);
  rb_define_method(SwigClassSvn_wc_diff_callbacks3_t.klass, "dir_opened", _wrap_svn_wc_diff_callbacks3_t_dir_opened_get, -1);
  rb_define_method(SwigClassSvn_wc_diff_callbacks3_t.klass, "dir_closed=", _wrap_svn_wc_diff_callbacks3_t_dir_closed_set, -1);
  rb_define_method(SwigClassSvn_wc_diff_callbacks3_t.klass, "dir_closed", _wrap_svn_wc_diff_callbacks3_t_dir_closed_get, -1);
  SwigClassSvn_wc_diff_callbacks3_t.mark = 0;
  SwigClassSvn_wc_diff_callbacks3_t.destroy = (void (*)(void *)) free_svn_wc_diff_callbacks3_t;
  SwigClassSvn_wc_diff_callbacks3_t.trackObjects = 0;
  
  SwigClassSvn_wc_diff_callbacks2_t.klass = rb_define_class_under(mWc, "Svn_wc_diff_callbacks2_t", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_svn_wc_diff_callbacks2_t, (void *) &SwigClassSvn_wc_diff_callbacks2_t);
  rb_define_alloc_func(SwigClassSvn_wc_diff_callbacks2_t.klass, _wrap_svn_wc_diff_callbacks2_t_allocate);
  rb_define_method(SwigClassSvn_wc_diff_callbacks2_t.klass, "initialize", _wrap_new_svn_wc_diff_callbacks2_t, -1);
  rb_define_method(SwigClassSvn_wc_diff_callbacks2_t.klass, "file_changed=", _wrap_svn_wc_diff_callbacks2_t_file_changed_set, -1);
  rb_define_method(SwigClassSvn_wc_diff_callbacks2_t.klass, "file_changed", _wrap_svn_wc_diff_callbacks2_t_file_changed_get, -1);
  rb_define_method(SwigClassSvn_wc_diff_callbacks2_t.klass, "file_added=", _wrap_svn_wc_diff_callbacks2_t_file_added_set, -1);
  rb_define_method(SwigClassSvn_wc_diff_callbacks2_t.klass, "file_added", _wrap_svn_wc_diff_callbacks2_t_file_added_get, -1);
  rb_define_method(SwigClassSvn_wc_diff_callbacks2_t.klass, "file_deleted=", _wrap_svn_wc_diff_callbacks2_t_file_deleted_set, -1);
  rb_define_method(SwigClassSvn_wc_diff_callbacks2_t.klass, "file_deleted", _wrap_svn_wc_diff_callbacks2_t_file_deleted_get, -1);
  rb_define_method(SwigClassSvn_wc_diff_callbacks2_t.klass, "dir_added=", _wrap_svn_wc_diff_callbacks2_t_dir_added_set, -1);
  rb_define_method(SwigClassSvn_wc_diff_callbacks2_t.klass, "dir_added", _wrap_svn_wc_diff_callbacks2_t_dir_added_get, -1);
  rb_define_method(SwigClassSvn_wc_diff_callbacks2_t.klass, "dir_deleted=", _wrap_svn_wc_diff_callbacks2_t_dir_deleted_set, -1);
  rb_define_method(SwigClassSvn_wc_diff_callbacks2_t.klass, "dir_deleted", _wrap_svn_wc_diff_callbacks2_t_dir_deleted_get, -1);
  rb_define_method(SwigClassSvn_wc_diff_callbacks2_t.klass, "dir_props_changed=", _wrap_svn_wc_diff_callbacks2_t_dir_props_changed_set, -1);
  rb_define_method(SwigClassSvn_wc_diff_callbacks2_t.klass, "dir_props_changed", _wrap_svn_wc_diff_callbacks2_t_dir_props_changed_get, -1);
  SwigClassSvn_wc_diff_callbacks2_t.mark = 0;
  SwigClassSvn_wc_diff_callbacks2_t.destroy = (void (*)(void *)) free_svn_wc_diff_callbacks2_t;
  SwigClassSvn_wc_diff_callbacks2_t.trackObjects = 0;
  
  SwigClassSvn_wc_diff_callbacks_t.klass = rb_define_class_under(mWc, "Svn_wc_diff_callbacks_t", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_svn_wc_diff_callbacks_t, (void *) &SwigClassSvn_wc_diff_callbacks_t);
  rb_define_alloc_func(SwigClassSvn_wc_diff_callbacks_t.klass, _wrap_svn_wc_diff_callbacks_t_allocate);
  rb_define_method(SwigClassSvn_wc_diff_callbacks_t.klass, "initialize", _wrap_new_svn_wc_diff_callbacks_t, -1);
  rb_define_method(SwigClassSvn_wc_diff_callbacks_t.klass, "file_changed=", _wrap_svn_wc_diff_callbacks_t_file_changed_set, -1);
  rb_define_method(SwigClassSvn_wc_diff_callbacks_t.klass, "file_changed", _wrap_svn_wc_diff_callbacks_t_file_changed_get, -1);
  rb_define_method(SwigClassSvn_wc_diff_callbacks_t.klass, "file_added=", _wrap_svn_wc_diff_callbacks_t_file_added_set, -1);
  rb_define_method(SwigClassSvn_wc_diff_callbacks_t.klass, "file_added", _wrap_svn_wc_diff_callbacks_t_file_added_get, -1);
  rb_define_method(SwigClassSvn_wc_diff_callbacks_t.klass, "file_deleted=", _wrap_svn_wc_diff_callbacks_t_file_deleted_set, -1);
  rb_define_method(SwigClassSvn_wc_diff_callbacks_t.klass, "file_deleted", _wrap_svn_wc_diff_callbacks_t_file_deleted_get, -1);
  rb_define_method(SwigClassSvn_wc_diff_callbacks_t.klass, "dir_added=", _wrap_svn_wc_diff_callbacks_t_dir_added_set, -1);
  rb_define_method(SwigClassSvn_wc_diff_callbacks_t.klass, "dir_added", _wrap_svn_wc_diff_callbacks_t_dir_added_get, -1);
  rb_define_method(SwigClassSvn_wc_diff_callbacks_t.klass, "dir_deleted=", _wrap_svn_wc_diff_callbacks_t_dir_deleted_set, -1);
  rb_define_method(SwigClassSvn_wc_diff_callbacks_t.klass, "dir_deleted", _wrap_svn_wc_diff_callbacks_t_dir_deleted_get, -1);
  rb_define_method(SwigClassSvn_wc_diff_callbacks_t.klass, "props_changed=", _wrap_svn_wc_diff_callbacks_t_props_changed_set, -1);
  rb_define_method(SwigClassSvn_wc_diff_callbacks_t.klass, "props_changed", _wrap_svn_wc_diff_callbacks_t_props_changed_get, -1);
  SwigClassSvn_wc_diff_callbacks_t.mark = 0;
  SwigClassSvn_wc_diff_callbacks_t.destroy = (void (*)(void *)) free_svn_wc_diff_callbacks_t;
  SwigClassSvn_wc_diff_callbacks_t.trackObjects = 0;
  rb_define_module_function(mWc, "svn_wc_check_wc2", _wrap_svn_wc_check_wc2, -1);
  rb_define_module_function(mWc, "svn_wc_check_wc", _wrap_svn_wc_check_wc, -1);
  rb_define_module_function(mWc, "svn_wc_has_binary_prop", _wrap_svn_wc_has_binary_prop, -1);
  rb_define_module_function(mWc, "svn_wc_text_modified_p2", _wrap_svn_wc_text_modified_p2, -1);
  rb_define_module_function(mWc, "svn_wc_text_modified_p", _wrap_svn_wc_text_modified_p, -1);
  rb_define_module_function(mWc, "svn_wc_props_modified_p2", _wrap_svn_wc_props_modified_p2, -1);
  rb_define_module_function(mWc, "svn_wc_props_modified_p", _wrap_svn_wc_props_modified_p, -1);
  rb_define_const(mWc, "Svn_wc_schedule_normal", SWIG_From_int((int)(svn_wc_schedule_normal)));
  rb_define_const(mWc, "Svn_wc_schedule_add", SWIG_From_int((int)(svn_wc_schedule_add)));
  rb_define_const(mWc, "Svn_wc_schedule_delete", SWIG_From_int((int)(svn_wc_schedule_delete)));
  rb_define_const(mWc, "Svn_wc_schedule_replace", SWIG_From_int((int)(svn_wc_schedule_replace)));
  rb_define_const(mWc, "SVN_WC_ENTRY_WORKING_SIZE_UNKNOWN", SWIG_From_int((int)((-1))));
  
  SwigClassSvn_wc_entry_t.klass = rb_define_class_under(mWc, "Svn_wc_entry_t", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_svn_wc_entry_t, (void *) &SwigClassSvn_wc_entry_t);
  rb_define_alloc_func(SwigClassSvn_wc_entry_t.klass, _wrap_svn_wc_entry_t_allocate);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "initialize", _wrap_new_svn_wc_entry_t, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "name=", _wrap_svn_wc_entry_t_name_set, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "name", _wrap_svn_wc_entry_t_name_get, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "revision=", _wrap_svn_wc_entry_t_revision_set, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "revision", _wrap_svn_wc_entry_t_revision_get, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "url=", _wrap_svn_wc_entry_t_url_set, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "url", _wrap_svn_wc_entry_t_url_get, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "repos=", _wrap_svn_wc_entry_t_repos_set, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "repos", _wrap_svn_wc_entry_t_repos_get, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "uuid=", _wrap_svn_wc_entry_t_uuid_set, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "uuid", _wrap_svn_wc_entry_t_uuid_get, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "kind=", _wrap_svn_wc_entry_t_kind_set, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "kind", _wrap_svn_wc_entry_t_kind_get, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "schedule=", _wrap_svn_wc_entry_t_schedule_set, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "schedule", _wrap_svn_wc_entry_t_schedule_get, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "copied=", _wrap_svn_wc_entry_t_copied_set, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "copied", _wrap_svn_wc_entry_t_copied_get, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "deleted=", _wrap_svn_wc_entry_t_deleted_set, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "deleted", _wrap_svn_wc_entry_t_deleted_get, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "absent=", _wrap_svn_wc_entry_t_absent_set, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "absent", _wrap_svn_wc_entry_t_absent_get, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "incomplete=", _wrap_svn_wc_entry_t_incomplete_set, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "incomplete", _wrap_svn_wc_entry_t_incomplete_get, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "copyfrom_url=", _wrap_svn_wc_entry_t_copyfrom_url_set, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "copyfrom_url", _wrap_svn_wc_entry_t_copyfrom_url_get, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "copyfrom_rev=", _wrap_svn_wc_entry_t_copyfrom_rev_set, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "copyfrom_rev", _wrap_svn_wc_entry_t_copyfrom_rev_get, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "conflict_old=", _wrap_svn_wc_entry_t_conflict_old_set, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "conflict_old", _wrap_svn_wc_entry_t_conflict_old_get, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "conflict_new=", _wrap_svn_wc_entry_t_conflict_new_set, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "conflict_new", _wrap_svn_wc_entry_t_conflict_new_get, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "conflict_wrk=", _wrap_svn_wc_entry_t_conflict_wrk_set, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "conflict_wrk", _wrap_svn_wc_entry_t_conflict_wrk_get, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "prejfile=", _wrap_svn_wc_entry_t_prejfile_set, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "prejfile", _wrap_svn_wc_entry_t_prejfile_get, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "text_time=", _wrap_svn_wc_entry_t_text_time_set, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "text_time", _wrap_svn_wc_entry_t_text_time_get, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "prop_time=", _wrap_svn_wc_entry_t_prop_time_set, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "prop_time", _wrap_svn_wc_entry_t_prop_time_get, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "checksum=", _wrap_svn_wc_entry_t_checksum_set, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "checksum", _wrap_svn_wc_entry_t_checksum_get, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "cmt_rev=", _wrap_svn_wc_entry_t_cmt_rev_set, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "cmt_rev", _wrap_svn_wc_entry_t_cmt_rev_get, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "cmt_date=", _wrap_svn_wc_entry_t_cmt_date_set, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "cmt_date", _wrap_svn_wc_entry_t_cmt_date_get, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "cmt_author=", _wrap_svn_wc_entry_t_cmt_author_set, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "cmt_author", _wrap_svn_wc_entry_t_cmt_author_get, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "lock_token=", _wrap_svn_wc_entry_t_lock_token_set, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "lock_token", _wrap_svn_wc_entry_t_lock_token_get, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "lock_owner=", _wrap_svn_wc_entry_t_lock_owner_set, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "lock_owner", _wrap_svn_wc_entry_t_lock_owner_get, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "lock_comment=", _wrap_svn_wc_entry_t_lock_comment_set, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "lock_comment", _wrap_svn_wc_entry_t_lock_comment_get, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "lock_creation_date=", _wrap_svn_wc_entry_t_lock_creation_date_set, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "lock_creation_date", _wrap_svn_wc_entry_t_lock_creation_date_get, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "has_props=", _wrap_svn_wc_entry_t_has_props_set, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "has_props", _wrap_svn_wc_entry_t_has_props_get, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "has_prop_mods=", _wrap_svn_wc_entry_t_has_prop_mods_set, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "has_prop_mods", _wrap_svn_wc_entry_t_has_prop_mods_get, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "cachable_props=", _wrap_svn_wc_entry_t_cachable_props_set, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "cachable_props", _wrap_svn_wc_entry_t_cachable_props_get, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "present_props=", _wrap_svn_wc_entry_t_present_props_set, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "present_props", _wrap_svn_wc_entry_t_present_props_get, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "changelist=", _wrap_svn_wc_entry_t_changelist_set, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "changelist", _wrap_svn_wc_entry_t_changelist_get, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "working_size=", _wrap_svn_wc_entry_t_working_size_set, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "working_size", _wrap_svn_wc_entry_t_working_size_get, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "keep_local=", _wrap_svn_wc_entry_t_keep_local_set, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "keep_local", _wrap_svn_wc_entry_t_keep_local_get, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "depth=", _wrap_svn_wc_entry_t_depth_set, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "depth", _wrap_svn_wc_entry_t_depth_get, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "tree_conflict_data=", _wrap_svn_wc_entry_t_tree_conflict_data_set, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "tree_conflict_data", _wrap_svn_wc_entry_t_tree_conflict_data_get, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "file_external_path=", _wrap_svn_wc_entry_t_file_external_path_set, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "file_external_path", _wrap_svn_wc_entry_t_file_external_path_get, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "file_external_peg_rev=", _wrap_svn_wc_entry_t_file_external_peg_rev_set, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "file_external_peg_rev", _wrap_svn_wc_entry_t_file_external_peg_rev_get, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "file_external_rev=", _wrap_svn_wc_entry_t_file_external_rev_set, -1);
  rb_define_method(SwigClassSvn_wc_entry_t.klass, "file_external_rev", _wrap_svn_wc_entry_t_file_external_rev_get, -1);
  SwigClassSvn_wc_entry_t.mark = 0;
  SwigClassSvn_wc_entry_t.destroy = (void (*)(void *)) free_svn_wc_entry_t;
  SwigClassSvn_wc_entry_t.trackObjects = 0;
  rb_define_const(mWc, "SVN_WC_ENTRY_THIS_DIR", SWIG_FromCharPtr(""));
  rb_define_module_function(mWc, "svn_wc_entries_read", _wrap_svn_wc_entries_read, -1);
  rb_define_module_function(mWc, "svn_wc_entry_dup", _wrap_svn_wc_entry_dup, -1);
  
  SwigClassSvn_wc_info_t.klass = rb_define_class_under(mWc, "Svn_wc_info_t", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_svn_wc_info_t, (void *) &SwigClassSvn_wc_info_t);
  rb_define_alloc_func(SwigClassSvn_wc_info_t.klass, _wrap_svn_wc_info_t_allocate);
  rb_define_method(SwigClassSvn_wc_info_t.klass, "initialize", _wrap_new_svn_wc_info_t, -1);
  rb_define_method(SwigClassSvn_wc_info_t.klass, "schedule=", _wrap_svn_wc_info_t_schedule_set, -1);
  rb_define_method(SwigClassSvn_wc_info_t.klass, "schedule", _wrap_svn_wc_info_t_schedule_get, -1);
  rb_define_method(SwigClassSvn_wc_info_t.klass, "copyfrom_url=", _wrap_svn_wc_info_t_copyfrom_url_set, -1);
  rb_define_method(SwigClassSvn_wc_info_t.klass, "copyfrom_url", _wrap_svn_wc_info_t_copyfrom_url_get, -1);
  rb_define_method(SwigClassSvn_wc_info_t.klass, "copyfrom_rev=", _wrap_svn_wc_info_t_copyfrom_rev_set, -1);
  rb_define_method(SwigClassSvn_wc_info_t.klass, "copyfrom_rev", _wrap_svn_wc_info_t_copyfrom_rev_get, -1);
  rb_define_method(SwigClassSvn_wc_info_t.klass, "checksum=", _wrap_svn_wc_info_t_checksum_set, -1);
  rb_define_method(SwigClassSvn_wc_info_t.klass, "checksum", _wrap_svn_wc_info_t_checksum_get, -1);
  rb_define_method(SwigClassSvn_wc_info_t.klass, "changelist=", _wrap_svn_wc_info_t_changelist_set, -1);
  rb_define_method(SwigClassSvn_wc_info_t.klass, "changelist", _wrap_svn_wc_info_t_changelist_get, -1);
  rb_define_method(SwigClassSvn_wc_info_t.klass, "depth=", _wrap_svn_wc_info_t_depth_set, -1);
  rb_define_method(SwigClassSvn_wc_info_t.klass, "depth", _wrap_svn_wc_info_t_depth_get, -1);
  rb_define_method(SwigClassSvn_wc_info_t.klass, "recorded_size=", _wrap_svn_wc_info_t_recorded_size_set, -1);
  rb_define_method(SwigClassSvn_wc_info_t.klass, "recorded_size", _wrap_svn_wc_info_t_recorded_size_get, -1);
  rb_define_method(SwigClassSvn_wc_info_t.klass, "recorded_time=", _wrap_svn_wc_info_t_recorded_time_set, -1);
  rb_define_method(SwigClassSvn_wc_info_t.klass, "recorded_time", _wrap_svn_wc_info_t_recorded_time_get, -1);
  rb_define_method(SwigClassSvn_wc_info_t.klass, "conflicts=", _wrap_svn_wc_info_t_conflicts_set, -1);
  rb_define_method(SwigClassSvn_wc_info_t.klass, "conflicts", _wrap_svn_wc_info_t_conflicts_get, -1);
  rb_define_method(SwigClassSvn_wc_info_t.klass, "wcroot_abspath=", _wrap_svn_wc_info_t_wcroot_abspath_set, -1);
  rb_define_method(SwigClassSvn_wc_info_t.klass, "wcroot_abspath", _wrap_svn_wc_info_t_wcroot_abspath_get, -1);
  rb_define_method(SwigClassSvn_wc_info_t.klass, "moved_from_abspath=", _wrap_svn_wc_info_t_moved_from_abspath_set, -1);
  rb_define_method(SwigClassSvn_wc_info_t.klass, "moved_from_abspath", _wrap_svn_wc_info_t_moved_from_abspath_get, -1);
  rb_define_method(SwigClassSvn_wc_info_t.klass, "moved_to_abspath=", _wrap_svn_wc_info_t_moved_to_abspath_set, -1);
  rb_define_method(SwigClassSvn_wc_info_t.klass, "moved_to_abspath", _wrap_svn_wc_info_t_moved_to_abspath_get, -1);
  SwigClassSvn_wc_info_t.mark = 0;
  SwigClassSvn_wc_info_t.destroy = (void (*)(void *)) free_svn_wc_info_t;
  SwigClassSvn_wc_info_t.trackObjects = 0;
  rb_define_module_function(mWc, "svn_wc_info_dup", _wrap_svn_wc_info_dup, -1);
  rb_define_module_function(mWc, "svn_wc_conflicted_p3", _wrap_svn_wc_conflicted_p3, -1);
  rb_define_module_function(mWc, "svn_wc_conflicted_p2", _wrap_svn_wc_conflicted_p2, -1);
  rb_define_module_function(mWc, "svn_wc_conflicted_p", _wrap_svn_wc_conflicted_p, -1);
  rb_define_module_function(mWc, "svn_wc_get_ancestry", _wrap_svn_wc_get_ancestry, -1);
  
  SwigClassSvn_wc_entry_callbacks2_t.klass = rb_define_class_under(mWc, "Svn_wc_entry_callbacks2_t", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_svn_wc_entry_callbacks2_t, (void *) &SwigClassSvn_wc_entry_callbacks2_t);
  rb_define_alloc_func(SwigClassSvn_wc_entry_callbacks2_t.klass, _wrap_svn_wc_entry_callbacks2_t_allocate);
  rb_define_method(SwigClassSvn_wc_entry_callbacks2_t.klass, "initialize", _wrap_new_svn_wc_entry_callbacks2_t, -1);
  rb_define_method(SwigClassSvn_wc_entry_callbacks2_t.klass, "found_entry=", _wrap_svn_wc_entry_callbacks2_t_found_entry_set, -1);
  rb_define_method(SwigClassSvn_wc_entry_callbacks2_t.klass, "found_entry", _wrap_svn_wc_entry_callbacks2_t_found_entry_get, -1);
  rb_define_method(SwigClassSvn_wc_entry_callbacks2_t.klass, "handle_error=", _wrap_svn_wc_entry_callbacks2_t_handle_error_set, -1);
  rb_define_method(SwigClassSvn_wc_entry_callbacks2_t.klass, "handle_error", _wrap_svn_wc_entry_callbacks2_t_handle_error_get, -1);
  SwigClassSvn_wc_entry_callbacks2_t.mark = 0;
  SwigClassSvn_wc_entry_callbacks2_t.destroy = (void (*)(void *)) free_svn_wc_entry_callbacks2_t;
  SwigClassSvn_wc_entry_callbacks2_t.trackObjects = 0;
  
  SwigClassSvn_wc_entry_callbacks_t.klass = rb_define_class_under(mWc, "Svn_wc_entry_callbacks_t", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_svn_wc_entry_callbacks_t, (void *) &SwigClassSvn_wc_entry_callbacks_t);
  rb_define_alloc_func(SwigClassSvn_wc_entry_callbacks_t.klass, _wrap_svn_wc_entry_callbacks_t_allocate);
  rb_define_method(SwigClassSvn_wc_entry_callbacks_t.klass, "initialize", _wrap_new_svn_wc_entry_callbacks_t, -1);
  rb_define_method(SwigClassSvn_wc_entry_callbacks_t.klass, "found_entry=", _wrap_svn_wc_entry_callbacks_t_found_entry_set, -1);
  rb_define_method(SwigClassSvn_wc_entry_callbacks_t.klass, "found_entry", _wrap_svn_wc_entry_callbacks_t_found_entry_get, -1);
  SwigClassSvn_wc_entry_callbacks_t.mark = 0;
  SwigClassSvn_wc_entry_callbacks_t.destroy = (void (*)(void *)) free_svn_wc_entry_callbacks_t;
  SwigClassSvn_wc_entry_callbacks_t.trackObjects = 0;
  rb_define_module_function(mWc, "svn_wc_walk_entries3", _wrap_svn_wc_walk_entries3, -1);
  rb_define_module_function(mWc, "svn_wc_walk_entries2", _wrap_svn_wc_walk_entries2, -1);
  rb_define_module_function(mWc, "svn_wc_walk_entries", _wrap_svn_wc_walk_entries, -1);
  rb_define_module_function(mWc, "svn_wc_mark_missing_deleted", _wrap_svn_wc_mark_missing_deleted, -1);
  rb_define_module_function(mWc, "svn_wc_ensure_adm4", _wrap_svn_wc_ensure_adm4, -1);
  rb_define_module_function(mWc, "svn_wc_ensure_adm3", _wrap_svn_wc_ensure_adm3, -1);
  rb_define_module_function(mWc, "svn_wc_ensure_adm2", _wrap_svn_wc_ensure_adm2, -1);
  rb_define_module_function(mWc, "svn_wc_ensure_adm", _wrap_svn_wc_ensure_adm, -1);
  rb_define_module_function(mWc, "svn_wc_maybe_set_repos_root", _wrap_svn_wc_maybe_set_repos_root, -1);
  rb_define_const(mWc, "Svn_wc_status_none", SWIG_From_int((int)(svn_wc_status_none)));
  rb_define_const(mWc, "Svn_wc_status_unversioned", SWIG_From_int((int)(svn_wc_status_unversioned)));
  rb_define_const(mWc, "Svn_wc_status_normal", SWIG_From_int((int)(svn_wc_status_normal)));
  rb_define_const(mWc, "Svn_wc_status_added", SWIG_From_int((int)(svn_wc_status_added)));
  rb_define_const(mWc, "Svn_wc_status_missing", SWIG_From_int((int)(svn_wc_status_missing)));
  rb_define_const(mWc, "Svn_wc_status_deleted", SWIG_From_int((int)(svn_wc_status_deleted)));
  rb_define_const(mWc, "Svn_wc_status_replaced", SWIG_From_int((int)(svn_wc_status_replaced)));
  rb_define_const(mWc, "Svn_wc_status_modified", SWIG_From_int((int)(svn_wc_status_modified)));
  rb_define_const(mWc, "Svn_wc_status_merged", SWIG_From_int((int)(svn_wc_status_merged)));
  rb_define_const(mWc, "Svn_wc_status_conflicted", SWIG_From_int((int)(svn_wc_status_conflicted)));
  rb_define_const(mWc, "Svn_wc_status_ignored", SWIG_From_int((int)(svn_wc_status_ignored)));
  rb_define_const(mWc, "Svn_wc_status_obstructed", SWIG_From_int((int)(svn_wc_status_obstructed)));
  rb_define_const(mWc, "Svn_wc_status_external", SWIG_From_int((int)(svn_wc_status_external)));
  rb_define_const(mWc, "Svn_wc_status_incomplete", SWIG_From_int((int)(svn_wc_status_incomplete)));
  
  SwigClassSvn_wc_status3_t.klass = rb_define_class_under(mWc, "Svn_wc_status3_t", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_svn_wc_status3_t, (void *) &SwigClassSvn_wc_status3_t);
  rb_define_alloc_func(SwigClassSvn_wc_status3_t.klass, _wrap_svn_wc_status3_t_allocate);
  rb_define_method(SwigClassSvn_wc_status3_t.klass, "initialize", _wrap_new_svn_wc_status3_t, -1);
  rb_define_method(SwigClassSvn_wc_status3_t.klass, "kind=", _wrap_svn_wc_status3_t_kind_set, -1);
  rb_define_method(SwigClassSvn_wc_status3_t.klass, "kind", _wrap_svn_wc_status3_t_kind_get, -1);
  rb_define_method(SwigClassSvn_wc_status3_t.klass, "depth=", _wrap_svn_wc_status3_t_depth_set, -1);
  rb_define_method(SwigClassSvn_wc_status3_t.klass, "depth", _wrap_svn_wc_status3_t_depth_get, -1);
  rb_define_method(SwigClassSvn_wc_status3_t.klass, "filesize=", _wrap_svn_wc_status3_t_filesize_set, -1);
  rb_define_method(SwigClassSvn_wc_status3_t.klass, "filesize", _wrap_svn_wc_status3_t_filesize_get, -1);
  rb_define_method(SwigClassSvn_wc_status3_t.klass, "versioned=", _wrap_svn_wc_status3_t_versioned_set, -1);
  rb_define_method(SwigClassSvn_wc_status3_t.klass, "versioned", _wrap_svn_wc_status3_t_versioned_get, -1);
  rb_define_method(SwigClassSvn_wc_status3_t.klass, "conflicted=", _wrap_svn_wc_status3_t_conflicted_set, -1);
  rb_define_method(SwigClassSvn_wc_status3_t.klass, "conflicted", _wrap_svn_wc_status3_t_conflicted_get, -1);
  rb_define_method(SwigClassSvn_wc_status3_t.klass, "node_status=", _wrap_svn_wc_status3_t_node_status_set, -1);
  rb_define_method(SwigClassSvn_wc_status3_t.klass, "node_status", _wrap_svn_wc_status3_t_node_status_get, -1);
  rb_define_method(SwigClassSvn_wc_status3_t.klass, "text_status=", _wrap_svn_wc_status3_t_text_status_set, -1);
  rb_define_method(SwigClassSvn_wc_status3_t.klass, "text_status", _wrap_svn_wc_status3_t_text_status_get, -1);
  rb_define_method(SwigClassSvn_wc_status3_t.klass, "prop_status=", _wrap_svn_wc_status3_t_prop_status_set, -1);
  rb_define_method(SwigClassSvn_wc_status3_t.klass, "prop_status", _wrap_svn_wc_status3_t_prop_status_get, -1);
  rb_define_method(SwigClassSvn_wc_status3_t.klass, "copied=", _wrap_svn_wc_status3_t_copied_set, -1);
  rb_define_method(SwigClassSvn_wc_status3_t.klass, "copied", _wrap_svn_wc_status3_t_copied_get, -1);
  rb_define_method(SwigClassSvn_wc_status3_t.klass, "revision=", _wrap_svn_wc_status3_t_revision_set, -1);
  rb_define_method(SwigClassSvn_wc_status3_t.klass, "revision", _wrap_svn_wc_status3_t_revision_get, -1);
  rb_define_method(SwigClassSvn_wc_status3_t.klass, "changed_rev=", _wrap_svn_wc_status3_t_changed_rev_set, -1);
  rb_define_method(SwigClassSvn_wc_status3_t.klass, "changed_rev", _wrap_svn_wc_status3_t_changed_rev_get, -1);
  rb_define_method(SwigClassSvn_wc_status3_t.klass, "changed_date=", _wrap_svn_wc_status3_t_changed_date_set, -1);
  rb_define_method(SwigClassSvn_wc_status3_t.klass, "changed_date", _wrap_svn_wc_status3_t_changed_date_get, -1);
  rb_define_method(SwigClassSvn_wc_status3_t.klass, "changed_author=", _wrap_svn_wc_status3_t_changed_author_set, -1);
  rb_define_method(SwigClassSvn_wc_status3_t.klass, "changed_author", _wrap_svn_wc_status3_t_changed_author_get, -1);
  rb_define_method(SwigClassSvn_wc_status3_t.klass, "repos_root_url=", _wrap_svn_wc_status3_t_repos_root_url_set, -1);
  rb_define_method(SwigClassSvn_wc_status3_t.klass, "repos_root_url", _wrap_svn_wc_status3_t_repos_root_url_get, -1);
  rb_define_method(SwigClassSvn_wc_status3_t.klass, "repos_uuid=", _wrap_svn_wc_status3_t_repos_uuid_set, -1);
  rb_define_method(SwigClassSvn_wc_status3_t.klass, "repos_uuid", _wrap_svn_wc_status3_t_repos_uuid_get, -1);
  rb_define_method(SwigClassSvn_wc_status3_t.klass, "repos_relpath=", _wrap_svn_wc_status3_t_repos_relpath_set, -1);
  rb_define_method(SwigClassSvn_wc_status3_t.klass, "repos_relpath", _wrap_svn_wc_status3_t_repos_relpath_get, -1);
  rb_define_method(SwigClassSvn_wc_status3_t.klass, "switched=", _wrap_svn_wc_status3_t_switched_set, -1);
  rb_define_method(SwigClassSvn_wc_status3_t.klass, "switched", _wrap_svn_wc_status3_t_switched_get, -1);
  rb_define_method(SwigClassSvn_wc_status3_t.klass, "locked=", _wrap_svn_wc_status3_t_locked_set, -1);
  rb_define_method(SwigClassSvn_wc_status3_t.klass, "locked", _wrap_svn_wc_status3_t_locked_get, -1);
  rb_define_method(SwigClassSvn_wc_status3_t.klass, "lock=", _wrap_svn_wc_status3_t_lock_set, -1);
  rb_define_method(SwigClassSvn_wc_status3_t.klass, "lock", _wrap_svn_wc_status3_t_lock_get, -1);
  rb_define_method(SwigClassSvn_wc_status3_t.klass, "changelist=", _wrap_svn_wc_status3_t_changelist_set, -1);
  rb_define_method(SwigClassSvn_wc_status3_t.klass, "changelist", _wrap_svn_wc_status3_t_changelist_get, -1);
  rb_define_method(SwigClassSvn_wc_status3_t.klass, "ood_kind=", _wrap_svn_wc_status3_t_ood_kind_set, -1);
  rb_define_method(SwigClassSvn_wc_status3_t.klass, "ood_kind", _wrap_svn_wc_status3_t_ood_kind_get, -1);
  rb_define_method(SwigClassSvn_wc_status3_t.klass, "repos_node_status=", _wrap_svn_wc_status3_t_repos_node_status_set, -1);
  rb_define_method(SwigClassSvn_wc_status3_t.klass, "repos_node_status", _wrap_svn_wc_status3_t_repos_node_status_get, -1);
  rb_define_method(SwigClassSvn_wc_status3_t.klass, "repos_text_status=", _wrap_svn_wc_status3_t_repos_text_status_set, -1);
  rb_define_method(SwigClassSvn_wc_status3_t.klass, "repos_text_status", _wrap_svn_wc_status3_t_repos_text_status_get, -1);
  rb_define_method(SwigClassSvn_wc_status3_t.klass, "repos_prop_status=", _wrap_svn_wc_status3_t_repos_prop_status_set, -1);
  rb_define_method(SwigClassSvn_wc_status3_t.klass, "repos_prop_status", _wrap_svn_wc_status3_t_repos_prop_status_get, -1);
  rb_define_method(SwigClassSvn_wc_status3_t.klass, "repos_lock=", _wrap_svn_wc_status3_t_repos_lock_set, -1);
  rb_define_method(SwigClassSvn_wc_status3_t.klass, "repos_lock", _wrap_svn_wc_status3_t_repos_lock_get, -1);
  rb_define_method(SwigClassSvn_wc_status3_t.klass, "ood_changed_rev=", _wrap_svn_wc_status3_t_ood_changed_rev_set, -1);
  rb_define_method(SwigClassSvn_wc_status3_t.klass, "ood_changed_rev", _wrap_svn_wc_status3_t_ood_changed_rev_get, -1);
  rb_define_method(SwigClassSvn_wc_status3_t.klass, "ood_changed_date=", _wrap_svn_wc_status3_t_ood_changed_date_set, -1);
  rb_define_method(SwigClassSvn_wc_status3_t.klass, "ood_changed_date", _wrap_svn_wc_status3_t_ood_changed_date_get, -1);
  rb_define_method(SwigClassSvn_wc_status3_t.klass, "ood_changed_author=", _wrap_svn_wc_status3_t_ood_changed_author_set, -1);
  rb_define_method(SwigClassSvn_wc_status3_t.klass, "ood_changed_author", _wrap_svn_wc_status3_t_ood_changed_author_get, -1);
  rb_define_method(SwigClassSvn_wc_status3_t.klass, "moved_from_abspath=", _wrap_svn_wc_status3_t_moved_from_abspath_set, -1);
  rb_define_method(SwigClassSvn_wc_status3_t.klass, "moved_from_abspath", _wrap_svn_wc_status3_t_moved_from_abspath_get, -1);
  rb_define_method(SwigClassSvn_wc_status3_t.klass, "moved_to_abspath=", _wrap_svn_wc_status3_t_moved_to_abspath_set, -1);
  rb_define_method(SwigClassSvn_wc_status3_t.klass, "moved_to_abspath", _wrap_svn_wc_status3_t_moved_to_abspath_get, -1);
  rb_define_method(SwigClassSvn_wc_status3_t.klass, "file_external=", _wrap_svn_wc_status3_t_file_external_set, -1);
  rb_define_method(SwigClassSvn_wc_status3_t.klass, "file_external", _wrap_svn_wc_status3_t_file_external_get, -1);
  rb_define_method(SwigClassSvn_wc_status3_t.klass, "actual_kind=", _wrap_svn_wc_status3_t_actual_kind_set, -1);
  rb_define_method(SwigClassSvn_wc_status3_t.klass, "actual_kind", _wrap_svn_wc_status3_t_actual_kind_get, -1);
  SwigClassSvn_wc_status3_t.mark = 0;
  SwigClassSvn_wc_status3_t.destroy = (void (*)(void *)) free_svn_wc_status3_t;
  SwigClassSvn_wc_status3_t.trackObjects = 0;
  
  SwigClassSvn_wc_status2_t.klass = rb_define_class_under(mWc, "Svn_wc_status2_t", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_svn_wc_status2_t, (void *) &SwigClassSvn_wc_status2_t);
  rb_define_alloc_func(SwigClassSvn_wc_status2_t.klass, _wrap_svn_wc_status2_t_allocate);
  rb_define_method(SwigClassSvn_wc_status2_t.klass, "initialize", _wrap_new_svn_wc_status2_t, -1);
  rb_define_method(SwigClassSvn_wc_status2_t.klass, "entry=", _wrap_svn_wc_status2_t_entry_set, -1);
  rb_define_method(SwigClassSvn_wc_status2_t.klass, "entry", _wrap_svn_wc_status2_t_entry_get, -1);
  rb_define_method(SwigClassSvn_wc_status2_t.klass, "text_status=", _wrap_svn_wc_status2_t_text_status_set, -1);
  rb_define_method(SwigClassSvn_wc_status2_t.klass, "text_status", _wrap_svn_wc_status2_t_text_status_get, -1);
  rb_define_method(SwigClassSvn_wc_status2_t.klass, "prop_status=", _wrap_svn_wc_status2_t_prop_status_set, -1);
  rb_define_method(SwigClassSvn_wc_status2_t.klass, "prop_status", _wrap_svn_wc_status2_t_prop_status_get, -1);
  rb_define_method(SwigClassSvn_wc_status2_t.klass, "locked=", _wrap_svn_wc_status2_t_locked_set, -1);
  rb_define_method(SwigClassSvn_wc_status2_t.klass, "locked", _wrap_svn_wc_status2_t_locked_get, -1);
  rb_define_method(SwigClassSvn_wc_status2_t.klass, "copied=", _wrap_svn_wc_status2_t_copied_set, -1);
  rb_define_method(SwigClassSvn_wc_status2_t.klass, "copied", _wrap_svn_wc_status2_t_copied_get, -1);
  rb_define_method(SwigClassSvn_wc_status2_t.klass, "switched=", _wrap_svn_wc_status2_t_switched_set, -1);
  rb_define_method(SwigClassSvn_wc_status2_t.klass, "switched", _wrap_svn_wc_status2_t_switched_get, -1);
  rb_define_method(SwigClassSvn_wc_status2_t.klass, "repos_text_status=", _wrap_svn_wc_status2_t_repos_text_status_set, -1);
  rb_define_method(SwigClassSvn_wc_status2_t.klass, "repos_text_status", _wrap_svn_wc_status2_t_repos_text_status_get, -1);
  rb_define_method(SwigClassSvn_wc_status2_t.klass, "repos_prop_status=", _wrap_svn_wc_status2_t_repos_prop_status_set, -1);
  rb_define_method(SwigClassSvn_wc_status2_t.klass, "repos_prop_status", _wrap_svn_wc_status2_t_repos_prop_status_get, -1);
  rb_define_method(SwigClassSvn_wc_status2_t.klass, "repos_lock=", _wrap_svn_wc_status2_t_repos_lock_set, -1);
  rb_define_method(SwigClassSvn_wc_status2_t.klass, "repos_lock", _wrap_svn_wc_status2_t_repos_lock_get, -1);
  rb_define_method(SwigClassSvn_wc_status2_t.klass, "url=", _wrap_svn_wc_status2_t_url_set, -1);
  rb_define_method(SwigClassSvn_wc_status2_t.klass, "url", _wrap_svn_wc_status2_t_url_get, -1);
  rb_define_method(SwigClassSvn_wc_status2_t.klass, "ood_last_cmt_rev=", _wrap_svn_wc_status2_t_ood_last_cmt_rev_set, -1);
  rb_define_method(SwigClassSvn_wc_status2_t.klass, "ood_last_cmt_rev", _wrap_svn_wc_status2_t_ood_last_cmt_rev_get, -1);
  rb_define_method(SwigClassSvn_wc_status2_t.klass, "ood_last_cmt_date=", _wrap_svn_wc_status2_t_ood_last_cmt_date_set, -1);
  rb_define_method(SwigClassSvn_wc_status2_t.klass, "ood_last_cmt_date", _wrap_svn_wc_status2_t_ood_last_cmt_date_get, -1);
  rb_define_method(SwigClassSvn_wc_status2_t.klass, "ood_kind=", _wrap_svn_wc_status2_t_ood_kind_set, -1);
  rb_define_method(SwigClassSvn_wc_status2_t.klass, "ood_kind", _wrap_svn_wc_status2_t_ood_kind_get, -1);
  rb_define_method(SwigClassSvn_wc_status2_t.klass, "ood_last_cmt_author=", _wrap_svn_wc_status2_t_ood_last_cmt_author_set, -1);
  rb_define_method(SwigClassSvn_wc_status2_t.klass, "ood_last_cmt_author", _wrap_svn_wc_status2_t_ood_last_cmt_author_get, -1);
  rb_define_method(SwigClassSvn_wc_status2_t.klass, "tree_conflict=", _wrap_svn_wc_status2_t_tree_conflict_set, -1);
  rb_define_method(SwigClassSvn_wc_status2_t.klass, "tree_conflict", _wrap_svn_wc_status2_t_tree_conflict_get, -1);
  rb_define_method(SwigClassSvn_wc_status2_t.klass, "file_external=", _wrap_svn_wc_status2_t_file_external_set, -1);
  rb_define_method(SwigClassSvn_wc_status2_t.klass, "file_external", _wrap_svn_wc_status2_t_file_external_get, -1);
  rb_define_method(SwigClassSvn_wc_status2_t.klass, "pristine_text_status=", _wrap_svn_wc_status2_t_pristine_text_status_set, -1);
  rb_define_method(SwigClassSvn_wc_status2_t.klass, "pristine_text_status", _wrap_svn_wc_status2_t_pristine_text_status_get, -1);
  rb_define_method(SwigClassSvn_wc_status2_t.klass, "pristine_prop_status=", _wrap_svn_wc_status2_t_pristine_prop_status_set, -1);
  rb_define_method(SwigClassSvn_wc_status2_t.klass, "pristine_prop_status", _wrap_svn_wc_status2_t_pristine_prop_status_get, -1);
  SwigClassSvn_wc_status2_t.mark = 0;
  SwigClassSvn_wc_status2_t.destroy = (void (*)(void *)) free_svn_wc_status2_t;
  SwigClassSvn_wc_status2_t.trackObjects = 0;
  
  SwigClassSvn_wc_status_t.klass = rb_define_class_under(mWc, "Svn_wc_status_t", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_svn_wc_status_t, (void *) &SwigClassSvn_wc_status_t);
  rb_define_alloc_func(SwigClassSvn_wc_status_t.klass, _wrap_svn_wc_status_t_allocate);
  rb_define_method(SwigClassSvn_wc_status_t.klass, "initialize", _wrap_new_svn_wc_status_t, -1);
  rb_define_method(SwigClassSvn_wc_status_t.klass, "entry=", _wrap_svn_wc_status_t_entry_set, -1);
  rb_define_method(SwigClassSvn_wc_status_t.klass, "entry", _wrap_svn_wc_status_t_entry_get, -1);
  rb_define_method(SwigClassSvn_wc_status_t.klass, "text_status=", _wrap_svn_wc_status_t_text_status_set, -1);
  rb_define_method(SwigClassSvn_wc_status_t.klass, "text_status", _wrap_svn_wc_status_t_text_status_get, -1);
  rb_define_method(SwigClassSvn_wc_status_t.klass, "prop_status=", _wrap_svn_wc_status_t_prop_status_set, -1);
  rb_define_method(SwigClassSvn_wc_status_t.klass, "prop_status", _wrap_svn_wc_status_t_prop_status_get, -1);
  rb_define_method(SwigClassSvn_wc_status_t.klass, "locked=", _wrap_svn_wc_status_t_locked_set, -1);
  rb_define_method(SwigClassSvn_wc_status_t.klass, "locked", _wrap_svn_wc_status_t_locked_get, -1);
  rb_define_method(SwigClassSvn_wc_status_t.klass, "copied=", _wrap_svn_wc_status_t_copied_set, -1);
  rb_define_method(SwigClassSvn_wc_status_t.klass, "copied", _wrap_svn_wc_status_t_copied_get, -1);
  rb_define_method(SwigClassSvn_wc_status_t.klass, "switched=", _wrap_svn_wc_status_t_switched_set, -1);
  rb_define_method(SwigClassSvn_wc_status_t.klass, "switched", _wrap_svn_wc_status_t_switched_get, -1);
  rb_define_method(SwigClassSvn_wc_status_t.klass, "repos_text_status=", _wrap_svn_wc_status_t_repos_text_status_set, -1);
  rb_define_method(SwigClassSvn_wc_status_t.klass, "repos_text_status", _wrap_svn_wc_status_t_repos_text_status_get, -1);
  rb_define_method(SwigClassSvn_wc_status_t.klass, "repos_prop_status=", _wrap_svn_wc_status_t_repos_prop_status_set, -1);
  rb_define_method(SwigClassSvn_wc_status_t.klass, "repos_prop_status", _wrap_svn_wc_status_t_repos_prop_status_get, -1);
  SwigClassSvn_wc_status_t.mark = 0;
  SwigClassSvn_wc_status_t.destroy = (void (*)(void *)) free_svn_wc_status_t;
  SwigClassSvn_wc_status_t.trackObjects = 0;
  rb_define_module_function(mWc, "svn_wc_dup_status3", _wrap_svn_wc_dup_status3, -1);
  rb_define_module_function(mWc, "svn_wc_dup_status2", _wrap_svn_wc_dup_status2, -1);
  rb_define_module_function(mWc, "svn_wc_dup_status", _wrap_svn_wc_dup_status, -1);
  rb_define_module_function(mWc, "svn_wc_status3", _wrap_svn_wc_status3, -1);
  rb_define_module_function(mWc, "svn_wc_status2", _wrap_svn_wc_status2, -1);
  rb_define_module_function(mWc, "svn_wc_status", _wrap_svn_wc_status, -1);
  rb_define_module_function(mWc, "svn_wc_walk_status", _wrap_svn_wc_walk_status, -1);
  rb_define_module_function(mWc, "svn_wc_get_status_editor5", _wrap_svn_wc_get_status_editor5, -1);
  rb_define_module_function(mWc, "svn_wc_get_status_editor4", _wrap_svn_wc_get_status_editor4, -1);
  rb_define_module_function(mWc, "svn_wc_get_status_editor3", _wrap_svn_wc_get_status_editor3, -1);
  rb_define_module_function(mWc, "svn_wc_get_status_editor2", _wrap_svn_wc_get_status_editor2, -1);
  rb_define_module_function(mWc, "svn_wc_get_status_editor", _wrap_svn_wc_get_status_editor, -1);
  rb_define_module_function(mWc, "svn_wc_status_set_repos_locks", _wrap_svn_wc_status_set_repos_locks, -1);
  rb_define_module_function(mWc, "svn_wc_copy3", _wrap_svn_wc_copy3, -1);
  rb_define_module_function(mWc, "svn_wc_copy2", _wrap_svn_wc_copy2, -1);
  rb_define_module_function(mWc, "svn_wc_copy", _wrap_svn_wc_copy, -1);
  rb_define_module_function(mWc, "svn_wc_move", _wrap_svn_wc_move, -1);
  rb_define_module_function(mWc, "svn_wc_delete4", _wrap_svn_wc_delete4, -1);
  rb_define_module_function(mWc, "svn_wc_delete3", _wrap_svn_wc_delete3, -1);
  rb_define_module_function(mWc, "svn_wc_delete2", _wrap_svn_wc_delete2, -1);
  rb_define_module_function(mWc, "svn_wc_delete", _wrap_svn_wc_delete, -1);
  rb_define_module_function(mWc, "svn_wc_add_from_disk3", _wrap_svn_wc_add_from_disk3, -1);
  rb_define_module_function(mWc, "svn_wc_add_from_disk2", _wrap_svn_wc_add_from_disk2, -1);
  rb_define_module_function(mWc, "svn_wc_add_from_disk", _wrap_svn_wc_add_from_disk, -1);
  rb_define_module_function(mWc, "svn_wc_add4", _wrap_svn_wc_add4, -1);
  rb_define_module_function(mWc, "svn_wc_add3", _wrap_svn_wc_add3, -1);
  rb_define_module_function(mWc, "svn_wc_add2", _wrap_svn_wc_add2, -1);
  rb_define_module_function(mWc, "svn_wc_add", _wrap_svn_wc_add, -1);
  rb_define_module_function(mWc, "svn_wc_add_repos_file4", _wrap_svn_wc_add_repos_file4, -1);
  rb_define_module_function(mWc, "svn_wc_add_repos_file3", _wrap_svn_wc_add_repos_file3, -1);
  rb_define_module_function(mWc, "svn_wc_add_repos_file2", _wrap_svn_wc_add_repos_file2, -1);
  rb_define_module_function(mWc, "svn_wc_add_repos_file", _wrap_svn_wc_add_repos_file, -1);
  rb_define_module_function(mWc, "svn_wc_remove_from_revision_control2", _wrap_svn_wc_remove_from_revision_control2, -1);
  rb_define_module_function(mWc, "svn_wc_remove_from_revision_control", _wrap_svn_wc_remove_from_revision_control, -1);
  rb_define_module_function(mWc, "svn_wc_resolved_conflict5", _wrap_svn_wc_resolved_conflict5, -1);
  rb_define_module_function(mWc, "svn_wc_resolved_conflict4", _wrap_svn_wc_resolved_conflict4, -1);
  rb_define_module_function(mWc, "svn_wc_resolved_conflict3", _wrap_svn_wc_resolved_conflict3, -1);
  rb_define_module_function(mWc, "svn_wc_resolved_conflict2", _wrap_svn_wc_resolved_conflict2, -1);
  rb_define_module_function(mWc, "svn_wc_resolved_conflict", _wrap_svn_wc_resolved_conflict, -1);
  rb_define_module_function(mWc, "svn_wc_queue_committed4", _wrap_svn_wc_queue_committed4, -1);
  rb_define_module_function(mWc, "svn_wc_queue_committed3", _wrap_svn_wc_queue_committed3, -1);
  rb_define_module_function(mWc, "svn_wc_queue_committed2", _wrap_svn_wc_queue_committed2, -1);
  rb_define_module_function(mWc, "svn_wc_queue_committed", _wrap_svn_wc_queue_committed, -1);
  rb_define_module_function(mWc, "svn_wc_process_committed_queue2", _wrap_svn_wc_process_committed_queue2, -1);
  rb_define_module_function(mWc, "svn_wc_process_committed_queue", _wrap_svn_wc_process_committed_queue, -1);
  rb_define_module_function(mWc, "svn_wc_process_committed4", _wrap_svn_wc_process_committed4, -1);
  rb_define_module_function(mWc, "svn_wc_process_committed3", _wrap_svn_wc_process_committed3, -1);
  rb_define_module_function(mWc, "svn_wc_process_committed2", _wrap_svn_wc_process_committed2, -1);
  rb_define_module_function(mWc, "svn_wc_process_committed", _wrap_svn_wc_process_committed, -1);
  rb_define_module_function(mWc, "svn_wc_crawl_revisions5", _wrap_svn_wc_crawl_revisions5, -1);
  rb_define_module_function(mWc, "svn_wc_crawl_revisions4", _wrap_svn_wc_crawl_revisions4, -1);
  rb_define_module_function(mWc, "svn_wc_crawl_revisions3", _wrap_svn_wc_crawl_revisions3, -1);
  rb_define_module_function(mWc, "svn_wc_crawl_revisions2", _wrap_svn_wc_crawl_revisions2, -1);
  rb_define_module_function(mWc, "svn_wc_crawl_revisions", _wrap_svn_wc_crawl_revisions, -1);
  rb_define_module_function(mWc, "svn_wc_check_root", _wrap_svn_wc_check_root, -1);
  rb_define_module_function(mWc, "svn_wc_is_wc_root2", _wrap_svn_wc_is_wc_root2, -1);
  rb_define_module_function(mWc, "svn_wc_is_wc_root", _wrap_svn_wc_is_wc_root, -1);
  rb_define_module_function(mWc, "svn_wc_get_actual_target2", _wrap_svn_wc_get_actual_target2, -1);
  rb_define_module_function(mWc, "svn_wc_get_actual_target", _wrap_svn_wc_get_actual_target, -1);
  rb_define_module_function(mWc, "svn_wc_get_update_editor4", _wrap_svn_wc_get_update_editor4, -1);
  rb_define_module_function(mWc, "svn_wc_get_update_editor3", _wrap_svn_wc_get_update_editor3, -1);
  rb_define_module_function(mWc, "svn_wc_get_update_editor2", _wrap_svn_wc_get_update_editor2, -1);
  rb_define_module_function(mWc, "svn_wc_get_update_editor", _wrap_svn_wc_get_update_editor, -1);
  rb_define_module_function(mWc, "svn_wc_get_switch_editor4", _wrap_svn_wc_get_switch_editor4, -1);
  rb_define_module_function(mWc, "svn_wc_get_switch_editor3", _wrap_svn_wc_get_switch_editor3, -1);
  rb_define_module_function(mWc, "svn_wc_get_switch_editor2", _wrap_svn_wc_get_switch_editor2, -1);
  rb_define_module_function(mWc, "svn_wc_get_switch_editor", _wrap_svn_wc_get_switch_editor, -1);
  rb_define_module_function(mWc, "svn_wc_prop_list2", _wrap_svn_wc_prop_list2, -1);
  rb_define_module_function(mWc, "svn_wc_prop_list", _wrap_svn_wc_prop_list, -1);
  rb_define_module_function(mWc, "svn_wc_get_pristine_props", _wrap_svn_wc_get_pristine_props, -1);
  rb_define_module_function(mWc, "svn_wc_prop_get2", _wrap_svn_wc_prop_get2, -1);
  rb_define_module_function(mWc, "svn_wc_prop_get", _wrap_svn_wc_prop_get, -1);
  rb_define_module_function(mWc, "svn_wc_prop_set4", _wrap_svn_wc_prop_set4, -1);
  rb_define_module_function(mWc, "svn_wc_prop_set3", _wrap_svn_wc_prop_set3, -1);
  rb_define_module_function(mWc, "svn_wc_prop_set2", _wrap_svn_wc_prop_set2, -1);
  rb_define_module_function(mWc, "svn_wc_prop_set", _wrap_svn_wc_prop_set, -1);
  rb_define_module_function(mWc, "svn_wc_is_normal_prop", _wrap_svn_wc_is_normal_prop, -1);
  rb_define_module_function(mWc, "svn_wc_is_wc_prop", _wrap_svn_wc_is_wc_prop, -1);
  rb_define_module_function(mWc, "svn_wc_is_entry_prop", _wrap_svn_wc_is_entry_prop, -1);
  rb_define_module_function(mWc, "svn_wc_canonicalize_svn_prop", _wrap_svn_wc_canonicalize_svn_prop, -1);
  rb_define_module_function(mWc, "svn_wc_get_diff_editor6", _wrap_svn_wc_get_diff_editor6, -1);
  rb_define_module_function(mWc, "svn_wc_get_diff_editor5", _wrap_svn_wc_get_diff_editor5, -1);
  rb_define_module_function(mWc, "svn_wc_get_diff_editor4", _wrap_svn_wc_get_diff_editor4, -1);
  rb_define_module_function(mWc, "svn_wc_get_diff_editor3", _wrap_svn_wc_get_diff_editor3, -1);
  rb_define_module_function(mWc, "svn_wc_get_diff_editor2", _wrap_svn_wc_get_diff_editor2, -1);
  rb_define_module_function(mWc, "svn_wc_get_diff_editor", _wrap_svn_wc_get_diff_editor, -1);
  rb_define_module_function(mWc, "svn_wc_diff6", _wrap_svn_wc_diff6, -1);
  rb_define_module_function(mWc, "svn_wc_diff5", _wrap_svn_wc_diff5, -1);
  rb_define_module_function(mWc, "svn_wc_diff4", _wrap_svn_wc_diff4, -1);
  rb_define_module_function(mWc, "svn_wc_diff3", _wrap_svn_wc_diff3, -1);
  rb_define_module_function(mWc, "svn_wc_diff2", _wrap_svn_wc_diff2, -1);
  rb_define_module_function(mWc, "svn_wc_diff", _wrap_svn_wc_diff, -1);
  rb_define_module_function(mWc, "svn_wc_get_prop_diffs2", _wrap_svn_wc_get_prop_diffs2, -1);
  rb_define_module_function(mWc, "svn_wc_get_prop_diffs", _wrap_svn_wc_get_prop_diffs, -1);
  rb_define_const(mWc, "Svn_wc_merge_unchanged", SWIG_From_int((int)(svn_wc_merge_unchanged)));
  rb_define_const(mWc, "Svn_wc_merge_merged", SWIG_From_int((int)(svn_wc_merge_merged)));
  rb_define_const(mWc, "Svn_wc_merge_conflict", SWIG_From_int((int)(svn_wc_merge_conflict)));
  rb_define_const(mWc, "Svn_wc_merge_no_merge", SWIG_From_int((int)(svn_wc_merge_no_merge)));
  rb_define_module_function(mWc, "svn_wc_merge5", _wrap_svn_wc_merge5, -1);
  rb_define_module_function(mWc, "svn_wc_merge4", _wrap_svn_wc_merge4, -1);
  rb_define_module_function(mWc, "svn_wc_merge3", _wrap_svn_wc_merge3, -1);
  rb_define_module_function(mWc, "svn_wc_merge2", _wrap_svn_wc_merge2, -1);
  rb_define_module_function(mWc, "svn_wc_merge", _wrap_svn_wc_merge, -1);
  rb_define_module_function(mWc, "svn_wc_merge_props3", _wrap_svn_wc_merge_props3, -1);
  rb_define_module_function(mWc, "svn_wc_merge_props2", _wrap_svn_wc_merge_props2, -1);
  rb_define_module_function(mWc, "svn_wc_merge_props", _wrap_svn_wc_merge_props, -1);
  rb_define_module_function(mWc, "svn_wc_merge_prop_diffs", _wrap_svn_wc_merge_prop_diffs, -1);
  rb_define_module_function(mWc, "svn_wc_get_pristine_contents2", _wrap_svn_wc_get_pristine_contents2, -1);
  rb_define_module_function(mWc, "svn_wc_get_pristine_contents", _wrap_svn_wc_get_pristine_contents, -1);
  rb_define_module_function(mWc, "svn_wc_get_pristine_copy_path", _wrap_svn_wc_get_pristine_copy_path, -1);
  rb_define_module_function(mWc, "svn_wc_cleanup4", _wrap_svn_wc_cleanup4, -1);
  rb_define_module_function(mWc, "svn_wc_cleanup3", _wrap_svn_wc_cleanup3, -1);
  rb_define_module_function(mWc, "svn_wc_cleanup2", _wrap_svn_wc_cleanup2, -1);
  rb_define_module_function(mWc, "svn_wc_cleanup", _wrap_svn_wc_cleanup, -1);
  rb_define_module_function(mWc, "svn_wc_upgrade", _wrap_svn_wc_upgrade, -1);
  rb_define_module_function(mWc, "svn_wc_relocate4", _wrap_svn_wc_relocate4, -1);
  rb_define_module_function(mWc, "svn_wc_relocate3", _wrap_svn_wc_relocate3, -1);
  rb_define_module_function(mWc, "svn_wc_relocate2", _wrap_svn_wc_relocate2, -1);
  rb_define_module_function(mWc, "svn_wc_relocate", _wrap_svn_wc_relocate, -1);
  rb_define_module_function(mWc, "svn_wc_revert6", _wrap_svn_wc_revert6, -1);
  rb_define_module_function(mWc, "svn_wc_revert5", _wrap_svn_wc_revert5, -1);
  rb_define_module_function(mWc, "svn_wc_revert4", _wrap_svn_wc_revert4, -1);
  rb_define_module_function(mWc, "svn_wc_revert3", _wrap_svn_wc_revert3, -1);
  rb_define_module_function(mWc, "svn_wc_revert2", _wrap_svn_wc_revert2, -1);
  rb_define_module_function(mWc, "svn_wc_revert", _wrap_svn_wc_revert, -1);
  rb_define_module_function(mWc, "svn_wc_restore", _wrap_svn_wc_restore, -1);
  rb_define_module_function(mWc, "svn_wc_create_tmp_file2", _wrap_svn_wc_create_tmp_file2, -1);
  rb_define_module_function(mWc, "svn_wc_create_tmp_file", _wrap_svn_wc_create_tmp_file, -1);
  rb_define_module_function(mWc, "svn_wc_translated_file2", _wrap_svn_wc_translated_file2, -1);
  rb_define_module_function(mWc, "svn_wc_translated_file", _wrap_svn_wc_translated_file, -1);
  rb_define_module_function(mWc, "svn_wc_translated_stream", _wrap_svn_wc_translated_stream, -1);
  rb_define_module_function(mWc, "svn_wc_transmit_text_deltas3", _wrap_svn_wc_transmit_text_deltas3, -1);
  rb_define_module_function(mWc, "svn_wc_transmit_text_deltas2", _wrap_svn_wc_transmit_text_deltas2, -1);
  rb_define_module_function(mWc, "svn_wc_transmit_text_deltas", _wrap_svn_wc_transmit_text_deltas, -1);
  rb_define_module_function(mWc, "svn_wc_transmit_prop_deltas2", _wrap_svn_wc_transmit_prop_deltas2, -1);
  rb_define_module_function(mWc, "svn_wc_transmit_prop_deltas", _wrap_svn_wc_transmit_prop_deltas, -1);
  rb_define_module_function(mWc, "svn_wc_get_default_ignores", _wrap_svn_wc_get_default_ignores, -1);
  rb_define_module_function(mWc, "svn_wc_get_ignores2", _wrap_svn_wc_get_ignores2, -1);
  rb_define_module_function(mWc, "svn_wc_get_ignores", _wrap_svn_wc_get_ignores, -1);
  rb_define_module_function(mWc, "svn_wc_match_ignore_list", _wrap_svn_wc_match_ignore_list, -1);
  rb_define_module_function(mWc, "svn_wc_add_lock2", _wrap_svn_wc_add_lock2, -1);
  rb_define_module_function(mWc, "svn_wc_add_lock", _wrap_svn_wc_add_lock, -1);
  rb_define_module_function(mWc, "svn_wc_remove_lock2", _wrap_svn_wc_remove_lock2, -1);
  rb_define_module_function(mWc, "svn_wc_remove_lock", _wrap_svn_wc_remove_lock, -1);
  
  SwigClassSvn_wc_revision_status_t.klass = rb_define_class_under(mWc, "Svn_wc_revision_status_t", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_svn_wc_revision_status_t, (void *) &SwigClassSvn_wc_revision_status_t);
  rb_define_alloc_func(SwigClassSvn_wc_revision_status_t.klass, _wrap_svn_wc_revision_status_t_allocate);
  rb_define_method(SwigClassSvn_wc_revision_status_t.klass, "initialize", _wrap_new_svn_wc_revision_status_t, -1);
  rb_define_method(SwigClassSvn_wc_revision_status_t.klass, "min_rev=", _wrap_svn_wc_revision_status_t_min_rev_set, -1);
  rb_define_method(SwigClassSvn_wc_revision_status_t.klass, "min_rev", _wrap_svn_wc_revision_status_t_min_rev_get, -1);
  rb_define_method(SwigClassSvn_wc_revision_status_t.klass, "max_rev=", _wrap_svn_wc_revision_status_t_max_rev_set, -1);
  rb_define_method(SwigClassSvn_wc_revision_status_t.klass, "max_rev", _wrap_svn_wc_revision_status_t_max_rev_get, -1);
  rb_define_method(SwigClassSvn_wc_revision_status_t.klass, "switched=", _wrap_svn_wc_revision_status_t_switched_set, -1);
  rb_define_method(SwigClassSvn_wc_revision_status_t.klass, "switched", _wrap_svn_wc_revision_status_t_switched_get, -1);
  rb_define_method(SwigClassSvn_wc_revision_status_t.klass, "modified=", _wrap_svn_wc_revision_status_t_modified_set, -1);
  rb_define_method(SwigClassSvn_wc_revision_status_t.klass, "modified", _wrap_svn_wc_revision_status_t_modified_get, -1);
  rb_define_method(SwigClassSvn_wc_revision_status_t.klass, "sparse_checkout=", _wrap_svn_wc_revision_status_t_sparse_checkout_set, -1);
  rb_define_method(SwigClassSvn_wc_revision_status_t.klass, "sparse_checkout", _wrap_svn_wc_revision_status_t_sparse_checkout_get, -1);
  SwigClassSvn_wc_revision_status_t.mark = 0;
  SwigClassSvn_wc_revision_status_t.destroy = (void (*)(void *)) free_svn_wc_revision_status_t;
  SwigClassSvn_wc_revision_status_t.trackObjects = 0;
  rb_define_module_function(mWc, "svn_wc_revision_status2", _wrap_svn_wc_revision_status2, -1);
  rb_define_module_function(mWc, "svn_wc_set_changelist2", _wrap_svn_wc_set_changelist2, -1);
  rb_define_module_function(mWc, "svn_wc_set_changelist", _wrap_svn_wc_set_changelist, -1);
  rb_define_module_function(mWc, "svn_wc_get_changelists", _wrap_svn_wc_get_changelists, -1);
  rb_define_module_function(mWc, "svn_wc_crop_tree2", _wrap_svn_wc_crop_tree2, -1);
  rb_define_module_function(mWc, "svn_wc_crop_tree", _wrap_svn_wc_crop_tree, -1);
  rb_define_module_function(mWc, "svn_wc_exclude", _wrap_svn_wc_exclude, -1);
  rb_define_module_function(mWc, "svn_wc_read_kind2", _wrap_svn_wc_read_kind2, -1);
  rb_define_module_function(mWc, "svn_wc_read_kind", _wrap_svn_wc_read_kind, -1);
  rb_define_module_function(mWc, "svn_wc_diff_callbacks4_invoke_file_opened", _wrap_svn_wc_diff_callbacks4_invoke_file_opened, -1);
  rb_define_module_function(mWc, "svn_wc_diff_callbacks4_invoke_file_changed", _wrap_svn_wc_diff_callbacks4_invoke_file_changed, -1);
  rb_define_module_function(mWc, "svn_wc_diff_callbacks4_invoke_file_added", _wrap_svn_wc_diff_callbacks4_invoke_file_added, -1);
  rb_define_module_function(mWc, "svn_wc_diff_callbacks4_invoke_file_deleted", _wrap_svn_wc_diff_callbacks4_invoke_file_deleted, -1);
  rb_define_module_function(mWc, "svn_wc_diff_callbacks4_invoke_dir_deleted", _wrap_svn_wc_diff_callbacks4_invoke_dir_deleted, -1);
  rb_define_module_function(mWc, "svn_wc_diff_callbacks4_invoke_dir_opened", _wrap_svn_wc_diff_callbacks4_invoke_dir_opened, -1);
  rb_define_module_function(mWc, "svn_wc_diff_callbacks4_invoke_dir_added", _wrap_svn_wc_diff_callbacks4_invoke_dir_added, -1);
  rb_define_module_function(mWc, "svn_wc_diff_callbacks4_invoke_dir_props_changed", _wrap_svn_wc_diff_callbacks4_invoke_dir_props_changed, -1);
  rb_define_module_function(mWc, "svn_wc_diff_callbacks4_invoke_dir_closed", _wrap_svn_wc_diff_callbacks4_invoke_dir_closed, -1);
  rb_define_module_function(mWc, "svn_wc_diff_callbacks3_invoke_file_changed", _wrap_svn_wc_diff_callbacks3_invoke_file_changed, -1);
  rb_define_module_function(mWc, "svn_wc_diff_callbacks3_invoke_file_added", _wrap_svn_wc_diff_callbacks3_invoke_file_added, -1);
  rb_define_module_function(mWc, "svn_wc_diff_callbacks3_invoke_file_deleted", _wrap_svn_wc_diff_callbacks3_invoke_file_deleted, -1);
  rb_define_module_function(mWc, "svn_wc_diff_callbacks3_invoke_dir_added", _wrap_svn_wc_diff_callbacks3_invoke_dir_added, -1);
  rb_define_module_function(mWc, "svn_wc_diff_callbacks3_invoke_dir_deleted", _wrap_svn_wc_diff_callbacks3_invoke_dir_deleted, -1);
  rb_define_module_function(mWc, "svn_wc_diff_callbacks3_invoke_dir_props_changed", _wrap_svn_wc_diff_callbacks3_invoke_dir_props_changed, -1);
  rb_define_module_function(mWc, "svn_wc_diff_callbacks3_invoke_dir_opened", _wrap_svn_wc_diff_callbacks3_invoke_dir_opened, -1);
  rb_define_module_function(mWc, "svn_wc_diff_callbacks3_invoke_dir_closed", _wrap_svn_wc_diff_callbacks3_invoke_dir_closed, -1);
  rb_define_module_function(mWc, "svn_wc_diff_callbacks2_invoke_file_changed", _wrap_svn_wc_diff_callbacks2_invoke_file_changed, -1);
  rb_define_module_function(mWc, "svn_wc_diff_callbacks2_invoke_file_added", _wrap_svn_wc_diff_callbacks2_invoke_file_added, -1);
  rb_define_module_function(mWc, "svn_wc_diff_callbacks2_invoke_file_deleted", _wrap_svn_wc_diff_callbacks2_invoke_file_deleted, -1);
  rb_define_module_function(mWc, "svn_wc_diff_callbacks2_invoke_dir_added", _wrap_svn_wc_diff_callbacks2_invoke_dir_added, -1);
  rb_define_module_function(mWc, "svn_wc_diff_callbacks2_invoke_dir_deleted", _wrap_svn_wc_diff_callbacks2_invoke_dir_deleted, -1);
  rb_define_module_function(mWc, "svn_wc_diff_callbacks2_invoke_dir_props_changed", _wrap_svn_wc_diff_callbacks2_invoke_dir_props_changed, -1);
  rb_define_module_function(mWc, "svn_wc_diff_callbacks_invoke_file_changed", _wrap_svn_wc_diff_callbacks_invoke_file_changed, -1);
  rb_define_module_function(mWc, "svn_wc_diff_callbacks_invoke_file_added", _wrap_svn_wc_diff_callbacks_invoke_file_added, -1);
  rb_define_module_function(mWc, "svn_wc_diff_callbacks_invoke_file_deleted", _wrap_svn_wc_diff_callbacks_invoke_file_deleted, -1);
  rb_define_module_function(mWc, "svn_wc_diff_callbacks_invoke_dir_added", _wrap_svn_wc_diff_callbacks_invoke_dir_added, -1);
  rb_define_module_function(mWc, "svn_wc_diff_callbacks_invoke_dir_deleted", _wrap_svn_wc_diff_callbacks_invoke_dir_deleted, -1);
  rb_define_module_function(mWc, "svn_wc_diff_callbacks_invoke_props_changed", _wrap_svn_wc_diff_callbacks_invoke_props_changed, -1);
  rb_define_module_function(mWc, "svn_wc_entry_callbacks2_invoke_found_entry", _wrap_svn_wc_entry_callbacks2_invoke_found_entry, -1);
  rb_define_module_function(mWc, "svn_wc_entry_callbacks2_invoke_handle_error", _wrap_svn_wc_entry_callbacks2_invoke_handle_error, -1);
  rb_define_module_function(mWc, "svn_wc_entry_callbacks_invoke_found_entry", _wrap_svn_wc_entry_callbacks_invoke_found_entry, -1);
  rb_define_module_function(mWc, "svn_wc_invoke_external_update", _wrap_svn_wc_invoke_external_update, -1);
  rb_define_module_function(mWc, "svn_wc_invoke_notify_func2", _wrap_svn_wc_invoke_notify_func2, -1);
  rb_define_module_function(mWc, "svn_wc_invoke_notify_func", _wrap_svn_wc_invoke_notify_func, -1);
  rb_define_module_function(mWc, "svn_wc_invoke_conflict_resolver_func2", _wrap_svn_wc_invoke_conflict_resolver_func2, -1);
  rb_define_module_function(mWc, "svn_wc_invoke_conflict_resolver_func", _wrap_svn_wc_invoke_conflict_resolver_func, -1);
  rb_define_module_function(mWc, "svn_wc_invoke_status_func4", _wrap_svn_wc_invoke_status_func4, -1);
  rb_define_module_function(mWc, "svn_wc_invoke_status_func3", _wrap_svn_wc_invoke_status_func3, -1);
  rb_define_module_function(mWc, "svn_wc_invoke_status_func2", _wrap_svn_wc_invoke_status_func2, -1);
  rb_define_module_function(mWc, "svn_wc_invoke_status_func", _wrap_svn_wc_invoke_status_func, -1);
  rb_define_module_function(mWc, "svn_wc_invoke_get_file", _wrap_svn_wc_invoke_get_file, -1);
  rb_define_module_function(mWc, "svn_wc_invoke_dirents_func", _wrap_svn_wc_invoke_dirents_func, -1);
  rb_define_module_function(mWc, "svn_wc_invoke_canonicalize_svn_prop_get_file", _wrap_svn_wc_invoke_canonicalize_svn_prop_get_file, -1);
  rb_define_module_function(mWc, "svn_wc_invoke_upgrade_get_repos_info", _wrap_svn_wc_invoke_upgrade_get_repos_info, -1);
  rb_define_module_function(mWc, "svn_wc_invoke_relocation_validator3", _wrap_svn_wc_invoke_relocation_validator3, -1);
  rb_define_module_function(mWc, "svn_wc_invoke_relocation_validator2", _wrap_svn_wc_invoke_relocation_validator2, -1);
  rb_define_module_function(mWc, "svn_wc_invoke_relocation_validator", _wrap_svn_wc_invoke_relocation_validator, -1);
  rb_define_module_function(mWc, "svn_changelist_invoke_receiver", _wrap_svn_changelist_invoke_receiver, -1);
  rb_define_module_function(mWc, "svn_wc_swig_init_asp_dot_net_hack", _wrap_svn_wc_swig_init_asp_dot_net_hack, -1);
  
  SwigClassSvn_wc_committed_queue_t.klass = rb_define_class_under(mWc, "Svn_wc_committed_queue_t", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_svn_wc_committed_queue_t, (void *) &SwigClassSvn_wc_committed_queue_t);
  rb_define_alloc_func(SwigClassSvn_wc_committed_queue_t.klass, _wrap_svn_wc_committed_queue_t_allocate);
  rb_define_method(SwigClassSvn_wc_committed_queue_t.klass, "initialize", _wrap_new_svn_wc_committed_queue_t, -1);
  SwigClassSvn_wc_committed_queue_t.mark = 0;
  SwigClassSvn_wc_committed_queue_t.destroy = (void (*)(void *)) free_svn_wc_committed_queue_t;
  SwigClassSvn_wc_committed_queue_t.trackObjects = 0;
  
  SwigClassSvn_wc_traversal_info_t.klass = rb_define_class_under(mWc, "Svn_wc_traversal_info_t", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_svn_wc_traversal_info_t, (void *) &SwigClassSvn_wc_traversal_info_t);
  rb_define_alloc_func(SwigClassSvn_wc_traversal_info_t.klass, _wrap_svn_wc_traversal_info_t_allocate);
  rb_define_method(SwigClassSvn_wc_traversal_info_t.klass, "initialize", _wrap_new_svn_wc_traversal_info_t, -1);
  SwigClassSvn_wc_traversal_info_t.mark = 0;
  SwigClassSvn_wc_traversal_info_t.destroy = (void (*)(void *)) free_svn_wc_traversal_info_t;
  SwigClassSvn_wc_traversal_info_t.trackObjects = 0;
}

